#ifndef BASETYPE_HPP
#define BASETYPE_HPP
//-----------------------------------------------------------------------------
/// @file
///
/// Define compiler independent types that can be used to move binary data
/// between architectures and into binary persistent files.
///
/// @section fixedenum Fixed Size Enumerations
///
/// The template classes CFixedEnum8<enumtype> and CFixedEnum16<enumtype>
/// are lightweight wrapper classes around the enumeration enumtype. Each
/// is of fixed size (8 or 16 bits), and so should be used for data shared
/// between platforms, or persistant data.
///
/// The classes define an assignment operator from as well as a conversion
/// to the encapsulated enumeration. In this way assignments between types
/// of this class and the enumeration are transparent. (BorlandC 4.5 has a
/// minor bug where it generates a warning on assignment to the fundamental
/// enumeration. It does do the correct thing however.
///
/// No constructor is defined for this class. This is to allow the class
/// to be used inside unions. This does hoever mean that the class cannot
/// be constructed with an enumeration value. It has to be constructed
/// uninitialized, and then assigned an initial value. If this turns out to
/// be a problem, an expicit conversion function (template) can be made to
/// help this.
///
/// Since these classes are arguably more complicated and more difficult to
/// use it makes sence to limit their use only to those situatins where it
/// is required. This means that enumerations are used without regard to
/// their size everywhere except in structures or classes that are either
/// meant for communicating between platforms, or for persistant data. In
/// those classes the fixed sized type is also defined, and the member
/// variable is made of that fixed size type. For Example:
///
///     enum    ERiseType
///     {
///         ceRinseTypeFast,
///         ceRinceTypeSlow
///     };
///     ...
///     struct  SComAnlaysisOpt
///     {
///         typedef CFixedEnum8<ERiseType>  E8RinseType;
///         ...
///         E8RinseType     m_instep;
///     };
///
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#if !defined(MIC_DM643)
#include <cstdint>
#include <limits>
#endif

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
/// @name Fundamental Types
/// These are the typedefs that setup the compiler independent types for the
/// various sized integers.
//@{
#if !defined(MIC_DM643)
using uint_8 = std::uint8_t;
using int_8 = std::int8_t;

// Note that armlinux has different endian, so we have binary compatibility
// issues even for fixed types.
using uint_16 = std::uint16_t;
using int_16 = std::int16_t;
using uint_32 = std::uint32_t;
using int_32 = std::int32_t;
using uint_64 = std::uint64_t;
using int_64 = std::int64_t;
using float_32 = float;           ///< 32 bit floating point
using float_64 = double;          ///< 64 bit floating point

static_assert( sizeof(float_32) == 4, "Unexpected fixed width floating point size.");
static_assert( sizeof(float_64) == 8, "Unexpected fixed width floating point size.");
static_assert( std::numeric_limits<float>::is_iec559, "Unexpected floating point representation.");
static_assert( std::numeric_limits<double>::is_iec559, "Unexpected floating point representation.");

#else
typedef unsigned char       uint_8;         ///< 8 bit unsigned integer
typedef signed   char       int_8;          ///< 8 bit signed integer
typedef unsigned short int  uint_16;        ///< 16 bit unsigned integer
typedef signed   short int  int_16;         ///< 16 bit signed integer
typedef unsigned int        uint_32;
typedef signed   int        int_32;
typedef unsigned long       uint_40;
typedef signed   long       int_40;
typedef unsigned long long  uint_64;
typedef signed   long long  int_64;

typedef float               float_32;       ///< 32 bit floating point
typedef double              float_64;       ///< 64 bit floating point
#endif

//@}

// For MFC builds '_TCHAR' is used in place of 'char' and literal strings are
// wrapped in '_T( )'.  Use 'char' and bare literal strings for other builds.
#if !defined(MIC_VCMFC)
typedef char _TCHAR;
#if !defined(MIC_TEST) && !defined(MIC_DM643)
#define _T(s) (u8##s)
#else
#define _T(s) (s)
#endif
#endif


#if defined(__cplusplus)

// --------------------------------------------------------------------------
#if defined(__GNUC__)
#pragma pack(push, 1)
#endif

/// Adapter for 8 bit Enumeration.  See @ref fixedenum
template<class Tenum>
class   CFixedEnum8
{
public:

    CFixedEnum8<Tenum>& operator=( Tenum e )
    { m_iData = (int_8)e; return *this; }

    operator Tenum() const { return (Tenum)m_iData; }

    int_8   m_iData;
};

/// Adapter for 16 bit Enumeration.   See @ref fixedenum
template<class Tenum>
class   CFixedEnum16
{
public:

    CFixedEnum16<Tenum>& operator=( Tenum e )   { m_iData = (int_16)e; return *this; }

    operator Tenum() const { return (Tenum)m_iData; }
#if defined (MIC_VCMFC)
    uint_16& Data() { return *((uint_16*)&m_iData); }
    const uint_16& Data() const { return *((uint_16*)&m_iData); }
#endif

    int_16  m_iData;
};

/// Adapter for 32 bit Enumeration.   See @ref fixedenum
template<class Tenum>
class   CFixedEnum32
{
public:

    CFixedEnum32<Tenum>& operator=( Tenum e )
    { m_iData = (int_32)e; return *this; }

    operator Tenum() const { return (Tenum)m_iData; }
    operator CFixedEnum8<Tenum>() const
    {
        CFixedEnum8<Tenum> r;
        r = (Tenum)m_iData;
        return r;
    }
    operator CFixedEnum16<Tenum>() const
    {
        CFixedEnum16<Tenum> r;
        r = (Tenum)m_iData;
        return r;
    }

    int_32  m_iData;
};


// --------------------------------------------------------------------------

/// Adapter for 8 bit boolean.
///
/// Specialization for bool due to performance warning in the int_8
/// conversion.
template<>
class   CFixedEnum8<bool>
{
public:

    CFixedEnum8<bool>& operator=( bool e )  { m_iData = (int_8)e; return *this; }

    operator bool() const { return m_iData ? true : false; }

    int_8   m_iData;
};

/// Adapter for 16 bit boolean
///
/// Specialization for bool due to performance warning in the int_16
/// conversion.
template<>
class   CFixedEnum16<bool>
{
public:

    CFixedEnum16<bool>& operator=( bool e )
    { m_iData = (int_16)e; return *this; }

    operator bool() const { return m_iData ? true : false; }
#if defined (MIC_VCMFC)
    uint_16& Data() { return *((uint_16*)&m_iData); }
#endif

    int_16 m_iData;
};

/// Specific implementation of CFixedEnum32 with special boolean handling.
///
/// Specialization for bool due to performance warning in the int_32
/// conversion.
template<>
class   CFixedEnum32< bool >
{
public:

    CFixedEnum32<bool>& operator=( bool e ) { m_iData = (int_32)e; return *this; }

    operator bool() const { return m_iData ? true : false; }
    operator CFixedEnum8<bool>() const
    {
        CFixedEnum8<bool> r;
        r = m_iData ? true : false;
        return r;
    }
    operator CFixedEnum16<bool>() const
    {
        CFixedEnum16<bool> r;
        r = m_iData ? true : false;
        return r;
    }

    int_32  m_iData;
};
#if !defined(MIC_DM643)
// DM643 does not support static_assert.
// Note that the array checking also checks the class alignment for the template.
static_assert( sizeof(CFixedEnum8<int>[4]) == 4, "Unexpected fixed width type size" );
static_assert( sizeof(CFixedEnum16<int>[4]) == 8, "Unexpected fixed width type size" );
static_assert( sizeof(CFixedEnum32<int>[4]) == 16, "Unexpected fixed width type size" );
static_assert( sizeof(CFixedEnum8<bool>[4]) == 4, "Unexpected fixed width type size" );
static_assert( sizeof(CFixedEnum16<bool>[4]) == 8, "Unexpected fixed width type size" );
static_assert( sizeof(CFixedEnum32<bool>[4]) == 16, "Unexpected fixed width type size" );
#endif

#if defined(__GNUC__)
#pragma pack(pop)
#endif

// --------------------------------------------------------------------------
/// @name Type Defines for Fixed Size Booleans.
/// These are the typedefs that setup the compiler independent types for the
/// various sized booleans.
//@{
typedef CFixedEnum8< bool > bool_8;         ///< 8 bit boolean type.
typedef CFixedEnum16< bool > bool_16;       ///< 16 bit boolean type
typedef CFixedEnum32< bool > bool_32;       ///< 32 bit boolean type
//@}

// --------------------------------------------------------------------------
/// @name Fixed Size conversions
/// These are the conversion functions that allow conversions between different
/// sized representations of the enumerations.
//@{
template<class Tenum>
inline CFixedEnum32<Tenum> ToFixed( Tenum e ) { CFixedEnum32<Tenum> r; r = e; return r; }
template<class Tenum>
inline CFixedEnum32<Tenum> ToFixed( CFixedEnum32<Tenum> e ) { CFixedEnum32<Tenum> r; r = (Tenum)e; return r; }
template<class Tenum>
inline CFixedEnum32<Tenum> ToFixed( CFixedEnum16<Tenum> e ) { CFixedEnum32<Tenum> r; r = (Tenum)e; return r; }
template<class Tenum>
inline CFixedEnum32<Tenum> ToFixed( CFixedEnum8<Tenum> e ) { CFixedEnum32<Tenum> r; r = (Tenum)e; return r; }
template<>
inline CFixedEnum32<bool> ToFixed( bool e ) { CFixedEnum32<bool> r; r = e; return r; }
///@}

#endif  //  defined(__cplusplus__)

#endif // BASETYPE_HPP
