#ifndef   BASEUNIT_HPP
#define   BASEUNIT_HPP
//-----------------------------------------------------------------------------
/// @file
/// The classes for units management.  They the support user changes to the
/// selections, but not the UI elements.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "cstr.hpp"
#include "json-util.hpp"

#include <vector>
#include <memory>
#include <functional>
#include <iosfwd>
#include <string>

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class CUnitSelection;
using PCUnitSelection = std::shared_ptr<CUnitSelection>;
namespace NBaseUnit
{
    struct SAppChoiceJSON;
    using VAppChoiceJSON = std::vector<SAppChoiceJSON>;
}

// ================================================================================================
/// @returns a JSON string with a "linked" field that is an array of objects with fields "name" and
/// "choice"
std::string UpdateLinkedUnits(const std::string& cszSelection, int iChoice);
/// @returns the json schema associated with UpdateLinkedUnits.
std::string UpdateLinkedUnitsJsonSchema();

/// @ returns a JSON string with the choices defined in @p cvSelData.
std::string EditUnitSelectionJson(const NBaseUnit::VAppChoiceJSON& cvSelData);
/// @returns the json schema associated with EditUnitSelectionJson
std::string EditUnitSelectionJsonSchema();

// ============================================================================
/// Namespace for free functions and types used in baseunit file.
namespace NBaseUnit
{
    /// The lookup table for source choice to linked choice for a linked selections.
    using CSrcDestLink = std::map<int, int>;


    /// The map of linked selection to choice lookup table.
    using CLinkedSelections = std::map<PCUnitSelection, CSrcDestLink>;

    /// A map of CUnitSelections to choices.  This is used to indicate
    /// temporary selection choices for unit conversions.
    using CSelectionMap = std::map<PCUnitSelection, int>;

    /// The data that defines the liked selections.
    struct SUnitSelLink
    {
        PCUnitSelection pSrc;
        PCUnitSelection pDest;
        CSrcDestLink  aSrcToDestLink;
    };

    /// The data for a choice in the selection.
    struct SChoice
    {
        int iChoice;              ///< The selection value.
        uint_16  uiChoiceNameID;  ///< The string ID for the choice name.
    };

    /// Structure to define the choices that the application can present to the
    /// user for selection of what he wants to see.
    struct  SAppChoice
    {
        PCUnitSelection pcType;     ///< The unit that can be selected.
        uint_16 uiStartID;          ///< The app specific ID that links to the UI
        std::vector<int> aiUnits;   ///< The set of units that can be selected.
    };

    /// The run time ser of choices that the application uses.
    using VAppChoice = std::vector<SAppChoice>;

    /// Structure to define the choices that the application can present to the
    /// user for selection of what he wants to see.
    struct SAppChoiceJSON
    {
        PCUnitSelection pType;  ///< The unit that can be selected.
        std::string szTitle;    ///< The name to show the user for the selection.
        /// The set of units that can be selected and what should be shown to the user.
        std::vector<std::pair<int, int>> aUnitsChoiceLabelID;
    };

    /// The run time ser of choices that the application uses.
    using VAppChoiceJSON = std::vector<SAppChoiceJSON>;

    /// The callback function type for iterating over all of the units.
    using FnUnitTypeCallback = std::function<void (
        const CStr& szName,
        PCUnitSelection pUnitSel )>;

    ///  Defined the data for a single converstion from internal units.
    struct SUnitChoice
    {
        int iUnit;            ///< Which choice it is.
        uint_16 uiShortIDS;   ///< String ID of this choice (short form)
        double dScale, dOffs; ///< Conversion from this choice: [display] = [file] * dScale + dOffs
    };

    /// This encapsulates as single conversation tied to one selection.  It can
    /// also degenerate to a single non conversion unit string.
    struct SSimpleUnit;
    using PSimpleUnit = std::shared_ptr<SSimpleUnit>;

    /// Multiple units can be linked to the same unit, This sets up the
    /// links. Whenever the pSrc link changes, it will automatically update the
    /// choices in all the destinations.
    /// All the source selections must be in the dialog list, and none of the
    /// destination selections can be in the global dialog list.
    void SetupLinks(const std::vector<SUnitSelLink>& rcLinkedUnits);

    /// Set the units and choices allowed to be changed by the application.
    void SetDialogData( const VAppChoice& cvDlgData );

    /// Give access to the unit choice data.
    const VAppChoice& GetDialogData();

    /// A mechanism for iterating over each of the units that can be changed.  This
    /// is used at the application level to support persistence of user selected
    /// choices.
    void ForEachDo( const CStr& cszName, FnUnitTypeCallback pFcn );
};

namespace NUnitSelection
{
    /// @return the Json schema for the CUnits ToJson method.
    std::string JsonSchema();

    /// @return the Json schema for the CUnits GlobalToJson method.
    std::string GlobalToJsonSchema();

    /// @return the Json schema for the UpdateCurrentSelectionsJson() method.
    std::string CurrentSelectionJsonSchema();
}   // namespace NUnitSelection

namespace NUnits
{
    /// @return the Json schema for the CUnits ToJson method.
    std::string JsonSchema();

    /// @return the Json schema for the CUnits GlobalToJson method.
    std::string GlobalToJsonSchema();
}  // namespace NUnits

//-----------------------------------------------------------------------------
/// Encapsulates a single choice that the user can make on how to display units
/// in the application.  Not that the application can link some selections to a
/// primary choice by using the SetupLinks method.
///
/// The class is meant to support persistence at the application level.  To
/// that end, each selection has a name (CStr) and each choice has a string ID
/// (CStr::CStr(uiResID)) that can be used to save the state of selections.
class CUnitSelection : public std::enable_shared_from_this<CUnitSelection>
{
    /// @return the json value for @p rData ready to be inserted into @p rDoc.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CUnitSelection& rData);

public:
    using SChoice = NBaseUnit::SChoice;
    using CLinkedSelections = NBaseUnit::CLinkedSelections;
    using CSrcDestLink = NBaseUnit::CSrcDestLink;

    /// Create a CUnitSelection.
    ///
    /// \param szSelName the name of the selection (as it will appear in
    /// persistence)
    /// \param aiChoices the SChoice array vector that defines the choices for
    /// the selection.
    static PCUnitSelection Create(
        const CStr& szSelName, const std::vector<SChoice>& aiChoices);

    /// @return json value for all the known CUnitSelection objects.
    static rapidjson::Value GlobalToJson(rapidjson::Document::AllocatorType& rAlloc);

    /// Update the user selections based on the json object string passed in.
    static void UpdateCurrentSelectionsJson(const std::string& cszSelectionJson);

    CUnitSelection(const CUnitSelection&) = delete;
    CUnitSelection& operator=(const CUnitSelection&) = delete;

    /// Removes itself forom the global lists, and from any linked units.
    ~CUnitSelection();

    /// Give access to all the linked units to this one.
    const CLinkedSelections& GetLinkedSelections() const { return m_aLinkedSelections; }

    /// What is the choice for this selection.
    int GetCurrentChoice() const { return m_iChoice; }
    /// Is \p iChoice the current choice for this selection.
    bool IsCurrentChoice(int iChoice) const { return m_iChoice == iChoice; }
    /// Set the choice for this selection to \p iChoice
    void SetCurrentChoice(int iChoice);
    /// Set the choice to the name that matched \p szChoice.
    void SetChoiceName( CStr szChoice );
    /// Return the name of the current selection.
    CStr GetChoiceName() const;

    /// Give access to the name of this selection.
    CStr GetSelectionName() const { return m_szSelectionName; }

    /// Adds the linked units to this selection.
    void AddLinkedUnit( PCUnitSelection pDest,
                        const CSrcDestLink&  aSrcToDestLink );

private:
    /// The constructor is private so these can conly be made through shared
    /// pointers.  See the make_shared method.
    CUnitSelection( const CStr& szSelName, const std::vector<SChoice>& aiChoices);

    /// See if the \p iChoice is in our list.
    bool HasChoice( int iChoice ) const;

    /// The name of this selection.
    CStr m_szSelectionName;
    /// The currently made choice.
    int m_iChoice;
    /// The list of available choices.
    std::vector<SChoice> m_aiChoices;
    /// The list of selections that are linked to this.
    CLinkedSelections  m_aLinkedSelections;
};

//-----------------------------------------------------------------------------
/// CUnits encapsulate the units for a quantity.  They work with CUnitSelection
/// to change how quantities are displayed to the user.
///
/// The class provides an interface for the units to display, and conversion of
/// quantities between internal and display units.  This also provides a
/// mechanism for unit conversion by the using temporary choices for the
/// selections.
class CUnits final
{
    /// @return the json value for @p rData ready to be inserted into @p rDoc.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc, const CUnits& rData);

public:
    using SUnitChoice = NBaseUnit::SUnitChoice;
    using CSelectionMap = NBaseUnit::CSelectionMap;

    /// @return json value for all the known CUnitSelection objects.
    static rapidjson::Value GlobalToJson(rapidjson::Document::AllocatorType& rAlloc);

    /// Create the no units unit (No conversions, no units.)
    CUnits( );
    /// Copy Consruction
    CUnits( const CUnits& ) = default;
    /// No converstion, just a fixed units string.
    explicit CUnits(const char* szName, uint_16 uiShortIDS);
    /// Units linked to a selection with a list of conversions and units for
    /// each selection choice.
    explicit CUnits(const char* szName,
        PCUnitSelection pUnitSel,
        const std::vector<SUnitChoice>& cValues,
        bool fUseDialogTxt);

    /// Are the two CUnits objects the same.
    bool operator==( const CUnits &op ) const;
    /// Are the two CUnits objects different.
    bool operator!=( const CUnits &op ) const
        { return !(*this == op); }

    /// Is there any overlap in the core units between the two units.
    bool operator&( const CUnits &op ) const;

    /// Make a compound unit object by adding the units from op to this.
    CUnits operator|( const CUnits &op ) const;

    /// Add additional units to make this a compound unit object by adding the
    /// units from op.
    CUnits& operator|=( const CUnits &op );

    /// Should the units in this be shown in the Dialog context.
    bool UseTextInDialogs() const;

    ///  Return the current display text for the measurement type.  If multiple
    ///  types are included concatenate the strings in the reverse order in
    ///  which they were OR ed.
    ///
    ///  If the mask contains only a single entry, its text is always returned.
    ///  If there are multiple entries, the ones are concatenated together
    ///  which have the 'use in dialog' flag set.
    CStr DisplayText() const;

    /// Convert the value to the display units.
    double  DisplayValue(double val) const;

    /// Return the scaling factor for the units currently selected.
    double  DisplayScale() const;

    ///@{
    /// Allows conversion of the \p val to or from internal units. The \p
    /// cSelMap holds the temproary choices that are used to do the
    /// conversion.
    double ConvertFromInternal( double val, CSelectionMap cSelMap ) const;
    double ConvertToInternal( double val, CSelectionMap cSelMap ) const;
    ///@}

private:
    /// The set of CSimpleUnit options that are combined into the aggregate
    /// unit as a whole.
    std::vector<NBaseUnit::PSimpleUnit> m_vpUnits;
};

/// The special "no units" compound units object to be used as a sentinel, or
/// default / no value value.  (This is needed by the .gentypes files).
const CUnits NULL_UNITS_MASK;

// --------------------------------------------------------------------------
/// The mixin base class for all objects that need to be notified when a Unit
/// selection is made.
///
class CUnitSelectionChangeNotify
{
public:

    /// Create and keep a list of these objects.
    CUnitSelectionChangeNotify();

    /// Remove from the list of all objects.
    virtual ~CUnitSelectionChangeNotify();

    /// Notify all objects derived from CUnitSelectionChangeNotify that some unit selections
    /// have changed.
    static void SelectionsChanged();

    ///@{
    /// Do not allow copies.
    CUnitSelectionChangeNotify( const CUnitSelectionChangeNotify& ) = delete;
    CUnitSelectionChangeNotify& operator=( const CUnitSelectionChangeNotify& ) = delete;
    ///@}

    /// The method to override that is called after all units in list have been
    /// changed.
    virtual void    OnUnitsChanged( ) {}

private:

};

//-----------------------------------------------------------------------------
// Declare the units that we make available.  All the units that are settable,
// or based on a fundamental settable type must be included in this list.
// Units that have only a single choice can be declared and defined at the
// application or library level.

/// \defgroup PredefinedUnitSelections Predefined Unit Selections.
// @{
PCUnitSelection UnitSelSmallLength();
PCUnitSelection UnitSelCircleArea();
PCUnitSelection UnitSelRelationValue();
PCUnitSelection UnitSelCumulative();
PCUnitSelection UnitSelPressure();
PCUnitSelection UnitSelGaugePressure();
PCUnitSelection UnitSelPressureRate();
PCUnitSelection UnitSelViscosity();
PCUnitSelection UnitSelGasAmountAdsorbed();
PCUnitSelection UnitSelBJHPoreDim();
PCUnitSelection UnitSelGasPressure();
PCUnitSelection UnitSelGasPressureFlowChem();
PCUnitSelection UnitSelGasPressureSym();
PCUnitSelection UnitSelAbsTemp();
PCUnitSelection UnitSelAbsAnlTemp();
PCUnitSelection UnitSelLengthVolume();
PCUnitSelection UnitSelVolumeAbsolute();
PCUnitSelection UnitSelVolumeSpecific();
PCUnitSelection UnitSelLength_cm();
PCUnitSelection UnitSelArea_cm();
PCUnitSelection UnitSelVolume_cm();
PCUnitSelection UnitSelForce();
PCUnitSelection UnitSelHeight();
PCUnitSelection UnitSelIntrusion();
/// @}

/// \defgroup PredefinedUnits Predefined Unit Objects.
const CUnits& fcn_UT_InvalidUnits();
#define UT_InvalidUnits fcn_UT_InvalidUnits()
const CUnits& fcn_UT_SmallLength();
#define UT_SmallLength fcn_UT_SmallLength()
const CUnits& fcn_UT_SmallLength_Sqr();
#define UT_SmallLength_Sqr fcn_UT_SmallLength_Sqr()
const CUnits& fcn_UT_SmallLength_Inv();
#define UT_SmallLength_Inv fcn_UT_SmallLength_Inv()
const CUnits& fcn_UT_CircleArea();
#define UT_CircleArea fcn_UT_CircleArea()
const CUnits& fcn_UT_CircleArea_Inv();
#define UT_CircleArea_Inv fcn_UT_CircleArea_Inv()
const CUnits& fcn_UT_RelationValue();
#define UT_RelationValue fcn_UT_RelationValue()
const CUnits& fcn_UT_Cumulative();
#define UT_Cumulative fcn_UT_Cumulative()
const CUnits& fcn_UT_Cumulative_Std();
#define UT_Cumulative_Std fcn_UT_Cumulative_Std()
const CUnits& fcn_UT_Cumulative_Sieve();
#define UT_Cumulative_Sieve fcn_UT_Cumulative_Sieve()
const CUnits& fcn_UT_Volume();
#define UT_Volume fcn_UT_Volume()
const CUnits& fcn_UT_Mass();
#define UT_Mass fcn_UT_Mass()
const CUnits& fcn_UT_Volume_STP();
#define UT_Volume_STP fcn_UT_Volume_STP()
const CUnits& fcn_UT_Energy();
#define UT_Energy fcn_UT_Energy()
const CUnits& fcn_UT_SurfaceArea();
#define UT_SurfaceArea fcn_UT_SurfaceArea()
const CUnits& fcn_UT_Area();
#define UT_Area fcn_UT_Area()
const CUnits& fcn_UT_Pressure();
#define UT_Pressure fcn_UT_Pressure()
const CUnits& fcn_UT_GaugePressure();
#define UT_GaugePressure fcn_UT_GaugePressure()
const CUnits& fcn_UT_Force();
#define UT_Force fcn_UT_Force()
const CUnits& fcn_UT_Height();
#define UT_Height fcn_UT_Height()
const CUnits& fcn_UT_Height_Sqr();
#define UT_Height_Sqr fcn_UT_Height_Sqr()
const CUnits& fcn_UT_Length_cm();
#define UT_Length_cm fcn_UT_Length_cm()
const CUnits& fcn_UT_Area_cm();
#define UT_Area_cm fcn_UT_Area_cm()
const CUnits& fcn_UT_Volume_cm();
#define UT_Volume_cm fcn_UT_Volume_cm()
const CUnits& fcn_UT_PressureRate();
#define UT_PressureRate fcn_UT_PressureRate()
const CUnits& fcn_UT_GaugePressureRate();
#define UT_GaugePressureRate fcn_UT_GaugePressureRate()
const CUnits& fcn_UT_Weight();
#define UT_Weight fcn_UT_Weight()
const CUnits& fcn_UT_Temperature();
#define UT_Temperature fcn_UT_Temperature()
const CUnits& fcn_UT_Time_Hours();
#define UT_Time_Hours fcn_UT_Time_Hours()
const CUnits& fcn_UT_Time_Minutes();
#define UT_Time_Minutes fcn_UT_Time_Minutes()
const CUnits& fcn_UT_Time_Seconds();
#define UT_Time_Seconds fcn_UT_Time_Seconds()
const CUnits& fcn_UT_Time_MicroSeconds();
#define UT_Time_MicroSeconds fcn_UT_Time_MicroSeconds()
const CUnits& fcn_UT_Viscosity();
#define UT_Viscosity fcn_UT_Viscosity()
const CUnits& fcn_UT_Pressure_Inv();
#define UT_Pressure_Inv fcn_UT_Pressure_Inv()
const CUnits& fcn_UT_Pressure_Inv_Sqr();
#define UT_Pressure_Inv_Sqr fcn_UT_Pressure_Inv_Sqr()
const CUnits& fcn_UT_GasAmountAdsorbed();
#define UT_GasAmountAdsorbed fcn_UT_GasAmountAdsorbed()
const CUnits& fcn_UT_GasLogAmountAdsorbed();
#define UT_GasLogAmountAdsorbed fcn_UT_GasLogAmountAdsorbed()
const CUnits& fcn_UT_GasAmountAdsorbedAbs();
#define UT_GasAmountAdsorbedAbs fcn_UT_GasAmountAdsorbedAbs()
const CUnits& fcn_UT_BJHPoreDim();
#define UT_BJHPoreDim fcn_UT_BJHPoreDim()
const CUnits& fcn_UT_BJHPoreDim_Inv();
#define UT_BJHPoreDim_Inv fcn_UT_BJHPoreDim_Inv()
const CUnits& fcn_UT_DifferentialVolume();
#define UT_DifferentialVolume fcn_UT_DifferentialVolume()
const CUnits& fcn_UT_DifferentialArea();
#define UT_DifferentialArea fcn_UT_DifferentialArea()
const CUnits& fcn_UT_GasPressure();
#define UT_GasPressure fcn_UT_GasPressure()
const CUnits& fcn_UT_GasPressureFlowChem();
#define UT_GasPressureFlowChem fcn_UT_GasPressureFlowChem()
const CUnits& fcn_UT_GasLogPressure();
#define UT_GasLogPressure fcn_UT_GasLogPressure()
const CUnits& fcn_UT_GasLnPressure();
#define UT_GasLnPressure fcn_UT_GasLnPressure()
const CUnits& fcn_UT_GasSqrtPressure();
#define UT_GasSqrtPressure fcn_UT_GasSqrtPressure()
const CUnits& fcn_UT_GasPressureInv();
#define UT_GasPressureInv fcn_UT_GasPressureInv()
const CUnits& fcn_UT_GasPressureRate();
#define UT_GasPressureRate fcn_UT_GasPressureRate()
const CUnits& fcn_UT_GasPressureRateMin();
#define UT_GasPressureRateMin fcn_UT_GasPressureRateMin()
const CUnits& fcn_UT_GasOutgasRate();
#define UT_GasOutgasRate fcn_UT_GasOutgasRate()
const CUnits& fcn_UT_GasVacLevel();
#define UT_GasVacLevel fcn_UT_GasVacLevel()
const CUnits& fcn_UT_GasPressureSym();
#define UT_GasPressureSym fcn_UT_GasPressureSym()
const CUnits& fcn_UT_GasPressureInvSym();
#define UT_GasPressureInvSym fcn_UT_GasPressureInvSym()
const CUnits& fcn_UT_AbsTemp();
#define UT_AbsTemp fcn_UT_AbsTemp()
const CUnits& fcn_UT_AbsTempRate();
#define UT_AbsTempRate fcn_UT_AbsTempRate()
const CUnits& fcn_UT_AbsTempInternalC();
#define UT_AbsTempInternalC fcn_UT_AbsTempInternalC()
const CUnits& fcn_UT_RelTemp();
#define UT_RelTemp fcn_UT_RelTemp()
const CUnits& fcn_UT_RelativePressure();
#define UT_RelativePressure fcn_UT_RelativePressure()
const CUnits& fcn_UT_InvRelativePressure();
#define UT_InvRelativePressure fcn_UT_InvRelativePressure()
const CUnits& fcn_UT_GasAmountAdsorbedInv();
#define UT_GasAmountAdsorbedInv fcn_UT_GasAmountAdsorbedInv()
const CUnits& fcn_UT_GasPo();
#define UT_GasPo fcn_UT_GasPo()
const CUnits& fcn_UT_freeSpaceVolume();
#define UT_freeSpaceVolume fcn_UT_freeSpaceVolume()
const CUnits& fcn_UT_Density();
#define UT_Density fcn_UT_Density()
const CUnits& fcn_UT_AbsAnlTemp();
#define UT_AbsAnlTemp fcn_UT_AbsAnlTemp()
const CUnits& fcn_UT_Percentage();
#define UT_Percentage fcn_UT_Percentage()
const CUnits& fcn_UT_MolarEnergy();
#define UT_MolarEnergy fcn_UT_MolarEnergy()
const CUnits& fcn_UT_TubeDiameter();
#define UT_TubeDiameter fcn_UT_TubeDiameter()
const CUnits& fcn_UT_Intrusion();
#define UT_Intrusion fcn_UT_Intrusion()
const CUnits& fcn_UT_IntrusionSpecific();
#define UT_IntrusionSpecific fcn_UT_IntrusionSpecific()
const CUnits& fcn_UT_IntrusionEquil();
#define UT_IntrusionEquil fcn_UT_IntrusionEquil()
const CUnits& fcn_UT_LengthVolume();
#define UT_LengthVolume fcn_UT_LengthVolume()
const CUnits& fcn_UT_VolumeAbsolute();
#define UT_VolumeAbsolute fcn_UT_VolumeAbsolute()
const CUnits& fcn_UT_VolumeSpecific();
#define UT_VolumeSpecific fcn_UT_VolumeSpecific()
const CUnits& fcn_UT_Conductivity();
#define UT_Conductivity fcn_UT_Conductivity()
const CUnits& fcn_UT_GasFlowRate();
#define UT_GasFlowRate fcn_UT_GasFlowRate()
const CUnits& fcn_UT_AtomicArea();
#define UT_AtomicArea fcn_UT_AtomicArea()
const CUnits& fcn_UT_GraphicSize();
#define UT_GraphicSize fcn_UT_GraphicSize()
const CUnits& fcn_UT_GasAmountPerMin();
#define UT_GasAmountPerMin fcn_UT_GasAmountPerMin()
const CUnits& fcn_UT_PercentPerMin();
#define UT_PercentPerMin fcn_UT_PercentPerMin()
const CUnits& fcn_UT_PercentPerMinSquared();
#define UT_PercentPerMinSquared fcn_UT_PercentPerMinSquared()
const CUnits& fcn_UT_PercentFullScale();
#define UT_PercentFullScale fcn_UT_PercentFullScale()
const CUnits& fcn_UT_PercentFullScalePerMin();
#define UT_PercentFullScalePerMin fcn_UT_PercentFullScalePerMin()
const CUnits& fcn_UT_Current();
#define UT_Current fcn_UT_Current()
/// @}

/// Identifiers used in the defined unit selections.
enum EMicUnit : unsigned short
{
    ceMicUnitNanometers          = 0,            // Small Length
    ceMicUnitAngstroms,

    ceMicUnitDiameter,                           // Circle Area
    ceMicUnitRadius,

    ceMicUnitPercent,                            // RelationValue
    ceMicUnitFraction,

    ceMicUnitCoarser,                            // Cumulative
    ceMicUnitFiner,

    ceMicUnitcm3_per_gram,                    // UT_Volume
    ceMicUnitcm3_per_gram_STP,                // UT_Volume_STP
    ceMicUnite_per_k,                         // UT_Energy
    ceMicUnitmeters2,                         // UT_SurfaceArea

    ceMicUnitPSIA,                            // UT_Pressure,
    ceMicUnitMegaPascals,                     // UT_Pressure,
    ceMicUnitmmHg,                            // UT_Pressure,
    ceMicUnitAtm,                             // UT_Pressure,
    ceMicUnitmbar,                            // UT_Pressure,
    ceMicUnitbar,                             // UT_Pressure,
    ceMicUnitMicrometers,

    ceMicUnitCentipoise,                      // UT_Viscosity,
    ceMicUnitMegaPascals_Second,

    ceMicUnitKiloPascals,                     // UT_Pressure
    ceMicUnitPascals,
    ceMicUnitKgForce_cm2,                     // Kgf/cm2

    ceMicUnitmeter,
    ceMicUnitmilimeter,

    ceMicUnitGas_mmol,
    ceMicUnitGas_ccSTP,
    ceMicUnitGas_umol,

    ceMicUnitBJHPore_width,
    ceMicUnitBJHPore_radius,
    ceMicUnitBJHPore_diameter,

    ceMicUnitGasPress_kPa,
    ceMicUnitGasPress_mbar,
    ceMicUnitGasPress_mmHg,
    ceMicUnitGasPress_Atm,
    ceMicUnitGasPress_bar,
    ceMicUnitGasPress_psia,

    ceMicUnitGasPressSym_iso,
    ceMicUnitGasPressSym_mic,

    ceMicUnitGasRelPressSym_iso,
    ceMicUnitGasRelPressSym_mic,

    ceMicUnitGasPoSym_iso,
    ceMicUnitGasPoSym_mic,

    // Force types
    ceMicUnitNewton,
    ceMicUnitkiloNewton,
    ceMicUnitpoundsforce,

    // Measures types.
    ceMicUnitLength_cm,
    ceMicUnitLength_in,

    ceMicUnitArea_cm2,
    ceMicUnitArea_in2,

    ceMicUnitVolume_cm3,
    ceMicUnitVolume_in3,

    // Temperature
    ceMicUnitTemp_C,
    ceMicUnitTemp_K,

    ceMicUnitcm3,  // UT_freeSpaceVolume

    ceMicUnitAnlTemp_C,
    ceMicUnitAnlTemp_K,

    ceMicUnitLength,      //  UT_LengthVolume
    ceMicUnitVolume,

    ceMicUnitVolAbs_um3,  //  UT_VolumeAbsolute
    ceMicUnitVolAbs_fL,

    ceMicUnitVolSpec_1mL, //  UT_VolumeSpecific
    ceMicUnitVolSpec_10mL,
    ceMicUnitVolSpec_100mL,
    ceMicUnitVolSpec_1L,

    ceMicUnitmL,          // UT_Intrusion
    ceMicUnitmm3,

    ceMicUnitGasPressFlowChem_Pa,    // UT_GasPressFlowChem
    ceMicUnitGasPressFlowChem_kPa,   // UT_GasPressFlowChem
    ceMicUnitGasPressFlowChem_MPa,   // UT_GasPressFlowChem
    ceMicUnitGasPressFlowChem_mbar,  // UT_GasPressFlowChem
    ceMicUnitGasPressFlowChem_mmHg,  // UT_GasPressFlowChem
    ceMicUnitGasPressFlowChem_atm,   // UT_GasPressFlowChem
    ceMicUnitGasPressFlowChem_bar,   // UT_GasPressFlowChem
    ceMicUnitGasPressFlowChem_psia,  // UT_GasPressFlowChem

    ceMicUnitSingleValue      = 32765,
    ceMicUnitValuesEnd        = 32766,
    ceMicUnitInvalid          = 32767
};


#endif
