#ifndef CALIBRATION_HPP
#define CALIBRATION_HPP
//-------------------------------------------------------------------------------------------------
/// @file The classes that are used for calibration.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include <utility>

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------

// ================================================================================================
/// The calibration information, and calculation for a point and slope calibration.
class CPointSlopeCal final
{
public:
    /// @returns a CPointSlopeCal object given two points.
    static constexpr CPointSlopeCal FromTwoPoints(float_32 flRaw1,
        float_32 flCal1,
        float_32 flRaw2,
        float_32 flCal2);

    constexpr CPointSlopeCal(float_32 flSlope, float_32 flRaw, float_32 flCal);
    CPointSlopeCal(const CPointSlopeCal&) = default;
    CPointSlopeCal(CPointSlopeCal&&) = default;
    CPointSlopeCal& operator=(const CPointSlopeCal&) = default;
    CPointSlopeCal& operator=(CPointSlopeCal&&) = default;
    ~CPointSlopeCal() = default;

    /// @returns the slope parameter of the calibration.
    float_32 Slope() const;
    /// @returns the <flRaw, flCal> values of the point of the calibration.
    std::pair<float_32, float_32> Point() const;

    /// @returns the calibrated value from the @p flRaw reading.
    float_32 CalFromRaw(float_32 flRaw) const;

    /// @returns the raw value that would have generated the @p flCal reading.
    float_32 RawFromCal(float_32 flCal) const;

private:
    ///@{ The parameters of the calibration so that c = m_flSlope * (r-m_flRaw) + m_flCal
    float_32 m_flSlope;
    float_32 m_flRaw;
    float_32 m_flCal;
    ///@}
};

// ================================================================================================
/// The calibration information, and calculation for a point and slope calibration for an RTD with
/// non-linearity lookup.
class CRTDCal final
{
public:
    /// @returns a CRTDCal object given two points.
    static CRTDCal FromTwoPoints(float_32 flRaw1,
        float_32 flCal1,
        float_32 flRaw2,
        float_32 flCal2);

    constexpr CRTDCal(float_32 flSlope, float_32 flRaw, float_32 flCal);
    constexpr CRTDCal();
    CRTDCal(CRTDCal&&) = default;
    CRTDCal& operator=(CRTDCal&&) = default;
    ~CRTDCal() = default;

    /// @returns the slope parameter of the calibration.
    float_32 Slope() const;
    /// @returns the <flRaw, flCal> values of the point of the calibration.
    std::pair<float_32, float_32> Point() const;

    /// @returns the calibrated value from the @p flRaw reading.
    float_32 CalFromRaw(float_32 flRaw) const;

    /// @returns the raw value that would have generated the @p flCal reading.
    float_32 RawFromCal(float_32 flCal) const;

private:

    /// The voltage to voltage calibration before the non-linearity is applied.
    CPointSlopeCal m_resistanceCal;
};

//-------------------------------------------------------------------------------------------------
constexpr CPointSlopeCal CPointSlopeCal::FromTwoPoints(float_32 flRaw1,
    float_32 flCal1,
    float_32 flRaw2,
    float_32 flCal2)
{
    return CPointSlopeCal(flRaw1 != flRaw2 ? (flCal2 - flCal1) / (flRaw2 - flRaw1) : 0.0f,
        flRaw1,
        flCal1);
}

constexpr CPointSlopeCal::CPointSlopeCal(float_32 flSlope, float_32 flRaw, float_32 flCal)
    : m_flSlope(flSlope),
      m_flRaw(flRaw),
      m_flCal(flCal)
{
}

//-------------------------------------------------------------------------------------------------
constexpr CRTDCal::CRTDCal(float_32 flSlope, float_32 flRaw, float_32 flCal)
    : m_resistanceCal(flSlope, flRaw, flCal)
{
}

constexpr CRTDCal::CRTDCal()
    : m_resistanceCal(1.0, 0.0, 0.0)
{
}

#endif // CALIBRATION_HPP
