#ifndef COMDTLNK_HPP
#define COMDTLNK_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "comphys.hpp"
#if !defined(MIC_DM643)
#include <cpp/tm1shot.hpp>
#else
#include <cpp/tmtimer.hpp>
#endif
#if defined(MIC_DM643)
#include "nucpp.h"
#endif
#if !defined(MIC_DM643) && !defined(MIC_VCMFC) && !defined(MIC_TEST)
#include <fstream>
#endif
//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
class CCommDataLink;
class CCommDataLinkChannel;
class CCommMonitor;
// ================================================================================================
///  This class handles the data-link layer of the communications 'stack'.  This provides the
///  functionality of sending packets error free between multiple chanels over the same physical
///  data link. It also provides confirmation 'callbacks' when the packet was received.
class CCommDataLink :
#if defined(MIC_DM643)
    private Task,
#endif
    public CCommPhysicalEvH
{
#if defined(MIC_DM643)
public:  // compiler bug workaround
#else
private:
#endif
#ifdef MIC_VCMFC
#pragma pack(1)
#endif

    struct SHeader
    {
        uint_16 uiDummy;        ///< Used to be crc, but not needed for tcp.
        uint_16 uiAck : 5;      ///< ack on the last packet
        uint_16 uiSeq : 5;      ///< seq = 0xF means reset.
        uint_16 uiChannel : 5;  ///< which channel to send.
        ///  The fExtra flag is zero for normal packets. It is 1 To indicate that another byte of
        ///  'header' information follows this structure.
        ///
        ///  The byte following this is interpreted as two 4 bit quantities. The high order four
        ///  bits must be zero to indicate that this is a compressed packet. If the high order bits
        ///  are zero, the 4 low order bits give the number of compression cycles used.
        uint_16 fExtra : 1;
    };
#ifdef MIC_VCMFC
#pragma pack()
#endif

public:
    ///   Set the timeout interval to the given time.
    static void SetTimeout(CTimeInterval cTimeout)
    {
        ms_ulTimeoutLengthmsec = cTimeout.MilliSeconds();
    }

#if defined(MIC_DM643)
    CCommDataLink(CCommPhysical* pPhys,
        UNSIGNED uiStackSize,
        OPTION uiPriority,
        UNSIGNED uiTimeSlice,
        BOOL fPreempt);
#else
    CCommDataLink(CCommPhysical* pPhys);
#endif

    /// @{
    ///  These two create/break the link to the event handlers there can be more than one registered
    ///  at any time.
    void Register(CCommDataLinkChannel* pChannel);
    void UnRegister(CCommDataLinkChannel* pChannel);
    /// @}

    /// @{
    /// These two create/break the link to the event handlers there can be only one registered at
    /// any time.
    void Register(CCommMonitor* pMon);
    void UnRegister(CCommMonitor* pMon);
    /// @}

    /// This is called to initialize the datalink layer. It will keep sending a reset packet over
    /// the physical layer unitl it is ack'ed at which point both sides are in their initial states
    /// and ready to handle data communications. When this happens, the InitFinished() event is sent
    /// to the Monitor.
    ///
    /// Only one of the sides (the front end) can issue the init command to send the reset. The
    /// other side (the mass) just creates this object, and it is reset as needed. When it is reset,
    /// it calls the Reset() event in the handler to notify it that all dataitems pending have been
    /// lost.
    void Init();

    /// This is called to reset the datalink. It's main purpose is to synchronize the one side if
    /// the other size desets (linke on an execute or reset message sent to the mass. It calls the
    /// Reset() of all the event handlers to notify that all dataitems pending have been lost.
    void Reset(bool fResetChannels = true);

    /// Called to indicate that one of the handlers has information that is ready to be sent. This
    /// will trigger the request of that information to be sent.
    void DataReadyToSend();

    void CommFailed();

    enum
    {
        /// The maximum size of a single packet that can be sent by this layer.  The -1 is for the
        /// extra byte on compressed packets.
        ceMaxPacketSize = CCommPhysical::ceMaxPacketSize - sizeof(SHeader) - 1,

        /// The number of retries on sending a packet before the CCommMonitor is notified of a
        /// potential problem.
        ceNumRetries = 3,

        /// The maximum number of chanels that handled by one connection.
        /// NOTE 1:     must be consistemt with channel in SHeader.
        /// NOTE 2:     there is an fixed array of these for quick lookup.
        ceMaxChannels = 32,

        /// The maximum number of packets that can be pending to confirm.  I.e. the number of
        /// packets that can be send before the first one is ack'ed. This number should be large so
        /// that many small packets are sent waiting for a large one to be returned with the ack.
        ///
        /// The ack and seq fields in the SHeader structure below must be large enough to hold this
        /// many plus 1 needed for the the sending algorithm, plus another one for the reset.
        ceMaxPending = 30,

        /// This is the ID used for the special reset packet.
        ceResetID = ceMaxPending + 1,

        /// This is the ack used for the special ConfirmConnect packet.
        ceConfirmPkt = ceResetID - 1,

        /// This is the timer used for deciding when to send the special ConfirmConnect packet.
        ceConfirmTimer = ceMaxPending + 1,

        /// This is the timer used for timeout on sending the special ConfirmConnect packet.
        ceConfirmAckTimer = ceMaxPending + 2
    };

    /// This is called when the ConfirmTimer expires.  It sends a special Confirm message to keep
    /// the link from timing out.
    void TimeExpired(uint_32 ulID);

    /// Fill in the status into the stream so that it can be shown.  szStr should be long enough to
    /// hold all the data.
    void FillDebugStatus(char* szStr) const;

#if !defined(MIC_DM643)
    /// Fill in the status into the stream so that it can be shown.
    void FillDebugStatus(std::ostream& os) const;
#endif

    CCommPhysical* GetPhys() const { return m_pPhys; }

    /// Set true if connection is made by a transport layer on a separate socket
    void SetSocketCom(bool fSocketCom);
    /// @returns true iff communicating by separate socket for each transport channel
    bool IsSocketCom() const;

private:
    enum ESpecialType
    {
        ceReset,
        ceResetAck,
        ceSimpleAck,
        ceConfirmConnect
    };

    struct SPending
    {
        const char* pHead;  ///<  The location of the header.
        int uiHeadSize;     ///<  The size of the header data.

        const char* pBuffer;  ///<  The location of the data.
        int uiBufferSize;     ///<  The size of the data.

        int uiChannel;  ///<  The channel to which this belongs.

        SHeader Header;  ///<  The header for the packet.
    };

    // CCommPhysicalEvH implementation.
    virtual void DataSent() /*override*/;
    virtual void DataAvailable() /*override*/;
    virtual void Connected() /*override*/;
    virtual void Disconnected() /*override*/;

    /// This is called when an expected ack was not received in the timeout period. This will
    /// trigger a retransmition of all pending messages. If there are more than several retries on
    /// on the same packet, the associated CCommMonitor is notified, to see if there should be
    /// further retries. If no Monitor is present, the retries are continued.
    void TimeExpired();

#if !defined(MIC_DM643)
    /// This will start the timer, if it is already started, it will reset the timer. If the timer
    /// expires, it will call the TimeOut() method.
    void StartTimer(uint_16 uiTimer);

    /// This will stop the timer so that it will not expire.
    void StopTimer(uint_16 uiTimer);

    /// Start the ConfirmTimer to keep the link from timing out.
    void StartConfirm() { StartTimer(ceConfirmTimer); }
#endif

#if defined(MIC_DM643)
    /// The function overwritten to be the background task.
    VOID Entry();
#endif

    /// This is the function that gets called to do the actual work of the receiving and sending the
    /// data. It will return true iff some work needed to be done.
    bool DoWork();

    /// This is is called (in the comm task) when a packet (potentialy invalid) has been received by
    /// the physical layer. This makes sure it is valid, It Handles special packets, and delivers
    /// all others to the necessary receiver.
    void HandleRecievePacket();

    /// This is called when it is possible to send the next packet. This will send another packet if
    /// available, if none is available, and an ack is necessary it will send an ack packet.
    bool HandleSendPacket();

    /// This will send the given packet to the physical layer.
    void SendPacket(uint_16 uiPacket);

    /// Send a special packet over the comm line.
    void SendSpecial(ESpecialType eType);

    /// This tells this object that there is something that needs servicing whenever is
    /// appropriate. In Windows, all the outstanding work to be done is handled immediately. In the
    /// embedded system, the comm task associated with this is notified that there is something to
    /// do.
    void NotifyWork();

    /// The length of time for the timeout interval in msec. The reason this is not a CTiemINterval
    /// class is that initializing an static integer is more efficient thatn a class.
    static uint_32 ms_ulTimeoutLengthmsec;
    /// The length of time for the confirm timeout interval in msec.
    static uint_32 ms_ulConfirmTimeoutmsec;

    bool m_fSocketCom;       ///< true iff communicating by separate socket for each channel
    CCommPhysical* m_pPhys;  ///< The physical transport layer to use to send packets.
    bool m_fPhysSending;     ///<  true iff the physical layer is buisy sending a packet.
    int m_uiNumRcv;          ///<  The number of packets that are available to receive.
    bool m_fLastSpecial;     //< true iff the last signal sent was a special (no ack) packet.
    CCommDataLinkChannel* m_apChannels[ceMaxChannels];  ///< The array of channels to interact with.
    CCommMonitor* m_pMonitor;                           ///<  The monitor that is linked to this.
    int m_uiRetryCount;  ///<  The number of retries without a sucessful transmition.

    /// True if the comm linke failed this will cause it to ignore
    /// all the requests except for Reset.
    bool m_fFailed;

    bool m_fWorking;       ///<  Flag to make sure the NotifyWork requests are handled sequntialy.
    bool m_fInitializing;  ///<  true iff This is in the middle of an initializing sequence.

    /// This is set if the Init() function was called while the datalink was working. This will
    /// scedule an init call at the end of the Do Work execution.
    bool m_fDoInit;

    uint_16 m_uiAckExpected;     ///<  The oldest as yet unack'ed packet.
    uint_16 m_uiPacketExpected;  ///<  The next packet expected.
    bool m_fAckSent;  ///<  This is true when the ack sent is up to date with what should be sent.
    SPending m_aPendingPacket[ceMaxPending + 1];
    SHeader m_ResetHead;  ///<  The reset and reset request packets.

    ///  The index into the m_aPendingPacket containing the next available packet to use.
    uint_16 m_uiNextPacket;

    /// The number of the next packet to resend. If this equals m_uiNextPacket, there are no more to
    /// resend.
    uint_16 m_uiNextResend;

    int m_uiNumBuffered;  ///< The number of packets that are buffered waiting for an ack.
    int m_uiLastChanel;   ///<  The last channel that sent information.

#if !defined(MIC_DM643)
    CTimeOneShot m_aTimer[ceMaxPending + 1];  ///<  The timers that this class uses.

    /// This timer is for sending ceConfirmConnect messages to avoid receive timeouts at the other
    /// end
    CTimeOneShot m_cConfirmTimer;
    /// This timer is for detecting Confirm send timeouts on this end
    CTimeOneShot m_cConfirmAckTimer;
#endif

#if defined(MIC_DM643)

    Event m_Event;  ///<  The event communication with the background task.
#endif

    DECLARE_MSG()
};

// ================================================================================================
/// This is the event handler for events generated by a specific object of type CCommDataLink. The
/// events sent are ResetChannel(), InitFinnished(), DataSend(), and DataReceived(). It also has
/// data request methods HasDataToSend(), and GetDataToSend().
class CCommDataLinkChannel
{
public:
    /// Create this handler associated with pSrc. This does not automaticaly register with pSrc
    /// since the derived object is not fully constructed.  To do the registering, simple call this
    /// classes Register() method.  The uiID must be unique for all the handlers for pSrc, and
    /// identifes it for recieving messages.
    CCommDataLinkChannel(CCommDataLink* pSrc, int uiID);

    /// If this was registered, it will unregister itself with pSrc.
    virtual ~CCommDataLinkChannel();

    /// This registers with the pSrc passed in at construction.
    void Register();

    /// @returns the associated datalink.
    CCommDataLink* GetSource() { return m_pSrc; }

    /// @returns the ID of this channel.
    int ID() const { return m_uiID; }

    /// This is called if the other end of the link requested a reset of the communications
    /// subsystem. See Init() above.
    virtual void ResetChannel() = 0;

    /// This is called once for each time that GetDataToSend() was called, unless the datalink layer
    /// was reset in which case the Reset() function was called.
    virtual void DataSent() = 0;

    /// This is called for every packet that was received by the DataLink layer. The data is valid
    /// only for the duration of this call, so if it needs to exist byond that, it needs to be
    /// copied. The header if any was sent is placed in the head of the package.
    virtual void DataReceived(const char* pData, int uiSize) = 0;

    /// This is called to see if this chanel has any data ready to send.
    virtual bool HasDataToSend() const = 0;

    /// This is called to retrieve the next data item that is ready to be sent. The data must
    /// continue to exist until the corresponding DataSent() event has been called, or if the chanel
    /// has been reset.  The header if any is placed in the front of the packet.
    virtual void GetDataToSend(const char** ppHead,
        int* puiHeadSize,
        const char** ppData,
        int* puiDataSize)
        = 0;

    /// Fill in the status into the stream so that it can be shown.  szStr should be long enough to
    /// hold all the data.
    virtual void FillDebugStatus(char* szStr) const = 0;

    /// Try to connect separate transport layer socket.
    virtual void Connect() = 0;

private:
    bool m_fRegistered;     ///< true iff this is registered.
    CCommDataLink* m_pSrc;  ///< The associated CCommDataLink
    int m_uiID;             ///<  The ID of this channel.

    DECLARE_MSG()
};
// ================================================================================================
/// This class is used to define the interface to the application for monitoring the communications
/// subsystem.
class CCommMonitor
{
public:
    /// Create a monitor associated with the given Datalink.
    CCommMonitor(CCommDataLink* pSrc);
    /// If this was registered, it will unregister itself with pSrc.
    virtual ~CCommMonitor();

    /// This registers with the pSrc passed in at construction.
    void Register();

    /// This is called when the initialization is finished.
    virtual void InitFinnished() = 0;

    /// This is called to indicate that there is a problem with the communications link. If this
    /// returns true, the datalink object that called this will retry the transmition, if it retruns
    /// false, it will not.
    virtual bool CommFailed(int uiRetryCount) = 0;

private:
    bool m_fRegistered;     ///< true iff this is registered.
    CCommDataLink* m_pSrc;  ///< The associated CCommDataLink

    DECLARE_MSG()
};

#endif  // COMDTLNK_HPP
