#ifndef COMLOCL_HPP
#define COMLOCL_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "comdtlnk.hpp"

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
// -------------------------------------------------------------------------------------------------
/// @{
/// A range of messages for the communications subsystem.  Be careful not to / overlap with the
/// messages defined in miclib/mfrmbase.hpp
const auto WM_USER_COMM_DELAY(WM_USER + 500);
const auto WM_USER_COMM_DELAY_END(WM_USER_COMM_DELAY + 20);
/// @}

// ===============================================================================================
/// This is the secondary physical layer in a local communications loop. It is internal to the
/// primary physical layer described below.  It also is the base class for the primary layer that
/// does all the work of transferring the packets.
///
/// The reason for this separation is to eliminate the need for dynamic memory allocation in these
/// classes.
class CCommPhysicalLocalBase : public CCommPhysical
{
public:
    /// Creates the secondary (hidden CommLocal).
    CCommPhysicalLocalBase(CCommPhysicalLocalBase* pTarget);

    virtual ~CCommPhysicalLocalBase();

    /// Overriden for the demo.
    virtual bool IsLocal() const { return true; }

    /// This will receive a packet.
    void Receive(const char* pHead1,
        int uiHead1Size,
        const char* pHead2,
        int uiHead2Size,
        const char* pData,
        int uiDataSize);

    /// Called by the main window when it receives the message posted in Send().  This allows the
    /// communications stack to unwind.
    void ReadyToReceive(int iMessageID);

    virtual void Send(const char* pHead1,
        int uiHead1Size,
        const char* pHead2,
        int uiHead2Size,
        const char* pData,
        int uiDataSize) override;
    virtual void ResetSend() override;

private:
    CCommPhysicalLocalBase* m_pTarget;  ///<  The sister physical layer.

    const char* m_pHead1;
    int m_uiHead1Size;
    const char* m_pHead2;
    int m_uiHead2Size;
    const char* m_pData;
    int m_uiDataSize;

    bool m_fReset;
    const int m_ciMessage;

    DECLARE_MSG()
};

//=================================================================================================
/// This is a local physical layer that eliminates the need for an RS-232 link or other medium. This
/// contains an internal secondary physical layer, as well as an accessible datalink layer that
/// receives the packets sent after this object and vice versa.
///
/// One of the uses of this class is in the simulation of back ends in the front end. In this case,
/// Create one of these instead of the RS-232 physical layer and link in the proper simulation
/// Transport layers to the TargetDataLink.
class CCommPhysicalLocal : public CCommPhysicalLocalBase
{
public:
    /// Create the local physical layer along with the target datalink.
    CCommPhysicalLocal();

    /// @retuns the datalink for the other end.
    CCommDataLink* GetTargetDataLink() { return &m_DataLink; }

private:
    CCommPhysicalLocalBase m_Secondary;  ///< The secondary local comm link.
    CCommDataLink m_DataLink;            ///< The datalink for the secondary physical layer.

    DECLARE_MSG()
};

#endif  // COMLOCL_HPP
