#ifndef COMPHYS_HPP
#define COMPHYS_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include <cpp/cstr.hpp>

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
class CCommPhysical;
class CCommPhysicalEvH;
// ================================================================================================
/// This defines the physical layer of the comm subsystem.
///
/// This class abstracts the physical level of the communications 'stack' This level provides the
/// functionality to send and receive packets of information. This does not provide an error free
/// protocall, Some packets may even be lost completely.  An object that uses an instance of this
/// class should be derived from CCommPhysicalEvH to receive notification of events.
///
/// This class also handles the receive buffers for all the derived classes.
class CCommPhysical
{
public:
    CCommPhysical();
    virtual ~CCommPhysical() {}

    /// @{
    /// These two create/break the link to the event handlers.
    void Register(CCommPhysicalEvH* pEvH);
    void UnRegister(CCommPhysicalEvH* pEvH);
    /// @}

    /// Called when the tcp should reconnect.
    virtual void Reconnect() {}

    /// Is this a local connection? Override and return true for demo.
    virtual bool IsLocal() const { return false; }

    /// See if the connection is made.  If the connection cannot tell, assume connected.
    virtual bool IsConnected() const { return true; }

    /// This will send a packet of information containing both the header, followed by the data
    /// through the channel. This call will return immediately and send the data 'in the
    /// background'. The buffers must be copied localy since there is no requirement that they are
    /// preserved after the return from this call.  Once this is sent, The class will call the
    /// DataSent() method of the associated handler.  It is illegal to call this function if the
    /// class is already in the middle of sending another packet.
    virtual void Send(const char* pHead1,
        int uiHead1Size,
        const char* pHead2,
        int uiHead2Size,
        const char* pData,
        int uiDataSize)
        = 0;

    /// This will stop any previous send commands in the middle, and get ready to start the next
    /// transfer.
    virtual void ResetSend() = 0;

    /// These give access to the size and buffer of the last received packet. It is illegal to call
    /// these if there is no data available.
    const char* GetPacketBuff() const;
    int GetPacketSize() const;

    /// This notifies this object that the last buffer received has been fully processed, and that
    /// memory is no longer needed.
    void FreePacket();

    virtual bool IsFatalComError() { return false; }
    int GetFatalComError() { return m_iFatalError; }
    void SetFatalComError(int iFatalError);

    enum
    {
        /// The maximum number of bytes for a packet.
        ///
        /// This number is chosen so that the 4+1 bytes for the datalink header and the (possible)
        /// 5 bytes for the transport header result in a nice size of transport/message packets.
        /// Also 1024 is large enough so the overhead is relatively small, but small enough so that
        /// the time to transfer is still only about 0.5 sec at 19.2.
        ceMaxPacketSize = 1024 + 4 + 5 + 1
    };

    /// Fill in the status into the stream so that it can be shown.  szStr should be long enough to
    /// hold all the data.
    virtual void FillDebugStatus(char* szStr) const { *szStr = 0; }

    /// @return a pointer to the instrument address string (if any)
    virtual CStr GetInstAddr() const { return CStr(_T("")); }

protected:
    /// This will return one of received buffers to be used to store the incoming bytes. If there
    /// are no more buffers available, 0 is returned. (and that packet is lost). If a buffer is
    /// returned, that will be the receive buffer until NotifyDataAvailable() is called to indicate
    /// that the buffer is full.
    char* GetReceiveBuff();

    /// This is called by the derived class whenever it has finished sending a packet of data. This
    /// will notify the m_pEvH if it exists.
    void NotifyDataSent();

    /// This is called by the derived class whenever it has finished recieving a packet of
    /// data. This will mark the current receive buffer as full to make it available as a received
    /// packet.  This will notify the m_pEvH if it exists.
    void NotifyDataAvailable(int uiSize);

    ///  The single event handler for this Physical connection.
    CCommPhysicalEvH* m_pEvH;

private:
    enum
    {
        /// The number of receive buffers that are available. Each contains the full ceMaxPacketSize
        /// room, There should be more than one of these so that the next packet can be received
        /// while the first is being handled.
        ceMaxBuffers = 8
    };

#if defined(MIC_DM643)
public:  // compiler bug workaround
#endif
    enum EBuffState
    {
        ceEmpty,    //  No data in the buffer.
        ceFilling,  //  Buffer is in the process of being filled.
        ceFull      //  Data is available in the buffer.
    };
#if defined(MIC_DM643)
private:
#endif

    /// Information needed for each of the ceMaxBuffers this holds.
    struct SBuffInfo
    {
        EBuffState eState;            //  The state of the buffer.
        int uiSize;                   //  The number of bytes in aBuff
                                      //  Valid if eState == ceFull.
        char aBuff[ceMaxPacketSize];  //  The storage area.
    };

    /// The buffers to store recieving packets into.
    SBuffInfo m_aReceiveBuffer[ceMaxBuffers];

    /// The index to the receive buffers that GetPacketBuff will return.
    int m_iNextPacket;

    /// The index into the receive buffers that GetReceiveBuff will return.
    int m_iNextBuffer;

    int m_iFatalError;

    DECLARE_MSG()
};
// ================================================================================================
/// This is an event handler for events generatrd by a specific object of type CCommPhysical. The
/// events are DataSent(), and DataAvailable().
class CCommPhysicalEvH
{
public:
    /// Create this handler associated with pSrc. This does not automatically register with pSrc
    /// since the derived object is not fully constructed.  To do the registering, simple call this
    /// classes Register() method.
    CCommPhysicalEvH(CCommPhysical* pSrc);

    /// Registers with the pSrc passed in at construction.
    void Register();

    /// If this was registered, it will unregister itself with pSrc.
    virtual ~CCommPhysicalEvH();

    /// @returns the CCommPhysical object with which this is associated.
    CCommPhysical* GetSource() { return m_pSrc; }

    /// This is called by the m_pSrc object whenever it completed sending a packet of information.
    virtual void DataSent() = 0;

    /// This is called by the m_pSrc object whenever there is a completed data packet received.
    virtual void DataAvailable() = 0;

    /// This is called by the m_pSrc object whenever there is a connection completed.
    virtual void Connected() = 0;

    /// This is called by the m_pSrc object whenever there is a disconnection.
    virtual void Disconnected() = 0;

private:
    bool m_fRegistered;     ///< true iff this is registered.
    CCommPhysical* m_pSrc;  ///< The associated CCommPhysical.

    DECLARE_MSG()
};

#endif  // COMPHYS_HPP
