#ifndef COMSTRTB_HPP
#define COMSTRTB_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "comtrns.hpp"
#include "queryinf.h"

// ================================================================================================
/// This is the base class for the two classes that handle the startup and initialization code in
/// the front end, and back end. The reason for this class is to make sure that the common
/// structures are well defined.
class CCommStartupBase : public CCommTransport
{
public:
#ifdef MIC_VCMFC
#pragma pack(1)
#endif
    ///  Indicates what type of download is to be performed.
    enum EDownloadType
    {
        ceBootBlock,
        ceApplication
    };

    ///  CPU Clock frequency
    enum EClockFrequency
    {
        ce25MHz = 0,
        ce20MHz,
        ce33MHz,
        ce16MHz
    };

    ///  The structure passed as part of the querry data message.
    struct SQueryInfo
    {
        uint_8 eInitState;               ///< EInitState of the application.
        uint_8 eBooted;                  ///< EBootType flag: who is running
        uint_8 eBootSelect;              ///< EBootType flag: who is selected
        uint_16 fPOSTError : 1;          ///< General POST error flag
        uint_16 fBootBlockCRCError : 1;  ///< Specific POST error flags
        uint_16 fDRAMError : 1;
        uint_16 fCom1Error : 1;
        uint_16 fCom2Error : 1;
        uint_16 fDebugComError : 1;
        uint_16 fApplicationCRCError : 1;
        uint_16 fEtherNetError : 1;
        uint_16 fCardInstalled : 1;        ///< PCMCIA card installed
        uint_16 fFlashWriteProtected : 1;  ///< Flash chip write protected
        uint_16 fEtherNetInstalled : 1;
        uint_16 fDRAMUsed : 1;          ///< DRAM is used (not just NVRAM)
        uint_16 fLegacyIOEnabled : 1;   ///< 1330/1360/23XX i/o enabled
        uint_16 fExternalComPorts : 1;  ///< Com1-4 on a separate board
        uint_16 eClockFrequency : 2;    ///< EClockFrequency
        uint_32 uiDRAMErrorAddress;     ///< Address of DRAM error if any
        uint_32 uiDRAMBytes;            ///< Bytes of DRAM
        uint_32 uiNonVolatileAppBytes;  ///< Bytes of NVRAM available to app
        char cBootBlockVersion[VERSION_LENGTH];
        char cApplicationVersion[VERSION_LENGTH];
        char achIPAddress[4];     ///< IP address for ethernet comm
        char achSubnetMask[4];    ///< IP subnet mask for ethernet comm
        BACKPLANEID BackplaneID;  ///< I2C Bus Backplane & Board IDs
        BOARDID BoardID[NUM_BOARD_SLOTS];
    };

    struct SQueryArmInfo
    {
        uint_32 eInitState;  ///< EInitState of the application.
        uint_32 eBooted;     ///< EBootType flag: who is running
        uint_32 uiAppCRC16;
        char cBootBlockVersion[VERSION_LENGTH];
        char cApplicationVersion[VERSION_LENGTH];
        char achIPAddress[4];   ///< IP address for ethernet comm
        char achSubnetMask[4];  ///< IP subnet mask for ethernet comm
    };

    struct SQuery643Info
    {
        SQueryArmInfo sArmInfo;
        BACKPLANEID BackplaneID;  ///< I2C Bus Backplane & Board IDs
        BOARDID BoardID[NUM_BOARD_SLOTS];
    };

protected:
    /// These are the command id's for the various commands handled by the classes..
    enum ECommand
    {
        ceReset = 0,
        ceResetRcv,

        ceExecute,
        ceExecuteRcv,

        ceQuery,
        ceQueryData,

        ceDownload,
        ceDownloadBlockReq,
        ceDownloadBlock,
        ceDownloadFin,

        ceSetIPData,

        ceQueryError,
        ceQueryErrorData,  ///< response from embedded with error info
        ceClearError,      ///< Clear errors reported by ceQueryError

        ceQueryArmData,
        ceQueryErrorArmData,  ///< response from embedded with error info

        ceQuery643Data,

        ceGetSetSessionID,  ///< Request the session ID; if it has not been set, use the one provided.
        ceRcvSessionID,     ///< Return the session ID.

        ceStartupLast
    };

    /// The possibilities for the results of attempting to download an application or boot block.
    enum EDownloadResult
    {
        ceDownloadOK,
        ceDownloadCannotProgram,
        ceDownloadBadFile,  ///<  Front end only.
        ceDownloadProgramFailed,
        ceDownloadCRCFailed
    };

    /// This is sent as part of the ceDownload command to indicate the size and type of download.
    struct SDownloadInfo
    {
        uint_8 eDownloadType;  ///<  The type of download.
        uint_32 uiFileSize;    ///<  The number of bytes to download.
    };

    ///  Sent as part of the ceDownloadBlockReq message to indicate which data to download.
    struct SDownloadBlockReq
    {
        uint_32 uiStart;
        uint_32 uiLength;
    };

    ///  Sent as part of the ceDownloadFin command to indicate any errors.
    struct SDownloadResult
    {
        uint_8 eResult;  //  The result of the download.
    };

    struct SIPData
    {
        char achIPAddress[4];   ///< IP address for ethernet comm
        char achSubnetMask[4];  ///< IP subnet mask for ethernet comm
    };

    struct SErrorInfo
    {
        uint_8 ubUnhandledInts;     ///< count of unhandled interrupts
        uint_16 uiLastIntVector;    ///< vector number of last unhandled interrupt
        uint_8 ubStackOverflows;    ///< count of stack overflow errors
        uint_8 ubProtectionFaults;  ///< count of protection fault interrupts
        uint_32 ulFaultAddr;        ///< Address of last protection fault
        uint_8 ubMathErrors;        ///< count of floating point math errors
        uint_32 ulMathAddr;         ///< Address of last math error fault
        uint_8 ubDivZeroErrors;     ///< count of divide by zero errors
        uint_32 ulDivZeroAddr;      ///< Address of last divide by 0 fault
    };

    struct SErrorArmInfo
    {
        uint_32 ulunused;
    };

    CCommStartupBase(CCommDataLink* pLink)
        : CCommTransport(pLink, ceTransStartup, false)  // FE should not connect yet
    {
    }
#ifdef MIC_VCMFC
#pragma pack()
#endif
};

#endif  // COMSTRTB_HPP
