#ifndef COMSTRTF_HPP
#define COMSTRTF_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "comstrtb.hpp"
#include <cpp/warningpragmas.hpp>

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------

// ================================================================================================
///  This class that forms the Front end Startup/initialization transport layer.
class CCommStartupFE : public CCommStartupBase, private CCommMonitor
{
    friend class CWaitSocketCom;

public:
    CCommStartupFE(CCommDataLink* pLink);
    virtual ~CCommStartupFE();

    /// Should be called to initialize the comm subsystem when this is first created. This will
    /// result in a call to the InitDone() method.
    void Init();

    /// This will query the back end for its query information. This will result in the QueryData()
    /// function being called.
    void Query();

    /// This will send the reset command to the back end. This will result in the resetting of the
    /// back end and execution from the boot block, so the datalink will be reset. After this
    /// happens the ResetDone() command will be called.
    void Reset();

    /// This will send the execute command to the back end. This will result in the resetting of the
    /// back end and execution from the application, so the datalink will be reset. After this
    /// happens the ExecuteDone() command will be called.
    void Execute();

    /// This will query the back end for its error information. This will result in the
    /// QueryErrorData() function being called.
    void QueryError();

    /// Clear any errors recorded in the embedded system
    void ClearError();

    /// @{
    /// This will handle the details of downloading the file to the back end. When the download is
    /// complete, the DownloadDone() method is called.
    void DownloadApp(const _TCHAR* szFilename) { DownloadFile(szFilename, ceApplication); }
    void DownloadBoot(const _TCHAR* szFilename) { DownloadFile(szFilename, ceBootBlock); }
    /// @}

    /// This will handle the details of downloading the file to the back end. When the download is
    /// complete, the DownloadDone() method is called.
    void DownloadFile(const _TCHAR* szFilename, EDownloadType eType);

    /// This will download the IP information to the mass. The input should be strings of the form
    /// n.n.n.n. If this is not the case, the function returns false and nothing is done, otherwise
    /// true is returned.
    bool DownloadIPData(const _TCHAR* szIPAddress, const _TCHAR* szSubnetMask);

protected:
    ///  This is called as a result of the Query() method when the back end sends the command.
    virtual void QueryData(const SQueryInfo& rInfo) = 0;

    /// This is called as a result of the Query() method when the back end sends the command.
    virtual void QueryArmData(const SQueryArmInfo& rInfo) = 0;

    /// This is called as a result of the Query() method when the back end sends the command.
    virtual void Query643Data(const SQuery643Info& rInfo) = 0;

    /// Called when a download begins
    virtual void DownloadBegin(const _TCHAR* szFilename, SDownloadInfo SDInfo) {}

    /// Called when a download block request is received
    virtual void DownloadingBlock(SDownloadBlockReq SBInfo) {}

    ///  Called when a DownloadApp(), or DownloadBoot() is completed.
    virtual void DownloadDone(EDownloadType eType, EDownloadResult eResult) {}

    ///  Called whenever the Reset operation is finished.
    virtual void ResetDone() = 0;

    ///  Called whenever the Reset operation is finished.
    virtual void ExecuteDone() = 0;

    ///  Called whenever the initial initialization operation is done.
    virtual void InitDone() = 0;

    /// Handle notification that the session ID was sent received from the backend.
    virtual void OnRcvSessionID(uint_32 uiSessionID) = 0;

    virtual void QueryErrorData(const SErrorInfo& rInfo) = 0;
    virtual void QueryErrorArmData(const SErrorArmInfo& rInfo) = 0;

    /// Request the backend session ID. This ID is used to determine whether the backend
    /// has restarted since the last successful connection.
    void ReqSessionID();

    // CCommTransport implementation.
    virtual void MessageSent(ID uiID, const char* pData, SIZE uiSize) override;
    virtual void MessageReset(ID uiID, const char* pData, SIZE uiSize) override;
    virtual char* GetMemory(ID uiID, SIZE uiSize) override;
    virtual void MessageRecieved(ID uiID, char* pData, SIZE uiSize) override;
    virtual void DiscardMemory(ID uiID, char* pData, SIZE uiSize) override;

    /// @{
    /// The information about the back end for the last query & error request.  These have been
    /// moved from private to give access to the derived class CInstInitStartup so that its
    /// duplicate members of the same structs can be eliminated.
    SQueryInfo m_sQueryInfo;
    SQueryArmInfo m_sQueryArmInfo;
    SErrorInfo m_sErrorInfo;
    SErrorArmInfo m_sErrorArmInfo;
    /// @

private:
    /// This is called by CWaitSocketCom() when its timer expires to finish
    /// initialization after waiting long enough for SocketCom to be initialized.
    void FinishInit();

    /// Called when the mass requests another block to download.
    void OnDownloadBlockReq();

    /// Called when the mass is finished requesting blocks to download.
    void OnDownloadFin();

    //  CCommMonitor Implementation.
    virtual void InitFinnished() override;
    virtual bool CommFailed(int uiRetryCount) override { return true; }

    /// This waits to see if SocketCom connection is made before finishing Init().
    CWaitSocketCom* m_pcWaitSocketCom;
    CCommDataLink* m_pLink;  ///<  The associated data link to reset.

    /// @{
    /// These flags indicate which command caused the initialization to proceed and
    /// whether it is finished.
    bool m_fReset;
    bool m_fExecute;
    bool m_fInitFinished;
    /// @}

    SDownloadInfo m_DownloadInfo;      ///< The information about the file to download.
    SDownloadBlockReq m_BlockInfo;     ///< The information about the next block to download.
    SDownloadResult m_DownloadResult;  ///< The information about the success of the download.
    int m_iFile;                       ///< The file handle for the file to download.
    char* m_pBlockMem;                 ///<  Where to store the data to download.
    SQuery643Info m_sQuery643Buff;  ///< buffer for storing 643 query before copying to other structs

    SIPData m_sIPData;          ///<  The data to hold the download information.
    uint_32 m_uiSessionIDRecv;  ///< Buffer for the received backend session ID. Initialized to zero.
    /// Buffer for the initial session ID sent to the backend. Implemented using as a
    /// timestamp.
    const uint_32 m_cuiSessionIDSent;

    DECLARE_MSG()
};

#endif  // COMSTRTF_HPP
