#ifndef COMTCPW_HPP
#define COMTCPW_HPP
//-------------------------------------------------------------------------------------------------
#if !defined(MIC_VCMFC)
#error This can only be used under windows.
#endif

//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "comphys.hpp"
#include <cpp/tm1shot.hpp>
#include <winsock.h>
#include <array>

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
class CCommPhysicalTCP;
class CCommPhysicalTCPNotify;
// ===============================================================================================
class CCommPhysicalTCP final : public CCommPhysical
{
    friend LRESULT CALLBACK s_TCPProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

public:
    /// Initializes this object to the idle state.
    CCommPhysicalTCP(CCommPhysicalTCPNotify* pNotify = 0);
    /// This will clean up the object and diconnect the class.
    ~CCommPhysicalTCP();

    /// Initializes the instrument model & serial number, IP address and port
    void Init(const _TCHAR* achModel,
        const _TCHAR* achSerial,
        const _TCHAR* szIPAddr,
        int iPort,
        const _TCHAR* szIPNICAddr);

    /// Broadcasts the instrument model, serial, and IP address, then connects
    void InstConnect();

    ///  This will broadcast the instrument model number, serial number, and IP address. The
    ///  instrument must receive this message before it will begin listening for a connection.
    void BroadcastInstAddr(char* achModel, char* achSerial, char* szIPAddr);

    ///  This will start the connection process. Then the class is connected, this will call the
    ///  Connected() method in the associated pNotify class (if any).
    void Connect(int iPort, const char* szIP);

    ///  This will setup this link to listen on that channel. Once the other end connects to that
    ///  channel, it will call the Connected() method in the associated pNotify.
    void Listen(int iPort);

    /// Remove the notification from the TCP class.
    void Remove(CCommPhysicalTCPNotify* pNotify);

    // The CCommPhysical overrides.
    virtual void Reconnect() override;
    virtual void Send(const char* pHead1,
        int uiHead1Size,
        const char* pHead2,
        int uiHead2Size,
        const char* pData,
        int uiDataSize) override;
    virtual void ResetSend() override;
    virtual bool IsFatalComError() override;

    virtual void FillDebugStatus(char* szStr) const override;
    virtual CStr GetInstAddr() const override;

private:
    enum EConnectionState
    {
        ceIdle,
        ceBroadcasting,
        ceConnecting,
        ceConnected,
        ceListening,
        ceForceReconnect,
        ceRetryBroadcast
    };

    /// This is called when the connection fais or should be terminated.  It will cleanup any data,
    /// and notify the associated object.
    void Disconnect();

    /// This is called whenever it is appropriate to send more data over the channel. If the last
    /// byte of the block is sent, this will call the NotifyDataSent() method to indicate that it is
    /// ready to send more information.
    void OnSendData();

    /// This is called whenever there is more data to be received from the TCP/IP link. When a full
    /// packet/message is received, the NotifyDataAvailable() method is called to indicate that a
    /// new message is present.
    void OnReceiveData();

    ///  This is called whenever there is activity on the TCP channel that This class needs to be
    ///  aware of.
    void OnTCPIP(LPARAM lParam);

    ///  This is called when a timeout occurs.
    void TimeExpired();

    CCommPhysicalTCPNotify* m_pNotify;  ///<  The class to notify when events happen here.
    std::array<char, 4> m_achModel;     ///< The instrument model number
    std::array<char, 4> m_achSerial;    ///< The instrument serial number
    std::array<char, 32> m_szIPAddr;  ///< The IP address that is used to connect to the instrument.
    int m_iPort;                      ///< The port or channel number to connect
    std::array<char, 32>
        m_szIPNICAddr;  ///< The IP address of the network card that will connect to the instrument(s)
    std::array<char, sizeof(uint_16) + ceMaxPacketSize>
        m_aSendBuff;    ///<  The buffer that stores the packet to send.
    uint_16 m_uiStart;  ///<  The starting index of the next block to send.
    uint_16 m_uiSize;   ///<  The number of bytes in the packet stored.
    char* m_aRcvBuff;   ///< The buffer to receive to.

    ///  The length of the buffer, The first 2 bytes are received directly into here. These lengths
    ///  include these two bytes.
    uint_16 m_uiLength;

    ///  Contains the number of bytes received so far. This includes the length bytes.
    uint_16 m_uiNumRcv;

    SOCKET m_Socket;             ///<  The cocket over which this object connects.
    static SOCKET ms_UDPSocket;  ///<  The UDP socket used for broadcasting instrument addresses.
    static struct sockaddr_in ms_UDPsin;  ///<  The socket address struct to use for UDP
    CTimeOneShot m_cTimer;                ///< The object that handles the time-out timing.
    unsigned int m_uiMsgID;               ///< The message ID for the TCP events for this object.
    int m_iBroadcastRetry;      ///< The number of attempts to broadcast the instrument address.
    EConnectionState m_eState;  ///<  The state of the connection.

    /// @{
    /// The linked list of available tcp objects. This is sorted by next expiration.
    static CCommPhysicalTCP* ms_pcHead;
    CCommPhysicalTCP* m_pcNext;
    /// @}

    int dm_iSendError;
    int dm_iRecvError;

    DECLARE_MSG()
};

// ================================================================================================
/// This class is the base class for objcts that would like to receive notification of the status on
/// the Physical TCPIP layer.
class CCommPhysicalTCPNotify
{
public:
    enum EError
    {
        ceInvalidSocket,
        ceCannotBind,
        ceCannotListen,
        ceCannotAccept,
        ceBadIp
    };

    virtual ~CCommPhysicalTCPNotify() {}

    ///  This is called whenever the connection is established.
    virtual void Connected(CCommPhysicalTCP* pcPhys) {}

    ///  This is called whenever the connection was unable to be established.
    virtual void Error(CCommPhysicalTCP* pcPhys, EError eError) {}

    ///  This is called whenever the physical connection is broken.
    virtual void Disconnected(CCommPhysicalTCP* pcPhys) {}
};

#endif  // COMTCPW_HPP
