#ifndef COMTRNS_HPP
#define COMTRNS_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#if defined(MIC_VCMFC)
#include <cpp/warningpragmas.hpp>
#include "tcpterm.hpp"
#include <cpp/tm1shot.hpp>
#endif
#include "comdtlnk.hpp"

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
#if defined(MIC_DM643)
class CCommTransRecvTask;
class CCommPhysicalTCP;
class CTCPTermDM643;
#endif

// ================================================================================================
/// This defines the facilities of transporting messages over a channel of the comm link. Each
/// message consists of a unique ID numbered 0 to 0x7E followed optionally by a block of data.
///
/// This class is meant to be a base class for classes that define and implement a specific channel
/// for some application. The derived class adds knowledge about the specific messages that are to
/// be sent and received by this class. This class defines the support functions for sending and
/// receiving messages.
///
/// Even though each message can be up to 2^32 bytes long. Messages under a packet(1K) in length do
/// not explicitly send the 4 bytes of length information. The full 4 byte length is only sent id
/// the data is over a packet in length.
#if defined(MIC_VCMFC)
class CCommTransport : public CCommDataLinkChannel, private CTCPTermFE::CObserver
#else
class CCommTransport : public CCommDataLinkChannel
#endif
{
public:
    typedef uint_8 ID;     ///< The type for the messageID.
    typedef uint_32 SIZE;  ///< The type for the message size.

    enum
    {
        ceTransStartup = 0,  ///< The transport layer ID for the startup channel.
        ceTransFirstFree
    };

    enum
    {
        /// The maximum size if the ID and size bytes.
        ceMaxHeaderSize = sizeof(ID) + sizeof(SIZE),

        /// This is the maximum packet size that this class will break the larger messages
        /// into. This is also the threshold between SMALL and other messages. It should really turn
        /// out that this is a nice number like 1024. since this is the most visible to the user.
        ceMaxPacketSize = static_cast<int>(CCommDataLink::ceMaxPacketSize) - ceMaxHeaderSize,

        /// The maximum number of messages that can be queued to send.
        ceMaxQueue = 64,

        /// The number of entries in the outgoing circular queue.
        ceMaxQueueSize = ceMaxQueue + 1,

        /// The maximum number of buffers that will be requested for a 'non-sequential' message.
        ceMaxAlloc = 2,

        /// The entries in the circular buffer queue.
        ceMaxAllocSize = ceMaxAlloc + 1
    };

#if defined(MIC_DM643)
    /// define the signature of the terminal creation function.
    // Note: no using nor std::function available in dm643.
    typedef CTCPTermDM643* (*FnCreateDM643Term)(int iID);

    /// Set the creation function for the dm643 terminal.
    static void SetTermCreationFcn(FnCreateDM643Term fnCreate);
#endif

    virtual ~CCommTransport();

    /// @returns true iff Connect() has been called and no FD_CONNECT has been handled.
    bool IsConnecting() const { return m_fIsConnecting; }

protected:
    ///  Create this transport layer on the datalink layer.
    CCommTransport(CCommDataLink* pSrc, int uiID, bool fConnect = true);

    /// This will return true iff there is no more room in the send queue for additional messages.
    bool IsSendQueueFull();

    /// This will return true iff there are no messages in the send queue.
    bool IsSendQueueEmpty();

    /// This will return the number of items that can still be queued.
    uint_16 SendQueueRoom();

    /// This will flush all the items from the queue that are not in the process of being sent.
    void FlushSendQueue();

    /// This will send the message over the datalink. The message data must continue to exist until
    /// the message is sent. When the message is sent, this will call the MessageSent() virtual
    /// function. It could happen that the data link is reset before this message is sent, in which
    /// case the MessageReset() function will be called.
    //
    /// If pData is 0 and uiSize is not, it is assumed that the data is not available completely in
    /// memory. In this case, The GetNextPacket() function will be called repeatedly to receive data
    /// for the message as necessary. The MessageSent() or MessageReset() method will still be
    /// called when the message is done. This is the mode of sending that will be used for
    /// downloading the application.
    void SendMessage(ID ubID, const char* pData, SIZE uiSize);

    /// This gets called once the message has been successfully received by the datalink layer. There
    /// is, however, no guarantee that the message was properly interpreted and processed by the
    /// other end.
    virtual void MessageSent(ID ubID, const char* pData, SIZE uiSize) {}

    /// This gets called when a message was in the middle of being sent, and the data-link was reset
    /// or initialized. The message will not be resent unless this requests a resend.
    virtual void MessageReset(ID ubID, const char* pData, SIZE uiSize) {}

    /// This is called when this is ready for the next packet of information for the given ID. It
    /// should fill in the next uiPacketSize bytes of information into an 'allocated' buffer which
    /// it should return.  The buffer must remain valid until the DiscardPacket() is called for that
    /// buffer. No more than ceMaxAllocSize of these 'allocated' buffers will be required at any
    /// time (i.e. GetNextPacket() could be called for a second time before the first call to
    /// DiscardPacket()). The call to DiscardPacket() is made before the call to MessageSent() or
    /// MessageReset(). The uiPacketSize is always lessthan or equal to ceMaxPacketSize.
    virtual const char* GetNextPacket(ID ubID, SIZE uiPacketSize) { return 0; }

    /// This signals that the object is done with the pPacket and this memory can be 'deallocated'
    virtual void DiscardPacket(ID ubID, const char* pPacket) {}

    /// This is called when the first packet of a message is received to 'allocate' the memory of
    /// the message. This memory is used internally until either the message is complete and
    /// MessageRecieved() is called, or the datalink is reset and DiscardMemory() is called.
    //
    /// If you do not want the whole message placed in a single memory buffer, this should return
    /// NULL, in which case whenever the next packet becomes available the NextPacketReceived()
    /// function is called.  The MessageRecieved(), or DiscardMemory() functions are still called at
    /// the end. This mode can be used for 'pass-through' mode where the receiver just has to pass
    /// the data to another process(or).
    virtual char* GetMemory(ID ubID, SIZE uiSize) { return 0; }

    /// This is called when a message has been sucessfully received. If the message was of zero size,
    /// pData will be null, and no GetMemory() call was made.
    virtual void MessageRecieved(ID ubID, char* pData, SIZE uiSize) {}

    /// This is called when the datalink was reset in the middle of receiving a message that
    /// required a call to GetMemory(). This allows the 'allocated' memory to be deallocated.
    virtual void DiscardMemory(ID ubID, char* pData, SIZE uiSize) {}

    /// This is called whenever a packet of a message that did not allocate a buffer is
    /// received. The data pPacket is only available for the duration of this call.
    virtual void NextPacketReceived(ID ubID, const char* pPacket, SIZE uiPacketSize) {}

    // CCommDatalinkChannel implementation.
    virtual void ResetChannel() /*override*/;
    virtual void Connect() /*override*/;

#if defined(MIC_DM643)
    /// Resize input buffer
    void ResizeBuffer(uint_32 uiSize);
#endif

private:
#if defined(MIC_DM643)
public:  // compiler bug workaround: The dm643 compiler will not allow usage of this type on the cpp
         // file unless it is public.
#endif

#if defined(MIC_VCMFC)
#pragma pack(1)
#endif
    /// SHeader is 1 byte ID + 4 byte SIZE: does not align for DM643
    struct SHeader
    {
        ID ubID;
        SIZE uiSize;
    };
#ifdef MIC_VCMFC
#pragma pack()
#endif

#if defined(MIC_DM643)
private:
#endif

#if defined(MIC_VCMFC)
#pragma pack(1)
#endif
    struct SSendEntry
    {
        SHeader Header;  ///< The header information of the datablock to be sent.
        ///  The header information packed into a buffer to be sent.
        char acHeadBuff[sizeof(ID) + sizeof(SIZE)];

        const char* pData;  ///<  The data block that is to be sent.

        /// The number of bytes of the message that have been sent so far.  If this is (SIZE)(-1) it
        /// indicates that the header was not even started.
        SIZE uiSent;

        int uiPending;  ///<  The number of packets that are pending to be confirmed.

        /// The buffers that are used to hold the pieces of information for a non sequential
        /// message.
        const char* apAllocBuff[ceMaxAllocSize];

        /// The next location in apAllocBuff to fill. uiNextAlloc-uiPending (circularly) is the next
        /// one to be confirmed.
        int uiNextAlloc;
    };
#ifdef MIC_VCMFC
#pragma pack()
#endif

    // CCommDatalinkChannel implementation.
    virtual void DataSent() /*override*/;
    virtual void DataReceived(const char* pData, int uiSize) /*override*/;
    virtual bool HasDataToSend() const /*override*/;
    virtual void GetDataToSend(const char** ppHead,
        int* puiHeadSize,
        const char** ppData,
        int* puiDataSize) /*override*/;
    virtual void FillDebugStatus(char* szStr) const /*override*/;

#if defined(MIC_VCMFC)
    // CTCPTermFE::CObserver overrides.
    virtual void OnSent() /*override*/;
    virtual void OnRecv(const VChar& vchData) /*override*/;
    virtual void OnConnect() /*override*/;
    virtual void OnDisconnect() /*override*/;
    virtual void OnError(EError eError) /*override*/;
    virtual void OnErrorMsg(const CStr& cszError) /*override*/;

    CTCPTermFE* m_pcTerm;  ///< CTCPTerm used by SocketCom instead of CCommDataLink

#elif defined(MIC_DM643)
    CTCPTermDM643* m_pcTerm;  ///< CTCPTerm used by SocketCom instead of CCommDataLink

    /// access control for ensuring SendMessage puts calls are atomic
    SEM_Handle m_hSendAccess;

    /// task for getting received data
    friend class CCommTransRecvTask;

    CCommTransRecvTask* m_pcRecvTask;

    /// The function that creates the terminal for this framework.
    static FnCreateDM643Term ms_fnTermCreation;
#endif

    /// true iff Connect() has been called and no FD_CONNECT has been handled.
    bool m_fIsConnecting;

    /// The transport layer ID for this object.
    int m_uiTransID;

    SSendEntry m_aSendQueue[ceMaxQueueSize];

    //  The (circular) queue of outgoing messages.
    /// The next location in the send queue that is to be filled by an outgoing message. If this is
    /// one before m_uiCurrentItem, the queue is full.
    int m_uiNextQueueItem;

    /// The location in the send queue of the current item that is being sent. If this is equal to
    /// m_uiNextQueueItem, there is nothing in the send queue.
    int m_uiCurrentItem;

    // Support for receiving data.
    bool m_fReceiving;    ///< True iff this is in the middle of receiving a message.
    SHeader m_RcvHeader;  ///< The header of the currently receiving header.
    char* m_pRcvData;     ///< The data item containing the data.
    SIZE m_uiReceived;    ///< The number of bytes received so far.

#if defined(MIC_DM643)
    Semaphore m_SendSem;  ///< The semaphore that protects the send queue.
#endif

    DECLARE_MSG()
};

/// ===============================================================================================
#if defined(MIC_DM643)
/// The interface to the dm643 required terminal functionality.
class CTCPTermDM643
{
public:
    /// Server init; initializes the object and starts the recv and send tasks as
    /// server tasks. Listens for incoming client connections from the frontend
    /// (CCommPhysicalTCP instance provided). Compare to InitAsClient.
    /// @param cpTCP Used to query for TCP system initialization and local IP address.
    /// @param uiPort Port to listen to for client connections.
    /// @param cTimeOut @see InitAsClient.
    /// @param fBufferInput @see InitAsClient.
    virtual void InitAsServer(const CCommPhysicalTCP* cpTCP,
        uint_16 uiPort,
        CTimeInterval cTimeout = CTimeInterval::Sec(0),
        bool fBufferInput = true)
        = 0;

    /// resize buffer
    virtual void Resize(uint_32 uiSize) = 0;

    virtual uint_32 puts(const char* buff,
        int_32 nChar = -1,
        UNSIGNED suspend = NU_SUSPEND,
        STATUS* pStatus = NULL)
        = 0;

    /// @returns True if connected (has valid socket), false otherwise.
    virtual bool IsConnected() = 0;

    /// Request for unbuffered receive.
    /// @return true iff request was successfully completed.
    virtual bool ReqUnbuffRecv(char* pchBuff, int iNumChars, uint_32 uiTimeout) = 0;

    /// Closes the connection and resets the char buffers.
    virtual void Reset() = 0;
};
#endif

/// ===============================================================================================
#if defined(MIC_DM643)
/// CCommTransRecvTask : Task for getting received data
class CCommTransRecvTask : public Task
{
public:
    CCommTransRecvTask(CCommTransport* pTrans);

    /// Notify recv task to discard previous message and reset state
    void Reset();

private:
    struct SRecvState
    {
        bool fGetHdr;
        uint_8 aubHdr[sizeof(CCommTransport::ID) + sizeof(CCommTransport::SIZE)];
        char* pBody;
        char* pNext;
        uint_32 uiBytesExpected;
        uint_32 uiWaitMSecs;
    };

    /// Set SRecvState to start a new message
    void InitState();

    /// Task body
    VOID Entry();

    CCommTransport* m_pTrans;
    SEM_Handle m_hSEM_Reset;
    SRecvState m_sState;

    DECLARE_MSG()
};
#endif  // MIC_DM643

#endif  // COMTRNS_HPP
