#ifndef CONSTANT_HPP
#define CONSTANT_HPP
//-----------------------------------------------------------------------------
// Library: baselib
// Description: Useful constants
//-----------------------------------------------------------------------------
/* Recommendation:
**      Once all reports have been verified against other instruments and are
**      operating correctly, extend the precision of these constants to the
**      maximum representable by type "double", or to its known limit of
**      accuracy, whichever occurs first for each constant.
**
**      Since this is intended to be library code, and since the software
**      is becoming increasingly decoupled from the hardware, we should
**      support future expanded instruments/computing platforms to the
**      maximum extent possible.
*/

// These defines are used for report generation
namespace NConstants
{
    const float STD_TEMP                  = 273.15f;     // standard temperature (kelvin)
    const double VOLGASTP                 = 22414.0;     // volume (cc) of 1 mole of gas at STP
    /// The Avogadro constant (1/mol) from 2010 CODATA recommended values
    const double AVOGADRO                 = 6.02214129e23;
    const double R                        = 8.31441e7;   // Gas constant (erg/K/mol)
    const double R2                       = 8.31441e-3;  // Gas constant (kJ/K/mol)
    const double PI                       = 3.14159265;  // value for pi
    const double M_CM                     = 1.0E-2;      // meters per centimeters
    const double M_A                      = 1.0E-10;     // meters per Angstrom
    const double A_M                      = 1.0E10;      // Angstroms per meter
    const double M_NM                     = 1.0E-9;      // meters per nanometer
    const double NM_M                     = 1.0E+9;      // nanometers per meter
    const double NM2_M2                   = 1.0E18;      // square nanometers per square meter
    const double CM2_A2                   = 1.0E-16;     // square centimeters per square Angstrom
    const double CM2_NM2                  = 1.0E-14;     // square centimeters per square nanometer
    const double M3_CC                    = 1.0E-06;     // cubic meters per cc
    const double CM3_M3                   = 1.0E6;       // centimeters cubed per meter cubed
    const double SPSATARGET               = 0.30;        // target relative pressure for single pt SA
    const double SPSACONST                = 4.35;        // single pt SA calculation constant

    // These defines are used for Thermal Transpiration correction
    const double THERMCON                 = 2.33;
    const double ALPHA                    = 0.033;
    const double BETA                     = 0.245;
    const double G                        = 2.50;
    const double H                        = 2.00;

    // These defines were taken from OPERATE.H
    const double STD_PR                   = 760.00;      // standard pressure (mmHg)
    const double ROOM_TEMP                = 298.00;      // approximate room temperature (kelvin)

    // Standard Temperature (kelvin).  Currently used by the ASAP2010 Status/Display
    const double CELCIUS_TO_KELVIN_OFFSET = STD_TEMP;

    /// Minutes to milliseconds conversion constant.
    const double cdMinuteToMS = 60. * 1000.;

    /// The maximum number of entries for target pressure and collected data tables in gas
    /// adsorption applications.  This really does not belong here, but it is
    /// used across multiple products and even some product line libraries.
    /// When it needs to change for a product that one should make it's own and
    /// eventually this will no longer be used.
    const int ciMaxPressureTableRows = 2000;

}

#endif // not CONSTANT_H
