#ifndef CSTR_HPP
#define CSTR_HPP
//-------------------------------------------------------------------------------------------------
/// @file
/// There are the following levels of functionality provided for CStr (and support) in this file:
/// - DM643: This does not support the stl nor streams this assumes accii only, the implementation
///   is based on an allocated string array.
/// - TEST and embedded linux: This assumes utf-8 representation and uses the std::string
///   implementation.
/// - VCMFC: This assumes utf-16 (no multibyte support I think) and uses the CString internally.
/// For these we define one of the following: MIC_CSTR_SIMPLE, MIC_CSTR_STD, MIC_CSTR_CSTRING. There is
/// a further define for MIC_CSTR_GETTEXT for MIC_CSTR_STD configurations that support gettext
/// translations (I.e. cmake based embedded linux that does not define any of the MIC_XXX).
#if defined(MIC_VCMFC)
#define MIC_CSTR_CSTRING
#elif defined(MIC_TEST)
#define MIC_CSTR_STD
#elif defined(MIC_DM643)
#define MIC_CSTR_SIMPLE
#else
#define MIC_CSTR_STD
#define MIC_CSTR_GETTEXT
#endif

//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "basetype.hpp"
#if defined(MIC_CSTR_STD)
#include <cctype>
#include <algorithm>
#include <string.h>
#elif defined(MIC_CSTR_CSTRING)
#include "stdafx.h"
#endif

#if defined(MIC_CSTR_SIMPLE)
#include "stddef.h"
#include <string.h>
#else
#include <boost/format/format_fwd.hpp>
#include <string>
#include <iosfwd>
#include <map>
#include <vector>
#endif

#if defined(MIC_CSTR_GETTEXT)
#include <libintl.h>
#endif

// ================================================================================================
class CStr;

#if !defined(MIC_CSTR_SIMPLE)
class CDataArchive;
#endif

bool operator==(const CStr& cStr1, const CStr& cStr2);
bool operator!=(const CStr& cStr1, const CStr& cStr2);
bool operator==(const _TCHAR* szStr1, const CStr& cStr2);
bool operator!=(const _TCHAR* szStr1, const CStr& cStr2);
bool operator==(const CStr& cStr1, const _TCHAR* szStr2);
bool operator!=(const CStr& cStr1, const _TCHAR* szStr2);

bool operator<=(const CStr& cStr1, const CStr& cStr2);
bool operator<(const CStr& cStr1, const CStr& cStr2);
bool operator<=(const _TCHAR* szStr1, const CStr& cStr2);
bool operator<(const _TCHAR* szStr1, const CStr& cStr2);
bool operator<=(const CStr& cStr1, const _TCHAR* szStr2);
bool operator<(const CStr& cStr1, const _TCHAR* szStr2);

bool operator>=(const CStr& cStr1, const CStr& cStr2);
bool operator>(const CStr& cStr1, const CStr& cStr2);
bool operator>=(const _TCHAR* szStr1, const CStr& cStr2);
bool operator>(const _TCHAR* szStr1, const CStr& cStr2);
bool operator>=(const CStr& cStr1, const _TCHAR* szStr2);
bool operator>(const CStr& cStr1, const _TCHAR* szStr2);

CStr operator+(const CStr& cStr1, const CStr& cStr2);
CStr operator+(const _TCHAR* szStr1, const CStr& cStr2);
CStr operator+(const CStr& cStr1, const _TCHAR* szStr2);

#if !defined(MIC_CSTR_SIMPLE)
void Serialize(CDataArchive& ar, CStr& szData);
void Serialize(CDataArchive& ar, CStr& szData, size_t uiFixedChar);
#endif

#if defined(MIC_CSTR_GETTEXT)
/// The gettext shortcut.
inline const char* _(const char* cszStr)
{
    return gettext(cszStr);
}

/// Setup the locale based on the strings passed in.
void SetTranslationLocale(const char* cszDomain, const char* cszLocaleLocation, const char* cszLocale);
#endif

#if !defined(MIC_CSTR_SIMPLE)
namespace NStr
{
    /// Append the provided string with the provided separator between.
    /// @tparam Iterable std container of CStr type.
    /// @param cStrings Container of strings to join.
    /// @param cszSep Separator string.
    /// @param fIgnoreEmpty If true, empty strings are ignored, otherwise they are joined.
    /// @return Joined new string. The first string in container is returned if container
    ///     size is one. Empty string if container size is zero.
    template<typename T>
    CStr Join(const T& cStrings, const CStr& cszSep, bool fIgnoreEmpty = false);

    /// Parse specified string into a list of words
    /// @param cszIn is the string to parse
    /// @param cbSepCharList are the word separators such as " .,\n"
    std::vector<CStr> Tokenize(const CStr& cszStr, const CStr& cszSep);

    /// Split the  string at the separator character.
    /// @param cszIn is the string to parse
    /// @param cbSepCharList are the word separators such as " .,\n"
    std::vector<CStr> Split(const CStr& cszStr, _TCHAR chSep);
}  // namespace NStr
#endif  // !defined(MIC_CSTR_SIMPLE)

// ================================================================================================
/// A cross-platform string class. Provides common string functionality and
/// interoperability with different underlying character encodings and platform-specific
/// string types. Where possible, members return a reference to the CStr instance to
/// allow chaining.
///
///  MIC_CSTR_CSTRING: Wide charaters (utf-16, mostly igroring 32 bit chars.)
///  MIC_CSTR_SIMPLE: Assumes ascii.
///  MIC_CSTR_STD: utf-8.
class CStr
{
public:
    /// Create an empty String.
    CStr();
    /// Create a copy of the \p cStr
    CStr(const CStr& cStr);
    /// Create a c opy of the null terminated raw c string.
    CStr(const _TCHAR* szString);

    /// Cleanup resources.
    virtual ~CStr();

    /// Allow conversation to the (constant) underlying raw null terminated array.
    operator const _TCHAR*() const { return GetStringPtr(); }

    /// Assign from raw null terminated C array.
    CStr& operator=(const _TCHAR* szString);
    /// Assign from CStr
    CStr& operator=(const CStr& cStr);
    /// Append from raw null terminated C array.
    CStr& operator+=(const _TCHAR* szString);
    /// Append from CStr
    CStr& operator+=(const CStr& cStr);

    /// Get the underlying raw char*.  On MIC_VCMFC this will encode the utf-16
    /// string to utf-8.
    const char* GetAsciiPtr() const;
    /// Get the underlying _TCHAR sting.
    const _TCHAR* GetStringPtr() const;

    /// @returns True iff the string is zero length.
    bool IsEmpty() const;

    /// Get the length (number of _TCHARs) of the current string.
    size_t GetLength() const;

    /// Makes string 0 size
    CStr& Empty();
    /// Makes string 0 size
    CStr& Clear();

    /// Compares two strings (raw)
    int Compare(const CStr& cStr) const;
    /// Compares two strings (raw)
    int Compare(const _TCHAR* pszSrc) const;
    /// Compares two strings (raw)
    int Compare(const CStr& cStr, int iStartPos) const;
    /// Compares two strings (raw)
    int Compare(const _TCHAR* pszSrc, int iStartPos) const;

    /// @returns the index of the first \p ch character starting at position \p
    /// iStartPos.  If \p ch is not found, -1 is returned.
    int Find(_TCHAR ch, size_t iStartPos = 0) const;

    /// Add Character of string to the end.
    CStr& Append(const _TCHAR ch);
    /// Add the string to the end.
    CStr& Append(const CStr& pszString);
    /// Add the raw C string to this, if \p pchEnd is given, only the part of
    /// \p pchStart before there.
    CStr& Append(const _TCHAR* pchStart, const _TCHAR* pchEnd = NULL);

    /// Append a string. If the string is not empty before appending, append the delimiter
    /// first.
    CStr& AppendDelimited(const CStr& cszString, const CStr& cszDelimiter);

    /// Delete all occurrences of \p szStr from this.
    void DeleteAll(const CStr& szStr);
    /// Convert the string to lower case.
    CStr& ConvertToLower();
    /// Convert the string to upper case.
    CStr& ConvertToUpper();

    static const CStr NULL_CSTR;

#if !defined(MIC_CSTR_SIMPLE)
    /// If available, create a string from the resource ID.
    explicit CStr(unsigned int uiResID);
    CStr(const boost::format& fmt);
    /// Copy from boost fomratter.
    CStr& operator=(const boost::format& fmt);
    /// Copy from utf-8 (as opposed to utf-16) raw pointer
    CStr& operator+=(const boost::format& fmt);

    CStr& Trim();
    CStr& TrimRight();
    CStr& TrimLeft();

    CStr Left(int iCount) const;
    CStr Right(int iCount) const;
    /// @returns String containing the given count of characters the given offset from the beginning
    /// of the string.  If count is -1, return from the given offset to the end of the string.
    CStr Substr(int iOff = 0, int iCount = -1) const;
    int Find(const _TCHAR* pszSrc, size_t iStartPos) const;
    int Find(const CStr& cStr, size_t iStartPos = 0) const;  ///< Find first substring

    /// Replace all occurences of \p szOld with \p szNew.
    int ReplaceAll(const CStr& szOld, const CStr& szNew);

    /// @returns True iff the string is zero length or contains only whitespace characters.
    bool IsBlank() const;
    /// (Non MIC_DM643): Compares two strings case sensitive
    int CompareNoCase(const CStr& cStr) const;
    /// (Non MIC_DM643): Compares two strings case sensitive
    int CompareNoCase(const CStr& cStr, int iStartPos) const;
    /// (Non MIC_DM643): Compares two strings case sensitive
    int CompareNoCase(const _TCHAR* pszSrc) const;
    /// (Non MIC_DM643): Compares two strings case sensitive
    int CompareNoCase(const _TCHAR* pszSrc, int iStartPos) const;

    /// Provide the CDataContainer interface to set the string to empty.
    CStr& SetDefaults();
    /// @param is Stream that is read from.
    /// @param iCharCount Number of characters to Read.
    /// @returns the string read.  It is always null terminated.
    CStr& ReadCharArray(std::istream& is, int iCharCount);
    /// @param os Stream that is written to.
    /// @param iCharCount Must be greater than 0. If CStr is longer than
    ///                   iCharCount, then the terminating NULL character
    ///                   will not be written.
    void WriteCharArray(std::ostream& os, size_t iCharCount) const;
    /// True iff a 16 bit style CStr can be read from the current stream
    /// position.  The Streamposition is moved back to the initial value in
    /// either case.
    static bool CanRead16(std::istream& rcStream);
    /// True iff a 32 bit CStr can be read from the current stream position.
    /// The Streamposition is moved back to the initial value in either case.
    static bool CanRead32(std::istream& rcStream);
    /// Read (and return) a string from the stream.  This can be in either 16
    /// or 32 bit format.
    CStr& Read(std::istream& is);
    /// Calculate the number of bytes that are necessary to store the sting in
    /// a stream.
    size_t GetSize() const;
    /// Returns sizeof string buffer in bytes needed to store string part of
    /// the Write
    size_t GetBufferSize() const;
    /// Write the string to the \p rcStream
    void Write(std::ostream& rcStream) const;
#endif

#if defined(MIC_CSTR_CSTRING)
    friend std::wostream& operator<<(std::wostream& os, const CStr& cStr);

    /// MIC_VCMFC specifc: Create a copy from the system CString
    CStr(const CString& cStr) : m_szString(cStr) {}
    /// MIC_VCMFC specifc: create string from the boost format mechanism.
    CStr(const boost::wformat& fmt);
    explicit CStr(const char* cszAsciiStr);
    /// MIC_VCMFC specifc: Allow conversion to access the internal CString.
    operator CString&() { return m_szString; }
    /// MIC_VCMFC specific: Access to the underlying CString object.
    CString& GetString() { return m_szString; }
    /// MIC_VCMFC specific: Access to the underlying CString object.
    const CString& GetString() const { return m_szString; }
    /// MIC_VCMFC specific: Copy from CString
    CStr& operator=(const CString& cStr)
    {
        m_szString = cStr;
        return *this;
    }
    /// MIC_VCMFC specific: Copy from boost fomratter.
    CStr& operator=(const boost::wformat& fmt);
    /// MIC_VCMFC specific: Copy from utf-8 (as opposed to utf-16) raw pointer
    CStr& operator=(const char* szString);
    /// MIC_VCMFC specific: Append from boost fomratter.
    CStr& operator+=(const boost::wformat& fmt);
    /// MIC_VCMFC specific: Method version of assignment.
    CStr& SetString(const char* szString);
    /// MIC_VCMFC specific: Method version of assignment.  Use at most \p uiLen characters.
    CStr& SetString(const char* szString, size_t uiLen);
    /// MIC_VCMFC specific: Method version of assignment.
    CStr& SetString(const CStringA& cStr);

    /// Remove line feeds and/or carriage returns at the end of the string.
    CStr& TrimLineEnd();

    int FindLast(_TCHAR c, size_t iStartPos = 0) const;              ///< Find last _TCHARacter
    int FindLast(const _TCHAR* pszSrc, size_t iStartPos = 0) const;  ///< Find last substring
    int FindLast(const CStr& cStr, size_t iStartPos = 0) const;      ///< Find last substring

    // Replace starting at iOff, return < 0 if error, else position of first pszOld after iOff
    int Replace(const CStr& szOld, const CStr& szNew, int iOff = 0);
    int Replace(_TCHAR chOld, _TCHAR chNew, int iOff = 0);
    int ReplaceAll(_TCHAR chOld, _TCHAR chNew);

    int Insert(int iIndex, _TCHAR ch);
    int Insert(int iIndex, const _TCHAR* psz);  ///< Inserts string or character at position
    CStr& Append(_TCHAR ch, int iCount);
    CStr& Append(const _TCHAR* pszString, int iCount);

    int Delete(const CStr& szStr, size_t iStartPos = 0);

    /// Copy characters to buffer including null terminator.
    /// @param pvBuf Destination buffer.
    /// @param uiLen Number of characters to copy or 0 to copy entire string.
    /// @return Characters copied.
    int CopyToBuf(_TCHAR* pvBuf, int uiLen = 0) const;
    /// Copy characters to buffer including null terminator.
    /// @param paBuf Destination buffer.
    /// @param uiLen Buffer size or 0 to copy entire string.
    /// @return Characters copied, max is (uiLen - 1) characters since last char is null.
    int CopyToBuf(char* paBuf, int uiLen = 0) const;

    _TCHAR GetAt(int iPos) const;      // Get _TCHARacter at given position
    CStr& SetAt(int iPos, _TCHAR ch);  ///< Sets a _TCHAR at a given position.

    size_t GetAllocLength() const;  /// <Return size of allocated memory for string
    CStr& Allocate(int nLength);    ///< Pre-allocate memory for string.
    int Delete(int iIndex, size_t iCount = 1);
#endif

#if defined(MIC_CSTR_STD)
    /// If not on MIC_DM643 nor MIC_VCMFC, also create a utf-8 string from a
    /// wide character array.
    CStr(const wchar_t* szStr);
    CStr(const std::string& szStr);
    /// Where available, Allow conversion to the the underlying std::string.
    operator std::string&() { return m_szString; }
    operator const std::string&() const { return m_szString; }
    /// The type for the look-up table for genstring dev-tool generated ID to
    /// string lookup.
    typedef const std::map<int, const _TCHAR*> TIDStrLookup;
    /// Setup the look-up table from ID to string.
    static void InitStringTable(std::initializer_list<const TIDStrLookup*> vpMap);
    static void InitStringTable(const TIDStrLookup& rmap);
#endif

private:
    /// Method version of assignment.
    CStr& SetString(const CStr& cStr);
    /// Method version of assignment.
    CStr& SetString(const _TCHAR* szString);

    /// Load the string based on the resource ID.  Either from the DLL string
    /// table (MIC_VCMFC) or from the CStr::InitStringTable() (others).
    void LoadResString(unsigned int uiResourceID);

#if defined(MIC_CSTR_CSTRING)
    // Unimplemented functions intended to catch accidental use of a resource ID argument
    // instead of a wide character.
    CStr& Append(int) = delete;
    CStr& Append(int, int iCount) = delete;
    int Find(int, size_t iStartPos = 0) const = delete;
    int FindLast(int, size_t iStartPos = 0) const = delete;
    int Insert(int iIndex, int) = delete;
    int DeleteAll(int, int iOff = 0) = delete;
    int Replace(int, int, int iOff = 0) = delete;
    int ReplaceAll(int, int) = delete;
    CStr& SetAt(int iPos, int) = delete;

    /// The actual implementation
    CString m_szString;

    ///  Used to cache the ascii version for GetAsciiPtr().
    mutable CStringA m_szASCII;
#elif defined(MIC_CSTR_STD)
    std::string m_szString;
#elif defined(MIC_CSTR_SIMPLE)
    enum
    {
        ceLoadMin = 128
    };

    char* m_szString;
    //  The actual string. If this is non null, it points to the first
    //  _TCHARater in the ascii string. The memory block allocated starts
    //  sizeof(size_t) bytes before this. Those bytes contain the reference
    //  count on the allocated string.

    size_t m_uStrLen;
    //  Current length of string. Not that this can be broken if the string
    //  returned by GetStrPtr() is adjusted.

    size_t m_uBufSize;
    //  Size of maximum length string that can be contained in this buffer.
    //  I.e. allocation size -1.
#else
#error One of the above needs to be defined.
#endif
};

#if !defined(MIC_CSTR_SIMPLE)
template<typename T> CStr NStr::Join(const T& cStrings, const CStr& cszSep, bool fIgnoreEmpty)
{
    typename T::const_iterator it = cStrings.begin();
    if(it == cStrings.end())
        return CStr();

    CStr szJoin = *it;
    if(fIgnoreEmpty)
    {
        for(++it; it != cStrings.end(); ++it)
            if(!it->IsEmpty())
                szJoin += cszSep + *it;
    }
    else
    {
        for(++it; it != cStrings.end(); ++it)
            szJoin += cszSep + *it;
    }
    return szJoin;
}
#endif  // !defined(MIC_CSTR_SIMPLE)

//-------------------------------------------------------------------------------------------------
#endif  // CSTR_HPP
