#ifndef DATA_VIEW_UPDATE_HPP
#define DATA_VIEW_UPDATE_HPP
//-----------------------------------------------------------------------------
/// @file
/// Defines the base class templates to use when you want to have a data ->
/// view link.  The TUpdateDataBase template creates a baseclass for the
/// CDataUpdate (template parameter) class.  It can be linked to zero or more
/// TUpdateViewBase derived objects.  Each TUpdateViewBase is liked to zero or
/// one TUpdateViewBase object.
///
/// The update notifications have a user defined unsigned long data parameter
/// to indicate what has changed.  A zero has special meaning that everything
/// changed.   There is also a (nested) freeze/restrore mechanism that caches
/// updates to be triggered when the last freeze is restored.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <set>

// ----------------------------------------------------------------------------
// Forward References
// ----------------------------------------------------------------------------
template<class TDATA> class TUpdateViewBase;

// ----------------------------------------------------------------------------
/// Acts as the serving class for the data. Maintains knowledge of all 'View'
/// classes that derive themselves from the partner TUpdateViewBase class. Links will
/// be internanlly updated if a 'View' object is destroyed or request removal
/// from this base class. Same is true for this 'Data' class, if it is destroyed,
/// the 'View' object's link to it will be removed.
///
/// UpdateView() is called to update all associated views.
///
/// FreezeViewUpdate() and RestoreViewUpdate() are called to stop and restart
/// the updates.  The class keeps an internal count so they must be called in
/// pairs, but they can also be nested.  The purpose is when there are multiple
/// updates to the data at once only a single Update will be called on the
/// views.
template<class TDATA> class TUpdateDataBase
{
public:
    friend TUpdateViewBase<TDATA>;

    TUpdateDataBase() {};
    TUpdateDataBase(const TUpdateDataBase<TDATA>&) = delete;
    virtual ~TUpdateDataBase()
    {
        // Destructor, removes all links.
        while(!m_aViewPtr.empty())
            RemoveLinkToView(*m_aViewPtr.begin());
    }
    TUpdateDataBase& operator=(const TUpdateDataBase<TDATA>&) = delete;

    /// Increment freeze counter. If Freeze count is greater than 0, the view
    /// should not be sent an update message.  While the updates are frozen,
    /// they are cached to be triggered when the updates are restored.
    void FreezeViewUpdate() { m_uiViewUpdateLockRefCount++; }

    // Decrement freese counter. When freeze counter is at 0, the view is
    // updated with any cached update requests.
    void RestoreViewUpdate()
    {
        if(--m_uiViewUpdateLockRefCount < 0)
            m_uiViewUpdateLockRefCount = 0;

        if(m_uiViewUpdateLockRefCount == 0)
        {
            // view is no longer locked, update view for all back update
            // flags. UpdateView() will remove them from the list if a view
            // exists.
            while(!m_aUpdateCache.empty())
                UpdateView(*m_aUpdateCache.begin());
        }
    }

    /// Mechanism to notify all 'View' classes that the contents of the 'Data'
    /// class have changed and that they should take appropriate action.
    /// uUpdateFlag is an unsigned long number that is defined by a derived class.
    /// It identifies the nature of the update call. 0 should be interpretted as a
    /// generic message and the view should interrigate and update all
    /// properties.

    /// Note that when the updates are frozen, these are cached.  If the same
    /// flag is cached more than once, only one Update trigger is called.  If
    /// there is any 0 flag, only it is used to update the views.
    void UpdateView(unsigned long uUpdateFlag = 0)
    {
        if(m_uiViewUpdateLockRefCount == 0)
        {
            for(auto pcView : m_aViewPtr)
                pcView->OnBaseChanged(uUpdateFlag);

            // if present, remove update flag from list.
            m_aUpdateCache.erase(uUpdateFlag);
        }
        else if(!m_aViewPtr.empty())
        {
            // generic update messages will incompass all other update messages
            // so remove all update events and insert a generic update event.
            if(uUpdateFlag == 0)
                m_aUpdateCache.clear();

            // the view does not exist or it is locked. If a generic update request (0)
            // is not already present and if a view does exist, buffer the update event.
            if(m_aUpdateCache.find(0) == m_aUpdateCache.end())
                m_aUpdateCache.insert(uUpdateFlag);
        }
    }

private:
    /// Add a link to the specified 'View' class.
    void AddLinkToView(TUpdateViewBase<TDATA>* pLink)
    {
        m_aViewPtr.insert(pLink);
        pLink->m_pLinkBase = (TDATA*)this;
    }

    /// Remove the link to the specified 'View' class.
    void RemoveLinkToView(TUpdateViewBase<TDATA>* pLink)
    {
        ASSERT_ERROR(pLink);
        if(!pLink)
            return;

        m_aViewPtr.erase(pLink);
        pLink->m_pLinkBase = nullptr;
    }

    /// All the associated views that need to be updated.
    std::set<TUpdateViewBase<TDATA>*> m_aViewPtr;

    /// The cache of Updates that were received while freezing was on.
    std::set<unsigned long> m_aUpdateCache;

    // Reference counting for the view update lock mechanism. Used so that sequential
    // (possibly multi-level functions calls) only actually update the view when the
    // reference count drops to zero (meaning that the sequential chain of modifications
    // are completed and the data is now ready to be visually updated)
    size_t m_uiViewUpdateLockRefCount = 0;
};

// ---------------------------------------------------------------------------
/// Acts as the client (View) class for the data. Maintains knowledge of the
/// single 'Data' class that derives itself from the partner TUpdateDataBase class.
/// Links will be internally updated if this 'View' object is destroyed or
/// if it is requested to remove link to 'Data' class. Same is true for the 'Data'
/// class, if it is destroyed, this 'View' object's link to it will be removed.
template<class TDATA> class TUpdateViewBase
{
public:
    friend TUpdateDataBase<TDATA>;

    TUpdateViewBase(TDATA* pcData) : m_pLinkBase(nullptr) { LinkToBase(pcData); }
    TUpdateViewBase() = default;
    TUpdateViewBase(const TUpdateViewBase&) = delete;
    virtual ~TUpdateViewBase() { RemoveLinkToBase(); }
    TUpdateViewBase& operator=(const TUpdateViewBase&) = delete;

    // The serving 'Data' object has changed and the derived class of this 'View'
    // link should take appropriate action. The meaning of @p uiUpdateFlag is defined by the
    // specific TDATA classes and their views.
    virtual void OnBaseChanged(unsigned int uiUpdateFlag) = 0;

    /// Set link to single 'Data' serving class.
    virtual void LinkToBase(TDATA* pcData)
    {
        ASSERT_ERROR(pcData);

        if(m_pLinkBase)
            m_pLinkBase->RemoveLinkToView(this);

        pcData->AddLinkToView(this);
    }

    /// Remove link to 'Data' serving class.
    virtual void RemoveLinkToBase()
    {
        if(m_pLinkBase)
            m_pLinkBase->RemoveLinkToView(this);
        ASSERT_ERROR(!m_pLinkBase);
    }

    /// Get a pointer to the base class (TUpdateDataBase*) of the linked 'Data' serving
    /// class.
    TDATA* GetLinkToBase() { return m_pLinkBase; }

private:
    /// the data class this is associated with.  It can be nullptr if there is
    /// no association.
    TDATA* m_pLinkBase = nullptr;
};

#endif  // DATA_VIEW_UPDATE_HPP
