#ifndef DATAARCHIVE_HPP
#define DATAARCHIVE_HPP
//-----------------------------------------------------------------------------
/// @file dataarchive
/// The class and functions to archive binary data.  It also has some compatibility
/// Serialize(std::istream&, xxx) and Serialize(std::ostream&, xxx) methods.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <boost/variant/variant.hpp>
#include <istream>
#include <ostream>
#include <vector>

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
// ============================================================================
/// The Archive for CDataContainer objects ensures that we only save binary
/// compatible types.
///
/// The object can be made from an input or output stream for loading or
/// saving.
class CDataArchive final
{
public:

    /// @returns true if this is running on a little endian machine.
    static bool IsLittleEndian();

    /// Create the archive for saving data.
    CDataArchive(std::istream& is);
    /// Create the archive for loading data.
    CDataArchive(std::ostream& os);
    CDataArchive(CDataArchive&&) = default;
    CDataArchive& operator=(const CDataArchive&) = delete;
    ~CDataArchive() = default;

    /// @return true if this archive is in loading mode.
    bool IsLoading() const;

    /// @return True if the owned stream has no errors.
    bool IsValid() const;

    /// @return true if the achive is loading and the data is valid.
    bool IsLoadingValid() const;

    /// Set the stream bad bit.
    void SetBad();

    /// The lowest level functionality to load/save data into the archive.
    void SerializeAtomicLittleEndian(char* pBuff, size_t n);

    /// Serialize a buffer this has no chance for endian conversion.  This
    /// should only be called for historical compatibility.
    [[deprecated]] void SerializeUnsafeBuffer(char* pBuff, size_t n);

    /// @return the istream for the friend functions. Only valid if loading.
    std::istream& GetIstream() const;

    /// @return the ostream for the friend functions. Only valid if saving
    std::ostream& GetOstream() const;

private:
    /// The input or output stream.
    boost::variant<std::istream*, std::ostream*> m_stream;
};

/// Serialize any number.
template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(CDataArchive& rAr, T& rubData)
{
    rAr.SerializeAtomicLittleEndian(reinterpret_cast<char*>(&rubData), sizeof rubData);
}

/// Serialize Fixed width enumerations
template<typename T> void Serialize(CDataArchive& rAr, CFixedEnum8<T>& rubData)
{
    rAr.SerializeAtomicLittleEndian(reinterpret_cast<char*>(&rubData), sizeof rubData);
}
/// Serialize Fixed width enumerations
template<typename T> void Serialize(CDataArchive& rAr, CFixedEnum16<T>& ruiData)
{
    rAr.SerializeAtomicLittleEndian(reinterpret_cast<char*>(&ruiData), sizeof ruiData);
}
/// Serialize Fixed width enumerations
template<typename T> void Serialize(CDataArchive& rAr, CFixedEnum32<T>& ruiData)
{
    rAr.SerializeAtomicLittleEndian(reinterpret_cast<char*>(&ruiData), sizeof ruiData);
}

/// Serialize non fixed width enumerations, but (static) assert if not uint_16 compatible.
template<typename T, typename std::enable_if<std::is_enum<T>::value>::type* = nullptr>
void Serialize(CDataArchive& rAr, T& reData)
{
    static_assert(sizeof(T) == 2, "Only 16 bit raw enums are allowed for archive compatibility.");
    rAr.SerializeAtomicLittleEndian(reinterpret_cast<char*>(&reData), sizeof reData);
}


/// Serialize arrays (of types that can be serialized.)
template<typename T, std::size_t N> void Serialize(CDataArchive& rAr, std::array<T, N>& raData)
{
    for(auto& data : raData)
        Serialize(rAr, data);
}

template<typename T> void Serialize(CDataArchive& rAr, std::vector<T>& rData)
{
    uint_32 uiLen = rData.size();
    Serialize(rAr, uiLen);
    if(rAr.IsLoadingValid())
        rData.resize(uiLen);
    for(auto& rItem : rData)
        Serialize(rAr, rItem);
}

/// Special case, serialize bool as uint_16 data item.
void Serialize(CDataArchive& rAr, bool& rData);

/// Special case, serialize bool vector as uint_16 data item.
void Serialize(CDataArchive& rAr, std::vector<bool>& raData);

/// Special case, serialize std::string
void Serialize(CDataArchive& rAr, std::string& rszData);

/// Allow serialization of structures, but only if load and save is guaranteed with same compiler.
template<typename T, std::enable_if_t<std::is_standard_layout<T>::value>* = nullptr>
void SerializeSameArch(CDataArchive& rAr, T& rData)
{
    if(rAr.IsLoading())
        rAr.GetIstream().read((char*)&rData, sizeof(rData));
    else
        rAr.GetOstream().write((char*)&rData, sizeof(rData));
}

///@{
/// These support the transition from the older CDataContainer::ReadData/WriteData to the new
/// serialize style.  Note that it includes specialization for bool type that is compatible
/// with Serialize above.
template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(std::istream& is, T& rData)
{
    is.read((char*)&rData, sizeof rData);
}

template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(std::ostream& os, const T& rData)
{
    os.write((char*)&rData, sizeof rData);
}

template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(std::istream& is, CFixedEnum8<T>& rData)
{
    is.read((char*)&rData, sizeof rData);
}

template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(std::ostream& os, const CFixedEnum8<T>& rData)
{
    os.write((char*)&rData, sizeof rData);
}

template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(std::istream& is, CFixedEnum16<T>& rData)
{
    is.read((char*)&rData, sizeof rData);
}

template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(std::ostream& os, const CFixedEnum16<T>& rData)
{
    os.write((char*)&rData, sizeof rData);
}

template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(std::istream& is, CFixedEnum32<T>& rData)
{
    is.read((char*)&rData, sizeof rData);
}

template<typename T, typename std::enable_if<std::is_arithmetic<T>::value>::type* = nullptr>
void Serialize(std::ostream& os, const CFixedEnum32<T>& rData)
{
    os.write((char*)&rData, sizeof rData);
}

template<>
void Serialize(std::istream& is, bool& rData);
template<>
void Serialize(std::ostream& os, const bool& rData);
/// @}

#endif  // DATAARCHIVE_HPP
