#ifndef DATACONT_HPP
#define DATACONT_HPP
//------------------------------------------------------------------------------
/// @file
/// Data container base class (unstructured persistent storage).
//------------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "metadata-traits.hpp"
#include "dataarchive.hpp"
#include <functional>
#include <memory>
#include <string>
#include <initializer_list>

//-----------------------------------------------------------------------------
// Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
// ============================================================================
/// CDataContainer is the base class for versioned, serializable types.
///
/// Data objects are read and written with the Read() and Write() methods.
/// Those methods handle the version byte and then call ReadData() and
/// WriteData() to allow the derived class to store the data.  There is also a
/// state.
///
/// GetVersion defines the current version of the serialization format
/// and is the version written on Write.
///
/// GetDiskVersion returns the version of the serialization format of
/// a serialized stream on Read. This is used to support backward
/// compatibility.
///
/// Only one CDataContainer-derived class should implement GetVersion.
/// Best practice is to have only one concrete class in a container
/// hierarchy and for only concrete classes to implement GetVersion.
/// <pre>
///              CDataContainer (ABC)
///                      |
///              CEvacParamsBase (ABC)
///             /               /
///    CEvacParamsStd (final)   CEvacParams3500 (final)
/// </pre>
///
/// Some older containers may encode a separate version byte independent
/// of the container GetVersion value. This should be avoided to prevent
/// container version complexity.
///
/// <pre>
///              CDataContainer (ABC)
///                      |
///              CScrAnlDoIfStart (concrete; implements GetVersion)
///                      |
///              CScrAnlDoIfStart2020 (concrete; separate version byte)
/// </pre>
///
/// This class provides undo functionality through RAII by using the
/// CDataContainer::CUndoContext class.  This is built on Read()/Write().
class CDataContainer : public NMeta::IsDataContainer
{
    friend void Serialize(CDataArchive& rAr, CDataContainer& rData);

public:
    class CUndoContext;

    CDataContainer();
    CDataContainer(const CDataContainer& rhs);
    CDataContainer(CDataContainer&& rhs) = default;
    virtual ~CDataContainer() = default;

    /// @return the number of bytes the @p rBaseData will use in a Write() call
    size_t WriteSize();

    /// Assignment operator sets the disk version of the container to that of
    /// the container provided. Does not copy the container contents.
    CDataContainer& operator=( const CDataContainer& crData );

    /// Set the container state to nominal values. Used to reset to a default
    /// state on error.
    virtual void SetDefaults() = 0;

    /// @returns Version byte of the serialized container obtained during
    ///     Read(std::istream&). Supports backward read compatibility.
    uint_8 GetDiskVersion() const { return m_bDiskVersion; }

    /// Set the version of the data on disk.
    [[deprecated("DiskVersion should only be used during ReadData")]] void SetDiskVersion(
        uint_8 bVersion)
    {
        m_bDiskVersion = bVersion;
    }

    /// Override to set the current data container version. The class that
    /// derives directly from CDataContainer manages the data container version.
    /// If multiple classes in a class hierarchy implement ReadData() and
    /// WriteData(), version info should be read and stored explicitly in all
    /// but the managing class, e.g., add a version byte to the persistent data
    /// read/written in ReadData()/WriteData().
    virtual uint_8 GetVersion() const;
    /// The default Read() behavior on a disk version error (disk version higher
    /// than current container version) is to skip ReadData, mark the archive as bad,
    /// and return.
    /// @return False for default Read behavior on disk version error, true to ignore
    [[deprecated("This is only for unusual historical funkiness")]] virtual bool AllowRead(
        uint_8 ubDiskVersion) const;

    /// Reads the disk version byte then calls ReadData(). Calls SetDefaults()
    /// if the version by read fails or ReadData() returns false.
    virtual bool Read(std::istream& is);
    /// Writes the disk version byte then calls WriteData().
    virtual bool Write(std::ostream& os) const;

protected:
    using FnSerialUndo = std::function<void(CDataArchive& rAr)>;

    /// Sets the function that is called to serialize (save/load) the undo
    /// information for this class.
    void SetSerializeUndo(FnSerialUndo fnSerialUndo);

    /// Read the version byte.
    virtual bool ReadData(std::istream& is);
    /// Write the version byte.
    virtual bool WriteData(std::ostream& os) const;

    /// The function to serialize ourselves
    virtual void SerializeData(CDataArchive& rAr, uint_8 ubVersion);

private:
    /// This is used to serialize the data for the CUndoContext.
    /// By default it calls SerializeData().
    void SerializeUndo(CDataArchive& rAr);

    /// @return the number of bytes the @p rBaseData will use in a WriteData() call
    size_t WriteDataSize();

    /// The function to save/load the state of data object for Undo functionality.
    FnSerialUndo m_fnSerialUndo;

    mutable uint_8 m_bDiskVersion;     ///< Version of the data container instance.
};

/// Serialize @p rData to the @p ar.  This includes version control, and
/// reading the known parts of future objects.
void Serialize(CDataArchive& rAr, CDataContainer& rData);

// ============================================================================
/// This provides an RAII context for CDataContainer for paired Add/Undo
/// that allows both Undo() and HasChanged() functionality.  It allows multiple
/// CDataContainer objects to be in the context, but it also has a simple
/// interface to a single object.
///
/// You can add the one or more CDataContainer objects that are saved for undo
/// at construction, or you can construct without any, and then add them
/// incrementally with the Add() method (or a combination of the two).  The
/// Remove() can be used to remove and object after it has been added.
class CDataContainer::CUndoContext final
{
public:
    CUndoContext() = default;
    CUndoContext(CDataContainer& rDataObject);
    CUndoContext(std::initializer_list<CDataContainer*> aDataObjects);

    CUndoContext(const CUndoContext&) = delete;
    CUndoContext(CUndoContext&&) = default;
    CUndoContext& operator=(const CUndoContext&) = delete;

    ~CUndoContext() = default;

    /// Adds an additional data object to the undo context.
    void Add(CDataContainer& rDataObject);

    /// Remove @param rDataObject  from the undo list.  If @param fUndo is
    /// true, it will also call undo any changes that may have been made.
    void Remove(CDataContainer& rDataObject, bool fUndo);

    /// @return true if any of the objects have changed since creation.
    bool HasChanged() const;

    /// Reset the objects back to the state they had at creation.
    void Undo() const;

private:
    /// Information for each object.
    struct SDataInfo
    {
        SDataInfo(CDataContainer* p, std::string undo) : pData(p), szUndoInfo(std::move(undo)) {}
        CDataContainer* pData;
        std::string szUndoInfo;
    };

    /// The info for each object.
    std::vector<SDataInfo> m_vDataInfo;
};

//-----------------------------------------------------------------------------
// Template functions
//-----------------------------------------------------------------------------
/// Helper template functions for iterable, ordered, standard containers (vector,
/// list, deque) to present a datacontainer-like read/write/get-size functions.
///
/// The length of the container is read and written; container must implement
/// resize(). Elements must be of a static size determinable by sizeof.
//-----------------------------------------------------------------------------
namespace NVarLenStdContainer
{
    //--------------------------------------------------------------------------
    /// Read into container a stream of element type.
    template<typename T> bool Read(std::istream& is, T& container)
    {
        uint_32 uiLen = 0;
        is.read( (char*)&uiLen, sizeof( uiLen ) );
        container.resize( uiLen );
        for( auto& el : container )
            is.read( (char*)&el, sizeof( el ) );
        return is.good();
    }

    //--------------------------------------------------------------------------
    /// Read into container a stream of specified type U. Assumes specified type
    /// is convertible to element type. Example usage is for a container of bools 
    /// stored as MMFLAGs.
    template<typename T, typename U> bool Read(std::istream& is, T& container)
    {
        uint_32 uiLen = 0;
        is.read( (char*)&uiLen, sizeof( uiLen ) );
        container.resize( uiLen );
        for(auto i = 0; i < uiLen; i++)
        {
            U readEl;
            is.read( (char*)&readEl, sizeof( U ) );
            container[i] = readEl;
        }
        return is.good();
    }

    //--------------------------------------------------------------------------
    /// Write container elements as element type to the stream.
    template<typename T> bool Write(std::ostream& os, const T& container)
    {
        uint_32 uiLen = container.size();
        os.write( (char*)&uiLen, sizeof( uiLen ) );
        for( const auto& el : container )
            os.write( (char*)&el, sizeof( el ) );
        return os.good();
    }

    //--------------------------------------------------------------------------
    /// Write container elements as specified type U to the stream. Assumes 
    /// elements are convertible to specified type U. Example usage is for a 
    /// container of bools stored as MMFLAGs.
    template<typename T, typename U> bool Write(std::ostream& os, const T& container)
    {
        uint_32 uiLen = container.size();
        os.write( (char*)&uiLen, sizeof( uiLen ) );
        for( const auto& el : container )
        {
            U writeEl = el;
            os.write( (char*)&writeEl, sizeof( U ) );
        }
        return os.good();
    }
}

//-----------------------------------------------------------------------------
/// Helper template functions for iterable, ordered, standard containers (vector,
/// list, deque) to present a datacontainer-like read/write/get-size functions.
///
/// The container must be of fixed length. Elements must be of a static size
/// determinable by sizeof.
//-----------------------------------------------------------------------------
namespace NFixedSizeStdContainer
{
    //--------------------------------------------------------------------------
    /// Read into container a stream of element type.
    template<typename T> bool Read(std::istream& is, T& container)
    {
        for( auto& el : container )
            is.read( (char*)&el, sizeof( el ) );
        return is.good();
    }

    //--------------------------------------------------------------------------
    /// Read into container a stream of specified type U. Assumes specified type
    /// is convertible to element type. Example usage is for a container of bools
    /// stored as MMFLAGs, e.g.,
    ///
    ///     NFixedSizeStdContainer<bool, MMFLAG>::Read( is, container );
    template<typename T, typename U> bool Read(std::istream& is, T& container)
    {
        for( auto& el : container )
        {
            U readEl;
            is.read( (char*)&readEl, sizeof( U ) );
            el = readEl;
        }
        return is.good();
    }

    //--------------------------------------------------------------------------
    /// Write container elements as element type to the stream.
    template<typename T> bool Write(std::ostream& os, const T& container)
    {
        for(const auto& el : container)
            os.write( (char*)&el, sizeof( el ) );
        return os.good();
    }

    //--------------------------------------------------------------------------
    /// Write container elements as specified type U to the stream. Assumes
    /// elements are convertible to specified type U. Example usage is for a
    /// container of bools stored as MMFLAGs, e.g.,
    ///
    ///     NFixedSizeStdContainer<bool, MMFLAG>::Write( os, container );
    template<typename T, typename U> bool Write(std::ostream& os, const T& container)
    {
        for(const auto& el : container)
        {
            U writeEl = el;
            os.write( (char*)&writeEl, sizeof( U ) );
        }
        return os.good();
    }
}

//-----------------------------------------------------------------------------
/// Helper template functions for iterable, ordered, standard containers
/// (vector, list, deque) of datacontainer-like objects.
///
/// The length of the container is read and written; container must implement
/// resize(). Elements must implement Read/Write.
//-----------------------------------------------------------------------------
namespace NVarLenDataContainer
{
    template<typename T> bool Read(std::istream& is, T& container);
    template<typename T> bool Write(std::ostream& os, const T& container);

    //---------------------------------------------------------------------------------------
    /// Duplicated from miclib/metadata.hpp, to reduce dependencies.
    /// @defgroup type_trait_templates_smart_pointer is-std-pointer
    /// Determine whether T is a std::shared_ptr or std::unique_ptr.
    /// @{
    //---------------------------------------------------------------------------------------
    template<class T>
    struct is_smart_ptr : std::false_type
    {};
    template<class T>
    struct is_smart_ptr<std::shared_ptr<T>> : std::true_type
    {};
    template<class T, class U>
    struct is_smart_ptr<std::unique_ptr<T, U>> : std::true_type
    {};
    /// Alias for is_smart_ptr<>::value.
    template<class T>
    constexpr bool is_smart_ptr_v = is_smart_ptr<std::remove_cv_t<T>>::value;
    /// @} type_trait_templates_smart_pointer

    /// Private helper functions for the public template functions.
    class CUtil
    {
        template<typename T> friend bool Read(std::istream& is, T& container);
        template<typename T> friend bool Write(std::ostream& os, const T& container);

        /// @returns Reference to non-const value given pointer to non-const value.
        template<typename T,
                 std::enable_if_t<std::is_pointer<T>::value
                 || is_smart_ptr_v<T>
                 && !std::is_const<std::remove_pointer_t<T>>::value>* = nullptr>
        static auto& GetContainer( T p )
        {
            return* p;
        }

        /// @returns Reference to const value given pointer to const value.
        template<typename T,
                 std::enable_if_t<std::is_pointer<T>::value
                 || is_smart_ptr_v<T>
                 && std::is_const<std::remove_pointer_t<T>>::value>* = nullptr>
        static const auto& GetContainer( T p )
        {
            return* p;
        }

        /// @returns Reference to non-const value given reference to non-const value.
        template<typename T,
                 std::enable_if_t<!std::is_pointer<T>::value
                 && !is_smart_ptr_v<T>
                 && !std::is_const<T>::value>* = nullptr>
        static auto& GetContainer( T& rv )
        {
            return rv;
        }

        /// @returns Reference to const value given reference to non-const value.
        template<typename T,
                 std::enable_if_t<!std::is_pointer<T>::value
                 && !is_smart_ptr_v<T>
                 && std::is_const<T>::value>* = nullptr>
        static const auto& GetContainer( T& rv )
        {
            return rv;
        }
    };

    //--------------------------------------------------------------------------
    /// Read into container a stream of datacontainer-like objects, i.e., elements
    /// implement Read( std::istream ).
    template<typename T> bool Read(std::istream& is, T& container)
    {
        uint_32 uiLen = 0;
        is.read( (char*)&uiLen, sizeof( uiLen ) );
        container.resize( uiLen );
        for( auto& el : container  )
            CUtil::GetContainer( el ).Read( is );
        return is.good();
    }

    //--------------------------------------------------------------------------
    /// Store container elements as element type.
    template<typename T> bool Write(std::ostream& os, const T& container)
    {
        uint_32 uiLen = container.size();
        os.write( (char*)&uiLen, sizeof( uiLen ) );
        for( auto& el : container )
            CUtil::GetContainer( el ).Write( os );
        return os.good();
    }
}

//-----------------------------------------------------------------------------
/// Helper template functions for iterable, ordered, standard containers
/// (vector, list, deque) of datacontainer-like objects.
///
/// The container must be of fixed length. Elements must implement Read/Write.
//-----------------------------------------------------------------------------
namespace NFixedSizeDataContainer
{
    //--------------------------------------------------------------------------
    /// Read into container a stream of datacontainer-like objects, i.e., elements
    /// implement Read( std::istream ).
    template<typename T> bool Read(std::istream& is, T& container)
    {
        for( auto& el : container )
            el.Read( is );
        return is.good();
    }

    //--------------------------------------------------------------------------
    /// Store container elements as element type.
    template<typename T> bool Write(std::ostream& os, const T& container)
    {
        for(const auto& el : container)
            el.Write( os );
        return os.good();
    }
}

#endif
