#ifndef DATACONTEXT_HPP
#define DATACONTEXT_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "formattr.hpp"
#include "internalunits.hpp"
#include "metadata.hpp"
#include "unitscontext.hpp"
#include "stbaselib.h"
#include <array>
#include <functional>
#include <map>
#include <set>
#include <type_traits>
#include <typeindex>
#include <typeinfo>
#include <variant>
#include <vector>

//=================================================================================================
class CSubsetOptions;  // Opaque class for this module.

//=================================================================================================
/// The NDataContext namespaces allows metadata to be associated with members of structures and
/// classes.  The member information that can be stored is:
/// - Name for printing.
/// - Name for logging changes.
/// - NumericFormat (for numeric values) that give presentation options for precision, units and
///   range.
/// - Default value
/// - Calculation for if it is relevant given the other values in the structure.  For example the
///   minimum axis value is not relevant if the axis is auto scaled.
/// - Calculation, based on CSubsetOptions if the value is used in this context.
/// - The set of sample status bits that would cause this to be non-editable.
///
/// The metadata is set as part of static initialization (pre-main). It should be used only post
/// main, or post initialization.  The NDataContext::CInit class is used to call a the
/// initialization function for the constructor for the static object.  In a file that wants static
/// initialization just make a static variable of this type, and pass in the initialization
/// function.
///
/// The NDataContext then provides this information to print, editor, and other generic
/// functionality.
namespace NDataContext
{
    struct SMemberID;
    struct SMemberData;

    /// A utility function that can convert simple (scalar) types to a CUnitString.
    template<typename T> CUnitsStr ToString(const T& cData);

    /// The type of the function that is used for calculating relevancy.  Note that it depends on a
    /// specific data object that provides the context for the relevancy.
    template<typename T> using TFnIsRelevant = std::function<bool(const T&)>;
    // The type of the function that is used to decide if a member is relevant.  Not that it does
    // not depend on an object, but just on the relevant CSubsetOptions for the data object.
    using FnIsUsed = std::function<bool(const CSubsetOptions&)>;
    /// A vector of member data information.
    using SMemberVect = std::vector<SMemberData>;

    bool operator==(const SMemberID& lhs, const SMemberID& rhs);

    /// The initialization helper class.
    class CInit;

    /// @{ Set Member information into the global data context
    template<typename T, typename S>
    void SetMember(S T::*pMember, ///< Which member to set.
        std::variant<int, CStr> prompt,  ///< The prompt for print format, or 0 to not print.
        std::variant<int, CStr> label, ///< The label for the change log.
        const TFormatter<S>* pFormatter ///< The formatter for the value.
        );
    template<typename T, typename S, typename Units>
    void SetMember(NUnits::TQuantity<S, Units> T::*pMember, ///< Which member to set.
        std::variant<int, CStr> prompt,  ///< The prompt for print format, or 0 to not print.
        std::variant<int, CStr> label, ///< The label for the change log.
        const TFormatter<S>* pFormatter ///< The formatter for the value.
        );
    template<typename T, typename S, typename U>
    void SetMember(S T::*pMember, ///< Which member to set.
        std::variant<int, CStr> prompt,  ///< The prompt for print format, or 0 to not print.
        std::variant<int, CStr> label, ///< The label for the change log.
        const U& cDefaultValue ///* The default value for the member.
        );
    /// @}

    /// Set the association between the class/member and its name.
    template<typename T, typename S>
    void SetPromptLabel(S T::*pMember, std::variant<int, CStr> prompt, std::variant<int, CStr> label);
    /// Does the type have members defined?
    bool HasMembers(const std::type_index& cType);
    /// @return the list of members known for the type.
    const SMemberVect& GetMembers(const std::type_index& cType);

    /// Does the member have a known formatter?
    bool FormatterMapHas(const SMemberID& csMember);
    /// Get the formatter associated with the member.
    template<typename T, typename S> const TFormatter<S>* GetFormatter(S T::*pMember);
    template<typename T, typename S, typename Units>
    const TFormatter<S>* GetFormatter(NUnits::TQuantity<S, Units> T::*pMember);
    const void* GetFormatterVoid(const SMemberID& rsID);

    /// Does the member have a known SetDefault method.
    bool SetDefaultMapHas(const SMemberID& csMember);
    /// Set the default values of the members of @p rData
    template<typename T> void SetDefaults(T& rData);
    void SetDefaultsVoid(const std::type_index& cType, void* pData);

    /// Set the function for calculating if the member is relevant in that context.
    template<typename T, typename S> void RelevantIf(S T::*pMember, TFnIsRelevant<T> fnRelevantIf);
    /// Mark @p pMember as relevent only if @p pDepend is relevant.
    template<typename T, typename S1, typename S2> void RelevantDep(S1 T::*pMember, S2 T::*pDepend);
    /// Does the member have known RelevantIf functions.
    bool IsRelevantIfMapHas(const SMemberID& csMember);
    /// Does the member have any dependent members.
    bool RelevantDepMapHas(const SMemberID& csMember);
    /// See if the member is relevant in the current objects context.
    template<typename T, typename S> bool IsRelevant(S T::*pMember, const T& cData);
    template<typename T> bool IsRelevant(const SMemberID& csMemberID, const T& cData);
    bool IsRelevantVoid(const SMemberID& csMemberID, const void* pClass);

    /// Set if the member is used.
    template<typename T, typename S> void UsedIf(S T::*pMember, FnIsUsed fnIsUsed);
    /// Does the member have UsedIf functions?
    bool UsedIfMapHas(const SMemberID& csMember);
    /// See if the member is used in the context.
    bool IsUsedVoid(const SMemberID& csMemberID, const CSubsetOptions& cSSOpt);

    /// Set the status mask for which cData should be disabled.
    template<typename T, typename S> void DisableOnStatus(S T::*pMember, uint_16 uiFlags);
    /// Does the member have a disable on status mask?
    bool DisableStatusMapHas(const SMemberID& csMember);
    /// @returns the disable on status mask (or 0).
    uint_16 GetDisableStatus(const SMemberID& csMember);

    /// Get the Prompt (with colon) of the member
    template<typename T, typename S> CStr GetPrompt(S T::*pMember);
    CStr GetPromptVoid(const SMemberID& cMembetID);

    /// Get the Label (for confirm logging) of the member
    template<typename T, typename S> CStr GetLabel(S T::*pMember);
    CStr GetLabelVoid(const SMemberID& cMembetID);

    /// Get the CUnitsStr of the member value. (It must be a scalar.)
    template<typename T, typename S> CUnitsStr GetUnitsStr(S T::*pMember, const T& rData);
    template<typename T, typename S> CUnitsStr GetUnitsStr(S T::*pMember, const S& rData);
    template<typename T> CUnitsStr GetUnitsStrVoidAgg(const SMemberID& csMemberID, const T& rData);
    template<typename S> CUnitsStr GetUnitsStrVoidScalar(const SMemberID& csMemberID, const S& rData);

    /// See if the member is a scalar value.
    template<typename T, typename S> bool IsScalar(S T::*pMember);
    bool IsScalarVoid(const SMemberID& csMemberID);

    template<typename T> void InitializeMetadata(const NMeta::SMemberLabel& csLabel);
};  // namespace NDataContext

/// The implementation namespace that supports the public NDataContext
namespace NDataContext::detail
{
    using namespace NDataContext;
    using FnIsRelevantVoid = std::function<bool(const void*)>;
    using FnSetDefault = std::function<void(void*)>;

    /// Set the formatting for the member.
    template<typename T, typename S>
    void SetFormatter(S T::*pMember, const TFormatter<S>* pFormatter);
    template<typename T, typename S, typename Units>
    void SetFormatter(NUnits::TQuantity<S, Units> T::*pMember, const TFormatter<S>* pFormatter);
    /// @{Set the default value for the member
    template<typename T, typename S>
    void SetDefault(S T::*pMember, const TFormatter<S>* pFormatter);
    template<typename T, typename S, typename Units>
    void SetDefault(NUnits::TQuantity<S, Units> T::*pMember, const TFormatter<S>* pFormatter);
    template<typename T, typename U>
    void SetDefault(CStr T::*pMember, const U& value);
    template<typename T, typename S, typename U>
    void SetDefault(S T::*pMember, const U& value);
    /// @}

    // See: Pointers to member functions are very strange animals
    // (https://devblogs.microsoft.com/oldnewthing/20040209-00/?p=40713)
    const size_t cuiMemberPtrSize = sizeof(int*) + sizeof(size_t);

    /// @{Update the global known information for the type/member.
    void AddTypeMember(const std::type_index& cType, const SMemberData& rsMemberData);
    void SetFormatter(const SMemberID& rsID, const void* pFormatter);
    void SetSetDefault(const SMemberID& rsID, FnSetDefault fnSetDefault);
    void AddIsRelevant(const SMemberID& rsID, FnIsRelevantVoid fnIsRelevant);
    void AddRelevantDep(const SMemberID& rsID, const SMemberID& rsDependent);
    void AddUsedIf(const SMemberID& rsID, FnIsUsed fnIsUsed);
    void SetDisableStatus(const SMemberID& rsID, uint_16 uiMask);
    /// @}

};  // namespace NDataContext::detail

/// This class is useful to do static initialization.  Just make a file static variable of this type
/// with an initialization function that does the initialization call.
class NDataContext::CInit final
{
public:
    /// Calls the initialization function.
    CInit(std::function<void(void)> fnInit);
};

/// The data that identifies a pointer to a member in a non template way.
struct NDataContext::SMemberID
{
    /// construct a generic pointer to member to SMemberID.
    template<typename T, typename S> SMemberID(S T::*pMember);

    /// Get the generic pointer from the SMemberID.
    template<typename T, typename S> S T::*Ptr() const;

    /// The type information of the class/object.
    std::type_index class_type;
    /// The type information of the member of thatmfc object.
    std::type_index member_type;
    /// The type erased representation of the member pointer.
    std::array<char, NDataContext::detail::cuiMemberPtrSize> hash;
};

/// All the information that is stored for a specific class/member pair.
struct NDataContext::SMemberData
{
    /// The function that returns full pointer to the member of the aggregate object.
    using FnGetPtr = std::function<void*(void*)>;
    /// The function that returns a full CUnitsConext for the member of the aggregate object.
    using FnToUnitsContext = std::function<CUnitsStr(const void*)>;
    /// The function that returns a NMeta::SMeta structure for the member.
    using FnGetSMeta = std::function<NMeta::SMeta()>;

    /// The kind object the member can be.  Different types of objects can be treated differently.
    enum EDataType
    {
        ceScalar,    ///< This is a single valued object.
        ceAggregate  ///< This is a structure or class that has members of its own.
    };

    NDataContext::SMemberID id;        ///< Which member this refers to.
    std::variant<int, CStr> prompt;    ///< The prompt of the member (for printing)
    std::variant<int, CStr> label;     ///< The label of the member (for change logging)
    EDataType eDataType;               ///< What kind of data it is.
    FnGetPtr fnGetPtr;                 ///< Get the pointer to the data object.
    FnToUnitsContext fnToUnitsString;  ///< Get the CUnitString for the member (scalar only)
    FnGetPtr fnGetEditPtr;             ///< Get the data pointer used by the editors.
    FnGetSMeta fnGetSMeta;             ///< Function to get the NMeta Info.
};

/// The hash algorithm that allows SMemberID's to be used as unordered_map indices.
template<> struct std::hash<NDataContext::SMemberID>
{
    std::size_t operator()(const NDataContext::SMemberID& cID) const;
};

// ================================================================================================
template<typename T> CUnitsStr NDataContext::ToString(const T& cData)
{
    if constexpr(std::is_same_v<T, CStr>)
        return cData;
    else if constexpr(std::is_same_v<T, bool>)
        return CStr(cData ? IDS_DIO_YES : IDS_DIO_NO);
    else if constexpr(std::is_arithmetic_v<T>)
        return CStr(std::to_string(cData).c_str());
    else if constexpr(NUnits::is_quantity_v<T>)
        return CStr(std::to_string(cData.Value()).c_str());
    else
        static_assert(NMeta::dependent_false_v<T>, "unhandled type");
}

template<typename T, typename S>
NDataContext::SMemberID::SMemberID(S T::*pMember)
    : class_type(std::type_index(typeid(T))),
      member_type(std::type_index(typeid(S)))
{
    static_assert(sizeof(pMember) <= NDataContext::detail::cuiMemberPtrSize);
    hash = decltype(hash){0};  // Make sure unused bits are zero;
    memcpy((char*)&hash, (char*)&pMember, sizeof(pMember));
}

template<typename T, typename S> S T::*NDataContext::SMemberID::Ptr() const
{
    S T::*pMember = nullptr;
    static_assert(sizeof(pMember) <= NDataContext::detail::cuiMemberPtrSize);
    memcpy((char*)&pMember, (char*)&hash, sizeof(pMember));
    return pMember;
}

template<typename T, typename S>
void NDataContext::SetMember(S T::*pMember,
    std::variant<int, CStr> prompt,
    std::variant<int, CStr> label,
    const TFormatter<S>* pFormatter)
{
    NDataContext::SetPromptLabel(pMember, prompt, label);
    NDataContext::detail::SetFormatter(pMember, pFormatter);
}

template<typename T, typename S, typename Units>
void NDataContext::SetMember(NUnits::TQuantity<S, Units> T::*pMember,
    std::variant<int, CStr> prompt,
    std::variant<int, CStr> label,
    const TFormatter<S>* pFormatter)
{
    NDataContext::SetPromptLabel(pMember, prompt, label);
    NDataContext::detail::SetFormatter(pMember, pFormatter);
}

template<typename T, typename S, typename U>
void NDataContext::SetMember(S T::*pMember,
    std::variant<int, CStr> prompt,
    std::variant<int, CStr> label,
    const U& value)
{
    NDataContext::SetPromptLabel(pMember, prompt, label);
    NDataContext::detail::SetDefault(pMember, value);
}

template<typename T, typename S>
void NDataContext::SetPromptLabel(S T::*pMember,
    std::variant<int, CStr> prompt,
    std::variant<int, CStr> label)
{
    using namespace NDataContext;
    using namespace NDataContext::detail;
    auto member = SMemberData{pMember, prompt, label};
    member.fnGetPtr = [pMember](void* ptr) {
        auto pData = static_cast<T*>(ptr);
        return static_cast<void*>(&(pData->*pMember));
    };
    member.fnGetEditPtr = [pMember](void* ptr) {
        auto pData = static_cast<T*>(ptr);
        return static_cast<void*>(&(pData->*pMember));
    };
    member.fnGetSMeta = [pMember]() {
        auto fnLabel = NMeta::LabelFn([pMember]() { return NDataContext::GetLabel(pMember); });
        auto fnValueDesc = NMeta::ValueDescFn<S>([pMember](const S& rValue) {
            return CUnitsContext::GetDefault()->ConvertString(NDataContext::GetUnitsStr(pMember, rValue));
        });
        return NMeta::Note(pMember, fnLabel, fnValueDesc);
    };
    if constexpr(std::is_same_v<S, CStr> || std::is_same_v<S, bool>)
    {
        member.eDataType = SMemberData::ceScalar;
        member.fnToUnitsString = [](const void* ptr) {
            auto pData = static_cast<const S*>(ptr);
            return ToString(*pData);
        };
    }
    else if constexpr(std::is_enum_v<S>)
    {
        member.eDataType = SMemberData::ceScalar;
        member.fnToUnitsString = [](const void* ptr) {
            auto pData = static_cast<const S*>(ptr);
            return NMeta::EnumDescription<S>(*pData);
        };
    }
    else if constexpr(std::is_arithmetic_v<S>)
    {
        member.eDataType = SMemberData::ceScalar;
        member.fnToUnitsString = [pMember](const void* ptr) -> CUnitsStr {
            auto pData = static_cast<const S*>(ptr);
            if(auto pFormat = GetFormatter(pMember))
                return TFormattedNumeric<S>(pFormat, *pData).ValueAndUnits();
            return ToString(*pData);
        };
    }
    else if constexpr(NUnits::is_quantity_v<S>)
    {
        member.eDataType = SMemberData::ceScalar;
        member.fnToUnitsString = [pMember](const void* ptr) -> CUnitsStr {
            auto pData = static_cast<const S*>(ptr);
            if(auto pFormat = GetFormatter(pMember))
                return TFormattedNumeric<typename S::value_type>(pFormat, pData->Value()).ValueAndUnits();
            return ToString(*pData);
        };
    }
    else if constexpr(NMeta::is_formatted_numeric_v<S>)
    {
        member.eDataType = SMemberData::ceScalar;
        member.fnGetEditPtr = [pMember](void* ptr) {
            auto pData = static_cast<T*>(ptr);
            return static_cast<void*>(&((pData->*pMember).InternalValueRef()));
        };
        if constexpr(std::is_same_v<S, TFormattedNumeric<double>>)
            member.fnToUnitsString = [](const void* ptr) {
                auto pData = static_cast<S*>(ptr);
                return CUnitsStr(*pData);
            };
        else
            member.fnToUnitsString = [](const void* ptr) {
                auto pData = static_cast<const S*>(ptr);
                return pData->ValueAndUnits();
            };
    }
    else
    {
        member.eDataType = SMemberData::ceAggregate;
    }

    AddTypeMember(std::type_index(typeid(T)), member);
}

template<typename T, typename S>
void NDataContext::detail::SetFormatter(S T::*pMember, const TFormatter<S>* pFormatter)
{
    ASSERT_ERROR(pFormatter);
    NDataContext::detail::SetFormatter(pMember, static_cast<const void*>(pFormatter));
    NDataContext::detail::SetDefault(pMember, pFormatter);
}

template<typename T, typename S, typename Units>
void NDataContext::detail::SetFormatter(NUnits::TQuantity<S, Units> T::*pMember,
    const TFormatter<S>* pFormatter)
{
    ASSERT_ERROR(pFormatter);
    NDataContext::detail::SetFormatter(pMember, static_cast<const void*>(pFormatter));
    NDataContext::detail::SetDefault(pMember, pFormatter);
}

template<typename T, typename S>
void NDataContext::detail::SetDefault(S T::*pMember, const TFormatter<S>* pFormatter)
{
    // Handle the normal case where we initialize with the same type.
    NDataContext::detail::SetSetDefault(pMember, [pMember, pFormatter](void* ptr) {
        auto pData = static_cast<T*>(ptr);
        pData->*pMember = pFormatter->GetDefault();
    });
}

template<typename T, typename S, typename Units>
void NDataContext::detail::SetDefault(NUnits::TQuantity<S, Units> T::*pMember,
    const TFormatter<S>* pFormatter)
{
    NDataContext::detail::SetSetDefault(pMember, [pMember, pFormatter](void* ptr) {
        auto pData = static_cast<T*>(ptr);
        pData->*pMember = NUnits::TQuantity<S, Units>(pFormatter->GetDefault());
    });
}

template<typename T, typename U>
void NDataContext::detail::SetDefault(CStr T::*pMember, const U& value)
{
    // Handle the case where we initialize a CStr with a string ID.
    static_assert(std::is_same_v<U, int>, "Type is not supported.");
    NDataContext::detail::SetSetDefault(pMember, [pMember, value](void* ptr) {
        auto pData = static_cast<T*>(ptr);
        pData->*pMember = CStr(value);
    });
}

template<typename T, typename S, typename U>
void NDataContext::detail::SetDefault(S T::*pMember, const U& value)
{
    // Handle the normal case where we initialize with the same type.
    static_assert(std::is_same_v<S, U>, "Type is not supported.");
    NDataContext::detail::SetSetDefault(pMember, [pMember, value](void* ptr) {
        auto pData = static_cast<T*>(ptr);
        pData->*pMember = value;
    });
}

template<typename T, typename S>
void NDataContext::RelevantIf(S T::*pMember, TFnIsRelevant<T> fnRelevantIf)
{
    using namespace NDataContext;
    using namespace NDataContext::detail;

    auto fnVoidRelevantIf = [fnRelevantIf](const void* ptr) {
        auto pData = static_cast<const T*>(ptr);
        return fnRelevantIf(*pData);
    };

    AddIsRelevant(SMemberID(pMember), std::move(fnVoidRelevantIf));
}

template<typename T, typename S1, typename S2>
void NDataContext::RelevantDep(S1 T::*pMember, S2 T::*pDepend)
{
    using namespace NDataContext;
    using namespace NDataContext::detail;
    AddRelevantDep(SMemberID(pMember), SMemberID(pDepend));
}

template<typename T, typename S> void NDataContext::UsedIf(S T::*pMember, FnIsUsed fnIsUsed)
{
    NDataContext::detail::AddUsedIf(pMember, fnIsUsed);
}

template<typename T, typename S> void NDataContext::DisableOnStatus(S T::*pMember, uint_16 uiFlags)
{
    NDataContext::detail::SetDisableStatus(pMember, uiFlags);
}

template<typename T, typename S> CStr NDataContext::GetPrompt(S T::*pMember)
{
    return NDataContext::GetPromptVoid(SMemberID(pMember));
}

template<typename T, typename S> CStr NDataContext::GetLabel(S T::*pMember)
{
    return NDataContext::GetLabelVoid(SMemberID(pMember));
}

template<typename T, typename S> CUnitsStr NDataContext::GetUnitsStr(S T::*pMember, const T& rData)
{
    return NDataContext::GetUnitsStrVoidAgg(SMemberID(pMember), rData);
}

template<typename T, typename S> CUnitsStr NDataContext::GetUnitsStr(S T::*pMember, const S& rData)
{
    return NDataContext::GetUnitsStrVoidScalar(SMemberID(pMember), rData);
}

template<typename T>
CUnitsStr NDataContext::GetUnitsStrVoidAgg(const SMemberID& csMemberID, const T& rData)
{
    using namespace NDataContext;
    using namespace NDataContext::detail;
    for(auto& data : GetMembers(csMemberID.class_type))
    {
        // Use const_cast here so that we don't need both a const and non const fnGetPtr.
        if(data.id == csMemberID)
            return data.fnToUnitsString(
                const_cast<const void*>(data.fnGetPtr(const_cast<T*>(&rData))));
    }

    ASSERT_ERROR(false);
    return CUnitsStr();
}

template<typename S>
CUnitsStr NDataContext::GetUnitsStrVoidScalar(const SMemberID& csMemberID, const S& rData)
{
    using namespace NDataContext;
    using namespace NDataContext::detail;
    for(auto& data : GetMembers(csMemberID.class_type))
    {
        // Use const_cast here so that we don't need both a const and non const fnGetPtr.
        if(data.id == csMemberID)
            return data.fnToUnitsString(
                static_cast<const void*>(&rData));
    }

    ASSERT_ERROR(false);
    return CUnitsStr();
}

template<typename T, typename S> const TFormatter<S>* NDataContext::GetFormatter(S T::*pMember)
{
    using namespace NDataContext::detail;
    if(!FormatterMapHas(pMember))
    {
        ASSERT_ERROR(false);
        return nullptr;
    }

    return static_cast<const TFormatter<S>*>(GetFormatterVoid(pMember));
}

template<typename T, typename S, typename Units>
const TFormatter<S>* NDataContext::GetFormatter(NUnits::TQuantity<S, Units> T::*pMember)
{
    using namespace NDataContext::detail;
    if(!FormatterMapHas(pMember))
    {
        ASSERT_ERROR(false);
        return nullptr;
    }

    return static_cast<const TFormatter<S>*>(GetFormatterVoid(pMember));
}

template<typename T, typename S> bool NDataContext::IsScalar(S T::*pMember)
{
    return NDataContext::IsScalarVoid(SMemberID(pMember));
}

template<typename T, typename S> bool NDataContext::IsRelevant(S T::*pMember, const T& cData)
{
    return NDataContext::detail::IsRelevant(NDataContext::SMemberID(pMember), cData);
}

template<typename T> bool NDataContext::IsRelevant(const SMemberID& csMemberID, const T& cData)
{
    return IsRelevantVoid(csMemberID, static_cast<const void*>(&cData));
}

template<typename T> void NDataContext::SetDefaults(T& rData)
{
    NDataContext::SetDefaultsVoid(std::type_index(typeid(T)), static_cast<void*>(&rData));
}

template<typename T> void NDataContext::InitializeMetadata(const NMeta::SMemberLabel& csLabel)
{
    using namespace NMeta;

    auto init = std::vector<SMeta>{};
    for(auto& rMember : GetMembers(std::type_index(typeid(T))))
        init.push_back(rMember.fnGetSMeta());

    ASSERT_ERROR(CGlobal::GetGlobalMap().count(typeid(T)) == 0);
    auto index = std::type_index(typeid(T));
    ASSERT_ERROR(CGlobal::GetGlobalMap().count(index) == 0);
    CGlobal::GetGlobalMap().insert({index, NoteRoot<T>(csLabel, std::move(init))});
}

#endif // DATACONTEXT_HPP
