#ifndef DEBUG_HPP
#define DEBUG_HPP
//-----------------------------------------------------------------------------
/// @file
/// The Debugging objects and mactros.
///
/// Notes:
///
/// 1.  Debug messages and invariant testing is enabled by the _DEBUG
///     define.
///
/// 2.  Header and cpp tracking is never enabled but present for
///     compatability.
///
/// 3.  The TEST_MSG?() macros are enabled on a per file basis with the
///     ENABLE_TEST_MSG macro is defined (default).
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "warningpragmas.hpp"
#if defined (MIC_VCMFC)
#include "stdafx.h"
#include <cstdarg>
#include <cstddef>
#include <boost/core/ignore_unused.hpp>
#else
#include <stdarg.h>
#include <stddef.h>
#endif
#include "basetype.hpp"
#include "debugc.h"
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------

// ========================================================================
///     This is the base class that handles the reporting of the debug
///     information. Each application platform library (386xlib and miclib)
///     should derive a class from this that can handle all the required
///     output. The module that does this needs to define the
///     CDebugHandler::CreateHandler() method to return an object of the
///     derived type. This function will only be called once.
///
///     *NOTE:* It is unusual for the derived class in the seperate library
///     to define one of the base classes member function. However,
///     this makes it possible to have the debug mecanism initialized
///     as needed (even befor main an independent of the initialization
///     order of the modules). There is precedence for this type of
///     workaround in the zApp::main() functionality.
class CDebugHandler
{
public:

    static  CDebugHandler*  Get();
    //  Thie retruns a pointer to the single CDebugHandler. If it does
    //  not yet exist, it is made.

    virtual ~CDebugHandler() {}
    //  Enshure that the destructor is virtual.

    virtual void    AssertFailed( int iLevel, const char* szFilename, int iLine ) {}
    //  The assertion of the given level failed.

    virtual void    ShowTraceString( const char* szStr ) {}
    virtual void    ShowTestString( const char* szStr ) {}
    //  The trace/test string should be shown.

protected:


private:

    static  CDebugHandler*  CreateHandler();
    //  NOTE:   This function needs to be defined by the application
    //          platform specific library to return an instance of
    //          a class that can handle the output appropriately.

    static  CDebugHandler*  ms_pHandler;
    //  THe class that handles the debug output.

};


/// The function that is called when one of the ASSERT_ macros is enabled and
/// fails.
void    MicAssertFailed( int iLevel, const char* FName, int iLine );


// --------------------------------------------------------------------------
/// @cond INTERNAL
#if !defined  (MIC_VCMFC)
#define THIS_FILE          __FILE__
#endif

#define ASSERT_FATAL_LEVEL      10
#define ASSERT_ERROR_LEVEL      20
#define ASSERT_WARNING_LEVEL    30
#define ASSERT_DEBUG_LEVEL      40

#ifdef  _DEBUG

    //  In the debugging version, turn on all levels of assertions.
    #define ASSERT_FATAL_ON
    #define ASSERT_ERROR_ON
    #define ASSERT_WARNING_ON
    #define ASSERT_DEBUG_ON

#else

    //  in the release version, turn on only the fatal errors.
    #define ASSERT_FATAL_ON

#endif
/// @endcond

// --------------------------------------------------------------------------
/// @name Assertions
///
/// There are 4 levels of assertions that are either enabled or disabled based
/// on the ASSERT_*_ON macro. If it is defined, the coresponding ASSERT_* macro
/// will generate an error, otherwise the argument of the assert will not be
/// executed. The VERIFY_* macro is the same as the ASSERT_* macro except that
/// the argument is executed in the release version also.
///
/// If `_DEBUG` is defined, all levels of assertions are enabled. If
/// it is not defined, only the `ASSERT_FATAL` is enabled.
//@{
    //  Enable the assertions based on which ones are turned on.
    //  _asm { int 3 } is the default MFC AfxDebugBreak();
#if defined (MIC_VCMFC)
    #ifdef  ASSERT_FATAL_ON
        #define ASSERT_FATAL( f )   VERIFY(f)
    #else
        #define ASSERT_FATAL( f )
    #endif
    #ifdef  ASSERT_ERROR_ON
        #define ASSERT_ERROR( f )   ASSERT(f)
    #else
        #define ASSERT_ERROR( f )
    #endif
    #ifdef  ASSERT_WARNING_ON
        #define ASSERT_WARNING( f ) ASSERT(f)
    #else
        #define ASSERT_WARNING( f )
    #endif
    #ifdef  ASSERT_DEBUG_ON
        #define ASSERT_DEBUG( f ) ASSERT(f)
    #else
        #define ASSERT_DEBUG( f )
    #endif
    #ifdef  ASSERT_ERROR_ON
        #define ASSERT_ERROR_RETURN( f, r ) if( !(f) ) {ASSERT(f); return (r); }
    #else
        #define ASSERT_ERROR_RETURN( f, r ) if ( !(f) ) { return (r); };
    #endif
#else
    #ifdef  ASSERT_FATAL_ON
        #define ASSERT_FATAL( f )   ((f) ? (void)0 : ::MicAssertFailed( ASSERT_FATAL_LEVEL, THIS_FILE, __LINE__))
    #else
        #define ASSERT_FATAL( f )
    #endif
    #ifdef  ASSERT_ERROR_ON
        #define ASSERT_ERROR( f )   ((f) ? (void)0 : ::MicAssertFailed( ASSERT_ERROR_LEVEL, THIS_FILE, __LINE__))
    #else
        #define ASSERT_ERROR( f )
     #endif
     #ifdef ASSERT_ERROR_ON
         #define    ASSERT_ERROR_RETURN( f, r ) if( !(f) ) {::MicAssertFailed( ASSERT_ERROR_LEVEL, THIS_FILE, __LINE__); return (r); } else;
     #else
        #define ASSERT_ERROR_RETURN( f, r ) if ( !(f) ) { return (r); } else;
    #endif
    #ifdef  ASSERT_WARNING_ON
        #define ASSERT_WARNING( f ) ((f) ? (void)0 : ::MicAssertFailed( ASSERT_WARNING_LEVEL, THIS_FILE, __LINE__))
    #else
        #define ASSERT_WARNING( f )
    #endif
    #ifdef  ASSERT_DEBUG_ON
        #define ASSERT_DEBUG( f )   ((f) ? (void)0 : ::MicAssertFailed( ASSERT_DEBUG_LEVEL, THIS_FILE, __LINE__))
    #else
        #define ASSERT_DEBUG( f )
    #endif
#endif

    //  Enable the verify to the same consistant level.
#ifdef  ASSERT_FATAL_ON
    #define VERIFY_FATAL( f )   ASSERT_FATAL( f )
#else
    #define VERIFY_FATAL( f )   ((void)(f))
#endif
#ifdef  ASSERT_ERROR_ON
    #define VERIFY_ERROR( f )   ASSERT_ERROR( f )
#else
    #define VERIFY_ERROR( f )   ((void)(f))
#endif
#ifdef  ASSERT_WARNING_ON
    #define VERIFY_WARNING( f ) ASSERT_WARNING( f )
#else
    #define VERIFY_WARNING( f ) ((void)(f))
#endif
#ifdef  ASSERT_DEBUG_ON
    #define VERIFY_DEBUG( f )   ASSERT_DEBUG( f )
#else
    #define VERIFY_DEBUG( f )   ((void)(f))
#endif
//@}

// --------------------------------------------------------------------------
/// @name Class Tracing
///
///     These macros are used to report information to the debug window.
/// The? is replaced by the number of arguments after the string (in
/// printf like format). The `TRACE_MSG?` macros are enabled when `_DEBUG`
/// is defined. The `TEST_MSG?` macros are enabled only when `_DEBUG` is
/// defined and `ENABLE_TEST_MSG` is defined (default).
///
/// Each of these can be used only in classes that have the DECLARE_MSG()
/// macro in their class declaration, Those classes also need to have
/// the DEFINE_MSG() macro in the cpp file.
///
/// Example:
///
///     class   CExample
///     {
///     public:
///         void Fun1( int i );
///     protected:
///     private:
///         DECLARE_MSG()
///     };
///
///     <in only 1 cpp.>
///     DEFINE_MSG( CExample, true, false );
///
///     void CExample::Fun1( int i )
///     {
///         TEST_MSG1( "CExample::Fun1( %d )\n", i );
///     }

//@{
#if defined (_DEBUG)
    #define TRACE_MSG       if( fTrace ) __DebugObj.TraceString
    #define TRACE_MSG0( s )                 TRACE_MSG( s )
    #define TRACE_MSG1( s, a1 )             TRACE_MSG( s, a1 )
    #define TRACE_MSG2( s, a1, a2 )         TRACE_MSG( s, a1, a2 )
    #define TRACE_MSG3( s, a1, a2, a3 )     TRACE_MSG( s, a1, a2, a3 )
    #define TRACE_MSG4( s, a1, a2, a3, a4 ) TRACE_MSG( s, a1, a2, a3, a4 )
    #define TRACE_MSG5( s, a1, a2, a3, a4, a5 ) TRACE_MSG( s, a1, a2, a3, a4, a5 )
    #define TRACE_MSG6( s, a1, a2, a3, a4, a5, a6 ) TRACE_MSG( s, a1, a2, a3, a4, a5, a6 )
    #define TRACE_MSG7( s, a1, a2, a3, a4, a5, a6, a7 ) TRACE_MSG( s, a1, a2, a3, a4, a5, a6, a7 )
    #define TRACE_MSG8( s, a1, a2, a3, a4, a5, a6, a7, a8 ) TRACE_MSG( s, a1, a2, a3, a4, a5, a6, a7, a8 )

    #define DECLARE_MSG()   static bool fTrace; static bool fTest; static   CDebugTrace     __DebugObj;
    #define DEFINE_MSG(Class, bTrace, bTest)    \
        bool Class::fTrace = bTrace; bool Class::fTest = bTest; CDebugTrace Class::__DebugObj = CDebugTrace (bTrace, bTest);
#else

    #define TRACE_MSG       1 ? (void)0 : (void)printf

    #define TRACE_MSG0( s )
    #define TRACE_MSG1( s, a1 )
    #define TRACE_MSG2( s, a1, a2 )
    #define TRACE_MSG3( s, a1, a2, a3 )
    #define TRACE_MSG4( s, a1, a2, a3, a4 )
    #define TRACE_MSG5( s, a1, a2, a3, a4, a5 )
    #define TRACE_MSG6( s, a1, a2, a3, a4, a5, a6 )
    #define TRACE_MSG7( s, a1, a2, a3, a4, a5, a6, a7 )
    #define TRACE_MSG8( s, a1, a2, a3, a4, a5, a6, a7, a8 )

    #define DECLARE_MSG()
    #define DEFINE_MSG(Class, bTrace, bTest)
#endif

#if defined (_DEBUG) && defined (ENABLE_TEST_MSG)

    #define TEST_MSG        if( fTest ) __DebugObj.TestString

    #define TEST_MSG0( s )                  TEST_MSG( s )
    #define TEST_MSG1( s, a1 )              TEST_MSG( s, a1 )
    #define TEST_MSG2( s, a1, a2 )          TEST_MSG( s, a1, a2 )
    #define TEST_MSG3( s, a1, a2, a3 )      TEST_MSG( s, a1, a2, a3 )
    #define TEST_MSG4( s, a1, a2, a3, a4 )  TEST_MSG( s, a1, a2, a3, a4 )
    #define TEST_MSG5( s, a1, a2, a3, a4, a5 )  TEST_MSG( s, a1, a2, a3, a4, a5 )
    #define TEST_MSG6( s, a1, a2, a3, a4, a5, a6 )  TEST_MSG( s, a1, a2, a3, a4, a5, a6 )
    #define TEST_MSG7( s, a1, a2, a3, a4, a5, a6, a7 )  TEST_MSG( s, a1, a2, a3, a4, a5, a6, a7 )
    #define TEST_MSG8( s, a1, a2, a3, a4, a5, a6, a7, a8 )  TEST_MSG( s, a1, a2, a3, a4, a5, a6, a7, a8 )

#else

    #define TEST_MSG        1 ? (void)0 : (void)printf

    #define TEST_MSG0( s )
    #define TEST_MSG1( s, a1 )
    #define TEST_MSG2( s, a1, a2 )
    #define TEST_MSG3( s, a1, a2, a3 )
    #define TEST_MSG4( s, a1, a2, a3, a4 )
    #define TEST_MSG5( s, a1, a2, a3, a4, a5 )
    #define TEST_MSG6( s, a1, a2, a3, a4, a5, a6 )
    #define TEST_MSG7( s, a1, a2, a3, a4, a5, a6, a7 )
    #define TEST_MSG8( s, a1, a2, a3, a4, a5, a6, a7, a8 )

#endif
//@}

//-----------------------------------------------------------------------------
/// This is the class that allows the TRACE_MSG and TEST_MSG macros to be used
/// to send debugging messages from inside member function of of classes that
///
/// @note
/// In the future, this will expand to allow the enabling
///     and disabling of messages for all classes from a central
///     point.
///
///     When debug is disabled, the macros TEST_MSG and TRACE_MSG must
///     be defined in such a way as to adsorb their arguement lists.
///     The "printf" function is currentlly used (see below).
///
/// @note Usage
///     In any class whose member functions are going to have use for Test and
///     trace messages Two things need to happen. In the declraration (probably
///     the private part, the DECLARE_MSG macro needs to be used to drclare a
///     static member of this class for the opbject to use (probably in the
///     private section). Then in the implementation the DEFINE_MSG(ClassName,
///     fShowTrace, fShowTest) macro needs to be used to define that static
///     member of the class. As a good default, the fShowTest should be false,
///     and the fShowTrace should be true.
///     Finaly, in any member function of theat class, the TEST_MSG and
///     TRACE_MSG macros can be used to send a test or trace message to the
///     output device.

#ifdef _DEBUG

class CDebugTrace
{
public:
    CDebugTrace(bool fShowTrace, bool fShowTest);
    CDebugTrace(int  fShowTrace, int  fShowTest);   //  Compatability only.

    /// Will show the test string (in printf form) if enabled.
    void    TestString(const char* szFmt, ... );
    /// Will show the trace string (in printf form) if enabled.
    void    TraceString(const char* szFmt, ... );

private:
    /// This will do the work of showing the string.
    static  void    ShowString(bool fTrace, const char* FmtStr, va_list& rArgs);

    /// Flag to see if Trace message should be shown.
    bool    m_fShowTrace;
    /// Flag to see if Test messages are to be shown.
    bool    m_fShowTest;
};

//-----------------------------------------------------------------------------
// Inline Methods
//-----------------------------------------------------------------------------
inline
CDebugTrace::CDebugTrace(
    bool    fShowTrace,
    bool    fShowTest
    )
{
    m_fShowTrace = fShowTrace;
    m_fShowTest = fShowTest;
}
inline
CDebugTrace::CDebugTrace(
    int     fShowTrace,
    int     fShowTest
    )
{
    m_fShowTrace = fShowTrace ? true : false;
    m_fShowTest = fShowTest ? true : false;
}
inline  void
CDebugTrace::TestString(
    const char* szFmt,
    ...
    )
{
    if (m_fShowTest) {
        va_list args;
        va_start(args, szFmt);

        ShowString(false, szFmt, args);
    }
}
inline  void
CDebugTrace::TraceString(
    const char* szFmt,
    ...
    )
{
    if (m_fShowTrace) {
        va_list args;
        va_start(args, szFmt);

        ShowString(true, szFmt, args);

    }
}
#endif

#endif // DEBUG_HPP
