#ifndef   FORMATTR_HPP
#define   FORMATTR_HPP
//-----------------------------------------------------------------------------
/// @file
/// Classes implementing self-formatting and verification functionality
/// for fundamental data types.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "baseunit.hpp"
#include "dataarchive.hpp"
#include "internalunits.hpp"

#include "../rapidjson/fwd.h"

#include <limits>
#include <sstream>
#include <stdlib.h>
#include <cmath>
#include <algorithm>

enum EFieldError
{
    ceFLDERR_NONE           = 0,
    ceFLDERR_REQUIRED,
    ceFLDERR_BOUNDS,
    ceFLDERR_NOTNUMBER,
    ceFLDERR_BADDATA,
    ceFLDERR_BADDEPENDENT,
    ceFLDERR_TABLE,
    ceFLDERR_BADFILENAME,
    ceFLDERR_BADFOLDER
};

#if defined(MIC_VCMFC)

#define TCSTOF(nptr, endptr) _tcstof((nptr), (endptr))
#define TCSTOD(nptr, endptr) _tcstod((nptr), (endptr))
#define TCSTOL(nptr, endptr, base) _tcstol((nptr), (endptr), (base))
#define TCSTOUL(nptr, endptr, base) _tcstoul((nptr), (endptr), (base))

#define TOSTREAM wostream
#define TOSTRINGSTREAM wostringstream

#else

#define TCSTOF(nptr, endptr) strtof((nptr), (endptr))
#define TCSTOD(nptr, endptr) strtod((nptr), (endptr))
#define TCSTOL(nptr, endptr, base) strtol((nptr), (endptr), (base))
#define TCSTOUL(nptr, endptr, base) strtoul((nptr), (endptr), (base))

#define TOSTREAM ostream
#define TOSTRINGSTREAM ostringstream

#endif

const unsigned long FMASK_NONE          = 0x00000000L;
const unsigned long FMASK_ALPHA         = 0x00000001L;
const unsigned long FMASK_INTEGER       = 0x00000002L;
const unsigned long FMASK_UINTEGER      = 0x00000004L;
const unsigned long FMASK_FLOAT         = 0x00000008L;
const unsigned long FMASK_HEXNUMERIC    = 0x00000010L;
const unsigned long FMASK_DISPLAY       = 0x00000020L;
const unsigned long FMASK_COMMENT       = 0x00000040L;  //Multiline edit field
const unsigned long FMASK_EXPFLOAT      = 0x00000080L;
const unsigned long FMASK_FILENAME      = 0x00000100L;
const unsigned long FMASK_FILENAMEWC    = 0x00000200L;  //File name with *and ?
const unsigned long FMASK_FILEBASENAME  = 0x00000400L;  // Valid file base name (no path or extension).

//Can be | with previous values.
const unsigned long FMASK_NOTREQUIRED   = 0x00010000L;
const unsigned long FMASK_TOUPPER       = 0x00020000L;
const unsigned long FMASK_TOLOWER       = 0x00040000L;

namespace NNumericFormat
{
    /// @return the Json schema for the CUnits ToJson method.
    std::string JsonSchema();
}  // namespace NNumericFormat

//-----------------------------------------------------------------------------
/// defines how a quantity should be shown to the user.
class  CNumericFormat
{
    /// @return the json value for @p rData ready to be inserted into @p rDoc.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CNumericFormat& rData);

public:
    /// The styles for how to show the number.
    enum EStyle
    {
        ceFixed,
        ceScientific,
        ceFixedOrScientific,
    };

    /// Constructor
    ///
    /// @param uiWidth The total output width (for leading zeros. of the
    /// output field including '.' and sign.
    /// @param uiPrecision How may digits past the '.' to show.
    /// @param eStyle How to show the number.
    /// @param dExpMin For ceFixedOrScientific the smallest exponent before
    /// swithcing to fixed.
    /// @param uiExpPrecision For ceFixedOrScientific the precision to show as
    /// in scientific notation.
    /// @param dExpMin For ceFixedOrScientific values (absolute) larger than
    /// this are shown in fixed format.
    /// @param dExpMax For ceFixedOrScientific values (absolute) smaller than
    /// this are shown in fixed format.
    CNumericFormat( int uiWidth = 0,
                    int uiPrecision = 0,
                    EStyle eStyle = ceFixed,
                    double dExpMin = 0.0,
                    int uiExpPrecision = 0,
                    double dExpMax = 0.0 )
        :   m_uiWidth( uiWidth ),
            m_uiPrecision( uiPrecision ),
            m_eStyle( eStyle ),
            m_dExpMin( dExpMin ),
            m_dExpMax( dExpMax ),
            m_uiExpPrecision( uiExpPrecision )
        {}

    int GetWidth() const { return m_uiWidth; }
    int GetPrecision() const { return m_uiPrecision; }
    EStyle GetStyle() const { return m_eStyle; }
    double GetExpMin() const;
    double GetExpMax() const;
    int GetExpPrecision() const;

private:
    int         m_uiWidth;      ///< Maximum number of characters for data entry including . and sign
    int         m_uiPrecision;  ///< Maximum number of characters after . (. not included)
    EStyle      m_eStyle;
    double      m_dExpMin;
    double      m_dExpMax;
    int         m_uiExpPrecision;

};

inline double CNumericFormat::GetExpMin() const
{
    ASSERT_ERROR( m_eStyle == ceFixedOrScientific );
    return m_dExpMin;
}

inline  double  CNumericFormat::GetExpMax() const
{
    ASSERT_ERROR( m_eStyle == ceFixedOrScientific );
    return m_dExpMax;
}

inline  int CNumericFormat::GetExpPrecision() const
{
    ASSERT_ERROR( m_eStyle == ceFixedOrScientific );
    return m_uiExpPrecision;
}

/// Setup the @p os to print the next number in the format specified by
/// @p fmt .
std::TOSTREAM& operator<<( std::TOSTREAM& os, const CNumericFormat& fmt );

//-----------------------------------------------------------------------------
// The value_traits and its specializations define some per type constants that
// are useful in the core templates of this file.
template <typename T> struct value_traits { };

template<> struct value_traits<float>
{
    typedef float value_type;
    static const unsigned long m_ulFilter = FMASK_FLOAT;
    static unsigned long GetFilter() { return m_ulFilter; }
    static bool FromString( const  CStr & szFromEdit, _TCHAR ** ppStopChar, float& value )
    {
        value = TCSTOF( szFromEdit, ppStopChar );
        if( **ppStopChar != 0 )
            return false;
        else
            return true;
    }
    static CNumericFormat DefaultFormat() { return CNumericFormat(5, 1 ); }
};

template < > struct value_traits<double>
{
    typedef double value_type;
    static const unsigned long m_ulFilter = FMASK_FLOAT;
    static unsigned long GetFilter() {return m_ulFilter;}
    static  bool FromString(const  CStr & szFromEdit, _TCHAR ** ppStopChar,
                            double & value)
    {
        value = TCSTOD(szFromEdit, ppStopChar);
        if( **ppStopChar != 0 )
            return false;
        else
            return true;
    }
    static CNumericFormat DefaultFormat() {return CNumericFormat(5, 1 );}
};

template<typename Units> struct value_traits<NUnits::TQuantity<double, Units>>
{
    typedef double value_type;
    static const unsigned long m_ulFilter = FMASK_FLOAT;
    static unsigned long GetFilter() { return m_ulFilter; }
    static bool FromString(const CStr& szFromEdit, _TCHAR** ppStopChar, double& value)
    {
        value = TCSTOD(szFromEdit, ppStopChar);
        if(**ppStopChar != 0)
            return false;
        else
            return true;
    }
    static CNumericFormat DefaultFormat() { return CNumericFormat(5, 1); }
};

template < > struct value_traits<short>
{
    typedef short short_type;
    static const unsigned long m_ulFilter = FMASK_INTEGER;
    static unsigned long GetFilter() {return m_ulFilter;}
    static bool FromString(const  CStr & szFromEdit, _TCHAR ** ppStopChar,
                            short & value)
    {
        long lValue = TCSTOL(szFromEdit, ppStopChar, 10);
        if( **ppStopChar == 0 && lValue >= std::numeric_limits<short>::lowest() &&
                                    lValue <= (std::numeric_limits<short>::max)())
        {
            value = static_cast<short>( lValue );
            return true;
        }
        return false;
    }
    static CNumericFormat DefaultFormat() {return CNumericFormat(5, 0 );}

  };

template < > struct value_traits<unsigned short> {
    typedef unsigned short ushort_type;
    static const unsigned long m_ulFilter = FMASK_UINTEGER;
    static unsigned long GetFilter() {return m_ulFilter;}
    static bool FromString(const  CStr & szFromEdit, _TCHAR ** ppStopChar,
                            unsigned short & value)
    {
        unsigned long lValue = TCSTOUL(szFromEdit, ppStopChar, 10);
        if( **ppStopChar == 0 && lValue <= (std::numeric_limits<unsigned short>::max)() )
        {
            value = static_cast<unsigned short>( lValue );
            return true;
        }
        return false;
    }
    static CNumericFormat DefaultFormat() {return CNumericFormat(5, 0 );}
};

template < > struct value_traits<long>
{
    typedef long long_type;
    static const unsigned long m_ulFilter = FMASK_INTEGER;
    static unsigned long GetFilter() {return m_ulFilter;}
    static bool FromString(const  CStr & szFromEdit, _TCHAR ** ppStopChar,
                            long & value)
    {
        value = TCSTOL(szFromEdit, ppStopChar, 10);
        if( **ppStopChar != 0 )
            return false;
        else
            return true;    }
    static CNumericFormat DefaultFormat() {return CNumericFormat(7, 0 );}
};

template < > struct value_traits<unsigned long>
{
    typedef unsigned long ulong_type;
    static const unsigned long m_ulFilter = FMASK_UINTEGER;
    static unsigned long GetFilter() {return m_ulFilter;}
    static unsigned long FromString(const  CStr & szFromEdit, _TCHAR ** ppStopChar,
                                    unsigned long & value)
    {
        value = TCSTOUL(szFromEdit, ppStopChar, 10);
        if( **ppStopChar != 0 )
            return false;
        else
            return true;
    }
    static CNumericFormat DefaultFormat() {return CNumericFormat(7, 0 );}
};

template < > struct value_traits<uint_32>
{
    typedef uint_32 ulong_type;
    static const unsigned long m_ulFilter = FMASK_UINTEGER;
    static unsigned long GetFilter() {return m_ulFilter;}
    static uint_32 FromString(const  CStr & szFromEdit, _TCHAR ** ppStopChar,
                              uint_32& value)
    {
        value = TCSTOUL(szFromEdit, ppStopChar, 10);
        if( **ppStopChar != 0 )
            return false;
        else
            return true;
    }
    static CNumericFormat DefaultFormat() {return CNumericFormat(7, 0 );}
};

template < > struct value_traits<int_32>
{
    typedef int_32 long_type;
    static const unsigned long m_ulFilter = FMASK_UINTEGER;
    static unsigned long GetFilter() {return m_ulFilter;}
    static int_32 FromString(const  CStr & szFromEdit, _TCHAR ** ppStopChar,
                             int_32& value)
    {
        value = TCSTOUL(szFromEdit, ppStopChar, 10);
        if( **ppStopChar != 0 )
            return false;
        else
            return true;
    }
    static CNumericFormat DefaultFormat() {return CNumericFormat(7, 0 );}
};

//-----------------------------------------------------------------------------
/// Defined the presentation for a quantity.
///
/// This combined a CUnits object, and a CNumericFormat object  (so far both a
/// print and edit that are similar).   It also provides the conversion
/// functions for the values and CNumericFormat based on the unit selections.
class CQuantityFormatter
{
public:
    /// Allow (testing) to overload the locale.
    static void OverrideLocale(const char* szLocale = "");

    /// Default constructor.  Used default units, fixed format with no
    /// 0 precision.
    CQuantityFormatter() = default;
    /// Constructor that sets the details of the formatting.
    CQuantityFormatter( CUnits mask, CNumericFormat editFormat, CNumericFormat prntFormat );
    ~CQuantityFormatter() = default;

    /// \returns the currently selected units.
    CStr ShortUnits() const { return m_cUnitsMask.DisplayText(); }
    /// \returns the CUnits associated with this.
    const CUnits &GetUnitsMask() const   { return m_cUnitsMask; }

    /// returns the CNumericFormat appropriate for the display based on the
    /// current unit selections.
    CNumericFormat  GetEditFormat( double dValue ) const
        { return GetFormat( UnitsFormatConvertion( m_cEditFormat ), dValue ); }

    /// \returns the internal format.
    const CNumericFormat&  GetInternalEditFormat() const { return m_cEditFormat; }

    /// returns the CNumericFormat appropriate for the display based on the
    /// current unit selections.
    CNumericFormat  GetPrintFormat( double dValue ) const
        { return GetFormat( UnitsFormatConvertion( m_cPrintFormat ), dValue ); }

    /// \returns the internal format.
    CNumericFormat  GetInternalPrintFormat() const {return m_cPrintFormat;}

    /// Convert the \param dInternalVal to the display value based in current
    /// units selections.
    double ToOutput( double dInternalVal ) const
        { return m_cUnitsMask.DisplayValue( dInternalVal ); }

    /// Convert the \param dDisplayVal to the internal value based in current
    /// units selections.
    double ToInternal( double dDisplayVal ) const
        { return GetUnitsMask().ConvertToInternal( dDisplayVal, {} ); }

    /// Convert the \param dInternalVal to a properly formatted sting using the
    /// edit formatter.
    CStr EditValue( double dInternalVal ) const;

    /// Convert the \param dInternalVal to a properly formatted sting using the
    /// print formatter.
    CStr Value ( double dInternalVal ) const;

    // NOTE: Only Used by grids in miclib, move there?
    static bool CheckChar( _TCHAR nChar, unsigned long ulFilter = FMASK_NONE );

    /// Give access to the units as a json string.
    std::string JsonUnits() const;
    /// Give access to the format as a json string.
    std::string JsonFormat() const;

protected:

    /// The default locale for the system.  This is used for string
    /// conversions.
    static std::locale ms_cLocale;

private:

    /// Update the CNumericFormat based on the scaling of the selected units.
    CNumericFormat UnitsFormatConvertion( CNumericFormat f ) const;

    /// Update the CNumericFormat based on the value value to be show.  This
    /// allows the proper selection for ceFixedOrScientific formatting.
    CNumericFormat GetFormat(CNumericFormat localFormat, double dValue ) const;

    CUnits m_cUnitsMask;
    CNumericFormat m_cEditFormat;
    CNumericFormat m_cPrintFormat;

    DECLARE_MSG()
};

//-----------------------------------------------------------------------------
/// The class that adds min, max, default to the CQuantityFormatter.  This is a
/// template to support different types of numbers.
template < class T > class TFormatter : public CQuantityFormatter
{
public:

    TFormatter (T tMin = std::numeric_limits<T>::lowest(),
                T tMax = (std::numeric_limits<T>::max)(), T tDefault = 0,
                CNumericFormat      editFormat  = value_traits<T>::DefaultFormat(),
                CNumericFormat      prntFormat  = value_traits<T>::DefaultFormat(),
                const CUnits    &mask       = CUnits());

    /// Access the data of the formatter.
    T   GetMinimum() const  { return m_tMinimum; }
    /// Access the data of the formatter.
    T   GetMaximum() const  { return m_tMaximum; }
    /// Access the data of the formatter.
    T   GetDefault() const  { return m_tDefault; }

    /// Set the allowed range of this formatter.
    void    SetRange(T min, T max);

    /// @param tInternalVal The internal value of a formatted value.
    /// @return True if the argument is in the range [min, max] when all are
    // converted and rounded to the precision of the print format.
    bool InRange( T tInternalVal ) const;

    /// @param szFromEdit The edit-format string representation of a formatted
    /// value.
    /// @param pStopChar A pointer to the first invalid character in szFromEdit
    /// or to the null terminator if the string has no invalid characters.
    /// @param rtInternalValue The internal value that corresponds szFromEdit
    /// will be returned through this reference if szFromEdit has no invalid
    /// characters.
    /// @return True if szFromEdit has no invalid characters.
    bool CheckData( const CStr& szFromEdit,
                    _TCHAR** pStopChar,
                    T& rtInternalValue ) const;

    /// @param tInternalValue The internal value of a formatted value.
    /// @return The output value obtained by converting the argument converted
    /// to the current units and rounding to the precision of the edit format.
    T RoundedOutputValue( T tInternalValue ) const;
    T RoundedInternalValue( T tInternalValue ) const;

private:

    T   m_tMinimum;
    T   m_tMaximum;
    T   m_tDefault;
};

template<class T> TFormatter<T>::TFormatter(
    T   tMin,
    T   tMax,
    T   tDefault,
    CNumericFormat  editFormat,
    CNumericFormat  prntFormat,
    const CUnits    &mask )
    : CQuantityFormatter( mask, editFormat, prntFormat ),
      m_tMinimum(tMin),
      m_tMaximum(tMax), m_tDefault(tDefault)
{
}

template<class T> void TFormatter<T>::SetRange(T min, T max)
{
    if(max >= min)
    {
        m_tMaximum = max;
        m_tMinimum = min;
    }
}

template<class T> bool TFormatter<T>::InRange( T tInternalValue ) const
{
    T tRounded = RoundedOutputValue( tInternalValue );
    return ( ( RoundedOutputValue( m_tMinimum ) <= tRounded )
             && ( tRounded <= RoundedOutputValue( m_tMaximum ) ) );
}

template<class T> bool TFormatter<T>::CheckData( const CStr& szFromEdit,
                                                 _TCHAR** ppStopChar,
                                                 T &tInternalValue ) const
{
    // Get the thousands grouping delimiter according to the locale.
    CStr szThousandsSep;
    szThousandsSep.Append( std::use_facet< std::numpunct< _TCHAR > >( ms_cLocale ).thousands_sep() );

    // Remove the delimiter: "1,000,222.5" => "1000222.5"
    CStr szNumber = szFromEdit;
    szNumber.DeleteAll( szThousandsSep );
    T tOutputValue;
    if ( value_traits< T >::FromString( szNumber, ppStopChar, tOutputValue ) )
    {
        auto dValue = ToInternal( tOutputValue );
        ASSERT_ERROR( (dValue >= std::numeric_limits<T>::lowest()
                       && dValue <= std::numeric_limits<T>::max())
                       || (std::numeric_limits<T>::has_quiet_NaN
                           && dValue == std::numeric_limits<T>::quiet_NaN()) );
        tInternalValue = static_cast<T>( dValue );
        return true;
    }
    return false;
}

template<class T> T TFormatter<T>::RoundedInternalValue( T tInternalValue ) const
{
    CStr szEditFormat( EditValue( tInternalValue ) );
    _TCHAR * pStopChar = NULL;
    CheckData( szEditFormat, &pStopChar, tInternalValue );
    return tInternalValue;
}

template<class T> T TFormatter<T>::RoundedOutputValue( T tInternalValue ) const
{
    auto dValue = ToOutput( RoundedInternalValue( tInternalValue ) );
    ASSERT_ERROR( (dValue >= std::numeric_limits<T>::lowest()
                   && dValue <= std::numeric_limits<T>::max())
                   || (std::numeric_limits<T>::has_quiet_NaN
                       && dValue == std::numeric_limits<T>::quiet_NaN()) );
    return static_cast<T>( dValue );
}

//-----------------------------------------------------------------------------
/// The class assosicates a TFormatter with an actual value.  It also adds
/// CDataContainer Read/Write functionality.
///
/// While this can be used directly, it is meant to be used with the
/// FORMATTED_*  macros to create a class/type that contains the formatting as
/// a static member that is passed onto this class.
///
template<class T> class TFormattedNumeric
{
public:
    using TNumericType = T;

    TFormattedNumeric( const TFormatter<T> &fmt, T tInternalVal )
        : m_pFormatter( &fmt ),
          m_tValue( tInternalVal )
    { }

    TFormattedNumeric( const TFormatter<T> *fmt, T tInternalVal )
        : m_pFormatter( fmt ),
          m_tValue(tInternalVal )
    { }

    TFormattedNumeric( const TFormattedNumeric<T> &d )
        : m_pFormatter( d.m_pFormatter ),
          m_tValue( d.m_tValue )
    { }
    virtual ~TFormattedNumeric(){}
    
    T  InternalValue() const { return m_tValue; }
    T& InternalValueRef() { return m_tValue; }
    const T& InternalValueRef() const { return m_tValue; }
    operator T () const { return m_tValue; }

    T OutputValue() const;

    T SetInternalValue(T tValue);
    T SetOutputValue(T tValue);

    // True if formatter's value is in the range [min, max] when all are
    // converted and rounded to the precision of the print format.
    bool InRange() const
        {   return m_pFormatter->InRange( m_tValue ); }

    // Test a value in internal units.
    bool InRange( T tInternalValue ) const
        {   return m_pFormatter->InRange( tInternalValue ); }

    // The remaining attributes are all extracted from the attached
    //  formatting object.
    T GetMinimum() const { return m_pFormatter->GetMinimum(); }
    T GetMaximum() const { return m_pFormatter->GetMaximum(); }
    T GetDefault() const { return m_pFormatter->GetDefault(); }
    CNumericFormat GetEditFormat() const { return m_pFormatter->GetEditFormat( m_tValue ); }
    CNumericFormat GetPrintFormat() const { return m_pFormatter->GetPrintFormat( m_tValue ); }
    const CUnits& GetUnitsMask() const { return m_pFormatter->GetUnitsMask(); }

    CStr EditValue() const { return m_pFormatter->EditValue( m_tValue ); }
    CStr Value() const { return m_pFormatter->Value(m_tValue); }
    CStr ShortUnits() const { return  m_pFormatter->ShortUnits(); }

    /// Reurn a CStr object that contians the value and short units string.
    CStr ValueAndUnits(uint_32 uiOverrideUnits) const;
    CStr ValueAndUnits() const { return ValueAndUnits(0); };

    /// Set The formatted numeric to the value that corresponds to edit-format
    /// string  szFromEdit.
    /// @param szFromEdit The edit-format string representation of a formatted
    /// value.
    /// @return The new internal value.
    virtual T StoreData( const CStr & szFromEdit );
    virtual bool CheckData( const CStr& szFromEdit,
                            _TCHAR** pStopChar,
                            T& tInternalValue ) const;

    // Return the formatter's value converted to the current units and rounded
    // to the precision of the edit format.  The formatter's value is not
    // changed.
    T RoundedOutputValue() const { return m_pFormatter->RoundedOutputValue( m_tValue ); }
    T RoundedInternalValue() const { return m_pFormatter->RoundedInternalValue( m_tValue ); }

    /// @returns the largest numeric value that when formatted to display units will display the
    /// same as the current value assuming round to nearest conversion to text.
    double RoundedInternalMin() const;
    /// @returns the smallest numeric value that when formatted to display units will display the
    /// same as the current value assuming round to nearest conversion to text.
    double RoundedInternalMax() const;

    /// Are these equal to withing the specified precision.
    bool    operator==( T value ) const
        { return m_pFormatter->Value(m_tValue) == m_pFormatter->Value(value); }
    /// Are these not equal to withing the specified precision.
    bool    operator!=( T value ) const
        { return !operator==(value); }
    /// Is it less than or with specified precision.
    bool    operator<=( const TFormattedNumeric< T > value ) const
        { return operator==(value) || m_tValue < value.m_tValue; }
    /// Is it less than or with specified precision.
    bool    operator>=( const TFormattedNumeric< T > value ) const
        { return operator==(value) || m_tValue >= value.m_tValue; }


    T Clip ();
    // Set the value such that minimum <= value <= maximum;

    /// Set the internal value to the formatter-specified default value.
    void    SetDefaults() { m_tValue = m_pFormatter->GetDefault(); }

    /// The binary size of this on Disk (Read/Write).
    size_t GetSize() const     { return sizeof m_tValue; }

    /// Read the data (just the value) from the stream.
    bool Read( std::istream &is );
    /// Write the data (just the value) to the stream.
    bool Write( std::ostream &os ) const;

    const TFormatter<T>* GetFormatter() const { return m_pFormatter; }
    void    SetRange(T min, T max) {((TFormatter<T> *)m_pFormatter)->SetRange(min, max);}

    /// Give access to the units as a json string.
    std::string JsonUnits() const;
    /// Give access to the format as a json string.
    std::string JsonFormat() const;

protected:

private:

    const TFormatter<T> *m_pFormatter;
    T m_tValue;
};

template<class T> T TFormattedNumeric<T>::OutputValue() const
{
    auto dValue = m_pFormatter->ToOutput( m_tValue );
    // Report an error for infinities or overly large numbers, but don't error
    // on quiet_NaN situations or sqrt(-1) which is also NaN
    ASSERT_ERROR( (dValue >= std::numeric_limits<T>::lowest()
                   && dValue <= std::numeric_limits<T>::max())
                   || std::isnan( dValue ) );
    return static_cast<T>( dValue );
}

template<class T>  T  TFormattedNumeric<T>::SetInternalValue(T tValue)
{
    m_tValue =  tValue;
    if(InRange())
        Clip();
    return  m_tValue;
}

template<class T> T TFormattedNumeric<T>::SetOutputValue(T tValue)
{
    m_tValue =  m_pFormatter->ToInternal( tValue);
    if(InRange())
        Clip();
    return  m_tValue;
}

template<class T> CStr TFormattedNumeric<T>::ValueAndUnits( uint_32 uiOverrideUnits ) const
{
    CStr szValue( Value() );
    CStr szUnits;
    if ( uiOverrideUnits )
        szUnits = CStr(uiOverrideUnits);
    else
        szUnits = ShortUnits();
    return  szValue + _T(" ") + szUnits;
}

template<class T> T TFormattedNumeric<T>::Clip()
{
    m_tValue = std::max( std::min(m_tValue, m_pFormatter->GetMaximum()), m_pFormatter->GetMinimum() );
    return m_tValue;
}

template<class T> bool TFormattedNumeric<T>::Read( std::istream &is )
{
    is.read( (char *)&m_tValue, sizeof m_tValue );
    return is.good() ? true : false;
}

template<class T> bool TFormattedNumeric<T>::Write( std::ostream &os ) const
{
    os.write( (char *)&m_tValue, sizeof m_tValue );
    return os.good() ? true : false;
}

template<class T> bool TFormattedNumeric<T>::CheckData(
    const  CStr & szFromEdit,
    _TCHAR** ppStopChar,
    T& tInternalValue ) const
{
    return m_pFormatter->CheckData(szFromEdit, ppStopChar, tInternalValue );
}
template<class T> T TFormattedNumeric<T>::StoreData(const  CStr & szFromEdit)
{
    _TCHAR *pStopChar;
    T tInternalValue;
    if ( CheckData( szFromEdit, &pStopChar, tInternalValue ) )
        SetInternalValue( tInternalValue );
    return m_tValue;
}

template<class T> double TFormattedNumeric<T>::RoundedInternalMin() const
{
    auto fmt = GetPrintFormat();
    auto dValue = RoundedOutputValue();
    auto dRound = std::pow(10.0, -fmt.GetPrecision())/2;
    return m_pFormatter->ToInternal(dValue - dRound);
}

template<class T> double TFormattedNumeric<T>::RoundedInternalMax() const
{
    auto fmt = GetPrintFormat();
    auto dValue = RoundedOutputValue();
    auto dRound = std::pow(10.0, -fmt.GetPrecision())/2;
    return m_pFormatter->ToInternal(dValue + dRound);
}

template<class T> std::TOSTREAM &operator<<(
    std::TOSTREAM &os,
    const TFormattedNumeric<T> &fd )
{
    os << fd.Value();
    return os;
}

template<class T> std::string TFormattedNumeric<T>::JsonUnits() const
{
    return m_pFormatter->JsonUnits();
}

template<class T> std::string TFormattedNumeric<T>::JsonFormat() const
{
    return m_pFormatter->JsonFormat();
}

#if !defined(MIC_DM643)
/// Serialize Formatted Numeric.
template<typename T> void Serialize(CDataArchive& rAr, TFormattedNumeric<T>& rxdData)
{
    if(rAr.IsLoadingValid())
        rxdData.Read(rAr.GetIstream());
    else if(rAr.IsValid())
        rxdData.Write(rAr.GetOstream());
}
#endif
//-----------------------------------------------------------------------------
// FORMATTED_* macros that define the specific classes for the
// TFormattedNumeric associated with specific formatters.

/// Macro for creating classed of TFormattedNumeric of type \a type and name \a
/// name.  The constructor can be empty, or take a numeric value.  The class
/// defines a GetStaticFormat() to give access to the created TFormatter.
///
/// For every class created, the creator must also define the "const
/// TFormatter<type> name::sm_Formatter" object that will define the formatting.
#define FORMATTED_TYPE( name, type ) \
    class name : public TFormattedNumeric<type> \
    { public: \
        static const TFormatter<type>& GetStaticFormat() { return sm_Formatter; } \
        name() : TFormattedNumeric< type >(sm_Formatter, sm_Formatter.GetDefault()) {}  \
        name( type d ) : TFormattedNumeric< type >( sm_Formatter, d ) {} \
      private: \
        static const TFormatter<type> sm_Formatter; \
    };

/// Creates a class declaration of type double with \a name
#define FORMATTED_DOUBLE( name ) FORMATTED_TYPE(name, double)
/// Creates a class declaration of type uint_16 with \a name
#define FORMATTED_UNSIGNED( name ) FORMATTED_TYPE(name, uint_16)
/// Creates a class declaration of type short with \a name
#define FORMATTED_INT( name ) FORMATTED_TYPE(name, short)
/// Creates a class declaration of type uint_32 with \a name
#define FORMATTED_ULONG( name ) FORMATTED_TYPE(name, uint_32)
/// Creates a class declaration of type long with \a name
#define FORMATTED_LONG( name ) FORMATTED_TYPE(name, long)

/// This is the similar to the FORMATTED_TYPE macro.  It just makes a class
/// that allows updating of the TFormatter at run time.  It provides Format()
/// methods for that.
#define ADJUSTABLE_FORMATTED_TYPE(name, type)                                                      \
    class name : public TFormattedNumeric<type>                                                    \
    {                                                                                              \
    public:                                                                                        \
        static const TFormatter<type>& GetStaticFormat() { return sm_Formatter; }                  \
        name() : TFormattedNumeric<type>(sm_Formatter, sm_Formatter.GetDefault()) {}               \
        name(type d) : TFormattedNumeric<type>(sm_Formatter, d) {}                                 \
        static void Format(type tMin,                                                              \
            type tMax,                                                                             \
            type tDefault,                                                                         \
            const CNumericFormat& crEdit,                                                          \
            const CNumericFormat& crPrint,                                                         \
            const CUnits& mask = CUnits())                                                         \
        {                                                                                          \
            sm_Formatter = TFormatter<type>(tMin, tMax, tDefault, crEdit, crPrint, mask);          \
        }                                                                                          \
        static void Format(TFormatter<type> formatter) { sm_Formatter = formatter; }               \
                                                                                                   \
    private:                                                                                       \
        static TFormatter<type> sm_Formatter;                                                      \
    };
/// Creates a class declaration of type double with \a name where the
/// formatting parameters are settable at run time.
#define ADJUSTABLE_FORMATTED_DOUBLE( name ) ADJUSTABLE_FORMATTED_TYPE(name, double)

#endif    // FORMATTR_HPP
