#ifndef INTERNALUNITS_HPP
#define INTERNALUNITS_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#if !defined(MIC_DM643)
#include <type_traits>
#endif
#include <vector>

// ================================================================================================
class CDataArchive;

#if !defined(MIC_DM643)
// ================================================================================================
/// The NUnits namespace supports zero overhead quantities that have units.  The NUnits::TQuantity
/// class, and the INTERNAL_UNIT macro for creating specific quantity types.
namespace NUnits
{
    struct SUnitsOfMeasure;
    template<typename T, typename Units> class TQuantity;

    template<typename T> struct is_quantity;

    /// The value template for the internal units check.
    template<typename T> inline constexpr bool is_quantity_v = NUnits::is_quantity<T>::value;
};  // namespace NUnits

// ================================================================================================
/// Base class for unit of measurement type annotations. Derived types are used for type /
/// annotation of concrete TQuantity subclasses.
///
/// Do not store data in this type or make it polymorphic. Derived types should be finalstruct
struct NUnits::SUnitsOfMeasure
{
};

// ================================================================================================
/// Zero run time overhead class that adds units to an arithmetic type.
///
/// It needs an explicit construction to create one, but has an implicit conversion to the
/// underlying type.  This way if an interface uses specific units, they have to be explicitly set,
/// but if a calculation is not unit aware, then this type can be used directly.  This is to support
/// incremental conversion of legacy code to be unit aware.
///
/// It also has implicit construction from one underlying type to another (as long as the underlying
/// types are convertible.)
///
/// Adding, subtracting, unary-, unary+ on two (or one) type will result in that same type.
/// Multiplying a quantity with units by a pure number will result in that same quantity with units.
///
/// This class also does not setup arithmetic on quantities with internal units.  (for example a cm
/// * cm united value does not return a cm2 united value.)  This could be added on a case by case
/// basis, but the design does not support a general mechanism for this as others do.  The tradeoff
/// here is for simplicity and to document (and enforce) the units on interfaces that choose to make
/// use of this.
template<typename T, typename Units> class NUnits::TQuantity final
{
public:
    // Make sure we are dealing with a simple arithmetic type.
    static_assert((std::is_integral_v<T> || std::is_floating_point_v<T>)
        && !std::is_same_v<T, bool>,
        "Must be on an arithmetic type.");
    static_assert(!std::is_const_v<T>, "Must not be a constant type.");
    static_assert(std::is_base_of_v<NUnits::SUnitsOfMeasure, Units>);
    static_assert(std::is_final_v<Units>, "Units type annotation must be final.");
    using value_type = T;  ///< The underlying arithmetic type.

    /// @{ construction and assignment.  Note that it must be explicitly constructed from the
    /// underlying type.
    explicit constexpr TQuantity(const T dVal) : m_dVal(dVal)
        {
            static_assert(sizeof(T) == sizeof(TQuantity<T, Units>(0)),
                "This must be binary compatible.");
        }
    TQuantity() = default;
    TQuantity(const TQuantity&) = default;
    TQuantity& operator=(const TQuantity&) = default;
    template<typename T2>TQuantity(NUnits::TQuantity<T2, Units> dVal)
        : TQuantity<T, Units>(dVal.operator T2()) {}
    /// @}

    /// Add conversion to the underlying type. This makes it possible to use these values in all
    /// calculations that are supported on the underlying type.
    operator T() const { return m_dVal; }

    /// @{ Also give explicit access to the underlying data.
    T& Value() { return m_dVal; }
    T Value() const { return m_dVal; }
    /// @}

private:
    T m_dVal;  ///< The actual value of the number.
};

/// @{ Allow unary - and + operators.
template<typename T, typename U>
NUnits::TQuantity<T, U> operator+(NUnits::TQuantity<T, U> rhs)
{
    return rhs;
}

template<typename T, typename U>
NUnits::TQuantity<T, U> operator-(NUnits::TQuantity<T, U> rhs)
{
    return NUnits::TQuantity<T, U>(-rhs.Value());
}
/// @}

/// @{ Allow sum and difference of two identical types with units to retain type.
template<typename T, typename U>
NUnits::TQuantity<T, U> operator+(NUnits::TQuantity<T, U> rhs, NUnits::TQuantity<T, U> lhs)
{
    return NUnits::TQuantity<T, U>(rhs.Value() + lhs.Value());
}

template<typename T, typename U>
NUnits::TQuantity<T, U> operator-(NUnits::TQuantity<T, U> rhs, NUnits::TQuantity<T, U> lhs)
{
    return NUnits::TQuantity<T, U>(rhs.Value() - lhs.Value());
}

template<typename T, typename U>
NUnits::TQuantity<T, U>& operator+=(NUnits::TQuantity<T, U>& rhs, NUnits::TQuantity<T, U> lhs)
{
    rhs = NUnits::TQuantity<T, U>(rhs.Value() + lhs.Value());
    return rhs;
}

template<typename T, typename U>
NUnits::TQuantity<T, U>& operator-=(NUnits::TQuantity<T, U>& rhs, NUnits::TQuantity<T, U> lhs)
{
    rhs = NUnits::TQuantity<T, U>(rhs.Value() - lhs.Value());
    return rhs;
}
/// @}

/// @{ Allow multiplication by arithmetic types.
template<typename T, typename T2, typename U, std::enable_if_t<std::is_arithmetic_v<T2>>* = nullptr>
auto operator*(NUnits::TQuantity<T, U> rhs, T2 lhs)
{
    return NUnits::TQuantity<decltype(rhs.Value()*lhs), U>(rhs.Value() * lhs);
}

template<typename T, typename T2, typename U, std::enable_if_t<std::is_arithmetic_v<T2>>* = nullptr>
auto operator*(T2 rhs, NUnits::TQuantity<T, U> lhs)
{
    return NUnits::TQuantity<decltype(rhs*lhs.Value()), U>(rhs * lhs.Value());
}

template<typename T, typename T2, typename U, std::enable_if_t<std::is_arithmetic_v<T2>>* = nullptr>
auto operator/(NUnits::TQuantity<T, U> rhs, T2 lhs)
{
    return NUnits::TQuantity<decltype(rhs.Value()/lhs), U>(rhs.Value() / lhs);
}
/// @}

/// @{ Allow in place multiplication and division of types with units by unitless quantities.
template<typename T, typename U, typename T2>
NUnits::TQuantity<T, U>& operator*=(NUnits::TQuantity<T, U>& rhs, T2 lhs)
{
    static_assert(std::is_arithmetic_v<T2>, "can only scale by numbers.");
    rhs = NUnits::TQuantity<T, U>(rhs.Value() * lhs);
    return rhs;
}

template<typename T, typename U, typename T2>
NUnits::TQuantity<T, U>& operator/=(NUnits::TQuantity<T, U>& rhs, T2 lhs)
{
    static_assert(std::is_arithmetic_v<T2>, "can only scale by numbers.");
    rhs = NUnits::TQuantity<T, U>(rhs.Value() / lhs);
    return rhs;
}
/// @}

/// The template structure that can check if a type is an internal units type.
template<typename T> struct NUnits::is_quantity : std::false_type
{
};

/// The specialization that sets the value to true for internal unit parameters.
template<typename T, typename Units>
struct NUnits::is_quantity<NUnits::TQuantity<T, Units>> : std::true_type
{
};

/// Allow internal unit types to be serialized the same as the underlying arithmetic type.  That way
/// changing a serialized data member from type T to TQuantity<T, xxx> does not change the
/// way data is serialized.
template<typename T, typename Units>
void Serialize(CDataArchive& rAr, NUnits::TQuantity<T, Units>& rData)
{
    Serialize(rAr, rData.Value());
}

/// Define an internal unit.  This means it has a double_<unit> and float_<unit> types, and a _<unit>
/// user defined literal suffix.
///
/// So, since we have a cm, and cm3 internal unit type, we can do the following:
/// using namespace SI;
/// double_cm d = 1.0_cm;
/// float_cm f = 1.0_cm;
/// but then d = 1.0; would fail, as would d = 1.0_cm3. as would f = d
/// To convert from one type of float to another you need an explicit cast (with value).
///   d = double_cm(f1.Value());
///   f1 = float_cm(static_cast<float>(f1.Value()));
/// This does not work: d = double_cm(f1) but I wish it would.
#define INTERNAL_UNIT(ns, unit)                                                                       \
    namespace NUnits                                                                                  \
    {                                                                                                 \
        struct unit final : SUnitsOfMeasure                                                           \
        {                                                                                             \
        };                                                                                            \
    };                                                                                                \
    namespace ns                                                                                      \
    {                                                                                                 \
        template<typename T> using unit = NUnits::TQuantity<T, NUnits::unit>;                         \
        namespace V                                                                                   \
        {                                                                                             \
            template<typename T> using unit = std::vector<NUnits::TQuantity<T, NUnits::unit>>;        \
        }                                                                                             \
    };                                                                                                \
    namespace ns::literal                                                                             \
    {                                                                                                 \
        inline constexpr NUnits::TQuantity<double, NUnits::unit> operator"" _##unit(long double dVal) \
        {                                                                                             \
            return NUnits::TQuantity<double, NUnits::unit>(double(dVal));                             \
        };                                                                                            \
    };

// ================================================================================================
/// Make sure the class is really zero overhead.
namespace NUnits
{
    struct SNoUnits final : SUnitsOfMeasure
    {
    };
    // Make sure these are really zero overhead.
    static_assert(sizeof(double) == sizeof(TQuantity<double, SNoUnits>));
    static_assert(sizeof(float) == sizeof(TQuantity<float, SNoUnits>));
    static_assert(sizeof(uint_16) == sizeof(TQuantity<uint_16, SNoUnits>));
    static_assert(sizeof(uint_8) == sizeof(TQuantity<uint_8, SNoUnits>));
}  // namespace NUnits

#else // !defined(MIC_DM643)
// For the DM643 platform, just define a new class in the right namespace for each unit that
// converts to and from the underlying type.
// The main difference is that for this platform it does not have user defined constants of the form
// 1.0_cm3.  And it allows direct assignment from a double to a SI::cm3<double>.

#define INTERNAL_UNIT(ns, unit)                       \
    namespace ns                                      \
    {                                                 \
        template<typename T> class unit               \
        {                                             \
        public:                                       \
            unit() {}                                 \
            unit(const T& dVal) : m_dVal(dVal) {}     \
            operator T() const { return m_dVal; }     \
        private:                                      \
            T m_dVal;                                 \
        };                                            \
    };                                                \

#endif

// ================================================================================================
// Define zero overhead quantities with units for common units. Note the conventions:
// 1. separate products with _
// 2. separate numerator units from denominator units with _per_.
// 3. for powers add the number after the unit with no separator.
//! TODO: These are the internal ones used in user selectable quantities.  We should review all the
// non user changeable ones and add those here as well.  That will make it possible to port
// application level code without adding quantities here.
#if defined min
#define MIC_SAVE_MIN min
#undef min
#endif
INTERNAL_UNIT(SI, pure) // no units.
INTERNAL_UNIT(SI, A)  // Angstrom
INTERNAL_UNIT(SI, A2)
INTERNAL_UNIT(SI, Amp)  // amperes.
INTERNAL_UNIT(SI, C)
INTERNAL_UNIT(SI, C_per_min)
INTERNAL_UNIT(SI, K)
INTERNAL_UNIT(SI, L) // volume: Liter
INTERNAL_UNIT(SI, S_per_m)  // Siemens per meter
INTERNAL_UNIT(SI, cm)
INTERNAL_UNIT(SI, cm2)
INTERNAL_UNIT(SI, cm3)
INTERNAL_UNIT(SI, cm3_STP_per_g)
INTERNAL_UNIT(SI, cm3_per_g)
INTERNAL_UNIT(SI, cm3_per_min)
INTERNAL_UNIT(SI, cm_per_s)
INTERNAL_UNIT(SI, cp)  // centipose (viscosity)
INTERNAL_UNIT(SI, dynes_per_cm3)
INTERNAL_UNIT(SI, e_per_k)
INTERNAL_UNIT(SI, g)
INTERNAL_UNIT(SI, g_per_cm3)
INTERNAL_UNIT(SI, g_per_mL)
INTERNAL_UNIT(SI, g_per_ml)
INTERNAL_UNIT(SI, hrs) // time: hours
INTERNAL_UNIT(SI, in)
INTERNAL_UNIT(SI, in2)
INTERNAL_UNIT(SI, kJ_per_mol)  // kilojoules per mole.
INTERNAL_UNIT(SI, kPa_per_min)
INTERNAL_UNIT(SI, lbf)  // pounds force.
INTERNAL_UNIT(SI, m2)
INTERNAL_UNIT(SI, m2_per_g)
INTERNAL_UNIT(SI, m2_per_gA)
INTERNAL_UNIT(SI, mAmp)  // miliamperes.
INTERNAL_UNIT(SI, mL)
INTERNAL_UNIT(SI, mg) // milligram.
INTERNAL_UNIT(SI, min) // time: minutes.
INTERNAL_UNIT(SI, ml)  // mililiters
INTERNAL_UNIT(SI, ml_per_g)
INTERNAL_UNIT(SI, ml_per_g_s)
INTERNAL_UNIT(SI, mmHg) // pressure: millimeters of Mercury.
INTERNAL_UNIT(SI, mol) // mole
INTERNAL_UNIT(SI, nm)
INTERNAL_UNIT(SI, nm2)
INTERNAL_UNIT(SI, ohms)
INTERNAL_UNIT(SI, pct)  // percent
INTERNAL_UNIT(SI, pct_per_min)
INTERNAL_UNIT(SI, pct_per_min2)
INTERNAL_UNIT(SI, per_A)
INTERNAL_UNIT(SI, per_ml)
INTERNAL_UNIT(SI, per_psia)
INTERNAL_UNIT(SI, per_psia2)
INTERNAL_UNIT(SI, psi_per_min)
INTERNAL_UNIT(SI, psia)
INTERNAL_UNIT(SI, psia_per_V)
INTERNAL_UNIT(SI, psig)
INTERNAL_UNIT(SI, s)
INTERNAL_UNIT(SI, uL_per_pF)
INTERNAL_UNIT(SI, ul)  // microleters
INTERNAL_UNIT(SI, um)
INTERNAL_UNIT(SI, um2)
INTERNAL_UNIT(SI, um3)
INTERNAL_UNIT(SI, umHg) // pressure: micrometers of Mercury.
INTERNAL_UNIT(SI, us)  // microseconds

#if defined MIC_SAVE_MIN
#define min MIC_SAVE_MIN
#undef MIC_SAVE_MIN
#endif


#endif // INTERNALUNITS_HPP
