#ifndef BOOST_PYTHON_CONVERT_HPP
#define BOOST_PYTHON_CONVERT_HPP
//-------------------------------------------------------------------------------------------------/
// @file
/// Conversion of generic aggregates to and from python.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "../cstr.hpp"
#include "../formattr.hpp"
#include "../tmdate.hpp"
#include "../internalunits.hpp"
#include <boost/python/class.hpp>
#include <boost/python/def.hpp>
#include <boost/python/list.hpp>
#include <boost/python/tuple.hpp>
#include <boost/python/to_python_converter.hpp>
#include <datetime.h>

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
template<typename T> struct IsStdArray : std::false_type
{
};
template<typename T, std::size_t N> struct IsStdArray<std::array<T, N>> : std::true_type
{
};

template<typename T> struct IsStdVector : std::false_type
{
};
template<typename T> struct IsStdVector<std::vector<T>> : std::true_type
{
};

// ================================================================================================
// Convert from CStr to python.
struct cstr_to_python_str final
{
    static PyObject* convert(const CStr& szStr);
};

// ================================================================================================
// Convert from python to CStr
struct cstr_from_python_str final
{
    cstr_from_python_str();
    static void* convertible(PyObject* obj_ptr);
    static void construct(PyObject* obj_ptr,
        boost::python::converter::rvalue_from_python_stage1_data* data);
};

// ================================================================================================
namespace NUnits
{
    // Convert from TQuantity to python.
    template<typename T>
    struct tquantity_to_python_float final
    {
        static_assert(is_quantity_v<T>, "Must be NUnits::TQuantity type.");
        static PyObject* convert(const T& dValue);
    };

    // Convert from python to TQuantity
    template<typename T>
    struct tquantity_from_python_float final
    {
        static_assert(is_quantity_v<T>, "Must be NUnits::TQuantity type.");
        tquantity_from_python_float();
        static void* convertible(PyObject* obj_ptr);
        static void construct(PyObject* obj_ptr,
            boost::python::converter::rvalue_from_python_stage1_data* data);
    };
}

// ================================================================================================
/// Convert an std:array<char, N> to python string
template<typename T> struct array_char_to_python_str final
{
    static_assert(IsStdArray<T>::value, "Can only be used for array types.");

    static PyObject* convert(const T& szStr);
};

// ================================================================================================
/// Convert a python string to an std:array<char, N>.
template<typename T> struct array_char_from_python_str final
{
    static_assert(IsStdArray<T>::value, "Can only be used for array types.");

    array_char_from_python_str();
    static void* convertible(PyObject* obj_ptr);
    static void construct(PyObject* obj_ptr,
        boost::python::converter::rvalue_from_python_stage1_data* data);
};

// ================================================================================================
/// Converter from std::pair to a python tuple.
template<typename T1, typename T2> struct pair_to_py_tuple final
{
    static PyObject* convert(const std::pair<T1, T2>& c_pair);
};

// ================================================================================================
/// Converter from std::array to a python tuple.
template<typename T> struct array_to_py_tuple final
{
    static_assert(IsStdArray<T>::value || IsStdVector<T>::value,
        "Can only be used for array/vector types.");

    static PyObject* convert(const T& c_array);
};

// ================================================================================================
/// Converter to std::array from a python tuple/sequence.
template<typename T> struct array_from_py_sequence final
{
    static_assert(IsStdArray<T>::value, "Can only be used for array types.");

    static void* convertible(PyObject* py_obj);
    static void construct(PyObject* py_obj,
        boost::python::converter::rvalue_from_python_stage1_data* data);
};

// ================================================================================================
/// Converter to std::array from a python tuple/sequence.
template<typename T> struct vector_from_py_sequence final
{
    static_assert(IsStdVector<T>::value, "Can only be used for array types.");

    static void* convertible(PyObject* py_obj);
    static void construct(PyObject* py_obj,
        boost::python::converter::rvalue_from_python_stage1_data* data);
};

// ================================================================================================
template<typename T> static void ExposeFormatted(const char* szName)
{
    using namespace boost::python;
    class_<T, bases<TFormattedNumeric<typename T::TNumericType>>>(szName,
        init<typename T::TNumericType>())
        .def(init<>());
}

// ================================================================================================
template<typename T> static void ExposeQuantity(const char* szName)
{
    using namespace boost::python;
    using namespace NUnits;
    static_assert(is_quantity_v<T>, "Must be of type NUnits::TQuantity.");

    to_python_converter<T, NUnits::tquantity_to_python_float<T>>();
    to_python_converter<std::vector<T>, array_to_py_tuple<std::vector<T>>>();

    NUnits::tquantity_from_python_float<T>();
    converter::registry::push_back(&vector_from_py_sequence<std::vector<T>>::convertible,
        &vector_from_py_sequence<std::vector<T>>::construct,
        type_id<std::vector<T>>());
}

// ================================================================================================
inline PyObject* cstr_to_python_str::convert(const CStr& szStr)
{
    return boost::python::incref(boost::python::object(szStr.GetStringPtr()).ptr());
}

inline cstr_from_python_str::cstr_from_python_str()
{
    boost::python::converter::registry::push_back(&convertible,
        &construct,
        boost::python::type_id<CStr>());
}

inline void* cstr_from_python_str::convertible(PyObject* obj_ptr)
{
    return PyUnicode_Check(obj_ptr) ? obj_ptr : 0;
}

inline void cstr_from_python_str::construct(PyObject* obj_ptr,
    boost::python::converter::rvalue_from_python_stage1_data* data)
{
    const char* value = PyUnicode_AsUTF8(obj_ptr);
    if(value == 0)
        boost::python::throw_error_already_set();
    void* storage
        = ((boost::python::converter::rvalue_from_python_storage<CStr>*)data)->storage.bytes;
    new(storage) CStr(value);
    data->convertible = storage;
}

template<typename T>
PyObject* NUnits::tquantity_to_python_float<T>::convert(const T& dValue)
{
    return boost::python::incref(boost::python::object(dValue.Value()).ptr());
}

template<typename T>
NUnits::tquantity_from_python_float<T>::tquantity_from_python_float()
{
    boost::python::converter::registry::push_back(&convertible,
        &construct,
        boost::python::type_id<T>());
}

template<typename T>
void* NUnits::tquantity_from_python_float<T>::convertible(PyObject* obj_ptr)
{
    return PyFloat_Check(obj_ptr) || PyLong_Check(obj_ptr) ? obj_ptr : 0;
}

template<typename T>
void NUnits::tquantity_from_python_float<T>::construct(PyObject* obj_ptr,
    boost::python::converter::rvalue_from_python_stage1_data* data)
{
    using S = typename T::value_type;
    auto dValue = S();
    if(PyFloat_Check(obj_ptr))
        dValue = S(PyFloat_AsDouble(obj_ptr));
    else if(PyLong_Check(obj_ptr))
        dValue = S(PyLong_AsLong(obj_ptr));
    void* storage
        = ((boost::python::converter::rvalue_from_python_storage<T>*)data)->storage.bytes;
    new(storage) T(dValue);
    data->convertible = storage;
}

template<typename T> PyObject* array_char_to_python_str<T>::convert(const T& szStr)
{
    return boost::python::incref(boost::python::object(&szStr[0]).ptr());
}

template<typename T> array_char_from_python_str<T>::array_char_from_python_str()
{
    boost::python::converter::registry::push_back(&convertible,
        &construct,
        boost::python::type_id<T>());
}

template<typename T> void* array_char_from_python_str<T>::convertible(PyObject* obj_ptr)
{
    return PyUnicode_Check(obj_ptr) ? obj_ptr : 0;
}

template<typename T>
void array_char_from_python_str<T>::construct(PyObject* obj_ptr,
    boost::python::converter::rvalue_from_python_stage1_data* data)
{
    auto value = PyUnicode_AsUTF8(obj_ptr);
    if(value == 0)
        boost::python::throw_error_already_set();
    auto storage = ((boost::python::converter::rvalue_from_python_storage<CStr>*)data)->storage.bytes;
    auto pArray = new(storage) T();
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstringop-overflow="
    // I believe this is correct because of the storage requirements calculated above, and because
    // it has not caused issues in the 1350 over much testing on real and simulated instruments.  It
    // does give a warning in the 1350 cross compile, but not in the 1350 native compile.
    std::strncpy(&(*pArray)[0], value, pArray->size());
#pragma GCC diagnostic pop
    pArray->back() = 0;
    data->convertible = storage;
}

template<typename T1, typename T2>
PyObject* pair_to_py_tuple<T1, T2>::convert(const std::pair<T1, T2>& c_pair)
{
    boost::python::list values;
    values.append(c_pair.first);
    values.append(c_pair.second);
    return boost::python::incref(boost::python::tuple(values).ptr());
}

template<typename T> PyObject* array_to_py_tuple<T>::convert(const T& c_array)
{
    boost::python::list values;
    for(auto& v : c_array)
    {
        if constexpr(NUnits::is_quantity_v<T>)
           values.append(v.Value());
        else
            values.append(v);
    }
    return boost::python::incref(boost::python::tuple(values).ptr());
}

template<typename T> void* array_from_py_sequence<T>::convertible(PyObject* py_obj)
{
    if(!PySequence_Check(py_obj))
        return 0;

    if(!PyObject_HasAttrString(py_obj, "__len__"))
        return 0;

    auto py_sequence
        = boost::python::object(boost::python::handle<>(boost::python::borrowed(py_obj)));
    if(std::tuple_size<T>::value != len(py_sequence))
        return 0;

    return py_obj;
}

template<typename T>
void array_from_py_sequence<T>::construct(PyObject* py_obj,
    boost::python::converter::rvalue_from_python_stage1_data* data)
{
    typedef boost::python::converter::rvalue_from_python_storage<T> storage_t;
    auto the_storage = reinterpret_cast<storage_t*>(data);
    auto memory_chunk = the_storage->storage.bytes;
    auto c_array = new(memory_chunk) T();
    data->convertible = memory_chunk;

    auto py_sequence
        = boost::python::object(boost::python::handle<>(boost::python::borrowed(py_obj)));
    for(int i = 0; i < boost::python::len(py_sequence) && i < c_array->size(); i++)
        (*c_array)[i] = boost::python::extract<typename T::value_type>(py_sequence[i]);
}

template<typename T> void* vector_from_py_sequence<T>::convertible(PyObject* py_obj)
{
    if(!PySequence_Check(py_obj))
        return 0;

    if(!PyObject_HasAttrString(py_obj, "__len__"))
        return 0;

    auto py_sequence
        = boost::python::object(boost::python::handle<>(boost::python::borrowed(py_obj)));

    return py_obj;
}

template<typename T>
void vector_from_py_sequence<T>::construct(PyObject* py_obj,
    boost::python::converter::rvalue_from_python_stage1_data* data)
{
    typedef boost::python::converter::rvalue_from_python_storage<T> storage_t;
    auto the_storage = reinterpret_cast<storage_t*>(data);
    auto memory_chunk = the_storage->storage.bytes;
    auto c_vect = new(memory_chunk) T();
    data->convertible = memory_chunk;

    auto py_sequence
        = boost::python::object(boost::python::handle<>(boost::python::borrowed(py_obj)));
    for(int i = 0; i < boost::python::len(py_sequence); i++)
        c_vect->emplace_back(boost::python::extract<typename T::value_type>(py_sequence[i]));
}

#endif  // BOOST_PYTHON_CONVERT_HPP
