#ifndef LOG_QUERY_PYTHON_UI_HPP
#define LOG_QUERY_PYTHON_UI_HPP
//-------------------------------------------------------------------------------------------------
/// @file
/// The log-query interface to python ui and storage.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "../log-query.hpp"
#include "../cstr.hpp"

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
class CEventWait;

namespace NLogQuery
{
    /// Initialize the logging system for a python script.  This calls the CLogUIPython::Init()
    /// method, and also sets up the CEventWait class that is needed for the interaction with the
    /// UI.  This can be called any number of times.  Only the first time does it do anything.
    void InitPythonWithEvent(const CStr& cszRootUrl);
    /// Put up a message on the instrument UI and wait for the user to press the OK or cancel
    /// buttons. If the user presses OK, this will return and allow the python script to continue
    /// executing.  If the user presses cancel, it will terminate the process (that is assumed to be
    /// a python script).
    void PythonOKCancel(const CStr& cszMessage);
    /// Put up a message on the instrument UI and wait for the user to press the yes, no or cancel
    /// buttons. If the user presses cancel, it will terminate the process (that is assumed to be
    /// a python script).  Otherwise it will return the CQueryResponse object.
    CQueryResponse PythonYesNoCancel(const CStr& cszMessage);
    /// Put up the data entry screen on the UI for the given context, and wait for the user to
    /// press the OK or cancel buttons. If the user presses OK, this will return and allow the
    /// python script to continue executing.  If the user presses cancel, it will terminate the
    /// process (that is assumed to be a python script).
    void PythonGetUserData(CLogUIInput& rInput, int iContext);
};

// ================================================================================================
/// The log-query interface to python ui and storage.
///
/// The main application process should initialize the NLogQuery system by Calling
/// Init(cszRootUrl).  This will associate all the (to be created) processes with the rest api at
/// the django python server rooted at cszRootUrl.  Then each process can use the (non modal)
/// NLogQuery methods to send information to the user or system log file.
///
/// If a process wants to use the modal queries (InformaWait, OKCancel etc.)  it must first
/// establish the link by calling SetEventWait() in the background process.  Then the application
/// UDP interface needs to expose the CLogUIPython::UserResponded() call that is the
/// callback from the UI.
class CLogUIPython : public CLogUIInterface
{
public:

    /// Initialize the log/query system with this interface.  This will check to make sure that the
    /// rest server is up, and if not, will continue to wait until it does come up.
    static void Init(const CStr& cszRootUrl);

    /// Set the wait event that is used by the ShowModal() this must be called at the beginning of
    /// the background process function if that function will allow the use of Modal Inform or query
    /// calls.  The event must continue to exist for the life of the process.
    static void SetEventWait(CEventWait& rEventWait);

    /// Called to indicate that the user responded to a message with CModalContext.  This will
    /// signal the correct process to stop waiting.
    ///
    /// @param iPID the process id of the process to signal.  This is set by the ShowModal()
    /// @param eResp The response from the user.
    static void UserResponded(CModalContext iPID, EUserResponse eResp);

    /// Called to indicate that the user responded to an input request with CModalContext.  This
    /// will signal the correct process to stop waiting.
    ///
    /// The data in cszDataJson is validated on the input side, so the values meet their formatting
    /// criteria if eResp is not a cancel.
    ///
    /// @param iPID the process id of the process to signal.  This is set by the ShowModal()
    /// @param eResp The response from the user.
    /// @param cszDataJson The data from the user.
    static void UserInputResponded(CModalContext iPID,
        EUserResponse eResp,
        const std::string& cszDataJson);

    CLogUIPython(const CLogUIPython&) = delete;
    CLogUIPython(CLogUIPython&& ) = delete;
    CLogUIPython& operator=(const CLogUIPython&) = delete;
    virtual ~CLogUIPython() = default;

private:

    CLogUIPython(const CStr& cszRootUrl);

    virtual bool CreateDB() override;
    virtual bool OpenDB() override;
    virtual VLogEntry GetData(const CStr& szLogFile,
        int64_t iFrom,
        int64_t iTo,
        std::vector<ESeverity>& veSeverities) override;
    virtual bool DidAppCrash() override;
    virtual void AddLogEntry(const SLogEntry& cLogEntry) override;
    virtual EUserResponse ShowModal(const SLogEntry& cLogEntry, CModalContext iPID) override;
    virtual EUserResponse GetUserData(CLogUIInput& rInput, int iInputContext) override;

    /// The root of the urls that interact with the log and query data.
    const CStr m_cszRootURL;
};

#endif // LOG_QUERY_PYTHON_UI_HPP
