#ifndef PROCESS_BASE_HPP
#define PROCESS_BASE_HPP
//-----------------------------------------------------------------------------
/// @file
/// The generic classes that support multi process work.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "../cstr.hpp"
#include "../datacont.hpp"
#include "../tmint.hpp"
#include "../internalunits.hpp"
#include <cerrno>
#include <ext/stdio_filebuf.h>
#include <functional>
#include <map>
#include <memory>
#include <netinet/in.h>
#include <sstream>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <type_traits>
#include <unistd.h>
#include <vector>

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class CProcess;
namespace NProcessBase
{
    template<typename T>
    struct is_serialize : std::is_scalar<T> {};
    template<>
    struct is_serialize<CStr>: std::true_type {};

    template<typename... Ts>
    using enable_if_serialize = std::enable_if_t<std::conjunction_v<is_serialize<Ts>...>>;
    template<typename... Ts>
    using enable_if_scalar = std::enable_if_t<std::conjunction_v<std::is_scalar<Ts>...>>;

}

// Allow NUnits::TQuantity objects for any units to be considered a scalar just like the underlying
// type for which they are defined.
template<typename T, typename S>
struct std::is_scalar<NUnits::TQuantity<T, S>> : std::is_scalar<T> {};

// ================================================================================================
/// The class for objects that allow a process to wait for a time, or an event on a file.
///
/// To use, just create an object, and add the event handlers for the files of interest with the
/// AddReadEventHandler(), AddWriteEventHandler(), or AddErroreventhandler().  Then call
/// WaitTimeOrEvent() to do the waiting.
class CEventWait final
{
    friend CProcess;

public:
    CEventWait() = default;
    CEventWait(const CEventWait&) = delete;
    CEventWait& operator=(const CEventWait&) = delete;
    ~CEventWait() = default;

    /// Add an event handler for a read event on a file.
    void AddReadEventHandler(int iFileID, std::function<void()> fcnHandler);
    /// Add an event handler for a write event on a file.
    void AddWriteEventHandler(int iFileID, std::function<void()> fcnHandler);
    /// Add an event handler for an error event on a file.
    void AddErrorEventHandler(int iFileID, std::function<void()> fcnHandler);

    /// Wait for the specified time, or an event to occur.
    /// @returns true if the time was expired without an event.
    bool WaitTimeOrEvent(CTimeInterval waittime);

    /// Wait for the specified time.  If an event occurs, handle it, but continue waiting.
    void WaitTime(CTimeInterval waittime);

private:
    /// Which type of event to wait for.
    enum EEventType
    {
        ceRead = 0,
        ceWrite,
        ceError,
        ceEventTypeNum,
    };

    /// The data for each event handler.
    struct SEntry
    {
        SEntry(int iFileID, EEventType eType, std::function<void()> fcnHandler);

        int iFileID;
        EEventType eType;
        std::function<void()> fcnHandler;
    };

    /// Add a generic event handler to the list.
    void AddEventHandler(int iFileID, EEventType eType, std::function<void()> fcnHandler);

    /// All the known event handlers.
    std::vector<SEntry> m_vsEventHandler;
};

// ================================================================================================
///  This is an RAII lock on the shared resources of a specific CProcess.  See the CProcess::Lock()
/// method.
class CProcessLock final
{
    friend class CProcess;

public:
    CProcessLock(CProcessLock&& rProcessLock) noexcept;
    CProcessLock(const CProcessLock&) = delete;
    CProcessLock& operator=(const CProcessLock&) = delete;
    ~CProcessLock();

private:
    CProcessLock(const CProcess* pProcess, unsigned short uiSem);

    /// The associated process for this lock.
    const CProcess* m_pProcess;

    /// Which semaphore is from the process is used to set to lock.
    const unsigned short m_cuiSem;

    DECLARE_MSG()
};

// ================================================================================================
/// The base class process that runs a function in a forked process.
///
/// You can set the process function to run in the forked process either on construction, or with
/// the SetProcessFcn() method.
///
/// To fork the process and run the function call RunProcess().  Then PID() will return 0 in the
/// forked process, and the PID of the forked process in the main process.
///
/// This also has the Lock() method to give single user access to any shared resources for the
/// process.
///
/// Finally, each process contains a CWaitEvent (accessible with the Wait() method).
class CProcess
{
    friend CProcessLock;

public:
    using FnProcess = std::function<void()>;

    /// Monitor all the processes and when one fails, kill all the others.
    static void Monitor();

    /// Kill all the child processes.
    static void KillProcesses();

    /// Construct the process
    /// @param fcnProcess The function that is called when the process runs.
    CProcess(FnProcess fcnProcess = FnProcess());
    CProcess(const CProcess&) = delete;
    CProcess(CProcess&&); // Not Implemented.
    CProcess& operator=(const CProcess&) = delete;
    virtual ~CProcess();

    /// Set the process function if it was not set at construction.
    void SetProcessFcn(FnProcess fcnProcess);

    /// The command to fork and run the process.  It returns to the calling process.
    ///
    /// In debug mode (_DEBUG), if cszDbgGdbEnv is given, the application will show the name and
    /// process ID for the forked process.  This way gdb can attach to that process.  If an
    /// environment variable with that name is set, it will also sleep forever so that the gdb
    /// attach can link to the process before anything is done.
    ///
    /// @param cszDbgGdbEnv: The name of the process and environment variable.
    void RunProcess(const CStr& cszDbgGdbEnv = CStr());

    /// @returns the PID of the forked process, or false if this is the forked process.
    pid_t PID() const;

    /// @returns the wait CEventWait associated with this process.
    CEventWait& Wait();

    /// @return the RAII lock for this process for shared memory.  The method cannot be used before
    /// RunProcess();
    CProcessLock Lock() const;

protected:
    /// @return the RAII lock for this process for the message queue.  The method cannot be used
    /// before RunProcess();
    CProcessLock LockMessageQueue() const;

private:
    /// Link this object to the event wait object.
    virtual void AddWaitHandler(CEventWait& rWait) {};

    /// The function to run the process.
    FnProcess m_fcnProcess;

    /// The process id that was forked of to another process, or false if it this is the forked off
    /// process.
    pid_t m_PID = 0;

    /// The Semaphore ID for this object.
    int m_iSemID = 0;

    /// The wait object for this process.
    CEventWait m_wait;

#if defined(_DEBUG)
    /// The number of locks that exist for this class.  Purely a debug safety check to assert that
    /// on destruction, there are no locks.
    mutable int m_iLockCount = 0;
#endif

    /// The list of all process objects.
    static std::vector<CProcess*> ms_vpProcess;
    DECLARE_MSG()
};

// ================================================================================================
/// This adds a queue of requests between the requesting processes and the working process that this
/// class manages. The specific commands must be defined at a higher level.
///
/// Both the requesting and working process share a set of ids that uniquely identify the request.
/// Each request also has one (or no) piece of data that is part of the request.  That piece of data
/// must be serializable through the CDataArchive interface.
///
/// On construction of the object, the request handlers have to be set with the SetRequestMap()
/// method.
///
/// The requesting process calls SendReq(iID, data) to send a request over the queue/pipe.
/// The forked process will receive the request, create the data object, read it from the pipe,
/// Serialize it, and call the user supplied handler.
///
/// The receipt of the requests is done during a CEventWait wait operation.
///
/// Finally, this class has support for requests to update the global user settings in the
/// subprocesses. This is done with the public Update* methods. If there are nested processes, those
/// are handled as well, but only if this class knows about them through the AddSubprocess() method.
/// The application needs to call the static SetGettextDomain() before any subprocesses are forked
/// to make sure that the domain is known on setting the language.
class CQueueProcess : public CProcess
{
public:
    using FnReceiveReq = std::function<void(CDataArchive&)>;
    using CIDCallBackMap = std::map<int, FnReceiveReq>;

    /// Set the text domain for the product.
    static void SetGettextDomain(const char* cszGettextDomain);
    /// Accessor to the static text domain for the product.
    static const char* GetGettextDomain();

    /// @{
    /// MakeReq are helper functions to setup the table that is passed to the SetRequestMap()
    /// method.  They take either an object pointer and a member function with zero to three data
    /// parameters, or they take a std::function that takes zero or one data parameter.  The single
    /// parameter can either be a const reference parameter, or, if it is a scalar, a value
    /// parameter.  The data member is optional for those messages that do not need data.
    /// @returns a FnReceiveReq that does the type handling for the reading.
    template<typename T> static FnReceiveReq MakeReq(T* pC, void (T::*pFcn)());
    template<typename T, typename TDATA>
    static FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(const TDATA&));
    template<typename T, typename TDATA, std::enable_if_t<std::is_scalar_v<TDATA>>* = nullptr>
    static FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(TDATA));
    template<typename T, typename TDATA1, typename TDATA2, NProcessBase::enable_if_serialize<TDATA1, TDATA2>* = nullptr>
    static FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(TDATA1, TDATA2));
    template<typename T,
        typename TDATA1,
        typename TDATA2,
        typename TDATA3,
        NProcessBase::enable_if_serialize<TDATA1, TDATA2, TDATA3>* = nullptr>
    static FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(TDATA1, TDATA2, TDATA3));
    static FnReceiveReq MakeReq(std::function<void()> fcn);
    template<typename T> static FnReceiveReq MakeReq(std::function<void(const T&)> fcn);
    template<typename T, std::enable_if<std::is_scalar_v<T>>* = nullptr>
    static FnReceiveReq MakeReq(std::function<void(T)> fcn);
    /// @}

    /// @param fBlock If false, send non-blocking requests. The process reading the request still
    /// blocks. SendReq ensures non-blocking requests are either sent atomically or dropped.
    /// Callers should check SendReq's return value and take appropriate actions if a
    /// non-blocking request is dropped. WARNING: if it fails to set the non-blocking flag
    /// the process will block; call IsBlocking() to check.
    CQueueProcess(FnProcess fcnProcess = FnProcess(), bool fBlock = true);
    CQueueProcess(const CQueueProcess&) = delete;
    CQueueProcess(CQueueProcess&&); // Not Implemented.
    CQueueProcess& operator=(const CQueueProcess&) = delete;
    virtual ~CQueueProcess() = default;

    /// Set the callback functions for each ID that is sent.
    /// @param rIDMap:  This is a map of iID to FnReceiveReq.  This calls the FnReceiveReq method
    /// when a request comes (to the "background" process) through the pipe from the SendReq(iID,
    /// data) calls from other processes.
    void SetRequestMap(CIDCallBackMap&& rIDMap);

    /// @{
    /// Send the serializable object(s) to the process with ID iID.  Note that if the process is
    /// blocking this will call CProcess::LockMessageQueue() to make sure that we have sequential
    /// access to the pipe.  So, it must not be called if the thread has a lock on the object.
    /// @return True if request was sent. For blocking processes, requests always send.
    /// For non-blocking processes, requests fail if the data is too big for an atomic non-blocking
    /// write or if the pipe is too full to write the request immediately.
    bool SendReq(int iID);
    template<typename T> bool SendReq(int iID, const T& cData);
    template<typename T1, typename T2> bool SendReq(int iID, const T1& cData1, const T2& cData2);
    template<typename T1, typename T2, typename T3>
    bool SendReq(int iID, const T1& cData1, const T2& cData2, const T3& cData3);
    /// @}

    /// @return True if process blocks when sending requests. Use this to check if process was
    /// constructed as intended.
    bool IsBlocking() const;

    /// Adds a sub-process to its list.  The updates will be propagated to it as well.  This can be
    /// called multiple times.
    void AddSubprocess(CQueueProcess* pProcess);

    /// Set the units to the selections identified by the json encoded @p cszSelectionJson in the
    /// background processes.
    void UpdateUnitSelections(const std::string& cszSels);

    /// Set the localeto  @p cszLocale in the background processes.
    void UpdateTranslationLocale(const std::string& cszLocale);

    /// Set the timezone to  @p cszTimeZone in the background processes.
    void UpdateTimeZone(const std::string& cszTimeZone);

private:
    /// The commands to the background process.  Note that these start at a high value so as not to
    /// interfere with the user ones.
    enum EReqID
    {
        ceUpdateUnitSelections = 10000,  ///< Request id for UpdateUnitSelections.
        ceUpdateTranslationLocale,       ///< Request id for UpdateTranslationLocale
        ceUpdateTimeZone,                ///< Request id for UpdateTimeZone
    };

    virtual void AddWaitHandler(CEventWait& rWait) override;

    /// @return True if the request was sent.
    bool SendRequest(int iID, std::ostringstream& rOs);

    /// There is a request in the queue, read it and forward.
    void OnReadReq();

    /// Update the unit Selections
    void DoUpdateUnitSelections(const std::string& cszSelectionJson);

    /// Update the translations for strings.
    void DoUpdateTranslationLocale(const std::string& cszLocale);

    /// Update the timezone for CDateTime.
    void DoUpdateTimeZone(const std::string& cszTimeZone);

    /// The map the associates message IDs with the callback function.
    CIDCallBackMap m_idCallbackMap;

    /// The gettext domain to be used for string translations.
    static const char* ms_cszGettextDomain;

    /// Buffer for non-blocking requests.
    static std::vector<char> ms_vszBuffer;

    /// @{
    /// The pipe handles for the pipe connecting the processes.
    int m_iReadID;
    int m_iWriteID;
    /// @}

    bool m_fBlocking;  ///< True if process blocks when sending a request.

    /// The objects that also need to be updated when this one is updated.
    std::vector<CQueueProcess*>  m_SubprocessVect;

    DECLARE_MSG()
};

// ================================================================================================
/// This is a CQueueProcess that also creates a shared memory object of type T.  Note that T must be
/// a POD type.
///
/// Access to the data is through the GetData() methods.  Note that accessing members of the data
/// must be guarded with a CProcess::Lock() call to ensure consistency.
///
/// For specific application use, this is the class that is most useful.  The expected usage is that
/// there is an application class that contains one of these objects.  It then adds the following
/// - Defines an enumeration for the id's for all the requests.
/// - Defines public methods for each request that the requester can call. The implementation is to
///   call the SendReq with the right data.
/// - Defines the implementation of the requests as private DoXXX methods.
/// - Link the DoXXX methods to the queue process by calling SetRequestMap in the constructor.
/// - Write accessors to the shared memory that lock access and return a copy of the information.
template<typename T> class TSharedMemQueueProcess final : public CQueueProcess
{
public:
    TSharedMemQueueProcess(FnProcess fcnProcess = FnProcess());
    TSharedMemQueueProcess(const TSharedMemQueueProcess&) = delete;
    TSharedMemQueueProcess(TSharedMemQueueProcess&&);  // Not Implemented.
    TSharedMemQueueProcess& operator=(const TSharedMemQueueProcess&) = delete;
    virtual ~TSharedMemQueueProcess();

    ///@returns a reference to the data
    T& GetData() { return *m_pData; }

    ///@returns a reference to the data
    const T& GetData() const { return *m_pData; }

private:
    /// The shared memory ID for the data object.
    int m_iDataID;

    /// The data for the shared memory.
    T* m_pData;
};

// ================================================================================================
/// The class that manages UDP requests and responses in a background process.
///
/// Each UDP request has a (uint_32) uiRequestID, followed optionally by a POD data type that
/// provides the parameters of the request.  Some messages have a response. That must also be a POD
/// data type.
///
/// For specific application use, this is the class that is most useful.  The expected usage is that
/// there is an application class that contains one of these objects.  It then adds the following
/// - Defines an enumeration for the id's for all the requests.
/// - Defines the implementation of the requests as private XXX methods.  Each of these methods
///   takes an optional request parameter (by value if it is a scalar, or by const ref if it is a
///   structure), and optionally returns a POD value that will be sent back to the requester.
/// - Link the XXX methods to the queue process by calling SetRequestMap in the constructor. The
///   MakeReq() template functions make it easy to create the table.
class CUDPProcess final : public CProcess
{
public:
    using FnReceiveReq = std::function<void(size_t, const char*)>;
    using CIDCallBackMap = std::map<uint_32, FnReceiveReq>;

    /// The UDP requests to the background process.  Note that these start at a high value so as not
    /// to interfere with the user ones.
    enum EReqID
    {
        ceUpdateUnitSelections = 10000, ///< Request id for UpdateUnitSelections
        ceSetTranslateLocale,           ///< Request id for SetTranslationLocale
        ceSetTimeZone,                  ///< Request id for SetTimeZone
        ceLogQueryUserResponded,        ///< Request id for LogQueryUserResponded
        ceLogQueryInputUserResponded,   ///< Request id for LogQueryInputUserResponded
    };

    CUDPProcess(int iPort);
    CUDPProcess(const CUDPProcess&) = delete;
    CUDPProcess& operator=(const CUDPProcess&) = delete;
    virtual ~CUDPProcess() = default;

    /// @{
    /// MakeReq are helper functions to setup the table that is passed to the SetRequestMap()
    /// method.  They take an object pointer and a member function with a zero to three data
    /// parameters. The single parameter can either be a const reference parameter, or, if it is a
    /// scalar, a value parameter.  The data member is optional for those requests that do not need
    /// data. The member function can return a POD value that will be sent back to the requester, or
    /// void if nothing is to be sent back.
    /// @returns a FnReceiveReq that does the type handling for the reading.
    template<typename T> FnReceiveReq MakeReq(T* pC, void (T::*pFcn)());
    template<typename T, typename TReq, std::enable_if_t<std::is_scalar_v<TReq>>* = nullptr>
    FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(TReq));
    template<typename T> FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(const std::string&));
    template<typename T, typename TReq1, typename TReq2, NProcessBase::enable_if_scalar<TReq1, TReq2>* = nullptr>
    FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(TReq1, TReq2));
    template<typename T, typename TReq, NProcessBase::enable_if_scalar<TReq>* = nullptr>
    FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(TReq, const std::string&));
    template<typename T, typename TReq1, typename TReq2, NProcessBase::enable_if_scalar<TReq1, TReq2>* = nullptr>
    FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(TReq1, TReq2, const std::string&));
    template<typename T,
        typename TReq1,
        typename TReq2,
        typename TReq3,
        NProcessBase::enable_if_scalar<TReq1, TReq2, TReq3>* = nullptr>
    FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(TReq1, TReq2, TReq3));
    template<typename T, typename TReq, std::enable_if_t<std::is_standard_layout_v<TReq>>* = nullptr>
    FnReceiveReq MakeReq(T* pC, void (T::*pFcn)(const TReq&));
    template<typename T, typename TRet, std::enable_if_t<std::is_standard_layout_v<TRet>>* = nullptr>
    FnReceiveReq MakeReq(T* pC, TRet (T::*pFcn)());
    template<typename T,
        typename TRet,
        typename TReq,
        std::enable_if_t<std::is_standard_layout_v<TRet> && std::is_scalar_v<TReq>>* = nullptr>
    FnReceiveReq MakeReq(T* pC, TRet (T::*pFcn)(TReq));
    template<typename T,
        typename TRet,
        typename TReq,
        std::enable_if_t<std::is_standard_layout_v<TRet> && std::is_standard_layout_v<TReq>>* = nullptr>
    FnReceiveReq MakeReq(T* pC, TRet (T::*pFcn)(const TReq&));
    /// @}

    /// Set the callback functions for each ID that is sent.
    /// @param rIDMap:  This is a map of iID to FnReceiveReq.  This calls the FnReceiveReq method
    /// when a request comes (to the "background" process) through the pipe from the SendReq(iID,
    /// data) call from other processes.
    /// @param rpvNotify:  The list of CQueueProcess that need to be notified for the internal
    /// messages.
    void SetRequestMap(CIDCallBackMap&& rIDMap, std::vector<CQueueProcess*>&& rpvNotify);

private:
    /// The listener for the udp requests.
    void ListenForRequests();

    /// notification that a request has been received.
    void OnRequest(size_t uiSize, const char* pData);

    /// Send a buffer with the reply.
    void SendReply(size_t uiSize, void* pData);

    /// Update the processes with new unit selections.
    void OnUpdateUnitSelections(const std::string& cszSelectionJson);

    /// Update the processes with new locale.
    void OnSetTranslationLocale(const std::string& cszLocale);

    /// Update the processes with new timezone
    void OnSetTimeZone(const std::string& cszTimeZone);

    /// Answer the user responded message for CLogUIPython.
    void OnLogQueryUserResponded(int iPID, int iResp);

    /// Answer the user responded message for CLogUIPython.
    void OnLogQueryInputUserResponded(int iPID, int iResp, const std::string& cszDataJson);

    /// The port on which the listening should occur.
    const int m_ciPort;

    /// The socket (and length) to reply to.  Only valid when replying inside a request handler.
    int m_iSockfd = 0;
    struct sockaddr_in m_sCliaddr = {0};
    socklen_t m_uiLen = 0;

    /// The map the associates message IDs with the callback function.
    CIDCallBackMap m_idCallbackMap;

    /// The processes to notify on standard messages.
    std::vector<CQueueProcess*> m_pvNotify;

    DECLARE_MSG()
};

// ================================================================================================
/// The class that sends a request to the server and waits for a response if one is coming.
class CUDPRequestResponse final
{
public:
    CUDPRequestResponse(int iPort, const char* szIP);
    CUDPRequestResponse(const CUDPRequestResponse&) = delete;
    CUDPRequestResponse(CUDPRequestResponse&&) = default;
    CUDPRequestResponse& operator=(const CUDPRequestResponse&) = delete;
    ~CUDPRequestResponse() = default;

    /// @{ Send the request (with @p iID, @p rsData) and return the response.
    void SendReq(int_32 iID);
    template<typename T, std::enable_if_t<std::is_standard_layout_v<T>>* = nullptr>
    void SendReq(int_32 iID, const T& rsData);
    void SendReq(int_32 iID, const std::string& cszData);
    template<typename T1,
        typename T2,
        std::enable_if_t<std::is_standard_layout_v<T1> && std::is_standard_layout_v<T2>>* = nullptr>
    void SendReq(int_32 iID, const T1& rsData1, const T2& rsData2);
    template<typename T, std::enable_if_t<std::is_standard_layout_v<T>>* = nullptr>
    void SendReq(int_32 iID, const T& rsData, const std::string& cszStr);
    template<typename T1,
             typename T2,
             std::enable_if_t<std::is_standard_layout_v<T1> && std::is_standard_layout_v<T2>>* = nullptr>
    void SendReq(int_32 iID, const T1& rsData1, const T2& rsData2, const std::string& cszStr);
    template<typename T1,
        typename T2,
        typename T3,
        std::enable_if_t<std::is_standard_layout_v<
                             T1> && std::is_standard_layout_v<T2> && std::is_standard_layout_v<T3>>* = nullptr>
    void SendReq(int_32 iID, const T1& rsData1, const T2& rsData2, const T3& rsData3);
    template<typename T, std::enable_if_t<std::is_standard_layout_v<T>>* = nullptr>
    std::pair<bool, T> SendReq(int_32 iID);
    /// @}

    void UpdateUnitSelections(const std::string& cszSelectionJson);
    void SetTranslationLocale(const std::string& cszLocale);
    void SetTimeZone(const std::string& cszTimeZone);
    void LogQueryUserResponded(int iPID, int iResp);
    void LogQueryInputUserResponded(int iPID, int iResp, const std::string& cszDataJson);

private:
    /// Send raw memory at @p pData of length @p uiLen over UDP.
    void Send(void* pData, size_t uiLen);
    /// Read the response.  It will assert that the length is @p uiLen. If all is correct, it will
    /// copy the memory into @p pData.  @return true if we got a response, or false otherwise.
    bool Response(void* pData, size_t uiLen);

    const int m_ciPort;             ///< Which port to send to.
    int m_iSockfd;                  ///< The socket id to use for sending/receiving.
    struct sockaddr_in m_servaddr;  ///< The server address to receive from.

    DECLARE_MSG()
};

// ================================================================================================
template<typename T> CQueueProcess::FnReceiveReq CQueueProcess::MakeReq(T* pC, void (T::*pFcn)())
{
    return [pC, pFcn](CDataArchive& rAr) { (pC->*pFcn)(); };
}

template<typename T, typename TDATA>
CQueueProcess::FnReceiveReq CQueueProcess::MakeReq(T* pC, void (T::*pFcn)(const TDATA&))
{
    return [pC, pFcn](CDataArchive& rAr) {
        TDATA data;
        Serialize(rAr, data);
        (pC->*pFcn)(data);
    };
}

template<typename T, typename TDATA, std::enable_if_t<std::is_scalar_v<TDATA>>* = nullptr>
CQueueProcess::FnReceiveReq CQueueProcess::MakeReq(T* pC, void (T::*pFcn)(TDATA))
{
    return [pC, pFcn](CDataArchive& rAr) {
        TDATA data;
        Serialize(rAr, data);
        (pC->*pFcn)(data);
    };
}

template<typename T, typename TDATA1, typename TDATA2, NProcessBase::enable_if_serialize<TDATA1, TDATA2>* = nullptr>
CQueueProcess::FnReceiveReq CQueueProcess::MakeReq(T* pC, void (T::*pFcn)(TDATA1, TDATA2))
{
    return [pC, pFcn](CDataArchive& rAr) {
        TDATA1 data1;
        TDATA2 data2;
        Serialize(rAr, data1);
        Serialize(rAr, data2);
        (pC->*pFcn)(data1, data2);
    };
}

template<typename T,
    typename TDATA1,
    typename TDATA2,
    typename TDATA3,
    NProcessBase::enable_if_serialize<TDATA1, TDATA2, TDATA3>* = nullptr>
CQueueProcess::FnReceiveReq CQueueProcess::MakeReq(T* pC, void (T::*pFcn)(TDATA1, TDATA2, TDATA3))
{
    return [pC, pFcn](CDataArchive& rAr) {
        TDATA1 data1;
        TDATA2 data2;
        TDATA3 data3;
        Serialize(rAr, data1);
        Serialize(rAr, data2);
        Serialize(rAr, data3);
        (pC->*pFcn)(data1, data2, data3);
    };
}

template<typename T>
CQueueProcess::FnReceiveReq CQueueProcess::MakeReq(std::function<void(const T&)> fcn)
{
    return [fcn](CDataArchive& rAr) {
        T data;
        Serialize(rAr, data);
        fcn(data);
    };
}

template<typename T, std::enable_if<std::is_scalar_v<T>>* = nullptr>
CQueueProcess::FnReceiveReq CQueueProcess::MakeReq(std::function<void(T)> fcn)
{
    return [fcn](CDataArchive& rAr) {
        T data;
        Serialize(rAr, data);
        fcn(data);
    };
}

template<typename T> bool CQueueProcess::SendReq(int iID, const T& cData)
{
    std::ostringstream os;
    auto ar = CDataArchive(os);
    Serialize(ar, const_cast<T&>(cData));

    return SendRequest(iID, os);
}

template<typename T1, typename T2>
bool CQueueProcess::SendReq(int iID, const T1& cData1, const T2& cData2)
{
    std::ostringstream os;
    auto ar = CDataArchive(os);
    Serialize(ar, const_cast<T1&>(cData1));
    Serialize(ar, const_cast<T2&>(cData2));

    return SendRequest(iID, os);
}

template<typename T1, typename T2, typename T3>
bool CQueueProcess::SendReq(int iID, const T1& cData1, const T2& cData2, const T3& cData3)
{
    std::ostringstream os;
    auto ar = CDataArchive(os);
    Serialize(ar, const_cast<T1&>(cData1));
    Serialize(ar, const_cast<T2&>(cData2));
    Serialize(ar, const_cast<T3&>(cData3));

    return SendRequest(iID, os);
}

// ============================================================================
template<typename T>
TSharedMemQueueProcess<T>::TSharedMemQueueProcess(FnProcess fcnProcess) : CQueueProcess(fcnProcess)
{
    static_assert(std::is_standard_layout_v<T>, "Type T must be a POD type.");
    static_assert(!std::is_const_v<T>, "Type T must not be a constant type.");
    m_iDataID = shmget(IPC_PRIVATE, sizeof(T), IPC_CREAT | IPC_EXCL | 0600);
    m_pData = new(shmat(m_iDataID, nullptr, 0)) T();
}

template<typename T> TSharedMemQueueProcess<T>::~TSharedMemQueueProcess()
{
    VERIFY_ERROR(shmctl(m_iDataID, IPC_RMID, nullptr) != -1);
    m_iDataID = 0;
    m_pData = nullptr;
}

// ================================================================================================
template<typename T, typename TRet, std::enable_if_t<std::is_standard_layout_v<TRet>>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, TRet (T::*pFcn)())
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize == 0);
        auto sRet = (pC->*pFcn)();
        SendReply(sizeof sRet, &sRet);
    };
}

template<typename T,
    typename TRet,
    typename TReq,
    std::enable_if_t<std::is_standard_layout_v<TRet> && std::is_scalar_v<TReq>>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, TRet (T::*pFcn)(TReq))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize == sizeof(TReq));
        auto sRet = (pC->*pFcn)(*reinterpret_cast<TReq*>(pData));
        SendReply(sizeof sRet, &sRet);
    };
}

template<typename T>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, void (T::*pFcn)(const std::string&))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        (pC->*pFcn)(std::string(pData, uiSize));
    };
}

template<typename T,
    typename TRet,
    typename TReq,
    std::enable_if_t<std::is_standard_layout_v<TRet> && std::is_standard_layout_v<TReq>>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, TRet (T::*pFcn)(const TReq&))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize == sizeof(TReq));
        auto sRet = (pC->*pFcn)(*reinterpret_cast<TReq*>(pData));
        SendReply(sizeof sRet, &sRet);
    };
}

template<typename T> CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, void (T::*pFcn)())
{
    FnReceiveReq ret = [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize == 0);
        (pC->*pFcn)();
    };
    return ret;
}

template<typename T, typename TReq, std::enable_if_t<std::is_scalar_v<TReq>>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, void (T::*pFcn)(TReq))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize == sizeof(TReq));
        (pC->*pFcn)(*reinterpret_cast<const TReq*>(pData));
    };
}

template<typename T, typename TReq1, typename TReq2, NProcessBase::enable_if_scalar<TReq1, TReq2>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, void (T::*pFcn)(TReq1, TReq2))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize == sizeof(TReq1) + sizeof(TReq2));
        (pC->*pFcn)(*reinterpret_cast<const TReq1*>(pData),
            *reinterpret_cast<const TReq2*>(pData + sizeof(TReq1)));
    };
}

template<typename T, typename TReq, NProcessBase::enable_if_scalar<TReq>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC,
    void (T::*pFcn)(TReq, const std::string& cszStr))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize >= sizeof(TReq));
        (pC->*pFcn)(*reinterpret_cast<const TReq*>(pData),
            std::string(pData + sizeof(TReq), uiSize - sizeof(TReq)));
    };
}

template<typename T, typename TReq1, typename TReq2, NProcessBase::enable_if_scalar<TReq1, TReq2>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC,
    void (T::*pFcn)(TReq1, TReq2, const std::string& cszStr))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize >= sizeof(TReq1) + sizeof(TReq2));
        (pC->*pFcn)(*reinterpret_cast<const TReq1*>(pData),
            *reinterpret_cast<const TReq2*>(pData + sizeof(TReq1)),
            std::string(
                pData + sizeof(TReq1) + sizeof(TReq2), uiSize - (sizeof(TReq1) + sizeof(TReq2))));
    };
}

template<typename T,
    typename TReq1,
    typename TReq2,
    typename TReq3,
    NProcessBase::enable_if_scalar<TReq1, TReq2, TReq3>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, void (T::*pFcn)(TReq1, TReq2, TReq3))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize == sizeof(TReq1) + sizeof(TReq2)  + sizeof(TReq3));
        (pC->*pFcn)(*reinterpret_cast<const TReq1*>(pData),
            *reinterpret_cast<const TReq2*>(pData + sizeof(TReq1)),
            *reinterpret_cast<const TReq3*>(pData + sizeof(TReq1) + sizeof(TReq2)));
    };
}

template<typename T, typename TReq, std::enable_if_t<std::is_standard_layout_v<TReq>>* = nullptr>
CUDPProcess::FnReceiveReq CUDPProcess::MakeReq(T* pC, void (T::*pFcn)(const TReq&))
{
    return [this, pC, pFcn](size_t uiSize, const char* pData) {
        ASSERT_ERROR(uiSize == sizeof(TReq));
        (pC->*pFcn)(*reinterpret_cast<const TReq*>(pData));
    };
}

// ============================================================================
template<typename T, std::enable_if_t<std::is_standard_layout_v<T>>* = nullptr>
void CUDPRequestResponse::SendReq(int_32 iID, const T& rsData)
{
    constexpr uint_32 cuiLen = sizeof(rsData);
    auto buffer = std::array<char, sizeof(iID) + sizeof(cuiLen) + cuiLen>();
    memcpy(&buffer[0], &iID, sizeof(iID));
    memcpy(&buffer[sizeof(iID)], &cuiLen, sizeof(cuiLen));
    memcpy(&buffer[sizeof(iID) + sizeof(cuiLen)], &rsData, sizeof(rsData));
    Send(&buffer[0], buffer.size());
}

template<typename T1,
    typename T2,
    std::enable_if_t<std::is_standard_layout_v<T1> && std::is_standard_layout_v<T2>>* = nullptr>
void CUDPRequestResponse::SendReq(int_32 iID, const T1& rsData1, const T2& rsData2)
{
    constexpr uint_32 cuiLen = sizeof(rsData1) + sizeof(rsData2);
    auto buffer = std::array<char, sizeof(iID) + sizeof(cuiLen) + cuiLen>();
    memcpy(&buffer[0], &iID, sizeof(iID));
    memcpy(&buffer[sizeof(iID)], &cuiLen, sizeof(cuiLen));
    memcpy(&buffer[sizeof(iID) + sizeof(cuiLen)], &rsData1, sizeof(rsData1));
    memcpy(&buffer[sizeof(iID) + sizeof(cuiLen) + sizeof(rsData1)], &rsData2, sizeof(rsData2));
    Send(&buffer[0], buffer.size());
}

template<typename T, std::enable_if_t<std::is_standard_layout_v<T>>* = nullptr>
void CUDPRequestResponse::SendReq(int_32 iID,
    const T& rsData,
    const std::string& cszStr)
{
    uint_32 cuiLen = sizeof(rsData) + cszStr.length();
    auto buffer = std::string();
    buffer.resize(sizeof(iID) + sizeof(cuiLen) + cuiLen);
    auto p = buffer.data();
    auto FillBuffer = [](char* p, const auto& cValue) {
        memcpy(p, &cValue, sizeof(cValue));
        return p + sizeof(cValue);
    };
    p = FillBuffer(p, iID);
    p = FillBuffer(p, cuiLen);
    p = FillBuffer(p, rsData);
    memcpy(p, cszStr.c_str(), cszStr.length());
    Send(buffer.data(), buffer.length());
}

template<typename T1,
         typename T2,
         std::enable_if_t<std::is_standard_layout_v<T1> && std::is_standard_layout_v<T2>>* = nullptr>
void CUDPRequestResponse::SendReq(int_32 iID,
    const T1& rsData1,
    const T2& rsData2,
    const std::string& cszStr)
{
    uint_32 cuiLen = sizeof(rsData1) + sizeof(rsData2) + cszStr.length();
    auto buffer = std::string();
    buffer.resize(sizeof(iID) + sizeof(cuiLen) + cuiLen);
    auto p = buffer.data();
    auto FillBuffer = [](char* p, const auto& cValue) {
        memcpy(p, &cValue, sizeof(cValue));
        return p + sizeof(cValue);
    };
    p = FillBuffer(p, iID);
    p = FillBuffer(p, cuiLen);
    p = FillBuffer(p, rsData1);
    p = FillBuffer(p, rsData2);
    memcpy(p, cszStr.c_str(), cszStr.length());
    Send(buffer.data(), buffer.length());
}

template<typename T1,
    typename T2,
    typename T3,
    std::enable_if_t<std::is_standard_layout_v<
                         T1> && std::is_standard_layout_v<T2> && std::is_standard_layout_v<T3>>* = nullptr>
void CUDPRequestResponse::SendReq(int_32 iID, const T1& rsData1, const T2& rsData2, const T3& rsData3)
{
    constexpr uint_32 cuiLen = sizeof(rsData1) + sizeof(rsData2) + sizeof(rsData3);
    auto buffer = std::array<char, sizeof(iID) + sizeof(cuiLen) + cuiLen>();
    auto p = buffer.data();
    auto FillBuffer = [](char* p, const auto& cValue) {
        memcpy(p, &cValue, sizeof(cValue));
        return p + sizeof(cValue);
    };
    p = FillBuffer(p, iID);
    p = FillBuffer(p, cuiLen);
    p = FillBuffer(p, rsData1);
    p = FillBuffer(p, rsData2);
    p = FillBuffer(p, rsData3);
    Send(&buffer[0], buffer.size());
}

template<typename T, std::enable_if_t<std::is_standard_layout_v<T>>* = nullptr>
std::pair<bool, T> CUDPRequestResponse::SendReq(int_32 iID)
{
    constexpr uint_32 cuiLen = 0;
    auto buffer = std::array<char, sizeof(iID) + sizeof(cuiLen)>();
    memcpy(&buffer[0], &iID, sizeof(iID));
    memcpy(&buffer[sizeof(iID)], &cuiLen, sizeof(cuiLen));
    Send(&buffer[0], buffer.size());

    T sRet;
    auto fGotResponse = Response(&sRet, sizeof(T));
    return std::make_pair(fGotResponse, sRet);
}

#endif  // PROCESS_BASE_HPP
