#ifndef LOG_QUERY_HPP
#define LOG_QUERY_HPP
//-------------------------------------------------------------------------------------------------
/// @file
/// Application-wide logging and simple queries.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "cstr.hpp"
#include "formattr.hpp"
#include "../rapidjson/fwd.h"
#include <memory>
#include <vector>

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
class CLogUIInterface;
class CLogUIInput;

/// Application-wide logging and simple queries.
///
/// The application (or application framework) must initialize and terminate the logging on app
/// startup and termination. See: SetCfrParameters(), InitializeLog() and CloseLog().  Messages can
/// be posted before InitializeLog(), but they will not be saved or shown till that is called.
///
/// This file provides several styles of notifying the user and support person:
/// - Log the message to the log file.  (LogToFileOnly)
/// - Show the user the message, but do not wait for acknowledgment. (PostXXX)
/// If the framework is a full GUI (MIC_VCMFC) the following are also provided.
/// - Show the message in a modal way so processing does not proceed until the user acknowledgment.
///   (InformWait)
/// - Query the user for yes/no/ok/cancel. (YesNo, YesNoCancel, OKCancel)
///
/// PostOnce messages are only posted once until the user acknowledges them.
/// Post messages can be updated with a PostUpdate() call if they are created with a iPostID.
namespace NLogQuery
{
#if defined(MIC_VCMFC)
    using CModalContext = CWnd*;
#else
    using CModalContext = int;
#endif

    class CQueryResponse;

    /// Determines the initialization error
    enum EInitState
    {
        ceNoInit,          ///< The Initialize method has not been called.
        ceUnableToOpenDb,  ///< A file I/O error occurred.
        ceIniNotSelected,  ///< No valid INI file was selected by the user.
        ceSuccess,         ///< No error.
    };

    /// Severity of message level of message
    enum ESeverity
    {
        ceImported,     ///< Message was imported from old plaintext .LOG file.
        ceVerbose,      ///< Message is considered noise and does not need to get
                        ///< displayed in audit trail report.
        ceDefaultError, ///< These messages get reported in the audit trail report. Error icon in display.
        ceDefaultInfo,  ///< These messages get reported in the audit trail report. Info icon in display.
    };

    /// User message-or-question dialog type.
    enum EUserAction
    {
        ceUActNone,  ///< Do not display user message.

        ceUActPostOnce,  ///< Display modeless user message (OK button); do not display if
                         ///< identical message already shown.
        ceUActPost,      ///< Display modeless user message (OK button).

        ceUActInformWait,   ///< Display modal user message (OK button).
        ceUActYesno,        ///< Display modal user question (Yes/No buttons).
        ceUActYesnocancel,  ///< Display modal user question (Yes/No/Cancel buttons).
        ceUActOkcancel,     ///< Display modal user question (OK/Cancel buttons).

        ceUActPostUpdate,  ///< Update the post with the same ID to the new message.

        ceUActInputForm,  ///< Requesting input from the user.
        ceUActSampleLog,  ///< Log the message into the sample log (if any).
    };

    /// Response to user query dialog.
    enum EUserResponse
    {
        ceRespOK,  ///< User pressed the OK or Yes button.
        ceRespYes = ceRespOK,
        ceRespNo,      ///< User pressed the No button.
        ceRespCancel,  ///< User pressed the Cancel button.
        ceRespNone,    ///< Signals an error; all dialogs must be closed by the user.
        ceRespNum
    };

    /// File system error codes that PostFileError can display.
    enum EFSError
    {
        // These start at a high value to make sure that they are different from ones in errno.h.
        ceFSErrorBase = 1000,
        ceFSErrorCouldNotRegister,
        ceFSErrorSubsetNotFound,
        ceFSErrorSeekFailed,
        ceFSErrorBadHeader,
        ceFSErrorOwnerDeny,
        ceFSErrorNotOurFile,
        ceFSErrorBadWriteLen,
        ceFSErrorReadError,
        ceFSErrorWriteError,
        ceFSErrorSeqsegCopy,
    };

    /// The data stored for a single log entry.
    struct SLogEntry final
    {
        SLogEntry(int64_t iTimeStamp,
            CStr szUserName,
            CStr szAccountType,
            EUserAction eUserAction,
            ESeverity eSeverity,
            CStr szMessage,
            int iPostID = 0);

        SLogEntry(const CStr& cszMessage,
            EUserAction eUser,
            ESeverity eSeverity = ceDefaultError);
        SLogEntry(const CStr& cszMessage,
            EUserAction eUser,
            int iPostID,
            ESeverity eSeverity = ceDefaultError);

        int64_t iTimeStamp;
        CStr szUserName;
        CStr szAccountType;
        EUserAction eUserAction;
        ESeverity eSeverity;
        CStr szMessage;
        int iPostID;
    };
    /// A list of log entries.
    using VLogEntry = std::vector<SLogEntry>;

    /// Set the cfr parameters that will be used in the log information.
    void SetCfrParameters(const CStr& szLoggedInUser, const CStr& szAccountType);

    /// Initialize the message log (log file, user interface). Should be called by
    /// mainframe during initialization.
    /// @param pLogUI Instance of the helper class that manages the global
    ///        application message log window and the display of message log
    ///        messages. Ownership of this object is transferred to the
    ///        application message log.
    EInitState InitializeLog(std::unique_ptr<CLogUIInterface>&& rpLogUI);
    /// @return True if initialized.
    bool IsInitialized();
    /// @returns the UI interface that this system is using, or nullptr if not initialized.
    CLogUIInterface* GetUIInterface();
    /// Close the application message log and any open windows. Should be called
    /// by the main frame on program exit.
    void CloseLog();

    /// @return the log data between time @p iFrom and @p iTo from the logfile identified by @p
    /// szLogFile
    VLogEntry GetAuditTrailData(const CStr& szLogFile, int64_t iFrom, int64_t iTo);
    /// @return all the the log data from the logfile identified by @p szLogFile
    VLogEntry GetAllData(const CStr& szLogFile);

    /// @{
    /// Log the message message in the application message log file. Do not notify the operator.
    /// @param eSeverity Used to mark the message level of the message for better filtering
    ///     capabilities.
    void LogToFileOnly(const CStr& cszMessage, ESeverity eSeverity = ceDefaultError);
    void LogToFileOnly(uint_32 uiRes, ESeverity eSeverity = ceDefaultError);
    /// @}

    /// Log the message to the sample log.
    void SampleLog(const CStr& cszMessage, ESeverity eSeverity = ceDefaultError);

    /// @{
    /// Log the message and show it a modeless way.
    ///
    /// The PostOnce version will only show the same message string once no matter how many PostOnce
    /// calls are made.
    /// @param eSeverity Used to mark the message level of the message for better
    ///     filtering capabilities.
    /// @param iPostID is used to replace the post once string even if the cszMessage has
    ///        changed.  These are not persistent, so they just have to be unique at runtime.  Also
    ///        see GetNextPostID()
    void PostOnce(const CStr& cszMessage, ESeverity eSeverity = ceDefaultError);
    void PostOnce(uint_32 uiRes, ESeverity eSeverity = ceDefaultError);
    void PostOnce(const CStr& cszMessage, int iPostID, ESeverity eSeverity = ceDefaultError);
    void Post(const CStr& cszMessage, ESeverity eSeverity = ceDefaultError);
    void Post(uint_32 uiRes, ESeverity eSeverity = ceDefaultError);
    void Post(const CStr& cszMessage, int iPostID, ESeverity eSeverity = ceDefaultError);
    void PostUpdate(const CStr& cszMessage, int iPostID);
    /// @}

    /// Call this and assign to a local static variable to allow you to reuse the ID for subsequent
    /// calls to PostOnce().
    /// @returns the next unique PostOnceID.  Always returns non zero.
    int GetNextPostID();

    /// This is a special error message for file errors.  The message has the form:
    ///     "2010-  Error accessing file <name>, error code = <number> (<description>)"
    /// @param iErrorCode must be EITHER a code from the "errno" variable or a EFSError value.
    /// @param eSeverity Used to mark the error level of the message for better
    ///     filtering capabilities.
    void PostFileError(int iErrCode, const CStr& cszFileName, ESeverity eSeverity = ceDefaultError);

    /// @{
    /// Query the user with the given question and wait for a Yes/No/OK/Cancel response. Messages
    /// are trimmed of leading/trailing whitespace (including newlines). resulting strings should
    /// not be empty.
    ///
    /// @param pWnd For VCMFC targets, this is parent window to use for the query dialog.  Can be
    ///     nullptr in which case the root/main application window will be used.  For other
    ///     targets, this is an integer that provides context for that target.
    /// @param eSeverity Used to mark the message level of the message for better filtering
    ///     capabilities.
    /// @returns CQueryResponse instance. This can be used to check the user's response.
    CQueryResponse YesNo(const CStr& cszMessage, CModalContext pWnd = 0, ESeverity eSeverity = ceDefaultError);
    CQueryResponse YesNo(uint_32 uiRes, CModalContext pWnd = 0, ESeverity eSeverity = ceDefaultError);
    CQueryResponse YesNoCancel(const CStr& cszMessage,
        CModalContext pWnd = 0,
        ESeverity eSeverity = ceDefaultError);
    CQueryResponse YesNoCancel(uint_32 uiRes, CModalContext pWnd = 0, ESeverity eSeverity = ceDefaultError);
    CQueryResponse OKCancel(const CStr& cszMessage,
        CModalContext pWnd = 0,
        ESeverity eSeverity = ceDefaultError);
    CQueryResponse OKCancel(uint_32 uiRes, CModalContext pWnd = 0, ESeverity eSeverity = ceDefaultError);
    /// @}

    /// Prompt the user and have him enter a numeric value.
    /// @ returns the ok/cancel response and the value entered.
    CQueryResponse GetUserData(CLogUIInput& rInput, int iInputContext);

    /// @{
    /// Log the  message and show the message to the user, and wait for the user to acknowledged
    /// before returning.
    /// @param pWnd The parent window to use for the query dialog.  Can be nullptr in which case the
    ///     root/main application window should be used.
    /// @param eSeverity Used to mark the message level of the message for better filtering
    ///     capabilities.
    void InformWait(const CStr& cszMessage, CModalContext pWnd = 0, ESeverity eSeverity = ceDefaultError);
    void InformWait(uint_32 uiRes, CModalContext pWnd = 0, ESeverity eSeverity = ceDefaultError);
    /// @}

    // ============================================================================================
    /// Objects encapsulate the response to a user question.
    class CQueryResponse
    {
    public:
        CQueryResponse(EUserResponse eResp = ceRespNone);
        CQueryResponse(const CQueryResponse& ) = default;
        CQueryResponse& operator=(const CQueryResponse& ) = default;

        /// @returns True if a request for operator action was acknowledge with the OK button.
        ///          Synonymous with IsYes().
        bool IsOK() const;
        /// @see IsOk()
        bool IsYes() const;
        /// @returns True if a request for operator action was acknowledge with the No button.
        bool IsNo() const;
        /// @returns True if a request for operator action was acknowledge with the Cancel button.
        bool IsCancel() const;

    private:
        /// The response.
        EUserResponse m_eUserResponse;
    };

#if defined(MIC_VCMFC)
    /// Predefined network errors that this call can interpret and display.
    enum ENetworkEvent
    {
        // socket function calls
        ceAccept,
        ceBind,
        ceCloseSocket,
        ceConnect,
        ceIoctlSocket,
        ceListen,
        ceRecv,
        ceSend,
        ceSendTo,
        ceGetSockOpt,
        ceSetSockOpt,
        ceSocket,

        // windows-specific extension function calls
        ceWSAStartup,
        ceWSAAsyncSelect,
        ceWSACleanup,

        ceNum
    };

    /// Debug mode only, assert, and then post the last windows error (GetLastError).
    void PostLastWindowsError();

    /// Log the error identified by @p iErrCode and @p eNetEvent to the error file.
    void LogToFileWinsockErr(int iErrCode, ENetworkEvent eNetEvent);

    /// @return the CStr associated with iErrCode
    CStr GetWinsockErrStr(int iErrCode);
    /// @return the CStr associated with iErrCode
    CStr GetWindowsError(int iErrCode);

#if defined(_DEBUG)
    /// Used to generate a callstack in string form. Useful when looking at logs with
    /// trace messages enabled.
    /// Ex: TRACE_MSG1("%s", NLogQuery::GetCurrentCallStack().GetStringPtr());
    CStr GetCurrentCallStack();
#endif

#endif  // defined(MIC_VCMFC)
}  // namespace NLogQuery

// ================================================================================================
/// Interface for the message log user interface implementation.
class CLogUIInterface
{
public:
    using SLogEntry = NLogQuery::SLogEntry;
    using VLogEntry = NLogQuery::VLogEntry;
    using ESeverity = NLogQuery::ESeverity;
    using EUserResponse = NLogQuery::EUserResponse;
    using CModalContext = NLogQuery::CModalContext;
    using CQueryResponse = NLogQuery::CQueryResponse;

    virtual ~CLogUIInterface() = default;

    /// Create the database table. Opens and closes the database.
    virtual bool CreateDB() = 0;
    /// Open the database. Asserts that the database is closed.
    virtual bool OpenDB() = 0;

    /// Gets the rows of the database as a vector of tuples.
    virtual VLogEntry GetData(const CStr& szLogFile,
        int64_t iFrom,
        int64_t iTo,
        std::vector<ESeverity>& veSeverities)
        = 0;

    /// Checks the last entry in the database to see if the "Log off" message was added
    /// or not.
    virtual bool DidAppCrash() = 0;

    /// Add the provided message message to the global message log window.
    virtual void AddLogEntry(const SLogEntry& cLogEntry) = 0;

    /// Display the provided query and wait for user to answer/confirm.
    /// @param cLogEntry Message string to display, and what buttons to show.
    /// @param pWnd Parent window. If nullptr, the application window is used.
    /// @return The user response corresponding to the button pressed.
    virtual EUserResponse ShowModal(const SLogEntry& cLogEntry, CModalContext pWnd) = 0;

    /// Prompt the user and have him enter a numeric value.
    /// @ returns the ok/cancel response and the value entered.
    virtual EUserResponse GetUserData(CLogUIInput& rInput, int iInputContext) = 0;
};

// ================================================================================================
/// The class that defines the InputForm that is used to allow the user to enter data during an
/// automatic operation.
///
/// To use, construct it with the prompt string that will be the title of the input dialog.  Then
/// call Add(szName, xdFormatter) or Add(szName, szStr, iMaxLength) to add an input field with the
/// name and given format.  You can add as many as you like.
///
/// To show the dialog, and let the user enter data, call the NLogQuery::GetUserData(fmt, context).
/// This will return the usual CQueryResponse object that you can interrogate for IsCancel().
///
/// If the user did not cancel, you can get the values the user entered with Calls to
/// GetDouble(szName) or GetString(szName).
class CLogUIInput final
{
public:
    CLogUIInput(const CStr& cszTitle = CStr());
    CLogUIInput(CLogUIInput&&) = default;
    ~CLogUIInput() = default;

    /// Access to the input form title.
    const CStr& GetTitle() const;

    /// Convert the object into the json string that can be used for react input form parameters.
    rapidjson::Value InputFormJson(rapidjson::Document::AllocatorType& rAlloc) const;

    /// @return the user entered value for @p cszName.
    double GetDouble(const CStr& cszName) const;

    /// @return the user entered value for @p cszName.
    CStr GetString(const CStr& cszName) const;

    /// Add a numeric parameter to show.
    void Add(const CStr& cszName, TFormattedNumeric<double> xdValue);

    /// Add a string to show.
    void Add(const CStr& cszName, const CStr& cszValue, int iMaxLength);

    /// Update the values stored in the object to those given by the json data.
    void UpdateValuesFromJson(const std::string& cszDataJson);

private:

    /// The data to save for each string input.
    struct  SStringData
    {
        CStr szName;    ///< The name of the parameter.
        CStr szValue;   ///< The value of the parameter.
        int iMaxLength; ///< Max allowed length of the string to input.
    };

    /// The title for the dialog.
    const CStr m_cszTitle;

    /// The order that the items were added for layout.
    std::vector<CStr> m_vszLayoutOrder;

    /// The vector of numeric values to read from the user. first is the name/prompt for the value,
    /// second is the default value along with the input formatting and range.
    std::vector<std::pair<CStr, TFormattedNumeric<double>>> m_vsDoublePrompts;

    /// The vector of string values to read from the user.
    std::vector<SStringData> m_vsStringPrompts;

    DECLARE_MSG()
};

#endif  // LOG_QUERY_HPP
