#ifndef METADATA_HPP
#define METADATA_HPP
//------------------------------------------------------------------------------
/// @file
/// Type info metadata support classes, functions, and templates.
//------------------------------------------------------------------------------
// Includes
//------------------------------------------------------------------------------
#include <type_traits>
#include <typeinfo>
#include <typeindex>
#include <list>

#ifdef max
#undef max
#endif
#ifdef min
#undef min
#endif

#include "formattr.hpp"
#include "basetype.hpp"
#include "cstr_fmt.hpp"
#include "metadata-traits.hpp"
#include "nummat.hpp"
#include <boost/optional.hpp>

#include <iomanip>
#include <sstream>
// Forward declarations
namespace NMeta
{
    struct SMeta;
    struct STypeValue;
    struct SMemberDesc;
    struct SMemberLabel;
}  // namespace NMeta

namespace NDataContext
{
    template<typename T> void InitializeMetadata(const NMeta::SMemberLabel& csLabel);
}

/// The metadata system
/// =====================================================================================
/// The metadata system is comprised of two parts.
///
/// The first part is the data container type annotation. This takes a description of each
/// serialized data member (i.e., data members read/written to disk) and base data
/// container class (i.e., base classes that have Read/WriteData called explicitly by the
/// annotated derived class). Annotation generates metadata functions that store type
/// information for a given data member or base class related to the annotated class.
///
/// The second part are the functions that take an instance of an annotated data
/// container and apply the annotation
///
/// All data containers must be annotated if they are serialized to disk either at the
/// derived level or as a base of a more derived class.
///
/// Typical structure of the serialized components of a data container
/// -------------------------------------------------------------------------------------
///
///     Root (most-derived CBaseDatContainer-derived class)
///         Base 1 (optional; CBaseDatContainer-derived base class)
///             Base N... (optional; CBaseDatContainer-derived base class)
///         member N... (member data of Root)
///
/// Structure of data container annotation
/// -------------------------------------------------------------------------------------
/// Given:
///
///     class Derived : public Base
///     class Base : public CDataContainer
///
/// Annotation of Derived (the root type) involves defining the static TypeInfo member
/// of the CTypeInfo<Derived> class. This definition includes the annotation of Derived
/// and is required to register Derived with the metadata system. As an example,
///
///     CTypeInfo<Derived>
///     CTypeInfo<Derived>::TypeInfo( 30000,  // Label for Derived
///     {
///         NoteBase<Base>( 30001 ),          // Base class annotation with optional description
///         Note( &Derived::dDouble, 30002 ), // Double data member.
///         Note( &Derived::eEnumA,  30003 ), // Enum data member. Requires support function.
///         Note( &Derived::eEnumB,  30004,   // Enum data member with a custom description
///               CustomEnumDescription ),    // function.
///         ...
///     } );
///
/// A custom data description function may be provided. It must take an instance of the
/// data member type. For smart pointers, it must take an instance of the wrapped type.
/// Enums require a data description function. All annotated enums must have a default
/// description function. See NMeta::EnumDescription.
///
/// @note Assumes root (top-level) type is derived from NMeta::IsDataContainer. All other
/// annotations are either NMeta::IsDataContainer base classes of the root or member
/// data.
/// @note CBaseDataContainer base classes should be annotated only if the base class
/// members are serialized, e.g., Derived::Read/WriteData call Base::Read/WriteData.
/// @note Reference members cannot be annotated. Pointer to reference member is disallowed
/// by the C++ standard.
/// @note Pointers and values only; no handles or std::weak_ptrs allowed.
/// @note C-style strings not unhandled; no char* allowed.
/// @note C-style array members cannot be annotated. Change to std::array.
/// @note Convert boost smart pointers and arrays (shared, scoped) to std equivalents.
/// @note Enums must be public. Enum values must fit in signed 64-bit int.
/// @note For SFINAE, unsupported classes/functions should use !is_same_v<T, T> to invoke
/// a compile time static_assert( false, error_string ).
namespace NMeta
{
/// Debug mechanism. Use WhichType<T> to get a compiler error that gives the type T.
template<typename...> struct WhichType;

#if defined(_DEBUG)
void TestMsg(const CStr& cszMsg);   ///< For debug log tracing.
void TraceMsg(const CStr& cszMsg);  ///< For debug log tracing.
#endif

/// A template-argument-dependent false value; a valid workaround used with static_assert. It is
/// required because static_assert(false) in a template is ill-formed.
template<typename>
inline constexpr bool dependent_false_v = false;

//=======================================================================================
/// @defgroup type_trait_templates Type trait templates
/// Type trait template structs use SFINAE to return true/false constants
/// given a type matching the template parameter signature.
/// @{
//---------------------------------------------------------------------------------------
/// @defgroup type_trait_templates_array_size Array size
/// Determine array size. Zero for unhandled types.
/// @{
//---------------------------------------------------------------------------------------
template<class T> struct array_size : std::integral_constant<size_t, 0>
{
};
/// N for arrays T[N].
template<class T, size_t N> struct array_size<T[N]> : std::integral_constant<size_t, N>
{
};
/// N for std::array<T, N>.
template<class T, size_t N>
struct array_size<std::array<T, N>> : std::integral_constant<size_t, N>
{
};
/// Alias for array_size<>::value
template<class T> constexpr size_t array_size_v = array_size<std::remove_cv_t<T>>::value;
/// @} type_trait_templates_array_size

//---------------------------------------------------------------------------------------
/// @defgroup type_trait_templates_IsDataContainer Is-NMeta::IsDataContainer
/// Determine whether T is an NMeta::IsDataContainer-derived type.
/// @{
//---------------------------------------------------------------------------------------
template<class T>
struct is_data_container
    : std::integral_constant<bool, std::is_base_of_v<NMeta::IsDataContainer, T>>
{
};
/// Alias for is_data_container<>::value.
template<class T> constexpr bool is_data_container_v = is_data_container<T>::value;
/// @} type_trait_templates_IsDataContainer

//---------------------------------------------------------------------------------------
/// @defgroup type_trait_templates_CFixedEnum Is-CFixedEnum
/// Determine whether T is a CFixedEnum.*-derived type.
/// @{
//---------------------------------------------------------------------------------------
template<class T> struct is_fixed_enum : std::false_type
{
};

template<template<typename> class Tpl, typename T>
struct is_fixed_enum<Tpl<T>>
    : std::integral_constant<bool,
          std::is_base_of_v<CFixedEnum8<T>, Tpl<T>> || std::is_base_of_v<CFixedEnum16<T>, Tpl<T>>
              || std::is_base_of_v<CFixedEnum32<T>, Tpl<T>>>
{
};
/// Alias for is_fixed_enum<>::value.
template<class T> constexpr bool is_fixed_enum_v = is_fixed_enum<T>::value;
/// Alias for is_enum_v or is_fixed_enum_v. C-style enum or CFixedEnum type.
template<class T> constexpr bool is_enum_type_v = std::is_enum_v<T> || is_fixed_enum_v<T>;
/// @} type_trait_templates_CFixedEnum

//---------------------------------------------------------------------------------------
/// @defgroup type_trait_templates_CStr is-CStr
/// Determine whether T is a CStr.
/// @{
//---------------------------------------------------------------------------------------
template<class T> struct is_str : std::integral_constant<bool, std::is_base_of_v<CStr, T>>
{
};
/// Alias for is_str<>::value.
template<class T> constexpr bool is_str_v = is_str<T>::value;
/// @} type_trait_templates_CStr

//---------------------------------------------------------------------------------------
/// @defgroup type_trait_templates_container is-std::array/vector, is-iterable, is-CNumMatrix.
/// Determine whether T is a container type: std::array, a std::vector, and TDynamicTableBase.
/// @{
//---------------------------------------------------------------------------------------
template<typename T> struct is_std_array : std::false_type
{
};

template<typename T, size_t N> struct is_std_array<std::array<T, N>> : std::true_type
{
};
/// Alias for is_std_array<>::value.
template<typename T> constexpr bool is_std_array_v = is_std_array<std::remove_cv_t<T>>::value;

//---------------------------------------------------------------------------------------
template<typename T> struct is_std_vector : std::false_type
{
};

template<typename T, typename U> struct is_std_vector<std::vector<T, U>> : std::true_type
{
};
/// Alias for is_std_vector<>::value.
template<typename T> constexpr bool is_std_vector_v = is_std_vector<std::remove_cv_t<T>>::value;

//---------------------------------------------------------------------------------------
template<typename T> struct is_std_list : std::false_type
{
};

template<typename T, typename U> struct is_std_list<std::list<T, U>> : std::true_type
{
};
/// Alias for is_std_list<>::value.
template<typename T> constexpr bool is_std_list_v = is_std_list<std::remove_cv_t<T>>::value;

//---------------------------------------------------------------------------------------
template<typename T>
struct is_meta_iterable : std::integral_constant<bool, std::is_base_of_v<NMeta::IsIterable, T>>
{
};
/// Alias for is_meta_iterable<>::value.
template<typename T>
constexpr bool is_meta_iterable_v = is_meta_iterable<std::remove_cv_t<T>>::value;

//---------------------------------------------------------------------------------------
template<typename T> struct is_num_matrix : std::false_type
{
};

template<typename T> struct is_num_matrix<CNumMatrix<T>> : std::true_type
{
};
/// Alias for is_num_matrix<>::value.
template<typename T> constexpr bool is_num_matrix_v = is_num_matrix<std::remove_cv_t<T>>::value;
/// @} type_trait_templates_container

//---------------------------------------------------------------------------------------
/// @defgroup type_trait_templates_smart_pointer is-std-pointer
/// Determine whether T is a std::shared_ptr or std::unique_ptr.
/// @{
//---------------------------------------------------------------------------------------
template<class T> struct is_smart_ptr : std::false_type
{
};

template<class T> struct is_smart_ptr<std::shared_ptr<T>> : std::true_type
{
};
template<class T, class U> struct is_smart_ptr<std::unique_ptr<T, U>> : std::true_type
{
};
/// Alias for is_smart_ptr<>::value.
template<class T> constexpr bool is_smart_ptr_v = is_smart_ptr<std::remove_cv_t<T>>::value;
/// @} type_trait_templates_smart_pointer

//---------------------------------------------------------------------------------------
/// @defgroup type_trait_templates_TFormattedNumeric Is-TFormattedNumeric<T>
/// Determine whether T is a TFormattedNumeric<>.*-derived type.
/// @{
//---------------------------------------------------------------------------------------
template<class T>
struct is_formatted_numeric : std::integral_constant<bool,
                                  std::is_base_of_v<TFormattedNumeric<float_32>, T>
                                      || std::is_base_of_v<TFormattedNumeric<float_64>, T>
                                      || std::is_base_of_v<TFormattedNumeric<char>, T>
                                      || std::is_base_of_v<TFormattedNumeric<int_8>, T>
                                      || std::is_base_of_v<TFormattedNumeric<int_16>, T>
                                      || std::is_base_of_v<TFormattedNumeric<int_32>, T>
                                      || std::is_base_of_v<TFormattedNumeric<uint_8>, T>
                                      || std::is_base_of_v<TFormattedNumeric<uint_16>, T>
                                      || std::is_base_of_v<TFormattedNumeric<uint_32>, T>>
{
};
template<template<typename> class Tpl, typename T>
struct is_formatted_numeric<Tpl<T>>
    : std::integral_constant<bool,
          std::is_base_of_v<TFormattedNumeric<float_32>, Tpl<T>>
                  || std::is_base_of_v<TFormattedNumeric<float_64>, Tpl<T>>
                  || std::is_base_of_v<TFormattedNumeric<char>, Tpl<T>>
                  || std::is_base_of_v<TFormattedNumeric<int_8>, Tpl<T>>
                  || std::is_base_of_v<TFormattedNumeric<int_16>, Tpl<T>>
                  || std::is_base_of_v<TFormattedNumeric<int_32>, Tpl<T>>
                  || std::is_base_of_v<TFormattedNumeric<uint_8>, Tpl<T>>
                  || std::is_base_of_v<TFormattedNumeric<uint_16>, Tpl<T>>
                  || std::is_base_of_v<TFormattedNumeric<uint_32>, Tpl<T>>>
    {
    };
    /// Alias for is_formatted_numeric<>::value.
    template<class T> constexpr bool is_formatted_numeric_v = is_formatted_numeric<T>::value;
    /// @} type_trait_templates_CFixedEnum

    //---------------------------------------------------------------------------------------
    /// @defgroup type_trait_templates_type Underlying type.
    /// Get the underlying type of class templates type wrappers via the get_arg_type_t
    /// template. This only unwraps one layer, e.g., get_arg_type_t<T> where T is
    /// std::shared_ptr<CFixedEnum8<EEnum>> will be the type CFixedEnum8 not EEnum.
    ///
    /// @note get_arg_type_t uses the return type of helper compile time functions; the helper
    /// functions are constructs for compile time type determination via return type and
    /// should not be invoked directly.
    /// @note Type identification is handled explicitly. New handled types must add a
    /// unique overload and disable handling of the added type to the general case
    /// @{
    //---------------------------------------------------------------------------------------
    /// Default handler. Returns pointer to provided type T.
    template<typename T,
        std::enable_if_t<!is_fixed_enum_v<T> && !is_formatted_numeric_v<T> && !is_meta_iterable_v<T>
            && !is_num_matrix_v<T> && !is_smart_ptr_v<T>>* = nullptr>
    constexpr auto get_arg_type(T*)
    {
        return (std::remove_cv_t<T>*)nullptr;
    }
    /// CFixedEnum<T> handler. Returns pointer to T.
    template<template<typename> class FixedEnum,
        typename T,
        std::enable_if_t<is_fixed_enum_v<FixedEnum<T>>>* = nullptr>
    constexpr auto get_arg_type(FixedEnum<T>*)
    {
        return (std::remove_cv_t<T>*)nullptr;
    }
    /// Derived class of TFormattedNumeric<T> handler. Returns pointer to T.
    template<class T, std::enable_if_t<is_formatted_numeric_v<T>>* = nullptr>
    constexpr auto get_arg_type(T*)
    {
        return (std::remove_cv_t<T>*)nullptr;
    }
    /// TFormattedNumeric<T> handler. Returns pointer to T.
    template<template<typename> class FormattedNumeric,
        typename T,
        std::enable_if_t<is_formatted_numeric_v<FormattedNumeric<T>>>* = nullptr>
    constexpr auto get_arg_type(FormattedNumeric<T>*)
    {
        return (std::remove_cv_t<T>*)nullptr;
    }
    /// std::array<T, N> handler. Returns pointer to T.
    template<typename T, size_t N> constexpr auto get_arg_type(std::array<T, N>*)
    {
        // Recurse until a non-pointer, non-array type is found.
        return get_arg_type((std::remove_pointer_t<T>*)nullptr);
    }
    /// std::vector<T, U> handler. Returns pointer to T.
    template<typename T, class U> constexpr auto get_arg_type(std::vector<T, U>*)
    {
        // Recurse until a non-pointer, non-array type is found.
        return get_arg_type((std::remove_pointer_t<T>*)nullptr);
    }
    /// NMeta::IsIterable derived class handler. Returns pointer to value_type.
    template<typename T, std::enable_if_t<is_meta_iterable_v<T>>* = nullptr>
    constexpr auto get_arg_type(T*)
    {
        return get_arg_type((typename T::value_type*)nullptr);
    }
    /// CNumMatrix<T> handler. Returns pointer to value_type.
    template<typename T, std::enable_if_t<is_num_matrix_v<T>>* = nullptr>
    constexpr auto get_arg_type(T*)
    {
        return get_arg_type((typename T::value_type*)nullptr);
    }
    /// std::shared_ptr<T> handler, Returns pointer to T.
    template<template<typename> class SharedPtr,
        typename T,
        std::enable_if_t<std::is_same_v<std::shared_ptr<T>, SharedPtr<T>>>* = nullptr>
    constexpr auto get_arg_type(SharedPtr<T>*)
    {
        // Recurse until a non-pointer, non-array type is found.
        return get_arg_type((std::remove_pointer_t<T>*)nullptr);
    }
    /// std::unique_ptr<T, U> handler, Returns pointer to T.
    template<template<typename, typename> class UniquePtr,
        typename T,
        typename U,
        std::enable_if_t<std::is_same_v<std::unique_ptr<T, U>, UniquePtr<T, U>>>* = nullptr>
    constexpr auto get_arg_type(UniquePtr<T, U>*)
    {
        // Recurse until a non-pointer, non-array type is found.
        return get_arg_type((std::remove_pointer_t<T>*)nullptr);
    }

    /// Use to retrieve the template type of a class template T<U>. If not a class
    /// template of this type, the provided type T is retrieved.
    template<typename T>
    using get_arg_type_t
        = std::remove_pointer_t<decltype(get_arg_type((std::remove_pointer_t<T>*)nullptr))>;
    /// @} type_trait_templates_type
    //---------------------------------------------------------------------------------------
    /// @} type_trait_templates

    //=======================================================================================
    /// @defgroup metadata_globals Metadata globals and initialization mechanism.
    /// These classes and functions handled runtime initialization of the global metadata.
    /// @{
    //---------------------------------------------------------------------------------------
    /// Global metadata for data container types and the private map of type IDs to metadata.
    /// @see CTypeInfo
    class CGlobal
    {
        /// Global map initialization type.
        template<class T>
        friend void InitializeMetadata(const SMemberLabel& csLabel, std::vector<SMeta>&& init);
        template<typename T>
        friend void NDataContext::InitializeMetadata(const NMeta::SMemberLabel& csLabel);

        /// The metadata access function.
        /// @returns Reference to the type metadata given a std::type_info if type has
        /// been registered by a CTypeInfo specialized on that type else nullptr.
        /// @note Returned pointer is guaranteed valid only within the context of the
        /// GetValueFn caller.
        friend const SMeta& GetMeta(const std::type_info&);

        /// @returns Global mapping of a data container type index to the data container
        /// metadata.
        static std::unordered_map<std::type_index, SMeta>& GetGlobalMap();
    };
    const SMeta& GetMeta(const std::type_info&);
    /// @} metadata_globals

    //=======================================================================================
    /// @defgroup instance_data Types for handling instance data
    /// These types store instance data, either from an instance of type with registered
    /// metadata or from the metadata values for that instance.
    //---------------------------------------------------------------------------------------
    /// Function type returning the instance values of a member or base class of a type.
    /// Stored in the metadata annotation SMeta struct.
    using ValueFn = std::function<STypeValue&(const SMemberDesc&, STypeValue&)>;
    /// Function type returning the metadata label/description for a member or base class of
    /// a type. Stored in the metadata annotation SMeta struct.
    using LabelFn = std::function<CStr()>;
    /// Groups all metadata functors for a given type. Used to allow adding functions to the
    /// metadata system registered. Use the instance-to-value-info ValueFn functor as an
    /// implementation guide.
    using MetaDataFuncs = std::tuple<ValueFn>;
    /// Type of the description function. Required for enum types, optional for others.
    /// A global default for an annotated enum must exist even if the default is not used.
    /// @note If template argument T is a pointer, the pointer is removed from the
    /// function argument type.
    /// @note If template argument T is an instance of a handled class template wrapper of
    /// a type, the wrapped type is used as the function argument type. For example, for
    /// T = TFormattedNumeric<double>, ValueDescFn<T> is CStr(double).
    /// @see EnumDescription
    template<typename T> using ValueDescFn = std::function<CStr(const std::remove_pointer_t<T>&)>;
    /// Utility struct for converting the label argument to Note. Creates label functions
    /// or stores custom label functions if provided.
    /// @see Note
    struct SMemberLabel
    {
    #if defined(MIC_VCMFC)
        /// Creates functor that returns a string from the provided resource ID. Resource IDs
        /// are used to reduce memory requirements and also to allow use prior to the
        /// initialization of the application string tables.
        SMemberLabel(uint_16 uiLabelID);
    #else
        /// For testing without string loading from resource.
        SMemberLabel(uint_16 uiLabelID) : Label([uiLabelID] { return CStr(Fmt(L"%u") % uiLabelID); })
        {
        }
    #endif
        /// For string literals.
        SMemberLabel(const wchar_t* achLabel);
        /// Creates functor using the provided nullary functor that returns a string.
        SMemberLabel(const LabelFn& Label);
        ~SMemberLabel() = default;

        LabelFn Label;
    };

    //=======================================================================================
    /// @defgroup label_value_utilities Utilities label and value-description functions.
    /// Utility functions for common label and value-description functions.
    /// @{
    //---------------------------------------------------------------------------------------
    /// Forward declaration for default enum value-to-string mapping function. Must be
    /// assignable to ValueDescFn. Must be implemented for all annotated enum types even if
    /// the default is not used. Only used by enums.
    template<typename Enum> CStr EnumDescription(const Enum& e);
    /// @returns "Yes" or "No" depending on whether the flag is true or false, respectively.
    CStr YesNo(bool fIsYes);
    /// @returns "Yes" or "No" depending on whether the type instance evaluates to true or
    /// false, respectively.
    template<typename T> CStr YesNoDescription(const T& t)
    {
        return YesNo(t);
    }
    /// @returns Parenthesized label function that returns "A (B)" given strings A and B.
    LabelFn ParenLabel(const SMemberLabel& csLabelA, const SMemberLabel& csLabelB);
    /// @returns Comma-separated label function that returns "A, B" given strings A and B.
    LabelFn CommaLabel(const SMemberLabel& csLabelA, const SMemberLabel& csLabelB);
    /// @returns Comma-separated label function that returns "A, B, C" given strings A, B, and C.
    LabelFn CommaLabel(const SMemberLabel& csLabelA,
        const SMemberLabel& csLabelB,
        const SMemberLabel& csLabelC);
    /// @returns Label function that performs Fmt(res) % i. Used with resource strings containing a
    /// %d string replacement, e.g., "Description %d"
    LabelFn NumberedLabel(uint_16 uiRes, int iNumber);
    /// The empty string label function.
    const LabelFn EmptyLabel = []() { return CStr(); };
    /// @returns True iff the label function provided is the EmptyLabel function.
    inline bool IsEmptyLabel(const LabelFn& Label)
    {
        return Label.target_type() == EmptyLabel.target_type();
    }
    /// @} label_value_utilities

    //=======================================================================================
    /// Output for Value functions. Contains type description and instance values. Note
    /// that the comparison value if for comparing to instances of a type and should not
    /// be affected by application parameters or configuration (e.g., fixed format and
    /// units). The display value is for logging differences and may contain unit strings.
    struct STypeValue
    {
        STypeValue() {}
        STypeValue(CStr&& rszCompVal, CStr&& rszDispVal, const LabelFn& Label)
            : szCompVal(rszCompVal),
              szDispVal(rszDispVal),
              szDesc(Label ? Label() : EmptyLabel())
        {
        }
        CStr szCompVal;                      ///< Type value for change comparison.
        CStr szDispVal;                      ///< Type value for display (e.g., with units).
        CStr szDesc;                         ///< Type description.
        std::vector<STypeValue> vMember;     ///< Type values for members of a class.
        std::vector<STypeValue> vBaseClass;  ///< Type values for base classes.
    };

    STypeValue& NullValue(const SMemberDesc&, STypeValue&);

    /// Comparison operator for STypeValue. Return true iff the comparison value string is the
    /// same and their member and base class STypeValue vectors are the same (same lengths,
    /// same comparison value strings).
    inline bool operator==(const STypeValue& rhs, const STypeValue& lhs)
    {
        return rhs.szCompVal == lhs.szCompVal && rhs.vMember == lhs.vMember
            && rhs.vBaseClass == lhs.vBaseClass;
    }

    //---------------------------------------------------------------------------------------
    /// Member data for an instance of type with registered metadata.
    struct SMemberDesc
    {
        /// Type-erased pointer to primary member. Cast to char when offset is added.
        const void* p;
        LabelFn Label;  ///< Label function of instance member.
        /// Optional type-erased pointer to secondary member. Cast to char when offset is
        /// added. Default is nullptr (unused).
        const void* p2;
    };
    /// @} instance_data

    //=======================================================================================
    /// @defgroup type_data Type metadata
    /// Types for storing type metadata that has been registered with the global metadata
    /// system.
    /// @see CGlobal
    /// @{
    //---------------------------------------------------------------------------------------
    /// @returns Offset in bytes of member of type U in a data structure of type T.
    template<class T, class U> constexpr size_t OffsetOf(U T::*member)
    {
        return reinterpret_cast<std::ptrdiff_t>(&(reinterpret_cast<T const volatile*>(0)->*member));
    }

    /// @returns Null value string, L"null value".
    CStr GetNullValue();

    //---------------------------------------------------------------------------------------
    /// Metadata annotation struct.
    ///
    /// Type metadata handling is implemented as stored functors and const member
    /// pointer-offset into the type. A valid initialized structure has associated value and
    /// description functions. An offset of zero may indicate (but does not guarantee) that
    /// this is the root (top-level) type of an annotation tree or its data container base
    /// class.
    struct SMeta
    {
    public:
        explicit SMeta(const LabelFn& Label,
            size_t uiOffset,
            boost::optional<size_t> puiOffset2 = boost::none)
            : Value(NullValue),
              Label(Label ? Label : EmptyLabel),
              uiOffset(uiOffset),
              puiOffset2(puiOffset2)
        {
        }
        ~SMeta() = default;
        /// @returns True if there is a value function defined. A valid value function defines
        /// a valid metadata annotation; the offset and label functions may be null.
        operator bool() const
        {
            return static_cast<bool>(Value);
        }  // std::function's bool cast is explicit.
        /// Function that returns the value of the member data or data container base class.
        ValueFn Value;
        /// Function that returns the label string of the member data or data container base
        /// class.
        /// @see SMemberLabel
        LabelFn Label;
        /// Pointer offset in bytes of the member being annotated; zero for root or base class.
        const size_t uiOffset;
        /// Optional offset in bytes of a secondary member required for annotation.
        const boost::optional<size_t> puiOffset2;
    };

    //=======================================================================================
    /// @defgroup metadata_function_factories Type metadata function factories.
    /// These factory functions return functors that handle applying metadata annotations
    /// to an instance of a type with registered metadata.
    /// @{
    //---------------------------------------------------------------------------------------
    /// Error handler for nullptr members of an instance of type with registered metadata.
    inline STypeValue& InvalidPointer(const SMemberDesc& csMember, STypeValue& rsClass)
    {
        // ASSERT_ERROR( false );
    #if defined(_DEBUG)
        TraceMsg(Fmt(L"Meta: invalid pointer class=\"%s\" member=\"%s\"") % rsClass.szDesc
            % csMember.Label());
    #endif
        rsClass.vMember.emplace_back(GetNullValue(), GetNullValue(), csMember.Label);
        return rsClass;
    }

    //---------------------------------------------------------------------------------------
    /// Overload for fundamental type members.
    template<typename T> MetaDataFuncs GetMetaDataFuncs(ValueDescFn<T> ValToDesc)
    {
        if constexpr(std::is_fundamental_v<T> || NUnits::is_quantity_v<T>)
        {
            // For floating point types, use the number of base-10 digits required to uniquely
            // identify all distinct values for the type as the precision but exclude the last
            // few digit to allow rounding of whole numbers.
            int iPrec;
            if constexpr(std::is_fundamental_v<T>)
                iPrec = std::is_floating_point<T>::value ? std::numeric_limits<T>::max_digits10 - 3
                                                         : 0;
            if constexpr(NUnits::is_quantity_v<T>)
                iPrec = std::is_floating_point<typename T::value_type>::value
                    ? std::numeric_limits<typename T::value_type>::max_digits10 - 3
                    : 0;
            return [=](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
                if(!csMember.p)
                    return InvalidPointer(csMember, rsClass);
                std::wostringstream os;
                auto v = *static_cast<const T*>(csMember.p);
                os << std::setprecision(iPrec) << v;
                CStr szCompVal{os.str().c_str()};
                CStr szDispVal{ValToDesc ? ValToDesc(v) : szCompVal};
#if defined(_DEBUG)
                TraceMsg(
                    Fmt(L"class=\"%s\", member=\"%s\", value=\"%s\" (fundamental compare value) %x")
                    % rsClass.szDesc % csMember.Label() % szCompVal % csMember.p);
#endif
                rsClass.vMember.emplace_back(std::move(szCompVal),
                    std::move(szDispVal),
                    csMember.Label);
                return rsClass;
            };
        }

        /// Overload for formatted numerics.
        else if constexpr(is_formatted_numeric_v<T>)
        {
            return [=](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
                if(!csMember.p)
                    return InvalidPointer(csMember, rsClass);
                std::wostringstream os;
                // Restore type.
                auto pxValue = static_cast<const T*>(csMember.p);
                auto v = pxValue->InternalValue();
                os << v;
                CStr szCompVal{os.str().c_str()};
                CStr szDispVal{ValToDesc ? ValToDesc(*pxValue) : pxValue->ValueAndUnits()};
#if defined(_DEBUG)
                TraceMsg(
                    Fmt(L"class=\"%s\", member=\"%s\", value=\"%s\" (numeric compare value) %x")
                    % rsClass.szDesc % csMember.Label() % szCompVal % csMember.p);
#endif
                rsClass.vMember.emplace_back(std::move(szCompVal),
                    std::move(szDispVal),
                    csMember.Label);
                return rsClass;
            };
        }

        /// Overload for enums and CFixedEnum.
        else if constexpr(is_enum_type_v<T>)
        {
            using Enum = get_arg_type_t<T>;
            // Default value-to-string mapping function required for enums by all annotated
            // enums and fixed enums.
            if(!ValToDesc)
                ValToDesc = EnumDescription<Enum>;
            return [=](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
                if(!csMember.p)
                    return InvalidPointer(csMember, rsClass);
                // Restore type.
                auto e = *static_cast<const T*>(csMember.p);
                CStr szCompVal(Fmt(L"%d") % e);
                CStr szDispVal = ValToDesc(e);
                if(szDispVal.IsEmpty())
                    szDispVal = GetNullValue();
#if defined(_DEBUG)
                TraceMsg(Fmt(L"class=\"%s\", member=\"%s\", value=\"%s\" (enum compare value) %x")
                    % rsClass.szDesc % csMember.Label() % szCompVal % csMember.p);
#endif
                rsClass.vMember.emplace_back(std::move(szCompVal),
                    std::move(szDispVal),
                    csMember.Label);
                return rsClass;
            };
        }

        /// Overload for CStr.
        else if constexpr(is_str_v<T>)
        {
            return [=](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
                if(!csMember.p)
                    return InvalidPointer(csMember, rsClass);

                CStr szCompVal{*static_cast<const CStr*>(csMember.p)};
                CStr szDispVal{ValToDesc ? ValToDesc(szCompVal) : Fmt(L"\"%s\"") % szCompVal};
#if defined(_DEBUG)
                TraceMsg(Fmt(L"class=\"%s\", member=\"%s\", value=\"%s\" (CStr compare value) %x")
                    % rsClass.szDesc % csMember.Label() % szCompVal % csMember.p);
#endif
                rsClass.vMember.emplace_back(std::move(szCompVal),
                    std::move(szDispVal),
                    csMember.Label);
                return rsClass;
            };
        }
        // Other types are not supported.
        else
            static_assert(dependent_false_v<T>, "Type is not supported.");
    }

    //---------------------------------------------------------------------------------------
    /// Handles special case: iterable containers (std::array, std::vector, TDynamicTableBase).
    /// IndexToDesc function, if valid, translates the index to a string.
    template<typename Array,
        typename IndexType = size_t,
        std::enable_if_t<is_std_array_v<Array> || is_std_vector_v<Array> || is_std_list_v<Array>
            || is_meta_iterable_v<Array>>* = nullptr>
    MetaDataFuncs GetMetaDataFuncsArrayMember(MetaDataFuncs funcs, ValueDescFn<IndexType> IndexToDesc)
    {
        ASSERT_ERROR(static_cast<bool>(IndexToDesc));  // std::function's bool cast is explicit.

        // Use value function for the element type.
        ValueFn Value;
        std::tie(Value) = funcs;
        return [Value, IndexToDesc](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
            if(!csMember.p)
                return InvalidPointer(csMember, rsClass);

            auto pArray = static_cast<const Array*>(csMember.p);
            auto sArrayVal = STypeValue();
            sArrayVal.szDesc = csMember.Label();

            size_t i = 0;
            for(const auto& cVal : *pArray)
            {
                // Use a temporary to get the element value. Element value is added to the
                // STypeValue member vector.
                auto sTemp = STypeValue();
                Value({&cVal, EmptyLabel}, sTemp);
                // There should be one and only one element in the member vector and the
                // rest of the type should be default.
                ASSERT_ERROR(sTemp.vMember.size() == 1 && sTemp.vBaseClass.empty()
                    && sTemp.szCompVal.IsEmpty() && sTemp.szDispVal.IsEmpty()
                    && sTemp.szDesc.IsEmpty());
                // Should only loop once.
                for(auto& rsElVal : sTemp.vMember)
                {
                    rsElVal.szDesc = IndexToDesc(static_cast<IndexType>(i)) + rsElVal.szDesc;
    #if defined(_DEBUG)
                    TraceMsg(Fmt(L"class=\"%s\", member=\"%s\", value=\"%s\" (array el desc) %x")
                        % rsClass.szDesc % csMember.Label() % rsElVal.szDesc % (&cVal));
    #endif
                    sArrayVal.vMember.push_back(std::move(rsElVal));
                }
                ++i;
            }
            rsClass.vMember.push_back(std::move(sArrayVal));
            return rsClass;
        };
    }

    //---------------------------------------------------------------------------------------
    /// Handles special case: iterable containers (std::array, std::vector, TDynamicTableBase).
    template<typename Array,
        std::enable_if_t<is_std_array_v<Array> || is_std_vector_v<Array> || is_std_list_v<Array>
            || is_meta_iterable_v<Array>>* = nullptr>
    MetaDataFuncs GetMetaDataFuncsArrayMember(MetaDataFuncs funcs)
    {
        return GetMetaDataFuncsArrayMember<Array>(funcs,
            [](const size_t& i) -> CStr { return Fmt(L"%u") % (i + 1); });
    }

    //---------------------------------------------------------------------------------------
    /// Handles special case: data container member.
    template<class DC, std::enable_if_t<is_data_container_v<DC>>* = nullptr>
    MetaDataFuncs GetMetaDataFuncsContainerMember()
    {
        // Data container members are pushed onto the member vector and, if a pointer,
        // may point to a derived type of the specified DC class.
        return [](const SMemberDesc& csDC, STypeValue& rsClass) -> STypeValue& {
            if(!csDC.p)
                return InvalidPointer(csDC, rsClass);

            // Get type info for the data container. This is a run-time type value; DC*
            // can be a base class pointer.
            auto& csMeta = GetMeta(typeid(*static_cast<const DC*>(csDC.p)));
            // Get the member data container metadata.
            auto sMember = STypeValue();
    #ifdef _DEBUG
            auto szLabelDC = csDC.Label();
            auto szLabelMeta = csMeta.Label();
            TraceMsg(Fmt(L"Container DC=\"%s\" Meta=\"%s\"") % szLabelDC % szLabelMeta);
    #endif
            sMember.szDesc = csDC.Label();
            ASSERT_ERROR(csMeta);
            if(csMeta)
                csMeta.Value({static_cast<const char*>(csDC.p), csMeta.Label}, sMember);
            // Add the member data container metadata.
            rsClass.vMember.emplace_back(std::move(sMember));
            return rsClass;
        };
    }

    //---------------------------------------------------------------------------------------
    /// Handles special case: data container base class.
    template<class DC, std::enable_if_t<is_data_container_v<DC>>* = nullptr>
    MetaDataFuncs GetMetaDataFuncsContainerBase()
    {
        // Data container bases are pushed onto the base class vector. They will
        // always be the specified DC class.
        return [](const SMemberDesc& csDC, STypeValue& rsClass) -> STypeValue& {
            if(!csDC.p)
                return InvalidPointer(csDC, rsClass);

            // Get type info for the data container. This is a compile-time type value.
            auto& csMeta = GetMeta(typeid(DC));
            // Get the base data container metadata.
            auto sBaseClass = STypeValue();
    #ifdef _DEBUG
            auto szLabelDC = csDC.Label();
            auto szLabelMeta = csMeta.Label();
            TraceMsg(Fmt(L"Container DC=\"%s\" Meta=\"%s\"") % szLabelDC % szLabelMeta);
    #endif
            sBaseClass.szDesc = csDC.Label();
            ASSERT_ERROR(csMeta);
            if(csMeta)
                csMeta.Value({static_cast<const char*>(csDC.p), csMeta.Label}, sBaseClass);
            // Add the base data container metadata.
            rsClass.vBaseClass.emplace_back(sBaseClass);
            return rsClass;
        };
    }

    //---------------------------------------------------------------------------------------
    /// Handles special case: root data container class.
    template<class DC, std::enable_if_t<is_data_container_v<DC>>* = nullptr>
    MetaDataFuncs GetMetaDataFuncsContainerRoot(std::vector<SMeta>&& vsMeta)
    {
        static_assert(is_data_container_v<DC>,
            "NMeta::GetMetaDataFuncs: NMeta::IsDataContainer-derived type expected");
        //! TODO enable to remove copy if necessary via rvalue move to lambda.
        //! return [vsMeta{ std::move( vsMeta ) }]( const SMemberDesc& csDC, STypeValue& rsClass )->STypeValue& {
        return [vsMeta](const SMemberDesc& csDC, STypeValue& rsClass) -> STypeValue& {
            if(!csDC.p)
                return InvalidPointer(csDC, rsClass);

            // If the class already has a description, retain it. This occurs if this
            // data container is a base class or member of the root data container type.
            for(const auto& s : vsMeta)
            {
                ASSERT_ERROR(s);
    #if defined(_DEBUG)
                TraceMsg(Fmt(L"Meta Value: \"%s\" %X = %X + %u (offset) %u (offset2)") % s.Label()
                    % (int)((const char*)csDC.p + s.uiOffset) % (int)csDC.p % s.uiOffset
                    % (s.puiOffset2 ? *s.puiOffset2 : 0));
    #endif
                if(s)
                    s.Value({static_cast<const char*>(csDC.p) + s.uiOffset,
                                s.Label,
                                s.puiOffset2 ? static_cast<const char*>(csDC.p) + *s.puiOffset2
                                             : nullptr},
                        rsClass);
            }
            return rsClass;
        };
    }

    //---------------------------------------------------------------------------------------
    /// Overload for unhandled type.
    inline MetaDataFuncs GetMetaDataFuncs()
    {
        ASSERT_ERROR(false);
        return MetaDataFuncs();
    }
    /// @} metadata_function_factories

    //=======================================================================================
    /// @defgroup metadata_custom_functions Common custom metadata functions.
    /// @{
    template<typename T> MetaDataFuncs GetMetaDataFuncsNumMatrix()
    {
        static_assert(std::is_fundamental_v<T>,
            "NMeta::NumMatrixFuncs: only fundamental types supported");
        // See GetMetaDataFuncs for fundamental types.
        const int ciPrec
            = std::is_floating_point<T>::value ? std::numeric_limits<T>::max_digits10 - 3 : 0;
        return [=](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
            if(!csMember.p)
                return InvalidPointer(csMember, rsClass);

            auto pMatrix = static_cast<const CNumMatrix<T>*>(csMember.p);
            auto sMatrixVal = STypeValue();
            // Handle null label functions.
            CStr szLabel = csMember.Label();
            sMatrixVal.szDesc = szLabel + (szLabel.IsEmpty() ? L"" : L" ")
                + Fmt(L"rows=%u cols=%u") % pMatrix->Rows() % pMatrix->Columns();
            for(size_t i = 0; i < pMatrix->Rows(); ++i)
            {
                CStr szRow(Fmt(L"%d:") % (i + 1));
                for(size_t j = 0; j < pMatrix->Columns(); ++j)
                {
                    std::wostringstream os;
                    auto v = *static_cast<const T*>(csMember.p);
                    os << std::setprecision(ciPrec) << v;
                    szRow += CStr{os.str().c_str()} + L",";
                }
                CStr szDispVal{szRow};
                rsClass.vMember.emplace_back(std::move(szRow), std::move(szDispVal), csMember.Label);
            }
            return rsClass;
        };
    }
    /// @} metadata_custom_functions

    //=======================================================================================
    /// @defgroup metadata_factories SMetaData factories
    /// Retrieve metadata structure containing description information for a member or
    /// base class, member pointer offsets (zero for base classes), and metadata function
    /// objects that encode the member of base class type information.
    /// @{
    /// Metadata for std::array with an enumerated index.
    template<typename E, class DC, typename T, size_t N>
    SMeta Note(std::array<T, N> DC::*member, SMemberLabel sLabel, ValueDescFn<T> ValToDesc)
    {
        static_assert(std::is_enum_v<E>,
            "NMeta::Note: std::array with enumerated index requires enum type argument.");
        auto m = SMeta(sLabel.Label, OffsetOf(member));
        auto el = Note((T DC::*)member, sLabel.Label, ValToDesc);
        std::tie(m.Value)
            = GetMetaDataFuncsArrayMember<std::array<T, N>, E>(el.Value, EnumDescription<E>);
        return m;
    }
    /// Metadata for std::array with an enumerated index.
    template<typename E, class DC, typename T, size_t N>
    SMeta Note(std::array<T, N> DC::*member, SMemberLabel sLabel)
    {
        static_assert(std::is_enum_v<E>,
            "NMeta::Note: std::array with enumerated index requires enum type argument.");
        auto m = SMeta(sLabel.Label, OffsetOf(member));
        auto el = Note((T DC::*)member, sLabel.Label);
        std::tie(m.Value)
            = GetMetaDataFuncsArrayMember<std::array<T, N>, E>(el.Value, EnumDescription<E>);
        return m;
    }
    //---------------------------------------------------------------------------------------
    /// @defgroup metadata_factories_custom Custom metadata function support
    /// Supports providing the set of metadata functions required for processing the given
    /// type
    /// @{
    //---------------------------------------------------------------------------------------
    template<class DC, class T> SMeta Note(T DC::*member, SMemberLabel sLabel, MetaDataFuncs funcs)
    {
        auto m = SMeta(sLabel.Label, OffsetOf(member));
        std::tie(m.Value) = funcs;
        return m;
    }

    //---------------------------------------------------------------------------------------
    /// Currently ignored entries. For annotating the annotation, future work.
    /// @returns Metadata structure with do-nothing functions.
    /// effectively do-nothing
    template<class DC, class T> SMeta NoteIgnored(T DC::*member)
    {
        auto m = SMeta(EmptyLabel, OffsetOf(member));
        m.Value = [](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
            // Do nothing.
            return rsClass;
        };
        return m;
    }
    /// @} metadata_factories_custom

    //---------------------------------------------------------------------------------------
    /// @defgroup metadata_factories_basecontainer NMeta::IsDataContainer support
    /// @note NMeta::IsDataContainer-derived types may be the root (top-level) type of the
    /// annotation tree, an annotated  base class of the root type or an annotated data member
    /// derived from NMeta::IsDataContainer. For hierarchical reasons, the different type
    /// usages are handled separately
    //---------------------------------------------------------------------------------------
    /// Metadata for the root (top-level) data container of an annotation tree. Provide
    /// the class description ID and metadata vector. Used internally during registration.
    /// @see CTypeInfo
    template<class DC> SMeta NoteRoot(const SMemberLabel& csLabel, std::vector<SMeta>&& init)
    {
        static_assert(is_data_container_v<DC>,
            "NMeta::Note: requires NMeta::IsDataContainer type argument.");
        auto m = SMeta(csLabel.Label, 0);
        std::tie(m.Value) = GetMetaDataFuncsContainerRoot<DC>(std::move(init));
        return m;
    }
    /// Metadata for the base data container type of the root data container type.
    template<class DC> SMeta NoteBase(SMemberLabel sLabel = LabelFn())
    {
        static_assert(is_data_container_v<DC>,
            "NMeta::NoteBase: requires NMeta::IsDataContainer type argument.");
        auto m = SMeta(sLabel.Label, 0);
        std::tie(m.Value) = GetMetaDataFuncsContainerBase<DC>();
        return m;
    }
    /// @} metadata_factories_basecontainer

    template<class DC, class T>
    SMeta Note(T DC::*member, SMemberLabel sLabel, ValueDescFn<get_arg_type_t<T>> ValToDesc)
    {
        /// Non-pointer, non-container types.
        if constexpr(!std::is_pointer_v<T> && !is_smart_ptr_v<T> && !is_data_container_v<T>
            && !is_std_array_v<T> && !is_std_vector_v<T> && !is_std_list_v<T>
            && !is_meta_iterable_v<T> && !is_num_matrix_v<T>)
        {
            static_assert(!std::is_array<T>::value,
                "NMeta::Note: C-style arrays are not supported. Replace with std::array");
            auto m = SMeta(sLabel.Label, OffsetOf(member));
            std::tie(m.Value) = GetMetaDataFuncs<T>(ValToDesc);
            return m;
        }
        /// Handler for metadata given a std::shared_ptr or std::unique_ptr member.
        else if constexpr(is_smart_ptr_v<T>)
        {
            // Get metadata for the managed type.
            using U = typename T::element_type;
            auto m = Note((U DC::*)member, sLabel.Label, ValToDesc);
            auto Value = m.Value;
            // Handle the dereference.
            m.Value = [Value](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
                using V = const T* const;
                if(!csMember.p || !static_cast<V>(csMember.p)->get())
                    return InvalidPointer(csMember, rsClass);
                return Value({static_cast<V>(csMember.p)->get(), csMember.Label}, rsClass);
            };
            return m;
        }
        /// Handler for metadata given a raw pointer member.
        else if constexpr(std::is_pointer_v<T>)
        {
            static_assert(!std::is_pointer_v<std::remove_pointer_t<T>>,
                "NMeta::Note: Handles disallowed");
            static_assert(!std::is_same_v<T, char*>,
                "NMeta::Note: C-style strings are not handled; char* disallowed");

            // Get metadata for the dereferenced type.
            using U = std::remove_pointer_t<T>;
            auto m = Note((U DC::*)member, sLabel, ValToDesc);
            auto Value = m.Value;
            // Handle the pointer dereference.
            m.Value = [Value](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
                using V = const U* const*;
                if(!csMember.p || !*static_cast<V>(csMember.p))
                    return InvalidPointer(csMember, rsClass);
                return Value({*static_cast<V>(csMember.p), csMember.Label}, rsClass);
            };
            return m;
        }
        /// Metadata for std::array.
        else if constexpr(is_std_array_v<T>)
        {
            using U = typename T::value_type;
            auto m = SMeta(sLabel.Label, OffsetOf(member));
            auto el = Note((U DC::*)member, sLabel.Label, ValToDesc);
            std::tie(m.Value) = GetMetaDataFuncsArrayMember<T>(el.Value);
            return m;
        }
        /// Metadata for std::vector.
        else if constexpr(is_std_vector_v<T> || is_std_list_v<T> || is_meta_iterable_v<T>)
        {
            using U = typename T::value_type;
            auto m = SMeta(sLabel.Label, OffsetOf(member));
            auto el = Note((U DC::*)member, sLabel.Label, ValToDesc);
            std::tie(m.Value) = GetMetaDataFuncsArrayMember<T>(el.Value);
            return m;
        }
        /// Support for CNumMatrix.
        else if constexpr(is_num_matrix_v<T>)
        {
            auto m = SMeta(sLabel.Label, OffsetOf(member), ValToDesc);
            std::tie(m.Value) = GetMetaDataFuncsNumMatrix<typename T::value_type>();
            return m;
        }
        /// Metadata for NMeta::IsDataContainer members of the root data container type.
        /// @note The value-to-string mapping function is required by interface design but
        /// should not be supplied.
        else if constexpr(is_data_container_v<T> && !std::is_enum_v<T> && !NUnits::is_quantity_v<T>
            && !std::is_arithmetic_v<T> && !is_meta_iterable_v<T>)
        {
            ASSERT_ERROR(!ValToDesc);
            auto m = SMeta(sLabel.Label, OffsetOf(member));
            std::tie(m.Value) = GetMetaDataFuncsContainerMember<T>();
            return m;
        }

        // Other types are not supported.
        else
            static_assert(dependent_false_v<T>, "Type is not supported.");
    }
    template<class DC, class T> SMeta Note(T DC::*member, SMemberLabel sLabel)
    {
        /// Non-pointer, non-container types.
        if constexpr(!std::is_pointer_v<T> && !is_smart_ptr_v<T> && !is_data_container_v<T>
            && !is_std_array_v<T> && !is_std_vector_v<T> && !is_std_list_v<T>
            && !is_meta_iterable_v<T> && !is_num_matrix_v<T>)
        {
            return Note(member, sLabel, ValueDescFn<get_arg_type_t<T>>());
        }

        /// Handler for metadata given a std::shared_ptr or std::unique_ptr member.
        else if constexpr(is_smart_ptr_v<T>)
        {
            // Get metadata for the managed type.
            using U = typename T::element_type;
            auto m = Note((U DC::*)member, sLabel.Label);
            auto Value = m.Value;
            // Handle the dereference.
            m.Value = [Value](const SMemberDesc& csMember, STypeValue& rsClass) -> STypeValue& {
                using V = const T* const;
                if(!csMember.p || !static_cast<V>(csMember.p)->get())
                    return InvalidPointer(csMember, rsClass);
                return Value({static_cast<V>(csMember.p)->get(), csMember.Label}, rsClass);
            };
            return m;
        }

        /// Handler for metadata given a raw pointer member.
        else if constexpr(std::is_pointer_v<T>)
        {
            return Note(member, sLabel, ValueDescFn<get_arg_type_t<T>>());
        }

        /// Metadata for std::array.
        else if constexpr(is_std_array_v<T>)
        {
            using U = typename T::value_type;
            auto m = SMeta(sLabel.Label, OffsetOf(member));
            auto el = Note((U DC::*)member, sLabel.Label);
            std::tie(m.Value) = GetMetaDataFuncsArrayMember<T>(el.Value);
            return m;
        }

        /// Metadata for std::vector.
        else if constexpr(is_std_vector_v<T> || is_std_list_v<T> || is_meta_iterable_v<T>)
        {
            using U = typename T::value_type;
            auto m = SMeta(sLabel.Label, OffsetOf(member));
            auto el = Note((U DC::*)member, sLabel.Label);
            std::tie(m.Value) = GetMetaDataFuncsArrayMember<T>(el.Value);
            return m;
        }

        /// Support for CNumMatrix.
        else if constexpr(is_num_matrix_v<T>)
        {
            auto m = SMeta(sLabel.Label, OffsetOf(member));
            std::tie(m.Value) = GetMetaDataFuncsNumMatrix<typename T::value_type>();
            return m;
        }

        /// Metadata for NMeta::IsDataContainer members of the root data container type.
        /// @note The value-to-string mapping function is required by interface design but
        /// should not be supplied.
        else if constexpr(is_data_container_v<T> && !std::is_enum_v<T> && !NUnits::is_quantity_v<T>
            && !std::is_arithmetic_v<T> && !is_meta_iterable_v<T>)
        {
            auto m = SMeta(sLabel.Label, OffsetOf(member));
            std::tie(m.Value) = GetMetaDataFuncsContainerMember<T>();
            return m;
        }

        // Other types are not supported.
        else
            static_assert(dependent_false_v<T>, "Type is not supported.");
    }
    //---------------------------------------------------------------------------------------
    /// @} metadata_factories

    //=======================================================================================
    /// @defgroup initialization_type Initialization for a type annotation.
    /// @{
    //---------------------------------------------------------------------------------------
    /// Used for initialization of type metadata with the global metadata system. Use to
    /// register all data container types. Create a static CTypeInfo<derived> instance and implement
    /// the derived::InitializeMetaData static method.
    /// @note Only implementation requirements are defining the TypeInfo static instance
    /// with the associated metadata structures.
    /// @note Derived classes must implement the static method InitializeMetaData that
    /// calls NMeta::InitializeMetaData
    template<class T> class CTypeInfo
    {
    public:
        CTypeInfo();
        ~CTypeInfo() = default;
        CTypeInfo& operator=(CTypeInfo&) = delete;
    };

    template<typename T> CTypeInfo<T>::CTypeInfo()
    {
        // InitializeMetaData(void) is a required implementation detail for registering type metadata.
        T::InitializeMetaData();
    }

    //---------------------------------------------------------------------------------------
    template<typename T>
    void InitializeMetadata(const SMemberLabel& csLabel, std::vector<SMeta>&& init)
    {
        static_assert(std::is_base_of_v<NMeta::IsDataContainer, T>,
            "NMeta::InitializeMetadata: Only NMeta::IsDataContainer-derived classes are supported "
            "for "
            "metadata annotation.");
        // Called once on instantiation of the static member.
        ASSERT_ERROR(CGlobal::GetGlobalMap().count(typeid(T)) == 0);

        auto index = std::type_index(typeid(T));
        ASSERT_ERROR(CGlobal::GetGlobalMap().count(index) == 0);
        CGlobal::GetGlobalMap().insert({index, NoteRoot<T>(csLabel, std::move(init))});
    }
    /// @} initialization_type

    //=======================================================================================
    /// @defgroup instance_functions Instance functions
    /// The functions apply metadata to extract value-based information for instances of
    /// types with registered metadata.
    /// @{
    //---------------------------------------------------------------------------------------
    /// @returns Type/value tree for an instance of the annotated type DC. DC must be an
    /// annotated data-container-like type, i.e., DC must derive from
    /// NMeta::IsDataContainer.
    template<class DC> STypeValue GetValues(DC* pDC)
    {
        static_assert(is_data_container_v<DC>,
            "NMeta::GetValues: requires NMeta::IsDataContainer instance for argument.");
        STypeValue ret;
        return pDC ? GetMeta(typeid(*pDC)).Value({pDC, 0}, ret) : ret;
    }

    //---------------------------------------------------------------------------------------
    template<class DC> STypeValue GetValues(std::shared_ptr<DC> pDC)
    {
        static_assert(is_data_container_v<DC>,
            "NMeta::GetValues: requires NMeta::IsDataContainer instance for argument.");
        STypeValue ret;
        return pDC ? GetMeta(typeid(*pDC)).Value({pDC.get(), 0}, ret) : ret;
    }

    //---------------------------------------------------------------------------------------
    /// Test function to print type/value tree to the console standard out. This function is recursive.
    inline void PrintValues(const STypeValue& csClass, int iLevel = 0, bool fIsBaseClass = true)
    {
        std::cout << std::string(2 * iLevel, ' ') << (fIsBaseClass || iLevel == 0 ? "" : "name: ")
                  << csClass.szDesc.GetAsciiPtr() << '\n';
        ++iLevel;
        for(const auto& c : csClass.vBaseClass)
            PrintValues(c, iLevel);
        for(const auto& m : csClass.vMember)
        {
            if(m.vMember.empty() && m.vBaseClass.empty())
                std::cout << std::string(4 * iLevel, ' ') << "name: " << m.szDesc.GetAsciiPtr()
                          << ", "
                          << "value: " << m.szDispVal.GetAsciiPtr() << ", "
                          << "comp: " << m.szCompVal.GetAsciiPtr() << '\n';
            else
                PrintValues(m, iLevel, false);
        }
    }
    /// @} instance_functions

}  // namespace NMeta

#endif //METADATA_HPP
