#ifndef NUCPP_H
#define NUCPP_H
/* nucpp.h - defines nucleus equivalents in TI 642 DSP equivalents */

#if !defined(MIC_DM643)
#error This file is only supported for dm643 targets.
#endif

#include <std.h>
#include <clk.h>
#include <lck.h>
#include <mbx.h>
#include <sem.h>
#include <tsk.h>
#include "tskpri.h"

typedef unsigned long           UNSIGNED;
typedef long                    SIGNED;
typedef unsigned char           DATA_ELEMENT;
typedef DATA_ELEMENT            OPTION;
typedef int                     STATUS;
typedef unsigned char           UNSIGNED_CHAR;
typedef char                    CHAR;
typedef int                     INT;
typedef void                    VOID;
typedef unsigned long *    UNSIGNED_PTR;
typedef unsigned char *    BYTE_PTR;
typedef unsigned char	BOOL;

typedef MBX_Handle		NU_MAILBOX;

#define         NU_AND                  2
#define         NU_AND_CONSUME          3
#define         NU_DISABLE_TIMER        4
#define         NU_ENABLE_TIMER         5
#define         NU_FALSE                0
#define         NU_FIFO                 6
#define         NU_FIXED_SIZE           7
#define         NU_NO_PREEMPT           8
#define         NU_NO_START             9
#define         NU_NO_SUSPEND           0UL
#define         NU_NULL                 0
#define         NU_OR                   0
#define         NU_OR_CONSUME           1
#define         NU_PREEMPT              10
#define         NU_PRIORITY             11
#define         NU_START                12
#define         NU_SUSPEND              0xFFFFFFFFUL
#define         NU_TRUE                 1
#define         NU_VARIABLE_SIZE        13
#define         NU_SUCCESS               0
#define         NU_TIMEOUT              -50
#define         NU_INVALID_TASK         -25
#define			NU_MAILBOX_EMPTY		-29
#define			NU_MAILBOX_FULL			-30

// nucleus task priorities
#define PRIO_MAX	0
#define PRIO_REALTIME_HI	50	//tasks requiring fast realtime response
#define PRIO_REALTIME_MED	100	//tasks requiring medium realtime response
#define PRIO_REALTIME_LO	150	//tasks requiring low realtime response
#define PRIO_I2C			175	//tasks talking over I2C Bus
#define PRIO_USER			200	//user interface tasks
#define PRIO_BATCH			240	//batch tasks that don't release control often 
#define PRIO_MIN			255

#define SLICE_DEFAULT		1	//default time slice value

//#define STACK_DEFAULT		16384 //8192	//use this unless you know requirements

VOID            NU_Sleep(UNSIGNED ticks);
uint_32 TimeoutFromNucSuspend(UNSIGNED suspend);
uint_16 PrioFromNucPrio(OPTION priority);

class Event
{
	public:
		Event();
		virtual ~Event();
		STATUS Set();
		STATUS Clear();
		STATUS Retrieve( BOOL consume = TRUE, UNSIGNED suspend = NU_SUSPEND );
	private:
		SEM_Handle m_hSEM;
};

class EventGroup
{
	public:
		EventGroup( CHAR *name=NULL );

		// Destructor for EventGroup object.
		virtual ~EventGroup();

		// Sets the specified event flags in this group. Any task 
		// waiting on the event group whose event flag request is
		// satisfied by the service is resumed.
		STATUS Set( UNSIGNED eventFlags, OPTION operation );

		// Retrieves the specified event flag combination from this
		// event group. If the combination is present, the service 
		// completes immediately.
		// NOTE: Retrieve is only guaranteed to work with one consumer. See code comments.
		STATUS Retrieve(UNSIGNED	requestedEvents, 
						OPTION	operation,
						UNSIGNED	*retrievedEvents,
						UNSIGNED	suspend = NU_SUSPEND);
		// same as above, except pass millisec for suspend
		STATUS RetrieveMS(UNSIGNED	requestedEvents, 
						  OPTION	operation,
						  UNSIGNED	*retrievedEvents,
						  uint_32	suspend = NU_SUSPEND);

		// tells whether a flag (or any flag in a mask) is set
		UNSIGNED Flags() { return m_uiFlags; }
		
	private:
		LCK_Handle m_hLCK;
		SEM_Handle m_hSEMChanged;
		UNSIGNED m_uiFlags;
};

class Semaphore
{
	public:
		// Constructor for Semaphore object.
		Semaphore(CHAR		*name,			// Name of semaphore.
				  UNSIGNED	initialCount,	// Initial count upon creation.
				  OPTION	suspend);		// Specifies how tasks suspend

		// Destructor for Semaphore object.
		virtual ~Semaphore();

		// Obtain sempahore request. Decrements the semaphore.
		STATUS Obtain( UNSIGNED suspend = NU_SUSPEND );

		// Release sempahore request. Increments the semaphore.
		STATUS Release();

		// Resets semaphore object to an initialCount.
		STATUS Reset( UNSIGNED initialCount = 1 );
	private:
		SEM_Handle m_hSEM;
};

class SystemClock
{
	public:
		// Constructor.
		SystemClock();

		// Destructor.
		virtual ~SystemClock();

		// Returns system clock.
		static UNSIGNED Retrieve();

		// Sets system clock.
		static void Set( UNSIGNED newValue );
};

class Task
{
	public:
		Task
				( 
				 CHAR*       name,       // The task's name (max of 7 char).

				 UNSIGNED    stackSize,  // Number of bytes for the task stack.

				 OPTION      priority,   // Priority between 0 and 255 (0 high).

				 UNSIGNED    timeSlice,  // The maximum number of clock ticks that 
									// can expire while executing this task
									// before other tasks of equal priority
									// are allowed to execute. A value of zero 
									// for timeslice disables time slicing for 
									// this task.

				 BOOL        preEmption  // TRUE indicates that this task is 
									// preemptable and FALSE indicates 
									// that it is not. Time slicing is 
									// disabled if the task is not 
									// preemptable.
				);

		// Destructor for task object.

		virtual ~Task();                                       

		STATUS Start();

		virtual VOID Entry() = 0;

		void Sleep(UNSIGNED uiTicks) { TSK_sleep(uiTicks * 10); }

		TSK_Handle Handle() { return m_hTSK; }
		
	private:
		enum { ceMaxNameLength = 32 };

		TSK_Handle	m_hTSK;
		TSK_Attrs	m_attrs;
		char        m_szName[ ceMaxNameLength ];
};

STATUS NU_Create_Mailbox(NU_MAILBOX *pNU_MBX,CHAR *name, OPTION suspend_type);
STATUS NU_Delete_Mailbox(NU_MAILBOX *pNU_MBX);
STATUS NU_Send_To_Mailbox(NU_MAILBOX *pNU_MBX,VOID *message,UNSIGNED suspend);
STATUS NU_Receive_From_Mailbox(NU_MAILBOX *pNU_MBX,VOID *message,UNSIGNED suspend);

#endif
