#ifndef NUMDIST_HPP
#define NUMDIST_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
// ============================================================================
///  This class is used to calculate 1-D statistics on a set of possibly
///  weighted numbers. To use the class just create an object of this type
///  (either empty or with an initial set of numbers) and then Add() all the
///  numbers of interest, finaly call the staticic functions that calulate
///  the statistics of interest.
class CNumDistribution
{
public:
    CNumDistribution();
    CNumDistribution(const CNumDistribution& src);
    CNumDistribution& operator=(const CNumDistribution& rhs);
    int operator==(const CNumDistribution& rhs) const;
    int operator!=(const CNumDistribution& rhs) const;

    ///  Creates a distribution of the @p N points of @p adX.
    CNumDistribution(size_t N, const double* adX);
    ///  Creates a distribution of the @p N points of @p aflX.
    CNumDistribution(size_t N, const float* aflX);
    ///  Creates an initialy empty distribution, and adds the single point x to it.
    CNumDistribution(double x);

    //  Creates a distribution with the @p N points @p cadX with wieghts @p cadW.
    CNumDistribution(size_t N, const double* cadX, const double* cadW);
    //  Creates a distribution with the @p N points @p caflX with wieghts @p caflW.
    CNumDistribution(size_t N, const float* caflX, const float* caflW);
    ///  Creates a distribution with a single point @p dX with weight @p dW.
    CNumDistribution(double dX, double dW);

    ///  Adds all the @p N  points @p cadX to the distribution.
    void Add(size_t N, const double* cadX);
    ///  Adds all the @p N  points @p aflX to the distribution.
    void Add(size_t N, const float* aflX);
    ///  Adds the single point $x$ to the distribution.
    void Add(double x);

    ///  Adds all the @p N  points @p adX to the distribution with wight @p adW.
    void Add(size_t N, const double* cadX, const double* cadW);
    ///  Adds all the @p N  points @p aflX to the distribution with wight @p aflW.
    void Add(size_t N, const float* caflX, const float* caflW);
    ///  Adds the single point @p dX to the distribution and gives it a weight dW.
    void Add(double dX, double dW);

    ///  Add a whole nother distribution to this one.
    void Add(const CNumDistribution& rDist);

    /// The total weight of the numbers so far. (or number if weights are not
    /// used.
    double Weight() const { return m_dSX0; }
    ///  Calculates the weighted average of the points in the distribution.
    ///  @f$ \bar x = { \sum x_i w_i \over \sum w_i }@f$
    double Ave() const;
    ///  Calculates the weighted variance of the points in the distribution.
    ///  @f$ \sigma ^2 = { \sum (x_i - \bar x)^2 w_i \over \sum w_i }@f$
    double Var() const;
    ///  Calculates the standard deviation (square root of the variance) of the
    ///  points in the distribution.
    ///  @f$ \sigma = \sqrt {\sigma ^2}
    ///            = \sqrt{( \sum (x_i - \bar x)^2 w_i \over \sum w_i )} @f$
    double StdDev() const;
    ///  Calculates the weighted variance of the points in the distribution.
    ///  @f$ \sigma _Sample ^2 = { \sum (x_i - \bar x)^2 w_i } \over { n-1 } @f$
    double SampleVar() const;
    ///  Calculates the standard deviation (square root of the variance) of the
    ///  points in the distribution.
    /// @f$ \sigma = \sqrt {\sigma _Sample ^2}
    ///            = \sqrt{ { \sum (x_i - \bar x)^2 w_i } \over { n-1 } } @f$
    double SampleStdDev() const;
    ///  Calculates the weighted coeficient of variance of the points in the
    ///  distribution.
    ///  @f$ CoefVar = { \sigma \over \bar x }@f$
    double CoefVar() const;
    ///  Calculates the weighted skewness of the points in the distribution.
    ///  @f$ \alpha _3 = { \mu _3 \over \sigma^3 } @f$
    ///  @f$\mu _3 = { \sum (x_i - \bar x) ^3 w_i \over \sum w_i } @f$
    ///  @f$ \sigma ^2 = { \sum (x_i - \bar x) ^2 w_i \over \sum w_i } @f$
    double Skewness() const;
    ///  Calculates the weighted skewness of the points in the distribution.
    ///  @f$ \alpha _4 = { \mu _4 \over \sigma^4 } - 3 @f$
    ///  @f$ \mu _4 = { \sum (x_i - \bar x) ^4 w_i \over \sum w_i } @f$
    ///  @f$ \sigma ^2 = { \sum (x_i - \bar x) ^2 w_i \over \sum w_i } @f$
    double Kurtosis() const;

    ///  Returns the maximum of all the $x_i$ values entered so far.
    ///  @f$ min\{ X_i \} @f$
    double Min() const { return m_dMin; }
    ///  Returns the maximum of all the $x_i$ values entered so far.
    ///  @f$ max\{ X_i \} @f$
    double Max() const { return m_dMax; }

private:
    /// Initialize the object on construction.  (DM643 does not support chained
    /// constructors.)
    void Init();

    //  Objects of this type store the running sums of powers of the
    //  items elements that are added to them. Then when the caller asks for
    //  a specific statistic, it can easily be calculated from the sums.
    //
    //  Each object keeps track of the following.
    //
    //      m\_dSX0 = \sum w_i
    //      m\_dSX1 = \sum x_i w_i
    //      m\_dSX2 = \sum x_i^2 w_i
    //      m\_dSX3 = \sum x_i^3 w_i
    //      m\_dSX4 = \sum x_i^4 w_i
    //      m\_dMin = min\{ x_i \}
    //      m\_dMax = max\{ x_i \}
    //

    double m_dSX0;
    double m_dSX1;
    double m_dSX2;
    double m_dSX3;
    double m_dSX4;

    double m_dMin;
    double m_dMax;
};

#endif  // NUMDIST_HPP
