#ifndef NUMFCN_HPP
#define NUMFCN_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
// ============================================================================

/// This class serves as a base interface class for 'long' functions. I.e.
/// functions that take an extended period of time to execute. Function
/// objects of this type can be Canceled in the middle of execution.
///
/// There is no support for  background execution with this.  This is for
/// cooperative "multitasking".  The functions of this class break up their
/// execution into pieces.  That way other processing can be done between the
/// pieces.
///
/// The standard usage of a CNumLongFunction is to create it, then call
/// IsDone() repeatedly until that returns true.  The Reset() start the
/// function execution from the beginning.  The Run() just does all the steps
/// to completion. And the Cancel() stop the execution in the middle.
///
/// Derived classes must override the NextStep() method to implement the
/// calculation.  They can override the InitExec() and StopExec() to setup the
/// calculation, and cleanup after the calculation is done.
class CNumLongFunction
{
public:
    CNumLongFunction() : m_eState(ceWaiting) {}

    virtual ~CNumLongFunction(){};

    ///  Cancel the execution of the function. This should only be called in
    ///  the ceRunning State.
    void Cancel();

    ///  Reset the function back to the initial waiting state. This should
    ///  only be done in the ceDone State.
    void Reset();

    ///  This takes another step in the execution of the function. It
    ///  returns true iff there is no more work to do (i.e. the function
    ///  is in the done state.
    bool IsDone();

    ///  This will run the function untill it is done executing.
    void Run();

protected:
    ///  This is called before the first NextStep() call to initialize the
    ///  function execution.  Override to setup the initial state for the
    ///  calculation.
    virtual void InitExec(){};

    ///  This is called whenever the execution is funnished, either as a
    ///  result of a cancel call, or as a result of the last NextStep()
    ///  indicating it is finished.  Override to cleanup if necessary.
    virtual void StopExec(){};

    ///  This should be overwritten to do an item of work. The function
    ///  is called repeatedly until there is no more work to do. The
    ///  function sould return true iff there is still more work to do.
    ///  Each call should not take very long since the function can only
    ///  be canceled between calls, also on non-multi-tasking systems
    ///  each step blocks the remainder of the processing.
    virtual bool NextStep() = 0;

private:
    enum
    {
        ceWaiting,  //  Ready to start
        ceRunning,  //  In the middle of running.
        ceDone      //  Completed the execution.

    } m_eState;
};

#endif  // NUMFCN_HPP
