#ifndef NUMHIST_HPP
#define NUMHIST_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "numvect.hpp"
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------

// ============================================================================
/// A class that supports making histograms from data.  These can be
/// linearly spaced or logarithmic.  The istogram keeps track of the
/// frequencies in each bin, and also outside the range of the bins.
///
/// This histogram is setup with the bins in the constructor.  Then data is
/// added with one od the Add() methods.  The data (or data and parameters )
/// can be Reset().
///
/// The bin information can be accessed with the NumBins(), MinRange(),
/// MaxRange(), IsLinear(), IsLogarithmic(), Midpoint() and EndPoint() methods.
///
/// Access to the collected frequency data is through the Frequency(),
/// Frequencies(), LowFrequency(), HighFrequency(), OutOfRangeFrequency(),
/// InRangeFrequency(), and TotalFrequency() methods.
class CNumHistogram
{
public:
    enum EChannelDist
    {
        ceLinear,
        ceLogarithmic
    };

    CNumHistogram();
    CNumHistogram(size_t uiBins, double dMin, double dMax, EChannelDist eDist = ceLinear);

    ///  Reset the frequencty data, but do no effect the binning.
    void Reset();

    ///  Reset the frequency data, and setup new bins.
    void Reset(size_t uiBins, double dMin, double dMax, EChannelDist eDist = ceLinear);

    /// @{
    //  Add frequency data at a specific location.  By default a single count
    //  is added.
    void Add(double dValue, double dFrequency = 1.0) { Bin(dValue) += dFrequency; }
    void Add(const CNumVector<double>& vdValues);
    void Add(const CNumVector<double>& vdValues, const CNumVector<double>& vdFrequencues);
    /// @}

    /// @{
    ///  Give access to the bin parameters of the histogram.
    size_t NumBins() const { return m_vdFreq.Dim(); }
    double MinRange() const { return m_dMin; }
    double MaxRange() const { return m_dMax; }
    bool IsLinear() const { return m_eDistType == ceLinear; }
    bool IsLogarithmic() const { return m_eDistType == ceLogarithmic; }

    ///  @return the midpoint of bin @p i.
    double Midpoint(size_t i) const;
    ///  @return a vector of the midpoints of all the bins.
    CNumVector<double> Midpoints() const;
    ///  @return the beginning boundary of bin @p i, or the ending boundary of
    ///  bin @p i-1.
    double Endpoint(size_t i) const;
    /// @return a vector of the bin boundaries.
    CNumVector<double> Endpoints() const;

    /// @return the frequency collected in bin @p i.
    double Frequency(size_t i) const;
    /// @return a vector of the frequencies in all the bins.
    const CNumVector<double>& Frequencies() const { return m_vdFreq; }

    /// @return the frequency below the lowest bin.
    double LowFrequency() const { return m_dLowFreq; }
    /// @return the frequency above the highest bin.
    double HighFrequency() const { return m_dHighFreq; }
    /// @return the frequency outside the range of the bins.
    double OutOfRangeFrequency() const { return LowFrequency() + HighFrequency(); }
    /// @return the frequency within the range of the bins.
    double InRangeFrequency() const;
    /// @return the total frequency added.
    double TotalFrequency() const { return OutOfRangeFrequency() + InRangeFrequency(); }

private:
    double& Bin(double dValue);
    //  Find the double that should accumulate the value.

    EChannelDist m_eDistType;
    double m_dMin;
    double m_dMax;

    double m_dHighFreq;
    double m_dLowFreq;
    CNumVector<double> m_vdFreq;
    //  All the frequencies added to this so far.

    double m_dLinMin;
    double m_dLinMax;
    //  cache of the calculated values to make the applications faster.
};

#endif  // NUMHIST_HPP
