#ifndef NUMINTRP_HPP
#define NUMINTRP_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#ifndef NUMVECT_HPP
#include "numvect.hpp"
#endif
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
///
/// CNumInterpolate
///
/// This class provides the functionality of a interpolation lookup table.
/// The table is set as an array of X/Y coordinates. These are sorted so
/// that the X=coordinates are in strictly ascending order. (this is done
/// with an O(n^2) insertion sort so if this takes too long, it can be
/// improved). If the same x-coordinate exists in multiple points, those
/// are merged by averaging the Y-coordinates.
///
/// So far this handles linear interpolation, and Akima Semi-Spline
/// interpolation. Other interpolation schemes (cubic spline or tension
/// spline) can be added to the framework when necessary.
///
/// This can also report the slope (Slope()) of the interpolation curve
/// at any point. Also, This can find the local extremes (first derivative
/// zero).
///
/// The Nothing is done with the second derivative. This is because the
/// Akima spline does not have a smooth second derivative (it is continuous
/// however). In order to get a second derivative it is best to make a new
/// CNumInterpolate object, with the slope values of this one, and then
/// use the first derivative with that one.
///
/// Some interpolation schemes require an initial set of parameters to be
/// pre-calculated. This class will do that on the first interpolation only
/// subsequent interpolation calls will use the cached values. If the class
/// is modified, the next interpolation call will recalculate the parameters.
/// What all this means is that the caller of the class does not need to
/// worry about doing this.
///
/// This does not handle Multiple cycles of increasing and decreasing
/// x-coordinates. These must be handled parametrically, i.e., have two
/// interpolation tables one for X and one for Y each based on a strictly
/// increasing parameter. The parameter can be the sums of the absolute
/// values of the x-steps in the data.
class CNumInterpolate
{
public:
    enum EType
    {
        ceLinear,
        ceAkima,

        ceNumType
    };

    /// Construct an empty interpolation object.
    CNumInterpolate(EType eType = ceLinear);

    /// Construct and initialize an interpolation object.
    CNumInterpolate(const CNumVector<double>& rvdX,
        const CNumVector<double>& rvdY,
        EType eType = ceLinear);

    /// Construct and initialize an interpolation object.
    CNumInterpolate(size_t iDim, const double* pdX, const double* pdY, EType eType = ceLinear);

    bool operator==(const CNumInterpolate& rhs) const;
    bool operator!=(const CNumInterpolate& rhs) const;

    /// Convenience method.
    double operator()(double x) const { return Interpolate(x); }
    /// Convenience method.
    CNumVector<double> operator()(const CNumVector<double>& rvdX) const;

    /// Set/reset the lookup table using a copy of the data provided.
    /// The points are sorted into increasing x-coordinates.
    void Set(const CNumVector<double>& rvdX, const CNumVector<double>& rvdY);
#if !defined(MIC_DM643)
    void Set(CNumVector<double>&& rvdX, CNumVector<double>&& rvdY);
#endif

    /// Set/reset the lookup table using a copy of the data provided.
    /// The points are sorted into increasing x-coordinates.
    void Set(size_t iDim, const double* pdX, const double* pdY);

    /// Set/reset the lookup table using a copy of the data provided. The
    /// x-coordinates are assumed to be in ascending order. Any points not in
    /// ascending order are ignored.
    void SetAscending(const CNumVector<double>& rvdX, const CNumVector<double>& rvdY);

    /// Set/reset the lookup table using a copy of the data provided. The
    /// x-coordinates are assumed to be in ascending order. Any points not in
    /// ascending order are ignored.
    void SetAscending(size_t iDim, const double* pdX, const double* pdY);

    /// Set/reset the lookup table using a copy of the data provided. The
    /// x-coordinates are assumed to be in descending order. Any points not in
    /// descending order are ignored.
    void SetDescending(const CNumVector<double>& rvdX, const CNumVector<double>& rvdY);

    /// Set/reset the lookup table using a copy of the data provided. The
    /// x-coordinates are assumed to be in descending order. Any points not in
    /// descending order are ignored.
    void SetDescending(size_t iDim, const double* pdX, const double* pdY);

    /// Set the interpolation type.
    void Set(EType eType);

    size_t Dim() const { return m_vdX.Dim(); }
    const double& X(size_t i) const { return m_vdX[i]; }
    const double& Y(size_t i) const { return m_vdY[i]; }

    /// Interpolate a y-coordinate given an x-coordinate.
    double Interpolate(double x) const;
    /// Interpolate a vector of y-coordinates given a vector of x-coordinates.
    CNumVector<double> Interpolate(const CNumVector<double>& rvdX) const;

    /// Calculate the slope at the given interpolation point.
    double Slope(double x) const;

    /// This finds and returns the smallest x-coordinate greater than the
    /// passed in one where the derivative is zero. If there are no more
    /// extrema the largest X-coordinate is returned.
    double NextExtreme(double x) const;

#if !defined(MIC_DM643)
    friend std::ostream& operator<<(std::ostream& os, const CNumInterpolate& rAxis);
    friend std::istream& operator>>( std::istream& is, CNumInterpolate& rAxis );
#endif

private:
    /// Release the used by the various interpolation schemes and clear any
    /// related is-valid flags.
    void ClearExtra();

    //
    //  Support for ceLinear interpolation.
    //
    /// Calculate the linear interpolation and return the proper value.
    double LinearInterp(double x) const;

    /// Calculate the slope at the given interpolation point using linear interpolation.
    double LinearSlope(double x) const;

    /// Find and return the index i such that X[i] <= x < X[i].
    size_t LinearFind(double x) const;

    //
    //  Support for Akima interpolation.
    //
    /// Initialize the Akima spline coefficients cache used with the Akima interpolation.
    void AkimaSetup() const;
    /// Akima spline version of LinearInterp().
    double AkimaInterp(double x) const;
    /// Akima spline version of LinearSlope().
    double AkimaSlope(double x) const;
    /// Return the smallest x-coordinate greater than the given point for which
    /// the derivative is zero. If there are no more extrema the largest X-coordinate
    /// in the interpolation vector is returned.
    double AkimaNextExtreme(double x) const;

    EType m_eType;             ///< Interpolation type.
    CNumVector<double> m_vdX;  ///< Interpolation x-coordinates.
    CNumVector<double> m_vdY;  ///< Interpolation y-coordinates.

    //
    // Mutable members used for caching.
    //
    /// Stores the last x-coordinate vector index used; since interpolation is
    /// generally done on sequential x-coordinates, the range lookup can be
    /// optimized by keeping track of the last range used.
    mutable size_t m_uiLast;
    ///< True iff the Akima spline coefficient cache has been initialized.
    mutable bool m_fExtraValid;
    ///< Akima spline coefficients cache; valid only if type is ceAkima and
    ///< m_fExtraValid is true.
    mutable CNumVector<double> m_vdAkimaM;
    ///< Akima spline coefficients cache; valid only if type is ceAkima and
    ///< m_fExtraValid is true.
    mutable CNumVector<double> m_vdAkimaZ;
};

// ================================================================================================
/// Template version of CNumInterpolate that support quantities with units for the x, y table.
#if !defined(MIC_DM643)
template<typename T1, typename T2>
class TNumInterpolate
{
public:
    TNumInterpolate(const TNumInterpolate<T1, T2>&) = default;
    TNumInterpolate(const std::vector<T1>& rvdX,
        const std::vector<T2>& rvdY,
        CNumInterpolate::EType eType = CNumInterpolate::ceLinear);
    TNumInterpolate(const std::vector<T1>& rvdX,
        const std::vector<T2>& rvdY,
        bool fIsAscending);
    TNumInterpolate(size_t uiNum, const T1* pdX, const T2* pdY, CNumInterpolate::EType eType = CNumInterpolate::ceLinear);

    /// @{ Set the data to be used.
    void Set(size_t uiNum, const T1* pdX, const T2* pdY);
    void Set(const std::vector<T1>& rvdX, const std::vector<T2>& rvdY);
    void SetDescending(size_t uiNum, const T1* pdX, const T2* pdY);
    void SetDescending(const std::vector<T1>& rvdX, const std::vector<T2>& rvdY);
    /// @}

    /// @{ Access to the table used for interpolation.
    size_t Dim() const;
    T1 X(size_t i) const;
    T2 Y(size_t i) const;
    /// @}

    /// @{ Interpolate in unit aware function.
    T2 Interpolate(T1 x) const;
    double operator()(double x) const { return Interpolate(x); }
    /// @}

private:

    /// The non unit aware interpolation object that does the work.
    CNumInterpolate m_interp;
};
#endif

//-----------------------------------------------------------------------------
/// Return the value of the line through (x1, y1) and (x2, y2) at x
inline double TwoPointInterpolate(double x1, double y1, double x2, double y2, double x)
{
    ASSERT_ERROR(x1 != x2);
    if(x1 == x2)
        return y1;
    return (x - x2) * (y2 - y1) / (x2 - x1) + y2;
}

// ================================================================================================
#if !defined(MIC_DM643)
template<typename T1, typename T2>
TNumInterpolate<T1, T2>::TNumInterpolate(const std::vector<T1>& rvdX,
    const std::vector<T2>& rvdY,
    CNumInterpolate::EType eType)
    : m_interp(eType)
{
    Set(rvdX, rvdY);
}

template<typename T1, typename T2>
TNumInterpolate<T1, T2>::TNumInterpolate(const std::vector<T1>& rvdX,
    const std::vector<T2>& rvdY,
    bool fIsAscending)
    : m_interp(CNumInterpolate::ceLinear)
{
    if(fIsAscending)
        Set(rvdX, rvdY);
    else
        SetDescending(rvdX, rvdY);
}

template<typename T1, typename T2>
TNumInterpolate<T1, T2>::TNumInterpolate(size_t uiNum, const T1* pdX, const T2* pdY, CNumInterpolate::EType eType)
    : m_interp(eType)
{
    Set(uiNum, pdX, pdY);
}

template<typename T1, typename T2>
void TNumInterpolate<T1, T2>::Set(const std::vector<T1>& rvdX, const std::vector<T2>& rvdY)
{
    ASSERT_ERROR(rvdX.size() == rvdY.size());
    Set(std::min(rvdX.size(), rvdY.size()), rvdX.data(), rvdY.data());
}

template<typename T1, typename T2>
void TNumInterpolate<T1, T2>::Set(size_t uiNum, const T1* pdX, const T2* pdY)
{
    auto vdx = CNumVector<double>(uiNum);
    auto vdy = CNumVector<double>(uiNum);
    for(auto i = 0u; i < uiNum; ++i)
    {
        vdx[i] = pdX[i];
        vdy[i] = pdY[i];
    }
    m_interp.Set(std::move(vdx), std::move(vdy));
}

template<typename T1, typename T2>
void TNumInterpolate<T1, T2>::SetDescending(const std::vector<T1>& rvdX, const std::vector<T2>& rvdY)
{
    ASSERT_ERROR(rvdX.size() == rvdY.size());
    SetDescending(std::min(rvdX.size(), rvdY.size()), rvdX.data(), rvdY.data());
}

template<typename T1, typename T2>
void TNumInterpolate<T1, T2>::SetDescending(size_t uiNum, const T1* pdX, const T2* pdY)
{
    auto vdx = CNumVector<double>(uiNum);
    auto vdy = CNumVector<double>(uiNum);
    for(auto i = 0u; i < uiNum; ++i)
    {
        vdx[uiNum-i-1] = pdX[i];
        vdy[uiNum-i-1] = pdY[i];
    }
    m_interp.Set(std::move(vdx), std::move(vdy));
}

template<typename T1, typename T2>
size_t TNumInterpolate<T1, T2>::Dim() const
{
    return m_interp.Dim();
}

template<typename T1, typename T2>
T1 TNumInterpolate<T1, T2>::X(size_t i) const
{
    return T1(m_interp.X(i));
}

template<typename T1, typename T2>
T2 TNumInterpolate<T1, T2>::Y(size_t i) const
{
    return T2(m_interp.Y(i));
}

template<typename T1, typename T2>

T2 TNumInterpolate<T1, T2>::Interpolate(T1 x) const
{
    return T2(m_interp(x));
}
#endif


#endif // NUMINTRP_HPP
