#ifndef NUMLMFIT_HPP
#define NUMLMFIT_HPP
//-----------------------------------------------------------------------------
/// @file
/// Gauss-Jordan elimination and Levenberg-Marquardt method of chi-squared
/// minimization (least-squares fit). Fits the sum of a series of parameterized
/// curves to a data set.
//-----------------------------------------------------------------------------
#include "nummat.hpp"
//-----------------------------------------------------------------------------
// Forward references.
//-----------------------------------------------------------------------------
class CNumParamCurve;

//-----------------------------------------------------------------------------
/// Gauss-Jordan elimination from Numerical Recipes in C, The Art of Scientific
/// Computing, Second Edition, by W.H.Press, S.A.Teukolsky, W.T.Vetterling,
/// B.P.Flannery Copyright (c) Cambridge University Press 1988, 1992 software
/// is Copyright (c) Numerical Recipes Software 1987, 1988, 1992 Section 2.1
/// page 39.
class CNumGaussJordan
{
public:
    /// a[0..n-1][0..n-1] is the input matrix. b[0..n-1][0..m-1] is input
    /// containing the m right-hand side vectors. On output, a is replaced
    /// by its matrix inverse, and b is replaced by the corresponding set of
    /// solution vectors.
    /// @return 1 for singular matrix at first check, 2 for singular matrix at
    ///         second check, else 0 for OK
    static int gaussj( CNumMatrix<double>& a, int n, CNumMatrix<double>& b, int m );
    DECLARE_MSG()
};

//-----------------------------------------------------------------------------
/// Levenberg-Marquardt method of chi-squared minimization (least-squares fit)
/// from Numerical Recipes in C, The Art of Scientific Computing, Second
/// Edition, by W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery
/// Copyright (c) Cambridge University Press 1988, 1992 software is Copyright
/// (c) Numerical Recipes Software 1987, 1988, 1992 Section 15.5 page 685, and
/// Gauss-Jordan elimination page 39.
class CNumLevMrqFit
{
public:
    CNumLevMrqFit() {}

    /// Levenberg-Marquardt method, attempting to reduce the value chi squared
    /// of a fit between a set of data points x[0..ndata-1], y[0..ndata-1] with
    /// individual standard deviations sig[0..ndata-1], and a nonlinear function
    /// dependent on ma coefficients a[0..ma-1]. The input array ia[0..ma-1]
    /// indicates by nonzero entries those components of a that should be fitted
    /// for, and by zero entries those components that should be held fixed at
    /// their input values. The program returns current best-fit values for the
    /// parameters a[0..ma-1], and chi squared = chisq. The arrays
    /// covar[0..ma-1][0..ma-1], alpha[0..ma-1][0..ma-1] are used as working
    /// space during most iterations. Supply a class pcCurve that evaluates the
    /// fitting function yfit, and its derivatives dyda[0..ma-1] with respect to
    /// the fitting parameters a at x. On the first call provide an initial guess
    /// for the parameters a, and set alamda<0 for initialization (which then sets
    /// alamda=.001). If a step succeeds chisq becomes smaller and alamda decreases
    /// by a factor of 10. If a step fails alamda grows by a factor of 10. You must
    /// call this routine repeatedly until convergence is achieved. Then, make one
    /// final call with alamda=0, so that covar[0..ma-1][0..ma-1] returns the
    /// covariance matrix, and alpha the curvature matrix. (Parameters held fixed
    /// will return zero covariances.)
    /// @param x Independent data set.
    /// @param y Dependent data set.
    /// @param sig Individual standard deviations.
    /// @param ndata Length of x, y and sig datasets.
    /// @param a Output best-fit values for parameters.
    /// @param ia The input array ia[0..ma-1] indicates by nonzero entries those
    ///           components of a that should be fitted for, and by zero entries
    ///           those components that should be held fixed at their input values.
    /// @param ma Length of a and ia, rank of covar and alpha.
    /// @return 1 or 2 if Gauss-Jordan elimination finds a singular matrix,
    ///         otherwise 0
    int mrqmin(const double* const x,
        const double* const y,
        const double* const sig,
        int ndata,
        double* a,
        const int* const ia,
        int ma,
        CNumMatrix<double>& covar,
        CNumMatrix<double>& alpha,
        double* chisq,
        const CNumParamCurve* const pcCurve,
        int nCurves,
        double* alamda);

    /// Sort the covariance matrix into the order of all ma parameters
    /// Expand in storage the covariance matrix covar, so as to take into account
    /// parameters that are being held fixed. (For the latter, return zero
    /// covariances)
    void covsrt(CNumMatrix<double>& covar, int ma, const int* const ia, int mfit);

    /// Evaluate fitting matrix alpha and vector beta, calculate chi squared
    void mrqcof(const double* const x,
        const double* const y,
        const double* const sig,
        int ndata,
        const double* const a,
        const int* const ia,
        int ma,
        CNumMatrix<double>& alpha,
        double* beta,
        double* chisq,
        const CNumParamCurve* const pcCurve,
        int nCurves);

private:
    int m_mfit;
    double m_ochisq;
    double* m_atry;
    double* m_beta;
    double* m_da;
    CNumMatrix<double> m_oneda;

    DECLARE_MSG()
};

#endif // NUMLMFIT_HPP
