#ifndef NUMLNFIT_HPP
#define NUMLNFIT_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <math.h>
// ----------------------------------------------------------------------------
///  Objects of this class are used to fit data to a straight line. The line
///  can be one of two types, either a general line fit (y = slope * x
///  + y-intercept) (ceGeneral) or forcing it to go through to go through the
///  origen ( y = slope * x ) (ceForceZero). The style of line can be choosen
///  at creation time, or at any time thereafter using the SetType() function.
///  The objects can be created with no points, or with vectors giving the
///  initial set of points (and possibly variances). Once created, more points
///  can be added simply by using the AddPoint() function.
///
///  The line of best fit is calculated assuming that the x-coordinates of the
///  points are exact, and the error is in the y-coordinate. This results in
///  the line that minimizes the sum of the squares of vertical distance from
///  the line to the datapoints ( as opposed to the general, or horizontal
///  distance).
///
///  The points can be entered in one of two ways (which probably should not
///  be mixed if any useful result is to result). The better way is to enter
///  the x- and y- coordinates along with the variance in the y-coordinate.
///  This will give the object the ability to give uncertaintes to the
///  calculated values, as well as an overall goodness of fit estimate. If
///  the variance information is not available, and it can be assumed that
///  all the datapoints have the same variance, it is possible to add the
///  points by just their x- and y- coordinates. This assumes that all the
///  y-coordinates have unit variance.
///
///  Once all the datapoints have been added, the characteristics of the line
///  can be calculated. Slope(), YIntercept() are the fundamental quantities.
///
///  If the data was entered with variance information, the VarSlope(),
///  VarYIntercept, CoVarSlopeYIntercept() give the variance, and covariance of
///  those items. The GoodnessOfFit() gives an indication of how likely the
///  data came from a line.
///
///  If the data was entered with no variance information, the RMSError() gives
///  the average error in the y-coordinates, and Correlation() gives the
///  correlation coefficient. If it is possible to assume constant variance
///  for all the datapoints, and assuming that the data does fit a straight line
///  the CalculateVariance() function can be called to add variance information
///  to the data.
///
///  The basic algorithms used in this object are somewat suseptable to roundoff
///  error, especialy when the magnitudes of the numbers are much larger than
///  their variation, or if there are so many datapoints that accuracy is lost
///  when adding the next point to the running totals. It is possible to
///  help the first problem by shifting the x-coordinates so that they are
///  centered near the origen. To help the second problem, create objects of
///  CNumLineFit() for subsets of the data of smaller size, and then combine
///  those using AddPoint( const CNumLineFit& ) into the full set. All
///  calculations are done internaly using double percision.
class CNumLineFit
{
public:
    enum ELineType
    {
        ceGeneral,
        ceForceZero,

        ceNumber  //  Last serves as cardinality.
    };

    ///  Create with no initial values.
    CNumLineFit(ELineType eLineType = ceGeneral);

    /// @{
    /// Creates an best fit object, and adds all the points X[i], Y[i] to it if
    /// the variances are given those are used as well.
    CNumLineFit(size_t N, const double cadX[], const double cadY[], ELineType eLineType = ceGeneral);
    CNumLineFit(size_t N,
        const double cadX[],
        const double cadY[],
        const double cadVar[],
        ELineType eLineType = ceGeneral);
    CNumLineFit(size_t N, const float caflX[], const float caflY[], ELineType eLineType = ceGeneral);
    CNumLineFit(size_t N,
        const float caflX[],
        const float caflY[],
        const float caflVar[],
        ELineType eLineType = ceGeneral);
    /// @}

    bool operator==(const CNumLineFit& rhs) const;
    bool operator!=(const CNumLineFit& rhs) const;

    /// @{
    /// Creates an best fit object, and adds all the points X[i], Y[i] to it if
    /// the variances are given those are used as well.
    void AddPoint(double dX, double dY);
    void AddPoint(double dX, double dY, double dVar);
    void AddPoint(size_t N, const double cadX[], const double cadY[]);
    void AddPoint(size_t N, const double cadX[], const double cadY[], const double cadVar[]);
    void AddPoint(size_t N, const float caflX[], const float caflY[]);
    void AddPoint(size_t N, const float caflX[], const float caflY[], const float caflVar[]);
    /// @}

    /// @{
    /// Calculate the simple statistics.
    void AddPoint(const CNumLineFit& rPoints);
    double AveX() const;
    double AveY() const;
    double VarX() const;
    double VarY() const;
    double SigmaX() const { return sqrt(VarX()); }
    double SigmaY() const { return sqrt(VarY()); }
    double CovarXY() const;

    /// @{
    /// Calculate linefit statistics.
    double Slope() const;
    double YIntercept() const;
    double XIntercept() const;
    double VarSlope() const;
    double VarYIntercept() const;
    double SigmaSlope() const { return sqrt(VarSlope()); }
    double SigmaYIntercept() const { return sqrt(VarYIntercept()); }
    double CovarSlopeYIntercept() const;
    double RMSError() const;
    double Correlation() const;
    /// @}

    ///  This rescales the variance of each point so that the overall variance
    ///  of the disrcepency from the line are accounted for.
    ///
    ///  Constraint:
    ///      Number of points must be bigger than 2, and the data cannot fit
    ///      the best fit line perfectly.
    ///      If the number of points is exacltly 2, then the special case of no
    ///      variance it returned in the slope and Y-Intercept.
    void CalculateVariance();

    /// @return the number of points added.
    int NumberOfPoints() const { return m_dN; }

private:
    //  All the information is stored as running totals.
    //  m_dN = number of points used.
    //  m_dS = sum 1/var sub i
    //  m_dSX = sum xi / var sub i
    //  m_dSXX = sum xi^2 / var sub i
    //  m_dSY = sum Yi / var sub i
    //  m_dSYY = sum Yi^2 / var sub i
    //  m_dSXY = sum Xi Yi / var sub i
    //
    //  The m_eLineType gives the style of the line that should be fit.

    int m_dN;
    double m_dS;
    double m_dSX, m_dSXX;
    double m_dSY, m_dSYY;
    double m_dSXY;

    ELineType m_eLineType;
};
/* ========================================================================= */

#if !defined(_WIN32) && (defined(MIC_CON) || defined(MIC_TEST))
/// Export the python interface for the linefit object.
void ExportLineFit();
#endif

#endif // NUMLNFIT_HPP
