#ifndef NUMMAT_HPP
#define NUMMAT_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#ifndef NUMVECT_HPP
#include "numvect.hpp"
#endif
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------

// ============================================================================
/// This template class defines a (linear algebra) matrix over the scalers
///  given / as the template paramter TType.
///
///  The class handles memory internaly. When SetDim( uiRows, uiCols ) is
///  called the class does one of the following.
///      A)  uiRows == 0     => all internal memory is free'd.
///      B)  uiRows larger than any previous matrix stored in this class
///          => The old memory is deleted, and a new block is allocated.
///      c)  otherwise => The memory is left alone, and just the dimention is
///          adjusted.
/// This results in the possibility of the following trick. If you know the
/// maximum size that will ever be used by the vector, Allocate it to that
/// size first, and then resize it as needed. This will result in new being
/// called only in the initial SetDim()/constructor.
///
///  TType should be a numeric quantity (ex. uint_8, double) and have all the
///  usual arithmatic operators defined on it (including division).
// ----------------------------------------------------------------------------
template<class TType> class CNumMatrix
{
public:
#if !defined(MIC_DM643)
    typedef TType value_type;
#endif

    ///  Creates this as uninitialized matrix of zero length. No memory
    ///  allocation is done.
    CNumMatrix() { SetDim(0u, 0u, false); }
    CNumMatrix(const CNumMatrix<TType>& src);
    ///  Creates this as a 2-dimentional matrix.
    CNumMatrix(const CNumVector<TType>& x, const CNumVector<TType>& y);
    ///  Creates this as a 3-dimentional matrix.
    CNumMatrix(const CNumVector<TType>& x, const CNumVector<TType>& y, const CNumVector<TType>& z);
    ///  Creates a matrix ofgiven dimention.
    CNumMatrix(size_t iRows, size_t iColumns) { SetDim(iRows, iColumns, false); }
    ///  Frees up any memory that is used by the matrix. This invalidates the result
    ///  of the  Data() call.
    ~CNumMatrix() { SetDim(0, 0); }

    CNumMatrix<TType>& operator=(const CNumMatrix<TType>& rhs);

    size_t Rows() const { return m_iRows; }
    size_t Columns() const { return m_iColumns; }

    CNumVector<TType>& operator[](size_t iRow);
    const CNumVector<TType>& operator[](size_t iRow) const;

    ///  Gives fast access to the array of TTypes that house the elemental
    ///  data.  If the dimention is zero, the return value is undefined.
    CNumVector<TType>* Data() { return m_aRow; }
    const CNumVector<TType>* Data() const { return m_aRow; }

    ///  Give fast (unsafe) access to the elemental data. The normal accessor
    ///  is just as fast in release mode.
    TType& operator()(size_t iRow, size_t iCol);
    const TType& operator()(size_t iRow, size_t iCol) const;

    ///  Reallocates the dimention of the vector to be equal to N. The data content
    ///  is undefined after the call to this method. And the pointer returned by
    ///  Data() is no longer valid.
    void SetDim(size_t iRows, size_t iColumns) { SetDim(iRows, iColumns, true); }

    ///  Set all the elements to 0.
    void Zero();

    int operator==(const CNumMatrix<TType>& rhs) const;
    int operator!=(const CNumMatrix<TType>& rhs) const;

    CNumMatrix<TType>& operator+=(const CNumMatrix<TType>& rhs);
    CNumMatrix<TType>& operator-=(const CNumMatrix<TType>& rhs);
    CNumMatrix<TType>& operator*=(const CNumMatrix<TType>& rhs);
    /// In place matrix multiplication. These must not change the size of this
    /// matrix (i.e. only square matricies are allowed for rhs.
    CNumMatrix<TType>& operator*=(TType rhs);

    CNumMatrix<TType> Transpose() const;

#if !defined(MIC_DM643)
    ///  Read the matrix (including its dimention) from the stream. If this
    ///  fails, the failbit in the stream is set.
    std::istream& Read(std::istream& is);
    ///  Read the contents of the matrix assuming that the dimention is
    ///  correct. If this fails, the failbit in the stream is set.
    std::istream& ReadData(std::istream& is);
    ///  Write the matrix (including its dimention) from the stream. If this fails,
    ///  the failbit in the stream is set.
    std::ostream& Write(std::ostream& os) const;
    ///  Write the contents of the matrix assuming that the dimention is correct. If
    ///  this fails, the failbit in the stream is set.
    std::ostream& WriteData(std::ostream& os) const;
#endif

    ///  return the number of bytes necessary to store this into a buffer using
    ///  PutBuff().
    size_t GetBuffSize() const;

    ///  return the number of bytes necessary to store this into a buffer using
    ///  PutBuffData().
    size_t GetBuffDataSize() const;

    ///  Read the vector (including its dimention) from the data. If this fails,
    ///  the *pfValid is set to false. If this was false on input nothing is done.
    const char* GetBuff(const char* pBuff, bool* pfValid);
    ///  Read the contents of the vector assuming that the dimention is
    ///  correct. If this fails, *pfValid is set to false. If this was false
    ///  on input nothing is done.
    const char* GetBuffData(const char* pBuff, bool* pfValid);
    //  Write the vector (including its dimention) from the stream.
    char* PutBuff(char* pBuff) const;
    ///  Write the contents of the vector assuming that the dimention is
    ///  correct.
    char* PutBuffData(char* pBuff) const;

private:
    ///  Reallocates the dimention of the matrix to be equal to iRow, iCol. The
    ///  data content is undefined after the call to this method. And the pointer
    ///  returned by Data() is no longer valid. If the fFree is false, it is assumed
    ///  that the data of this has never been initialized, and so is noot freed to
    ///  make room for the new one.
    void SetDim(size_t iRows, size_t iColumns, bool fFree);

    size_t m_iAllocRows;        ///< Amount of room in m_aRow.
    size_t m_iRows;             ///< Current number of rows
    size_t m_iColumns;          ///< Current number of columns.
    CNumVector<TType>* m_aRow;  ///< Array of rows.
};
//-----------------------------------------------------------------------------
// NOTE: The template functions are marked inline below so that the DM643
// compiler will link them.  Since inline is only a hint, it should not cause
// problems for the other compilers.
template<class TType> inline CNumMatrix<TType>::CNumMatrix(const CNumMatrix<TType>& src)
{
    SetDim(src.Rows(), src.Columns(), false);
    for(size_t r = 0; r < Rows(); r++)
        m_aRow[r] = src[r];
}

template<class TType>
inline CNumMatrix<TType>::CNumMatrix(const CNumVector<TType>& x, const CNumVector<TType>& y)
{
    ASSERT_ERROR(x.Dim() == y.Dim());
    SetDim(2u, x.Dim(), false);
    m_aRow[0] = x;
    m_aRow[1] = y;
}

template<class TType>
inline CNumMatrix<TType>::CNumMatrix(const CNumVector<TType>& x,
    const CNumVector<TType>& y,
    const CNumVector<TType>& z)
{
    ASSERT_ERROR(x.Dim() == y.Dim());
    SetDim(3u, x.Dim(), false);
    m_aRow[0] = x;
    m_aRow[1] = y;
    m_aRow[2] = z;
}

template<class TType>
inline CNumMatrix<TType>& CNumMatrix<TType>::operator=(const CNumMatrix<TType>& rhs)
{
    SetDim(rhs.Rows(), rhs.Columns(), true);
    for(size_t r = 0; r < Rows(); r++)
        m_aRow[r] = rhs[r];

    return *this;
}

template<class TType> inline CNumVector<TType>& CNumMatrix<TType>::operator[](size_t iRow)
{
    ASSERT_ERROR(iRow < Rows());
    return m_aRow[iRow];
}

template<class TType>
inline const CNumVector<TType>& CNumMatrix<TType>::operator[](size_t iRow) const
{
    ASSERT_ERROR(iRow < Rows());
    return m_aRow[iRow];
}

template<class TType> inline TType& CNumMatrix<TType>::operator()(size_t iRow, size_t iCol)
{
    return m_aRow[iRow].Data()[iCol];
}

template<class TType>
inline const TType& CNumMatrix<TType>::operator()(size_t iRow, size_t iCol) const
{
    return m_aRow[iRow].Data()[iCol];
}

template<class TType> inline void CNumMatrix<TType>::Zero()
{
    for(size_t r = 0; r < Rows(); r++)
        m_aRow[r].Zero();
}

template<class TType> inline int CNumMatrix<TType>::operator==(const CNumMatrix<TType>& rhs) const
{
    if(Rows() != rhs.Rows() || Columns() != rhs.Columns())
        return 0;

    for(size_t r = 0; r < Rows(); r++)
        if(m_aRow[r] != rhs[r])
            return 0;

    return 1;
}

template<class TType> inline int CNumMatrix<TType>::operator!=(const CNumMatrix<TType>& rhs) const
{
    return !(*this == rhs);
}

template<class TType>
inline CNumMatrix<TType>& CNumMatrix<TType>::operator+=(const CNumMatrix<TType>& rhs)
{
    ASSERT_ERROR(Rows() == rhs.Rows() && Columns() == rhs.Columns());
    if(Rows() != rhs.Rows())
        return *this;

    for(size_t r = 0; r < Rows(); r++)
    {
        m_aRow[r] += rhs[r];
    }
    return *this;
}

template<class TType>
inline CNumMatrix<TType>& CNumMatrix<TType>::operator-=(const CNumMatrix<TType>& rhs)
{
    ASSERT_ERROR(Rows() == rhs.Rows() && Columns() == rhs.Columns());
    if(Rows() != rhs.Rows())
        return *this;

    for(size_t r = 0; r < Rows(); r++)
    {
        m_aRow[r] -= rhs[r];
    }
    return *this;
}

template<class TType>
inline CNumMatrix<TType>& CNumMatrix<TType>::operator*=(const CNumMatrix<TType>& rhs)
{
    ASSERT_ERROR(Columns() == rhs.Rows() && Columns() == rhs.Columns());
    if(Columns() != rhs.Rows() || Columns() != rhs.Columns())
        return *this;

    CNumVector<TType> CurRow(Columns());
    for(size_t r = 0; r < Rows(); r++)
    {
        CurRow = m_aRow[r];
        for(size_t c = 0; c < Columns(); c++)
        {
            TType dSum = 0;
            for(size_t i = 0; i < Columns(); i++)
                dSum += (TType)(CurRow[i] * rhs[i][c]);
            m_aRow[r][c] = dSum;
        }
    }
    return *this;
}

template<class TType> inline CNumMatrix<TType>& CNumMatrix<TType>::operator*=(TType rhs)
{
    for(size_t r = 0; r < Rows(); r++)
        m_aRow[r] *= rhs;
    return *this;
}

template<class TType> inline CNumMatrix<TType> CNumMatrix<TType>::Transpose() const
{
    CNumMatrix<TType> Ret(Columns(), Rows());

    for(size_t r = 0; r < Rows(); r++)
        for(size_t c = 0; c < Columns(); c++)
            Ret[c][r] = (*this)[r][c];

    return Ret;
}

#if !defined(MIC_DM643)

template<class TType> std::istream& CNumMatrix<TType>::Read(std::istream& is)
{
    uint_32 iRows = 0;
    uint_32 iColumns = 0;
    is.read((char*)&iRows, sizeof(iRows));
    is.read((char*)&iColumns, sizeof(iColumns));

    //  Make sure that the size is in an acceptable range,
    if(iRows > size_t(-1) / sizeof(m_aRow[0]) || iColumns > size_t(-1) / sizeof(m_aRow[0][0]))
    {
        is.clear(std::ios::failbit);
        return is;
    }

    SetDim(size_t(iRows), size_t(iColumns));
    return ReadData(is);
}

template<class TType> std::istream& CNumMatrix<TType>::ReadData(std::istream& is)
{
    for(size_t r = 0; r < Rows(); r++)
        m_aRow[r].ReadData(is);

    return is;
}

template<class TType> std::ostream& CNumMatrix<TType>::Write(std::ostream& os) const
{
    uint_32 iRows = Rows();
    uint_32 iColumns = Columns();
    os.write((char*)&iRows, sizeof(iRows));
    os.write((char*)&iColumns, sizeof(iColumns));
    return WriteData(os);
}

template<class TType> std::ostream& CNumMatrix<TType>::WriteData(std::ostream& os) const
{
    for(size_t r = 0; r < Rows(); r++)
        m_aRow[r].WriteData(os);
    return os;
}

#endif

template<class TType> inline size_t CNumMatrix<TType>::GetBuffSize() const
{
    return 2 * sizeof(uint_32) + GetBuffDataSize();
}

template<class TType> inline size_t CNumMatrix<TType>::GetBuffDataSize() const
{
    return Rows() * Columns() * sizeof m_aRow[0][0];
}

template<class TType>
inline const char* CNumMatrix<TType>::GetBuff(const char* pBuff, bool* pfValid)
{
    //  Only do anything if still valid.
    if(pfValid != NULL && !*pfValid)
        return pBuff;

    //  Read the dimention and make sure that the size is in an acceptable range.
    uint_32 iRows = *(uint_32*)pBuff;
    pBuff += sizeof(uint_32);
    uint_32 iColumns = *(uint_32*)pBuff;
    pBuff += sizeof(uint_32);
    if(iRows > size_t(-1) / sizeof(m_aRow[0]) || iColumns > size_t(-1) / sizeof(m_aRow[0][0]))
    {
        if(pfValid != NULL)
            *pfValid = false;
        return pBuff;
    }

    SetDim(size_t(iRows), size_t(iColumns));
    return GetBuffData(pBuff, pfValid);
}

template<class TType>
inline const char* CNumMatrix<TType>::GetBuffData(const char* pBuff, bool* pfValid)
{
    for(size_t r = 0; r < Rows(); r++)
        pBuff = m_aRow[r].GetBuffData(pBuff, pfValid);
    return pBuff;
}

template<class TType> inline char* CNumMatrix<TType>::PutBuff(char* pBuff) const
{
    //  Write the data.
    *(uint_32*)pBuff = (uint_32)Rows();
    pBuff += sizeof(uint_32);
    *(uint_32*)pBuff = (uint_32)Columns();
    pBuff += sizeof(uint_32);
    return PutBuffData(pBuff);
}

template<class TType> inline char* CNumMatrix<TType>::PutBuffData(char* pBuff) const
{
    for(size_t r = 0; r < Rows(); r++)
        pBuff = m_aRow[r].PutBuffData(pBuff);
    return pBuff;
}

template<class TType>
inline void CNumMatrix<TType>::SetDim(size_t iRows, size_t iColumns, bool fFree)
{
    if(!fFree)
    {
        m_aRow = NULL;
        m_iAllocRows = 0;
    }

    if(m_iAllocRows < iRows || iRows == 0)
    {
        delete[] m_aRow;
        m_aRow = (iRows != 0u ? new CNumVector<TType>[iRows] : NULL);
        m_iAllocRows = iRows;
    }

    if(m_iRows != iRows || m_iColumns != iColumns)
    {
        for(size_t r = 0; r < iRows; r++)
            m_aRow[r].SetDim(iColumns);
    }

    m_iRows = iRows;
    m_iColumns = iColumns;
}

template<class TType>
inline CNumMatrix<TType> operator+(const CNumMatrix<TType>& lhs, const CNumMatrix<TType>& rhs)
{
    CNumMatrix<TType> Ret(lhs);
    Ret += rhs;
    return Ret;
}

template<class TType>
inline CNumMatrix<TType> operator-(const CNumMatrix<TType>& lhs, const CNumMatrix<TType>& rhs)
{
    CNumMatrix<TType> Ret(lhs);
    Ret -= rhs;
    return Ret;
}

template<class TType>
inline CNumMatrix<TType> operator*(const CNumMatrix<TType>& lhs, const CNumMatrix<TType>& rhs)
{
    CNumMatrix<TType> Ret(lhs.Rows(), rhs.Columns());
    ASSERT_ERROR(lhs.Columns() == rhs.Rows());
    if(lhs.Columns() != rhs.Rows())
        return Ret;

    for(size_t iRow = 0; iRow < Ret.Rows(); iRow++)
        for(size_t iCol = 0; iCol < Ret.Columns(); iCol++)
        {
            TType& dVal = Ret.Data()[iRow].Data()[iCol];
            dVal = 0;
            for(size_t i = 0; i < lhs.Columns(); i++)
                dVal += lhs.Data()[iRow].Data()[i] * rhs.Data()[i].Data()[iCol];
        }

    return Ret;
}

template<class TType> inline CNumMatrix<TType> operator*(TType lhs, const CNumMatrix<TType>& rhs)
{
    CNumMatrix<TType> Ret(rhs.Rows(), rhs.Columns());

    for(size_t iRow = 0; iRow < Ret.Rows(); iRow++)
        for(size_t iCol = 0; iCol < Ret.Columns(); iCol++)
            Ret.Data()[iRow].Data()[iCol] = lhs * rhs.Data()[iRow].Data()[iCol];

    return Ret;
}

template<class TType> inline CNumMatrix<TType> operator*(const CNumMatrix<TType>& lhs, TType rhs)
{
    CNumMatrix<TType> Ret(lhs.Rows(), lhs.Columns());

    for(size_t iRow = 0; iRow < Ret.Rows(); iRow++)
        for(size_t iCol = 0; iCol < Ret.Columns(); iCol++)
            Ret.Data()[iRow].Data()[iCol] = lhs.Data()[iRow].Data()[iCol] * rhs;

    return Ret;
}

template<class TType>
inline CNumVector<TType> operator*(const CNumMatrix<TType>& lhs, const CNumVector<TType>& rhs)
{
    CNumVector<TType> Ret(lhs.Rows());
    ASSERT_ERROR(lhs.Columns() == rhs.Dim());
    if(lhs.Columns() != rhs.Dim())
        return Ret;

    for(size_t iRow = 0; iRow < Ret.Dim(); iRow++)
    {
        TType& dVal = Ret.Data()[iRow];
        dVal = 0;
        for(size_t i = 0; i < lhs.Columns(); i++)
            dVal += lhs.Data()[iRow].Data()[i] * rhs.Data()[i];
    }

    return Ret;
}

template<class TType>
inline CNumVector<TType> operator*(const CNumVector<TType>& lhs, const CNumMatrix<TType>& rhs)
{
    CNumVector<TType> Ret(rhs.Columns());
    ASSERT_ERROR(lhs.Dim() == rhs.Rows());
    if(lhs.Dim() != rhs.Rows())
        return Ret;

    for(size_t iRow = 0; iRow < Ret.Dim(); iRow++)
    {
        TType& dVal = Ret.Data()[iRow];
        dVal = 0;
        for(size_t i = 0; i < rhs.Rows(); i++)
            dVal += lhs.Data()[i] * rhs.Data()[i].Data()[iRow];
    }

    return Ret;
}

/* ========================================================================= */
#endif  // NUMMAT_HPP
