#ifndef NUMNNLS_HPP
#define NUMNNLS_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "numfcn.hpp"
#include "nummat.hpp"

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
// ============================================================================
/// This class implements the CNumLongFunction interface for the problem
/// of non-negative least squares fitting.
///
/// <pre>
/// age (from Windft):
///
/// CNumNNLS*   p = new CNumNNLS( a, b, x );
/// p -> Tau( sc_dTau );
/// p -> MaxIter( size_t NUM_ITER );
/// p -> InitCond( EInitCond eCond );
/// if ( STEP_SIZE == 2 )
/// {
///     p -> Cutoff( bmax * m_dThresh );
///     p -> WMaxType( CNNLS::m_ceMostDecrease );
/// }
/// else if ( STEP_SIZE == 1 )
/// {
///     p -> Cutoff( sc_dXScale * m_dThresh );
///     p -> WMaxType( CNNLS::m_ceLargestStep, NORM );
/// }
/// else
/// {
///     p -> Cutoff( m_dThresh * bmax / sc_dXScale );
///     p -> WMaxType( CNNLS::m_ceSteepestDescent );
/// }
/// </pre>
class CNumNNLS : public CNumLongFunction
{
public:
    enum EWMaxType
    {
        ceSteepestDescent,
        ceLargestStep,
        ceMostDecrease,

        m_ceNumber
    };

    enum EInitCond
    {
        ceNone,
        ceAll
    };
    CNumNNLS();
    CNumNNLS(const CNumMatrix<float_64>& a, const CNumVector<float_64>& b);
    ~CNumNNLS();

    void Set(const CNumMatrix<float_64>& a, const CNumVector<float_64>& b);
    void Tau(float_64 dTau)
    {
        Reset();
        m_dTau = dTau;
    }
    void MaxIter(size_t MaxIter)
    {
        Reset();
        m_uiMaxIter = MaxIter;
    }
    void InitCond(EInitCond eCond)
    {
        Reset();
        m_eInitCond = eCond;
    }
    void Cutoff(double dCutoff)
    {
        Reset();
        m_dCutoff = dCutoff;
    }
    void WMaxType(EWMaxType eType, float_64* NormX = NULL);
    void RepeatCheck(uint_16 uiRepeat);

    const CNumVector<float_64>& X()
    {
        Run();
        return m_x;
    }

protected:
#if defined(MIC_DM643)
    virtual void InitExec();
    virtual void StopExec();
    virtual bool NextStep();
#else
    virtual void InitExec() override;
    virtual void StopExec() override;
    virtual bool NextStep() override;
#endif

private:
    void SetDefault();

    ///  Subroutine for CNumNNLS( ... ),
    ///  Returns t such that w[t] = max{ w[i] i in Zero }, -1 to indicate Zero is
    ///  empty, or all W[j] <= 0 for j in Zero.
    ///
    ///  The algorithm follows Lawson and Hanson, Solving Least Squares Problems,
    ///  (1974), Algorithm NNLS() p. 161.
    ///
    ///  This does steps 2, 3, 4.
    int WMax(const CNumMatrix<float_64>& a,  //  a[0..m-1][0..n-1]
        const CNumVector<float_64>& b,       //  b[0..m-1]
        CNumVector<float_64>& x,             //  x[0..n-1]
        CNumVector<uint_8>& Zero);

    //  Subalgorithm for the least squates fit.
    int RemClass();

    ///  Subalgorithm for the least squates fit.
    ///
    ///  The algorithm follows Lawson and Hanson, Solving Least Squares Problems,
    ///  (1974), Algorithm H1() p. 57.
    void h1(int p,
        int l,
        int m,
        int n,
        CNumMatrix<float_64>& v,
        float_64* h,
        CNumMatrix<float_64>& c,
        int cj,
        int nu,
        int row);

    ///  Subalgorithm for the least squates fit.
    ///
    ///  The algorithm follows Lawson and Hanson, Solving Least Squares Problems,
    ///  (1974), Algorithm H2() p. 57.
    void h2(int p,
        int l,
        int m,
        int n,
        CNumMatrix<float_64>& v,
        float_64* h,
        CNumMatrix<float_64>& c,
        int cj,
        int nu,
        int row);

    ///  The regular least squares solution. Returns the rank of the matrix e.
    ///  ( 0 indicates an error ).
    ///
    ///  The algorithm follows Lawson and Hanson, Solving Least Squares Problems,
    ///  (1974), Algorithm HFTI() p. 81.
    ///
    ///  This is sped up (considerably ?) since the matrix e really only
    ///  has some columns non zero. Thus the Lookup[] is used to convert the
    ///  indicies of the 'real' e matrix to the scrunched version, r.
    int ls(CNumMatrix<float_64>& e,
        const CNumVector<float_64>& b,
        CNumVector<float_64>& z,
        CNumVector<uint_8>& Zero);

    enum
    {
        ceStart,
        ceInitLoop,
        ceLoop,
        ceEnd
    } m_eState;

    //  Algorithm parameters.
    float_64 m_dTau;
    size_t m_uiMaxIter;
    EInitCond m_eInitCond;
    float_64 m_dCutoff;
    EWMaxType m_eWMaxType;
    float_64* m_adNormX;
    uint_16 m_uiRepeatCheck;

    //  Input paremeters.
    const CNumMatrix<float_64>* m_a;
    const CNumVector<float_64>* m_b;

    //  Output paremeters.
    CNumVector<float_64> m_x;

    //  'Local' paremeters.
    CNumMatrix<float_64> Ep;
    CNumVector<float_64> m_z;
    CNumVector<uint_8> m_Zero;
    CNumVector<uint_8> Plus;
    size_t m_uiNumIter;
    CNumMatrix<uint_8> m_muiZeroHist;

    CNumVector<uint_8> Lookup;
    CNumMatrix<float_64> Zint;
    CNumMatrix<float_64> r;
    CNumVector<float_64> m_h;
    CNumMatrix<float_64> m_c;
    CNumVector<float_64> g;
    CNumVector<uint_8> m_p;
};
// ============================================================================
/// This class implements the CNumLongFunction interface for the problem
/// of second derivative regularization non-negative least squares fitting.
class CNumD2RegNNLS
{
public:
    typedef CNumNNLS::EInitCond EInitCond;
    typedef CNumNNLS::EWMaxType EWMaxType;

    CNumD2RegNNLS() {}
    CNumD2RegNNLS(const CNumMatrix<float_64>& a0,
        const CNumVector<float_64>& b,
        float_64 dGamma,
        float_64 dThresh)
    {
        Set(a0, b, dGamma, dThresh);
    }
    ~CNumD2RegNNLS() {}

    void Set(const CNumMatrix<float_64>& a0,
        const CNumVector<float_64>& b,
        float_64 dGamma,
        float_64 dThresh);
    void Tau(float_64 dTau) { m_cNNLS.Tau(dTau); }
    void MaxIter(size_t MaxIter) { m_cNNLS.MaxIter(MaxIter); }
    void InitCond(EInitCond eCond)
    {
        Reset();
        m_cNNLS.InitCond(eCond);
    }
    void WMaxType(EWMaxType eType, float_64* NormX = NULL) { m_cNNLS.WMaxType(eType, NormX); }
    void RepeatCheck(uint_16 uiRepeat) { m_cNNLS.RepeatCheck(uiRepeat); }

    /// Cancel the execution of the function. This should only be called in
    /// the ceRunning State.
    void Cancel() { m_cNNLS.Cancel(); }

    ///  Reset the function back to the initial waiting state. This should
    ///  only be done in the ceDone State.
    void Reset() { m_cNNLS.Reset(); }

    /// This takes another step in the execution of the function. It
    /// returns true iff there is no more work to do (i.e. the function
    /// is in the done state.
    bool IsDone() { return m_cNNLS.IsDone(); }

    ///  This will run the function untill it is done executing.
    void Run() { m_cNNLS.Run(); }

    const CNumVector<float_64>& X() { return m_cNNLS.X(); }

protected:
private:
    CNumNNLS m_cNNLS;

    CNumMatrix<float_64> m_aPrime;
    CNumVector<float_64> m_bPrime;
};
#endif // NUMNNLS_HPP
