#ifndef NUMPLFIT_HPP
#define NUMPLFIT_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "numvect.hpp"
#include "numsvd.hpp"
#include "numpoly.hpp"

// ============================================================================
///  Objects of this class are used to fit data to a polynomial. The
///  polynomial can be one of two types, either a general polynomial fit
///  (y = sum a_i * x^i for i >= 0) (ceGeneral) or forcing it to go
///  through to go through the origen ( y = sum a_i * x^i for i > 0 )
///  (ceForceZero). The stype of the line and the points must be chosen
///  at creation time.
///
///  The polynomial of best fit is calculated assuming that the
///  x-coordinates of the points are exact, and the error is in the
///  y-coordinate. This results in the line that minimizes the sum of
///  the squares of vertical distance from the line to the datapoints (
///  as opposed to the general, or horizontal distance).
class CNumPolynomialFit : public CNumLongFunction
{
public:
    enum EPolynomialType
    {
        ceGeneral,
        ceForceZero,

        ceNumber  //  Last serves as cardinality.
    };

    CNumPolynomialFit(size_t iDegree,
        size_t N,
        const double X[],
        const double Y[],
        EPolynomialType eLineType = ceGeneral);
    CNumPolynomialFit(size_t iDegree,
        size_t N,
        const float X[],
        const float Y[],
        EPolynomialType eLineType = ceGeneral);
    CNumPolynomialFit(const CNumVector<uint_16>& cvuiAllowedExp,
        const CNumVector<float_64>& cvdX,
        const CNumVector<float_64>& cvdY,
        const CNumVector<float_64>& cvdStdDev);

    ///  The coeficient of the x^i term of the polynomial. For the
    ///  ceForceZero type the 0 entry will always be zero.
    double Coefficient(size_t i);

    /// Get the root-mean-squared residual of the fit.
    double GetRMSResidual();

    ///  Get the R^2 value for the fit. That is, the fraction
    ///  of the y variance that is fit by the polynomial.
    double GetRSquared();

    ///  Give access to the fitted polynomial.
    const CNumPolynomial& GetPolynomial();

protected:
#if defined(MIC_DM643)
    virtual void InitExec();
    virtual void StopExec();
    virtual bool NextStep();
#else
    virtual void InitExec() override;
    virtual void StopExec() override;
    virtual bool NextStep() override;
#endif

private:
    ///  Build the matrix a that solved the A Coef = Y.
    void BuildMatrix();

    ///  Find the coeficient of best fit.
    void FitData();

    enum
    {
        ceBuildMatrix,
        ceDoSVD,
        ceFitData,
        ceDone
    } m_eNextState;
    //  Where is it in the execution of the function.

    CNumSVD m_cSVD;
    CNumVector<float_64> m_vdYWeighted;
    //  The auxiliary function todo the SVD fit.

    CNumPolynomial m_cFitPoly;
    //  The results.

    CNumVector<float_64> m_vdX;
    CNumVector<float_64> m_vdY;
    CNumVector<float_64> m_vdStdDev;
    //  The points to fit.

    CNumVector<uint_16> m_vuiAllowedExp;
    //  THe parameters of the fit.

    DECLARE_MSG()
};

#endif  // NUMPLFIT_HPP
