#ifndef NUMPMCRV_HPP
#define NUMPMCRV_HPP
//-----------------------------------------------------------------------------
//// @file
//-----------------------------------------------------------------------------
#if defined(MIC_VCMFC)
#include <iostream>
#include "basetype.hpp"
#endif

// ============================================================================
/// Peak-shaped parameterized curve
///
/// (Peak-shaped) parameterized curve class. Can be any of a family of functions
/// for fitting peaks that are characterized by a similar set of parameters,
/// including height, center, and width, and sometimes one or two shape
/// parameters. Partial derivatives with respect to each parameter are included.
/// Only Gaussian and Log Normal types are currently implemented. Other
/// possibilities include Lorentzian, Logistic, LaPlace, Extreme Value, Error,
/// Student t, Gamma, Weibull, Beta, Inverted Gamma, F-Variance, Chi-Squared
/// Pearson IV.
/// See http://www.systat.com/products/PeakFit/help for the above types.
class CNumParamCurve
{
public:
    enum ECurveType
    {
        ceGaussian = 0,  // symmetric 3-parameter Gaussian (Shape param not used)
        ceLogNormal4,    // asymmetric 4-parameter log normal
        ceLogNormal3,
        ceLorentzian,
        cePearsonVII,
        cePseudoVoigt,
    };

    CNumParamCurve();
    CNumParamCurve( const CNumParamCurve& rhs );
    CNumParamCurve& operator=( const CNumParamCurve& rhs );
    virtual ~CNumParamCurve() {}

#if defined( MIC_VCMFC )
    size_t StreamSize() const;
    std::istream& Read(std::istream& is);
    std::ostream& Write( std::ostream& os ) const;
#endif

    ECurveType Type() const { return m_eType; }
    int NumParams() const;
    float_64 Param(int iPrm) const { return m_adParam[iPrm]; }
    float_64& Param(int iPrm) { return m_adParam[iPrm]; }

    const uint_8& Fixed(int iPrm) const { return m_afFixed[iPrm]; }
    uint_8& Fixed(int iPrm)
    {
        return const_cast<uint_8&>(static_cast<const CNumParamCurve&>(*this).Fixed(iPrm));
    }

    const float_64& Height() const { return m_adParam[0]; }
    float_64& Height()
    {
        return const_cast<float_64&>(static_cast<const CNumParamCurve&>(*this).Height());
    }

    void SetCenter(double dCenterTime) { m_adParam[1] = dCenterTime; }
    float_64 Center() const { return m_adParam[1]; }

    const float_64& Width() const { return m_adParam[2]; }
    float_64& Width()
    {
        return const_cast<float_64&>(static_cast<const CNumParamCurve&>(*this).Width());
    }

    const float_64& Shape() const { return m_adParam[3]; }
    float_64& Shape()
    {
        return const_cast<float_64&>(static_cast<const CNumParamCurve&>(*this).Shape());
    }

    /// Curve y value and (optional) trial parameters.
    /// Optionally returns partial derivatives by parameter for given x
    double Y(double dX, const double* const pA = NULL, double* pD = NULL) const;

    void ChangeCurve(ECurveType eCurveType, double dShape = 0.0);

private:
    ECurveType m_eType;
    float_64 m_adParam[4];
    uint_8 m_afFixed[4];
};

#endif  // NUMPMCRV_HPP
