#ifndef NUMPOLY_HPP
#define NUMPOLY_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#ifndef NUMVECT_HPP
#include "numvect.hpp"
#endif
#include <string.h>
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
// ============================================================================
/// A Polynomial over the doubles.  The exponents could be sparse.  It is also
/// streamable.
class CNumPolynomial
{
public:
    ///  Construct the zero polynomial.
    CNumPolynomial();

    ///  Construct a sparse polynomial with the given coeficients for the given
    ///  exponents.
    CNumPolynomial(const CNumVector<uint_16>& rvuiExp, const CNumVector<double>& rvdCoef);

    ///  Construct a (non sparse polynomial from the coefficient vector.  The
    ///  power is the index.
    CNumPolynomial(const CNumVector<double>& rvdCoef);

    ///  Construct a linear (degree 1) polynomial.
    CNumPolynomial(double dC0, double dC1);

    /// Construct a quadratic (degree 2) polynomial.
    CNumPolynomial(double dC0, double dC1, double dC2);

    ///  @return the degree of the polynomial.
    uint_16 Degree() const;

    /// Evaluate the polynomial using function notation.
    double operator()(double x) const { return Eval(x); }
    /// Evaluate the polynomial
    double Eval(double x) const;

    /// Evaluate the derivative of the polynomial
    double Derivative(double x) const;
    //  Evaluate the (first) derivative at x.

    ///  See if this sparse polynomial has this coefficient. (it could still be
    ///  zero.)
    bool HasCoefficient(uint_16 i) const;

    ///  @return the coeficient of the x^i term.
    double Coefficient(uint_16 i) const;

    /// @name Stream and Buffer I/O.
    /// @{

#if !defined(MIC_DM643)
    ///  Read the polynomial from the stream. If this fails, the failbit in the
    ///  stream is set.
    std::istream& Read(std::istream& is);

    ///  Write the polynomial from the stream. If this fails, the failbit in the
    ///  stream is set.
    std::ostream& Write(std::ostream& os) const;

#endif

    ///  @return the number of bytes necessary to store this into a buffer
    ///  using PutBuff().  (Also the Read/Write size.)
    size_t GetBuffSize() const;

    ///  Read the polynomial from the data. If this fails, the *pfValid is set
    ///  to false. If this was false on input nothing is done. @returns the
    ///  point in the buffer after this data.
    const char* GetBuff(const char* pBuff, bool* pfValid);

    ///  Write the polynomial (including its dimention) from the
    ///  stream. @returns the point in the buffer after this data.
    char* PutBuff(char* pBuff) const;

    /// @}

protected:
private:
    CNumVector<uint_16> m_vuiExp;  ///< Corresponding exponents for m_vdCoef
    CNumVector<double> m_vdCoef;   ///< The coefficients for the exponents.

#if !defined(MIC_DM643)
    friend std::ostream& operator<<(std::ostream& os, const CNumPolynomial& rhs);
    friend std::istream& operator>>(std::istream& is, CNumPolynomial& rhs);
#endif
};

#if !defined(MIC_DM643)

/// Write the CNumPolynomial to a stream as text.
std::ostream& operator<<(std::ostream& os, const CNumPolynomial& rhs);

/// Readthe CNumPolynomial to a stream as text.
std::istream& operator>>(std::istream& is, CNumPolynomial& rhs);

#endif

// ========================================================================
/// Polynomial over a range that optionally does either linear or constant
/// extrapolation on either end of the range. (By default it does not
/// extrapolation.
class CNumPolynomialExtrap
{
#if !defined(MIC_DM643)
    friend std::ostream& operator<<(std::ostream& os, const CNumPolynomialExtrap& rhs);
    friend std::istream& operator>>(std::istream& is, CNumPolynomialExtrap& rhs);
#endif

public:
    /// The type of extrapolation at the end.
    enum EExtrapType
    {
        ceNone,
        ceConstant,
        ceLinear
    };

    ///  Construct the zero polynomial.
    CNumPolynomialExtrap();

    /// Construct an extrapolation polynomialbased on the standard polynomial
    /// (with no endpoints.)
    CNumPolynomialExtrap(const CNumPolynomial& rcPoly);

    /// Construct a full polynomial with endpoints set.
    CNumPolynomialExtrap(const CNumPolynomial& rcPoly,
        double dUpper,
        EExtrapType eExtrapUpper,
        double dLower,
        EExtrapType eExtrapLower);

    ///  Set the underlying polynomial
    void SetPolynomial(const CNumPolynomial& rcPoly);

    ///  Setup how the extrapolation is to be done on upper end.
    void SetUpperExtrap(double dUpper, EExtrapType eExtrap);
    ///  Setup how the extrapolation is to be done on lower end.
    void SetLowerExtrap(double dLower, EExtrapType eExtrap);

    ///  @return the degree of the polynomial.
    uint_16 Degree() const;

    /// Evaluate the polynomial using function notation.
    double operator()(double x) const { return Eval(x); }
    ///  Evaluate the polynomial at passed in value.
    double Eval(double x) const;

    ///  Evaluate the (first) derivative at x.
    double Derivative(double x) const;

    ///  See if this sparse polynomial has this coefficient. (It could still be
    ///  zero.)
    bool HasCoefficient(uint_16 i) const;

    ///  @return the coeficient of the x^i term.
    double Coefficient(uint_16 i) const;

    /// @{
    /// Accessors for the extrapolation points.
    const CNumPolynomial& GetPolynomial() const;
    EExtrapType GetUpperExtrapType() const;
    double GetUpperLimit() const;
    double GetUpperSlope() const;
    double GetUpperOffset() const;
    EExtrapType GetLowerExtrapType() const;
    double GetLowerLimit() const;
    double GetLowerSlope() const;
    double GetLowerOffset() const;
    /// @}

    /// @name Stream and Buffer I/O.
    /// @{

#if !defined(MIC_DM643)
    ///  Read the polynomial from the stream. If this fails, the failbit in the
    ///  stream is set.
    std::istream& Read(std::istream& is);

    ///  Write the polynomial from the stream. If this fails, the failbit in
    ///  the stream is set.
    std::ostream& Write(std::ostream& os) const;

#endif

    ///  @return the number of bytes necessary to store this into a buffer
    ///  using PutBuff().  (Also the Read/Write size.)
    size_t GetBuffSize() const;

    ///  Read the polynomial from the data. If this fails, the *pfValid is set
    ///  to false. If this was false on input nothing is done. @returns the
    ///  point in the buffer after this data.
    const char* GetBuff(const char* pBuff, bool* pfValid);

    ///  Write the polynomial (including its dimention) from the
    ///  stream. @returns the point in the buffer after this data.
    char* PutBuff(char* pBuff) const;

    /// @}

private:
    ///  Updates the interally chached values that speed up evaluation.
    void UpdateCache();

    //  The data of this class.
    CNumPolynomial m_Poly;
    uint_16 m_eLowerExtrap;
    double m_dLower;
    uint_16 m_eUpperExtrap;
    double m_dUpper;

    //  Chached values to make the evaluation fast.
    double m_dLowerSlope;
    double m_dLowerOffset;
    double m_dUpperSlope;
    double m_dUpperOffset;
};

#if !defined(MIC_DM643)

/// Write the CNumPolynomialExtrap to a stream as text.
std::ostream& operator<<(std::ostream& os, const CNumPolynomialExtrap& rhs);

/// Read the CNumPolynomialExtrap to a stream as text.
std::istream& operator>>(std::istream& is, CNumPolynomialExtrap& rhs);

#endif

//-----------------------------------------------------------------------------
#endif // NUMPOLY_HPP
