#ifndef NUMSMTHS_HPP
#define NUMSMTHS_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "numvect.hpp"
#include "numutil.hpp"
#if !defined(MIC_DM643)
#include <vector>
#endif

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class CCircularQueue;

// ============================================================================
/// Class for performing Savitsky-Golay smooth and smoothed first derivative.
/// Performs a least squares fit of a set points to a polynomial.
///
/// The smoothing filter is defined by three quantities, NL, NR, and M. NL Is
/// the number of points in the smoothing array to the left of the point of
/// interest, NR is the number of points to the right. Including the point of
/// interest this requires NL+NR+1 total points for the smoothing to be
/// initialized.  M is the degree of the polynomial to fit.
///
/// For backward compatability, the tree quantities above can be replaced by a
/// single number DIM which is the total number of points (and must be odd) in
/// that case it is assumed that NL = NR = (DIM-1)/2 and M = 2.
///
/// The points are assumed to be evenly spaced on the abscissa.  After
/// initializing with the selected number of points (one at a time or as an
/// array), the smoothed value (0th derivative) and slope (1st derivative) can
/// be obtained.  The 1st derivative is the rate of change per abcissa interval
/// between points.
///
/// Additional points can be added one at a time, each causing the oldest
/// (first) point in the array to be deleted and succeeding points shifted
/// down, providing a sliding window on a long series of points. The smoothed
/// values returned are always for the appropriate point, i.e. for NR intervals
/// before the last point added, where the array width is NL+NR+1.
///
/// Initial implementation gives Smoothed and 1st derivative values for a range
/// of NL, NR, and M. See the s_asCoef array int the cpp file for the complete
/// list.
///
/// The class can be expanded as needed to offer higher order polynomials,
/// higher order derivatives, and greater array widths. See Smoothing and
/// Differentiation of Data by Simplified Least Squares Procedures, Abraham
/// Savitzky and Marcel J. E. Golay, Analytical Chemistry, Vol. 36, No. 8, July
/// 1964 (Note: There are errors in some of the higher order coefficients in
/// the original publication) or Numerical Recipes in C, by Press, Teukolsky,
/// Vetterling, and Flannery, 2nd Edition, Ch. 14.8, which has C code for
/// generating coefficients for various array widths, polynomial orders, and
/// derivative orders.
///
/// The SG.EXE application (built as a part of the 32bit console version of
/// baselib) will give the proper values for the coefficient arrays for any
/// combination of NL, NR, and M. The \p l value for the application is for the
/// oder of the derivative i.e. 0 - smooth 1 - first derivative.
///
/// WARNING: Note that (if using non 'standard') smoothing windows, this class
/// is not safe for concurrent use in multiple threads. The reason for this is
/// that the internal cashe may get updated when a new smoothing window is
/// required. The rules needed to make this safe are as follows. (If these are
/// too restictive, the class can be expanded to make this safer.)
///
///     1. No smoothing window of larger size than any previous one can be
///     requested during a function call in this class.
///     2. No two objects can request new smoothing windows at the same time.
///
/// In a multi threaded environment it is best to make sure all the windows are
/// pre-allocated before multiple threads can create and use objects without
/// care.
class CNumSmoothSG
{
public:
    /// Constructor with array dimension; must be an odd number.
    /// Must be 5,7,9, or 11 if Smooth() is to be used. Array is uninitialized.
    CNumSmoothSG(size_t uiNL, size_t uiNR, size_t uiM);
    CNumSmoothSG(size_t uiDim);

    /// constructor with vector of initial values
    CNumSmoothSG(size_t uiNL, size_t uiNR, size_t uiM, const CNumVector<float_64>& cX);
    CNumSmoothSG(const CNumVector<float_64>& cX);
    CNumSmoothSG(size_t uiNL, size_t uiNR, size_t uiM, const float_64* pdX);
    CNumSmoothSG(size_t uiDim, const float_64* pdX);

    virtual ~CNumSmoothSG();

    // copy constructor, assignment operator: default is OK.

    /// Have enough points been added to initialize the whole array?
    /// Deriv values are invalid unless true
    bool IsInitialized() const;

    /// returns array dimension
    size_t Dim() const;

    /// changes array dimension; values already in array are used to reinit
    /// useful if early results are needed from a short array before enough
    /// data is available for the final array dimension
    void SetDim(size_t uiNL, size_t uiNR, size_t uiM);
    void SetDim(size_t uiDim) { SetDim(uiDim / 2, uiDim / 2, 2); }

    /// add a point to the array, shift out the oldest value
    void AddPoint(float_64 dX);

    /// Shifts the data by an amount indicated.  This is useful when stitching
    /// from one data stream to another that has an offset ion it.
    void ShiftData(float_64 dShift);

    /// smoothed value (0th derivative) at center of array
    /// returns 0 if uninitialized or array dimension is invalid (not 5,7,9,11)
    float_64 Smooth() const;

    /// smoothed slope (1st derivative) at center of array
    /// returns 0 if uninitialized or array dimension is invalid (even)
    float_64 Deriv1() const;

    /// smoothed curvature (2ond derivative) at center of array
    /// returns 0 if uninitialized.
    float_64 Deriv2() const;

    /// return to the uninitialized state.
    void Clear();

    /// return pointer to the array of data
    const float_64* Data() const;

    ///  Returns the array of coeficients.
    static const float_64* FindSmooth(size_t uiNL, size_t uiNR, size_t uiM, size_t uiDeriv);

private:
    void SetCoeff(size_t uiNL, size_t uiNR, size_t uiM, const float_64* pFloats, size_t uiSize);

    CCircularQueue* m_pX;
    const float_64* m_pdSmooth;
    const float_64* m_pdDeriv1;
    const float_64* m_pdDeriv2;
};

// ============================================================================
/// This class will do smoothing on a full dataset. This includes handling
/// the endpoints by smoothing over the same size window. There are two modes
/// of itneracting with this class, one is to smooth a full dataset at once,
/// the other is to iterate over the points as they become available. This
/// iteration is complicated by the fact that while the output points are 1-1
/// with the input points they are not available immediately, and com in
/// spurts at the beginning and the end. pseudo code for the iteration is
/// given here:
///
/// <pre>
///   CNumSmoothSGFull    cSmooth( 30, 30, 2 )
///   ....
///   do
///   {
///       if ( cSmooth.NeedsRaw() && [end of data is reached] )
///           cSmooth.EndOfData();
///       while (cSmooth.NeedsRaw() )
///           cSmooth.AddRaw( [next data] );
///       while ( cSmooth.HasSmooth() )
///           [do something]( cSmooth.GetSmooth() )
///   }
///   while ( !cSmooth.IsIdle() )
/// </pre>
///
///   This reflects the following properites.
///
///  1. The object maintains an internal window of raw (unsmoothed) points of
///     width L + R + 1 from which to calculate smoothed points.
///  2. AddPoint() feeds the next raw point into the window.
///  3. GetPoint() gets the next smoothed point out.
///  4. NeedsRaw() = true indicates that more raw points must be Added (or
///     EndOfData called) before Getting more smoothed points.
///  5. HasSmooth() = true indicates that sufficient raw data has been Added to
///     Get the next smoothed point.
///  6. NeedsRaw() and HasSmooth() are always opposites except before
///     initialization and after Getting the last point.
///
///   WARNING: Note that this class is not safe for concurrent use in multiple
///   threads. The reason for this is that the internal cashe may get updated
///   when a new smoothing window is required. The rules needed to make this safe
///   are as follows. (If these are too restictive, the class can be expanded to
///   make this safer.)
///
///       1. No smoothing window of larger size than any previous one can be
///       requested during a function call in this class.
///       2. No two objects can request new smoothing windows at the same time.
///
///   In a multi threaded environment it is best to make sure all the windows are
///   pre-allocated before multiple threads can create and use objects without
///   care.
///
class CNumSmoothSGFull
{
public:
    ///  Create the object.
    CNumSmoothSGFull(uint_16 uiNL, uint_16 uiNR, uint_16 uiM);

    ///  This will smooth the data in adIn and return the smoothed values in
    ///  adOut.
    void Smooth(size_t uiNum, const float_64 adIn[], float_64 adOut[]);
#if !defined(MIC_DM643)
    ///  @return the smoothed the data from \p cvdIn
    template<typename T>
    std::vector<T> Smooth(const std::vector<T>& cvdIn);
    ///  @return the smoothed the data from \p cvdIn
    template<typename T>
    std::vector<double> Deriv1(const std::vector<T>& cvdIn, double dSpacing);
#endif

    ///  Is the iteration complete.
    bool IsIdle() const { return m_eState == ceIdle ? true : false; }

    ///  Is tehre room to store the next reading.
    bool NeedsRaw() const { return m_fNeedsRaw; }

    ///  Is the next point available.
    bool HasSmooth() const { return m_fHasSmooth; }

    ///  Add the next point in the data stream.
    void AddRaw(float_64 dValue);

    ///  Read the enxt point from the data.
    float_64 GetSmooth();

    ///  Mark the fact that the end of the inoput has been received.
    void EndOfData();

protected:
private:
    uint_16 m_uiNL;
    uint_16 m_uiNR;
    uint_16 m_uiM;
    //  The parameters of the fit.

    CNumVector<float_64> m_cX;
    //  The array of the data points that were added.

    enum
    {
        ceIdle,
        ceInitLoad,  //  The initial phase where a full window is loaded.
        ceInitGet,   //  The initial phase where the skewed windows are
                     //  calculated.
        ceBodyGet,   //  The body of the calculation using the standard
                     //  window ready to return a new point.
        ceBodyPut,   //  The body of the calculation using the standard
                     //  window ready for a new point.
        ceFinalGet,  //  The tail end of the getting using a skewed window.
        ceFewPoints  //  Got to the end of the data before the required
                     //  number of points.
    } m_eState;
    uint_16 m_uiCount;
    uint_16 m_uiCurrent;
    const float_64* m_aCoef;
    //  The state and secondary data for the 'iterator'.

    bool m_fNeedsRaw;
    bool m_fHasSmooth;
    //  Ccalculated value for if there is room for a new point or if a point is
    //  available.
};

// ============================================================================
/// Class for performing Savitsky-Golay smooth and smoothed
/// first derivative, and ensuring the returned smooth value
/// falls within the range of the values in the smoothing window.
///
/// In the case where the oldest value in the smoothing window is
/// much larger magnitude than the rest of the values, a standard
/// Savitsky-Golay quadratic smooth can return a value outside the
/// range of the values in the window, and even a value with the
/// opposite sign. In gas adsorption equilibration, for instance,
/// this can happen with very rapid adsorption at low pressure, as
/// with microporous samples. This class enherits all functionality
/// from CNumSmoothSG, but adds a NNumUtil::ClipToArray() of the value
/// returned by Smooth() ensure the returned value is within the
/// range of the smoothing window. This class should be used for
/// gas adsorption equilibration in preference to CNumSmoothSG for
/// this reason.
class CNumSmoothSGClip : public CNumSmoothSG
{
public:
    /// Constructor with array dimension; must be an odd number.
    /// Must be 5,7,9, or 11 if Smooth() is to be used. Array is uninitialized.
    CNumSmoothSGClip(uint_16 uiNL, uint_16 uiNR, uint_16 uiM) : CNumSmoothSG(uiNL, uiNR, uiM) {}
    CNumSmoothSGClip(uint_16 uiDim) : CNumSmoothSG(uiDim) {}

    /// constructor with vector of initial values
    CNumSmoothSGClip(uint_16 uiNL, uint_16 uiNR, uint_16 uiM, CNumVector<float_64> cX)
        : CNumSmoothSG(uiNL, uiNR, uiM, cX)
    {
    }
    CNumSmoothSGClip(CNumVector<float_64> cX) : CNumSmoothSG(cX) {}
    CNumSmoothSGClip(uint_16 uiNL, uint_16 uiNR, uint_16 uiM, const float_64* pdX)
        : CNumSmoothSG(uiNL, uiNR, uiM, pdX)
    {
    }
    CNumSmoothSGClip(uint_16 uiDim, const float_64* pdX) : CNumSmoothSG(uiDim, pdX) {}

    // copy constructor, assignment operator: default is OK.

    /// smoothed value (0th derivative) at center of array
    /// clipped to the range spanned by the smoothing window values
    float_64 Smooth() const { return NNumUtil::ClipToArray(CNumSmoothSG::Smooth(), Data(), Dim()); }

private:
};


// ================================================================================================
#if !defined(MIC_DM643)
template<typename T>
std::vector<T> CNumSmoothSGFull::Smooth(const std::vector<T>& cvdIn)
{
    auto vdOut = std::vector<T>(cvdIn.size());
    auto uiNum = cvdIn.size();
    auto uiWndSize = m_uiNL + m_uiNR + 1;

    // Make sure there is enough data to do the smoothing.
    if(uiNum < uiWndSize)
    {
        ASSERT_DEBUG(false);
        for(size_t i = 0; i < uiNum; i++)
            vdOut[i] = cvdIn[i];
        return vdOut;
    }

    // Handle the starting Ramp up.
    for(auto iL = 0u; iL < m_uiNL; ++iL)
    {
        auto adCoefL = CNumSmoothSG::FindSmooth(iL, m_uiNL + m_uiNR - iL, m_uiM, 0);
        vdOut[iL] = T(0.0);
        for(size_t iW = 0; iW < uiWndSize; iW++)
            vdOut[iL] += adCoefL[iW] * cvdIn[iW];
    }

    // Handle the body
    auto adCoefM = CNumSmoothSG::FindSmooth(m_uiNL, m_uiNR, m_uiM, 0);
    for(auto iM = m_uiNL; iM < uiNum - m_uiNR; ++iM)
    {
        vdOut[iM] = T(0.0);
        for(size_t iW = 0; iW < uiWndSize; iW++)
            vdOut[iM] += adCoefM[iW] * cvdIn[iM - m_uiNL + iW];
    }

    // Handle the tail
    for(auto iR = 0; iR < m_uiNR; iR++)
    {
        auto adCoefR = CNumSmoothSG::FindSmooth(m_uiNL + m_uiNR - iR, iR, m_uiM, 0);
        vdOut[uiNum - 1 - iR] = T(0.0);
        for(auto iW = 0u; iW < uiWndSize; iW++)
            vdOut[uiNum - 1 - iR] += adCoefR[iW] * cvdIn[uiNum - uiWndSize + iW];
    }
    return vdOut;
}

template<typename T>
std::vector<double> CNumSmoothSGFull::Deriv1(const std::vector<T>& cvdIn, double dSpacing)
{
    auto vdOut = std::vector<double>(cvdIn.size());
    auto uiNum = cvdIn.size();
    auto uiWndSize = m_uiNL + m_uiNR + 1;

    // Make sure there is enough data to do the smoothing.
    if(uiNum < uiWndSize)
    {
        ASSERT_DEBUG(false);
        vdOut.assign(cvdIn.begin(), cvdIn.end());
        return vdOut;
    }

    // Handle the starting Ramp up.
    for(auto iL = 0u; iL < m_uiNL; ++iL)
    {
        auto adCoefL = CNumSmoothSG::FindSmooth(iL, m_uiNL + m_uiNR - iL, m_uiM, 1);
        vdOut[iL] = T(0.0);
        for(size_t iW = 0; iW < uiWndSize; iW++)
            vdOut[iL] += adCoefL[iW] * cvdIn[iW];
    }

    // Handle the body
    auto adCoefM = CNumSmoothSG::FindSmooth(m_uiNL, m_uiNR, m_uiM, 1);
    for(auto iM = m_uiNL; iM < uiNum - m_uiNR; ++iM)
    {
        vdOut[iM] = T(0.0);
        for(size_t iW = 0; iW < uiWndSize; iW++)
            vdOut[iM] += adCoefM[iW] * cvdIn[iM - m_uiNL + iW];
    }

    // Handle the tail
    for(auto iR = 0; iR < m_uiNR; iR++)
    {
        auto adCoefR = CNumSmoothSG::FindSmooth(m_uiNL + m_uiNR - iR, iR, m_uiM, 1);
        vdOut[uiNum - 1 - iR] = T(0.0);
        for(auto iW = 0u; iW < uiWndSize; iW++)
            vdOut[uiNum - 1 - iR] += adCoefR[iW] * cvdIn[uiNum - uiWndSize + iW];
    }

    ASSERT_ERROR(dSpacing != 0.0);
    dSpacing = dSpacing == 0.0 ? 1.0 : dSpacing;
    for(auto i = 0u; i < vdOut.size(); ++i)
        vdOut[i] /= dSpacing;

    return vdOut;
}
#endif

#endif  // NUMSMTHS_HPP
