#ifndef NUMUTIL_HPP
#define NUMUTIL_HPP
//-----------------------------------------------------------------------------
// Library:         BaseLib
// Numeric utilites.
//
//-----------------------------------------------------------------------------

#if !defined(MIC_DM643)
#include <cstdlib>
#include <cmath>
#include <climits>
#include <bitset>
#include <algorithm>
#else
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#endif

// Make sure we're getting the template functions and not the macros.
#if defined(MIC_DM643)
namespace std
{
    template<typename T> inline const T min(const T a, const T b) { return a < b ? a : b; }
    template<typename T> inline const T max(const T a, const T b) { return a > b ? a : b; }
}  // namespace std
#endif

namespace NNumUtil
{
    /// @return the value of @p x clipped to the range [@p low, @p high].
    template<class T> inline T Clip(T x, T low, T high) { return std::max(std::min(x, high), low); }

    /// @return the value of @p x clipped to the range of the elements of array.
    template<class T> inline T ClipToArray(T x, const T* array, size_t iSize)
    {
        if(iSize == 0)
            return x;

        T xMin = array[0];
        T xMax = array[0];
        for(size_t i = 1; i < iSize; i++)
        {
            xMin = std::min(xMin, array[i]);
            xMax = std::max(xMax, array[i]);
        }

        return Clip(x, xMin, xMax);
    }

    /// @return the log base @p dBase of @p dArg
    inline double LogBase(double dBase, double dArg)
    {
        if((dBase <= 1.0) || (dArg <= 0.0))
        {
            ASSERT_ERROR(dBase > 1.0);
            ASSERT_ERROR(dArg > 0.0);
            return 1.0;
        }

        return log(dArg) / log(dBase);
    }

    /// @return true if @p dLHS and @p dRHS are within @p dTolerance
    inline bool Close(double dLHS, double dRHS, double dTolerance)
    {
        return fabs(dLHS - dRHS) <= dTolerance;
    }

#if defined(MIC_VCMFC)
    /// @param dX The value to check.
    /// @param f32BitTolerance True to use the float epsilon (~1e-7) even though
    /// it is checking if a double is near zero. Use this when quantities
    /// between +/- 1e-7 might as well be zero.
    inline bool NearZero(double dX, bool f32BitTolerance)
    {
        return f32BitTolerance ? Close(dX, 0.0, std::numeric_limits<float>::epsilon())
                               : Close(dX, 0.0, std::numeric_limits<double>::epsilon());
    }
#endif

    inline bool Xor(bool a, bool b) { return (!a && b) || (a && !b); }

    inline bool Xand(bool a, bool b) { return (a && b) || (!a && !b); }

    /// @return the natural log but makes sure @p x is clipped to above 1e-9.
    inline double SafeLog(double x) { return x > 0.0 ? log(x) : 1.0e-9; }

    /// @return the base 10 log but makes sure @p x is clipped to above 1e-9.
    inline double SafeLog10(double x) { return x > 0.0 ? log10(x) : 1.0e-9; }

#if defined(MIC_VCMFC)
    inline size_t CountSetBits(uint_32 uiNum)
    {
        return std::bitset<sizeof uint_32 * CHAR_BIT>(uiNum).count();
    }
#endif

}  // namespace NNumUtil

// for dm643 compatibility also define macro
#if defined(MIC_DM643)
#undef min
#undef max
#define min(A, B) ((A > B) ? (B) : (A))
#define max(A, B) ((A > B) ? (A) : (B))
#endif

#endif // not NUMUTIL_HPP
