#ifndef NUMVECT_HPP
#define NUMVECT_HPP
//-----------------------------------------------------------------------------
//
// Library:         BaseLib
//
// Description: A template based numerical vector class.
//
//
// Contents:
//
//  class   CNumVector<T>;
//
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#if !defined(MIC_DM643)
#include "dataarchive.hpp"
#include <iostream>
#include <cmath>
#include <cstring>
#include <vector>
#endif

#if defined(MIC_DM643) || defined(MIC_TEST) || defined(MIC_CON)
#include <string.h>
#endif

#if defined(MIC_VCMFC) || defined(MIC_TEST)
#ifndef METADATA_TRAITS_HPP
#include "metadata-traits.hpp"
#endif
#endif

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
#if defined(MIC_CON) || defined(MIC_TEST)
#pragma warn - inl
#endif
// ----------------------------------------------------------------------------
///  This template class defines a (linear algebra) vector over the scalers given
///  as the template paramter T.
///
///  The class handles memory internaly. When SetDim( uiSize ) is called the
///  class does one of the following.
///      A)  uiSize == 0     => all internal memory is free'd.
///      B)  uiSize larger than any previous vector stored in this class
///          => The old memory is deleted, and a new block is allocated.
///      C)  otherwise => The memory is left alone, and just the dimention is
///          adjusted.
///
///  To resize a vector while preserving the values of the elements,
///  use Resize( uiSize ).  As with SetDim(), reallocation is only done
///  in case B)
///
///  If you know the maximum size that will ever be used by the vector,
///  initialize it to that size first, and then Resize() it as needed.
///  Memory will be allocated only on initialization.
///
///  T should be a numeric quantity (ex. uit_8, float_64) and have all the
///  usual arithmatic operators defined on it (including division).

// NOTE: We cannot change this name to TNumVector which would be more
// appropriate because this does not work with the gencom generation that has
// these for parameters.  In particular 1340embd/oper.gencom fails.
template<class T>
#if defined(MIC_VCMFC) || defined(MIC_TEST)
class CNumVector : public NMeta::IsIterable
#else
class CNumVector
#endif
{
public:
    /// Copy construction.
    CNumVector(const CNumVector<T>& src);
    /// Creates this as a 2-dimentional vector.
    CNumVector(T x, T y);
    /// Creates this as a 3-dimentional vector.
    CNumVector(T x, T y, T z);
    // Create the vector with an array of initial values.
    CNumVector(size_t N, const T* src);
    /// Creates a vector of dimention iDim, with the elements uninitialized.
    CNumVector(size_t N) { SetDim(N, false); }
    /// Creates this as uninitialized vector of zero length. No memory
    /// allocation is done.
    CNumVector() { SetDim(0u, false); }
    /// Frees up any memory that is used by the vector. This invalidates the
    /// result of the Data() call.
    ~CNumVector() { SetDim(0u, true); }

    /// The usual assignment operator.
    CNumVector& operator=(const CNumVector<T>& rhs);

#if !defined(MIC_DM643)
    CNumVector(const std::vector<T>& src) : CNumVector(src.size(), &src[0]) {}

    //-----------------------------------------------------------------------------
    // Support for iterable-like standard container interface.
    using value_type = T;

    size_t size() const { return m_iDim; }
    bool empty() const { return m_iDim == 0; }
    T* begin() { return m_aValue; }
    T* end() { return m_aValue + m_iDim; }
    const T* begin() const { return m_aValue; }
    const T* end() const { return m_aValue + m_iDim; }

    operator std::vector<T>() const { return std::vector<T>(m_aValue, m_aValue + m_iDim); }
#endif  // defined(MIC_VCMFC) || defined(MIC_TEST)

    /// Returns the dimension of this vector.
    size_t Dim() const { return m_iDim; }

    T& operator[](size_t i);
    const T& operator[](size_t i) const;

    ///  Give fast (unsafe) access to the elemental data. The normal accessor
    ///  is just as fast in release mode.
    T& operator()(size_t i) { return m_aValue[i]; }
    const T& operator()(size_t i) const { return m_aValue[i]; }

    ///  Gives fast access to the array of Ts that house the elemental data.
    ///  If the dimention is zero, the return value is undefined.
    T* Data() { return m_aValue; }
    const T* Data() const { return m_aValue; }

    /// Reallocates the dimention of the vector to be equal to N. The data
    /// content is undefined after the call to this method. And the pointer
    /// returned by Data() is no longer valid.
    void SetDim(size_t N) { SetDim(N, true); }

    /// Change the size of the vector to N, reallocating if necessary.
    /// Elements of the original vector that are within the bounds of the new
    /// vector are preserved.
    void Resize(size_t N);

    ///  Set all the elements to 0.
    void Zero();

    bool operator==(const CNumVector<T>& rhs) const;
    bool operator!=(const CNumVector<T>& rhs) const;

    CNumVector<T>& operator+=(const CNumVector<T>& rhs);
    CNumVector<T>& operator-=(const CNumVector<T>& rhs);
    CNumVector<T>& operator*=(T rhs);
    CNumVector<T>& operator/=(T rhs);
    ///  Useful in-place arithmatic function *this = *this + Factor*rhs.
    CNumVector<T>& Add(T Factor, const CNumVector<T>& rhs);

#if !defined(MIC_DM643)
    /// Read the vector (including its dimention) from the stream. If this
    /// fails, the failbit in the stream is set.
    std::istream& Read(std::istream& is);
    /// Read the contents of the vector assuming that the dimention is
    /// correct. If this fails, the failbit in the stream is set.
    std::istream& ReadData(std::istream& is);
    /// Write the vector (including its dimention) from the stream. If this
    /// fails, the failbit in the stream is set.
    std::ostream& Write(std::ostream& os) const;
    /// Write the contents of the vector assuming that the dimention is
    /// correct. If this fails, the failbit in the stream is set.
    std::ostream& WriteData(std::ostream& os) const;

#endif

    ///  return the number of bytes necessary to store this into a buffer
    ///  using PutBuff().
    size_t GetBuffSize() const { return sizeof(uint_32) + GetBuffDataSize(); }
    ///  return the number of bytes necessary to store this into a buffer
    ///  using PutBuffData().
    size_t GetBuffDataSize() const { return Dim() * sizeof m_aValue[0]; }

    /// Read the vector (including its dimention) from the data. If this fails,
    ///  the *pfValid is set to false. If this was false on input nothing is
    ///  done.
    const char* GetBuff(const char* pBuff, bool* pfValid);
    /// Read the contents of the vector assuming that the dimention is
    /// correct. If this fails, *pfValid is set to false. If this was false on
    /// input nothing is done.
    const char* GetBuffData(const char* pBuff, bool* pfValid);
    /// Write the vector (including its dimension) from the stream.
    char* PutBuff(char* pBuff) const;
    /// Write the contents of the vector assuming that the dimension is
    /// correct.
    char* PutBuffData(char* pBuff) const;

private:
    /// Reallocates the dimention of the vector to be equal to N. The data
    /// content is undefined after the call to this method. And the pointer
    /// returned by Data() is no longer valid. If the fFree is false, it is
    /// assumed that the data of this has never been initialized, and so is
    /// noot freed to make room for the new one/
    void SetDim(size_t N, bool fFree);

    size_t m_iDim;    ///< Current dimention of the vector
    size_t m_iAlloc;  ///< Amount of room in m_aValue
    T* m_aValue;      ///< The array ov values.
};
// ----------------------------------------------------------------------------
// NOTE: The template functions are marked inline below so that the DM643
// compiler will link them.  Since inline is only a hint, it should not cause
// problems for the other compilers.
template<typename T> inline CNumVector<T>::CNumVector(const CNumVector<T>& src)
{
    SetDim(src.Dim(), false);
    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] = src.Data()[i];
}

template<typename T> inline CNumVector<T>::CNumVector(T x, T y)
{
    SetDim(2u, false);
    m_aValue[0] = x;
    m_aValue[1] = y;
}

template<typename T> inline CNumVector<T>::CNumVector(T x, T y, T z)
{
    SetDim(3u, false);
    m_aValue[0] = x;
    m_aValue[1] = y;
    m_aValue[2] = z;
}

template<typename T> inline CNumVector<T>::CNumVector(size_t N, const T* src)
{
    SetDim(N, false);
    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] = *(src++);
}

template<typename T> inline CNumVector<T>& CNumVector<T>::operator=(const CNumVector<T>& rhs)
{
    SetDim(rhs.Dim());
    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] = rhs.Data()[i];
    return *this;
}

template<typename T> inline T& CNumVector<T>::operator[](size_t i)
{
    ASSERT_ERROR(i < Dim());
    return m_aValue[i];
}

template<typename T> inline const T& CNumVector<T>::operator[](size_t i) const
{
    ASSERT_ERROR(i < Dim());
    return m_aValue[i];
}

template<typename T> inline void CNumVector<T>::Resize(size_t N)
{
    if(N > m_iDim)
    {
        // Save the current vector for restoring after reallocation.
        CNumVector<T> aOldVector = *this;
        SetDim(N, true);
        // Restore the old values, new values are undefined.
        for(size_t i = 0; i < aOldVector.Dim(); i++)
            m_aValue[i] = aOldVector[i];
    }
    else
        m_iDim = N;
}

template<typename T> inline void CNumVector<T>::Zero()
{
    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] = (T)0;
}

template<typename T> inline bool CNumVector<T>::operator==(const CNumVector<T>& rhs) const
{
    if(Dim() != rhs.Dim())
        return false;
    for(size_t i = 0; i < Dim(); i++)
        if(m_aValue[i] != rhs.m_aValue[i])
            return false;
    return true;
}

template<typename T> inline bool CNumVector<T>::operator!=(const CNumVector<T>& rhs) const
{
    return !(*this == rhs);
}

template<typename T> inline CNumVector<T>& CNumVector<T>::operator+=(const CNumVector<T>& rhs)
{
    ASSERT_ERROR(Dim() == rhs.Dim());
    if(Dim() != rhs.Dim())
        return *this;

    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] += rhs.m_aValue[i];
    return *this;
}

template<typename T> inline CNumVector<T>& CNumVector<T>::operator-=(const CNumVector<T>& rhs)
{
    ASSERT_ERROR(Dim() == rhs.Dim());
    if(Dim() != rhs.Dim())
        return *this;

    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] -= rhs.m_aValue[i];
    return *this;
}

template<typename T> inline CNumVector<T>& CNumVector<T>::operator*=(T rhs)
{
    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] *= rhs;
    return *this;
}

template<typename T> CNumVector<T>& CNumVector<T>::operator/=(T rhs)
{
    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] /= rhs;
    return *this;
}
//  The usual in-place matrix operations.

template<typename T> inline CNumVector<T>& CNumVector<T>::Add(T Factor, const CNumVector<T>& rhs)
{
    ASSERT_ERROR(Dim() == rhs.Dim());
    if(Dim() != rhs.Dim())
        return *this;

    for(size_t i = 0; i < Dim(); i++)
        m_aValue[i] += (T)(Factor * rhs.m_aValue[i]);
    return *this;
}

#if !defined(MIC_DM643)
template<class T> std::istream& CNumVector<T>::Read(std::istream& is)
{
    //  Read the dimention and make sure that the size is in an acceptable range.
    uint_32 iDim = 0;
    is.read((char*)&iDim, sizeof(iDim));

    if(iDim > 10000000)
    {
        is.clear(std::ios::failbit);
        ASSERT_ERROR(is.good());
        return is;
    }

    SetDim(size_t(iDim));
    if(iDim > 0)
        return ReadData(is);
    else
    {
        ASSERT_ERROR(is.good());
        return is;
    }
}

template<class T> std::istream& CNumVector<T>::ReadData(std::istream& is)
{
    is.read((char*)m_aValue, Dim() * sizeof m_aValue[0]);
    ASSERT_ERROR(is.good());
    return is;
}

template<class T> std::ostream& CNumVector<T>::Write(std::ostream& os) const
{
    uint_32 iDim = Dim();
    os.write((char*)&iDim, sizeof(iDim));
    if(iDim > 0)
        return WriteData(os);
    else
        return os;
}

template<class T> std::ostream& CNumVector<T>::WriteData(std::ostream& os) const
{
    os.write((char*)m_aValue, Dim() * sizeof m_aValue[0]);
    return os;
}

#endif

template<class T> inline const char* CNumVector<T>::GetBuff(const char* pBuff, bool* pfValid)
{
    //  Only do anything if still valid.
    if(pfValid != NULL && !*pfValid)
        return pBuff;

    //  Read the dimention and make sure that the size is in an acceptable range.
    uint_32 iDim;
    memcpy((char*)&iDim, pBuff, sizeof(uint_32));
    pBuff += sizeof(uint_32);
    if(iDim > size_t(-1))
    {
        if(pfValid != NULL)
            *pfValid = false;
        return pBuff;
    }

    SetDim(size_t(iDim));
    return GetBuffData(pBuff, pfValid);
}

template<class T> inline const char* CNumVector<T>::GetBuffData(const char* pBuff, bool* pfValid)
{
    if(pfValid == NULL || *pfValid)
    {
        memcpy((char*)m_aValue, pBuff, Dim() * sizeof m_aValue[0]);
        pBuff += Dim() * sizeof m_aValue[0];
    }
    return pBuff;
}

template<class T> inline char* CNumVector<T>::PutBuff(char* pBuff) const
{
    //  Write the data.
    uint_32 iDim = (uint_32)Dim();
    memcpy(pBuff, (char*)&iDim, sizeof(uint_32));
    pBuff += sizeof(uint_32);
    return PutBuffData(pBuff);
}

template<class T> inline char* CNumVector<T>::PutBuffData(char* pBuff) const
{
    memcpy(pBuff, (char*)m_aValue, Dim() * sizeof m_aValue[0]);
    pBuff += Dim() * sizeof m_aValue[0];
    return pBuff;
}

template<class T> inline void CNumVector<T>::SetDim(size_t N, bool fFree)
{
    //  If it shouldnt be freed (i.e. Called in the constructor) setup the
    //  Setup from scratch.
    if(!fFree)
    {
        m_aValue = NULL;
        m_iAlloc = 0;
    }

    //  If necessary, reallocate the structure.
    if(m_iAlloc < N || N == 0)
    {
        delete[] m_aValue;

        m_iAlloc = N;
        m_aValue = (m_iAlloc != 0 ? new T[m_iAlloc] : NULL);
    }

    //  Set the dimention.
    m_iDim = N;
}

// ----------------------------------------------------------------------------
template<class T> inline CNumVector<T> operator+(const CNumVector<T>& lhs, const CNumVector<T>& rhs)
{
    return CNumVector<T>(lhs) += rhs;
}

template<typename T> CNumVector<T> operator+(const CNumVector<T>& lhs, T rhs)
{
    auto vRet = std::vector<T>(lhs);
    for(auto i = 0; i < lhs.size(); ++i)
        vRet[i] += rhs;
    return vRet;
}

template<class T> inline CNumVector<T> operator-(const CNumVector<T>& lhs, const CNumVector<T>& rhs)
{
    return CNumVector<T>(lhs) -= rhs;
}

template<typename T> CNumVector<T> operator-(const CNumVector<T>& lhs, T rhs)
{
    auto vRet = std::vector<T>(lhs);
    for(auto i = 0; i < lhs.size(); ++i)
        vRet[i] -= rhs;
    return vRet;
}

template<class T> inline CNumVector<T> operator*(const CNumVector<T>& lhs, T rhs)
{
    return CNumVector<T>(lhs) *= rhs;
}

template<class T> inline CNumVector<T> operator*(T lhs, const CNumVector<T>& rhs)
{
    return CNumVector<T>(rhs) *= lhs;
}

template<class T> inline CNumVector<T> operator/(const CNumVector<T>& lhs, T rhs)
{
    return CNumVector<T>(lhs) /= rhs;
}

template<class T> inline T operator*(const CNumVector<T>& lhs, const CNumVector<T>& rhs)
{
    ASSERT_ERROR(lhs.Dim() == rhs.Dim());
    if(lhs.Dim() != rhs.Dim())
        return 1;

    T uiDot = 0;
    for(size_t i = 0; i < rhs.Dim(); i++)
    {
        uiDot += lhs.Data()[i] * rhs.Data()[i];
    }

    return uiDot;
}

template<class T> inline CNumVector<T> pow(const CNumVector<T>& lhs, const T& rhs)
{
    CNumVector<T> cRet(lhs.Dim());
    for(size_t i = 0; i < cRet.Dim(); i++)
    {
        cRet.Data()[i] = pow(lhs.Data()[i], rhs);
    }
    return cRet;
}

#if !defined(MIC_DM643)
/// Write a vector to a stream in the form [ v1 v2 v3 ... vn ].
template<class T> std::ostream& operator<<(std::ostream& os, const CNumVector<T>& rVector)
{
    os << rVector.Dim() << " [ ";
    for(size_t i = 0; i < rVector.Dim(); i++)
        os << rVector[i] << ' ';
    return os << ']';
}

/// True if the next character read from the stream is what we expect.
inline bool CheckCharacter(std::istream& is, char cExpectedCharacter)
{
    char cActualCharacter;
    is >> cActualCharacter;

    if( cActualCharacter == cExpectedCharacter )
        return true;

    is.clear( std::ios::failbit );
    return false;
}

/// Read a vector from a stream in the form [ v1 v2 v3 ... vn ].  The
/// vector must have the correct size before reading.
template<class T> std::istream& operator>>(std::istream& is, CNumVector<T>& rVector)
{
    size_t iSize;
    is >> iSize;
    if( !is ) // Make sure we read an integer
        return is;
    rVector.SetDim( iSize );

    if(!CheckCharacter(is, '['))
        return is;

    for(size_t i = 0; i < rVector.Dim(); i++)
    {
        is >> rVector[i];
    }

    CheckCharacter( is, ']' );
    return is;
}
#endif  // not streams

// ----------------------------------------------------------------------------
#if !defined(MIC_DM643)

/// Serialize a CNumVector (including the size)
template<typename T> void Serialize(CDataArchive& rAr, CNumVector<T>& rvdData)
{
    if(rAr.IsLoadingValid())
        rvdData.Read(rAr.GetIstream());
    else if(rAr.IsValid())
        rvdData.Write(rAr.GetOstream());
}

/// Serialize a CNumVector, but only the contents.
template<typename T> void SerializeContent(CDataArchive& rAr, CNumVector<T>& rvdData)
{
    if(rAr.IsLoadingValid())
        rvdData.ReadData(rAr.GetIstream());
    else if(rAr.IsValid())
        rvdData.WriteData(rAr.GetOstream());
}
#endif

// ----------------------------------------------------------------------------
CNumVector<double> ToDouble( const CNumVector<float>& rhs );
CNumVector<bool> ToBool( const CNumVector<uint_8>& rhs );

#endif // NUMVECT_HPP
