#ifndef RSAXISDATA_HPP
#define RSAXISDATA_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "colordef.hpp"
#include "rsdefaults.hpp"
#include "cstr.hpp"
#include "unitscontext.hpp"

// ============================================================================
///  This is a simple data container for all the options that pertain to a
///  single axis.
///
///  Each axis can be on vertical (right/left), horizontal (top/bottom) of
///  depth (left/right). Each axis has an ID to identify itself in the context
///  of a graph.  And each can be shown or not.
///
///  Axes are units aware including inversion (finer/coarser) for the title and
///  scale.  All quantities (for min/max) are in internal units.  The object
///  provides methods for converting from internal to display values.  Note
///  that there is a difference with an inverted scale (high values on
///  left/bottom) and a units inversion (the quantities are x-value).
///
///  The axis has a tile is CUnitsStr object that can update itself based in
///  the unit selections.  If the associated units have a units string, that is
///  appended to the title as well.
///
///  Both ends of the scale can be autoscaled independently.  If an end is not
///  autoscaled, the value to use for the end must be set (in internal units).
///  The axis can be either linear or logarithmic. The minimum and maximum ends
///  are the numericaly minimum and maximum values, and will be shown left to
///  right or bottom to top unless The scale is set to be revered.
class CRSGraphAxisData final
{
public:
    ///  The type of axis this is, and where it is to be shown.
    enum EType
    {
        ceBottom,
        ceTop,
        ceLeft,
        ceRight,
        ceDepthLeft,
        ceDepthRight
    };

    ///  Create a graph axis data object of the givn type and ID.
    CRSGraphAxisData(EType eType, int uiID);

    ~CRSGraphAxisData() = default;

    /// @returns the ID of the axis.
    int GetID() const { return m_uiID; }
    /// @returns the type of the axis.
    EType GetType() const { return m_eType; }
    /// Show or Hide the axis (useful for a 2nd y or x axis that
    /// might be shown or hidden with interactive reports )
    void SetShow(bool fShow) { m_fShow = fShow; }
    /// @returns if the axis should be shown.
    bool IsShown() const { return m_fShow; }

    /// @name Units Support
    /// @{

    /// Set the CUnits and optionally the CUnitsContext for the axis.
    void SetUnits(const CUnits& rcUnits) { m_Units = rcUnits; }
    /// Sets the context to be used by the title.
    void SetUnitsContext(std::shared_ptr<CUnitsContext> pUnitsContext);
    /// Sets up the column to be inverted in @p pUnitSel has a current
    /// selection of @p iSelForInvert.  The inversion is the @p dInvertValue -
    /// internal value of the column.
    void SetInverter(std::shared_ptr<CUnitSelection> pUnitSel, int iSelForInvert, double dInvertValue);

    /// Convert internal value to Display units
    double DisplayValue(double dInternal) const;
    /// Convert the display units to internal units.
    double ConvertToInternal(double dDisplay) const;

    /// @}

    /// @name Title Options
    /// @{

    /// @return the title string for the axis if @p fAddUnits is true, the
    /// units (if any) will be appended.
    CStr GetTitleText(bool fAddUnits = true) const;

    void SetTitleText(const CUnitsStr& rcStr) { m_TitleText = rcStr; }
    void SetTitleColor(COLORREF ulColor) { m_ulTitleColor = ulColor; }
    COLORREF GetTitleColor() const { return m_ulTitleColor; }

    /// @returns the text orientation that should be used for the title
    int GetTitleOrientation() const;

    /// @}

    /// @name Scale Options.
    /// @{

    bool IsLinear() const { return m_fIsLinear ? true : false; }
    bool IsLogarithmic() const { return !m_fIsLinear ? true : false; }
    bool AllowLogAxis() const { return m_fAllowLogAxis; }
    bool IsMinAutoscale() const { return m_fMinAuto ? true : false; }
    bool IsMaxAutoscale() const { return m_fMaxAuto ? true : false; }
    double GetMinScale() const;
    double GetMaxScale() const;
    double GetMinSeparation() { return m_dMinSeparation; }
    bool IsScaleInverted() const { return m_fScaleInverted ? true : false; }
    bool GetAllowGrid() const { return m_fAllowGrid; }
    COLORREF GetScaleColor() const { return m_ulScaleColor; }

    /// Convenience function to set the scaling.
    void Setup(bool fLinear,
        bool fAutoScaleMin = true,
        bool fAutoScaleMax = true,
        double dMin = 0.0,
        double dMax = 0.0);

    void SetLinear() { m_fIsLinear = true; }
    void SetLogarithmic() { m_fIsLinear = false; }
    void SetAllowLogAxis(bool fAllow) { m_fAllowLogAxis = fAllow; }
    void SetMinAutoscale(bool fMinAuto) { m_fMinAuto = fMinAuto; }
    void SetMaxAutoscale(bool fMaxAuto) { m_fMaxAuto = fMaxAuto; }
    void SetAutoscale(bool fMinAuto, bool fMaxAuto);
    void SetDisplayMinMaxScale(double dMin, double dMax);
    void SetMinScale(double dMin) { m_dMinValue = dMin; }
    void SetMaxScale(double dMax) { m_dMaxValue = dMax; }
    void SetScale(double dMin, double dMax);
    void SetMinSeparation(double dMin) { m_dMinSeparation = dMin; }
    void SetScaleInverted(bool fInvert) { m_fScaleInverted = fInvert; }
    void SetAllowGrid(bool fAllowGrid) { m_fAllowGrid = fAllowGrid; }
    void SetScaleColor(COLORREF ulColor) { m_ulScaleColor = ulColor; }

    /// Save the current set of the scale limits.  This supports the "Reset to
    /// Initial" option for the UI.  See also ReloadInitial().
    void MarkInitial();
    /// Restore the saved scale values from MarkInitial().  This supports the "Reset to
    /// Initial" option for the UI.
    void ReloadInitial();
    /// @}

    /// @returns true if there is a non trivial unit conversion in the axis scaling. Debug only.
    bool HasUnitsConversion() const;

    /// Allow the axis to save the range of the data.
    [[deprecated]] void SetDataRange(double dMin, double dMax);

    /// Give acces to the data range.
    [[deprecated]] double GetDataMin() const;
    [[deprecated]] double GetDataMax() const;

private:
    bool IsInvertUnitSel() const;

    const EType m_eType;
    const int m_uiID;

    //  Title information.
    CUnitsStr m_TitleText;
    COLORREF m_ulTitleColor;

    std::shared_ptr<CUnitsContext> m_pUnitContext;
    CUnits m_Units;

    std::shared_ptr<CUnitSelection> m_pInvertUnitSel;
    int m_iSelForInvert;
    double m_dInvertValue;

    //  Scale information.
    bool m_fIsLinear;
    bool m_fAllowLogAxis;  // Indicate whether to allow selecting log axis.
    bool m_fMinAuto;
    bool m_fMaxAuto;
    double m_dMinValue;
    double m_dMaxValue;
    double m_dMinSeparation;
    bool m_fScaleInverted;
    bool m_fAllowGrid;
    COLORREF m_ulScaleColor;

    bool m_fInitialMarked;
    double m_dInitMinValue;
    double m_dInitMaxValue;
    bool m_fInitMinAuto;
    bool m_fInitMaxAuto;

    /// Flag to indicate whether or not to draw the view associated with this data.
    bool m_fShow;

    /// The axis stores the data min max.  This supports temporary suspension of
    /// autoscale.  Only used in SetDataRange(), GetDataMin(), GetDataMax()
    double m_dDataMin, m_dDataMax;
};

#endif  // RSAXISDATA_HPP
