#ifndef RSAXISVIEW_HPP
#define RSAXISVIEW_HPP
#if _MSC_VER > 1000
#pragma once
#endif  // _MSC_VER > 1000
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "rsitemview.hpp"
#include "rsaxisdata.hpp"
#include "rsgraphdata.hpp"
#include <functional>

// ================================================================================================
/// The axis view is used for all axes (vertical and horizontal). The terms used are for those
/// describing a horizontal axis. For vertical axes, they stand for the analogous parts.
class CRSGraphAxisView final
{
public:
    using FnInvalidateRect = std::function<void(const SRect& rRect, bool fErase)>;
    static int_32 CalcMaxScaleSize(CReportDisplay& rDisp,
        const std::vector<CRSGraphAxisView*>& cvpViews);

    /// Create the axis view linked to the @p pData
    CRSGraphAxisView(CRSGraphAxisData* pData);
    CRSGraphAxisView(CRSGraphAxisView&&) = delete;
    ~CRSGraphAxisView();

    CRSGraphAxisView& operator=(const CRSGraphAxisView&) = delete;

    /// @{ access to the data object.
    CRSGraphAxisData* GetData() { return m_pData; }
    CRSGraphAxisData::EType GetType() const { return m_pData->GetType(); }
    int GetID() const { return m_pData->GetID(); }
    bool IsLinear() const { return m_pData->IsLinear(); }
    bool IsLogarithmic() const { return m_pData->IsLogarithmic(); }
    double DisplayValue(double dInternal) const;
    double ConvertToInternal(double dInternal) const;
    /// @}

    /// Set the range of the data associated with this axis.
    void SetDataScaleRange(double dDataMin, double dDataMax);

    /// This will calculate all the internal sizes used by the object.
    /// @param iAxisWidth Used to calculate the title size (@see GetTitleSize).
    ///    NOTE that iAxisWidth will be ignored in the case of the Z-axis.
    void CalcSizes(CReportDisplay& rDisp,
        int_32 iRectX,
        int_32 iRectY,
        int_32 iAxisWidth);

    /// Returns the full height of the axis including the title, scale, and tick marks.
    int_32 Height() const { return (int)(1.5 * m_iTitleHeight) + m_iScaleHeight + m_iTickHeight; }

    /// Returns the amount of extra space (past the plotting area necessary at the ends of the
    /// axis.
    int_32 LowExtra() const { return m_iLowExtra; }
    int_32 HighExtra() const { return m_iHighExtra; }

    /// Returns the extra height and width needed for this axis.
    int_32 DepthHeight() const { return m_iDepthHeight; }
    int_32 DepthWidth() const { return m_iDepthWidth; }

    /// Sets the reference points for the axis.
    void SetPlotRange(int_32 iXLow, int_32 iYLow, int_32 iXHigh, int_32 iYHigh, bool fLastIter);

    /// Sets the extra space between the edge of the plot area and the start of the graph.
    void SetExtraHeight(int_32 iExtraHeight) { m_iExtraHeight = iExtraHeight; }

    /// Draws the axis items as required by this class.
    /// @param iStackedGraphNum If negative, do not stagger the y-axis labels.
    void Paint(CReportDisplay& rDisp, int iStackedGraphNum);
    void Paint(CReportDisplay& rDisp);
    void Paint(CReportDisplay& rDisp, int_32 iLow, int_32 iHigh, bool fDrawScale);

    /// Draws the grid lines of thei axis as required by this class.
    void PaintGrid(CReportDisplay& rDisp, int_32 iLow = -99, int_32 iHigh = -99);

    /// This will show the position on the axis for this axis.
    void ShowPosition(CReportDisplay& rDisp,
        size_t ui,
        bool fShow,
        int_32 x,
        int_32 y,
        const FnInvalidateRect& fnInvalidateRect);

    /// This calculates the position on the axis of interest the value dVal should be. If the axis
    /// is a log scale, the dValue is already the log of the number to look up.
    int_32 DevicePos(double dValue) const;

    /// Give the device position from the raw user value of the Axis.
    int_32 DeviceFromUser(double dValue, bool& rfValid) const;

    /// @return the number of pixels to move along the axis to indicate the depth to be used.
    int_32 DepthX(double dDepth) const;
    int_32 DepthY(double dDepth) const;

    /// Does the inverse calculation of DevicePos. Except that if the scale
    /// is logarithmic, it will return the real value, and not the logarithm.
    double LogicalPos(int_32 iDevicePos) const;

    /// Give access to the extreames.
    double GetShowMin() const { return m_dShowMin; }
    double GetShowMax() const { return m_dShowMax; }

    void SetScaleHeight(int_32 iScaleHeight);

private:
    /// @param iStackedGraphNum If negative, do not stagger the y-axis labels.
    void Paint(CReportDisplay& rDisp, int_32 iLow, int_32 iHigh, bool fDrawScale, int iStackedGraphNum);

    /// Gets the minimum area to be shown by this graph independent of the autoscaling adjustments
    /// (if any).
    /// @returns the std::parin of min, max.
    std::pair<double, double> GetShowMinMax();

    /// This sets up the spacing (m_dShowMin, m_dShowMax) assuming a spacing of dd between numbers,
    /// and a minimum range of dMin to dMax. This also returns the number of items that will be
    /// shown.
    int_32 SetupSpacing(double dd, double dTd, double dMin, double dMax);

    /// Find the start and distance between the numbers to be printed, also do any autoscaling if
    /// necessary. The distances for the numbers are of the form 1, 2, or 5 x 10^n. The smallest
    /// number with acceptable less than or equal to iNumMax lables is to be used.
    void SetupSpacing(int iNumMax = 25);

    /// Calculates the height and width of the title rectangle as well as the number of lines that
    /// the title string will use.
    /// @param iAxisWidth Used to calculate the title size (@see GetTitleSize).
    ///    NOTE that iAxisWidth will be ignored in the case of the Z-axis.
    /// @returns a tuple of iWidth, iHeight, iNumLines
    /// iWidth: This is an output variable which tells the width of the title rectangle.
    /// iHeight: This is an output variable which tells the height of the title rectangle.
    /// iNumLines: This is an output variable which tells the number of lines that the title
    ///    string will span in the title rectangle.
    std::tuple<int_32, int_32, int_32> GetTitleSize(CReportDisplay& rDisp, int_32 iAxisWidth);
    /// @returns the tuple of iHeight, iLowExtra, iHightExtra
    std::tuple<int_32, int_32, int_32> GetScaleSize(CReportDisplay& rDisp);
    int_32 GetTickSize(CReportDisplay& rDisp_32);

    void DrawScale(CReportDisplay& rDisp);
    /// Draw the title string at the x, y position. If the string is longer than the window's
    /// dimensions, it is split into multiple lines and each line is drawn with a separate draw call
    /// in order to support vertical strings.
    void DoDrawTitle(CReportDisplay& rDisp, int_32 x, int_32 y);
    /// @param iStackedGraphNum If negative, do not stagger the labels.
    void DrawTitle(CReportDisplay& rDisp, int_32 iLow, int_32 iHigh, int iStackedGraphNum);
    void DrawTitleStacked(CReportDisplay& rDisp, int_32 iYLow, int_32 iYHigh);
    void DrawScaleStacked(CReportDisplay& rDisp, int_32 iYLow, int_32 iYHigh);

    using FnScale = std::function<void(double)>;
    /// @param dMin Minimum of the scale along the associated axis.
    /// @param dMax Maximum of the scale along the associated axis.
    /// @param dInc Scale increment being tested, or having already been established,
    /// by the caller.
    /// @param fnScale Function to be performed at each location established by
    /// incrementing by dInc between the scale boundaries.
    /// @param pData Data to be modified/used by pFcn each time it is called.
    void ScaleIter(double dMin, double dMax, double dInc, const FnScale& fcn);

    /// Where to draw the tick or number.
    struct STickData
    {
        int_32 x, y;
        int_32 xTick, yTick;
    };

    /// The callback function to draw the major tick mark and number.
    void DrawNumberCallback(double dValue, CReportDisplay& rDisp, STickData& rData);

    /// The callback function to draw the grid lines.
    void DrawGridlineCallback(double dValue, CReportDisplay& rDisp);

    /// The callback function to draw the minor tick mark.
    void DrawTickCallback(double dValue, CReportDisplay& rDisp, STickData& rData);

    /// The associated data object.
    CRSGraphAxisData* m_pData;

    /// The range to show for the axis.
    double m_dShowMin = 0.0, m_dShowMax = 0.0;
    double m_dShowLogMin = 0.0, m_dShowLogMax = 0.0;
    bool m_fShowExp = false, m_fShowMant = false;

    /// The sclale parameters.
    double m_dScaleFirst = 0.0, m_dScaleStep = 0.0, m_dTickStep = 0.0;

    /// The dimentions needed for positioning the graph.
    int_32 m_iTitleHeight = 0, m_iTitleWidth = 0, m_iNumLines = 0, m_iScaleHeight = 0,
           m_iTickHeight = 0;
    int_32 m_iLowExtra = 0, m_iHighExtra = 0;

    /// The high and low corners of the lot range for this axis.
    bool m_fHavePlotArea = false;
    int_32 m_iPlotXLow = 0, m_iPlotYLow = 0, m_iPlotXHigh = 0, m_iPlotYHigh = 0;
    int_32 m_iExtraHeight = 0;

    /// How much extra room is needed for the depth axis.
    int_32 m_iDepthWidth = 0, m_iDepthHeight = 0;

    void ShowPositionMarkers(CReportDisplay& rDisp);
    enum
    {
        ceMaxPosShow = 2
    };
    std::array<CStr, ceMaxPosShow> m_acPosStr;
    std::array<CStr, ceMaxPosShow> m_acPosStrExp;      ///< Position strings.
    std::array<bool, ceMaxPosShow> m_afPosShown = {};  ///< True iff the position is shown.
    std::array<bool, ceMaxPosShow> m_afPosShouldShow
        = {};                                          ///< True iff the position should be shown.
    std::array<SRect, ceMaxPosShow> m_acPosRect;       ///< Where the item should be shown.

    DECLARE_MSG()
};

#endif  // #define RSAXISVIEW_HPP
