#ifndef RSCURVEDATA_HPP
#define RSCURVEDATA_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "colordef.hpp"
#include "rsdataset.hpp"
#include "numvect.hpp"

//-----------------------------------------------------------------------------
// Forward Declarations
//-----------------------------------------------------------------------------
class CRSGraphOverlayData;
class CRSGraphCurveData;
class CRSGraphAxisData;
class CDataSetColumn;

namespace NRSGraphCurveData
{
    /// The axis association for the datasets.
    enum EAxis
    {
        ceHorz = 0,
        ceVert,
        ceDepth,
        ceLabel,
        ceFillFlag,
        ceAxisNum
    };
    /// The type of data to show for the curve.
    enum EType
    {
        ceHistogram = 0,
        ceCurve,
        cePoints,
        ceCurveAndPoints,
        ceTypeNum
    };
    /// The symbol to show when symbols are shown.
    enum ESymbol
    {
        ceSymLeftTriangle,
        ceSymRightTriangle,
        ceSymUpTriangle,
        ceSymDnTriangle,
        ceSymCircle,
        ceSymSquare,
        ceSymDiamond,
        ceSymPlus,
        ceSymCross,
        ceSymHourGlass,
        ceSymHorzHourGlass,
        ceSymNone,
        ceSymbolNum
    };
    /// The type of interpolation to use.
    enum EInterp
    {
        ceLinear = 0,
        ceAkimaSpline,
        ceInterpNum
    };
    /// Where (relative to the data point) to put the label.
    enum ELabelLoc
    {
        ceTopLeft = 0,
        ceTop,
        ceTopRight,
        ceLeft,
        ceCenter,
        ceRight,
        ceBottomLeft,
        ceBottom,
        ceBottomRight
    };

    /// @return an entry from the standard color palette.
    COLORREF ColorPalette(uint_16 uiIndex);

    /// @return an entry from the standard symbol palette.
    ESymbol SymbolPalette(uint_16 uiIndex);

    /// Change the saturation the COLOREF ulColor by a factor of dFactor.
    COLORREF SaturateColor(COLORREF ulColor, float_32 dFactor);

    /// @return the EType for @p fPoints and/or @p fCurve.
    EType TypeFromPointsCurve(bool fPoints, bool fCurve);
}  // namespace NRSGraphCurveData

// ============================================================================
/// This is a simple data container for all the options that pertain to a
/// single curve on a graph.
///
/// The curve has a title (that can be CUnitsContext aware), and multiple style
/// options.
///
/// It has columns of data associated with horizontal, vertical, and
/// optionally depth axes.  Each axis is associated with an axis ID that is
/// maintained in the graph.
///
/// Each curve can also have textual labels, subgraphs, and CRSGraphOverlayData
/// to show.
///
/// There is a historical interface that is still used, but is deprecated that
/// allows the user to work with the CDataSetColumn directly.  Once that is not
/// longer used we can change the datasets to std::vectors<double> (and CStr
/// for labels) directly.
class CRSGraphCurveData final
{
public:
    using EAxis = NRSGraphCurveData::EAxis;
    using EType = NRSGraphCurveData::EType;
    using ESymbol = NRSGraphCurveData::ESymbol;
    using EInterp = NRSGraphCurveData::EInterp;
    using ELabelLoc = NRSGraphCurveData::ELabelLoc;

    /// Create a curve object with the @p szTitle, and associated axis IDs.
    CRSGraphCurveData(const CUnitsStr& szTitle, int uiXAxisID, int uiYAxisID, int uiZAxisID = -1);
    ~CRSGraphCurveData() = default;

    /// @name Title Options.
    /// @{

    CStr GetTitleText() const;
    void SetTitleText(CUnitsStr rcStr) { m_TitleText = rcStr; }
    void SetTitleText(CStr rcStr) { SetTitleText(CUnitsStr(rcStr)); }

    void SetUnitsContext(std::shared_ptr<CUnitsContext> pContext);

    /// @name Display Parameters
    /// @{

    /// Method that copies the style parameters of pCurve to the current curve.
    void CopyStyle(const CRSGraphCurveData* pCurve);

    EType GetType() const { return m_eType; }
    ESymbol GetSymbol() const { return m_eSymbol; }
    bool GetSymbolHollow() const { return m_fHollowSymbol; }
    EInterp GetInterp() const { return m_eInterp; }
    bool IsShown() const { return m_fShow; }
    COLORREF GetPenColor() const { return m_ulPenColor; }
    int GetPenStyle() const { return m_iPenStyle; }
    COLORREF GetBrushColor() const { return m_ulBrushColor; }
    int GetBrushType() const { return m_iBrushType; }
    bool GetDefThickness() const { return m_fDefThickness ? true : false; }
    int GetThickness() const { return m_iThickness; }

    /// Currently only used by exporting functionality to determine if the curve object's
    /// data gets included in the output or not.
    bool IsIncludeInExport() const;

    /// Set the curve's basic properties
    /// @param uiMainIndex indexes the data-set (for instance the sample)
    /// @param uiSubIndex  indexes the sub-curve (an associated curve like a line-fit overlay)
    void SetProperties(EType eType, uint_16 uiMainIndex, uint_16 uiSubIndex = 0);

    void SetType(EType eType);
    void SetSymbol(ESymbol eSymbol, bool fHollowSymbol = false);
    void SetInterp(EInterp eInterp);
    void SetShow(bool fShow);
    void SetShow();
    void SetHide();
    void SetPenColor(COLORREF ulColor);
    void SetPenStyle(int iStyle);
    void SetBrushColor(COLORREF ulColor);
    void SetBrushType(int iType);
    void SetDefThickness(bool fDefault);
    void SetThickness(int iThickness);

    /// When a graph report gets exported, all curve data gets copied to the exported
    /// output (XLS/TXT etc.). Sometimes this is not desirable such as in the case
    /// of baseline curves for flowing-chemi graphical reports. This method allows
    /// the client to customize this behavior.
    void IncludeInExport(bool fInclude);

    /// @name  Curve Data Options.
    /// @{

    void ClearData();
    template<typename X, typename Y>
    void SetData(const X* adX, const Y* adY, size_t uCount);
    template<typename X, typename Y>
    void SetData(const std::vector<X>& adX, const std::vector<Y>& adY);
    void SetData(const CNumVector<double>& adX, const CNumVector<double>& adY);
    void AppendData(double dX, double dY);
    void AppendData(double dX, double dY, const CUnitsStr& szLabel);

    void SetGapIndexes(const std::vector<size_t>& cvuiGapIdx);

    const std::vector<size_t>& GetGapIndexes() { return m_vuiGapIdx; }

    // checks to see if the curve has zero data.  Optionally checks the subcurves also
    bool IsDataSetEmpty(bool fCheckSubCurves = true) const;

    int GetHorzAxisID() const { return m_aiAxisID[NRSGraphCurveData::ceHorz]; }
    int GetVertAxisID() const { return m_aiAxisID[NRSGraphCurveData::ceVert]; }
    int GetDepthAxisID() const { return m_aiAxisID[NRSGraphCurveData::ceDepth]; }
    int GetAxisID(EAxis eAxis) const { return m_aiAxisID[eAxis]; }

    bool IsUsed(EAxis eAxis) const { return m_abAxisUsed[eAxis] ? true : false; }

    /// @}

    /// @name Label, Subcurve, and Overlay methods.
    /// @{

    ELabelLoc GetLabelLocation() const { return m_eLabelLoc; }
    PCDataSetColumn GetLabelDataSet() { return m_pacDataSet[NRSGraphCurveData::ceLabel]; }
    std::shared_ptr<const CDataSetColumn> GetLabelDataSet() const;
    bool HasLabels() const { return m_abAxisUsed[NRSGraphCurveData::ceLabel] ? true : false; }

    void SetLabelLocation(ELabelLoc eLoc) { m_eLabelLoc = eLoc; }
    void SetLabelDataSet(PCDataSetColumn pDataSet);
    COLORREF GetLabelColor() const { return m_ulLabelColor; }
    void SetLabelColor(COLORREF ulColor) { m_ulLabelColor = ulColor; }

    const CRSGraphCurveData* GetSubCurve() const;
    CRSGraphCurveData* GetSubCurve();

    /// @param fForward If true, sub curves are added to the end of the list and drawn
    ///     in the order in which they were added. Otherwise, it's the reverse order.
    void SetForwardSubCurveDrawOrder(bool fForward);

    /// Creates a subcurve, and sets all the parameters to be the same as this
    /// curve.  Any differences can be overwritten.
    CRSGraphCurveData* AddSubCurve(const CUnitsStr& szTitle = CUnitsStr());
    void AddSubCurve(std::unique_ptr<CRSGraphCurveData> pCurve);
    //  Add the passed in curve as a 'sub-curve' of this object. There
    //  can be several of these that are chained together.
    void RemoveSubCurves();

    CRSGraphOverlayData* AddOverlayLine(double dSlope, double dYInt);
    CRSGraphOverlayData* GetOverlay() { return m_pOverlay.get(); }
    //  Adds a line overlay to the curve.

    /// @}

    /// @name Depricated functions
    /// @{

    CRSGraphCurveData();
    CRSGraphCurveData(EType eType,
        ESymbol eSymbol,
        ELabelLoc eLabelLoc,
        EInterp eInterp,
        COLORREF uiColor,
        int uiPenStyle,
        CStr szTitle);

    /// Used for changing the y-axis association
    void SetVertDataSetAxisID(int iAxisID);

    // Set CDataSetColumn that contains data points describing curve along
    // the Horz, vert, and depth axis (number of data points in all used axis
    // for curve must be same and in matching order) Specifying a depth data
    // will create a 3D-Curve that does not lie in any one x-y plane (as long
    // as the dataset contents are not all the same value)
    // The link to the Axis information. This is done by Axis ID values to make
    // the persistance and restoration easier.
    void SetHorzDataSet(PCDataSetColumn pDataSet, int iAxisID);
    void SetVertDataSet(PCDataSetColumn pDataSet, int iAxisID);
    void SetDepthDataSet(PCDataSetColumn pDataSet, int iAxisID);
    void SetDataSet(EAxis eAxis, PCDataSetColumn pDataSet, int iAxisID);

    // Get pointer to Horz, Vert, Depth DataSet used for curve
    std::shared_ptr<const CDataSetColumn> GetHorzDataSet() const
    {
        return m_pacDataSet[NRSGraphCurveData::ceHorz];
    }
    PCDataSetColumn GetHorzDataSet() { return m_pacDataSet[NRSGraphCurveData::ceHorz]; }
    std::shared_ptr<const CDataSetColumn> GetVertDataSet() const
    {
        return m_pacDataSet[NRSGraphCurveData::ceVert];
    }
    PCDataSetColumn GetVertDataSet() { return m_pacDataSet[NRSGraphCurveData::ceVert]; }
    std::shared_ptr<const CDataSetColumn> GetDepthDataSet() const
    {
        return m_pacDataSet[NRSGraphCurveData::ceDepth];
    }
    std::shared_ptr<const CDataSetColumn> GetDataSet(EAxis eAxis) const
    {
        return m_pacDataSet[eAxis];
    }
    PCDataSetColumn GetDataSet(EAxis eAxis) { return m_pacDataSet[eAxis]; }

    /// @}

private:
    DECLARE_MSG()

    /// The units context to use for string formatting.
    std::shared_ptr<CUnitsContext> m_pContext;

    /// If this is false, do not inclue this curve data in any export format.
    bool m_fIncludeInExport = true;

    //  Title information.
    CUnitsStr m_TitleText;

    //  Display options.
    EType m_eType;
    ESymbol m_eSymbol;
    bool m_fHollowSymbol;
    EInterp m_eInterp;
    bool m_fShow;

    COLORREF m_ulPenColor, m_ulBrushColor;
    //  The colors for the pen to draw the items, and the brush used for the
    //  internals of the histograms.

    int m_iBrushType, m_iPenStyle;
    //  The hash value for the brush, or -1 if it should be solid.

    bool m_fDefThickness;
    int m_iThickness;
    // use default fill type and thickness;

    ELabelLoc m_eLabelLoc;
    COLORREF m_ulLabelColor;
    //  Label information.

    // flags to indicate if information for indexed curve axis data is valid.
    bool m_abAxisUsed[NRSGraphCurveData::ceAxisNum];

    //  The ID that is used to identify the axis that each of the curves
    //  belongs to. This is bone By ID to make the persistance easier. It is
    //  up to the graph object to containing this and the axis to handle the
    //  explicit link.
    int m_aiAxisID[NRSGraphCurveData::ceAxisNum];

    // pointer to the DataSetColumn objects being used for curve data.
    PCDataSetColumn m_pacDataSet[NRSGraphCurveData::ceAxisNum];

    std::vector<size_t> m_vuiGapIdx;  // Indexes of points where no line should be drawn.

    std::unique_ptr<CRSGraphCurveData> m_pNextSubCurve;
    //  The next in the linked list of sub curves of this object.

    /// The pointer to the one overlay item, or nullptr if none available.
    std::unique_ptr<CRSGraphOverlayData> m_pOverlay;

    /// By default, sub-curves are added, drawn, and exported in reverse order.
    bool m_fForwardSubCurveDrawOrder = true;
};

// ================================================================================================
template<typename X, typename Y>
void CRSGraphCurveData::SetData(const X* adX, const Y* adY, size_t uCount)
{
    ASSERT_ERROR(GetHorzDataSet() && GetVertDataSet());
    GetHorzDataSet()->FillData(adX, uCount);
    GetVertDataSet()->FillData(adY, uCount);
}

template<typename X, typename Y>
void CRSGraphCurveData::SetData(const std::vector<X>& adX, const std::vector<Y>& adY)
{
    if(std::min(adX.size(), adY.size()) > 0)
        SetData(&adX[0], &adY[0], std::min(adX.size(), adY.size()));
    else
        SetData(static_cast<const Y*>(nullptr), static_cast<const Y*>(nullptr), 0);
}

#endif  // RSCURVEDATA_HPP
