#ifndef RSCURVEVIEW_HPP
#define RSCURVEVIEW_HPP
#if _MSC_VER > 1000
#pragma once
#endif  // _MSC_VER > 1000
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "rsitemview.hpp"
#include "rscurvedata.hpp"
#include "numintrp.hpp"

//-------------------------------------------------------------------------------------------------
class CRSGraphOverlayView;
class CRSGraphAxisView;
class CDataSetColumn;
// ================================================================================================
class CRSGraphCurveView final
{
public:
    CRSGraphCurveView(CRSGraphCurveData* pData, CRSGraphAxisView* pXAxis, CRSGraphAxisView* pYAxis);
    CRSGraphCurveView(CRSGraphCurveView&&) = default;
    ~CRSGraphCurveView() = default;

    CRSGraphCurveView& operator=(const CRSGraphCurveView&) = delete;

    ///  See if the curve has any points attached.
    bool HasPoints() const { return m_ulNumPoints > 0; }

    ///  This will calculate all the internal sizes used by the object.  It is assumes that the font
    ///  to be used is already set up.  The *pfHeightChanged is set to true of false depending on if
    ///  the overall height changed from the last call. The *pfWidthChanged is set simmilarly if the
    ///  Width changed.
    void CalcSizes(CReportDisplay& rDisp,
        bool* pfLegendHeightChanged = NULL,
        bool* pfLegendWidthChanged = NULL);

    ///  Give access to the dimensions of this legend item.
    int GetLegendHeight() const { return m_iTotalLegendHeight; }
    int GetLegendWidth() const { return m_iTotalLegendWidth; }

    CRSGraphAxisView* GetXAxisView() { return m_pXAxis; }
    CRSGraphAxisView* GetYAxisView() { return m_pYAxis; }

    ///  Sets the reference points for the axis.
    void SetPlotRange(int iXLow, int iYLow, int iXHigh, int iYHigh);

    ///  This will draw the legend with the top left corner being x, y.
    void PaintLegend(CReportDisplay& rDisp, int x, int y);

    ///  returns true iff the legends for these two curves are the same.
    bool LegendEqual(const CRSGraphCurveView& rhs) const;

    ///  returns true iff the should not be shown.
    bool LegendEmpty() const;

    ///  Draws the curve items as required by this class.
    void Paint(CReportDisplay& rDisp);

    ///  Give access to the underlying object.
    CRSGraphCurveData* GetData() { return m_pData; }
    const CRSGraphCurveData* GetData() const { return m_pData; }

    bool IsPrinting() { return m_fPrinting; }
    void SetPrinting(bool fPrinting);
    CRSGraphCurveData::ESymbol GetPrintMarker() { return m_ePrintMarker; }
    ///  Curve marker for printing on non-color printer.
    void SetPrintMarker(CRSGraphCurveData::ESymbol eSymbol) { m_ePrintMarker = eSymbol; }

    /// If set, curve drawing will service the window message queue preventing application lockup
    /// during lengthy drawing operations. If cleared, curve drawing will block until complete.
    ///
    /// If non-blocking, it is the caller's responsibility to ensure that the curve data is not
    /// modified while the drawing is taking place.
    void SetNonBlockingDraw(bool fSet);

private:
    /// Init Akima Interp. info for initial draw or curve change
    void InitInterpolation();

    //  Draw the curve in it various incarnations..
    void DrawHistogram(CReportDisplay& rDisp);
    void DrawCurve(CReportDisplay& rDisp);
    void DrawPoints(CReportDisplay& rDisp, bool fFilledPoints);
    void DrawLabels(CReportDisplay& rDisp);

    /// Draws the part of the line that is visible in the plotting area.
    /// @param fSkip True to skip drawing line and just move pointer to next position.
    void DrawClippedLine(CReportDisplay& rDisp,
        int iX1,
        int iY1,
        int iX2,
        int iY2,
        bool fSkip,
        bool& rfInStroke);

    ///  The associated data set.
    CRSGraphCurveData* m_pData;

    ///  The cached (possibly log10'ed) coordinate points.
    std::unique_ptr<double[]> m_adX;
    std::unique_ptr<double[]> m_adY;
    std::shared_ptr<const CDataSetColumn> m_pXCol;
    std::shared_ptr<const CDataSetColumn> m_pYCol;

    std::vector<size_t> m_vuiGapIdx;  ///< Indexes of points where no line should be drawn.

    std::vector<CUnitsStr> m_cstrStrList;

    ///  The data datasets to use from m_pData.
    uint_32 m_ulNumPoints;
    uint_32 m_ulNumLabels;

    ///  The plot area to show.
    int m_iPlotXLow = 0, m_iPlotYLow = 0, m_iPlotXHigh = 0, m_iPlotYHigh = 0;

    ///  The axes for the curve.
    CRSGraphAxisView* m_pXAxis;
    CRSGraphAxisView* m_pYAxis;

    /// If true, Paint() will service the window message queue to prevent application lockup during
    /// lengthy draw operations. If false, Paint() will block until complete.  @see
    /// SetNonBlockingDraw
    bool m_fIsNonBlockingDraw = false;
    /// Prevent reentrancy during a draw operation.
    bool m_fIsDrawing = false;

    /// The object that does the interpolation of the curves.
    /// This stores log values if the axis is logarithmic.
    std::vector<CNumInterpolate> m_vInterp;

    /// The dimensions for the legend.
    int m_iLegendHeight, m_iLegendWidth, m_iLegendSymbolWidth;
    /// The total dimensions for the legend including subcurves.
    int m_iTotalLegendHeight, m_iTotalLegendWidth;

    bool m_fPrinting;
    ///  Curve marker for printing on non-color printer.
    CRSGraphCurveData::ESymbol m_ePrintMarker;

    /// The next in the linked list of sub curves of this object.
    std::unique_ptr<CRSGraphCurveView> m_pNextSubCurve;

    ///  The view for the associated overlay or NULL if none.
    std::unique_ptr<CRSGraphOverlayView> m_pOverlayView;

    DECLARE_MSG()
};

//{{AFX_INSERT_LOCATION}}
#endif  // #define RSCURVEVIEW_HPP
