#ifndef RSDTASET_H
#define RSDTASET_H
//-----------------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------------
#include "basetype.hpp"
#include "cstr.hpp"
#include "unitscontext.hpp"

//-----------------------------------------------------------------------------
// Formward Declarations and typedefs
//-----------------------------------------------------------------------------
class CDataSetColumn;

//=============================================================================
/// A CDataSet can be understood as one column of data (text or
/// numerical).  Note that textual data used the CUnitsStr so that they can be
/// dynamically updated on the display when the unit selections change.
///
/// Each data set has a label.  This is not used for much except when the data
/// needs to be exported without other identifying information it can be used
/// to label the data.
class CDataSetColumn
{
public:
    using CPtr = std::shared_ptr<CDataSetColumn>;

    /// The type of storage if this is a numerical data.
    typedef std::vector<double> CFloatList;

    /// CDataSetColumn is only available through shared pointers.
    static CPtr Create();

    CDataSetColumn(CDataSetColumn&&) = default;
    ~CDataSetColumn() = default;

    CDataSetColumn& operator=(const CDataSetColumn&) = delete;

    /// Removes the contents of the data arrays and resets data identifiers
    void Clear();

    /// Checks to see if there is any data in the dataset.
    bool IsEmpty() const;

    /// Are the contents of this DataSet solely an array of strings (CStr)
    bool IsTypeString() const { return m_eSetType == ceText ? true : false; }

    /// Are the contents of this DataSet solely an array of numbers
    bool IsTypeNumber() const { return m_eSetType == ceFloat ? true : false; }

    /// Get the read only contents assuming that this is a sting list.
    const auto& GetStringList() const { return m_StrList; }
    /// Get the read only contents assuming that this is a sting list.
    auto& GetStringList() { return m_StrList; }
    /// Get the read only contents assuming that this is a numerical list.
    const auto& GetFloatList() const { return m_FloatList; }
    /// Give access to the numerical contents for updating.
    auto& GetFloatList() { return m_FloatList; }

    /// Return the number of data items contained in the singular array of
    /// items contained in this DataSet.
    size_t GetItemCount() const { return IsTypeString() ? m_StrList.size() : m_FloatList.size(); }

    /// Set the contents as a string list.
    void FillData(CStr* acDataBuf, const size_t ulItemCount);

    /// Set the contents of as a numerical list.
    template<typename T>
    void FillData(const T* pdDataBuf, const size_t ulItemCount);

    /// Clear list and fill with data in crvFloatList.
    void FillData(const CFloatList& crvFloatList);

    /// Add string items to the end of the current set.  This must be a string
    /// list or empty.
    void AppendItem(const CUnitsStr& szItem);
    /// Add string items to the end of the current set.  This must be a string
    /// list or empty.
    void AppendItem(const CStr& szItem);

    /// Add the contents of this DataSet's singular array of data items to a list of strings.
    void AppendItem(CStr* acDataBuf, const size_t ulItemCount);

    /// Add @p ulItemCount items to the end of the data starting at @p pdItem.
    template<typename T>
    void AppendItem(const T* pdItem, const size_t ulItemCount = 1);

    /// Append an empy Item to the list which is appropriate for this column (string or numeric)
    void AppendBlank();

    ///  Return the minimum value of the data in this item. This is only valid
    ///  for numerical data.
    double GetMinValue() const;
    ///  Return the /maximum value of the data in this item. This is only valid
    ///  for numerical data.
    double GetMaxValue() const;

private:
    CDataSetColumn();

    /// The options for the type of data.
    enum EDataType
    {
        ceEmpty,
        ceText,
        ceFloat,
        ceEndOfEnum
    };

    /// The type of data that this is.
    EDataType m_eSetType;

    /// The list of strings if this is a string dataset.
    std::vector<CUnitsStr> m_StrList;

    /// The list of values if this is a numerical dataset.
    CFloatList m_FloatList;
};

/// Standard type for shared pointer to CDataSetColumn
using PCDataSetColumn = std::shared_ptr<CDataSetColumn>;

// ================================================================================================
template<typename T>
void CDataSetColumn::FillData(const T* pdDataBuf, const size_t ulItemCount)
{
    Clear();
    AppendItem(pdDataBuf, ulItemCount);
}

template<typename T>
void CDataSetColumn::AppendItem(const T* pdItem, const size_t ulItemCount /*= 1*/)
{
    static_assert(!(std::is_same_v<T, char> || std::is_same_v<T, wchar_t>),
        "Cannot be used with constant strings.");
    if(m_eSetType != ceFloat)
    {
        ASSERT_ERROR(m_eSetType == ceEmpty);
        m_eSetType = ceFloat;
    }
    for(auto i = 0u; i < ulItemCount; ++i)
        m_FloatList.push_back(pdItem[i]);
}

#endif  // RSDTASET_H
