#ifndef RSDEFAULTS_HPP
#define RSDEFAULTS_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "basetype.hpp"
#include "cstr.hpp"
#include "datacont.hpp"
#include "formattr.hpp"
#include "rsitemdata.hpp"
#include <array>

// ================================================================================================
class CUnitsStr;
class CUnitsContext;

namespace NRSDefaults
{
    ///  The type of the grid to associate with this axis
    enum EGrid
    {
        ceGridNone,
        ceGridSolid,
        ceGridDotted
    };

    enum EGridStyle : uint_16
    {
        ceGridStyleSolid,
        ceGridStyleDotted
    };

#pragma pack(1)  // set to byte packing
    /// The style information for the default curve parameters.
    struct SCurveStye
    {
        uint_16 uiThickness;
        uint_32 ulBrushStyle;
        uint_16 uiBorderThickness;
    };
#pragma pack()  // restore compiler default data packing size

    /// @{ Formatted doubles for header image size.
    FORMATTED_DOUBLE(DRSBitmapWidth)
    FORMATTED_DOUBLE(DRSBitmapHeight)
    /// @}

    /// The platform independent style information for a font.
    class SSingleFontData : public CDataContainer
    {
    public:
        SSingleFontData() { SetDefaults(); }

        virtual void SetDefaults() override;
        virtual void SerializeData(CDataArchive& ar, uint_8 ubVersion) override;
        virtual uint_8 GetVersion() const override;

        CStr szFaceName;
        uint_32 ulHeight;  // Height in 1/100th of an inch
        uint_32 ulWidth;
        uint_32 lfOrientation;  //  Angle of rotation.

        uint_8 ubRed, ubGreen, ubBlue;  // Text Color.

        uint_8 fBold;
        uint_8 fItalic;
        uint_8 fUnderline;
        uint_8 fStrikeout;
        uint_8 fProtected;
        uint_8 fAutoColor;
        
        uint_8 lfCharSet; // Character set
    };

    /// The parameters that define what appears on all headers
    struct SRSCommonHeader : public CDataContainer
    {
    public:
        SRSCommonHeader();

        virtual void SetDefaults() override;
        virtual void SerializeData(CDataArchive& ar, uint_8 ubVersion) override;

        CStr szReportTitle;
        bool fShowTitle;
        bool fShowBitmap;
        CStr szBitmapFile;
        DRSBitmapHeight xdBitmapHeight;
        DRSBitmapWidth xdBitmapWidth;
    };

    /// All the style data that is written to disk.
    class SReportStyleData : public CDataContainer
    {
    public:
        static SReportStyleData& Get();
        static bool Save();

        virtual void SetDefaults() override;
        virtual void SerializeData(CDataArchive& ar, uint_8 ubVersion) override;

        SRSCommonHeader sHeader;

        std::array<SSingleFontData, CReportDataItemBase::ceFontTypes> asFontData;

        uint_16 uiThickness;
        uint_32 ulBrushStyle;
        uint_16 uiBorderThickness;

        bool fGridHorzLinMajor;
        bool fGridHorzLinMinor;
        bool fGridHorzLogMajor;
        bool fGridHorzLogMinor;
        bool fGridVertLinMajor;
        bool fGridVertLinMinor;
        bool fGridVertLogMajor;
        bool fGridVertLogMinor;
        EGridStyle eGridStyleMajor;
        EGridStyle eGridStyleMinor;

        float_64 dSplitterFraction;

    private:
        SReportStyleData();
    };

    /// Set the location of the default graphic to be shown in the report header. Should be called
    /// before the first call to SReportStyleData::Get()
    void SetDefaultMicEmfFilename(const CStr& szPath);
    /// Set the filename of where the report defaults are stored. Must be called before the first
    /// call to SReportStyleData::Get()
    void SetSaveFilename(const CStr& szPath);

    /// Fill in the curve style with default information.
    void GetCurveStye(SCurveStye* psCS);

    /// @return the height (in 1/100 #include) of the font @p eType.
    uint_32 Height(CReportDataItemBase::EFontType eType);

    /// @{Access to the default grid lines
    EGrid GetMajorGrid(bool fVert, bool fLinear);
    EGrid GetMinorGrid(bool fVert, bool fLinear);
    /// @}

    /// @return the global header default parameters.
    const SRSCommonHeader& GetHeaderDefaults();

    /// @return the report view splitter window location.
    double GetSplitterFraction();

    /// Update the global report view splitter location.
    void SetSplitterFraction(double dFraction);

    /// Access the platform independent font information for the type.
    const SSingleFontData& GetFontData(CReportDataItemBase::EFontType eType);

    /// @returns Path for the default Micromeritics image file
    CStr GetDefaultMicEmf();

    /// Serialize the CUnitsStr item.  Note that on saving the CUnitsStr are converted to the
    /// explicit CStr with the pContext (or the default if not passed).  If it is different on
    /// reading, that will not change the strings.
    void Serialize(CDataArchive& rAr, CUnitsStr& rData, CUnitsContext* pContext);

}  // namespace NRSDefaults

// ================================================================================================
/// Class that is notified if any defaults change.
class CReportDefaultsNotify
{
public:
    /// Notify all of these objects to update themselves.
    static void ViewDefaultsChanged();

    CReportDefaultsNotify();
    CReportDefaultsNotify(const CReportDefaultsNotify& rNotify);
    CReportDefaultsNotify(CReportDefaultsNotify&& rNotify);
    virtual ~CReportDefaultsNotify();

    CReportDefaultsNotify& operator=(const CReportDefaultsNotify&) = default;
    CReportDefaultsNotify& operator=(CReportDefaultsNotify&&) = default;

    /// Called whenever the defaults are updated.
    virtual void OnReportDefaultsUpdated() = 0;
};

#endif
