#ifndef RSDISPLAY_HPP
#define RSDISPLAY_HPP
//-------------------------------------------------------------------------------------------------
/// @file The abstract display (screen/printer/etc) interface that is used by reports.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "rsitemdata.hpp"

// ================================================================================================
/// The abstraction for a display context that is used to draw the CReportDataItem's.
///
/// Note that all calls to its methods must be within a LockContext. This is because some
/// CReportDisplay objects have different contexts at different times.
///
/// There are also contexts for Font, Pen, Brush and ClipRegion.  These are created and reset in
/// RAII fashion.
///
class CReportDisplay
{
    friend class CReportDisplayFwd;

public:
    class CLockContext;
    class CFontContext;
    class CPenContext;
    class CBrushContext;
    class CClipRgnContext;
    friend class CLockContext;

    using EFontType = CReportDataItemBase::EFontType;
    enum ETextAlign
    {
        ceTextAlignLeft,
        ceTextAlignCenter,
        ceTextAlignRight,
    };

    CReportDisplay() = default;
    CReportDisplay(CReportDisplay&&) = default;
    virtual ~CReportDisplay() = default;

    /// Create the RAII context for interacting with the display.  These can be nested, but there
    /// must be at least one before around any of the actual context calls.  This is necessary
    /// because some Display types (i.e. the CHWnd based one) could have different actual contexts
    /// at different times in it lifetime.
    CLockContext LockContext();

    /// @{
    /// Create the RAII context for the font to be used in text methods.
    CFontContext FontContext(EFontType eType);
    CFontContext FontContext(EFontType eType, uint_32 ulColor);
    CFontContext FontContext(EFontType eType, uint_32 ulColor, int iOrientation);
    /// @}

    /// Create the RAII context for the pen to be used in line drawing methods.
    CPenContext PenContext(int iType, int iWidth, uint_32 ulColor);

    /// @{
    /// Create the RAII context for the brush to be used in filling methods.
    CBrushContext BrushContext(uint_32 ulColor);
    CBrushContext BrushContext(int iType, uint_32 ulColor);
    CBrushContext NullBrushContext();
    /// @}

    /// Create the RAII context for the clip region to be used in drawing methods.
    CClipRgnContext ClipRgnContext(int x1, int y1, int x2, int y2);

    /// @Return the contexts x and y resolution of the display.
    virtual std::pair<int, int> GetPixelsPerInch() const = 0;
    /// Is it a raster of vector display.
    virtual bool IsRaster() const = 0;
    /// @Return the width of the page for printing, or windows width if on screen.
    virtual int PrintingResOrWndWidth() const = 0;
    /// @Return true if this display is black and white printing.
    virtual bool IsPrintingBW() const = 0;
    /// @return the x and y size of a page when printing.
    virtual std::pair<int, int> GetPaperSize() const = 0;
    /// @return the default background color for the display.
    virtual uint_32 GetBackgroundColor() const = 0;

    /// @{
    /// Functions to convert to and from display and internal resolution.  Internal resolution is at
    /// 100 per inch.
    int FromDisplayX(int iDisplayX) const;
    int FromDisplayY(int iDisplayY) const;
    int ToDisplayX(int iX) const;
    int ToDisplayY(int iY) const;
    SRect FromDisplay(const SRect& rsRect);
    SRect ToDisplay(const SRect& rsRect);
    /// @}

    /// @return the size of the boarder of a standard frame rectangle.
    virtual int CalcFrameHeight() const = 0;
    /// @return the height of a default window edge.
    virtual int CalcCYEdge() const = 0;
    /// @return the x and y size of @p szText when drawn in the current context.
    virtual std::pair<int, int> CalcTextExtent(const CStr& cszText) const = 0;
    /// @return the height of @p szText when drawn in the current context with wrapping at @p
    /// iWidth.
    virtual int CalcTextHeight(const CStr& cszText, ETextAlign eAlign, int iWidth) const = 0;
    /// @return the height of @p szText when drawn in the current context with wrapping at @p
    /// iWidth, but add in the frame height.
    int CalcTextHeightWithFrame(const CStr& cszText, ETextAlign eAlign, int iWidth) const;
    /// @return the nominal text height in the current context.
    virtual int CalcTextNominalHeight() const = 0;

    /// @{
    /// Set the background mode or color for text operations.
    virtual void SetBkModeTransparent() = 0;
    virtual void SetBkModeOpaque() = 0;
    virtual void SetBkColor(uint_32 uiColor) = 0;
    /// @}

    /// @{
    /// Show the test at the location given in the current font context.
    virtual void TextOut(int x, int y, const CStr& szText) = 0;
    virtual void DrawTextWithBreak(const CStr& szText, ETextAlign eAlign, const SRect& cRect) = 0;
    virtual void DrawText(const CStr& szText, ETextAlign eAlign, const SRect& cRect) = 0;
    /// @}

    /// @{
    /// Draw the geometry requested with the current pen and brush context.
    virtual void MoveTo(int x, int y) = 0;
    virtual void LineTo(int x, int y) = 0;
    virtual void Stroke() = 0;
    virtual void Ellipse(int x1, int y1, int x2, int y2) = 0;
    virtual void Polygon(const SPoint* lpPoints, int nCount) = 0;
    virtual void Rectangle(int x1, int y1, int x2, int y2) = 0;
    virtual void FillRect(uint_32 uColor, const SRect& cRect) = 0;
    /// @}

#if defined(MIC_VCMFC)
    virtual void PlayEnhMetaFile(HENHMETAFILE hemf, RECT* psRect) = 0;
    virtual void DrawBitmap(BITMAPINFO* psInfo, char* pBits, int x1, int y1, int x2, int y2) = 0;
#endif

private:
    /// @{
    /// The current font information.
    virtual void DoPushFont(EFontType eType, uint_32 ulColor, int iOrientation) = 0;
    virtual void DoPopFont() = 0;
    /// @}

    /// @{
    /// The current pen information.
    virtual void DoPushPen(int iPen, int iWidth, uint_32 ulColor) = 0;
    virtual void DoPopPen() = 0;
    /// @}

    /// @{
    /// The current brush information.
    virtual void DoPushBrush(uint_32 ulColor) = 0;
    virtual void DoPushBrush(int iType, uint_32 ulColor) = 0;
    virtual void DoPushNullBrush() = 0;
    virtual void DoPopBrush() = 0;
    /// @}

    /// @{
    /// The current clip region.
    virtual void DoPushClipRgn(int x1, int y1, int x2, int y2) = 0;
    virtual void DoPopClipRgn() = 0;
    /// @}

    /// @{
    /// For some display objects, the actual display context can change during the life of this
    /// object.  For these we have to make sure it is allocated and freed.  This must be done in
    /// pairs through instances of the CReportDisplayContext objects.
    virtual void DoAllocateDisplay() {}
    virtual void DoFreeDisplay() {}
    void AllocateDisplay();
    void FreeDisplay();
    int m_iContextCount = 0;
    /// @}
};

// ============================================================================================
/// An RAII context variable that fixes the pdc in the CReportDisplay.
class CReportDisplay::CLockContext final
{
public:
    CLockContext(CReportDisplay& rDisp);
    CLockContext(const CLockContext&) = delete;
    CLockContext& operator=(const CLockContext&) = delete;
    ~CLockContext();

private:
    CReportDisplay& m_rDisp;
};

// ============================================================================================
/// An RAII context variable that sets and restores the font used in text operations.
class CReportDisplay::CFontContext
{
public:
    CFontContext(CReportDisplay& rDisp) : m_pDisp(&rDisp){};
    CFontContext(CFontContext&& rhs) : m_pDisp(nullptr) { std::swap(m_pDisp, rhs.m_pDisp); }
    CFontContext(const CFontContext&) = delete;
    ~CFontContext()
    {
        if(m_pDisp)
            m_pDisp->DoPopFont();
    }

private:
    CReportDisplay* m_pDisp;
};

// ============================================================================================
/// An RAII context variable that sets and restores the pen used in drawing operations.
class CReportDisplay::CPenContext
{
public:
    CPenContext(CReportDisplay& rDisp) : m_pDisp(&rDisp){};
    CPenContext(CPenContext&& rhs) : m_pDisp(nullptr) { std::swap(m_pDisp, rhs.m_pDisp); }
    CPenContext(const CPenContext&) = delete;
    ~CPenContext()
    {
        if(m_pDisp)
            m_pDisp->DoPopPen();
    }

private:
    CReportDisplay* m_pDisp;
};

// ============================================================================================
/// An RAII context variable that sets and restores the fill brush used in drawing operations.
class CReportDisplay::CBrushContext
{
public:
    CBrushContext(CReportDisplay& rDisp) : m_pDisp(&rDisp){};
    CBrushContext(CBrushContext&& rhs) : m_pDisp(nullptr) { std::swap(m_pDisp, rhs.m_pDisp); }
    CBrushContext(const CBrushContext&) = delete;
    ~CBrushContext()
    {
        if(m_pDisp)
            m_pDisp->DoPopBrush();
    };

private:
    CReportDisplay* m_pDisp;
};

// ============================================================================================
/// An RAII context variable that sets and restores the clipping region.
class CReportDisplay::CClipRgnContext
{
public:
    CClipRgnContext(CReportDisplay& rDisp) : m_pDisp(&rDisp){};
    CClipRgnContext(CClipRgnContext&& rhs) : m_pDisp(nullptr) { std::swap(m_pDisp, rhs.m_pDisp); }
    CClipRgnContext(const CClipRgnContext&) = delete;
    ~CClipRgnContext()
    {
        if(m_pDisp)
            m_pDisp->DoPopClipRgn();
    };

private:
    CReportDisplay* m_pDisp;
};

// ================================================================================================
/// This is the report display for times when we really don't have one.  It does not do anything,
/// but is safe to call.  It does not assert for each function call since it is only created in one
/// place, and that place asserts.
class CReportDisplayNull : public CReportDisplay
{
public:
    virtual std::pair<int, int> GetPixelsPerInch() const override;
    virtual bool IsRaster() const override;
    virtual int PrintingResOrWndWidth() const override;
    virtual bool IsPrintingBW() const override;
    virtual std::pair<int, int> GetPaperSize() const override;
    virtual uint_32 GetBackgroundColor() const override;

    virtual void SetBkModeTransparent() override;
    virtual void SetBkModeOpaque() override;
    virtual void SetBkColor(uint_32 uiColor) override;

    virtual int CalcFrameHeight() const override;
    virtual int CalcCYEdge() const override;
    virtual std::pair<int, int> CalcTextExtent(const CStr& szText) const override;
    virtual int CalcTextHeight(const CStr& szText, ETextAlign eAlign, int iWidth) const override;
    virtual int CalcTextNominalHeight() const override;
    virtual void TextOut(int x, int y, const CStr& szText) override;
    virtual void DrawTextWithBreak(const CStr& szText,
        ETextAlign eAlign,
        const SRect& cRect) override;
    virtual void DrawText(const CStr& szText, ETextAlign eAlign, const SRect& cRect) override;

    virtual void MoveTo(int x, int y) override;
    virtual void LineTo(int x, int y) override;
    virtual void Stroke() override;
    virtual void Ellipse(int x1, int y1, int x2, int y2) override;
    virtual void Polygon(const SPoint* lpPoints, int nCount) override;
    virtual void Rectangle(int x1, int y1, int x2, int y2) override;
    virtual void FillRect(uint_32 uColor, const SRect& cRect) override;

#if defined(MIC_VCMFC)
    virtual void PlayEnhMetaFile(HENHMETAFILE hemf, RECT* psRect) override;
    virtual void DrawBitmap(BITMAPINFO* psInfo, char* pBits, int x1, int y1, int x2, int y2) override;
#endif

private:
    virtual void DoPushFont(EFontType eType, uint_32 ulColor, int iOrientation) override;
    virtual void DoPopFont() override;
    virtual void DoPushPen(int iPen, int iWidth, uint_32 ulColor) override;
    virtual void DoPopPen() override;
    virtual void DoPushBrush(uint_32 ulColor) override;
    virtual void DoPushBrush(int iType, uint_32 ulColor) override;
    virtual void DoPushNullBrush() override;
    virtual void DoPopBrush() override;
    virtual void DoPushClipRgn(int x1, int y1, int x2, int y2) override;
    virtual void DoPopClipRgn() override;
};

#endif // RSDISPLAY_HPP
