#ifndef RSGRAPHDATA_HPP
#define RSGRAPHDATA_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include "colordef.hpp"
#include "rsaxisdata.hpp"
#include "rscurvedata.hpp"
#include "rslegenddata.hpp"
#include "rstextdata.hpp"
#include "rsoverlaydata.hpp"
#include "rssurfacedata.hpp"
#include "unitscontext.hpp"

//-----------------------------------------------------------------------------
// Forward Declarations
//-----------------------------------------------------------------------------
class CRSGraphAxisData;
class CRSGraphCurveData;
class CReportGraphData;
class CDataSetColumn;
class CRSGraphSurfaceData;

namespace NReportGraphData
{
    std::string JsonSchema();
}

//=============================================================================
/// The CReportDataItemBase derived class that contains all the graph data and
/// parameters. Graphs are typically X-Y charts, but can also be X-Y-Z surfaces.
///
/// Graphs have a title (and possibly subtitle) along with CRSGraphAxisData and
/// CRSGraphCurveData (or CRSGraphSurfaceData).  They have legends. and
/// optional overlay.
///
/// The usage is can be simple.  Just create the graph object and set the
/// title, subtitle and units context (which will be passed onto the current
/// and future components).  Add, and configure the axes, then add and
/// configure the curves. Then add the data to the curves. Finally, add any
/// overlays that are needed.

class CReportGraphData final : public CReportDataItemBase
{
    /// @return the json value for @p rData ready to be inserted into @p cData.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CReportGraphData& cData);

public:
    using CPtr = std::shared_ptr<CReportGraphData>;
    using EAxisType = CRSGraphAxisData::EType;

    /// Create an empty graph object.
    CReportGraphData();
    CReportGraphData(const CReportGraphData& rcCopy) = delete;
    CReportGraphData& operator=(const CReportGraphData& rcCopy) = delete;
    ~CReportGraphData() = default;

    void SerializeData(CDataArchive& rAr, const VDataItem& cvDataItems) override;

#if defined(_DEBUG)
    virtual CStr DbgName() const override;
#endif

    // CReportDataItemViewBase override
    virtual bool CanExpandView() const override;

    /// Set the CUnitsContext for this, and all contained objects.
    void SetUnitsContext(std::shared_ptr<CUnitsContext> pUnitsContext);

    /// @name Title Methods.
    /// @{

    void SetTitle(const CUnitsStr& rcTitle);
    void SetSubTitle(const CUnitsStr& rcTitle);
    void ShowTitle(bool bArg);
    void SetBorderColor(COLORREF ulColor) { m_ulBorderColor = ulColor; }

    CReportTextData* GetTitleData();
    CReportTextData* GetSubTitleData();
    const CReportTextData* GetTitleData() const;
    const CReportTextData* GetSubTitleData() const;
    COLORREF GetBorderColor() const { return m_ulBorderColor; }

    /// @}

    /// @name Axis support.
    /// @{

    CRSGraphAxisData* AddAxis(EAxisType eType, int iAxisID, CStr rcTitle);
    CRSGraphAxisData* AddAxis(EAxisType eType, int iAxisID, CUnitsStr rcTitle);
    bool HasAxis(int iAxisID) const;
    const CRSGraphAxisData* GetAxis(int iAxisID) const;
    CRSGraphAxisData* GetAxis(int iAxisID);

    /// This supports the UI element "Restore Axis scaling to Initial defaults"
    /// functionality by marking all the axis data objects.  It should be
    /// called by the display window during construction to setup the initial
    /// values.
    void MarkInitialScale();

    /// @}

    /// @name Curve Support.
    /// @{

    /// Create a new curve based on the specified DataSets. Each data
    /// set is linked to an axis through the axis ID which must
    /// (eventualy) correspond to an axis of the proper type.
    CRSGraphCurveData* AddCurve(const CUnitsStr& szTitle,
        int iHorzAxisID,
        int iVertAxisID,
        int iDepthAxisID = 0);
    /// Create a new curve based on the specified DataSets. Each data
    /// set is linked to an axis through the axis ID which must
    /// (eventualy) correspond to an axis of the proper type.
    CRSGraphCurveData* AddCurve(const CStr& szTitle,
        int iHorzAxisID,
        int iVertAxisID,
        int iDepthAxisID = 0);

    [[deprecated]] CRSGraphCurveData* AddCurve(const CStr& cTitle,
        PCDataSetColumn pHorzData,
        int iHorzAxisID,
        PCDataSetColumn pVertData,
        int iVertAxisID,
        PCDataSetColumn pDepthData = nullptr,
        int iDepthAxisID = 0,
        std::vector<size_t> vuiGapIdx = std::vector<size_t>());

    size_t GetNumCurves() const { return m_aCurve.size(); }
    CRSGraphCurveData* GetCurve(int uiWhichCurve);
    const CRSGraphCurveData* GetCurve(int uiWhichCurve) const;
    const std::vector<std::unique_ptr<CRSGraphCurveData>>& GetCurves() const;

    /// Removes all the curves from the graph.
    void DeleteAllCurves();

    /// Resets the graph object to its initial creation state (no axis, curve,
    /// overlay data).
    void Reset();

    /// @}

    /// @name Overlay and Legend methods.
    /// @{

    ///  Create a new linear overlay. Each data set is linked to an axis
    ///  through the axis ID which must (eventualy) correspnd to an axis of the
    ///  proper type.
    CRSGraphOverlayData* AddOverlayLine(double dSlope,
        double dYIntercept,
        int iHorzAxisID,
        int iVertAxisID);
    ///  Create a new linear overlay. Each data set is linked to an axis
    ///  through the axis ID which must (eventualy) correspnd to an axis of the
    ///  proper type.
    CRSGraphOverlayData* AddOverlayVerticalLine(double dXPos, int iHorzAxisID, int iVertAxisID);
    ///  Create a new polynomial overlay. Each data set is linked to an axis
    ///  through the axis ID which must (eventualy) correspnd to an axis of the
    ///  proper type.
    CRSGraphOverlayData* AddOverlayPolynomial(const CNumPolynomialExtrap& rcPoly,
        int iHorzAxisID,
        int iVertAxisID);

    const std::vector<std::unique_ptr<CRSGraphOverlayData>>& GetOverlays() const;

    //  Set the legend to the passed in type and return a pointer to it.
    //  The reason it is called add is for consistancy with the other items.
    CRSGraphLegendData* AddLegend(CRSGraphLegendData::EType eType);

    //  Accessors to the legend.
    CRSGraphLegendData* GetLegend() { return &m_LegendData; }
    const CRSGraphLegendData* GetLegend() const { return &m_LegendData; }

    /*
    **  Surface Support.
    */
    CRSGraphSurfaceData* AddSurface(int iXAxisID, int iYAsixID, int iZAxisID);
    CRSGraphSurfaceData* GetSurface() { return m_pSurface.get(); }
    const CRSGraphSurfaceData* GetSurface() const { return m_pSurface.get(); }

    /// Draw the curves from first to last. By default, curves are drawn
    /// last to first.
    [[deprecated]] void SetForwardDrawOrder();
    /// see SetForwardDrawOrder
    [[deprecated]] bool IsForwardDrawOrder() const;

private:
    bool m_fShowTitle;

    /// When set to true the curves are drawn from first to last.
    bool m_fUseForwardDrawOrder;

    COLORREF m_ulBorderColor;

    /// The units Context for this item
    std::shared_ptr<CUnitsContext> m_pUnitContext;

    /// The uiID to axis map for the graph.
    std::map<int, std::unique_ptr<CRSGraphAxisData>> m_aAxisMap;
    /// All the curves in this graph.
    std::vector<std::unique_ptr<CRSGraphCurveData>> m_aCurve;
    /// All the overlays that this graph knows about.
    std::vector<std::unique_ptr<CRSGraphOverlayData>> m_aOverlay;
    /// The surface display.
    std::unique_ptr<CRSGraphSurfaceData> m_pSurface;

    enum EGraphTextItem
    {
        ceGphTitle = 0,
        ceGphSubTitle,
        ceGphItemCount
    };
    std::array<std::unique_ptr<CReportTextData>, ceGphItemCount> m_aTxtBox;

    CRSGraphLegendData m_LegendData;
};

#endif  // RSGRAPHDATA_HPP
