#ifndef RSGRAPHVIEW_HPP
#define RSGRAPHVIEW_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "rsgraphdata.hpp"
#include "rslegendview.hpp"
#include "rsitemview.hpp"
#include "rssurfaceview.hpp"
#include "rsdensity-graph-view.hpp"
#include "rstitleview.hpp"
#include "rsoverlayview.hpp"
#include "rscurveview.hpp"

// ------------------------------------------------------------------------------------------------
// Forward references.
// ------------------------------------------------------------------------------------------------
class CRSGraphSurfaceView;
class CRSGraphAxisView;
class CRSGraphCurveView;
class CRSGraphOverlayView;
class CRSGraphLegendView;
class CRSGraphTitleView;

struct SCurveWithAxes
{
    SCurveWithAxes(CRSGraphCurveView* ppCurve) : ppCurve(ppCurve) {}

    CRSGraphCurveView* ppCurve;
    int iPlotXLow = 0;
    int iPlotYLow = 0;
    int iPlotXHigh = 0;
    int iPlotYHigh = 0;
};

// ================================================================================================
/// This is a graph controller base class.  The functions below are set by the view, but used by the
/// graph window and its derived classes.  This allows that communication to happen without full
/// dependency of the window classes on the view classes.
class CReportGraphCtrl
{
public:
    virtual ~CReportGraphCtrl() = default;

    /// Allow derived classes to setup a minimum autoscale for this axis.
    /// Returns true if the scale was setup.  false oterwise.
    virtual bool SetMinAutoscale(CRSGraphAxisView* pAxis, float_64& dMin, float_64& dMax) = 0;
    /// @return True if we always want to pad top of Y axis when autoscaling.
    virtual bool AlwaysPadYMax() = 0;
};

//=================================================================================================
/// The view for the CReportGraphData object.
class CReportGraphView : public TReportDataItemView<CReportGraphData>
{
public:
    using CAxisViewList = std::vector<std::unique_ptr<CRSGraphAxisView>>;
    using CCurveViewList = std::vector<std::unique_ptr<CRSGraphCurveView>>;
    using COverlayViewList = std::vector<std::unique_ptr<CRSGraphOverlayView>>;
    using CCurveWithAxesList = std::vector<SCurveWithAxes>;
    using EAxis = CRSGraphCurveData::EAxis;
    using ESymbol = CRSGraphCurveData::ESymbol;

    virtual ~CReportGraphView();

    virtual void LinkToBase(CReportDataItemBase* pData) override;
    virtual void OnBaseChanged(unsigned int uiUpdateFlag) override;
    virtual void PrintPartial(uint_32 uPageIndex, const CPageInfoVec& cPageInfo) override;
    virtual void DrawItem(const SRect& cLocation) override;

    void SetCtrl(CReportGraphCtrl* pCtrl);
    size_t GetHeight() const;
    size_t GetWidth() const;

    CCurveViewList& GetCurveList() { return m_cCurveList; }
    const CCurveViewList& GetCurveList() const { return m_cCurveList; }
    CCurveWithAxesList& GetCurveWithAxesList() { return m_vsCurveWithAxesList; }
    const CCurveWithAxesList& GetCurveWithAxesList() const { return m_vsCurveWithAxesList; }
    CAxisViewList& GetAxisList() { return m_vpAxisList; }
    const CAxisViewList& GetAxisList() const { return m_vpAxisList; }
    /// Find the axis view with this ID, or NULL if none available. If the View is not in the list,
    /// but is available in the data, the view is created.
    CRSGraphAxisView* GetAxis(uint_32 uiID);

    void SetDisplayMode(bool fStacked) { m_fStacked = fStacked; }
    bool GetDisplayMode() { return m_fStacked; }
    void SetAllowStackedColor(bool fAllowSetColor) { m_fAllowSetColor = fAllowSetColor; }
    void SetNonBlockingDraw(bool fSet) { m_fIsNonBlockingDraw = fSet; }

    void DisableAutoScale(bool fDisabled) { m_fDisableAutoScale = fDisabled; }

    bool GetStacked() const { return m_fStacked; }
    const SRect& GetViewingArea() const { return m_sViewingAreaRect; }
    int GetPlotXLow() const { return m_iPlotXLow; }
    int GetPlotXHigh() const { return m_iPlotXHigh; }
    int GetPlotYLow() const { return m_iPlotYLow; }
    int GetPlotYHigh() const { return m_iPlotYHigh; }
    bool GetAllowStackedColor() const { return m_fAllowSetColor; }
    bool IsNonBlockingDrawDrawing() const { return m_fIsNonBlockingDraw && m_fIsDrawing; }

    void Draw();

    // This recalculates the positions of the various subobjects contained in this class and if
    // necessary moves them into their new positions, as specified by the rsRect rectangle.
    void UpdateLayout(const SRect& csRect);

    /// This will create the views that are necessary for the display.
    void SetupViews(bool fAutoScale = true);

private:
    /// Get the min and max values of the curve within the current x-range. Values may be set larger
    /// than the curve min/max to allow room for labels.
    /// @param eAxis Axis on which to get min/max values.
    /// @return True if *pdMin or *pdMax was modified.
    bool GetCurveMinMax(CRSGraphCurveData* pData, EAxis eAxis, float_64* pdMin, float_64* pdMax);

    /// Get the min and max values of the curve within the current x-range.
    /// @param eAxis Axis on which to get min/max values.
    /// @return True if *pdMin or *pdMax was modified.
    bool GetSingleCurveMinMax(CRSGraphCurveData* pData, EAxis eAxis, float_64* pdMin, float_64* pdMax);
    // Get the min and max values of the curve on this axis, taking into account the x-axis range
    // if necessary. This returns true if there is data to return.
    void AdjustMinMaxForLabel(CRSGraphCurveData* pDataMain,
        double& rdCurveMax,
        double& rdCurveMin,
        EAxis eAxis) const;

    std::pair<int, int> CalculateLeftRightMargins() const;
    int GetGraphCurveCount();
    void DoOverLayedUpdateLayout(const SRect& csRect);
    void DoStackedUpdateLayout(SRect sRect);
    void DrawStacked();
    void DrawOverlay();

    /// Find the curve marker symbol that is not used by another curve.
    ESymbol GetSymbol(const CRSGraphCurveView* pCurve) const;

    /// The associates control object if any.
    CReportGraphCtrl* m_pCtrl = nullptr;

    /// If true, curve drawing will service the window message queue to prevent application lockup
    /// during lengthy draw operations. If false, curve drawing will block until complete.  @see
    /// SetNonBlockingDraw
    bool m_fIsNonBlockingDraw = false;
    /// Prevent reentrancy during a draw operation.
    bool m_fIsDrawing = false;
    /// Allow to disable (temporarily) the autoscale updates.
    bool m_fDisableAutoScale = false;
    /// graph mode - true when graph is to be shown in STACKED mode
    bool m_fStacked = false;
    /// Allows the user to change the axis label color.
    bool m_fAllowSetColor = false;

    std::unique_ptr<CRSGraphTitleView> m_pTitle;

    /// The object that handles the responcibility for the depth in a "3-D" history plot.
    CRSGraphAxisView* m_pDepthAxis = nullptr;
    std::unique_ptr<CRSGraphSurfaceView> m_pSurface;
    std::unique_ptr<CDensityGraphView> m_pDensityView;

    /// The contents of the curve.
    CCurveViewList m_cCurveList;

    CCurveWithAxesList m_vsCurveWithAxesList;

    /// The contents of the curve.
    COverlayViewList m_cOverlayList;

    /// The legend information.
    std::unique_ptr<CRSGraphLegendView> m_pLegendView;

    /// The contents of the curve.
    CAxisViewList m_vpAxisList;

    /// The plot rectangle in this window.
    int m_iPlotXLow = 0;
    int m_iPlotYLow = 0;
    int m_iPlotXHigh = 0;
    int m_iPlotYHigh = 0;

    int m_iTitleOffset = 0;

    /// The position of the legend.
    int m_iLegendPosX = 0, m_iLegendPosY = 0;

    /// the graph bottom Y-coordinate
    int m_iLargestYLow = 0;

    SRect m_sViewingAreaRect = {0, 0, 0, 0};

    DECLARE_MSG()
};

#endif // RSGRAPHVIEW_HPP
