#ifndef RSHEADERDATA_HPP
#define RSHEADERDATA_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "datacontext.hpp"
#include "rsitemdata.hpp"
#include "rsdataset.hpp"
#include "rstablehelper.hpp"
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class CReportTextData;
class CReportTableData;
class CReportHeaderData;
class CReportImageData;
class CReportPageData;
class CReportModel;

namespace NReportHeaderData
{
    std::string JsonSchema();
}

// ============================================================================
/// Lightweight class to generate a comment report when the sample-comment
/// string is longer than five lines. The reason is that PDF reports with
/// a sample comment longer than five lines can cause formatting problem in
/// some report tables of the PDF document.  It should be noted that some
/// report selections generate acceptable PDF output when the comment is longer
/// than five lines, but the formatting errors are apparent for different selections.
/// This class is a workaround to fix this issue.
class CCommentData
{
public:
    CCommentData(const CStr& szComment);
    /// @return true if the comment requires its own report page.
    bool NeedsCommentReport();
    /// Add a single report to the report model for a long comment.
    void AddCommentReport(CReportPageData* pReport);

private:
    const CStr m_cszComment;
};

// ============================================================================
/// CReportDataItemBase derived Report Data Item. Responsible for storing Header
/// configuration information and text and table items contents. Entire Header
/// will be made up of text and table item types.
class CReportHeaderData : public CReportDataItemBase
{
    /// @return the json value for @p cData.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CReportHeaderData& cData);

public:
    /// Create standard full detail header item;
    CReportHeaderData();
    /// Create the simple header with just the one table at the top.
    CReportHeaderData(size_t uiColumns);
    virtual ~CReportHeaderData() = default;

    CReportHeaderData(const CReportHeaderData& rcCopy) = delete;
    CReportHeaderData& operator=(const CReportHeaderData& rcCopy) = delete;

#if defined(_DEBUG)
    virtual CStr DbgName() const override;
#endif

    void SerializeData(CDataArchive& rAr, const VDataItem& cvDataItems) override;

    /// @returns true if the header configuration is a simple style.
    bool IsSimple() const;

    /// @name Setup the data for the header.
    /// @{

    void UpdateDefaultTitle();

    /// Allow special reports to set a special title.  (e.g. sample print/export)
    void OverrideTitle(const CStr& szTitle);

    /// Set the name and version of the application that is producing the report.
    void SetReportingAppName(const CStr& szAppName);
    /// Set the information about the application and instrument that were used to
    /// collected the data. The port number will not be shown if not specified or < 1.
    void SetAnalysisAppInfo(const CStr& cszAppName,
        const CStr& cszSerialNumber = CStr(),
        int iUnitNumber = 0,
        int iPortNumber = 0,
        const CStr& cszModelName = CStr());
    /// @see SetAnalysisAppInfo( const CStr&, const CStr&, int, int, int ).
    void SetAnalysisAppInfo(const CStr& cszAppName,
        const CStr& cszSerialNumber,
        const CStr& cszUnitNumber,
        const CStr& cszPortNumber);
    /// @see SetAnalysisAppInfo( const CStr&, const CStr&, int, int, int ).
    void SetAnalysisAppInfo(const CStr& cszAppName,
        const CStr& cszSerialNumber,
        const CStr& cszUnitNumber,
        const CStr& cszPortNumber,
        const CStr& cszModelName);
    void SetPageNumber(int iPage);
    void SetSample(const CStr& szPrompt, const CStr& szSample);
    void SetOperator(const CStr& szPrompt, const CStr& szOperator);
    void SetSubmitter(const CStr& szPrompt, const CStr& szSubmitter);
    void SetBarCode(const CStr& szPrompt, const CStr& szSubmitter);
    void SetFile(const CStr& szFilename, const CStr& cszCaption = CStr());

    /// Set page number in appropriate field for printing purposes. This is a change
    /// to the header data and so an update will reflect the change.
    void SetPrintPageNumber(uint_32 uiPageNumber, uint_32 uiNumPages);
    /// Set the text for the page number field. Does nothing if a header-like table is
    /// not found.
    void SetPageNumberText(const CStr& cszPageNumber);
    /// Get the text for the page number field. Returns empty string
    CStr GetPageNumberText() const;

    void SetHeaderItem(const CUnitsStr& szPrompt, const CUnitsStr& szDesc);
    void SetHeaderItem(int iResID, const TFormattedNumeric<double>& xdVal);
    void SetHeaderItem(const CUnitsStr& szPrompt, const TFormattedNumeric<double>& xdVal);
    void SetHeaderItem(const CStr& szPrompt, const CStr& szDesc);
    void SetHeaderItem(const int iResID, const CStr& szDesc);
    void SetHeaderItem(const int iResID1, const int iResID2);
    template<typename T, typename S>
    void SetHeaderItem(const int iResID1, S T::*pMember, const T& rData);

    void SetNote(const CStr& szNote);
    void SetComment(const CStr& szData);

    void SetPrepTableColumns(int iCols);
    void AddPrepTableData(int iCol, const CStr& szValue);

    /// @}

    /// @{ Set the text for the next item in the column of the simple header.
    void SetSimpleItem(size_t uiColumn, const CUnitsStr& cszText);
    void SetSimpleItem(size_t uiColumn, const CStr& cszText);
    /// @}

    /// @name View Support Items
    /// @{

    bool GetShowTitle() const { return m_fShowTitle ? true : false; }
    bool GetShowBitmap() const { return m_fShowBitmap ? true : false; }

    auto GetBitmapItem() const { return m_pImage.get(); }
    auto GetTitleItem() const { return m_pTitleText.get(); }
    auto GetTopTableItem() const { return m_pTopTable.get(); }
    auto GetCommentItem() const { return m_pCommentText.get(); }
    /// @}

    auto size() const { return m_DataItems.size(); }
    auto& operator[](size_t i) const { return m_DataItems[i]; }
    auto& operator[](size_t i) { return m_DataItems[i]; }
    auto begin() const { return m_DataItems.begin(); }
    auto end() const { return m_DataItems.end(); }
    auto begin() { return m_DataItems.begin(); }
    auto end() { return m_DataItems.end(); }

    virtual void SetFramePos(int_32 lPosX, int_32 lPosY, bool fUpdate = true) override;
    virtual void SetFrameSize(int_32 lWidth, int_32 lHeight, bool fUpdate = true) override;

private:
    void SetShowTitle(bool f) { m_fShowTitle = f; }
    void SetTitle(const CStr& szTitle);
    void SetShowBitmap(bool f) { m_fShowBitmap = f; }
    void SetHeaderImageFile(const CStr& szFile) const;
    void SetImageFrame(double dHeight, double dWidth);

    /// Which kind of header is it.
    bool m_fSimpleHeader = false;

    /// The table used for both simple and full headers
    std::unique_ptr<CReportTableData> m_pTopTable;

    /// @{ The data for the full header.
    bool m_fOverrideDefaultTitle = false;
    bool m_fShowBitmap = false;
    std::unique_ptr<CReportImageData> m_pImage;
    bool m_fShowTitle = false;
    std::unique_ptr<CReportTextData> m_pTitleText;
    std::unique_ptr<CReportTableData> m_pDescriptionTable;
    std::unique_ptr<CReportTableData> m_pOptionTable;
    int m_iCurrRow = 0;  ///< Row index for option table.
    int m_iCurrCol = 0;  ///< Column index for option table.
    std::unique_ptr<CReportTableData> m_pNoteTable;
    int m_iNoteTblRow = 0;  ///< Row Index for note table
    std::unique_ptr<CReportTextData> m_pCommentText;
    std::unique_ptr<CReportTableData> m_pPrepTable;
    int m_iPrepTableCols;
    /// @}

    /// An array of the above items to treat them as a group;
    std::vector<CReportDataItemBase*> m_DataItems;
};

template<typename T, typename S>
void CReportHeaderData::SetHeaderItem(const int iResID1, S T::*pMember, const T& rData)
{
    if constexpr(NUnits::is_quantity_v<S>)
    {
        auto xdValue = TFormattedNumeric<typename S::value_type>(NDataContext::GetFormatter(pMember),
            (rData.*pMember).Value());

        SetHeaderItem(iResID1, xdValue);
    }
    else
        SetHeaderItem(iResID1,
            TFormattedNumeric<S>(NDataContext::GetFormatter(pMember), rData.*pMember));
}

#endif  // RSTEXTDATA_HPP
