#ifndef RSITEMDATA_H
#define RSITEMDATA_H
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "data-view-update.hpp"
#include <vector>
#include <memory>
#include <array>

//-----------------------------------------------------------------------------
// Enumerations, Forward Declarations etc...
//-----------------------------------------------------------------------------
class CDataArchive;
class CReportDataItemBase;
class CReportDataItemViewBase;
class CStr;

//-------------------------------------------------------------------------------------------------
/// A Simple Point structure (x, y). Used by the layout scheme of CReportDataItemBase(s).
struct SPoint final
{
    SPoint() = default;
    SPoint(int_32 x, int_32 y);

    int_32 x = 0;
    int_32 y = 0;
};

//-------------------------------------------------------------------------------------------------
/// A Simple Size structure (cx, cy). Used by the layout scheme of CReportDataItemBase(s).
struct SSize final
{
    SSize() = default;
    SSize(int_32 cx, int_32 cy);

    int_32 cx = 0;
    int_32 cy = 0;

    SSize& operator=(const SSize& sSize) = default;

    friend bool operator==(const SSize& sSizeL, const SSize& sSizeR);
    friend bool operator!=(const SSize& sSizeL, const SSize& sSizeR);

#if defined(MIC_VCMFC)
    SSize(const CSize& cSize);
    CSize MFCSize() const;
    operator const CSize();
    operator const tagSIZE();
    SSize& operator=(const CSize cSize);
#endif
};

//-----------------------------------------------------------------------------
/// A Simple Rectangle structure. (left, top) is the top left corner of the rect
/// and (width, height) is the width and height of the rectangle.  Used by the
/// layout scheme of CReportDataItemBase(s).
struct SRect final
{
    SRect(){};
    SRect(const int_32 _left, const int_32 _top, const int_32 _right, const int_32 _bottom);
    SRect(const SPoint& cTopLeft, const SSize& cSize);

    int_32 left;
    int_32 top;
    int_32 right;
    int_32 bottom;

    void Clear()
    {
        left = 0;
        top = 0;
        right = 0;
        bottom = 0;
    }
    void SetRect(const int_32 _left, const int_32 _top, const int_32 _right, const int_32 _bottom);
    void SetUL(const int_32 _top, const int_32 _left);
    void SetBR(const int_32 _bottom, const int_32 _right);
    void SetWidth(const int_32 _width) { right = left + _width - 1; }
    void SetHeight(const int_32 _height) { bottom = top + _height - 1; }
    void SetRect(const SRect& sRect);
    void DeflateRect(const int_32 _iX, const int_32 _iY);
    void InflateRect(const int_32 _iX, const int_32 _iY);
    void OffsetRect(const int_32 _iDeltaX, const int_32 _iDeltaY);

    bool IsPointInRect(const int_32 _iX, const int_32 _iY) const;

    SPoint TopLeft() const { return SPoint(left, top); }
    SSize Size() const { return SSize(right - left, bottom - top); }
    int_32 Width() const { return right - left + 1; }
    int_32 Height() const { return bottom - top + 1; }

    SRect& operator=(const SRect& sRect);

    friend bool operator==(const SRect& sRectL, const SRect& sRectR);
    friend bool operator!=(const SRect& sRectL, const SRect& sRectR);
    friend SRect& operator+=(SRect& sRectL, const SSize& cSize);
    friend SRect operator|(const SRect& sRectL, const SRect& sRectR);

#ifdef __AFXWIN_H__
public:
    SRect(const CRect& rcRect);
    CRect MFCRect() const;
    operator const CRect();
    operator const tagRECT();
    SRect& operator=(const CRect cRect);
#endif
};

/// Allow the SRect structure to be serialized.
void Serialize(CDataArchive& rAr, SRect& rData);

//=============================================================================
/// Base class for all the report system data objects.
///
/// This class provides a consistant mechanism for accessing common attributes
/// of Report Data Items and to allow for simplified expansion for more data
/// item types.
///
/// This class defines the layout mechanism for report data items in a
/// CReportPageData.  On construction Call the SetFrameRect() method, and then
/// if this item should be placed relative to another call the
/// SetLayoutBottom() or SetLayoutRight() to do the positioning.
///
/// This class also manages some behavior parameters for all types of data
/// objects.   See \ref behavior.
///
/// Objects derived from this type are designed to be linked to TUpdateViewBase
/// views. This allows the application to work with all data aspects of a
/// visual representation for a report item without having any dependencies of
/// GUI libraries. There exists a sister 'view' class CReportDataItemViewBase
/// for this data class, which handles the GUI related manipulations. This Data
/// base class and the View class communicate through the update mechanism.
class CReportDataItemBase : public TUpdateDataBase<CReportDataItemBase>
{
    friend class CReportDataItemViewBase;

public:
    using VDataItem = std::vector<std::unique_ptr<CReportDataItemBase>>;

    /// List of font items that can be used in reports.
    enum EFontType : uint_16
    {
        ceNone = 0,         // Type Not Set
        ceText,             // TextData Item font
        ceTableTitle,       // Table Title font
        ceTableHeader,      // Table Header font
        ceTableColumn,      // Table Column font
        ceGraphTitle,       // Graph Title font
        ceGraphAxisTitle,   // Graph Axis Title font
        ceGraphAxisScale,   // Graph Axis Scale font
        ceGraphCurveLabel,  // Graph Curve Label font
        ceGraphLegend,      // Graph Legend font
        ceHeaderTable,      // The table text in the header
        ceFontTypes         // Number of Font Types
    };

    CReportDataItemBase();
    CReportDataItemBase(const CReportDataItemBase& rcObject) = delete;
    CReportDataItemBase& operator=(const CReportDataItemBase& rcObject) = delete;

    virtual ~CReportDataItemBase() = default;

    /// This should be overwritten to write the serialize itself.  The derived classes
    /// implementation should call the SerializeBaseData() method to save its part of the state.
    virtual void SerializeData(CDataArchive& rAr, const VDataItem& cvDataItems) = 0;

#if defined(_DEBUG)
    virtual CStr DbgName() const = 0;
#endif

    /// @return true if this report item can be stretched to fit the enclosing window
    virtual bool CanExpandView() const;

    /// @name layout
    /// Position and Size modifiers. Units should be in 1/100".
    /// @{

    /// Set size and position of report item in units of 1/100" relative to the upper left
    /// corner of a page.
    virtual void SetFrameRect(const SRect& rcRect, bool fUpdate = true);

    /// Specify Report Data Item after which this item is to be placed. lSpacing is used
    /// to space the two items apart. The FrameRect left and right positions
    /// will be used to position the item horizontally, but the top and bottom
    /// values will be adjusted at view creation time to position this item below
    /// the specified item.
    void SetLayoutBottom(CReportDataItemBase* pReportDataItem, int_32 lSpacing);

    /// Specify Report Data Item to the right of which this item is to be placed.
    /// lSpacing is used to space the two items apart. The FrameRect top and bottom
    /// positions will be used to position the item vertically, but the left and right
    /// values will be adjusted at view creation time to position this item to the right
    /// of the specified item.
    void SetLayoutRight(CReportDataItemBase* pReportDataItem, int_32 lSpacing);
    // Set position of report item in units of 1/100" relative to the upper left
    // corner of a page.

    /// The Frame Rect is used to position this item upon view creation. (This is the default)
    void SetLayoutAbsolute();

    virtual void SetFramePos(int_32 lPosX, int_32 lPosY, bool fUpdate = true);

    // Set size of report item in units of 1/100" relative to the upper left
    // corner of the report item.
    virtual void SetFrameSize(int_32 lWidth, int_32 lHeight, bool fUpdate = true);

    // Return (READ only) the bounding rectangle (left, top, right bottom) of
    // the report item in units of 1/100" relative to the upper left corner of
    // the page.
    const SRect& GetFrameRect() const { return m_sFrameRect; }

    int_32 Width() const { return m_sFrameRect.Width(); }
    int_32 Height() const { return m_sFrameRect.Height(); }
    int_32 Left() const { return m_sFrameRect.left; }
    int_32 Right() const { return m_sFrameRect.right; }
    int_32 Top() const { return m_sFrameRect.top; }
    int_32 Bottom() const { return m_sFrameRect.bottom; }

    // Get ID for Report Item used for ralative layout.
    CReportDataItemBase* GetTopItem() const { return m_apRelativeItem[ceLayoutTop]; }
    CReportDataItemBase* GetLeftItem() const { return m_apRelativeItem[ceLayoutLeft]; }

    int_32 GetTopItemSpacing() const { return m_alRelativeItemSpacing[ceLayoutTop]; }
    int_32 GetLeftItemSpacing() const { return m_alRelativeItemSpacing[ceLayoutLeft]; }

    uint_32 GetLayoutFlags() const;
    bool IsLayoutAbsolute() const;
    bool IsLayoutHorz() const;
    bool IsLayoutVert() const;

    bool IsLayoutRight() const { return (m_eHorzLayoutType == ceLayoutRight) ? true : false; }
    bool IsLayoutLeft() const { return (m_eHorzLayoutType == ceLayoutLeft) ? true : false; }
    bool IsLayoutTop() const { return (m_eVertLayoutType == ceLayoutTop) ? true : false; }
    bool IsLayoutBottom() const { return (m_eVertLayoutType == ceLayoutBottom) ? true : false; }
    /// @}

    /// @name behavior
    /// @{
    // Change flag associated with the item's size changeability via GUI mechanisms
    void Sizeable(const bool bArg) { m_fIsSizeable = bArg; }
    // Is the item's size changeable using GUI mechanisms.
    bool IsSizeable() const { return m_fIsSizeable ? true : false; }

    void AllowInteraction(bool bArg) { m_fAllowInteraction = bArg; }
    bool IsInteractionAllowed() const { return m_fAllowInteraction ? true : false; }

    void SetAutoSize(bool bArg) { m_fAutoSize = bArg; }
    bool IsAutoSize() const { return m_fAutoSize ? true : false; }
    /// @}

    // Flags (non-maskable) that identify the nature of an UpdateView(...)
    // message to a view object. A derived class should use ceViewUpdateFlagBase
    // as the starting enumeration number for it's own update enumeration.
    enum EViewUpdateFlag
    {
        ceGeneric = 0,
        ceFramePos,
        ceFrameSize,
        ceFramePosSize,
        ceLayoutChanged,
        ceDataChanged,
        ceViewUpdateFlagBase
    };

protected:

    /// This should be overwritten to write the serialize itself.  The derived classes
    /// implementation should call the SerializeBaseData() method to save its part of the state.
    void SerializeBaseData(CDataArchive& rAr, const VDataItem& cvDataItems);

private:
    /// Layout mechanism to be used for this report item.
    enum EItemLayout : uint_16
    {
        ceLayoutTop = 0,
        ceLayoutBottom = 1,
        ceLayoutRight = 2,
        ceLayoutLeft = 3,
        ceLayoutAbsolute
    };

    // Frame representing bounding (left, top, right, middle) for item's visible
    // region. Used for visual placement of data on screen, printer, and file.
    // Units are in 1/100" representing the  most points of the outer frame rectangle.
    SRect m_sFrameRect;

    // Specify primary layout alignment
    EItemLayout m_eHorzLayoutType;
    EItemLayout m_eVertLayoutType;

    // -ve values are treated as absolute layout.
    std::array<CReportDataItemBase*, 4> m_apRelativeItem;
    std::array<int_32, 4> m_alRelativeItemSpacing;

    bool m_fIsSizeable;
    bool m_fAllowInteraction;
    bool m_fAutoSize;
};

//-----------------------------------------------------------------------------
#endif  // RSITEMDATA_H
