#ifndef RSITEMVIEW_HPP
#define RSITEMVIEW_HPP
//-------------------------------------------------------------------------------------------------
/// @file The non MFC views on report items.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "rsitemdata.hpp"
#include "rsdisplay.hpp"
#include <array>
#include <memory>
#include <map>
#include <stack>
#include <vector>

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
class CReportDisplay;
class CPageInfoVec;
class CStr;
class CReportDataItemView;
class CLayoutLink;
class CLayoutSupport;

using PCReportDataItemView = std::shared_ptr<CReportDataItemView>;

// ================================================================================================
/// The infomation needed by a CReportDataItemView about how to show on a page.
struct SPageInfo
{
    uint_32 uPageIndex = 0;     // relative to it's first page (0).
    uint_32 uItemIndexMin = 0;  // index of the first item to be displayed (0 relative)
    uint_32 uItemIndexMax = 0;  // index of the last item to be displayed  (0 relative)
    SRect sRect = {};           // area ocupied by item in device untis
    bool bIsFirstPage = false;  // is this the first page the item appears on
};

// ================================================================================================
/// The information about all the pages for a CReportDataItemView
class CPageInfoVec final : public std::vector<SPageInfo>
{
public:
    /// @return the page info for the requested page.
    const SPageInfo* GetPageInfo(uint_32 uPageIndex) const;
    /// @return the page info for last page that this item has content.
    const SPageInfo* GetLastPageInfo() const;
    /// @return where this item will be shown on the requested page.
    SRect GetPageRect(uint_32 uPageIndex) const;
    /// @return true if the item is visible on the requested page.
    bool IsVisibleOnPage(uint_32 uPageIndex) const;
};

// ================================================================================================
/// Support functionality for the CReportDataItemView class to handle the layout of report items for
/// the view.
class CLayoutLink final
{
    friend CLayoutSupport;
    enum ELayoutDir
    {
        ceLayoutLeft = 0,
        ceLayoutRight = 1,
        ceLayoutTop = 2,
        ceLayoutBottom = 3,
        ceLayoutAbsolute,
    };

public:
    CLayoutLink(CReportDataItemView* pParent);
    CLayoutLink(CLayoutLink&&) = default;
    CLayoutLink(const CLayoutLink&) = delete;
    ~CLayoutLink() = default;
    CLayoutLink& operator=(const CLayoutLink&) = delete;

    bool IsValid() const { return m_bIsValid; }
    bool IsLayoutAbsolute() const;
    bool IsLayoutRight() const { return (m_eHorzLayoutType == ceLayoutRight); }
    bool IsLayoutBottom() const { return (m_eVertLayoutType == ceLayoutBottom); }

    CReportDataItemView* GetLayoutItemLeft() const { return m_apItemPtr[ceLayoutLeft]; }
    CReportDataItemView* GetLayoutItemRight() const { return m_apItemPtr[ceLayoutRight]; }
    CReportDataItemView* GetLayoutItemTop() const { return m_apItemPtr[ceLayoutTop]; }
    CReportDataItemView* GetLayoutItemBottom() const { return m_apItemPtr[ceLayoutBottom]; }

    // Return position in units of 1/100 th of an Inch.
    int_32 GetLayoutPosMostBottom(const CLayoutSupport& rLayout,
        ELayoutDir eDir = ceLayoutBottom) const;

    // Return position in units of 1/100 th of an Inch.
    int_32 GetLayoutPosMostTop(const CLayoutSupport& rLayout, ELayoutDir eDir = ceLayoutTop) const;

    // Return pointer to Report View Item the has the longest vertical print
    // extent in a level (items connected left and reight of each other).
    CReportDataItemView* GetPrintLayoutLongestItem(const CLayoutSupport& rLayout,
        ELayoutDir eDir = ceLayoutBottom) const;

private:
    std::array<CReportDataItemView*, 4> m_apItemPtr;
    CReportDataItemView* m_pParent;

    bool m_bIsValid;

    ELayoutDir m_eHorzLayoutType;
    ELayoutDir m_eVertLayoutType;
};

// ================================================================================================
/// The class that keeps all the layout links for a report page.
class CLayoutSupport final
{
public:
    using VPCReportDataItemView = std::vector<PCReportDataItemView>;
    using CLayoutItemSet = std::set<CReportDataItemView*>;

    CLayoutSupport() = default;

    const CLayoutLink& GetLink(CReportDataItemView* pView) const;
    CLayoutLink& GetLink(CReportDataItemView* pView);

    uint_32 GetPrintPageCount() const;
    const SRect& GetPageRect() const;
    const SRect& GetWorkRect() const;

    void SetLayoutAbsolute(CReportDataItemView* pView);
    void SetLayoutRight(CReportDataItemView* pView, CReportDataItemView* pRight);
    void SetLayoutBottom(CReportDataItemView* pView, CReportDataItemView* pBottom);

    /// Initialize the size of the printable area on the output page (in printer device units).
    void SetPrintPageRect(CReportDisplay& rDisp,
        const SRect& sPageRect,
        CReportDataItemView* pHeaderView);

    void LayoutReportItems(VPCReportDataItemView& vItemViews, SSize sPageWndSize, bool fAllowStrech);
    uint_32 LayoutPrintItems(VPCReportDataItemView& vItemViews,
        std::shared_ptr<CReportDisplay> pDisplay);

    ///  This sets up the links between the the view items whose position depends on each other.
    void UpdateLayoutLinks(VPCReportDataItemView& vItemViews);

    void SetPageInfo(CReportDataItemBase* pItem, const CPageInfoVec& rPageInfo);
    const CPageInfoVec& GetPageInfo(CReportDataItemBase* pItem) const;

private:
    /// Get the view associated with the item.
    PCReportDataItemView GetReportViewItem(CReportDataItemBase* pDataItem,
        VPCReportDataItemView& vItemViews);

    std::map<CReportDataItemView*, CLayoutLink> m_links;
    std::map<CReportDataItemBase*, CPageInfoVec> m_pageInfo;

    SRect m_sPageRect = SRect{0, 0, 0, 0};  /// size of printable area on page (after margins)
    SRect m_sWorkRect = SRect{0, 0, 0, 0};  /// size of work area on a page (after margin and header)
    uint_32 m_uPrintPageCount = 0;          /// total pages to print for this item.
};

// ================================================================================================
/// The base layout algorithm.  This handles the sizing and relative positioning of the report
/// items through a CLayoutSupport object.
///
/// It is specialized for screen and printer layouts.
///
/// The expected usage is to create the derived type and run the algorithm:
///    CLayoutAlgoXXX(...).LayoutItems().
class CLayoutAlgo
{
public:
    using VPCReportDataItemView = std::vector<PCReportDataItemView>;
    using CLayoutItemSet = std::set<CReportDataItemView*>;
    virtual ~CLayoutAlgo() = default;

    /// Run the layout algorithm to update the layout support object on the views that were passed
    /// on construction.
    uint_32 LayoutItems();

protected:
    /// Create the basic algorithm object.  Once it is created the LayoutItems() should be called to
    /// do the layout.
    CLayoutAlgo(VPCReportDataItemView& rvItemViews, CLayoutSupport& rLayout);

    /// The layout object that is filled in with this algorithm.
    CLayoutSupport& m_rLayout;

private:
    /// Override if the layout handles a single item layout specially.  Returning true means that no
    /// further processing is to be done.
    virtual bool DoHandleSingleItem(CReportDataItemView* pItem);
    /// Override in screen layout to fill a single graph window to the available space.
    virtual void DoStretchToWindow(CReportDataItemView* pItem){};
    /// Override in print layout to make sure a single item is not bumped out of the first page.
    virtual void DoSingleBody(CReportDataItemView* pItem){};
    /// Override in print layout to calculate the last page of the item.
    virtual uint_32 DoComputeLastPage(CReportDataItemView* pItem) { return 0; }

    /// Function which is called []recursively to position a set of items in a view
    /// relative to one another.  The first call to the function should pass in
    /// two empty linked lists. This function will adjust the lists as it works.
    /// @param pItem is the current item to position.
    uint_32 LayoutItem(CReportDataItemView* pItem);
    void StretchToWindow(CReportDataItemView* pItem);

    VPCReportDataItemView& m_rvItemViews;
    /// The map  of items which have been completed.
    CLayoutItemSet m_Completed;
    /// The map of items which have started but NOT completed.
    CLayoutItemSet m_Pending;

    /// The location of the bottom of the header that appears on every page.
    uint_32 m_ulHeaderBottom = 0;

    DECLARE_MSG()
};

// ================================================================================================
/// The specialization of the layout algorithm to handle the special cases of screen layout.
class CLayoutAlgoScreen : public CLayoutAlgo
{
public:
    /// @param fAllowStrech is if stretching of the window is allowed.
    /// @param parentWndSize is the parent window size.
    CLayoutAlgoScreen(VPCReportDataItemView& rvItemViews,
        CLayoutSupport& rLayout,
        bool fAllowStrech,
        SSize sPageWndSize);

private:
    virtual bool DoHandleSingleItem(CReportDataItemView* pItem) override;
    virtual void DoStretchToWindow(CReportDataItemView* pItem) override;

    bool m_fAllowStrech;
    SSize m_PageWndSize;
};

// ================================================================================================
/// The specialization of the layout algorithm to handle the special cases of print layout.
class CLayoutAlgoPrint : public CLayoutAlgo
{
public:
    CLayoutAlgoPrint(VPCReportDataItemView& rvItemView,
        CLayoutSupport& rLayout,
        std::shared_ptr<CReportDisplay> pDisplay);

private:
    virtual void DoSingleBody(CReportDataItemView* pItem) override;
    virtual uint_32 DoComputeLastPage(CReportDataItemView* pItem) override;

    std::shared_ptr<CReportDisplay> m_pDisplay;
};

// ================================================================================================
/// The base class for all item views.
///
/// Each view has a stack of displays.  The top display is used for the drawing and support.  The
/// reason for the stack is that the same view for the screen could be used for printing as well. (I
/// don't think this is still used, but it is hard to confirm.)
class CReportDataItemView : public TUpdateViewBase<CReportDataItemBase>
{
public:
    class CDisplayContext;

    CReportDataItemView() = default;
    CReportDataItemView(const CReportDataItemView&) = delete;
    virtual ~CReportDataItemView() = default;
    CReportDataItemView& operator=(const CReportDataItemView&) = delete;

    /// Given the printer DC, the page index (based on what was returned by
    /// ComputePrintPageCount()), and a printable rect (for the specified page).  It will print the
    /// contents (replicating the header if needed) and return the size of the printed rect on the
    /// specified page (in printer device units).
    virtual void PrintPartial(uint_32 uPageIndex, const CPageInfoVec& cPageInfo) = 0;

    virtual void DrawItem(const SRect& cLocation) = 0;

    virtual void AutoSize() {}

    CDisplayContext DisplayLock(std::shared_ptr<CReportDisplay> pDisplay);
    void PushDisplay(std::shared_ptr<CReportDisplay> pDisplay);
    void PopDisplay();
    CReportDisplay& GetDisplay() const;

    // Given a device to map the units to, a page rect, and a printable rect on the page. Calculate
    // the last page index the item will occupy. The page index is 0 relative and incriments to the
    // last page index the item will use.  If the item is splittable (table), it will return the
    // last page index the item occupies, taking into consideration any duplication of headers
    // required for this particular item.
    uint_32 ComputeLastPageIndex(CReportDisplay& rDisp,
        const SRect& sPageRect,
        const SRect& sPrintRect,
        CLayoutSupport& rLayout);
    /// This does the same thing as ComputeLastPageIndex, but never bumps the page number of the
    /// start.  It resizes the item so that it fits onto this page.
    uint_32 ComputeLastPageIndexNoBump(CReportDisplay& rDisp,
        const SRect& sPageRect,
        const SRect& sPrintRect,
        CLayoutSupport& rLayout);

    /// This is called to indicate that the size and position of this window should no longer be
    /// linked to the data item associated.
    void SetLocalPos();
    /// This is called to relink the data item's size and position information.
    void UnsetLocalPos();
    bool IsLocalPos() const { return m_fLocalRect; }

    /// @{ The frame rectangle (in 1/100 of an inch) for the data item.
    SRect GetFrameRect();

    inline int_32 GetFrameLeft() { return GetFrameRect().left; }
    inline int_32 GetFrameRight() { return GetFrameRect().right; }
    inline int_32 GetFrameTop() { return GetFrameRect().top; }
    inline int_32 GetFrameBottom() { return GetFrameRect().bottom; }
    inline int_32 GetFrameWidth() { return GetFrameRect().Width(); }
    inline int_32 GetFrameHeight() { return GetFrameRect().Height(); }

    void SetFrameRect(const SRect& rsRect, bool fUpdate = true);
    void SetFramePos(int_32 lPosX, int_32 lPosY, bool fUpdate = true);
    void SetFrameSize(int_32 lSizeX, int_32 lSizeY, bool fUpdate = true);
    /// @}

    /// @{
    /// The screen rectangle is the linked to the frame rectangle, but in screen coordinates, and
    /// the vertical offset adjustment.
    SRect GetScreenRect();

    void SetScreenRect(const SRect& rsRect, bool fUpdate = true);
    void SetScreenPos(int_32 lPosX, int_32 lPosY, bool fUpdate = false);
    void SetScreenSize(int_32 lSizeX, int_32 lSizeY, bool fUpdate = false);
    /// @}

    /// Returns the item's frame rect in printer units (does not include consideration for splitting
    /// accross multiple pages and replicating headers and things of the sort). Offset is scaled
    /// relative to on-screen placement and length takes into consideration actual size (if item was
    /// displayed as a scrollable item).
    void SetPrintRect(CReportDisplay& rDisp, const SRect& rsPrintRect);
    SRect GetPrintRect(CReportDisplay& rDisp);

protected:
    /// Sets the length of the data item to match the iDisplaySizeY length calculated with the
    /// currently active CReportDisplay.
    void SetDisplayLength(int iDisplaySizeY);

private:
    virtual CPageInfoVec CalculatePrintInfo(uint_32 uPageOffset,
        uint_32 uStartPageIndex,
        const SRect& sItemRect,
        const SRect& sPrintRect,
        CLayoutSupport& rLayout);

    std::stack<std::shared_ptr<CReportDisplay>> m_displayStack;

    /// If m_fLocalRect is true the size and position of this window is unlinked from the data
    /// object, and stored localy in m_sLocalFrameRect (in the ususal 1/100 inch units for frame
    /// rectangles.
    bool m_fLocalRect = false;
    SRect m_sLocalScreenRect;
};

// ================================================================================================
/// A context variable that locks the context for a particular item view.
class CReportDataItemView::CDisplayContext final
{
public:
    CDisplayContext(CReportDataItemView& rView, std::shared_ptr<CReportDisplay> pDisplay);
    CDisplayContext(CDisplayContext&& rhs) : m_pView(nullptr) { std::swap(m_pView, rhs.m_pView); }
    CDisplayContext(const CDisplayContext&) = delete;
    ~CDisplayContext();

private:
    CReportDataItemView* m_pView;
};

using PCReportDataItemView = std::shared_ptr<CReportDataItemView>;

// ================================================================================================
/// A helper base class that links a concrete CReportDataItemView class to concrete
/// CReportDataItemBase type.
template<typename T> class TReportDataItemView : public CReportDataItemView
{
public:
    /// Give a good name to the parameter type for items derived from specific instances.
    using TData = T;

    T* GetLinkToBase();
};

template<typename T> T* TReportDataItemView<T>::GetLinkToBase()
{
    auto pBase = CReportDataItemView::GetLinkToBase();
    auto pT = dynamic_cast<T*>(pBase);
    // Assert that if we have one, it is of the right type.
    ASSERT_ERROR((pBase && pT) || (!pBase && !pT));
    return pT;
}

// ================================================================================================
/// This is a CReportDisplay, but it forwards the requests to another diaplay that is contained in a
/// CReportDataItemView.  The usefulness of this is for nested report item views like in a report
/// header where the nested items need to use the display of the containing view.
class CReportDisplayFwd : public CReportDisplay
{
public:
    CReportDisplayFwd(CReportDataItemView& rView);
    virtual ~CReportDisplayFwd() = default;

    virtual std::pair<int, int> GetPixelsPerInch() const override;
    virtual bool IsRaster() const override;
    virtual int PrintingResOrWndWidth() const override;
    virtual bool IsPrintingBW() const override;
    virtual std::pair<int, int> GetPaperSize() const override;
    virtual uint_32 GetBackgroundColor() const override;

    virtual void SetBkModeTransparent() override;
    virtual void SetBkModeOpaque() override;
    virtual void SetBkColor(uint_32 uiColor) override;

    virtual int CalcFrameHeight() const override;
    virtual int CalcCYEdge() const override;
    virtual std::pair<int, int> CalcTextExtent(const CStr& szText) const override;
    virtual int CalcTextHeight(const CStr& szText, ETextAlign eAlign, int iWidth) const override;
    virtual int CalcTextNominalHeight() const override;
    virtual void TextOut(int x, int y, const CStr& szText) override;
    virtual void DrawTextWithBreak(const CStr& szText, ETextAlign eAlign, const SRect& cRect) override;
    virtual void DrawText(const CStr& szText, ETextAlign eAlign, const SRect& cRect) override;

    virtual void MoveTo(int x, int y) override;
    virtual void LineTo(int x, int y) override;
    virtual void Stroke() override;
    virtual void Ellipse(int x1, int y1, int x2, int y2) override;
    virtual void Polygon(const SPoint* lpPoints, int nCount) override;
    virtual void Rectangle(int x1, int y1, int x2, int y2) override;
    virtual void FillRect(uint_32 uColor, const SRect& cRect) override;

#if defined(MIC_VCMFC)
    virtual void PlayEnhMetaFile(HENHMETAFILE hemf, RECT* psRect) override;
    virtual void DrawBitmap(BITMAPINFO* psInfo, char* pBits, int x1, int y1, int x2, int y2) override;
#endif

private:
    virtual void DoPushFont(EFontType eType, uint_32 ulColor, int iOrientation) override;
    virtual void DoPopFont() override;
    virtual void DoPushPen(int iPen, int iWidth, uint_32 ulColor) override;
    virtual void DoPopPen() override;
    virtual void DoPushBrush(uint_32 ulColor) override;
    virtual void DoPushBrush(int iType, uint_32 ulColor) override;
    virtual void DoPushClipRgn(int x1, int y1, int x2, int y2) override;
    virtual void DoPopClipRgn() override;
    virtual void DoPushNullBrush() override;
    virtual void DoPopBrush() override;
    virtual void DoAllocateDisplay() override;
    virtual void DoFreeDisplay() override;

    CReportDataItemView& m_rView;
};

#endif // RSITEMVIEW_HPP
