#ifndef RSTYPEDATA_HPP
#define RSTYPEDATA_HPP
//-----------------------------------------------------------------------------
/// @file
/// Defined CReportPageData, a single tab in a CReportModel.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "rsdataset.hpp"
#include "rsitemdata.hpp"
#include "cstr.hpp"
#include "data-view-update.hpp"

//-----------------------------------------------------------------------------
// Forward References
//-----------------------------------------------------------------------------
class CReportTextData;
class CReportTableData;
class CReportHeaderData;
class CReportGraphData;
class CReportImageData;
class CReportPageData;

// ============================================================================
namespace NReportPageData
{
    std::string JsonSchema();
}

// ============================================================================
/// Defines a single page of a report.
///
/// It may print on more than one page.  It is simply a container of
/// CReportDataItemBase* objects that it created and stored.  It also has a
/// name that is used when multiple pages are shown in a single (tabbed)
/// window.
class CReportPageData final : public TUpdateDataBase<CReportPageData>
{
    /// @return the json value for @p rData ready to be inserted into @p cData.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CReportPageData& cData);
    friend void Serialize(CDataArchive& rAr, CReportPageData& rData);

public:
    using VDataItem = std::vector<std::unique_ptr<CReportDataItemBase>>;

    /// @{ Define iterator and const_iterator for this container to convert unique_ptr interface to
    /// raw pointer.  This is done so that the CReportPageData class can be iterated over like a
    /// container of pointers, not smart pointers.
    struct iterator : public VDataItem::iterator
    {
        iterator(VDataItem::iterator&& rIter);
        CReportDataItemBase* operator*();
    };
    struct const_iterator : public VDataItem::const_iterator
    {
        const_iterator(VDataItem::const_iterator&& rIter);
        const CReportDataItemBase* operator*();
    };
    /// @}

    CReportPageData(const CStr& szName);
    CReportPageData(const CReportPageData&) = delete;
    CReportPageData(CReportPageData&&) = default;
    ~CReportPageData() = default;
    CReportPageData& operator=(const CReportPageData&) = delete;
    CReportPageData& operator=(CReportPageData&&) = default;

    /// @return the name of the report.
    const CStr& GetName() const;

    /// @return a pointer to the Report Header Data Item to be used for
    /// page header presentation.
    CReportHeaderData* GetReportHeaderData();
    const CReportHeaderData* GetReportHeaderData() const;

    /// @{ Container interface for the CReprotpointers this stores.
    void clear();
    bool empty() const;
    const_iterator begin() const;
    const_iterator end() const;
    iterator begin();
    iterator end();
    /// @}

    /// @{ Add a specific report items to the page.
    CReportTextData* AddTextData();
    CReportTableData* AddTableData();
    CReportHeaderData* AddHeaderData();
    CReportHeaderData* AddHeaderData(size_t uiColumns);
    CReportGraphData* AddGraphData(CReportDataItemBase* pPrevItem = nullptr,
        int iLeft = 75,
        int iRight = 775);
    CReportImageData* AddImageData();
    /// @}

    /// @returns a sorted list of the report items based on vertical positioning (for export).
    void OrderItems() const;

    /// Sets flag indicating report data has been truncated due to max window height
    /// limit in the preview report frame.
    void SetDataTruncatedFlag();
    /// @return true if this report page has truncated data (marked by the above method).
    bool IsDataTruncated() const;

private:
    CStr m_szName;                       ///< The name of the page.
    VDataItem m_DataItems;               ///< The CReportDataItems on this page.
    CReportHeaderData* m_pFirstHdrItem;  ///< The header item (if any) of this page.
    bool m_fIsDataTruncated = false;     ///< Indicates whether any report data
                                         ///< have been truncated.
};

/// Serialize the contents of the CReportPageData object.
void Serialize(CDataArchive& rAr, CReportPageData& rData);

#endif  // RSTYPEDATA_HPP
