#ifndef RSPAGEVIEW_HPP
#define RSPAGEVIEW_HPP
//-------------------------------------------------------------------------------------------------
/// @file The view associated with CReportPageData.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "rspagedata.hpp"
#include "rsheaderview.hpp"

// ================================================================================================
/// The class that holds all the views for the items on a single report page.
class CReportPageView final : public TUpdateViewBase<CReportPageData>
{
public:
    enum EType
    {
        ceHeader,
        ceBody,
        ceFull
    };

    CReportPageView(EType eType = ceFull);
    CReportPageView(const CReportPageView&) = delete;
    CReportPageView(CReportPageView&&) = default;
    virtual ~CReportPageView() = default;
    CReportPageView& operator=(const CReportPageView&) = delete;

    virtual void OnBaseChanged(unsigned int uiUpdateFlag) override;

    bool IsBody() const { return m_ceType == ceBody; }
    bool IsHeader() const { return m_ceType == ceHeader; }
    std::shared_ptr<CReportHeaderView> GetHeaderView() { return m_pHeaderView; }

    PCReportDataItemView GetViewFromData(CReportDataItemBase* pData);

    void UpdateLayout(SSize pageWndSize, bool fAllowStrech);

    // Initialize the size of the printable area on the output page (in printer device units).
    void SetPrintPageRect(CReportDisplay& rDisp, const SRect& sPageRect);

    // return the number of pages required for print job.
    uint_32 DoPrintLayout(std::shared_ptr<CReportDisplay> pDisplay);

    uint_32 GetPrintPageCount() const;

    /// Given the printer DC and the page index (based on what was returned by DoPrintLayout()). It
    /// will print the contents (replicating the header if needed) and return the size of the
    /// printed rect on the specified page.
    SRect PrintPage(std::shared_ptr<CReportDisplay> pDisplay,
        uint_32 uiPageIndex,
        uint_32 uiPageNumber,
        uint_32 uiNumPages);

private:
    const EType m_ceType;  //  The style of the view.

    std::vector<PCReportDataItemView> m_cItemViews;

    /// Pointer to Report Header Item in m_cItemViews.
    std::shared_ptr<CReportHeaderView> m_pHeaderView;

    /// The data for handling the layout of the contained items.
    CLayoutSupport m_layout;
};

// ================================================================================================
/// The generic implementation of the pdf output.
class CPDFReports
{
public:
    using CReportViewDeque = std::deque<std::shared_ptr<CReportPageView>>;

    CPDFReports(CReportModel& rModel);
    virtual ~CPDFReports() = default;

    void GenerateFile(const CStr& szFileName);

private:
    uint_32 CalculateNumberOfPagesToPrint(CReportViewDeque& aReportView,
        std::shared_ptr<CReportDisplay> pDisp);

    virtual bool IsValid() = 0;
    virtual void SaveToFile(const CStr& szFileName) = 0;
    /// Called as part of the print layout mechanism if the there is no room under the header for
    /// the body. Overwritten implementations should returns true if this should cancel the report
    /// generation.  And if so notify the user.
    virtual bool WarnHeaderCancel() = 0;
    /// Called during printing to allow derived classes to setup the first page of the print.
    virtual std::shared_ptr<CReportDisplay> GetFirstPageDisplay() = 0;
    /// Called during printing to allow derived classes to close the previous page and start the
    /// next page of the print.
    virtual std::shared_ptr<CReportDisplay> GetNextPageDisplay(uint_32 uiIndex) = 0;
    /// Called during printing to cleanup after the last page has been generated.
    virtual void LastPageComplete() = 0;

    CReportModel& m_rModel;
};

#endif // RSPAGEVIEW_HPP
