#ifndef RSPRINT_HPP
#define RSPRINT_HPP
//-------------------------------------------------------------------------------------------------
/// @file Support for the print/export functionality.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "datacontext.hpp"
#include "rstablehelper.hpp"

// ================================================================================================
/// Helper class to print out data items given CDataContext.  This handles the names and relevancy
/// of each data object.
class CReportPrintTable final
{
public:
    /// Create the object for the @p rInfo data, and produce the output into the @p rTable.
    CReportPrintTable(CReportPageData* pReport, uint_16 uiTitleID, bool fShowLabels);
    CReportPrintTable(const CSubsetOptions* pSSOpt,
        CReportPageData* pReport,
        uint_16 uiTitleID,
        bool fShowLabels);
    CReportPrintTable(CReportPrintTable&&) = default;

    /// Position this table below the specified previous item.
    /// @return the item for this table for inclusion in the item chain.
    CReportDataItemBase* SetLayoutBottom(CReportDataItemBase* pPrevItem);

    /// Show the string as a subheading only.
    void AddSubHeading(uint_16 uiIDS);

    /// Add a blank line as a separator.
    void AddBlank();

    /// Backward compatibility for complicated things.
    CReportOptionsTable& GetOptionsTable();

    /// Add all the known, relevant members of the data object.
    template<typename T> void Show(const T& cData);

    /// Add the data item into the table if it is relevant.
    template<typename T, typename S> void Show(S T::*pMember, const T& cData);

private:
    /// Add the specific supported types, properly formatted into the table.
    template<typename T> void Add(const NDataContext::SMemberID& cMemberID, const T& cData);

    template<typename T> void ShowVoid(const NDataContext::SMemberID& cMemberID, const T& cData);

    /// The table to print into.
    CReportOptionsTable m_Table;

    /// The subset options that decide on used if.
    const CSubsetOptions* m_pSSOpt;
};

// ================================================================================================
template<typename T>
void CReportPrintTable::Add(const NDataContext::SMemberID& cMemberID, const T& cData)
{
    using namespace NDataContext;
    if(IsScalarVoid(cMemberID))
    {
        auto szPrompt = GetPromptVoid(cMemberID);
        if(!szPrompt.IsEmpty())
            m_Table.AddItem(szPrompt, GetUnitsStrVoidAgg(cMemberID, const_cast<T&>(cData)));
    }
    else
        ASSERT_ERROR(false);
}

template<typename T> void CReportPrintTable::Show(const T& cData)
{
    using namespace NDataContext;

    auto type = std::type_index(typeid(T));
    if(!HasMembers(type))
        return;

    for(auto& rMember : GetMembers(type))
        ShowVoid(rMember.id, cData);
}

template<typename T, typename S> void CReportPrintTable::Show(S T::*pMember, const T& cData)
{
    ShowVoid(NDataContext::SMemberID(pMember), cData);
}

template<typename T>
void CReportPrintTable::ShowVoid(const NDataContext::SMemberID& cMemberID, const T& cData)
{
    using namespace NDataContext;

    auto fUsed = !m_pSSOpt || IsUsedVoid(cMemberID, *m_pSSOpt);
    if(fUsed && IsRelevantVoid(cMemberID, static_cast<const void*>(&cData)))
        Add(cMemberID, cData);
}

#endif // RSPRINT_HPP
