#ifndef REPORTMODEL_HPP
#define REPORTMODEL_HPP
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "rspagedata.hpp"
#include "rsdataset.hpp"

// ================================================================================================
namespace NReportModel
{
    std::string JsonSchema();
}

// ================================================================================================
/// The report object that contains all the pages of a multipage report.
///
/// To Use, just create with the name of the report.  Call SetDataTruncatedFlag() and
/// SetPreviewOnlyFlag() if necessary.  Finally, use AddReportType() to add the individual pages.
///
/// To access the pages use the usual container methods.
class CReportModel final
{
    /// @return the json value for @p rData ready to be inserted into @p cData.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CReportModel& cData);
    friend void Serialize(CDataArchive& rAr, CReportModel& rData);

public:
    using VPageData = std::vector<std::shared_ptr<CReportPageData>>;

    /// @{ Define iterator and const_iterator for this container to convert unique_ptr interface to
    /// raw pointer.  This is done so that the CReportModel class can be iterated over like a
    /// container of pointers, not smart pointers.
    struct iterator : public VPageData::iterator
    {
        iterator(VPageData::iterator&& rIter);
        CReportPageData* operator*();
    };
    struct const_iterator : public VPageData::const_iterator
    {
        const_iterator(VPageData::const_iterator&& rIter);
        const CReportPageData* operator*();
    };
    /// @}

    CReportModel(const CStr& szTitle);
    ~CReportModel() = default;
    CReportModel(const CReportModel&) = delete;
    CReportModel(CReportModel&&) = default;
    CReportModel& operator=(const CReportModel&) = delete;
    CReportModel& operator=(CReportModel&&) = default;

    /// @returns the title of the report model.
    const CStr& GetTitle() const;

    /// @returns True if any of the report pages have data which has been truncated.
    bool IsDataTruncated() const;
    /// @returns True if the preview-only flag is set; indicates that the report should
    /// not be printable/exportable and that "Preview" should appear in the report header.
    bool IsPreviewOnly() const;

    /// Sets the preview-only flag. Once set, it cannot be cleared.
    void SetPreviewOnlyFlag();

    /// Create a CReportPageData, insert it into linked list
    /// @return a pointer to item.
    /// NOTE: A unique string is required when adding a report page.  If
    /// there are duplicate report name strings, only the last page gets shown
    std::shared_ptr<CReportPageData> AddPage(const CStr& cName);

    /// Convert the whole model to a json string.
    std::string AsJson() const;

    /// @{ Container interface for the CReportPageData pointers this stores.
    const_iterator begin() const;
    const_iterator end() const;
    iterator begin();
    iterator end();
    bool empty() const;
    /// @}

    // Get the report pointer by name.
    [[deprecated]] CReportPageData* GetReportType(const CStr& cName);

private:
    VPageData m_cReportTypeData;  ///< The list of report pages.

    CStr m_cTitle;            ///< Title of report window. Usually the file name.
    bool m_fIsPreviewOnly;    ///< Indicates if saving of the generated report is allowed.
                              ///< Also adds "Preview" to the report header.

    DECLARE_MSG()
};

/// Serialize the contents of the CReportPageData object.
void Serialize(CDataArchive& rAr, CReportModel& rData);

#endif  // #ifndef REPORTMODEL_HPP
