#ifndef RSSURFACEDATA_HPP
#define RSSURFACEDATA_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "rsdataset.hpp"

// ============================================================================
/// This hold the information for a surface displayed on a three axis graph.
///
/// The surface is defined by a set of horizontal data values.  There are
/// multiple vertical data sets that each correspond to the same horizontal
/// values.  The depth values for each vertical set is either just its number
/// (insertion order), or the depth data set added.
class CRSGraphSurfaceData final
{
public:
    using VPolygonVertex = std::vector<std::pair<float_64, float_64>>;

    /// Construct a surface linked to the three axes
    CRSGraphSurfaceData(int iXAxisID, int iYAxisID, int iZAxisID);
    ~CRSGraphSurfaceData() = default;

    ///  Set the horizontal data set.  These are the x-coordinates for each
    ///  vertical data set added.
    void SetHorzDataSet(PCDataSetColumn pDataSet);

    /// Add a vertical data set that corresponds to the one horizontal.
    void AppendVertDataSet(PCDataSetColumn pDataSet);
    /// Delete all the vertical data sets so they can be restarted.
    void ClearVertData();

    /// Set the depth data set. There should be an entry for every vertical
    /// data set added.  If this is not given, it is assumes that the depth set
    /// is just an ordinal count of the vertical sets.
    void SetDepthDataSet(PCDataSetColumn pDataSet);

    /// @return true if contours have been set.
    bool HasContours() const;
    /// Clear all contour information.
    void ClearContours();
    /// Add a closed contour with specified layer height to the list of contours.
    void AddContour(float_64 dLayer, const VPolygonVertex& vPolgonVertex);
    /// @return the list of layers values.
    const std::vector<float_64>& GetContourLayers() const;
    /// @return the list of contours.
    const std::vector<VPolygonVertex>& GetContours() const;

    /// @name Data accessors needed by the view.
    /// @{

    //  The axis links
    int GetXAxisID() const { return m_iXAxisID; }
    int GetYAxisID() const { return m_iYAxisID; }
    int GetZAxisID() const { return m_iZAxisID; }

    auto GetHorzDataSet() const { return m_pXDataSet; }
    PCDataSetColumn GetHorzDataSet() { return m_pXDataSet; }

    //  The vertical data sets.
    size_t GetVertDataSetNum() const { return m_apYDataSet.size(); }
    std::shared_ptr<const CDataSetColumn> GetVertDataSet(int iNum) const;
    PCDataSetColumn GetVertDataSet(int iNum);

    //  The depth data set.
    auto GetDepthDataSet() const { return m_pZDataSet; }
    PCDataSetColumn GetDepthDataSet() { return m_pZDataSet; }

    /// @}

private:
    ///  The associated axes.
    int m_iXAxisID, m_iYAxisID, m_iZAxisID;

    ///  The X-Axis dataset.
    PCDataSetColumn m_pXDataSet;

    ///  The table of Y-Axis datasets.
    std::vector<PCDataSetColumn> m_apYDataSet;

    //  The depth values for the Y datasets if present.
    PCDataSetColumn m_pZDataSet;

    /// Contour information which is added by calls to AddContour() and removed
    /// with a call to ClearContours().  The contour vector is a list of vertices
    /// in the x-z plane which begin and end with the same vertex. The layer height
    /// is the average height for the dataset in the contour region. Both vectors
    /// have the same length.
    std::vector<VPolygonVertex> m_vContour;
    std::vector<float_64> m_vdLayer;

    DECLARE_MSG()
};

#endif  // RSSURFACEDATA_HPP
