#ifndef RSTABLEDATA_HPP
#define RSTABLEDATA_HPP
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "rsdataset.hpp"
#include "rsitemdata.hpp"
#include "formattr.hpp"
#include "unitscontext.hpp"

#include <math.h>

//-----------------------------------------------------------------------------
// Forward Declarations
//-----------------------------------------------------------------------------
namespace NTableColumnInfo
{
    std::string JsonSchema();
}

namespace NReportTableData
{
    std::string JsonSchema();
}

//-----------------------------------------------------------------------------
/// Data object that holds the information pertinent to a column of data for a
/// table.
///
/// There are three functional areas for objects of this class: The data of the
/// column, including the header; The parameters for how the column should be
/// displayed; And the Units and formatting options for numerical columns.
///
/// Note that if units are used, the unit string is automatically added to the
/// report header, and the numerical values are automatically converted.  Note
/// that units settings include both CUnits, and the SetInverter() that is
/// applied to the whole column (for coarser/finer type columns).
struct CTableColumnInfo
{
    /// @return the json value for @p rData ready to be inserted into @p cData.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CTableColumnInfo& cData);
    friend void Serialize(CDataArchive& rAr, CTableColumnInfo& rData);

public:
    /// Construct linked to the data.
    CTableColumnInfo(PCDataSetColumn pDataSet);

    /// @name Access to the Column Data
    /// @{

    /// Set the header string (without the units).
    void SetHeaderString(const CUnitsStr& szHeader) { m_szHeaderString = szHeader; }

    /// Calculates the header string by evaluating the CUnitsStr with the
    /// current Context (if any).  If @p fAddUnits is true (default) it will
    /// also add the units string to the column header.
    CStr HeaderString(bool fAddUnits = true) const;

    /// constant access to the CDataSetColumn data.
    std::shared_ptr<const CDataSetColumn> GetDataSetPtr() const { return m_pDataSetColumn; }
    /// access to the CDataSetColumn data.
    PCDataSetColumn GetDataSetPtr() { return m_pDataSetColumn; }

    /// Get the string item, properly formatted to the @p uItemIndex item.
    CStr GetItemString(const size_t uItemIndex) const;
    /// Get the numeric value of the item in display units.
    double GetItemValue(const size_t uItemIndex) const;

    /// @{
    /// Change the contents of an indexed item in the associated
    /// DataSet.
    void SetItem(const size_t uItemIndex, const double dNumber);
    void SetItem(const size_t uItemIndex, const CStr& szStr);
    void SetItem(const size_t uItemIndex, const CUnitsStr& szStr);
    /// @}

    size_t GetItemCount() const { return m_pDataSetColumn->GetItemCount(); }
    /// @}

    /// @name Display Parameter Access
    /// @{

    /// Set the width of column in 1/100 Inches.
    void SetColumnWidth(size_t uColumnWidth) { m_uColumnWidth = uColumnWidth; }
    /// Access the width of column in 1/100 Inches.
    size_t GetColumnWidth() const { return m_uColumnWidth; }

    // Should the column be visible
    bool IsVisible() const { return m_fVisible ? true : false; }
    void SetVisible(bool f) { m_fVisible = f; }

    // Is column width sizing allowed via GUI mechanisms.
    bool IsSizeable() const { return m_fSizeable ? true : false; }
    void Sizeable(bool f) { m_fSizeable = f; }

    void AlignLeft() { m_eAlign = ceAlignLeft; }
    bool IsAlignLeft() const { return m_eAlign == ceAlignLeft ? true : false; }

    void AlignRight() { m_eAlign = ceAlignRight; }
    bool IsAlignRight() const { return m_eAlign == ceAlignRight ? true : false; }

    void AlignCenter() { m_eAlign = ceAlignCenter; }
    bool IsAlignCenter() const { return m_eAlign == ceAlignCenter ? true : false; }
    /// @}

    /// @name Unit and Formatting Options
    /// @{

    /// Set the format of the column to match the @p rcFmt.  This also enables
    /// dynamic units for the table so that the column header's unit string may
    /// change when the global unit selections change.
    void SetFormat(const CQuantityFormatter& rcFmt);

    /// Sets up the column to be inverted in @p pUnitSel has a current
    /// selection of @p iSelForInvert.  The inversion is the @p dInvertValue -
    /// internal value of the column.
    void SetInverter(std::shared_ptr<CUnitSelection> pUnitSel, int iSelForInvert, double dInvertValue);

    /// Set the units context that will be used for the header string.
    void SetUnitsContext(std::shared_ptr<CUnitsContext> pUnitsContext);

    /// Access to the currently selected units string.
    CStr ShortUnits() const { return m_OutputFormat.ShortUnits(); }

    /// Output format is in exponential format as output by printf( "%.nE" )
    [[deprecated]] void SetExpFormat(size_t uDeciPoint);

    /// Output format has fixed number of units after decimal point.
    [[deprecated]] void SetFixedFormat(size_t uDeciPoint);

    /// Output format is in fixed or exponenital depending on the value.
    [[deprecated]] void SetFixedOrExpFormat(size_t uDeciPoint,
        double dExpMin,
        double dExpMax,
        size_t uExpPrecision);
    /// @}

private:
    enum EAlign : uint_16
    {
        ceAlignLeft,
        ceAlignCenter,
        ceAlignRight
    };

    /// How to align the column.
    EAlign m_eAlign;
    /// Should the column be shown.
    bool m_fVisible;
    /// Is the column allowed to be resized?
    bool m_fSizeable;
    /// Width of column in 1/100 Inches.
    size_t m_uColumnWidth;

    /// String to be used for the column header.
    CUnitsStr m_szHeaderString;

    /// The formatter to use for converting and displaying numerical values.
    CQuantityFormatter m_OutputFormat;

    /// The inversion data.
    std::shared_ptr<CUnitSelection> m_pInvertUniSel;
    int m_iSelForInvert;
    double m_dInvertValue;

    // The CUnitsContext used for the display in the header.
    std::shared_ptr<CUnitsContext> m_pUnitsContext;

    // Pointer to DataSet to use for the contents of this column.
    PCDataSetColumn m_pDataSetColumn;
};

/// Serialize the CTableColumnInfo and data.
void Serialize(CDataArchive& rAr, CTableColumnInfo& rData);

//=============================================================================
/// This stores the information for a table in the report system.  Note that
/// this should not be used directly, but the CReportStdTable and
/// CReportOptionsTable helper functions should be used instead.
///
/// Tables are units aware (as are the columns), so when the units are setup
/// properly, the resulting output will be updated when unit selections
/// change.
class CReportTableData final : public CReportDataItemBase
{
    friend class CReportTableWin;

    /// @return the json value for @p rData ready to be inserted into @p cData.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CReportTableData& cData);

public:
    /// @returns true if the @p dValue indicates the spot should be left empty.
    static bool DontShowNum(double dValue);

    /// Create a new table object.
    CReportTableData();

    CReportTableData(const CReportTableData&) = delete;
    CReportTableData& operator=(const CReportTableData& cArg) = delete;

    ~CReportTableData() = default;

    void SerializeData(CDataArchive& rAr, const VDataItem& cvDataItems) override;

#if defined(_DEBUG)
    virtual CStr DbgName() const override;
#endif

    /// Resets all the data and parameters for the table.
    void Reset();

    /// Removes all the data from the columns, but keeps the metadata.
    void DeleteAllColumnData();

    /// Setup the title.
    void SetTitle(const CUnitsStr& rcszStr);
    void SetTitle(const CStr& rcszStr) { SetTitle(CUnitsStr(rcszStr)); }

    /// Get access to the title, this is formatted through the set or default
    /// units context.
    CStr GetTitle() const;

    /// Set the units context for the table and all current and future
    /// columns.
    void SetUnitsContext(std::shared_ptr<CUnitsContext> pUnitsContext);

    /// Is this a table that has parameter name and value columns.
    bool IsKeyValueTable() const;

    /// @name Column Data Access
    /// @{

    /// Remove all columns from table.
    void Clear();

    // Set the header string for the specified column.
    void SetColumnHeader(const size_t ulCol, const CStr& szHeader);

    // Add a column at the end of the list.
    // - Set the column header to the specified header string (if not nullptr).
    // - Return a pointer to the new column.
    // - NOTE: At this point, the column is NOT linked to a DataSet source.
    std::shared_ptr<CTableColumnInfo> AddColumn(const CUnitsStr& szColumnHeader,
        PCDataSetColumn pDataSet = nullptr);
    std::shared_ptr<CTableColumnInfo> AddColumn(const CStr& szColumnHeader,
        PCDataSetColumn pDataSet = nullptr);

    // Return the number of columns in Table.
    size_t GetColumnCount() const { return m_Column.size(); }

    // Return a pointer to the indexed CTableColumnInfo structure containing
    // formatting and column data information.
    std::shared_ptr<const CTableColumnInfo> GetIndexedColumnInfo(const size_t uColumnIndex) const;
    std::shared_ptr<CTableColumnInfo> GetIndexedColumnInfo(const size_t uColumnIndex);

    std::shared_ptr<CTableColumnInfo> GetVisibleIndexedColumnInfo(int lColumnIndex) const;
    /// @}

    /// @name Display parameters.
    /// @{

    /// The style of font to use in the table data.
    void SetTableFont(EFontType eType) { m_eTableFontType = eType; }
    EFontType GetTableFont() const { return m_eTableFontType; }

    /// Should the column header text be displayed?
    void ShowHeader(bool f);
    bool IsHeaderVisible() const { return m_fHdrVisible ? true : false; }

    /// Should the table title be visible.
    void ShowTitle(bool f);
    bool IsTitleVisible() const { return m_fShowTitle ? true : false; }

    // Set page width to use for column snaking in units of 1/100th of an inch.
    // Setting the width does NOT enable column snaking (see EnableColumnSnaking(...) )
    void SetColumnSnakingPageWidth(int lPageWidth);
    int GetColumnSnakingPageWidth() const { return m_lSnakingPageWidth; }

    void EnableColumnSnaking(bool b);
    bool IsColumnSnaking() const { return m_fColumnSnaking ? true : false; }

    /// @}

private:
    /// Show column headers?
    bool m_fHdrVisible;
    /// Shoe report title?
    bool m_fShowTitle;
    /// Should it show snaking columns?
    bool m_fColumnSnaking;

    /// Font to be used for the body of the table.
    EFontType m_eTableFontType;

    /// The Title information if for the table.
    CUnitsStr m_szTitle;

    /// The width of the snaked columns.
    int m_lSnakingPageWidth;

    /// The vector of columns that is the table.
    std::vector<std::shared_ptr<CTableColumnInfo>> m_Column;

    /// The CUnitsContext used for the display in the header.  (and to pass
    /// onto future columns.
    std::shared_ptr<CUnitsContext> m_pUnitsContext;
};

#endif  // RSTABLEDATA_HPP
