#ifndef RSTBLHLP_HPP
#define RSTBLHLP_HPP
//-----------------------------------------------------------------------------
//
// Library:     mic-cpp-lib
// Description: Classes and functions for commonly used report items.
//
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "rsdataset.hpp"
#include "rstabledata.hpp"
#include "numvect.hpp"
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class CReportPageData;
class CReportDataItemBase;
class CReportModel;
class CReportTableData;
struct CTableColumnInfo;
class CNumericFormat;
template<typename T> class TFormattedNumeric;

//-----------------------------------------------------------------------------
/// This is the way to create CReportTableData that is setup as columns of
/// quantities.  The table handles the unit conversion and display.
///
/// To use, first create the table with the right number of columns and title.
/// You can create this as part of a CReportPageData or stand alone.  You also
/// have the option of choosing to not show the column headers.  You can set
/// the Units context through the SetUnitsContext() method.  This is used in
/// the column headers and title.  Typically after creation the
/// SetLayoutBottom() is called to set the display order.
///
/// Then for each column, set the title (SetTitle()), and setup the column
/// parameters through the GetColInfo() method.  This is how the units and
/// possible inversion is set.  See the CTableColumnInfo methods.
///
/// Then you can add the data either one at a time or set the whole set with
/// the AddItem() methods.
///
/// Finally, if this is a stand alone table with an associated view, you can
/// set the parameters for that configuration.
class CReportStdTable
{
public:
    using CPtr = std::shared_ptr<CReportStdTable>;

    /// Creates the table with title @p szTitle and @p uiCols columns.
    CReportStdTable(CReportPageData* pReport,
        const CUnitsStr& szTitle,
        uint_16 uiCols,
        int_32 iLeftEdge = 75,
        int_32 iRightEdge = 775);
    /// Constructor with shared_ptr and no sizes to support python interface.
    CReportStdTable(std::shared_ptr<CReportPageData> pReport, const CUnitsStr& szTitle, uint_16 uiCols);
    CReportStdTable(std::shared_ptr<CReportPageData> pReport, const CStr& szTitle, uint_16 uiCols);
    /// Creates the table with title @p szTitle and @p uiCols columns. If @p
    /// fShowHeader is false, this will hide the column headers.
    CReportStdTable(bool fShowHeader,
        CReportPageData* pReport,
        const CUnitsStr& szTitle,
        uint_16 uiCols,
        int_32 iLeftEdge = 75,
        int_32 iRightEdge = 775);
    /// Constructor with shared_ptr and no sizes to support python interface.
    CReportStdTable(bool fShowHeader,
        std::shared_ptr<CReportPageData> pReport,
        const CUnitsStr& szTitle,
        uint_16 uiCols);
    CReportStdTable(bool fShowHeader,
        std::shared_ptr<CReportPageData> pReport,
        const CStr& szTitle,
        uint_16 uiCols);

    ~CReportStdTable() = default;

    /// Resets all the data and parameters for the table.
    void Reset();

    /// Removes all the data from the columns, but keeps the metadata.
    void DeleteAllColumnData();

    /// Set the units context for the report and column titles.  Note that you
    /// have to set the units in each column separately through GetColInfo().
    void SetUnitsContext(std::shared_ptr<CUnitsContext> pUnitsContext);

    /// Position this table below the specified previous item.
    /// @return the item for this table for inclusion in the item chain.
    CReportDataItemBase* SetLayoutBottom(CReportDataItemBase* pPrevItem);

    /// Setup the column.  It also sets the default width.  It also sets up the
    /// column in the actual table, so it must be called before GetColInfo()
    void SetTitle(uint_16 uiCol, const CUnitsStr& szTitle);
    /// Setup the column.  It also sets the default width. It also sets up the
    /// column in the actual table, so it must be called before GetColInfo()
    void SetTitle(uint_16 uiCol, const CStr& szTitle);

    /// Set the formatting for the @p uiCol column to @p rcFmt
    void SetFormat(uint_16 uiCol, const CQuantityFormatter& rcFmt);
    /// Set the formatting with the formatted double passed in.
    template<typename T> void SetFormat(uint_16 uiCol, const TFormattedNumeric<T>& xdFmt)
    {
        SetFormat(uiCol, *xdFmt.GetFormatter());
    }
    /// Set the width of the @p uiCol to @p uiColumnWidth (1/100 inch).
    void SetColumnWidth(uint_16 uiCol, size_t uiColumnWidth);
    /// @{ Set the alignment of column @p uiCol
    void AlignLeft(uint_16 uiCol);
    void AlignRight(uint_16 uiCol);
    void AlignCenter(uint_16 uiCol);
    /// @}

    /// Sets up the column to be inverted if @p pUnitSel has a current
    /// selection of @p iSelForInvert.  The inversion is the @p dInvertValue -
    /// internal value of the column.
    void SetInverter(uint_16 uiCol,
        std::shared_ptr<CUnitSelection> pUnitSel,
        int iSelForInvert,
        double dInvertValue);

    /// @return the CTableColumnInfo for column @p uiCol.
    [[deprecated]] std::shared_ptr<CTableColumnInfo> GetColInfo(uint_16 uiCol);

    /// @return Number of columns.
    uint_16 GetNumCol() const;

    ///  Allow the table to use snaking.
    void SetSnaking(bool fEnable);

    /// @name Data Setting Methods
    /// @{

    // Add entries into the table.
    /// Add a single string @p szVal to columne @p uiCol
    void AddItem(uint_16 uiCol, const CStr& szVal);
    /// Add a single units string @p szValue to column @p uiCol
    void AddItem(uint_16 uiCol, const CUnitsStr& szValue);
    /// Add a double @p dVal to column @p uiCol
    void AddItem(uint_16 uiCol, double dVal);
    /// Add a whole vector @p adVals to column @p uiCol
    template<typename T>
    void AddItem(uint_16 uiCol, const std::vector<T>& adVals);
    /// Add a whole CNumVector @p adVals to column @p uiCol
    void AddItem(uint_16 uiCol, const CNumVector<double>& adVals);

    /// Insert an empty row.
    void AddBlank();

    /// Clear all the information from the data in the column.
    void ClearData(uint_16 uiCol);

    /// Clear all the information from the data in the whole table.
    void ClearData();

    /// @}

    /// Set the frame rectangle if tied to a full window.
    void SetFrameRect(const SRect& rRect);

    /// Allow the application to disable Interactions.
    void AllowInteraction(bool bArg);

    ///  Give access to the item for chaining more to the layout
    CReportDataItemBase* GetItem() const;

    ///  Give access to the table itself to link directly to a window.
    CReportTableData* GetTableItem() const;

    /// Convenience function to give access to the associated ReportPage (if any).
    [[deprecated]] CReportPageData* GetReportTypeData() { return m_pReport; }

private:
    /// Adds all the columns into the underlying table.
    void SetupColumns();

    /// The report (if any) that this is associated with.  Purely convenience
    /// for the established GetReportTypeData() method.
    CReportPageData* m_pReport;
    /// The actual table to work with.
    CReportTableData* m_pTable;
    /// The number of columns
    uint_16 m_uiCols;

    std::vector<PCDataSetColumn> m_apDataSet;
};

//-----------------------------------------------------------------------------
/// This is the way to create CReportTableData that is setup as a pair of
/// ID/Value columns.
///
/// To use, first create the table.  You can create this as part of a
/// CReportPageData or stand alone.  You also have the option of choosing to
/// not show the title nor the ID column.  You can set the units context
/// through the SetUnitsContext() method.  Note that each value can be unit
/// aware as well.
///
/// Then you can setup the parameters of how the table should be displayed.
///
/// Then you can add the data using any of the many convenience functions to
/// show the result.
class CReportOptionsTable
{
public:
    using CPtr = std::shared_ptr<CReportOptionsTable>;

    /// Creates the table with title @p szTitle
    CReportOptionsTable(CReportPageData* pReport,
        const CUnitsStr& szTitle,
        int_32 iLeftEdge = 75,
        int_32 iRightEdge = 775);
    /// Constructor with shared_ptr and no sizes to support python interface.
    CReportOptionsTable(std::shared_ptr<CReportPageData> pReport, const CUnitsStr& szTitle);
    CReportOptionsTable(std::shared_ptr<CReportPageData> pReport, const CStr& szTitle);

    /// Creates the table with title @p szTitle. If @p fLabels is false, this
    /// will not show the table title, nor the ID column.
    CReportOptionsTable(bool fLabels,
        CReportPageData* pReport,
        const CUnitsStr& szTitle,
        int_32 iLeftEdge = 75,
        int_32 iRightEdge = 775);
    /// Constructor with shared_ptr and no sizes to support python interface.
    CReportOptionsTable(bool fLabels,
        std::shared_ptr<CReportPageData> pReport,
        const CUnitsStr& szTitle);
    CReportOptionsTable(bool fLabels,
        std::shared_ptr<CReportPageData> pReport,
        const CStr& szTitle);

    ~CReportOptionsTable();

    /// Removes all the data from the columns, but keeps the metadata.
    void DeleteAllColumnData();

    /// Sets the units context used for the title.
    void SetUnitsContext(std::shared_ptr<CUnitsContext> pUnitsContext);

    /// @return the item for this table for inclusion in the item chain.
    CReportDataItemBase* GetItem() const;

    /// @return the table items in case of stand alone usage.
    CReportTableData* GetTableItem() const;

    /// Position this table below the specified previous item.
    /// @return the item for this table for inclusion in the item chain.
    CReportDataItemBase* SetLayoutBottom(CReportDataItemBase* pPrevItem);

    /// Set the actual rectangle, useful if the table is used in a standalone way.
    void SetFrameRect(const SRect& rRect);
    /// Allow the caller to disable user interaction in the table.
    void AllowInteraction(bool bArg);

    /// @name Table Setup
    /// @{
    void SetLeftColFrac(float_64 dLeftFrac);
    void HideLeftColumn() { m_fHideLeft = true; }
    void CenterRightColumn() { m_fCenterRight = true; }

    /// Change the table title.
    void SetTitle(const CStr& cszTitle);
    /// Hide the table title.
    void HideTitle();
    /// Show the table title.
    void ShowTitle();

    /// @}

    /// @name Data Methods
    /// @{

    /// Insert a string in each column.
    void AddItem(const CUnitsStr& szLabel, const CUnitsStr& szValue);
    void AddItem(const CStr& szLabel, const CStr& szValue);
    void AddItem(int iLabelID, const CStr& szValue);
    void AddItem(int iLabelID, int iValueID = 0);

    /// Insert an empty row.
    void AddBlank();
    /// Add an item to a validation report.
    void AddValidation(const CStr& szLabel, const CStr& szValue, const CStr& szWarning);
    /// Add the state of a yes/no selection.
    void AddYesNoLine(int iLabelID, bool fValue);
    /// Add the state of a yes/no selection.
    void AddYesNoLine(const CStr& szLabel, bool fValue);
    /// Add the state of a yes/no selection.
    void AddYesNoLine(const CUnitsStr& szLabel, bool fValue);
    /// Add a line for an integer.
    void AddNumericLine(int iLabelID, int iValue, const _TCHAR* szUnits = nullptr);
    /// Add a line for a numeric value.
    void AddNumericLine(int iLabelID, int iPrecision, double dValue, const _TCHAR* szUnits = nullptr);
    /// Add a line for a numeric value with uncertainty.
    void AddNumericLine(const CStr& szLabel,
        int iPrecision,
        double dValue,
        const _TCHAR* szUnits = nullptr);
    /// Add a line for a numeric value with uncertainty.
    void AddNumericLine(int iLabelID,
        int iPrecision,
        double dValue,
        double dUncertainty,
        const _TCHAR* szUnits = nullptr);
    /// Add a line for a numeric value with units.
    template<typename T>
    void AddNumericLine(int iLabelID,
        const TFormattedNumeric<T>& crValue,
        const _TCHAR* szOverrideUnits = nullptr);
    /// Add a line for a numeric value with units.
    template<typename T>
    void AddNumericLine(const CStr& szLabel,
        const TFormattedNumeric<T>& crValue);
    /// Add a line for a numeric value with units.
    template<typename T>
    void AddNumericLine(const CStr& szLabel,
        const TFormattedNumeric<T>& crValue,
        const _TCHAR* szOverrideUnits);
    /// Add a line for a numeric value with units and uncertainty.
    template<typename T>
    void AddNumericLine(int iLabelID,
        const TFormattedNumeric<T>& crValue,
        const TFormattedNumeric<T>& crUncertainty,
        const _TCHAR* szOverrideUnits = nullptr);
    /// Add a line for a numeric value with units and uncertainty.
    template<typename T>
    void AddNumericLine(const CStr& szLabel,
        const TFormattedNumeric<T>& crValue,
        const TFormattedNumeric<T>& crUncertainty,
        const _TCHAR* szOverrideUnits = nullptr);

    /// Add a range (min to max units)
    template<typename T>
    void AddAxisRange(int iLabelID,
        const TFormattedNumeric<T>& crMin,
        const TFormattedNumeric<T>& crMax,
        const _TCHAR* szOverrideUnits = nullptr);

    /// Add a range (min to max units)
    template<typename T>
    void AddAxisRange(const CStr& szLabel,
        const TFormattedNumeric<T>& crMin,
        const TFormattedNumeric<T>& crMax,
        const _TCHAR* szOverrideUnits = nullptr);

    /// Clear all the information from the data in the table.
    void ClearData();

    /// @}

private:
    CReportTableData* m_pTable;
    PCDataSetColumn m_pDataSetID;
    PCDataSetColumn m_pDataSetVal;
    int_32 m_iWidth;
    bool m_fHideLeft;
    bool m_fCenterRight;
    float_64 m_dLeftFrac;
    bool m_fInitDone;
    bool m_fIgnoreLabels;
};

//-----------------------------------------------------------------------------
template<typename T>
void CReportStdTable::AddItem(uint_16 uiCol, const std::vector<T>& adVals)
{
    ASSERT_ERROR(uiCol < m_uiCols);
    if(uiCol < m_uiCols && adVals.size() > 0)
        m_apDataSet[uiCol]->AppendItem(&adVals[0], adVals.size());
}

//-----------------------------------------------------------------------------
template<typename T>
void CReportOptionsTable::AddNumericLine(int iLabelID,
    const TFormattedNumeric<T>& crNumber,
    const _TCHAR* szOverrideUnits)
{
    CStr szLabel(iLabelID);
    AddNumericLine(szLabel, crNumber, szOverrideUnits);
}

//-----------------------------------------------------------------------------
template<typename T>
void CReportOptionsTable::AddNumericLine(const CStr& cszLabel,
    const TFormattedNumeric<T>& crNumber)
{
    AddNumericLine(cszLabel, crNumber, nullptr);
}

template<typename T>
void CReportOptionsTable::AddNumericLine(const CStr& szLabel,
    const TFormattedNumeric<T>& crNumber,
    const _TCHAR* szOverrideUnits)
{
    CUnitsStr szValue(szOverrideUnits ? CStr(_T("{%0-v%} ")) + CStr(szOverrideUnits) : CStr(_T("{%0%}")),
        {crNumber});
    AddItem(szLabel, szValue);
}

//-----------------------------------------------------------------------------
template<typename T>
void CReportOptionsTable::AddNumericLine(const CStr& szLabel,
    const TFormattedNumeric<T>& crNumber,
    const TFormattedNumeric<T>& crUncertainty,
    const _TCHAR* szOverrideUnits)
{
    CUnitsStr szValue(szOverrideUnits ? CStr(_T("{%0-v%} \u00B1 {%1-v%} ")) + CStr(szOverrideUnits)
                                      : CStr(_T("{%0-v%} \u00B1 {%1-v%} {%0-u%}")),
        {crNumber, crUncertainty});
    AddItem(szLabel, szValue);
}
//-----------------------------------------------------------------------------
template<typename T>
void CReportOptionsTable::AddNumericLine(int iLabelID,
    const TFormattedNumeric<T>& crNumber,
    const TFormattedNumeric<T>& crUncertainty,
    const _TCHAR* szOverrideUnits)
{
    CStr szLabel(iLabelID);
    AddNumericLine(szLabel, crNumber, crUncertainty, szOverrideUnits);
}
//-----------------------------------------------------------------------------
template<typename T>
void CReportOptionsTable::AddAxisRange(const CStr& szLabel,
    const TFormattedNumeric<T>& crMin,
    const TFormattedNumeric<T>& crMax,
    const _TCHAR* szOverrideUnits /* = nullptr */)
{
    CUnitsStr szValue(szOverrideUnits ? CStr(_T("{%0-v%} to {%1-v%} ")) + CStr(szOverrideUnits)
                                      : CStr(_T("{%0-v%} to {%1-v%} {%0-u%}")),
        {crMin, crMax});
    AddItem(szLabel, szValue);
}
//-----------------------------------------------------------------------------
template<typename T>
void CReportOptionsTable::AddAxisRange(int iLabelID,
    const TFormattedNumeric<T>& crMin,
    const TFormattedNumeric<T>& crMax,
    const _TCHAR* szOverrideUnits /* = nullptr */)
{
    CStr szLabel(iLabelID);
    AddAxisRange(szLabel, crMin, crMax, szOverrideUnits);
}

//-----------------------------------------------------------------------------

#endif  // not RSTBLHLP_HPP
