#ifndef SCRCTRL_HPP
#define SCRCTRL_HPP
//-----------------------------------------------------------------------------
//
// Library:			5200/MASS
//
// Description:
//
//	This describes the class that handles the script control Skip/Suspend/Cancel
//	and support for a particular execution.
//
// Contents:		CScrControl, CScrControlMon
//
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#if defined(MIC_DM643)
#include <cpp/nucpp.h>
#endif
#include "scrprcs.hpp"
#include "eventgroup.hpp"
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class	CScrControlMon;

class	CScrControlCancel;
class	CScrLogMonitor;
class	CScrRecData;
class	CScrRecTask;
/* ========================================================================
** CScrControl
**
**	This class handles the script control Skip/Suspend/Cancel and support 
**	for a particular execution.
**
**	The part the interfaces to the 'user' consists of the OnCancel(), 
**	OnSuspend(), OnResume(), and OnSkip() methods which handle the user 
**	requests, as well as the IsSuspended() call which provides status.
**
**	The interface to the script execution (microsteps). are the Wait(), 
**	WaitNoSuspend() and WaitResume() methods. Whenever a micrstep needs 
**	to 'delay' or 'Wait' during script execution it should call the Wait() 
**	method provided by an object of this class. This gives the synconization 
**	point to the microstep to consider interuptions. (If a step wants to 
**	handle interuptions, but not delay it can call Wait() with zero time()). 
**	microsteps that require no synconization (like manual control commands)
**	do not have to use this class at all.
**
**	The Wait() methods are virutal so that I an application has a well 
**	defined suspend behavoiur it can override these functions to handle that
**	behaviour automaticaly.
**
** ======================================================================== */
class CScrControl
   :	public	CScrProcessMon
{
public:

#if !defined(MIC_DM643)
    typedef uint_32 UNSIGNED;
#endif

    CScrControl(CScrProcess* pScript);
	//	This will initialize the object to control the script execution. 

	virtual	~CScrControl();
	//	Make sure that the destructor is virtual.
	
	CScrLogMonitor*	GetLogMonitor() const { return m_pcLogMonitor; }
	CScrRecData*	GetRecData() const { return m_pcRecordData; }
	CScrRecTask*	GetRecTask() const { return m_pcRecordTask; }
	CScrProcess*	GetScrProcess() const { return m_pScript; }
	//	These are opaque pointers for possible use in the script steps.


    uint_16 GetID() const;
	//	Return he ID associated with this script.
	
	void	SetLogMonitor( CScrLogMonitor* pcMon ) { m_pcLogMonitor = pcMon; }
	void	SetRecData( CScrRecData* pcData ) { m_pcRecordData = pcData; }
	void	SetRecTask( CScrRecTask* pcTask ) { m_pcRecordTask = pcTask; }
	//	These are opaque pointers for possible use in the script steps.

	bool	HasChanged() const { return m_fChanged; }
	//	returns true iff any of the accessors would return different 
	//	information since the last call to ClearChanged();

	void	ClearChanged() { m_fChanged = false; }
	//	Resets the changed indication (see HasChanged()).

	bool	IsSuspended() const { return m_fSuspended; }
	//	Will return thre if the script execution is suspended.

	uint_16	GetWaitCount() const { return m_uiWaitCount; }
	bool	IsWaiting() const { return m_fWaiting; }
	uint_8	GetWaitType() const { return m_ubWaitType; }
	uint_16	GetWaitID() const { return m_uiWaitID; }
	//	These define the wait for external continue state of the instrument.

	uint_16	GetWaitDataCount() const { return m_uiWaitDataCount; }
	bool	IsWaitingData() const { return m_fWaitingData; }
	uint_16	GetWaitDataID() const { return m_uiWaitDataID; }
	//	These define the wait for external continue state of the instrument.

#if defined(MIC_DM643)
	EventGroup*	GetEventGroup() { return &m_cEvents.GetEventGroup(); }
	//	This returns the event group that will be used in the Wait() method.
	//	The high 8-bits are used internaly to handle the suspend etc. 
	//	functionality, So all other events should use bits 0-23.
#endif
    
	/*
	**	Script control commands.
	*/
	void	OnContinueExternal();
	//	Called when the script is in a wait for external continue state
	//	to indicate that it should continue.

	void	OnContinueExternalData( float_32 flValue );
	//	Called when the script is in a wait for external data state to indicate
	//	that it should continue.

	void	OnCancel();
	//	Called when the script should be canceled. This will notify the 
	//	script that it sould cancel (i.e. skip to the terminating step)
	//	when the current step is done. It will then immediately terminate
	//	any pending wait with a cancel return.

	void	OnSuspend();
	//	This will cause this object to go into suspending a suspending state.
	//	This means that the next (or current) wait that is requested will 
	//	immediately terminate with a susped return. (given the usual 
	//	priorities the execution will be in always be wait state at this 
	//	point).

	void	OnResume();
	//	This will cause the object to go into a non-suspending state.
	//	The current or (next if none current) WaitResume() method will return
	//	immediately with running.

	void	OnSkip();
	//	This will cause the object to go into a skipping state. The current
	//	(or next if none is current) Wait() method will terminate with a 
	//	skip return.

	/*
	**	Support Commands.
	*/
	enum	EState
	{
		ceContinue,		//	Continue with normal flow.
		ceError,		//	The Error mask bit was set.
		ceTimeout,		//	Max time to wait expired.
		ceSkip,			//	User requested a skip.
		ceSuspend,		//	User requested a suspend.
		ceCancel		//	User requested a cancel.
	};

	virtual	EState	Wait( CTimeInterval ulMaxWait );
	virtual	EState	Wait( UNSIGNED uContinueEvent, UNSIGNED uErrorEvent, 
						  CTimeInterval ulMaxWait );
	EState	WaitNoSuspend( CTimeInterval ulMaxWait, CTimeInterval* pcRemain = NULL );
	EState	WaitNoSuspend( UNSIGNED uContinueEvent, UNSIGNED uErrorEvent, 
						   CTimeInterval ulMaxWait, CTimeInterval* pcRemain = NULL );
	//	These functions will wait either for a specified amount of time, 
	//	or for waiting on an event. The function will return one of the EState
	//	enumerations to indicate what caused the return, See below for 
	//	full explanation.
	//
	//	The methods will return ceCancel, ceSkip if either of these are 
	//	pending. The Wait() methods will handle the suspention internaly.
	//	The WaitNoSuspend() will return ceSuspend when the user wants to 
	//	suspend. This should result in a call to WaitResume (if necessary)
	//	to handle the suspention. Any of these cases take priority over 
	//	external events which will remain in the event group.
	//
	//	Passing in a zero time is allowed. This will just check the state of 
	//	the events without suspending (i.e. sumulate a suspention of 0 time).
	//
	//	The possible return values are as follows:
	//
	//	ceContinue	- The wait ended sucessfuly.
	//	ceError		- The wait ended with the error bit set.
	//	ceTimeout	- The requsted event did not occur in the time allotted.
	//	ceSkip		- The user would like to skip the current operation.
	//	ceSuspend	- The useer would like to suspend (only by WaitNoSuspend())
	//	ceCancel	- The user would like to cancel.
	//	
	//	The WaitNoSuspend methods take an optional parameter which is the 
	//	time remaining on the origional wait if the function returns with 
	//	the suspended state. It is OK if thiis is &ulMaxWait.

	EState	WaitResume();
	//	This should be called whenever the script should be placed into a 
	//	suspended state. This is usually a result of a user request (a call
	//	to Wait() returning a ceSkip), or some recoverable processing eror.
	//	The function will return one of the following.
	//
	//	ceContinue	- Continue with normal flow.
 	//	ceCancel	- The user requested a cancel. This should drop out of the
	//				  step as quickly as possible and return to the script 
	//				  processor to handle any termination.

	void	StartExternalWait( uint_8 ubType, uint_16 uiID );
	void	StopExternalWait();
	// These setup and clear the state information that the indicates if the
	// back end is waiting for the front end. These are ment to be used in the
	// situation where the mass has to do periodic processing (like
	// status updates) while waiting for the user to continue. These
	// functions would be called outside the loop to setup the state,
	// and then the WaitExternal() should be called inside the loop to
	// see if the user is ready to continue.

	EState	WaitExternal( uint_8 ubType, uint_16 uiID );
	EState	WaitExternal( uint_8 ubType, uint_16 uiID, CTimeInterval ulMaxWait );
	//	This will handle the wait for the external continue. It handles the
	//	suspends internaly, and will wait until OnContinueExternal(),
	//	OnSkip(), or OnCancel() methods are called. It returns one of the 
	//	following:
	//
	//	ceContinue	- User asked to continue or skip.
	//	ceCancel	- User asked to cancel.
	//	ceTimeout	- ulMaxWait value was exceeded.
	//
	// NOTE:	If this is being called betwee a StartExternalWait(),
	// StopExternalWait() block, The ubType, and uiID must match. In
	// this case this will not update the prompt status.

	EState	WaitExternalData( uint_16 uiID, float_64* pdData );
	EState	WaitExternalData( uint_16 uiID, float_32* pflData );
	//	This will handle the wait for the external data. It handles the
	//	suspends internaly, and will wait until OnExternalData(), OnCancel()
	//	methods are called (OnSkips() are silently consumed.) . It returns one
	//	of the following:
	//
	//	ceContinue	- User asked to continue (Data is filled in).
	//	ceCancel	- User asked to cancel.

	EState	WaitEvent( UNSIGNED uContinueEvent, UNSIGNED uErrorEvent, 
					   CTimeInterval ulMaxWait,
					   CScrControlCancel* pCancel, uint_32 uiID );
	//	This waits for an external event simmilar the the Wait above.
	//	This however will not return until either of the two events
	//	are recieved, or the timeout value expires.
	//	
	//	The possible return values are as follows:
	//
	//	ceCancel	- The user would like to cancel.
	//	ceTimeout	- The requsted event did not occur in the time allotted.
	//	ceError		- The wait ended with the error bit set.
	//	ceSkip		- The user would like to skip the current operation.
	//	ceContinue	- The wait ended sucessfuly. 
	//
	//	Since this can possibly wait for multiple events, it must decide
	//	which event has prescidence. The precidence is in the order listed
	//	above. 

	void	Add( CScrControlMon* pcMon );
	void	Remove( CScrControlMon* pcMon );
	//	Setup and terminate the monitor for this class.

	bool	IsActive() const { return m_fActive; }
	//	This is used to indicate whether or not the script is active. When the 
	//	script enters the active state all pending events are cleared. This 
	//	will eliminate any pending cancels.
	
	/*
	**	CScrProcessMon implementation.
	*/
	virtual	void	OnReset( CScrProcess* ) { SetActive( false ); }
	//	This function is called whenever the script Enters the Idle state.
	//	indicates that this script is not longer active.

	virtual	void	OnStart( CScrProcess* ) { SetActive( true ); }
	//	Called to indicate that the script is starting an run.
	//	Indicates that thsi is now active.

	virtual	void	OnTerminate( CScrProcess* ) { ClearEvents(); }
	//	This function is called whenever the script enters the 
	//	terminating state, clear any pending cancel etc.

private:

	void	SetActive( bool fActive );

	void	ClearEvents();
	//	This will clear all the events from the event queue if any.

	bool	m_fActive;
	//	True iff the script is currently active.

	bool	m_fSuspended;
	//	True iff the script is currently in a suspended state.

	uint_16	m_uiWaitCount;
	bool	m_fWaiting;
	uint_8	m_ubWaitType;
	uint_16	m_uiWaitID;
	//	These define the wait for external continue state.

	uint_16		m_uiWaitDataCount;
	bool		m_fWaitingData;
	uint_16		m_uiWaitDataID;
	float_32	m_flExternalData;
	//	These define the wait for external Data continue state.

	enum
	{
		//	Event masks for the builtin events.
		ceSkipMask		= 1<<24,
		ceSuspendMask	= 1<<25,
		ceResumeMask	= 1<<26,
		ceCancelMask	= 1<<27,
		ceWaitMask		= 1<<28,	//	Wait for external continue

		ceReservedMask	= 0xff << 24
	};

#if defined(MIC_DM643)
	CEventGroup		m_cEvents;
	//	The event group that scripts and steps can use to pend on. This
	//	is used to handle the skip/suspend/resume/cancel events, as well
	//	as any application specific used..
#endif

	CScrProcess*	m_pScript;
	//	The associated script.

	CScrLogMonitor*	m_pcLogMonitor;
	CScrRecData*	m_pcRecordData;
	CScrRecTask*	m_pcRecordTask;
	//	These are opaque pointers for possible use in the script steps.

	bool	m_fChanged;			//	Indicate if content changed since last clear.

	CScrControlMon*	m_apcMon[5];
	//	The class that gets notified on state changes.

	DECLARE_MSG()
};
/* ========================================================================
** CScrControlMon
**
**	This class provides the interface needed to monitor the the script 
**	control object.
** ======================================================================== */
class CScrControlMon
{
public:

	virtual	~CScrControlMon() {}
	//	Make sure the destructor is virtual.

	virtual	void	OnSuspend( CScrControl* pCtrl ) {}
	//	Called whenever the the script enteres the suspend state.

	virtual	void	OnResume( CScrControl* pCtrl ) {}
	//	Called whenever the the script exits the suspended state.

	virtual	void	OnCancel( CScrControl* pCtrl ) {}
	//	Called whenever the script is canceled.
};

/* ========================================================================
** CScrControlCancel
**
**	This is the base class for objects that need to do a WaitEvent() call
**	in the CScrControl class above. This provides the virtual function that
**	will be called whn the operation needs to be canceled.
** ======================================================================== */
class CScrControlCancel
{
public:

	virtual	~CScrControlCancel() {}
	//	Make sure the destructor is virtual.
	
	virtual	void	OnCancelWait( uint_32 uiID ) {}
	//	called when the wait should be canceled the uiID is the one
	//	passed into the CScrControl::WaitEvent().
};

#endif // SCRCTRL_HPP
