#ifndef SCRDATA_HPP
#define SCRDATA_HPP
//-----------------------------------------------------------------------------
//
// Library:			BaseLib
//
// Description:
//
//		This contains the data objects that are used to transfer/hold 
//		information recorded by the script.
//
// Contents:
//
//		class	CScrRecPoint
//		class	CSrcRecBlock
//		class	CSrcRecBlockSet
//
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <cpp/tmint.hpp>
#include <cpp/tmtimer.hpp>
#if defined(MIC_DM643)
#include <cpp/nucpp.h>
#endif
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
/* ========================================================================
** CScrRecPoint
**
**	The purpose of this class to to make it easy to work with the memory
**	block that is corresponds to a single data point. This is a very small
**	lightweight class therefore it can be copied and passed around at will.
** ======================================================================== */
class CScrRecPoint
{
public:

    /// Set up the recording point to work with the data provided.
    /// @param pData The start of the memory location for this block.
    CScrRecPoint( char* pData );
    /// Sets this block's memory location to NULL.
    ~CScrRecPoint();
    /// @param uiNumValues Number of data items in a point.
    /// @return The number of bytes that a single point of the
    /// passed in number of values will take.
    static uint_16 Size( uint_16 uiNumValues );
    /// @return The time at which the point was taken relative to
    /// the start of the script.
    const CTimeInterval& GetTime() const;
    /// @param uiIndex Index of data item.
    /// @return The value of the uiIndex'th data item for the point.
    float_32 Get( size_t uiIndex ) const;
    /// @param cTime This sets the time at which the data point was taken.
    void SetTime( const CTimeInterval& cTime );
    /// @param uiIndex/flValue This sets the value of the uiIndex'th data item
    /// for the point to flValue.
    void Set( size_t uiIndex, float_32 flValue ) const;
    /// @param uiIndex/dValue This sets the value of the uiIndex'th data item
    /// for the point to dValue.
    void Set( size_t uiIndex, float_64 dValue ) const;

private:

    char* m_pData;
};

/* ========================================================================
** CScrRecBlock
**
**	The porpose of this class is to allow the easy use of a block of memory
**	that contains a list of points. It can either manage the memory of this 
**	block either internaly, or externaly.
** ======================================================================== */
class CScrRecBlock
{
public:

	CScrRecBlock();
	//	This creates an empty block. It will allocate the normal amount
	//	of information for the block, but leave the data uninitialized.

	CScrRecBlock( bool fPreFilled, char* pData );
	//	Creates this class from the memory block passed in. If fPreFilled is
	//	true, then the pData is assumed to hold data of a full memory block.
    //  This class never takes ownership of the  memory block, and so will
    //  never delete
	//	it when the class no longer needs it. 
	//	

	~CScrRecBlock();
	//	Frees up the resources of this class.

	void	SetData( uint_16 uiBlockNum, uint_16 uiTableID, uint_16 uiNumValues );
	//	This creates a single block to house information for the given 
	//	table. 

	uint_16		CopyBlock( char** ppData ) const;
	//	This allocates the memory for a message into the data pointer passed
	//	in. It returns the size of the block allocated. The data for this
	//	message must be freed by the caller.

	uint_16	GetTableID() const { return ((SBlockHeader*)(void*)m_pData)->uiTableID; }
	//	retruns the associated table ID for the block.

	uint_16	GetNumValues() const { return ((SBlockHeader*)(void*)m_pData)->uiNumValues; }
	//	Returns the number of values in each point of the table.

	uint_16	GetBlockNum() const { return ((SBlockHeader*)(void*)m_pData)->uiBlockNum; }
	//	Returns the block number that this uses.

	bool	IsFull() const { return ((SBlockHeader*)(void*)m_pData)->fFull ? true : false; }
	//	Returns true if the block reports itself full.

	void	MarkFull() { ((SBlockHeader*)(void*)m_pData)->fFull = 1; }
	//	Sets this block to indicate that is is full.

	uint_16	GetMaxPoints() const;
	//	Returns the maximum number of points that can be added to this block.

	uint_16	GetNumPoints() const { return ((SBlockHeader*)(void*)m_pData)->uiNumPoints; }
	//	Returns the number of points that are available to that class.

	CScrRecPoint	GetPoint( uint_16 uiPointNum ) const;
	//	This returns the address in the block of the requested point.

	void	SetNumPoints( uint_16 uiNum );
	//	Sets the number of points that are avalable in this block. This 
	//	should be called after the data for that point is filled in.

	// public so miclib's showmsgs can use it
	struct	SBlockHeader
	{
		uint_16		uiTableID;
		uint_16		uiNumValues;
		uint_16		fFull;
		uint_16		uiBlockNum;
		uint_16		uiNumPoints;
        uint_16     uiFill;
	};

	enum
	{
		ceDefAllocSize = 1024		//	Default memory size to allocate.
	};


protected:

private:

	static	uint_16	CalcSize( uint_16 uiNumValues, uint_16 uiNumPoints );
	//	This calculates the size of the memeory to hold a block
	//	of the given size.

	char*	m_pData;
	//	The block of data associated with this file. The format is the 
	//	SBlockHeader structure, followed by the data for the indevidual 
	//	points.

    bool    m_fManageMemory;
    //  true if and only if the m_pData was allocated from the heap and and is
    //  the responcibility of this class to delete. 

	uint_16	m_uiAllocSize;
	//	The amount of memory allocated for the m_pData;
};
/* ========================================================================
** CScrRecBlockSet
**
**	The purpose of this task is to keep track of the data that is 
**	collected into the various tables under the script processing.
**	The access to the internal structures is serializes so multiple 
**	threads can access the class.
** ======================================================================== */
class CScrRecBlockSet
{
public:

    static  inline  size_t MemNeededFor( uint_16 uiMaxBlocks )
        { return ( (uiMaxBlocks+1)
                   * (CScrRecBlock::ceDefAllocSize
                      + sizeof(CScrRecBlock) ) ); }
    //  Retuns the memory needed for uiMaxBlocks to be passed into the
    //  constructor. 
    


    CScrRecBlockSet( uint_16 uiMaxBlocks,
                     char*pData );
	//	This sets up the subsystem to get ready to record information.
	//	It is assumed that the recording is done for the script processor
	//	that is passed in. The uiMaxBlocks is the maximum number of blocks
	//	That this will allocate before loosing data.  pData must point to a
	//	block of memory at least MemNeededFor( uiMaxBlocks ) large. 
    
	CScrRecBlockSet( uint_16 uiMaxBlocks );
	//	This sets up the subsystem to get ready to record information.
	//	It is assumed that the recording is done for the script processor
	//	that is passed in. The uiMaxBlocks is the maximum number of blocks
	//	That this will allocate before loosing data.

	virtual	~CScrRecBlockSet();
	//	Free up the resources used by this class.

	bool	HasChanged() const { return m_fChanged; }
	//	returns true iff any of the accessors would return different 
	//	information since the last call to ClearChanged();

	void	ClearChanged() { m_fChanged = false; }
	//	Resets the changed indication (see HasChanged()).

	void	SetNumTables( uint_16 uiNumTables );
	//	This indicates how many of the maximum number of tables will be 
	//	used for this script. This should be called before any SetupTable()
	//	method.

	void	SetupTable( uint_16 uiID, uint_16 uiNumValues );
	//	This will setup the table of the given ID. This should be called before 
	//	any data is added to the table.

	CScrRecPoint	GetNextPoint( uint_16 uiID );
	//	This will return a CScrRecPoint structure to easily
	//	fill in the data. If there is no more room in the current block, 
	//	it will create another block. This may cause an earlier block 
	//	to be lost.

	void	StoreNextPoint( uint_16 uiID );
	//	This will store the next point in in the table. It is assumed that 
	//	the point has been filled in.

	uint_16	GetNumTables() const { return m_uiNumTables; }
	//	Returns the number of tables that are currently active.

	uint_16	GetMinBlockNum( uint_16 uiID ) const;
	uint_16	GetMaxBlockNum( uint_16 uiID ) const;
	//	This returns the first block, and one past the last block number 
	//	that are available for a given table.

	uint_16	GetNumPoints( uint_16 uiID ) const;
	//	This will get the number of points that are stored in the last block
	//	of the table.

	uint_16		CopyBlock( uint_16 uiID, uint_16 uiNum, char** ppData );
	//	This allocates the memory for a message into the data pointer passed
	//	in. It returns the size of the block allocated. The data for this
	//	message must be freed by the caller. If the requested block is no
	//	longer available, different block for that table is returned.

	void	Reset();
	//	This function is called whenever the script Enters the Idle state.
	//	It will clean up the memory used by this recording session.

	void	Start();
	//	Called to indicate that the script is starting an run.
	//	It will reset the timer so that time recordings are relative to 
	//	that time.

	void	Done();
	//	This is called whenever the script enters the done state.
	//	This will finish all the currently active blocks to indicate 
	//	That no more data is to be collected.

protected:

private:

    /// Returns the requested block for that table. If that block is no
    /// longer available, different block for that table is returned.
    const CScrRecBlock& GetBlock( uint_16 uiID, uint_16 uiNum ) const;
    CScrRecBlock&       GetBlock( uint_16 uiID, uint_16 uiNum );

	/*
	**	The data for the tables.
	*/
	enum
	{
		ceMaxTables = 128
        // NOTE: Coordinate with the one in the following files: scrdata.hpp,
        // scrtrnsm.hpp, scrcoll.hpp (needs version control), scrmon.hpp (could
        // have version control), scrtrnsf.
	};

	CScrRecBlock*	m_aBlocks;
	//	The array of all the blocks used. There are m_uiMaxBlocks+1 of 
	//	these. The last one is used for tables that are unused.

    bool    m_fManageMemory;
    //  true if and only if this class is responcible for the memory allocated
    //  for the m_aBlocks;

	uint_16			m_uiMaxBlocks;
	//	The maximum number of blocks that will be used in this class.

	uint_16			m_uiNumTables;
	//	The number of tables in this script, or zero outside of the script.

	struct
	{
		uint_16			uiMinBlock;		//	Smallest block number available.
		uint_16			uiMaxBlock;		//	Largest block number available.
		uint_16			uiNumBlock;		//	Total blocks for this table.

		uint_16			uiNumValues;	//	Number of values recorded.

		CScrRecBlock*	aBlock;			//	Fisr block in the list.

	}				m_aTable[ ceMaxTables ];
	//	This contains the information about the data collection of each 
	//	table.

	bool			m_fEnableRec;
	//	true iff the recording is enabled. 

	CTimeTimer		m_cTimer;
	//	The class that handles the elapsed time since the start of the script.

#if defined(MIC_DM643)
	Semaphore	m_Critical;		//	Semaphore for critical access.
#endif

	bool	m_fChanged;			//	Indicate if content changed since last clear.

	DECLARE_MSG()
};

#endif // SCRDATA_HPP
