#ifndef SCRLOGM_HPP
#define SCRLOGM_HPP
//-----------------------------------------------------------------------------
//
// Library:			5200/MASS
//
// Description:
//
//		This contains the functionality for the holding the log information for
//		the mass.
//
// Contents:
//
//		class	CScrLogMonitor
//
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#if defined(MIC_DM643)
#include <cpp/nucpp.h>
#endif

#include "scrlogd.hpp"
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
/* ========================================================================
** CScrLogData
**
**	This class stores any information logged during the execution of the 
**	mass. This class serializes the calls to GetNextLog()/StoreNextLog()
**	and CopyData(). So that multiple threads can use these without 
**	disrupting the data.
** ======================================================================== */
class CScrLogMonitor
{
public:

	CScrLogMonitor( uint_16 uiMaxData );
	//	Create and get ready to start handling the logging.

    CScrLogMonitor(
        char*pData, uint_16 uiDataSize );
    //  Create this using the shared memory passed in. 
    
    
	~CScrLogMonitor();
	//	Free up any resources.

	bool	HasChanged() const { return m_fChanged; }
	//	returns true iff any of the accessors would return different 
	//	information since the last call to ClearChanged();

	void	ClearChanged() { m_fChanged = false; }
	//	Resets teh changed indication (see HasChanged()).

	CScrLogItem		GetNextLog( uint_16 uiNumArgs );
	//	Ensures that there is enough room for the logitem, and returns 
	//	the LogItem ready to be filled in. The caller must fill in the 
	//	data, and then call StoreNextLog. 

	void	StoreNextLog();
	//	This should be called after the call to GetNextLog(), and after 
	//	the Data is filled in. It will add the proper index and make it 
	//	available in the table.

	uint_16	GetLowAvail() const { return m_uiLowLog; }
	uint_16	GetHighAvail() const { return m_uiHighLog; }
	//	Get the low and high numbers of the available log items. If 
	//	Low == High there are no logs available.

	uint_16	CopyData( uint_16 uiNum, char** ppData );
	//	This allocates the memory for a message into the data pointer passed
	//	in. It returns the size of the block allocated. The data for this
	//	message must be freed by the caller. The block will contain all the 
	//	available log messages starting with uiNum. It may be possible that 
	//	some of the earlier messages are not available, and so will not be 
	//	sent. The data is in the form used by the CScrLogItemIter.
	//	The entries below uiNum are removed under the assumption that since 
	//	they were not asked for, they have already been processed.

    void    ClearAllMessages();
    //  Clear all the messages as though they were sent to the front end. 

protected:

private:

    void    RemoveLogsUpTo( uint_16 uiNum );

	char*	m_pData;			//	The data block allocated for this.
	uint_16	m_uiAllocSize;		//	The size allocated.
    bool    m_fDelete;          //  true iff this class owns m_pData;

	char*	m_pLowLog;			//	Pointer to the lowest log available.
	char*	m_pHighLog;			//	Pointer to the next log (assuming no wrap).
	char*	m_pWrap;			//	Location in m_pData where block wraps.

	uint_16	m_uiLowLog;			//	The number of the lowest log avaialble.
	uint_16	m_uiHighLog;		//	The number of the next log to be added.

#if defined(MIC_DM643)

    Semaphore   m_Critical;	    //	Semaphore for critical access.

#endif

	bool	m_fChanged;			//	Indicate if content changed since last clear.

	DECLARE_MSG()
};
#endif // SCRLOGM_HPP
