#ifndef SCRPRCS_HPP
#define SCRPRCS_HPP
//-----------------------------------------------------------------------------
//
// Library:			BaseLib
//
// Description:
//
//		This contains functionality of the script processor.
//
// Contents:
//
//	class	CScrProcess;
//	class	CScrProcessMon;
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "scrcom.hpp"
#include <cpp/tmtimer.hpp>
#if defined(MIC_DM643)
#include <cpp/nucpp.h>
#endif

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class	CScrProcess;
class	CScrProcessMon;
class	CScrControl;		//	Used only as opaque pointer.
/* ========================================================================
** CScrProcess
**
**	The objects of this class supervise the execution of a script. Each one 
**	can run a single script at a time. If there is need to run multiple 
**	scripts at a time, several objects of this type are needed. 
**
**	Before any scripts processors are created the application should call the
**	static SetMicroSteps() method. This will make the application dependent
**	script step executors available to the subsystem. After that the 
**	application should create a CScrProcess for every script that may be 
**	executed in parallel.
**
**	There are two ways this class is meant to be used. One is for running
**	service tests or analysis scripts. In this use the SetScript(), 
**	StartScript(), CancelScript(), and DoneScript() are called individually
**	as a result of a specific front end command. 
**	
**	The other way to use this is for execution of 'instantaneous' commands 
**	(typically manual control commands). In this mode the SetScript(),
**	StartScript(), and FinishScript() methods are called as a result of a
**	single front end command. In this method, the communications subsystem 
**	will block for the duration of the script execution, so the script 
**	execution should be fast. The blocking makes sense in this case because
**	the manual control commands are fast anyway, and blocking ensures that
**	there is no overlap manual control scripts.
**
**	The script processor also allows objects to monitor the the execution 
**	of the script. To do this the object must be of a class derived from 
**	the CScrProcessMon type described below, and register itself with
**	a script processor using the AddMon() and RemoveMon() methods. When a 
**	monitor is registered, it will receive notification of pertinent
**	information about the script execution.
**
**	The format of the script is a sequential sequence of script steps. 
**	each step consists of a ScriptID (uint_16), followed by a a Length
**	(uint_16), followed by a variable amount of data. The ScriptID is
**	either a built-in command described below, or an index into the
**	microstep table. The Length field gives the length of the data 
**	associated with the script step. This length does not include the 
**	size of the ScriptID nor the Length. 
**
**	The script processing walks through the script and executes each step
**	by calling the appropriate execution function with the step data. The 
**	processor also handles the script step ID's defined by the EScriptStep
**	enumeration	internally as follows.
**
**	Step		- This sets the number of the 'macro' step that this 
**				  script is on. This is used in support of some form of 
**				  synchronization or possibly logging with the front end,
**				  The data of this step is a single uint_16.
**	LogSteps	- This tells the script processor that Step commands
**				  should be logged (i.e., sent to any registered monitor).
**				  The data of this step is a single uint_16 which will 
**				  also be the ID used to log the step. This is followed by 
**				  another uint_16 which is the log mask used to log the 
**				  step.
**	Repeat		- This starts a repeat loop. The data associated is a 
**				  single uint_16 which gives the number of times that 
**				  the loop will be executed. This must be at least 1.
**				  The repeats can be nested.
**  RepeatCond  - Similar to Repeat but is associated additionally with a
**                uint_16 index into an array of conditional functions and
**                optionally a segment of data that can be used by the
**                conditional function.
**	EndRepeat	- This signals the end of a repeat loop. It will either
**				  go back to the first step after the corresponding repeat
**				  or to the following step this one. There is no data 
**				  associated with this step, but if a valid conditional index
**				  is found for the repeat, the repeat will also end if the
**				  indexed conditional function returns false.
**	DoIf        - Similar to RepeatCond but evaluates the condition at the
**	              start of the block. If the condition evaluates to true,
**	              the next step will be executed. If the condition evaluates to
**	              false, the script is scanned for a matching EndDoIf, and
**	              the next step after that is executed. The DoIf blocks can be
**	              nested.
**	EndDoIf     - This signals the end of a DoIf block. It is used as a marker
**	              in the event that the DoIf conditional evaluates to false.
**	              There is no data associated with this step.
**	Terminate	- This indicates that the termination processing of the 
**				  script is to begin. This will place the script into the 
**				  terminating state. There is no data associated with 
**				  this step.
**
**	NOTE:	If it is required we can add more generic looping constructs
**			and 'if' statements relatively easily.
**
**
** ======================================================================== */
class CScrProcess
#if defined(MIC_DM643)
	:	private	Task
#endif
{
public:

	typedef	void	(*MICROSTEP)( void* pData, uint_32 uiLen, CScrControl* pCtrl );
	static	void	SetMicrosteps( MICROSTEP aSteps[], uint_16 uiNum );
	//	This gives the scripting subsystem the list of available
	//	microsteps. This function should be called once during initialization
	//	of the application, before any CScrProcess objects are created.

	typedef	bool	(*REPEATCOND)( void* pData, uint_32 uiLen, CScrControl* pCtrl );
	static	void	SetRepeatCondition( REPEATCOND aRepeatCond[], uint_16 uiNum );
	//	This gives the scripting subsystem the list of available
	//	conditions for repeats. This function should be called once (if any
	//	conditional repeats are supported) during initialization of the
	//	application, before any CScrProcess objects are created.

    typedef	void	(*REPEATEND)( uint_16 uiCondIndex, uint_16 uiMaxCount, CScrControl* pCtrl );
    static	void	SetRepeatEndCallback( REPEATEND pRepeatEndCallback );
    //	This gives the scripting subsystem the callback function for
    //	the end-repeats. This function should be called once during
    //  initialization of the application, before any CScrProcess objects
    //  are created.

    typedef bool    (*DOIFCOND)( void* pData, uint_32 uiLen, CScrControl* pCtrl );
    static  void    SetDoIfCondition( DOIFCOND aDoIfCond[], uint_16 uiNum );
    //  This gives the scripting subsystem the list of available conditions
    //  for conditional execution. This function should be called once (if
    //  conditional execution is supported) during initialization of the
    //  application, before any CScrProcess objects are created.

	CScrProcess( uint_32 uiID
#if defined(MIC_DM643)
				 , UNSIGNED uiStackSize, OPTION Option,
				 UNSIGNED uiTimeSlice, BOOL fPreempt
#endif
        );
	//	This sets up the class. The parameters passed in are those used to
	//	create the script execution task. The script is 

	~CScrProcess();
	//	Clean up any resources.

	void	SetScrControl( CScrControl* pScrCtrl) { m_pScrControl = pScrCtrl; }
	//	Sets the opaque pointer to be forwarded to the microsteps.

	uint_32	GetID() const { return m_uiID; }
	//	Give access to the associated ID.
	
	bool	HasChanged() const { return m_fChanged; }
	//	returns true iff any of the accessors would return different 
	//	information since the last call to ClearChanged();

	void	ClearChanged() { m_fChanged = false; }
	//	Resets the changed indication (see HasChanged()).

	EScriptState	GetScriptState() const { return m_eState; }
	//	Returns the current state that the script processor is in.

	CTimeInterval	GetDuration() const { return m_Time.Elapsed(); }
	//	Returns the amount of time elapsed since the start of the script.

    /// @return The current (macro) step executed in the current script.
	uint_16		GetCurrStep() const { return m_uiCurrStep; }
	/// @return the last (macro) step in the current script.
	uint_16		GetMaxStep() const  { return m_uiMaxStep;  }
	/// @return The number of (macro) steps completed in the current script.
	uint_32     GetStepsDone() const { return m_uiStepsDone; }

	void	SetScript( uint_8* pubScript, uint_32 ulSize, bool fDelete );
	//	This stores the script in this object. This command is only valid 
	//	when the script is in the idle state. This is usually called as a
	//	result of a download from the front end. 
	//	If the fDelete flag is true, the data pointed to by pubScript is
	//	deleted when the script goes back to idle. If the fDelete flag 
	//	if false the data is discarded, but not deleted when the script 
	//	is done.

	void	StartScript();
	//	This is called whenever the the previously set script should be 
	//	Executed. The script is executed in a different task. This is only
	//	valid if the script is in the idle state. This places the script 
	//	in the running state.

	void	CancelScript();
	//	This is called during the execution of the script to request the
	//	script to cancel. This will not cancel the currently active step
	//	but when that step returns, this will skip all the remaining steps
	//	up to the Terminate step if any. The termination process is then
	//	executed. The state of the script execution immediately changes to 
	//	terminating.

	void	DoneScript();
	//	This should be called when the script is in the done or canceled state
	//	 to indicate that it is OK to do any kind of cleanup. This places the 
	//	script into the idle state, 

#if defined(MIC_DM643)
	void	FinishScript();
	//	This will wait for the script to reach its done or canceled state, and
	//	 then call the DoneScript() method to place the script back into an idle 
	//	state.
#endif

	void	Add( CScrProcessMon* pcMon );
	void	Remove( CScrProcessMon* pcMon );
	//	Setup and terminate links to classes that monitor the execution 
	//	of the script. 

protected:

private:

	void    GetStepParameters( uint_8*&      rpData,
                               size_t&       ruiDataSize,
                               uint_8*&      rpNextStep,
                               uint_32&      ruiID,
                               uint_32&      ruiLen );
    //  Increments the step pointer to the next step in the script. Deletes the
	//  data buffer argument, then creates a new buffer and fills it with the
	//  step data. The buffer length and step data length well as the step ID
	//  are copied into the respective arguments. In general, the variables used
	//  as arguments to this function should not be modified except by this
	//  function.

	enum
	{
		ceMaxMon			= 8,
		ceMaxActiveLoop 	= 16
	};

	uint_8*	m_pScript;
	//	The pointer to the script currently executing.

	uint_32	m_ulSize;
	//	THe size of the script.

    uint_8* m_pData;
    size_t  m_uiDataSize;
    //  This is storage for the current step data.  It only ever grows. 
    
	bool	m_fDelete;
	//	true iff the script should be deleted when it is finished.

	uint_32			m_uiID;
	//	The identifier for this script.
	
	EScriptState	m_eState;
	//	The current state of the script.

	uint_16			m_uiCurrStep;
	uint_16			m_uiMaxStep;
	//	The current and largest (macrostep) executed in this script.

	/// Number of steps completed. May be larger than max-step as it counts
	/// all steps completed (e.g., steps in a loop are counted multiple times).
	uint_32         m_uiStepsDone;

	uint_16			m_uiActiveLoops;
	//	The number of nested loops active in the script.

	struct SLoopStatus
	{
        SLoopStatus();

		uint_8*	pRetStep;		//	The step to return to on looping.
		uint_16	uiLoopCount;	//	The number of times through the loop.
		uint_16	uiMaxCount;		//	The total time the loop should be run.
        uint_16 uiCondIndex;    //  Index into the repeat condition table, or
                                //  past the end if unconditional.
        uint_8* pCondData;      //  The pointer to the data of the conditional
                                //  loop.
        uint_32 uiCondDataLen;  //  The length of the data for the conditional
                                //  loop.
	};

	SLoopStatus     m_aLoopStatus[ ceMaxActiveLoop ];
	//	The information for each of the active loops.

	bool			m_fCanceling;
	//	true iff the script is in a canceling state.

	CTimeTimer		m_Time;
	//	The time since the start of the script execution.

	bool			m_fLogSteps;
	//	If this is true all (macro) steps are logged.

	CScrProcessMon*	m_apcMonitor[ ceMaxMon ];
	//	The monitors for this script (if any).

	CScrControl*		m_pScrControl;
	//	This is an opaque pointer passed on the the script steps in case
	//	it is useful there.

#if defined(MIC_DM643)
	Event				m_StartEvent;		
	//	The event communication with the script task. Posting to this 
	//	wills tart the script execution.

	Event				m_DoneEvent;
	//	The event communication with the script task. Posting to this 
	//	indicates that the script is in the done or canceled state.

	VOID	Entry();
	//	The function overwritten to be the script task.
#endif
    
    void    ExecuteScript();
    //  Run the currently defined script from start to done

	bool	m_fChanged;			//	Indicate if content changed since last clear.

	DECLARE_MSG()
};
/* ========================================================================
** CScrProcessMon
**
**	This class provides the interface needed to monitor the a script 
**	processor. Objects that wish to receive this information should be
**	derived from this class and overwrite the pertinent virtual functions
**	to receive the notification. They of course must also register with the
**	script processor. 
**
**	The main usefulness of the monitoring is so that subsystems can do 
**	some sort of cleanup processing based on the state of the script.
**	For example the time based recording subsystem might need to turn off
**	recording when the script goes terminating, or idle. Or the collected 
**	data subsystem can clean up its collected data.
**
** ======================================================================== */
class CScrProcessMon
{
public:

	virtual	~CScrProcessMon() {}
	//	Make sure destructor is virtual for a base class.

	virtual	void	OnReset( CScrProcess* pSource ) {}
	//	This function is called whenever the script Enters the Idle state.

	virtual	void	OnStart( CScrProcess* pSource ) {}
	//	Called to indicate that the script is starting an run.

	virtual	void	OnTerminate( CScrProcess* pSource ) {}
	//	This function is called whenever the script enters the terminating state.

	virtual	void	OnDone( CScrProcess* pSource ) {}
	//	This is called whenever the script enters the done or canceled state.

	virtual	void	OnStep( uint_16 uiStep, CScrProcess* pSource ) {}
	//	This function is called whenever the script start a new macro step,
	//	It is only called if the script had executed a LogSteps microstep.
	//	The uiLogID is the argument of the LogSteps command, and the uiStep
	//	is the argument of the Step command.

	//virtual	void	Error( ???? arguments );
	//	This is called whenever there is an error in the script processor.

protected:

private:

};
#endif // SCRPRCS_HPP
