#ifndef SCRREC_HPP
#define SCRREC_HPP
//-----------------------------------------------------------------------------
//
// Library:			BaseLib
//
// Description:
//
//		This contains the recording functionality of the script subsystem.
//		This includes recording data into multiple tables of data sets.
//
// Contents:
//
//		class	CScrRecData
//		class	CScrRecTask
//
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#if defined(MIC_DM643)
#include <cpp/nucpp.h>
#endif
#include "scrprcs.hpp"
#include "scrdata.hpp"
#include <cpp/tmtimer.hpp>
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
/* ========================================================================
** CScrRecData
**
**	The purpose of this task is to keep track of the data that is 
**	collected into the various tables under the script processing.
**	The access to the internal structures is serializes so multiple 
**	threads can access the class.
** ======================================================================== */
class CScrRecData	:	public	CScrRecBlockSet, private	CScrProcessMon
{
public:

#if defined(MIC_DM643)
	CScrRecData( CScrProcess* pScr, uint_16 uiMaxBlocks );
	//	This sets up the subsystem to get ready to record information.
	//	It is assumed that the recording is done for the script processor
	//	that is passed in. The uiMaxBlocks is the maximum number of blocks
	//	That this will allocate before loosing data.
#endif

	CScrRecData( CScrProcess* pScr,
                 uint_16 uiMaxBlocks,
                 char*   pData );
	//	This sets up the subsystem to get ready to record information.  It is
	//	assumed that the recording is done for the script processor that is
	//	passed in. The uiMaxBlocks is the maximum number of blocks That this
	//	will allocate before loosing data.  pData must point to a block of
	//	memory at least MemNeededFor( uiMaxBlocks ) large.

	~CScrRecData();
	//	Free up the resources used by this class.

protected:

private:

	/*
	**	CScrProcessMon implementation.
	*/
	virtual	void	OnReset( CScrProcess* ) { CScrRecBlockSet::Reset(); }
	//	This function is called whenever the script Enters the Idle state.
	//	It will clean up the memory used by this recording session.

	virtual	void	OnStart( CScrProcess* )  { CScrRecBlockSet::Start(); }
	//	Called to indicate that the script is starting an run.
	//	It will reset the timer so that time recordings are relative to 
	//	that time.

	virtual	void	OnDone( CScrProcess* ) { CScrRecBlockSet::Done(); }
	//	This is called whenever the script enters the done state.
	//	This will finish all the currently active blocks to indicate 
	//	That no more data is to be collected.


	CScrProcess*	m_pScript;
	//	The associated script.
};
/* ========================================================================
** CScrRecTask
**
**	The purpose of this class is to run a background task that will 
**	periodicay record a set of floating point readings into a table.
**
** ======================================================================== */
class CScrRecTask
#if defined(MIC_DM643)
	:	private	Task,
#endif
		private	CScrProcessMon
{
public:

	typedef	float_64	(*SIGNAL)( void );
	static	void	SetSignals( SIGNAL aSignals[], uint_16 uiNum );
	typedef	bool	(*SIGNALEXTRA)( CScrControl* pCtrl, float_64* pdData );
	static	void	SetSignalsExtra( SIGNALEXTRA aSignals[], uint_16 uiNum );
	//	This gives the script recorder the list of available channels that
	//	be used to record the data.

	CTimeInterval	GetRecInterval() const { return m_Interval; }

#if defined(MIC_DM643)
	CScrRecTask( CScrProcess* 	   	   pScr, 
				 CScrRecData*		   pcRecData,
				 UNSIGNED uiStackSize, OPTION Option,
				 UNSIGNED uiTimeSlice, BOOL fPreempt );
	//	This creates the object and gets ready to create the task.
#endif
    
	~CScrRecTask();
	//	This frees up the resources.

	void	StartRecord( uint_16 uiID, const CTimeInterval& rInterval,
						 uint_16 uiNumSignal, const uint_16 aSignals[] );
	//	This will start the recording session 

	void	StopRecord();
	//	This will stop the recording of the time based information.

    void    RecordData( uint_16 uiTableID, uint_16 uiNum, const uint_16 auiSignal[],
                        CScrControl* pCtrl );
    //  Record all the signals into this table. 

    
protected:

private:

	enum
	{
		ceMaxSignals = 64		//	Maximum signals to record.
	};


	/*
	**	CScrProcessMon implementation.
	*/
	virtual	void	OnReset( CScrProcess* pSource );
	//	This function is called whenever the script Enters the Idle state.
	//	It will clean up the memory used by this recording session.

	virtual	void	OnDone( CScrProcess* pSource );
	//	This is called whenever the script enters the done state.
	//	This will end the periodic recording.

	/*
	**	Task overwritten.
	*/
#if defined(MIC_DM643)
	VOID	Entry();
	//	The function overwritten to be the background recording task.
#endif

	Event	m_Signal;
	//	The event signal that communicates with the recording task.
	CScrProcess*	m_pcScript;
	//	The associated script.

	CScrRecData*	m_pcRecData;
	//	The the recorder.

	uint_16			m_uiID;			//	The Table ID of this data.
	CTimeInterval	m_Interval;		//	The time between readings.
	uint_16			m_uiNumSignal;	//	How many signals to record.
	uint_16			m_aSignals[ceMaxSignals];	//	The signals.

	bool			m_fRecording;
	//	true iff the state should be recording.

	DECLARE_MSG()
};

#endif // SCRREC_HPP
