#ifndef SCRSIM_HPP
#define SCRSIM_HPP
//-----------------------------------------------------------------------------
//
// Library:			baselib
//
// Description:		The simulated script transport layer.
//
//
// Contents:		CScrCommTransportSim
//
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "scrtrnsb.hpp"
#include "scrdata.hpp"

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class	CScrStepSim;
/* ========================================================================
** CScrCommTransportSim
**
**	The simulation link for the script handler. Note that this does not
**	include simulating the details of the script execution, and data
**	collection. It can be expanded to do so however.
** ======================================================================== */
class CScrCommTransportSim : public CScrCommTransportBase
{
public:

	CScrCommTransportSim( CCommDataLink* pLink, uint_16 eWhich );
	//	Create the simulation link for the script handler.

	~CScrCommTransportSim();
	//	Free up any resources.

	void	SetStepSim( CScrStepSim* pStepSim ) { m_pStepSim = pStepSim; }
	//	Setup the operation script.

	void ExecuteScriptCmd( SIZE uiBufferSize, const char* pBuffer );

	/*
	**	Access to the collected data tables.
	*/
	void	LockTable() { m_uiTableLock++; };
	void	UnlockTable() { if ( --m_uiTableLock == 0 ) SendStatus(); }
	//	Allow for multiple updates with a single send status.
	
	void	SetNumTables( uint_16 uiNumTables )
		{ LockTable(); m_cTableBlock.SetNumTables( uiNumTables ); UnlockTable(); }
	//	This indicates how many of the maximum number of tables will be 
	//	used for this script. This should be called before any SetupTable()
	//	method.

	void	SetupTable( uint_16 uiID, uint_16 uiNumValues )
		{ LockTable(); m_cTableBlock.SetupTable( uiID, uiNumValues ); UnlockTable(); }
	//	This will setup the table of the given ID. This should be called before 
	//	any data is added to the table.

	CScrRecPoint	GetNextPoint( uint_16 uiID )
		{ return m_cTableBlock.GetNextPoint( uiID ); }
	//	This will return a CScrRecPoint structure to easily
	//	fill in the data. If there is no more room in the current block, 
	//	it will create another block. This may cause an earlier block 
	//	to be lost.

	void	StoreNextPoint( uint_16 uiID )
		{ LockTable(); m_cTableBlock.StoreNextPoint( uiID ); UnlockTable(); }
	//	This will store the next point in in the table. It is assumed that 
	//	the point has been filled in.

	void	SendStatus();
	//	Send the status to the front end, make sure status is not overwritten
	//	during the send.

    EScriptState	GetState() const { return (EScriptState)m_sStatus.eState; }
    //	Give read access to the state.

    void    SetWaiting( bool fWaiting, uint_8 ubWaitType, uint_16 uiWaitID  );
    //  Set the wait state.

    void    PauseScriptSimulation( bool fIsPause )
        { m_fIsPaused = fIsPause; }
    //  Pause the simulator's timed execution loop. This is specific to the
    //  simulator and used to interact with the script execution mechanism.

    /// If true, the last script step is repeated until this method is called again with a false
    /// argument. This can be used to keep the script from ending prematurely.
    void RepeatAtEnd( bool fRepeat ) { m_fRepeatAtEnd = fRepeat; }

protected:

	virtual	void	MessageSent( ID uiID, const char* pData, SIZE uiSize );
	//	If there are more status updates pending send them as well.

	virtual	char*	GetMemory( ID uiID, SIZE uiSize );
	//	This is called when the first packet of a message is received to
	//	'allocate' the memory of the message. This memory is used internally
	//	until either the message is complete and MessageRecieved() is called,
	//	or the datalink is reset and DiscardMemory() is called.
	//

	virtual	void	MessageRecieved( ID uiID, char* pData, SIZE uiSize );
	//	This is called when a message has been successfully received. If the
	//	message was of zero size, pData will be null, and no GetMemory() call
	//	was made.

    bool    IsPaused() const { return m_fIsPaused; }

private:

	void	SetSuspended( bool fSuspended )
		{ m_sStatus.fSuspended = fSuspended; SendStatus(); }
	//	Set the suspended flag.

	bool	IsSuspended() const { return (bool)m_sStatus.fSuspended; }
	//	Give access to the suspended state.

	void	SetState( EScriptState eState )
		{ m_sStatus.eState = eState; SendStatus(); }
	//	Set the execution state.

        void TimeExpired();
        //	The timer that walks through the script expired.

        SReqLogData m_sReqLogData;
        SReqTableBlock m_sReqTableBlock;
        //	The memory to contain the request data.

        CScrRecBlockSet m_cTableBlock;
        uint_16 m_uiTableLock;
        //	The simulated table block.

        SScriptStatus m_sStatus;
        bool m_fPending;
        bool m_fSending;
        //	The memory for the status upload.

        char* m_pScriptStart;
        char* m_pScriptNext;
        char* m_pScriptEnd;
        //	The script that is currently executing.

        uint_16 m_uiRepeatLeft;
        //	How many times to execute the current loop.

        char* m_pLoopNext;
        //	Where to continue the loop on repeat.

        CTimeOneShot m_cTimer;
        //	The timer that keeps the script moving.

        CScrStepSim* m_pStepSim;
        //	The operation script.

        bool m_fIsPaused;
        bool m_fIsCancelled;

        int m_iRange_Min;
        int m_iRange_Max;

        /// Repeat the last script step while true. This variable is set to false
        /// when the done step is requested. As a result, there is only a need
        /// to set this variable at the beginning of an operation, if required.
        bool m_fRepeatAtEnd;

        DECLARE_MSG()
};

//-----------------------------------------------------------------------------
/// Interface for the CScrCommTransportSim class supporting the execution of script steps.
class CScrStepSim
{
public:

    virtual ~CScrStepSim() {}

	/// The required interface for script step simulator classes intending to
    /// use CScrCommTransportSim. Implementing classes will simulate the backend
    /// script step behavior in response to ExecStep() using the provided script
    /// ID and related script step data.
    virtual void ExecStep( CScrCommTransportSim* pSim,
                           uint_32 uiID, uint_32 uiLen, void* pData ) = 0;

    /// Supports extended steps (steps that do not complete after ExecStep is called).
    /// If false is returned, the execution of the next script step is delayed.
    /// @returns True iff the last step executed by ExecStep is completed.
	virtual bool IsStepComplete()
	    { return true; }
	/// Supports application-specific simulated data. If application-specific data is
	/// supported, add the data points to the table indicated for the signals indicated
	/// and return true. Otherwise, random values are generated by default.
	/// @see CScrScriptBase::RecData
	/// @returns True iff default data generation is being replaced by application
	///     specific simulated values.
    virtual bool DoStepRecordData( CScrCommTransportSim* pSim,
                                   uint_16 uiTableID , uint_16 uiNum, const uint_16* auiSignal )
	    { return false; }
};

#endif // SCRSIM_HPP
