#ifndef SCRTRNSB_HPP
#define SCRTRNSB_HPP
//-----------------------------------------------------------------------------
//
// Library:			BaseLib
//
// Description:		The FrontEnd/mass base class for the Script/Service test
//					connection.
//
//
// Contents:		CScrCommTransportBase
//
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <cpp/comtrns.hpp>
#include "scrcom.hpp"
#include <cpp/tmint.hpp>
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
/* ========================================================================
** CScrCommTransportBase
**
**	The base class for script/service test handling support.
**
**	Message:				Data sent
**
**	ceReqScriptStatus,		(none)
**	ceReqLogData,			SReqLogData (First log entry.)
**	ceReqTableBlock,		SReqTableBlock	(Table ID, block number).
**	ceReqExecuteScript,		script buffer.
**	ceReqConinueScript,		(none)
**	ceReqSkipScript,		(none)
**	ceReqSuspendScript,		(none)
**	ceReqResumeScript,		(none)
**	ceReqCancelScript,		(none)
**	ceReqDoneScript,		(none)
**
**	ceScriptStatus,			SStatus followed by a variable number of STableStatus
**	ceLogData,				variable size buffer to use with CScrLoItemIter
**							defined in scrlogd.hpp
**	ceTableBlock,			variable sized block ready to be used with
**							CScrRecBlock defined in scrdata.hpp
**
** ======================================================================== */
class CScrCommTransportBase
	:	public	CCommTransport
{
public:
#ifdef MIC_VCMFC
#pragma pack(1)
#endif
	static const uint_32 s_cuiMaxScriptSize = 0xFA000;	// 1,024,000 bytes
	
	enum	ELogID
	{
		ceScriptLogStep = 0,
		ceScriptLogSuspend,
		ceScriptLogResume,
		ceScriptLogCancel
	};

	// below items are public so miclib's showmsgs can use
	enum	EMessage	
	{
		//-----------------------------------------------------------------
		// Front end to controller messages
		//-----------------------------------------------------------------
		ceReqScriptStatus = 0,      // none -> ceScriptStatus
		ceReqLogData,               // SReqLogData
		ceReqTableBlock,            // SReqTableBlock
		ceReqExecuteScript,         //  <script buffer>
		ceReqContinueScript,        //  none
		ceReqContinueDataScript,	//	float_32
		ceReqSkipScript,            //  none
		ceReqSuspendScript,         //  none
		ceReqResumeScript,          //  none
		ceReqCancelScript,          //  none
		ceReqDoneScript,            //  none

		//-----------------------------------------------------------------
		// Controller to front end messages
		//-----------------------------------------------------------------
		ceScriptStatus,             // SScriptStatus, uiNumTables*STableStatus
		ceLogData,                  // <log Data>
		ceTableBlock,               // <table Data>

	};

	struct	SReqTableBlock
	{
		uint_16		uiTableID;		//	The ID of the table of interest.
		uint_16		uiBlockNum;		//	Which block to upload.
	};


protected:

	CScrCommTransportBase( CCommDataLink* pLink, uint_16 eTransportID )
		:	CCommTransport( pLink, eTransportID ) {}		
	//	Setup the transport layer.
	//	NOTE:	The reason the TransportID is passed in is that this 
	//			allows potentialy multiple scripts and connections to 
	//			handle them on an instrument. Also this is the only thing
	//			that would make this Instrument specific.

#if defined(MIC_DM643)
public:
	typedef	uint_8						E8ScriptState;
	typedef	uint_8						E8Bool;
#else
	typedef	CFixedEnum8<EScriptState>	E8ScriptState;
	typedef	CFixedEnum8<bool>			E8Bool;
#endif

	struct	SReqLogData
	{
	    SReqLogData() : uiMinLog( 0 ), uiFill( 0 ) {}


		uint_16		uiMinLog;		//	Lowest log number to upload.
        uint_16     uiFill;         //  Filler to 32-bit boundry.
	};


	struct	SScriptStatus
	{
	    SScriptStatus()
        {
	        uiStepsDone = 0;
	        eState = ceScriptStateIdle;
	        fSuspended = false;
	        uiCurrStep = 0;
	        uiMaxStep = 0;
	        uiWaitCount = 0;
	        fWaiting = false;
	        ubWaitType = 0;
	        uiWaitID = 0;
	        uiWaitDataCount = 0;
	        fWaitingData = false;
	        ubFill = true;
	        uiWaitDataID = 0;
	        uiLowLog = 0;
	        uiHighLog = 0;
	        uiNumTables = 0;
        }

		//	Script processor state.
        uint_32         uiStepsDone;//  Number of steps executed.
		E8ScriptState	eState;		//	The execution state of the script.
		E8Bool			fSuspended;	//	True iff the script is suspended.
		uint_16			uiCurrStep;	//	Step currently executing.
		uint_16			uiMaxStep;	//	Largest step executed.

		uint_16			uiWaitCount;	//	how many time fWaiting has changed.
		E8Bool			fWaiting;		//	true iff script is waiting for
										//	front end.
		uint_8			ubWaitType;		//	For what reason the script is
										//	waiting.
		uint_16			uiWaitID;		//	The identifier of which wait.

		uint_16			uiWaitDataCount;//	how many time fWaitingData has
										//	changed.
		E8Bool			fWaitingData;	//	true iff script is waiting for
										//	front end.
        uint_8          ubFill;         //  Fill to 16 bit boundary. 
		uint_16			uiWaitDataID;	//	The identifier of which wait.

		//	Log data state
		uint_16	uiLowLog;		//	Get the low and one past the high numbers
		uint_16	uiHighLog;		//	 of the available log items. 

		//	table info.
		uint_16	uiNumTables;	//	number of tables used in the current script.
		//	This number os STableStatus structures follow.
	};
	//	The structure giving all the information for the status on the script.
	//	This is followed by uiNumTables STableStatis structures defining the 
	//	indevidual table status.

	struct	STableStatus
	{
	    STableStatus() : uiLowBlock( 0 ), uiHighBlock( 0 ), uiNumPoints( 0 ), uiFill( 0 ) {}

		uint_16	uiLowBlock;			//	first block, and one past the last block
		uint_16	uiHighBlock;		//	number that are available
		uint_16	uiNumPoints;		//	number of points in the last block.
        uint_16 uiFill;             //  Filler to 32-bit boundry.
	};
#ifdef MIC_VCMFC
#pragma pack()
#endif

private:

};
#endif // SCRTRNSB_HPP
