#ifndef SIMULATE_PLUMBING_HPP
#define SIMULATE_PLUMBING_HPP
//-----------------------------------------------------------------------------
/// @file
/// Simulated plumbing for an instrument.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <vector>

// ================================================================================================
/// Simulate a region of gas pressure.
class CSimRegion
{
public:
    CSimRegion() = default;
    CSimRegion(const CSimRegion&) = delete;
    CSimRegion& operator=(const CSimRegion&) = delete;
    virtual ~CSimRegion() = default;

    virtual float Pressure() const = 0;
    virtual void AddGas(float flGas) = 0;
};

// ================================================================================================
/// Simulated connection interface between two regions.
class CSimConnection
{
public:
    CSimConnection() = default;
    CSimConnection(const CSimConnection&) = delete;
    CSimConnection& operator=(const CSimConnection&) = delete;
    virtual ~CSimConnection() = default;

    /// Calculate the gas flow across the connection from region 1 to region 2
    /// in the time dt.
    virtual float GasFlow(float dt, const CSimRegion& rR1, const CSimRegion& rR2) const = 0;
};

// ================================================================================================
/// Simulated pressure source region (that is a region that is always at the same pressure).
class CSimSourceRegion final : public CSimRegion
{
public:
    CSimSourceRegion() = default;
    CSimSourceRegion(const CSimSourceRegion&) = delete;
    CSimSourceRegion& operator=(const CSimSourceRegion&) = delete;
    virtual ~CSimSourceRegion() = default;

    /// Set the pressure of the region.
    void SetPressure(float flPressure);

    virtual float Pressure() const override;
    virtual void AddGas(float flGas);

private:
    /// The (fixed) pressure of the region.
    float m_flPressure = 760.0f;
};

// ================================================================================================
/// Simulated gas volume region.
class CSimVolumeRegion final : public CSimRegion
{
public:
    CSimVolumeRegion() = default;
    CSimVolumeRegion(const CSimVolumeRegion&) = delete;
    CSimVolumeRegion& operator=(const CSimVolumeRegion&) = delete;
    virtual ~CSimVolumeRegion() = default;

    virtual float Pressure() const override;
    virtual void AddGas(float flGas) override;

    /// Set the simulated temperature of the volume.
    void SetTemp(float flTemp);
    /// Set the physical volume of the region.
    void SetVolume(float flVolume);
    /// Set the quantity of gas to achieve the requested pressure.
    void SetPressure(float flPressure);

private:
    /// The physical volume of the region.
    float m_flVolume = 1.0f;
    /// The absolute temperature of the region.
    float m_flTemp = 300.0f;
    /// The amount of gas in the region.
    float m_flGas = 0.0f;
};

// ================================================================================================
/// Simulated valve with a specific flow constant.
class CSimValve final : public CSimConnection
{
public:
    CSimValve() = default;
    CSimValve(const CSimValve&) = delete;
    CSimValve& operator=(const CSimValve&) = delete;
    virtual ~CSimValve() = default;

    // Set the flow constant of the valve
    void SetFlowConstant(float flFlowConst);

    /// Set the state of the valve;
    void SetValve(bool fIsOpen);

    virtual float GasFlow(float dt, const CSimRegion& rR1, const CSimRegion& rR2) const override;

private:
    /// The flow rate of the valve.
    float m_flFlowConst = 80.0f / 760.0f;
    /// If the valve is open or closed.
    bool m_fIsOpen = false;
};

// ================================================================================================
/// Region and connectivity class; represents the simulated network of instrument plumbing.
///
/// It is required that the connections and regions that this objects of this type hold are at the
/// same scope, and therefore at the same lifetime, as the object itself.
class CSimPlumbingNetwork final
{
public:
    /// The initialization structure defining the connection between regions.
    struct SConnection
    {
        CSimConnection& rConn;
        CSimRegion& rR1;
        CSimRegion& rR2;
    };

    CSimPlumbingNetwork() = default;
    CSimPlumbingNetwork(const CSimPlumbingNetwork&) = delete;
    CSimPlumbingNetwork& operator=(const CSimPlumbingNetwork&) = delete;
    virtual ~CSimPlumbingNetwork() = default;

    /// Setup the network of gas connections.
    void SetPlumbing(std::initializer_list<SConnection> aConnections);

    /// Flow the gas for the given time in iSteps steps. The more steps the more accurate the
    /// simulation, but it takes longer.
    void FlowGas(float flTime, size_t iSteps = 1);

private:
    struct SConnData
    {
        SConnData(CSimConnection& _rConn, CSimRegion& _rR1, CSimRegion& _rR2);
        CSimConnection& rConn;
        CSimRegion& rR1;
        CSimRegion& rR2;
        float flGas;  ///< Used in the running of the steps.
    };

    /// The connections to be made.
    std::vector<SConnData> m_vConnections;
};

#endif  // SIMULATE_PLUMBING_HPP
