#ifndef TCPTERM_HPP
#define TCPTERM_HPP
//-----------------------------------------------------------------------------
/// @file
/// A TCP/IP-based terminal class for direct data communication.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <cpp/tm1shot.hpp>

#include <memory>
#include <vector>

#if !defined(MIC_VCMFC)
#error only vcmfc target supported.
#endif

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class CStr;

namespace NTCPTermFE
{
    enum EError
    {
        ceInvalidSocket,
        ceCannotBind,
        ceCannotListen,
        ceCannotAccept
    };

    enum
    {
        ceBufferLen = 2048
    };
}

//-----------------------------------------------------------------------------
/// A simple TCP/IP-based terminal for direct data communication.
class CTCPTermFE final
{
    friend static LRESULT CALLBACK TCPTermProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

public:

    /// Observer interface for CTCPTermFE data notification.
    class CObserver
    {
    public:
        typedef NTCPTermFE::EError EError;
        typedef std::vector<char> VChar;

        virtual ~CObserver() {}
        /// Callback for data sent.
        virtual void OnSent() {}
        /// Callback for new data received.
        /// @param chChar Byte data received.
        virtual void OnRecv( const VChar& vchData ) {}
        /// Callback when Connect() is successful. Also called if AddObserver() is called
        /// while connected.
        virtual void OnConnect() {}
        /// Callback when Disconnect() is called.
        virtual void OnDisconnect() {}
        /// Callback for socket errors in CTCPTermFE.
        virtual void OnError( EError eError ) {}
        /// Callback for Windows errors.
        /// @param cszErr Error string provided by GetLastError and FormatMessage.
        virtual void OnErrorMsg( const CStr& cszError ) {}
    };

    typedef std::vector<char> VChar;

    /// Initialize the terminal but do not connect. @see Connect.
    /// @param cszIPAddr IP octets in string form.
    /// @param uiPort Port number.
	/// @param cTimeOut recheck send/receive after timeout (0 for no timeout)
    CTCPTermFE( const CStr& cszIPAddr, unsigned short uiPort, CTimeInterval cTimeOut = CTimeInterval::Sec(0) );
    ~CTCPTermFE();

	/// Port number.
	unsigned short Port();
    /// Establish connection to the remote IP address and port; notify observers.
    void Connect();
    /// Disconnect from the remote IP address and port; notify observers. If not connected
    /// observers are still notified. Any data waiting to send is NOT discarded.
    void Disconnect();
    /// @returns True if connected (has valid socket), false otherwise.
    bool IsConnected() const;
    /// Buffer data to send.
    /// @returns True if connected and the data is buffered for sending.
    bool SendData( const VChar& vchData );
    /// Clear the send buffer.
    void ClearSendData();
	/// Disconnect the socket and discard any data waiting to be sent.
	void Reset() { Disconnect(); ClearSendData(); }
	
    /// Add a terminal observer.
    void AddObserver( CObserver* pObserver );
    /// Remove a terminal observer.
    void RemoveObserver( CObserver* pObserver );

private:
    void TimeExpired();

    void NotifyObservers( NTCPTermFE::EError eError );
    void NotifyObservers( const CStr& cszError );
    CStr GetLastWindowsError() const;
    /// Callback for network activity for this terminal.
    void OnTCPIP( LPARAM lParam );
    void OnSendData();
    void OnReceiveData();

    typedef std::vector<CTCPTermFE*> VTCPTermFE;

    struct SImpl;
    std::unique_ptr<SImpl> m_pState;

	bool m_fIsConnected;

	CTimeInterval m_cTimeOut;
	
    DECLARE_MSG()
};

#endif /* TCPTERM_HPP */
