#ifndef TM1SHOT_HPP
#define TM1SHOT_HPP
//-----------------------------------------------------------------------------
/// @file
/// The classes that allow use of a one time callback after a certain amount of
/// time: CTimeOneShot
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "tmint.hpp"
#include "tmtimer.hpp"
#if defined(MIC_DM643)
#error The CTimeOneShot class is not supported in the MIC_DM643 environment.
#endif
#include <functional>
//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------

// ============================================================================
/// A one-shot timer. When the timer expires, the registered handler function
/// is called.  Total duration is limited to 0x7FFFFFFF ms, a little under 25
/// days.  The timers are not very accurate or fast since they depend on the
/// timing of the callback, but they do ensure that a minimum time elapses
/// before the callback.
///
/// For MIC_DM643 this is not supported.  (It has its own mechanism for timing
/// callbacks. For MIC_VCMFC this the callbacks are linked to the windows API
/// SetTimer() call that runs every 0.1 seconds. For all other targets, the
/// application must call the CTimeOneShot::CheckTimers() function periodically to
/// allow the timers to be checked.
class CTimeOneShot final
{
public:
    CTimeOneShot();
    CTimeOneShot(const CTimeOneShot&) = delete;
    CTimeOneShot& operator=(const CTimeOneShot&) = delete;

    ///  stops the timer in case it was still running.
    ~CTimeOneShot();

    ///  This sets the handler for this object. The handler should
    ///  continue to exist for the lifetime of this class.
    void SetHandler(std::function<void()> fnHandler);

    ///  This will start the timer running. It must have a Current handler.
    ///  If it is already running, this will stop it and restart.
    void Start(CTimeInterval uiTime);

    ///  Supend the running timer, or if not running do nothing.
    void Suspend();

    ///  Resume the suspended timer, or if not suspended no nothing.
    void Resume();

    ///  This causes the timer to stop execution. If it was already stopped
    ///  it will do nothing.
    void Stop();

    /// @return True iff the timer is running and not suspended or idle.
    bool IsRunning() const;

    /// The tick function that is called periodically to see if any timers need
    /// to be expired. The fidelity of this will result in the fidelity of the
    /// one shot timers of this.  On windows this is called automatically, for
    /// other frameworks the application needs to make sure this is called.
    static void CheckTimers();

private:
    ///  The function that gets called when the timer expires.
    void OnTimer();

    ///  The registered handler for the timer.
    std::function<void()> m_fnHandler;

    ///  The linked list of available timers. This is sorted by next
    ///  expiration.
    static CTimeOneShot* ms_pcHead;
    CTimeOneShot* m_pcPrev;
    CTimeOneShot* m_pcNext;

    ///  When the timer should expire.
    CTimeInterval m_cExpireTime;

    ///  The state of the timer object.
    enum
    {
        ceIdle,
        ceRunning,
        ceSuspended
    } m_eState;

    /// @{
    ///  If running, m_cTotalTime is the total time that was requested.
    ///  and m_cElapsedTime counts the time since then. This is either
    ///  from the initial Start, or the remaining info from Resume time.
    ///  If it is Stopped or paused, the m_cTotalTime is the time
    ///  remaining at the time it was stopped.
    CTimeInterval m_cTotalTime;
    CTimeTimer m_cElapsed;
    /// @}

    DECLARE_MSG()
};

#endif  // TM1SHOT_HPP
