#ifndef TMDATE_HPP
#define TMDATE_HPP
//-----------------------------------------------------------------------------
/// @file
/// The basic class the work with date/time objects.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#if defined(MIC_VCMFC)
#include <ctime>
#else
#include <time.h>
#endif
#include "cstr.hpp"

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
class CDataArchive;
class CTimeDate;

/// Serialize the time instance
void Serialize(CDataArchive& rAr, CTimeDate& rData);

// ============================================================================
/// This object encapsulates the concept of a calendar data/time value. It has
/// the ability to convert from struct tm and time_t. It can also convert to
/// human readable strings (according to the system settings if any). The
/// default value is a special NULL value which is converted into empty
/// strings.
///
/// On Windows, it should be constructed such that its internal SLocalTM struct is in local time.
///
/// On Linux, it should be constructed such that its internal SLocalTM struct is in UTC time, AND
/// the internal Linux system time should also be in UTC (run timedatectl to check).
///
/// This object has the usual persistence methods for data classes. Read(),
/// Write(), and GetSize() are defined in the usual way. This class can
/// read in the old (miclib) CTimeDate objects.
class CTimeDate
{
public:
    ///  @return the object for the current time.
    static CTimeDate Now();

#if !defined(MIC_VCMFC) && !defined(MIC_DM643) && !defined(MIC_TEST)
    /// Set the locale to be used to format the data and time
    static void SetLocale(const std::string& cszLocale);
    /// Set the local timezone to be used
    static void SetTimeZone(const std::string& cszTimeZone);
#endif

#if !defined(MIC_DM643)
    CTimeDate(const CTimeDate&) = default;
#endif
    CTimeDate() { SetDefaults(); }
    CTimeDate(const struct tm& d) { SetDate(d); }
    CTimeDate(time_t t) { SetDate(t); }
    //  The usual constructors.

    /// Set this to be the default (NULL) date.
    void SetDefaults();
    ///@{
    ///  Set the date of this object.
    void SetDate(const struct tm& d);
    void SetDate(time_t t);
    ///@}

    /// @return true if this is the default date
    bool IsDefault() const;
    /// @return true if the year is before 1971 and likely a derivative of the
    /// default date (start of the Epoch) or some other invalid date.
    bool IsBeforeEpoch() const;

#if defined(MIC_VCMFC)
    ///@{
    ///  Access the parts of the time structure in human readable form.
    static CStr GetStdUsaDateFormat();
    static CStr GetStdUsaTimeFormat();
    const CStr& GetDate() const;
    const CStr& GetTime() const;
    const CStr& GetBoth() const;
    const CStr& GetBothStdUsa() const;
    ///@}

#else
    ///@{
    ///  Access the parts of the time structure in human readable form.
    const char* GetDate() const;
    const char* GetTime() const;
    const char* GetBoth() const;
    ///@}

#endif

    ///  Convert this into a time_t object.
    operator time_t(void) const { return GetTime_t(); }
    ///  Convert this into a time_t object.
    time_t GetTime_t(void) const;

    /// @returns Local time. On Linux based systems this is the time in the timezone defined by
    /// s_pTimeZone, NOT the time in the system's timezone.
    struct tm GetLocalTime() const;

    bool operator==(const CTimeDate& rSrc) const;
    bool operator!=(const CTimeDate& rSrc) const;
    bool operator<(const CTimeDate& rSrc) const;
    bool operator>(const CTimeDate& rSrc) const;

#if !defined(MIC_DM643)
    void Read( std::istream& is );
    void Write( std::ostream& os ) const;
#endif

#if !defined(MIC_VCMFC)

    static const char* GetTimeSep();
    //  This returns the string to be placed between the hour, minute and
    //  seconds of the time printout.
#endif

private:

#if defined(MIC_VCMFC)
#pragma pack (1)
#endif
    struct SLocalTM
    {
        int_16 tm_sec;
        int_16 tm_min;
        int_16 tm_hour;
        int_16 tm_mday;
        int_16 tm_mon;
        int_16 tm_year;
        int_16 tm_wday;
        int_16 tm_yday;
        int_16 tm_isdst;
    };
#if defined(MIC_VCMFC)
#pragma pack ()
#endif
    // This is a copy of the std c library struc tm, but using
    // compiler independent types. int_16's were used since that is
    // the size of the ones used in CTimeDate (micmos).

    SLocalTM m_sTM;
    //  The time referred to, the special value of all zeros indicates
    //  an uninitialized object corresponding to a null string.

    /// The string  members cache the strings returned by the Get*() functions
    /// above for memoization and are mutable because object const-ness is
    /// independent of the state of these strings.
    mutable CStr m_szDate;
    mutable CStr m_szTime;
    mutable CStr m_szBoth;
};
#endif // TMDATE_HPP
