#ifndef TMINT_HPP
#define TMINT_HPP
//-----------------------------------------------------------------------------
/// @file
/// The class to work with an interval of time.
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "basetype.hpp"
#if !defined(MIC_DM643)
#include "dataarchive.hpp"
#include <iostream>
#endif

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
// ============================================================================
/// This is a lightweight class that stores a time. This makes it easy to use
/// various units of time.  This is accurate to 1 ms and will wrap every 49.71
/// days.
///
/// CTimeInterval can be created for various units of time with the factory
/// methods: MilliSec(), Sec(), Hour().
///
/// CTimeInterval objects can be compared, added and subtracted.  They can also
/// be multiplied (and divided) by pure numerical quantities.  In place
/// assignment operations are also provided.  Note that all these operators can
/// cause integer overflow (commonly caused by subtraction).
///
/// The length of the interval can be accessed in various units, either in full
/// precision (as a double), or rounded down as an integral type.
///
/// Finally, for the platforms that support streams, the standard streaming
/// interface is supported.
class CTimeInterval
{
#if !defined(MIC_DM643)
    friend void Serialize(CDataArchive& rAr, CTimeInterval& rData);
#endif

public:
    CTimeInterval() : m_ulData(0) {}

#if defined (MIC_VCMFC)
    /// This ignores the date stored in CTime and is used primarily by the
    /// Schedule diagnostic wizard.
    CTimeInterval( const CTime cTime )
    {
        uint_32 iHourInMilliSec = cTime.GetHour()   * 60 * 60 * 1000;
        uint_32 iMinsInMilliSec = cTime.GetMinute() * 60 * 1000;
        uint_32 iSecsInMilliSec = cTime.GetSecond() * 1000;
        m_ulData           = iHourInMilliSec + iMinsInMilliSec + iSecsInMilliSec;
    }
#endif

    ///@{
    /// Static constructors for various time units from numeric types.
    static CTimeInterval MilliSec(uint_32 i) { return CTimeInterval(i); }
    static CTimeInterval Sec(uint_32 i) { return CTimeInterval(i * 1000L); }
    static CTimeInterval Min(uint_32 i) { return CTimeInterval(i * 60000L); }
    static CTimeInterval Hour(uint_32 i) { return CTimeInterval(i * 3600000L); }
    static CTimeInterval MilliSec(int i) { return CTimeInterval(i); }
    static CTimeInterval Sec(int i) { return CTimeInterval(i * 1000L); }
    static CTimeInterval Min(int i) { return CTimeInterval(i * 60000L); }
    static CTimeInterval Hour(int i) { return CTimeInterval(i * 3600000L); }
    static CTimeInterval MilliSec(double d) { return CTimeInterval(uint_32(d)); }
    static CTimeInterval Sec(double d) { return CTimeInterval(uint_32(d * 1000.)); }
    static CTimeInterval Min(double d) { return CTimeInterval(uint_32(d * 60000.)); }
    static CTimeInterval Hour(double d) { return CTimeInterval(uint_32(d * 3600000.)); }
#if defined(MIC_CON) || defined(MIC_TEST)
    static CTimeInterval MilliSec( long unsigned int i ) { return CTimeInterval(uint_32(i)); }
    static CTimeInterval Sec( long unsigned int i ) { return CTimeInterval(uint_32(i*1000.)); }
    static CTimeInterval Min( long unsigned int i ) { return CTimeInterval(uint_32(i*60000.)); }
    static CTimeInterval Hour( long unsigned int i ) { return CTimeInterval(uint_32(i*3600000.)); }
#endif
    ///@}

    ///@{
    /// Numerical multiplication for time intervals is supported.
    friend CTimeInterval operator*(uint_16 rhs, CTimeInterval lhs);
    friend CTimeInterval operator*(uint_32 rhs, CTimeInterval lhs);
    friend CTimeInterval operator*(int_16 rhs, CTimeInterval lhs);
    friend CTimeInterval operator*(int_32 rhs, CTimeInterval lhs);
#if !defined(MIC_DM643)
    // On DM643 platform, int and int_32 are identical.
    friend CTimeInterval operator*(int rhs, CTimeInterval lhs);
#endif
    friend CTimeInterval operator*(CTimeInterval rhs, uint_16 lhs);
    friend CTimeInterval operator*(CTimeInterval rhs, uint_32 lhs);
    friend CTimeInterval operator*(CTimeInterval rhs, int_16 lhs);
    friend CTimeInterval operator*(CTimeInterval rhs, int_32 lhs);
#if !defined(MIC_DM643)
    // On DM643 platform, int and int_32 are identical.
    friend CTimeInterval operator*(CTimeInterval rhs, int lhs);
#endif
    friend CTimeInterval operator*(double rhs, CTimeInterval lhs);
    friend CTimeInterval operator*(CTimeInterval rhs, double lhs);
    CTimeInterval operator*=(uint_16 lhs);
    CTimeInterval operator*=(uint_32 lhs);
    CTimeInterval operator*=(int_16 lhs);
    CTimeInterval operator*=(int_32 lhs);
    friend CTimeInterval operator/(CTimeInterval rhs, double lhs);
    friend double operator/(CTimeInterval rhs, CTimeInterval lhs);
    ///@}

    ///@{
    /// Numerical multiplication for time intervals is supported.
    friend CTimeInterval operator+(CTimeInterval rhs, CTimeInterval lhs);
    friend CTimeInterval operator-(CTimeInterval rhs, CTimeInterval lhs);
    CTimeInterval operator+=(CTimeInterval lhs);
    CTimeInterval operator-=(CTimeInterval lhs);
    CTimeInterval operator*=(double lhs);
    CTimeInterval operator/=(double lhs);
    ///@}

    ///@{
    ///  Comparison operators.
    friend bool operator<(CTimeInterval rhs, CTimeInterval lhs);
    friend bool operator<=(CTimeInterval rhs, CTimeInterval lhs);
    friend bool operator>(CTimeInterval rhs, CTimeInterval lhs);
    friend bool operator>=(CTimeInterval rhs, CTimeInterval lhs);
    friend bool operator==(CTimeInterval rhs, CTimeInterval lhs);
    friend bool operator!=(CTimeInterval rhs, CTimeInterval lhs);
    ///@}

    ///@{
    /// @returns Time interval in the specified units rounded down to the
    /// nearest integral value.
    uint_32 Hours() const { return m_ulData / 3600000L; }
    uint_32 Minutes() const { return m_ulData / 60000L; }
    uint_32 Seconds() const { return m_ulData / 1000L; }
    uint_32 DeciSeconds() const { return m_ulData / 100L; }
    uint_32 CentiSeconds() const { return m_ulData / 10L; }
    uint_32 MilliSeconds() const { return m_ulData; }
    // left misspelled function in case it's used
    ///@}

    ///@{
    /// @returns Time interval in the specified units with full precision.
    double FullPrecHours() const        { return (double)m_ulData / 3600000.; }
    double FullPrecMinutes() const      { return (double)m_ulData / 60000.; }
    double FullPrecSeconds() const      { return (double)m_ulData / 1000.; }
    double FullPrecDeciSeconds() const  { return (double)m_ulData / 100.; }
    double FullPrecCentiSeconds() const { return (double)m_ulData / 10.; }
    double FullPrecMilliSeconds() const { return (double)m_ulData; }
    ///@}

#if !defined(MIC_DM643)
    bool Read( std::istream& is )
        {
            unsigned char ubVer = 1;
            is.read( (char*)&ubVer, sizeof( ubVer ) );
            is.read( (char*)&m_ulData, sizeof( m_ulData ) );
            return is.good();
        }
    bool Write( std::ostream& os ) const
        {
            // Version 1: Initial implementation.
            unsigned char ubVer = 1;
            os.write( (char*)&ubVer, sizeof( ubVer ) );
            os.write( (char*)&m_ulData, sizeof( m_ulData ) );
            return os.good();
        }
    //  Write the contents of the matrix assuming that the dimention is correct. If
    //  this fails, the failbit in the stream is set.
#endif

private:

    // Create this (privately) in the internal units.
#if !defined(MIC_DM643)
    explicit CTimeInterval( uint_32 ulTime ) : m_ulData( ulTime ) {}
#else
    CTimeInterval( uint_32 ulTime ) : m_ulData( ulTime ) {}
#endif

    ///  The actual time stored in units of milliseconds.
    uint_32 m_ulData;
};

//-----------------------------------------------------------------------------
inline CTimeInterval operator*(uint_16 rhs, CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs * lhs.m_ulData);
}
inline CTimeInterval operator*(uint_32 rhs, CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs * lhs.m_ulData);
}
inline CTimeInterval operator*(int_16 rhs, CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs * lhs.m_ulData);
}
inline CTimeInterval operator*(int_32 rhs, CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs * lhs.m_ulData);
}
inline CTimeInterval operator*(CTimeInterval rhs, uint_16 lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs.m_ulData * lhs);
}
inline CTimeInterval operator*(CTimeInterval rhs, uint_32 lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs.m_ulData * lhs);
}
inline CTimeInterval operator*(CTimeInterval rhs, int_16 lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs.m_ulData * lhs);
}
inline CTimeInterval operator*(CTimeInterval rhs, int_32 lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs.m_ulData * lhs);
}
inline CTimeInterval operator+(CTimeInterval rhs, CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs.m_ulData + lhs.m_ulData);
}
inline CTimeInterval operator-(CTimeInterval rhs, CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval(rhs.m_ulData - lhs.m_ulData);
}
inline CTimeInterval operator*(double rhs, CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval((uint_32)(rhs * lhs.m_ulData + 0.5));
}
inline CTimeInterval operator*(CTimeInterval rhs, double lhs)
{
    // Clients are responsible for handling overflow.
    return CTimeInterval((uint_32)(rhs.m_ulData * lhs + 0.5));
}
inline CTimeInterval operator/(CTimeInterval rhs, double lhs)
{
    return CTimeInterval((uint_32)(rhs.m_ulData / lhs + 0.5));
}
inline double operator/(CTimeInterval rhs, CTimeInterval lhs)
{
    return ((double)rhs.m_ulData / lhs.m_ulData);
}

//-----------------------------------------------------------------------------
inline CTimeInterval CTimeInterval::operator*=(uint_16 lhs)
{
    // Clients are responsible for handling overflow.
    m_ulData *= lhs;
    return *this;
}
inline CTimeInterval CTimeInterval::operator*=(uint_32 lhs)
{
    // Clients are responsible for handling overflow.
    m_ulData *= lhs;
    return *this;
}
inline CTimeInterval CTimeInterval::operator*=(int_16 lhs)
{
    // Clients are responsible for handling overflow.
    m_ulData *= lhs;
    return *this;
}
inline CTimeInterval CTimeInterval::operator*=(int_32 lhs)
{
    // Clients are responsible for handling overflow.
    m_ulData *= lhs;
    return *this;
}
inline CTimeInterval CTimeInterval::operator+=(CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    m_ulData += lhs.m_ulData;
    return *this;
}
inline CTimeInterval CTimeInterval::operator-=(CTimeInterval lhs)
{
    // Clients are responsible for handling overflow.
    m_ulData -= lhs.m_ulData; return *this;
}
inline CTimeInterval CTimeInterval::operator*=(double lhs)
{
    // Clients are responsible for handling overflow.
    m_ulData = (uint_32)(m_ulData * lhs + 0.5);
    return *this;
}
inline CTimeInterval CTimeInterval::operator/=(double lhs)
{
    m_ulData = (uint_32)(m_ulData / lhs + 0.5);
    return *this;
}

//-----------------------------------------------------------------------------
inline bool operator<(CTimeInterval rhs, CTimeInterval lhs)
{
    return bool(rhs.m_ulData < lhs.m_ulData);
}
inline bool operator<=(CTimeInterval rhs, CTimeInterval lhs)
{
    return bool(rhs.m_ulData <= lhs.m_ulData);
}
inline bool operator>(CTimeInterval rhs, CTimeInterval lhs)
{
    return bool(rhs.m_ulData > lhs.m_ulData);
}
inline bool operator>=(CTimeInterval rhs, CTimeInterval lhs)
{
    return bool(rhs.m_ulData >= lhs.m_ulData);
}
inline bool operator==(CTimeInterval rhs, CTimeInterval lhs)
{
    return bool(rhs.m_ulData == lhs.m_ulData);
}
inline bool operator!=(CTimeInterval rhs, CTimeInterval lhs)
{
    return bool(rhs.m_ulData != lhs.m_ulData);
}

#if !defined(MIC_DM643)
inline std::ostream& operator<<(std::ostream& os, const CTimeInterval& crInterval)
{
    return os << crInterval.MilliSeconds();
}

inline std::istream& operator>>(std::istream& is, CTimeInterval& rInterval)
{
    uint_32 uiMilliseconds;
    is >> uiMilliseconds;
    rInterval = CTimeInterval::MilliSec( uiMilliseconds );
    return is;
}

inline void Serialize(CDataArchive& rAr, CTimeInterval& rData)
{
    Serialize(rAr, rData.m_ulData);
}
#endif // not stream

#endif // TMINT_HPP
