#ifndef   UNITSCONTEXT_HPP
#define UNITSCONTEXT_HPP
//-------------------------------------------------------------------------------------------------
/// @file
/// Class for formatting strings within the context of selected units and
/// formatted doubles and units as units
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "cstr.hpp"
#include "baseunit.hpp"
#include "formattr.hpp"
#include "../rapidjson/fwd.h"

//-------------------------------------------------------------------------------------------------
/// The lookup table from a quantity name to a formatter.
using CQuantitiesMap = std::map<std::string, const CQuantityFormatter*>;

/// @param cQuantities The lookup table from quantity name to CQuantityFormatter table that will be
/// added to the json in the
/// quantities value.
/// @return a JSON string of all the global unit related tables (unit selections, simple units,
/// unitscontexts as a JSON string.
std::string GlobalUnitsJson(const CQuantitiesMap& cQuantities);

/// @returns the json schema associated with GlobalUnitsJson
std::string GlobalUnitsJsonSchema();

namespace NUnitsStr
{
    /// @return the Json schema for the CUnits ToJson method.
    std::string JsonSchema();
}  // namespace NUnitsStr

namespace NUnitsContext
{
    /// @returns the json schema associated with GlobalToJson
    std::string GlobalToJsonSchema();
}  // namespace NUnitsContext

//---------------------------------------------------------------------------------------------
/// This class combines a mask string and vector of TFormattedNumeric<double>
/// parameters.  It is meant to be used with CUnitsContext to allow definition
/// of strings that can be evaluated at presentation time.  (and also updated
/// if unit selections change.)  In particular see the
/// CUnitsContext::ConvertString  method for the details of the formatting.
class CUnitsStr final
{
    /// @return the json value for @p rData ready to be inserted into @p rDoc.
    friend rapidjson::Value ToJson(rapidjson::Document::AllocatorType& rAlloc,
        const CUnitsStr& rData);

public:
    using CFormattedDouble = TFormattedNumeric<double>;

    /// Default Constructor
    CUnitsStr() = default;

    /// Constructor to allow CUnitsStr to pretend to be CStr
    CUnitsStr( const CStr& szFormat );

    /// Constructor to allow CUnitsStr for a single TFormattedNumeric
    explicit CUnitsStr(const CFormattedDouble& cxdVal);

    /// Constructor. Note that it can be constructed in-line with szFmt, { x1,
    /// x2 }.
    CUnitsStr(CStr szFormat, const std::vector<CFormattedDouble>& cvxdValues);

    /// Access to the format string.
    const CStr& GetFormat() const { return m_szFormat; }

    /// Access to the list of values.
    const std::vector<CFormattedDouble>& GetValues() const { return m_vxdValues; }

private:
    /// The string to use for formatting.
    CStr m_szFormat;

    /// The formatted doubles to format into the string.
    std::vector<CFormattedDouble> m_vxdValues;
};

//-------------------------------------------------------------------------------------------------
/// Creates a context for CUnitsStr or CStr formatting based on the current
/// units.   The context consists of a [name], CUnitSelection pair.  That
/// replaces {%US:[name]%} masks in the conversion with the current
/// ChoiceName().  There is a second context for [name CUnits pairs.  Those
/// are used to replace the {%UT:[name]%} with the currently selected units for
/// that name.
class CUnitsContext
{
public:
    using Ptr = std::shared_ptr<CUnitsContext>;

    /// Create a CUnitsContext.  All are shared pointers so that they can be
    /// passed around easily.
    static Ptr Create(CStr&& szName,
        std::map<CStr, PCUnitSelection>&& aUnitSels,
        std::map<CStr, CUnits>&& aUnits);

    /// @return json value for all the known CUnitSelection objects.
    static rapidjson::Value GlobalToJson(rapidjson::Document::AllocatorType& rAlloc);

    /// Give access to a single default Context object.  This does not have any
    /// actual context, but can be used to expand CFormattedDouble
    /// within the CUnitsStr class.
    static Ptr GetDefault();

    /// @return the name of this units context.
    const CStr& GetName() const;

    /// Convert a string that has {%US:[name]%} and/or {%UT:[name]%} patterns,
    /// and fill int CUnitSelection::GetChoiceName() and CUnits.ShortUnits()
    /// strings instead.
    CStr ConvertString(CStr szStr) const;

    /// Convert a CUnitsStr to a string.  This uses the same conversion for the
    /// context as ConvertString( szStr ).  It also adds the {%n%} {%n-v%} and
    /// {%n-u%} to replace with the 'n'th formatted double value and untis,
    /// value, or units.
    CStr ConvertString(const CUnitsStr& cUnitsStr) const;

private:
    /// The constructor is private so that only shared pointers to this are
    /// allowed.
    CUnitsContext(CStr&& cszName,
        std::map<CStr, PCUnitSelection>&& aUnitSels,
        std::map<CStr, CUnits>&& aUnits);

    CUnitsContext(const CUnitsContext&) = delete;
    CUnitsContext& operator=(const CUnitsContext&) = delete;
    ~CUnitsContext();

    /// The name of this units context.
    const CStr m_cszName;

    /// The list of CUnitSelection to find lookup under {%US:[name]%}
    std::map<CStr, PCUnitSelection> m_UnitSelMap;

    /// The list of [name], CUnit to find lookup under {%UT:[name]%}
    std::map<CStr, CUnits> m_UnitMap;
};

#endif  // UNITSCONTEXT_HPP
