#ifndef CALC_CFP_HPP
#define CALC_CFP_HPP
//-------------------------------------------------------------------------------------------------
/// @file
/// The fundamental cfp calculations.  The calculations proceed from the transducer
/// readings/collected data to the final reports in the following way.
/// 1. Start with collected data --- these are pressure, flow, temperature from the transducers.
/// 2. Calculate Raw data by applying the lohm correction to the pressures.
/// 3. Calculate FilterFlow by combining the wet and dry data raw flow vs pressure data.  Then
///    calculate the pores size distribution: cumulative and incremental:  pore quantity vs
///    diameter.
/// 4. Calculate statistics on the pore size distribution.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include "dataporometry.hpp"
#include <cpp/internalunits.hpp>
#include <vector>
#include <algorithm>

//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
namespace NCfp
{
    struct SPoreSizeDist;
    struct SPoreData;
    struct SPoreSizeDistSummary;
    struct SCumulativeFilterFlowData;

    /// @return the smoothed and interpolated pressures and flow for the pressure and flow curve.
    std::pair<SI::V::psig<double>, SI::V::cm3_per_min<double>>
    SmoothBin(const SI::V::psig<double>& cvdPress, const SI::V::cm3_per_min<double>& cvdFlow);

    /// @return calculated pore diameter at the differential pressure.
    SI::um<double> CalcPoreDiameter(
        SI::psig<double> dDifferentialPressure,  ///< the differential pressure
        SI::pure<double> dShapeFactor,           ///< pore shape factor.
        SI::dynes_per_cm3<double> dSurfTension   ///< wetting fluid surface tension
    );

    /// @return the calculated differential pressure for the pore diameter.
    SI::psig<double> CalcDifferentialPressure(
        SI::um<double> dPoreDiameter,            ///< the pore diameter.
        SI::pure<double> dShapeFactor,           ///< pore shape factor.
        SI::dynes_per_cm3<double> dSurfTension   ///< wetting fluid surface tension
    );

    /// @return calculated full pore size distribution.
    SPoreSizeDist CalcPSDData(
        const SWetDryData& csWetDryData,        ///< The raw (but lohm corrected) collected data.
        bool fIsRamp,                           ///< If the data was collected in ramp mode.
        SI::pure<double> dShapeFactor,          ///< the shape factor for pores.
        SI::dynes_per_cm3<double> dSurfTension  ///< The wetting fluid surface tension
    );

    /// @returns the bubble point based on @p csData.
    SPoreData CalcBubblePoint(const SPoreSizeDist& csData);
    /// @returns the smallest pore size based on @p csData.
    SPoreData CalcSmallestPore(const SPoreSizeDist& csData);
    /// @returns the largest pore size based on @p csData.
    SPoreData CalcLargestPore(const SPoreSizeDist& csData);
    /// @returns the mean flow point based on @p csData.
    SPoreData CalcMeanFlowPoint(const SPoreSizeDist& csData);

    /// @returns values for pore size distribution at the largest incremental size
    SPoreSizeDistSummary CalcPoreSizeDistSummary(const SPoreSizeDist& csData);
    /// @returns the diameters at a few fractional flow rates.
    SCumulativeFilterFlowData CalcCumulativeFilterFlowData(const SPoreSizeDist& csData);
}  // namespace NCfp

/// The pressure, pore diameter and cumulative flow for a point on the pore data curve.
struct NCfp::SPoreData
{
    SI::um<double> dPoreDiameter{0.0};        ///< Pore diameter (microns)
    SI::pct<double> dFilterFlowPercent{0.0};  ///< Cumulative filter flow (%)
    SI::psig<double> dPorePressure{0.0};      ///< Pore pressure (psig)
};

/// The data for the cumulative pore size distribution.
struct NCfp::SPoreSizeDist
{
    SI::V::psig<double> vdPressure;        ///< Pressures for each filter flow point.
    SI::V::cm3_per_min<double> vdWetFlow;  ///< Source wet flow that was used in calculation.
    SI::V::cm3_per_min<double> vdDryFlow;  ///< Interpolated dry flow that was used in calculation.
    SI::V::pct<double> vdFilterFlowPercent;      ///< Cumulative distribution for the pore sizes.
    SI::V::um<double> vdPoreDiameter;            ///< Pore diameters for the distribution.
    SI::V::pct<double> vdIncrementalFilterFlow;  ///< Incremental filter flow.
    std::vector<double> vdPoreDistribution;      ///< Differential pore distribution. (% per um)
    SI::V::um<double> vdAverageDiameter;         ///< Average diameters for the vdPoreDistribution

    /// The bubble point information if available.
    NCfp::SPoreData sBubblePoint = NCfp::SPoreData{};

    /// Do we have enough data to calculate a bubble point.
    bool HasBubblePoint() const { return sBubblePoint.dPoreDiameter > 0.0; }
    /// Do we have enough data for a pore distribution.
    bool HasPSD() const { return vdPressure.size() > 1; }
};

/// pore size distribution data
struct NCfp::SPoreSizeDistSummary
{
    double dMaxPoreSizeDist = 0.;        ///< Maximum pore size distribution value
    SI::um<double> dDiameterAtMax{0.0};  ///< Pore diameter at max pore size dist.
};

/// cumulative filter flow data (pore size at flow %)
struct NCfp::SCumulativeFilterFlowData
{
    SI::um<double> d10PercentFlow{0.0};  ///< pore size at 10% CFF
    SI::um<double> d25PercentFlow{0.0};  ///< pore size at 25% CFF
    SI::um<double> d75PercentFlow{0.0};  ///< pore size at 75% CFF
    SI::um<double> d90PercentFlow{0.0};  ///< pore size at 90% CFF
};

#endif  // CALC_CFP_HPP
