#ifndef DATAPOROMETRY_HPP
#define DATAPOROMETRY_HPP
//-----------------------------------------------------------------------------
/// @file Common enumerations, constants, and functions
//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <cpp/formattr.hpp>

//-----------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-----------------------------------------------------------------------------
namespace NCfp
{
    enum EFluid : uint_16;
    enum EShape : uint_16;
    enum ETType : uint_16;
    enum ETMode : uint_16;
    enum EFlowMeterToUse : uint_16;
    enum class EBubblePointMethod : uint_16;
    enum class EPoreSizeRange : uint_16;

    struct SFluid;
    struct SSampleShape;
    struct SPoreShape;
    struct SRawCurve;
    struct SWetDryData;
}  // namespace NCfp

PCUnitSelection UnitSelCfpSampleDimension();
PCUnitSelection UnitSelCfpSampleThickness();
PCUnitSelection UnitSelSLMFromSCCM();

enum ECfpUnitChoices
{
    ceCfpSampleDimensioncm,
    ceCfpSampleDimensionmm,
    ceCfpSampleDimensionin,
    ceCfpSampleThicknesscm,
    ceCfpSampleThicknessum,
    ceCfpSampleThicknessmm,
    ceCfpFlowSLM,
};

const CUnits& fcn_UT_SurfaceTension();
#define UT_SurfaceTension fcn_UT_SurfaceTension()
const CUnits& fcn_UT_CfpLength();
#define UT_CfpLength fcn_UT_CfpLength()
const CUnits& fcn_UT_CfpArea();
#define UT_CfpArea fcn_UT_CfpArea()
const CUnits& fcn_UT_CfpSampleThickness();
#define UT_CfpSampleThickness fcn_UT_CfpSampleThickness()
const CUnits& fcn_UT_CfpGaugePressurePSIG();
#define UT_CfpGaugePressurePSIG fcn_UT_CfpGaugePressurePSIG()
const CUnits& fcn_UT_CfpGaugePressurePSIGPerSec();
#define UT_CfpGaugePressurePSIGPerSec fcn_UT_CfpGaugePressurePSIGPerSec()
const CUnits& fcn_UT_Microns();
#define UT_Microns fcn_UT_Microns()
const CUnits& fcn_UT_SCCM();
#define UT_SCCM fcn_UT_SCCM()
const CUnits& fcn_UT_SLM_FROM_SCCM();
#define UT_SLM_FROM_SCCM fcn_UT_SLM_FROM_SCCM()

// Application formatted doubles
FORMATTED_DOUBLE(DTimeSeconds)
FORMATTED_DOUBLE(DCfpPressure)
FORMATTED_DOUBLE(DCfpPressurePerS)
FORMATTED_DOUBLE(DCfpSurfaceTension)
FORMATTED_DOUBLE(DSilwickSurfaceTension)
FORMATTED_DOUBLE(DGalwickSurfaceTension)
FORMATTED_DOUBLE(DPorewickSurfaceTension)
FORMATTED_DOUBLE(DCfpLength)
FORMATTED_DOUBLE(DCfpArea)
FORMATTED_DOUBLE(DCfpShapeFactor)
FORMATTED_DOUBLE(DCfpInitialPressure)
FORMATTED_DOUBLE(DCfpFinalPressure)
FORMATTED_DOUBLE(DCfpPoreSize)
FORMATTED_DOUBLE(DCfpSampleThickness)
FORMATTED_UNSIGNED(UCfpNumAnalysisPoints)
FORMATTED_UNSIGNED(UCfpPrecision)
FORMATTED_UNSIGNED(UCfpResolution)
FORMATTED_DOUBLE(DCfpFlow)
FORMATTED_DOUBLE(DCfpFlowSLM)
FORMATTED_DOUBLE(DCfpFlowPercent)
FORMATTED_DOUBLE(DCfpCVValue)

enum NCfp::EFluid : uint_16
{
    ceFluidPorewick,
    ceFluidSilwick,
    ceFluidGalwick,
    ceFluidOther,
};

enum NCfp::EShape : uint_16
{
    ceShapeCircle,
    ceShapeSquare,
    ceShapeOther,
};

/// The type of test being run
enum NCfp::ETType : uint_16
{
    ceTTypeBP = 1,       ///< Bubble Point
    ceTTypeWUDD = 2,     ///< Wet-Up/Dry-Down
    ceTTypeWUDU = 3,     ///< Wet-Up/Dry-Up
    ceTTypeDUWU = 4,     ///< Dry-Up/Wet-Up
    ceTTypeWULD = 5,     ///< Wet-Up/Linear-Dry-Calc
};

/// The test mode being used
enum NCfp::ETMode : uint_16
{
    ceTModePressureScan = 0,  ///< Pressure scan more(normal)
    ceTModePressureStep = 1,  ///< Pressure step mode
};

enum NCfp::EFlowMeterToUse : uint_16
{
    ceUseFlowMeterLow,
    ceUseFlowMeterHigh,
    ceUseFlowMeterBoth,
};

enum class NCfp::EBubblePointMethod : uint_16
{
    ceMeasured,
    ceEstimated,
};
enum class NCfp::EPoreSizeRange : uint_16
{
    ceSmall,
    ceMedium,
    ceLarge,
};

struct NCfp::SFluid
{
    CStr szName;
    EFluid eFluid;
    double dGalwickSurfaceTension;
    double dSilwickSurfaceTension;
    double dPorewickSurfaceTension;
    double dOtherSurfaceTension;
};

void Serialize(CDataArchive& ar, NCfp::SFluid& rData);

struct NCfp::SSampleShape
{
    CStr szName;
    EShape eShape;
    double dCircleDiameter;
    double dSquareSide;
    double dOtherArea;
};

void Serialize(CDataArchive& ar, NCfp::SSampleShape& rData);
struct NCfp::SPoreShape
{
    CStr szName;
    double dShapeFactor;
};

void Serialize(CDataArchive& ar, NCfp::SPoreShape& rData);

/// The data for a curve.  This can be either the collected/raw transducer readings, or the lohm
/// corrected information.
struct NCfp::SRawCurve
{
    /// Which pressure transducer was used in the point.
    enum EWhichPressure : uint_16
    {
        ceLowPressure,
        ceHighPressure,
    };

    enum EWhichFlow : uint_16
    {
        ceLowFlow,
        ceMediumFlow,
        ceHighFlow,
    };

    SI::V::s<double> vdTime;              ///< Time after start of analysis of the point.
    std::vector<EWhichPressure> vePressureGauge;  ///< Which transducer was used for pressure reading.
    std::vector<EWhichFlow> veFlowMeter;  ///< Which transducer was used for the flow reading.
    SI::V::psig<double> vdPressure;       ///< The pressure of the collected data point.
    SI::V::cm3_per_min<double> vdFlow;    ///< The flow of the collected data point.
    SI::V::C<double> vdTemperature;       ///< The temperature of the collected data point.
};

void Serialize(CDataArchive& ar, NCfp::SRawCurve& rData);

/// The data needed for the wet/dry flow vs pressure graph.
struct NCfp::SWetDryData
{
    NCfp::SRawCurve sWetCurve;  ///< The raw data for the wet curve.
    NCfp::SRawCurve sDryCurve;  ///< The raw data for the dry curve, or empty if that was not collected.
};

void Serialize(CDataArchive& ar, NCfp::SWetDryData& rData);

#endif
