#ifndef CALCACCU_HPP
#define CALCACCU_HPP
//-------------------------------------------------------------------------------------------------
/// @file Calculation objects for the base accupyc data reduction.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include <cpp/numvect.hpp>
#include <cpp/numdist.hpp>
//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------

// ================================================================================================
/// Calculate the sample volume using the traditional P1, P2 measurements.
class CCalcSampleVolumeP1P2 final
{
public:
    typedef CNumVector<float_64> TVector;

    CCalcSampleVolumeP1P2() = default;
    CCalcSampleVolumeP1P2(const CCalcSampleVolumeP1P2&) = delete;
    CCalcSampleVolumeP1P2(CCalcSampleVolumeP1P2&&) = default;
    CCalcSampleVolumeP1P2& operator=(const CCalcSampleVolumeP1P2&) = delete;
    ~CCalcSampleVolumeP1P2() = default;

    void SetCellVolume(double dVolume);
    void SetExpansionVolume(double dVolume);
    void SetIsReversed(bool fIsReversed);
    void SetCupVolume(double dCupVolume);

    double Volume(double dP1, double dP2) const;
    TVector Volume(const TVector& adP1, const TVector& adP2) const;

private:
    double m_dCellVolume = 0.;
    double m_dExpansionVolume = 0.;
    bool m_fIsReversed = false;
    double m_dCupVolume = 0.;
};

// ================================================================================================
class CCalcAccuPycCore
{
public:
    typedef CNumVector<float_64> TVector;
    typedef CNumVector<uint_8> TFlagVector;
    typedef CNumVector<bool> TBoolVector;

    CCalcAccuPycCore() = default;
    CCalcAccuPycCore(double dMass);

    void SetSampleMass(double dMass);

    double Density(double dVolume) const;
    TVector Density(const TVector& adVolume) const;

    double SpecificGravity(double dVolume, double dT) const;
    TVector SpecificGravity(const TVector& adVolume, const TVector& adT) const;

    // Asphalt-specific calculations.  Quantities are adjusted to 60 deg F.
    double AsphaltVolume60F(double dVolume, double dT) const;
    TVector AsphaltVolume60F(const TVector& adVolume, const TVector& adT) const;

    double AsphaltDensity60F(double dVolume, double dT) const;
    TVector AsphaltDensity60F(const TVector& adVolume, const TVector& adT) const;

    double AsphaltSpecificGravity60F(double dVolume, double dT) const;
    TVector AsphaltSpecificGravity60F(const TVector& adVolume, const TVector& adT) const;

    /// @{
    /// These are convenience functions for calculating statistics on
    /// a collected quantity; if exclusion flags are used, only includes runs
    /// the user has not excluded.
    static CNumDistribution Statistics(const TVector& adQuant);
    static CNumDistribution Statistics(const TVector& adQuant, const TFlagVector& afExclude);
    static CNumDistribution Statistics(const TVector& adQuant, const TBoolVector& afExclude);
    /// @}

private:
    double m_dSampleMass = 0.;
};

// ================================================================================================
class CCalcAccuPyc : public CCalcAccuPycCore
{
public:
    CCalcAccuPyc() = default;
    CCalcAccuPyc(double dMass, double dEnvelopeDensity, double dSolidsDensity, double dLiquidDensity);

    void SetEnvelopeDensity(double dDensity);
    void SetSolidsDensity(double dDensity);
    void SetLiquidDensity(double dDensity);

    double PercentPorosity(double dVolume) const;
    TVector PercentPorosity(const TVector& adVolume) const;

    double TotalPoreVolume(double dVolume) const;
    TVector TotalPoreVolume(const TVector& adVolume) const;

    double TotalSolidsConcentration(double dVolume) const;
    TVector TotalSolidsConcentration(const TVector& adVolume) const;
    bool CanCalculateTotalSolidsConc() const;

private:
    double m_dSampleDensity = 0.;
    double m_dEnvelopeDensity = 0.;
    double m_dSolidsDensity = 0.;
    double m_dLiquidDensity = 0.;
};

//-------------------------------------------------------------------------------------------------
#endif  // CALCACCU_HPP
