#ifndef CALC_ACCUPYC_HPP
#define CALC_ACCUPYC_HPP
//-------------------------------------------------------------------------------------------------
/// @file
/// The fundamental accupyc calculations.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include <cpp/cstr.hpp>
#include <vector>
#include <pycnometry/calcaccu.hpp>
//-------------------------------------------------------------------------------------------------
// Globals, Constants, Defines, Macros, Forward References
//-------------------------------------------------------------------------------------------------
class CDataArchive;

// ================================================================================================
namespace NPyc
{
    namespace NP1P2P0
    {
        /// The direction of fill, either sample chamber first, or reference chamber first.
        enum EFillDirection : uint_16
        {
            ceFillSample,
            ceFillReference,
        };
        /// Calculate the compressibility of the gas using CoolProp.
        /// @param dPpsia The pressure of the gas in psia units.
        /// @param dTAbsolute The temperature of the gas in K.
        /// @param szGas The name of the gas to be passed to CoolProp.  This must be a fluid that
        /// CoolProp knows, otherwise infinity is returned.
        /// @returns The compressibility of the gas.
        double GasCompressibilityPsiaKelvin(double dPpsia, double dTAbsolute, const CStr& cszGas);

        /// @param dP0 The pressure of the ambient chamber before the equilibration.
        /// @param dP1 The pressure of the fill chamber before the equilibration.
        /// @param dP2 The pressure of the both chambers after the equilibration.
        /// @param dT The temperature at which the measurements were taken.
        /// @param cszGas The name of the gas that was used.
        /// @returns the Vfill over VExpand ratio for the given run data. (internal units of psia and
        /// degrees C).
        double CalcVolumeRatio(double dP0, double dP1, double dP2, double dT, const CStr& cszGas);

        /// The information for a chamber calibration.
        struct SChamberCalibration;

        /// The information for an insert calculation.
        struct SInsertCalibration;

        /// The calculation class for volume calibration.  Construct one with the fixed parameters of
        /// an analysis, and then call CalcVolumes for each cycle.
        class CCalcVolumeCalibration;

        /// The calculation class for the sample volume.  Construct one with the fixed parameters of an
        /// analysis, and then call CalcVolume for each cycle.
        class CCalcSampleVolume;
    }
}

struct NPyc::NP1P2P0::SChamberCalibration
{
    /// @{ The chamber volumes for the Fill-Sample direction.
    double dFSVolumeSample = 10.0;
    double dFSVolumeReference = 10.0;
    /// @}

    /// @{ The chamber volumes for the Fill-Reference direction.
    double dFRVolumeSample = 10.0;
    double dFRVolumeReference = 10.0;
    /// @}

    /// The temperature for the volumes.
    double dTemperature = 20.0;

    /// The thermal expansion coefficient for the chamber.
    double dThermalExpansion = 0.0;
};
void Serialize(CDataArchive& rAr, NPyc::NP1P2P0::SChamberCalibration& rsChamberCal);

struct NPyc::NP1P2P0::SInsertCalibration
{
    SInsertCalibration(double dVolume, double dTemperature, double dThermalExpansion);
    SInsertCalibration() = default;
    SInsertCalibration(const SInsertCalibration&) = default;
    SInsertCalibration& operator=(const SInsertCalibration&) = default;

    double dVolume = 0.0;            ///< The volume of the insert.
    double dTemperature = 20.0;      ///< The temperature at which the volume was measured.
    double dThermalExpansion = 0.0;  ///< The thermal expansion coefficient for the insert.
};
void Serialize(CDataArchive& rAr, NPyc::NP1P2P0::SInsertCalibration& rsInsertCal);

// ============================================================================================

class NPyc::NP1P2P0::CCalcVolumeCalibration final
{
public:
    using EFillDirection = NPyc::NP1P2P0::EFillDirection;
    using SInsertCalibration = NPyc::NP1P2P0::SInsertCalibration;

    /// The return value from the chamber calibration calculation.
    struct SChamberVolumes
    {
        double dSampleVolume;     ///< The calculated sample chamber volume.
        double dReferenceVolume;  ///< The calculated reference chamber volume.
    };

    CCalcVolumeCalibration(EFillDirection eFillDirection,
        double dChamberThermalExpansion,
        double dVolumeTemperature,
        const SInsertCalibration& csReferenceVolume,
        bool fHasCup,
        const SInsertCalibration& csCupVolume);
    CCalcVolumeCalibration(const CCalcVolumeCalibration&) = default;
    CCalcVolumeCalibration(CCalcVolumeCalibration&&) = default;
    CCalcVolumeCalibration& operator=(const CCalcVolumeCalibration&) = default;
    ~CCalcVolumeCalibration() = default;

    /// @param dRempty Volume ratio for the empty expansion.
    /// @param dTempty Temperature at the time of the empty expansion.
    /// @param dRreference Volume ratio for the expansion with reference volume.
    /// @param dTreference Temperature at the time of the reference expansion.
    /// @returns The calibrated volumes from the calibration data.
    SChamberVolumes CalcVolumes(double dRempty,
        double dTempty,
        double dRreference,
        double dTreference);

private:
    const EFillDirection m_eFillDirection;    ///< The fill direction when the data was taken.
    const double m_dChamberThermalExpansion;  ///< The thermal expansion coefficient of the chamber.
    const double m_dVolumeTemperature;  ///< The temperature at which to calculate the volumes.
    const SInsertCalibration m_sReferenceVolume;  ///< The calibration ball information.
    const bool m_fHasCup;                         ///< Do we have a cup in the calibration.
    const SInsertCalibration m_sCupVolume;        ///< What are the cup parameters.
};

// ============================================================================================
/// The calculation class for the sample volume.  Construct one with the fixed parameters of an
/// analysis, and then call CalcVolume for each cycle.
class NPyc::NP1P2P0::CCalcSampleVolume final
{
public:
    using EFillDirection = NPyc::NP1P2P0::EFillDirection;
    using SChamberCalibration = NPyc::NP1P2P0::SChamberCalibration;
    using SInsertCalibration = NPyc::NP1P2P0::SInsertCalibration;

    CCalcSampleVolume(const CStr& cszGas,
        EFillDirection eFillDirection,
        const SChamberCalibration& csChamberCal,
        std::initializer_list<SInsertCalibration> asInserts);
    CCalcSampleVolume(const CStr& cszGas,
        EFillDirection eFillDirection,
        const SChamberCalibration& csChamberCal,
        std::vector<SInsertCalibration>&& rvInserts);
    CCalcSampleVolume(const CStr& cszGas,
        EFillDirection eFillDirection,
        const SChamberCalibration& csChamberCal,
        const std::vector<SInsertCalibration>& cvInserts);
    CCalcSampleVolume(const CStr& cszGas,
        EFillDirection eFillDirection,
        const SChamberCalibration& csChamberCal);
    CCalcSampleVolume(const CStr& cszGas,
        EFillDirection eFillDirection,
        const SChamberCalibration& csChamberCal,
        const SInsertCalibration& rInsert1);
    CCalcSampleVolume(const CStr& cszGas,
        EFillDirection eFillDirection,
        const SChamberCalibration& csChamberCal,
        const SInsertCalibration& rInsert1,
        const SInsertCalibration& rInsert2);
    CCalcSampleVolume(const CCalcSampleVolume&) = default;
    CCalcSampleVolume(CCalcSampleVolume&&) = default;
    CCalcSampleVolume& operator=(const CCalcSampleVolume&) = default;
    ~CCalcSampleVolume() = default;

    /// @returns the nominal (non temperature adjusted) volume available for the sample.
    double NominalCellVolume() const;
    /// @param dP0 The pressure of the ambient chamber before the equilibration.
    /// @param dP1 The pressure of the fill chamber before the equilibration.
    /// @param dP2 The pressure of the both chambers after the equilibration.
    /// @param dT The temperature at which the measurements were taken.
    /// @returns the sample volume for the run with the given pressures and temperature.
    double CalcVolume(double dP0, double dP1, double dP2, double dT) const;

private:
    const CStr m_cszGas;  ///< The name of the gas that was used during the run.
    const EFillDirection m_ceFillDirection;              ///< The fill/expansion order.
    const SChamberCalibration m_csChamberCal;            ///< The parameters of the chamber.
    const std::vector<SInsertCalibration> m_cvsInserts;  ///< All the inserts/cups used.
};


#endif  // CALC_ACCUPYC_HPP
