#ifndef CALCFOAM_HPP
#define CALCFOAM_HPP
//-------------------------------------------------------------------------------------------------
/// @file Calculation objects for the FoamPyc data reduction.
//-------------------------------------------------------------------------------------------------
// Includes
//-------------------------------------------------------------------------------------------------
#include <cpp/numvect.hpp>
#include <math.h>

//=================================================================================================
class CCalcFoamPycCommon
{
public:
    // Method parameter setup. Only one method setup should be called to setup the appropriate
    // parameters.
    void SetEnteredGeometryVolume(double dVolume);
    void SetCubeGeometry(double dLength, int iNumberOfPieces);
    void SetCylinderGeometry(double dDiameter, double dLength, int iNumberOfPieces);
    void SetRectangleGeometry(double dLongEdge,
        double dShortEdge,
        double dOtherEdge,
        int iNumberOfPieces);

    void SetSampleMass(double dMass);       ///< Set the sample mass in grams.
    void SetResinDensity(double dDensity);  ///< Set the resin density in grams per cc.

    double GeometricVolume() const;
    double ResinVolume() const;
    double ResinVolumePercent() const;

protected:
    int NumberOfPieces() const;
    double CubeLength() const;
    double CylinderDiameter() const;
    double CylinderLength() const;
    double RectangleLongEdge() const;
    double RectangleShortEdge() const;
    double RectangleOtherEdge() const;

private:
    double m_dVolume = 0.0;  ///< Pre-calculated geometric volume.  (cm^3)

    double m_dSampleMass = 0.0; /// (g)
    double m_dResinDensity = 0.0; /// (g/cm^3)

    // Cached values for Method A Active area calculation.
    int m_iNumberOfPieces = 0;
    double m_dCubeLength = 0.0; /// (cm)
    double m_dCylinderDiameter = 0.0; /// (cm)
    double m_dCylinderLength = 0.0; /// (cm)
    double m_dRectangleLongEdge = 0.0; /// (cm)
    double m_dRectangleShortEdge = 0.0; /// (cm)
    double m_dRectangleOtherEdge = 0.0; /// (cm)
};

//=================================================================================================
class CCalcFoamPycMethodA final : public CCalcFoamPycCommon
{
public:
    // Method parameter setup. Only one method setup should be called to setup the appropriate
    // parameters. Note: base class must be initialized first.
    void SetEnteredGeometryArea(double dArea);
    void SetCubeSkins(int iNumberOfSkins);
    void SetCylinderSkins(int iNumberOfSkins);
    void SetRectangleSkins(int iNumberOfLargeSkins, int iNumberOfSmallSkins, int iNumberOfOtherSkins);
    // Measurement parameter setup. Only one measurement setup should be called.
    void SetChordLength(double dChordLength);
    void SetCellDiameter(double dCellDiameter);

    double ActiveArea() const;
    double VolumeCellsCutOpen() const;
    double OpenCellPercent(double dVSamp) const;
    double ResinPlusClosedCellPercent(double dVSamp) const;

private:
    double m_dActiveArea = 0.0; /// (cm^2)
    double m_dVolCellsCutOpen = 0.0; /// (cm^3)
};

//=================================================================================================
class CCalcFoamPycMethodB final : public CCalcFoamPycCommon
{
public:
    double VolumeCellsCutOpen(double dVSamp1, double dVSamp2) const;
    double ResinPlusClosedCellVolume(double dVSamp1, double dVSamp2) const;
    double OpenCellVolume(double dVSamp1, double dVSamp2) const;
    double ResinPlusClosedCellPercent(double dVSamp1, double dVSamp2) const;
    double OpenCellPercent(double dVSamp1, double dVSamp2) const;
};

//=================================================================================================
class CCalcFoamPycMethodC final : public CCalcFoamPycCommon
{
public:
    double OpenCellVolume(double dVSamp) const;
    double OpenCellPercent(double dVSamp) const;
    double ResinPlusClosedCellPercent(double dVSamp) const;
};

//=================================================================================================
class CCalcFoamPycMethodD final : public CCalcFoamPycCommon
{
public:
    using TVector = CNumVector<float_64>;
    using TFlagVector = CNumVector<uint_8>;
    using TBoolVector = CNumVector<bool>;

    double AverageVolumeChange(const TVector& adVSamp,
        const TVector& adP,
        const TFlagVector& afExclude) const;
    double AverageVolumeChange(const TVector& adVSamp,
        const TVector& adP,
        const TBoolVector& afExclude) const;
    double PercentAverageVolumeChange(double dAverageVolumeChange) const;
    double PercentAverageVolumeChange(const TVector& adVSamp,
        const TVector& adP,
        const TFlagVector& afExclude) const;
};

//=================================================================================================
class CCalcFoamPycMethodE final : public CCalcFoamPycCommon
{
public:
    double VolumeChange(double dVSamp1, double dVSamp3) const;
    double PercentFracturedCells(double dVSamp1, double dVSamp3) const;
};

#endif  // CALCFOAM_HPP
