'''Tests for integration of ui, rest, and analysis of i1350 application'''

import os
import sys
import pathlib
import subprocess
import time
from contextlib import contextmanager
from django.test import LiveServerTestCase, tag
from django.contrib.auth.models import User
from django.utils import timezone
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec
from selenium.common.exceptions import (
    StaleElementReferenceException, NoSuchElementException, ElementClickInterceptedException
)
from selenium.webdriver.common.keys import Keys
from pyc_sample.models import Insert
from mic_python_lib.models import LogMessage

ROOT_PATH = str(
    os.path.abspath(
        pathlib.Path(__file__).parent / pathlib.Path('..') / pathlib.Path('..') /
        pathlib.Path('..')
    )
)
REST_PATH = str(ROOT_PATH / pathlib.Path('rest'))


@tag("integration")
class TestRegressions(LiveServerTestCase):
    '''Test case for regression integration tests'''

    # specify port so that rest server and analysis application can communicate
    port = 8000

    # re-create initital data from migrations
    serialized_rollback = True

    @classmethod
    def setUpClass(cls):
        "Start analysis, UI, and web driver processes"

        cls.exceptions = (NoSuchElementException, StaleElementReferenceException)

        build_path = default_build_path()
        set_ld_library_path(build_path)
        set_python_path(build_path)
        cls.app = run_1350app(build_path)
        cls.ui = run_client()
        super(TestRegressions, cls).setUpClass()
        time.sleep(5)  # wait for ui and analysis to start
        options = Options()
        options.add_experimental_option("excludeSwitches", ['enable-automation'])
        options.add_argument('--headless')
        options.add_argument('disable-popup-blocking')
        cls.driver = webdriver.Chrome(
            executable_path='/usr/local/bin/chromedriver', chrome_options=options
        )

    @classmethod
    def tearDownClass(cls):
        "Politely ask analysis app and UI to stop, and wait for termination"

        subprocess.call(["killall", "1350analysis"])
        cls.driver.quit()
        cls.app.terminate()
        cls.ui.terminate()
        cls.app.wait()
        cls.ui.wait()
        super(TestRegressions, cls).tearDownClass()

    def setUp(self):
        "Setup database for each test method"

        # assert that analysis app and UI are still running
        self.assertIsNone(self.app.poll())
        self.assertIsNone(self.ui.poll())

    def click(self, element_id):
        "Wait for element to be clickable and click it"

        self.assertTrue(
            WebDriverWait(self.driver, 20, 1,
                          self.exceptions).until(ec.element_to_be_clickable((By.ID, element_id)))
        )
        while True:
            try:
                self.driver.find_element(By.ID, element_id).click()
                break
            except (NoSuchElementException, StaleElementReferenceException):
                time.sleep(1)
                continue

    def press_virtual_key(self, key):
        "Wait for virtual keyboard element to exist and then click on it"

        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.presence_of_element_located((By.XPATH, f'//div[@data-skbtn={key}]'))
            )
        )
        self.driver.find_element(By.XPATH, f'//div[@data-skbtn={key}]').click()

    def send_keys(self, element_id, keys):
        "Wait for element to exist and then send keys to it"

        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.presence_of_element_located((By.ID, element_id))
            )
        )
        self.driver.find_element(By.ID, element_id).send_keys(keys)

    def wait_forever_until_exists(self, element_search, search_by=By.ID):
        "Wait forever for a process to finish until an element exists"

        while True:
            try:
                self.driver.find_element(search_by, element_search)
                break
            except self.exceptions:
                time.sleep(1)
                continue

    def wait_for_url(self, url):
        "Wait for url to be updated"

        self.assertTrue(WebDriverWait(self.driver, 20, 1, self.exceptions).until(ec.url_to_be(url)))

    def wait_until_not_exists(self, element_search, search_by=By.ID):
        "Wait until an element disappears"

        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until_not(
                ec.presence_of_element_located((search_by, element_search))
            )
        )

    def test_general_settings_go_back_discard(self):
        "Bug 555: General settings doesn't navigate away on save/discard changes"

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # go to settings page
        self.click("nav-link_settings")
        # click on first setting (language)
        self.click("form_language")
        # click on second option (Chinese)
        self.click("language_zn-CN")
        # click check button to accept change
        self.click("keyboard-check-btn")
        # click go back button
        self.click("action-btn_previous")
        # click discard changes button
        self.click("query-btn_discard")
        # wait for page change
        self.wait_for_url("http://localhost:3000/#/analysis")
        # click on settings
        self.click("nav-link_settings")
        # assert changes have been discarded
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.element_to_be_clickable((By.ID, "form_language"))
            )
        )
        self.assertEqual(self.driver.find_element(By.ID, "form_language").text, "English")

    def test_general_settings_go_back_save(self):
        "Bug 555: General settings doesn't navigate away on save/discard changes"

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # go to settings page
        self.click("nav-link_settings")
        # click on first setting (language)
        self.click("form_language")
        # click on second option (Chinese)
        self.click("language_zn-CN")
        # click check button to accept change
        self.click("keyboard-check-btn")
        # click go back button
        self.click("action-btn_previous")
        # click save changes button
        self.click("query-btn_save")
        # wait for page change
        self.wait_for_url("http://localhost:3000/#/analysis")
        # click on settings
        self.click("nav-link_settings")
        # assert changes have been saved
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.element_to_be_clickable((By.ID, "form_language"))
            )
        )
        self.assertEqual(self.driver.find_element(By.ID, "form_language").text, "Chinese")

    def test_analysis_error_go_back(self):
        "Bug 561: Cancel and Okay buttons not working when starting analysis without calibration"

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # click on mass field
        self.click("form_mass")
        # input mass
        self.send_keys("mass_keyboard", "5.0")
        # click check button to accept input
        self.click("keyboard-check-btn")
        # click start button
        self.click("action-btn_start")
        # click start button again
        self.click("action-btn_start")
        # assert error message
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.presence_of_element_located((By.ID, "start-analysis-error"))
            )
        )
        self.assertEqual(
            self.driver.find_element(By.ID, "start-analysis-error").text,
            "Cannot start an analysis. No calibration available."
        )
        # click go back button
        self.click("action-btn_previous")
        # click cancel button
        self.click("action-btn_cancel")
        # assert analysis dialog is visible
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.presence_of_element_located((By.ID, "form_mass"))
            )
        )

    def test_keyboard_enter_accept_values(self):
        "Bug 574: Connected keyboard enter key does not accept values"

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # click on description field
        self.click("form_name")
        # input value
        self.send_keys("name_keyboard", "test")
        # press enter key
        self.send_keys("name_keyboard", Keys.ENTER)
        # click on mass field
        self.click("form_mass")
        # input value
        self.send_keys("mass_keyboard", "1.00")
        # press enter key
        self.send_keys("mass_keyboard", Keys.ENTER)
        # wait for keyboard dialog to close
        self.wait_until_not_exists("mass_keyboard")
        # assert changes were accepted
        self.assertEqual(self.driver.find_element(By.ID, "form_name").text, "test")
        self.assertEqual(self.driver.find_element(By.ID, "form_mass").text, "1.0000")

    def test_blank_screen_pressure_calibration(self):
        "Bug 629: Screen goes blank during pressure calibration"

        User.objects.create_user("service", password="107.104.103", is_active=False)

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # go to settings page
        self.click("nav-link_settings")
        # go to service page
        self.click("menu-item_service")
        # go to pressure calibration
        self.click("menu-item_calpress")
        # wait for start button
        self.wait_forever_until_exists("action-btn_start")
        # press start button
        self.click("action-btn_start")
        # wait for gauge SN field
        self.wait_forever_until_exists("form_Gauge SN")
        # click on field
        self.click("form_Gauge SN")
        # click close button
        self.click("dialog-close-btn")
        # assert form is still there and value is still blank
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.element_to_be_clickable((By.ID, "form_Gauge SN"))
            )
        )
        self.assertEqual(self.driver.find_element(By.ID, "form_Gauge SN").text, "")

    def test_temperature_configuration_initialized_blank(self):
        """Bug 623: Service, set serial number screen: temperature configuration is initialized
        as blank"""

        User.objects.create_user("service", password="107.104.103", is_active=False)

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # go to settings page
        self.click("nav-link_settings")
        # go to service page
        self.click("menu-item_service")
        # go to set serial number
        self.click("menu-item_serial")
        # assert temperature configuration is not blank
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.element_to_be_clickable((By.ID, "form_is_temp_control"))
            )
        )
        self.assertEqual(self.driver.find_element(By.ID, "form_is_temp_control").text, "TS")

    def test_invalid_value_messages_disappear(self):
        "Bug 637: Invalid value message disappear after a split second"

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # click on mass field
        self.click("form_mass")
        # input value
        self.send_keys("mass_keyboard", ".")
        # wait a "split-second"
        time.sleep(1)
        # assert error appears
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1,
                          self.exceptions).until(ec.element_to_be_clickable((By.ID, "mass_error")))
        )
        self.assertEqual(
            self.driver.find_element(By.ID, "mass_error").text, "Mass is not a number."
        )

    def test_close_dialog_with_open_keyboard(self):
        "Bug 802: Screen is unresponsive when keyboard is open and popup dialog appears"

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # click on description field
        self.click("form_name")
        # input value
        self.send_keys("name_keyboard", "test")
        # click on method field
        with self.assertRaises(ElementClickInterceptedException):
            self.click("form_parameters")
        # assert description field still has focus
        self.assertEqual(
            self.driver.find_element_by_id("name_keyboard"), self.driver.switch_to.active_element
        )
        # open a popup dialog
        LogMessage.objects.create(
            time=timezone.now(), action=3, severity=3, message="Test", acknowledged=False
        )
        # assert description field still has focus
        self.assertEqual(
            self.driver.find_element_by_id("name_keyboard"), self.driver.switch_to.active_element
        )
        # close popup dialog
        self.click("query-btn_ok")
        # input more text
        self.send_keys("name_keyboard", "123")
        # click check button to accept input
        self.click("keyboard-check-btn")
        # assert changes were accepted
        self.assertEqual(self.driver.find_element(By.ID, "form_name").text, "test123")

    def test_service_password(self):
        "Test that password widget works with virtual keyboard"

        User.objects.create_user("service", password="1234", is_active=True)

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # go to settings page
        self.click("nav-link_settings")
        # go to service page
        self.click("menu-item_service")
        # click on password field
        self.click("form_password")
        # click on "4 3 2 1" keys on virtual keyboard
        for number in "4321":
            self.press_virtual_key(number)
        # click check button to accept input
        self.click("keyboard-check-btn")
        # assert error message
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.presence_of_element_located((By.ID, "form-errors"))
            )
        )
        # click on password field
        self.click("form_password")
        # click on "1 2 3 4" keys on virtual keyboard
        for number in "1234":
            self.press_virtual_key(number)
        # click check button to accept input
        self.click("keyboard-check-btn")
        # click check button to submit form
        self.click("action-btn_next")
        # assert service menu appears
        self.assertTrue(
            WebDriverWait(self.driver, 20, 1, self.exceptions).until(
                ec.presence_of_element_located((By.ID, "menu-item_serial"))
            )
        )

    @tag("slow")
    def test_run_calibration_and_analysis(self):
        "Test that the calibration and analysis run sucessfully"

        Insert.objects.create(
            name="ref-vol",
            volume=5,
            temperature=20,
            thermal_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="R"
        )
        Insert.objects.create(
            name="cup",
            volume=0.5,
            temperature=20,
            thermal_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="C"
        )
        User.objects.create_user("service", password="107.104.103", is_active=False)

        # go to analysis page and refresh to sync ui with reset database
        self.driver.get("http://localhost:3000/#/analysis")
        self.driver.refresh()
        # go to settings page
        self.click("nav-link_settings")
        # go to service page
        self.click("menu-item_service")
        # go to volume calibration
        self.click("menu-item_calvol")
        # press start button
        self.click("action-btn_start")
        # wait forever until okay button appears
        self.wait_forever_until_exists("query-btn_ok")
        # accept "make sure the sample chamber is empty" message
        self.click("query-btn_ok")
        # wait for query to be dismissed so that next wait works
        self.wait_until_not_exists("query-btn_ok")
        # wait forever until okay button appears
        self.wait_forever_until_exists("query-btn_ok")
        # accept "insert the cup and reference volume" message
        self.click("query-btn_ok")
        # wait until calibration is complete
        self.wait_forever_until_exists("action-btn_check")
        # click on check button to accept results
        self.click("action-btn_check")
        # go to volume tab (no id on link built by Report Model)
        self.driver.get("http://localhost:3000/#/settings/service/all_cal/page-3")
        # wait for page to load
        self.wait_forever_until_exists("tr", By.TAG_NAME)
        # assert there is one calibration (and one header row)
        self.assertEqual(len(self.driver.find_elements(By.TAG_NAME, "tr")), 2)

        # go to analysis page
        self.click("nav-link_analysis")
        # click on mass field
        self.click("form_mass")
        # input mass
        self.send_keys("mass_keyboard", "5.0")
        # click check button to accept input
        self.click("keyboard-check-btn")
        # click start button
        self.click("action-btn_start")
        # click start button again on analysis wizard page
        self.click("action-btn_start")
        # wait until analysis is complete
        self.wait_forever_until_exists("action-btn_check")
        # click on check button to go back to analysis page
        self.click("action-btn_check")
        # go to records page
        self.click("nav-link_record")
        # wait for page to load
        self.wait_forever_until_exists("ag-cell", By.CLASS_NAME)
        # assert there is one analyis result
        self.assertEqual(len(self.driver.find_elements(By.CLASS_NAME, "ag-cell")), 1)


def default_build_path():
    "Return path to cmake build directory"

    return str(ROOT_PATH / pathlib.Path('build-debug'))


def run_1350app(build_path):
    "Start analysis c++ application"

    try:
        return subprocess.Popen([f'{build_path}/analysis/1350analysis/1350analysis', '--simulate'],
                                stdout=subprocess.DEVNULL,
                                stderr=subprocess.STDOUT)
    except OSError:
        print("Could not start C++ analysis application")
        raise


def run_client():
    "Start ui typescript application"

    with cd_context(ROOT_PATH + "/ui"):
        try:
            return subprocess.Popen(['npm', 'run', 'start'],
                                    stdout=subprocess.DEVNULL,
                                    stderr=subprocess.STDOUT)
        except OSError:
            print("Could not start Typescript UI")
            raise


def python_extension_path(build_path):
    "Get the path to the python boost extension"

    return str(build_path / pathlib.Path('analysis/python'))


def analysis_extension_path(build_path):
    "Get the path to the analysis boost extension"

    return str(build_path / pathlib.Path('mic-cpp/python'))


def set_ld_library_path(build_path):
    "Set the ld_library_path environment variable"

    cpplib_path = build_path / pathlib.Path('mic-cpp/cpp')
    accupyclib_path = build_path / pathlib.Path('analysis/accupyc')
    ld_library_path = ':'.join([str(cpplib_path), str(accupyclib_path)])
    os.environ['LD_LIBRARY_PATH'] = ld_library_path


def set_python_path(build_path):
    "Set the pythonpath environment variable"

    ext = python_extension_path(build_path)
    analysis_path = analysis_extension_path(build_path)
    sys.path.insert(0, ext)
    sys.path.insert(0, REST_PATH)
    sys.path.insert(0, analysis_path)
    os.environ['PYTHONPATH'] = ":".join([ext, REST_PATH, analysis_path])


@contextmanager
def cd_context(path):
    "Context managed change of current directory."

    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)
