"Support functionality to limit the number of UIs."
import json
import datetime
from django.http import JsonResponse
from django.utils import timezone
from .models import UIAttached

# The maximum number of external displays that are allowed.
MAX_EXTERNAL_UI = 2

# =================================================================================================
def update_ui_data(request):
    "@return False if the request from that ip/uuid should be blocked."

    ip = _get_client_ip(request)
    uuid = _get_uuid(request)

    # don't do any checks for local UI.
    if ip == '127.0.0.1':
        return True

    # if it is already connected, all is OK.
    if _is_connected(ip, uuid):
        return True;

    # if it is blocked, don't show this.
    if _is_blocked(ip, uuid):
        return False;

    # see if we are allowed another connection?
    if _num_ui_connected() >= MAX_EXTERNAL_UI:
        return False

    # if we were blocked, unblock
    try:
        ui = UIAttached.objects.get(source_ip=ip, source_uuid=uuid, blocked=True)
        ui.last_date_time = timezone.now()
        ui.blocked = False
        ui.save()
    except UIAttached.DoesNotExist:
        ui = UIAttached.objects.create(source_ip=ip, source_uuid=uuid, last_date_time =
                                       timezone.now(), blocked=False)

    return True


def remove_from_list_view(request):
    "Remove the UI from the list of active uis."

    # actually this just blocks the ip/uuid pair.  That is so that if there is a new request queued,
    # it will not undo the removal.
    ip = _get_client_ip(request)
    uuid = _get_uuid(request)
    try:
        ui = UIAttached.objects.get(source_ip=ip, source_uuid=uuid)
        ui.blocked = True
        ui.save()
    except UIAttached.DoesNotExist:
        pass
    return JsonResponse({})


def state_view(request):
    "Return (in json) the information that the react BlockedContents component needs."
    active = [
        { "ip": ui.source_ip, "uuid": ui.source_uuid }
        for ui in UIAttached.objects.filter(blocked=False)
    ]
    return JsonResponse({
        "active": active,
        "source_ip": _get_client_ip(request),
    })


def block_view(request):
    "View the blocks the given ui and allows this UI to be setup."
    source_ip = _get_client_ip(request)
    source_uuid = _get_uuid(request);
    block_ip, block_uuid = _get_block(request)
    for ui in UIAttached.objects.filter(source_ip=block_ip, source_uuid=block_uuid):
        ui.blocked = True
        ui.blocked_by_ip = source_ip
        ui.blocked_by_guild = source_uuid
        ui.save()
    UIAttached.objects.filter(source_ip=source_ip, source_uuid=source_uuid).delete()
    UIAttached.objects.create(
        source_ip=source_ip, source_uuid=source_uuid, blocked=True, last_date_time=timezone.now())
    return JsonResponse({})

# =================================================================================================
def _get_client_ip(request):
    "@return the ip address of ui request."
    ip = '127.0.0.1'
    forwarded = request.META.get('HTTP_FORWARDED')
    if forwarded:
        first = forwarded.split()[0]
        if first.startswith("for="):
            ip = first.split("=")[1]
    return ip


def _get_uuid(request):
    "@return the json uuid parameter from the request body."
    return json.loads(request.body.decode('utf-8'))['uuid']


def _get_block(request):
    "@return the json uuid parameter from the request body."
    data = json.loads(request.body.decode('utf-8'))
    return data['blockIP'], data['blockUUID']


def _clear_old_uiattached():
    "Remove the uis from the known list have not been heard from in 5 minutes."
    UIAttached.objects.filter(
        last_date_time__lte=timezone.now() - datetime.timedelta(minutes=5)
    ).delete()


def _num_ui_connected():
    "@return how many uis are currently connected."
    _clear_old_uiattached()
    return UIAttached.objects.filter(blocked=False).count()


def _is_connected(ip, uuid):
    "@return True if the ip and uuid is already connected."
    try:
        ui = UIAttached.objects.get(source_ip=ip, source_uuid=uuid, blocked=False)
        ui.last_date_time = timezone.now()
        ui.save()
        return True
    except UIAttached.DoesNotExist:
        return False


def _is_blocked(ip, uuid):
    "@return True if the ip and uuid is available, but blocked."
    try:
        ui = UIAttached.objects.get(source_ip=ip, source_uuid=uuid, blocked=True)
        ui.last_date_time = timezone.now()
        ui.save()
        return True
    except UIAttached.DoesNotExist:
        return False
