'''
The python functionality that supports datetime functionality
'''
import datetime
import icu
import pytz
from django import forms
from django.urls import resolve
from django.urls.exceptions import Resolver404
from django.utils import timezone
from .models import Settings
from .settings import set_language, udp_set_translate_locale

class ICUDateTimeFormField(forms.DateTimeField):
    "Field that parses datetime input using the ICU api and user selected locale/timezone"

    def to_python(self, value):
        "Validate that the input can be interpreted in local format"

        if value in self.empty_values:
            return timezone.now()
        if isinstance(value, datetime.datetime):
            utc = icu.ICUtzinfo.getDefault()
            return value.astimezone(utc)

        return super().to_python(value)


class SelectTimeZoneMapWidget(forms.Widget):
    """
    Custom widget for displaying timezone map. This works with mic-react-lib/InputChoice,
    not the default django widgets, to display a touch-selectable map of timezones.
    The input_type is passed by rest_input as prop to the InputChoice, which does the right thing.
    """

    input_type = "select-time-zone"


class TimezoneMiddleware:  # pylint: disable=too-few-public-methods
    "Activate the selected timezone for all views"

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        try:
            view, _, _ = resolve(request.get_full_path())
            ignore_time = hasattr(view, 'ignore_time') and view.ignore_time
        except Resolver404 as e:
            ignore_time = False
        if not ignore_time:
            settings = Settings.objects.get()
            set_language(settings.locale, settings.time_zone)
            udp_set_translate_locale()
            tzname = settings.time_zone
            if tzname:
                timezone.activate(pytz.timezone(tzname))
            else:
                timezone.deactivate()
        return self.get_response(request)


def get_timezone_display_name(tzname):
    "Returns the string that we want to show for the timezone setting."
    settings = Settings.objects.get()
    time_zone = icu.TimeZone.createTimeZone(tzname)
    return time_zone.getDisplayName(icu.Locale(settings.locale))
