"""
This module provides support functionality for running instrument scripts through the normal
instrument ui, and saving results to the database.
"""
import __main__
import os
import importlib
import django
from django.utils import timezone
from mic_cpp import cpp

APP_NAME = None

# =================================================================================================
def django_setup(site, app_name):
    """
    Called at the beginning of a script to setup the links to the site and application so that we
    can have access to the django models.
    """
    global APP_NAME
    APP_NAME = app_name
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", site + ".settings")
    django.setup()


def get_model_and_page_with_header():
    """
    Create a report model and first page for the script.  This will also add the default header to
    the page.

    @returns: model, page (pair)
    """
    name = _get_script_name()
    model = cpp.CReportModel(name)
    page = model.AddPage(name)
    _add_default_header(page)
    return model, page


def add_result_table(page, passed):
    """
    Adds an options table to the page. It also adds the proper Results of passed failed as the first
    line of the results.

    @returns the options table in case the script wants to add more information.
    """
    table = cpp.CReportOptionsTable(page, "Passed" if passed else "FAILED")
    table.AddItem("", "")
    return table


def add_std_table(page, headers, title=None):
    """
    Adds a standard table to the results page

    @param page: The page to add to.
    @param headers: The list of column headers for the table.
    @param title: The title of the table (if any)

    @returns: The CReportStdTable that is setup with the titles, ready to add data.
    """
    table = cpp.CReportStdTable(page, title if title else "", len(headers))
    for i, header in enumerate(headers):
        table.SetTitle(i, header)
    return table


def script_results(passed, model=None):
    """
    Called to update the currently running script in the database.  If there is none, it is created
    and marked complete.  So it should only be called once at the end of a script.

    @param passed: if the script passed or failed.
    @param model: (optional) the report model if the script did not just create its own.
    """
    from mic_python_lib.models import InstrumentScriptRun
    active = InstrumentScriptRun.objects.active_or_create(_get_script_name())

    active.passed = passed
    active.end_time = timezone.now()
    active.state = 'C'
    active.report_model = cpp.SaveReportModel(model) if model else None
    active.save()


# =================================================================================================
def OKCancel(message: str):
    """
    Put up a message on the instrument UI and wait for the user to press the OK or cancel
    buttons. If the user presses OK, this will return and allow the python script to continue
    executing.  If the user presses cancel, it will terminate the process.
    """
    cpp.NLogQuery_InitPythonWithEvent("127.0.0.1:8000/rest/log_query")
    cpp.NLogQuery_PythonOKCancel(message)


def YesNoCancel(message: str):
    """
    Put up a message on the instrument UI and wait for the user to press the yes, no or cancel
    buttons. If the user presses cancel, it will terminate the process. Otherwise, the function
    returns a CQueryResponse object that can be queried of IsYes or IsNo.
    """
    cpp.NLogQuery_InitPythonWithEvent("127.0.0.1:8000/rest/log_query")
    return cpp.NLogQuery_PythonYesNoCancel(message)


def CLogUIInput(title: str):
    """
    Create and return a CLogUIInput (c++ exposed) object that can be setup to define that
    information to get from the user.
    """
    return cpp.NLogQuery_CLogUIInput()


def GetUserData(ui: cpp.NLogQuery_CLogUIInput):
    """
    Put up the data entry screen on the instrument UI for this script and wait for the user to
    press the OK or cancel buttons.  If the user presses OK, this will return and allow the python
    script to continue executing.  If the user presses cancel, it will terminate the process.
    """
    cpp.NLogQuery_InitPythonWithEvent("127.0.0.1:8000/rest/log_query")
    from mic_python_lib.models import InstrumentScriptRun
    active = InstrumentScriptRun.objects.active_or_create(_get_script_name())
    cpp.NLogQuery_PythonGetUserData(ui, active.user_data_context)


# =================================================================================================
def _get_model_serial():
    m = importlib.import_module(APP_NAME + ".models")
    local = m.Instrument.objects.get(is_local=True)
    model_no = local.model
    serial = local.serial
    return model_no, serial


def _get_script_name():
    return os.path.basename(__main__.__file__)[:-3]


def _add_default_header(page):
    from mic_python_lib.models import InstrumentScriptRun
    active = InstrumentScriptRun.objects.active_or_create(_get_script_name())

    model_no, serial = _get_model_serial()
    name = _get_script_name()
    header = cpp.CReportHeaderHelper(page, 2)
    header.SetSimpleItem(0, f"{model_no} SN: {serial}")
    header.SetSimpleItem(1, active.start_time.strftime("%Y-%m-%d %H:%M"))
    header.SetSimpleItem(0, name)
    header.SetSimpleItem(1, active.operator)
