'''Python support code for embedded touchscreen scripting system.'''
import datetime
import gettext
import glob
import json
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from django.contrib.auth.models import User
from django.http import JsonResponse
from django.urls import path
from django.utils import timezone
from .rest import make_json_view
from .instrument import is_instrument
from .log_query import post
from .mic_network import get_usb_drive_mounts, zip_dir_to_usb
from .models import InstrumentScriptRun, LogMessage
from .results import report_show_export_print_view
from mic_cpp import cpp

_ = gettext.gettext
# =================================================================================================
# The location of where the installed scripts are.  For the development system this is in a
# different place than on the released system.
SCRIPT_LOCATION = "/var/mic-scripts" if is_instrument() else "/tmp/mic-scripts"
# Where we create a directory to zip up into the script export.
EXPORT_LOCATION = "/tmp/mic-script-export/"

# =================================================================================================
class ScriptInterface(ABC):
    """
    The interface to the application for the functionality that the script subsystem needs.
    """
    @abstractmethod
    def model_serial(self):
        "Return the model and serial number pair for the instrument."


    @abstractmethod
    def run_from_oper(self, oper_id: int):
        """
        Return the InstrumentScriptRun object that is associated with the operation index of
        oper_id.
        """

    @abstractmethod
    def script_start(self, name: str, operator: str):
        """Start the python script of the given name and associate the operator with it."""

    @abstractmethod
    def script_cancel(self):
        """Cancel the currently running script."""

    @abstractmethod
    def app_export(self, location: str, is_service: bool):
        """
        Add all the application specific export information to the directory at location.  If
        is_service is true, then assume is post reset to factory defaults, and don't export the
        customer's data.
        """

# =================================================================================================
def start_get_status():
    "For the script start page, get the status of all known scripts."
    names = _get_installed_script_names()
    scripts = []
    for name in names:
        last = InstrumentScriptRun.objects.filter(
            name=name, state="C").order_by("start_time").last()
        scripts.append({
            "name": name,
            "num": InstrumentScriptRun.objects.filter(name=name, state="C").count(),
            "date": last.start_time.strftime("%Y-%m-%d %H:%M") if last else "",
            "passed": last.passed if last else False,
            "id": last.id if last else 0,
        })

    last = InstrumentScriptRun.objects.filter(
        name__in=names).order_by("start_time").last()
    operator = last.operator if last else "Unknown"
    return {"operator": operator, "scripts": scripts }


def start_set_error(error: str):
    "Set the error in the data for the currently running script."
    run = InstrumentScriptRun.objects.filter(state='P').order_by("start_time").last()
    if run:
        run.error_output = error
        run.save()


def start_review(request, script_id, interface: ScriptInterface):
    "Return the report for the script"
    run = InstrumentScriptRun.objects.get(id=script_id)
    model = _get_script_report(run, interface)
    if not request.body:
        return JsonResponse(json.loads(model.AsJson()))
    basename = f'{run.name}-{run.start_time.strftime("%Y%m%d%H%M")}'
    return report_show_export_print_view(request, model, basename)


def start_oper_review(request, oper_id, interface: ScriptInterface):
    "Return the report for the script that is associated with the oper_id."
    return start_review(request, interface.run_from_oper(oper_id), interface)

# =================================================================================================
def export_get_status():
    "Return the status of all the known scripts for the export page."
    failed, not_run, complete = _get_run_status()
    summary_text = _("All tests have passed.")  if complete else _("Not all tests have passed.")

    return {
        "summary_text": summary_text,
        "complete": complete,
        "failed": sorted(failed, key=lambda f: f["name"]),
        "not_run": sorted(not_run, key=lambda f: f["name"]),
    }

def export_export(interface: ScriptInterface):
    "Do the script export action."
    try:
        servicePasswordActive = User.objects.get(username="service").is_active
    except User.DoesNotExist:
        servicePasswordActive = False

    # make sure we have one usb drive plugged in, if not don't bother generating the export files.
    usb_dirs = get_usb_drive_mounts()
    if len(usb_dirs) > 1:
        post(_("Cannot export to USB. Multiple USB drives found."))
        return {}
    elif len(usb_dirs) == 0:
        post(_("Cannot export to USB. No USB drive found."))
        return {}

    _unused, _unused, complete = _get_run_status()
    incomplete = "" if complete else "-incomplete"
    shutil.rmtree(EXPORT_LOCATION, ignore_errors=True)
    Path(EXPORT_LOCATION).mkdir(parents=True, exist_ok=True)
    _add_script_export(interface)
    _add_instrument_log(interface)
    interface.app_export(EXPORT_LOCATION, servicePasswordActive)
    model, serial = interface.model_serial()
    if servicePasswordActive:
        zip_name = f'{model}-{serial}-{timezone.now().strftime("%Y%m%d%H%M")}.zip'
    else:
        zip_name = f'{model}-pcp{incomplete}-{serial}-{timezone.now().strftime("%Y%m%d%H%M")}.zip'
    zip_dir_to_usb(EXPORT_LOCATION, zip_name)
    shutil.rmtree(EXPORT_LOCATION, ignore_errors=True)
    return {}

# =================================================================================================
def manage_get_files(interface: ScriptInterface):
    """
    Return an array of objects with action and name attributes that define what should be shown on
    the manage scripts UI for the user to know what update and replace will do.
    """
    replace, added, remove = _get_script_update(interface.model_serial()[0])
    actions = []
    for rep in replace:
        actions.append({"action": _("Replace:"), "name": rep.split("/")[-1][:-3]})
    for add in added:
        actions.append({"action": _("Add:"), "name": add.split("/")[-1][:-3]})
    for rem in remove:
        actions.append({"action": _("Remove on replace:"), "name": rem.split("/")[-1][:-3]})
    return actions


def manage_update(interface: ScriptInterface):
    "Do the script/manage/update to add the new/replace scripts to the instrument."
    Path(SCRIPT_LOCATION).mkdir(parents=True, exist_ok=True)
    replace, added, _ = _get_script_update(interface.model_serial()[0])
    for fname in replace + added:
        shutil.copy(fname, SCRIPT_LOCATION)
    return {}


def manage_replace(interface: ScriptInterface):
    "Do the script/manage/replace to add the new/replace scripts and remove old to the instrument."
    # On replace, just delete the existing, and then move the new.
    for fname in glob.glob(SCRIPT_LOCATION + "/*.py"):
        Path(fname).unlink(missing_ok=True)
    return manage_update(interface)


# =================================================================================================
def script_from_name(name: str) -> str:
    "Return the full pathname for the script."
    return f"{SCRIPT_LOCATION}/{name}.py"

def script_report(location: str, run: InstrumentScriptRun, interface: ScriptInterface):
    "Generate the output files for the script report."
    model = _get_script_report(run, interface)
    serial = interface.model_serial()[1]
    name = f'{run.name}-{serial}-{run.start_time.strftime("%Y%m%d%H%M")}'
    cpp.ExportReportModelXLSX(model, EXPORT_LOCATION, f"{name}.xlsx")
    cpp.ExportReportModelPDF(model, EXPORT_LOCATION, f"{name}.pdf")

def url_list(base, interface: ScriptInterface):
    "Make a list of all the url and action pairs that this module provides."
    return [
        path(base + 'start/get_status/', make_json_view(start_get_status)),
        path(base + 'start/start/',
             make_json_view(lambda name, operator: interface.script_start(name, operator))),
        path(base + 'start/set_error/', make_json_view(start_set_error)),
        path(base + 'start/cancel/',
             make_json_view(lambda: interface.script_cancel())),
        path(base + 'start/review/<int:script_id>/',
             lambda request, script_id: start_review(request, script_id, interface)),
        path(base + 'start/oper_review/<int:oper_id>/',
             lambda request, oper_id: start_oper_review(
                 request, oper_id, interface)),
        path(base + 'export/get_status/',
             make_json_view(export_get_status)),
        path(base + 'export/export/',
             make_json_view(lambda: export_export(interface))),
        path(base + 'manage/get_files/',
             make_json_view(lambda: manage_get_files(interface))),
        path(base + 'manage/update/',
             make_json_view(lambda: manage_update(interface))),
        path(base + 'manage/replace/',
             make_json_view(lambda: manage_replace(interface))),
    ]

# =================================================================================================
def _get_installed_scripts():
    return _get_python_files(SCRIPT_LOCATION)


def _get_installed_script_names():
    scripts = [name.split('/')[-1][:-3] for name in _get_installed_scripts()]
    return [name for name in scripts if not name.startswith("lib")]


def _get_usb_scripts(model):
    "Return a list if all the python files in /<usb>/<model>/scripts/ directory."
    usb_dirs = get_usb_drive_mounts()
    if not len(usb_dirs):
        return []
    location = usb_dirs[0] + "/" + model + "/scripts"
    return _get_python_files(location)


def _get_python_files(location: str):
    if location[-1] != '/':
        location = location + '/'
    loc_len = len(location)
    return sorted(glob.glob(location+"*.py"))


def _get_script_update(model):
    "Returns a list of full path filenames for scripts to be replaced, added, and removed."
    installed = _get_installed_scripts()
    available = _get_usb_scripts(model)

    ins_name = [fname.split('/')[-1][:-3] for fname in installed]
    avl_name = [fname.split('/')[-1][:-3] for fname in available]

    # If none available, don't show anything.
    if not available:
        return [], [], []

    replace = [av for av, avn in zip(available, avl_name) if avn in ins_name]
    add = [av for av, avn in zip(available, avl_name) if avn not in ins_name]
    remove = [ins for ins, insn in zip(installed, ins_name) if insn not in avl_name]

    return replace, add, remove


def _get_run_status():
    current_names = _get_installed_script_names()
    failed = []
    run = {}
    for last_run  in InstrumentScriptRun.objects.raw('''
            SELECT r1.id, r1.name, r1.passed, r1.end_time
            FROM mic_python_lib_instrumentscriptrun r1, (
                SELECT name, MAX(end_time) AS last_end
                FROM mic_python_lib_instrumentscriptrun
                GROUP BY name
            ) AS r2
            WHERE r1.name = r2.name AND r1.end_time = r2.last_end
            '''):
        if last_run.name not in current_names:
            continue
        if not last_run.passed:
            failed.append({"name": last_run.name, "end_time": str(last_run.end_time)})
        run[last_run.name] = last_run.end_time

    not_run = [{"name": name} for name in current_names if name not in run.keys()]
    complete = not failed and not not_run
    return failed, not_run, complete


def _get_script_report(run: InstrumentScriptRun, interface: ScriptInterface):
    if run.report_model and run.state == "C":
        model = cpp.CReportModel(run.name)
        cpp.ReadReportModel(run.report_model, model)
    else:
        model = cpp.CReportModel(run.name)
        page = model.AddPage(run.name)
        header = cpp.CReportHeaderHelper(page, 2);
        model_no, serial = interface.model_serial()
        header.SetSimpleItem(0, f"{model_no} SN: {serial}")
        header.SetSimpleItem(1, cpp.CTimeDate(timezone.now()).GetBoth())
        header.SetSimpleItem(0, run.name)
        header.SetSimpleItem(1, run.operator)

        # Note: this is not translated because the script generated ones are not either.
        result = run.get_state_display()
        if run.state == 'C':
            result = "Passed" if run.passed else "FAILED"
        table = cpp.CReportOptionsTable(page, result)
        table.AddItem("", "")
        if run.state == 'E' and run.error_output:
            table.AddItem("Output:", run.error_output)
    return model


def _add_script_export(interface: ScriptInterface):
    model = cpp.CReportModel("Summary")
    page = model.AddPage("Summary")
    header = cpp.CReportHeaderHelper(page, 2);
    model_no, serial = interface.model_serial()
    header.SetSimpleItem(0, f"{model_no} SN: {serial}")
    header.SetSimpleItem(1, cpp.CTimeDate(timezone.now()).GetBoth())
    table = cpp.CReportStdTable(page, _("PCP Script Summary"), 3)
    table.SetTitle(0, _("Date"))
    table.SetTitle(1, _("Script"))
    table.SetTitle(2, _("Result"))
    table.AlignCenter(2)
    for run  in InstrumentScriptRun.objects.filter(state__in="C").order_by("end_time"):
        table.AddItem(0, cpp.CUnitsStr(cpp.CTimeDate(run.end_time).GetBoth()) if run.end_time else "")
        table.AddItem(1, cpp.CUnitsStr(run.name))
        table.AddItem(2, cpp.CUnitsStr(_("Passed") if run.passed else _("FAILED")))
        script_report(EXPORT_LOCATION, run, interface)
    name = f'summary-{serial}-{timezone.now().strftime("%Y%m%d%H%M")}'
    cpp.ExportReportModelPDF(model, EXPORT_LOCATION, f'{name}.pdf')
    cpp.ExportReportModelXLSX(model, EXPORT_LOCATION, f'{name}.xlsx')


def _add_instrument_log(interface: ScriptInterface):
    "Export the instrument Log to the export directory."
    model = cpp.CReportModel("Instrument Log")
    page = model.AddPage(_("Instrument Log"))
    header = cpp.CReportHeaderHelper(page, 2);
    model_no, serial = interface.model_serial()
    header.SetSimpleItem(0, f"{model_no} SN: {serial}")
    header.SetSimpleItem(1, cpp.CTimeDate(timezone.now()).GetBoth())
    table = cpp.CReportStdTable(page, _("Instrument Log"), 2)
    table.SetTitle(0, _("Date"))
    table.SetTitle(1, _("Message"))
    start = timezone.now() - datetime.timedelta(days=90)
    for log in LogMessage.objects.filter(time__gte=start).order_by("-time")[:660]:
        table.AddItem(0, cpp.CUnitsStr(cpp.CTimeDate(log.time).GetBoth()))
        table.AddItem(1, cpp.CUnitsStr(log.message))
    name = f'instrumentlog-{serial}-{timezone.now().strftime("%Y%m%d%H%M")}'
    cpp.ExportReportModelXLSX(model, EXPORT_LOCATION, f'{name}.xlsx')
