"""The functionality of the log and query messages."""
from datetime import datetime
import json
import pytz
from django.urls import path
from .models import LogAction, LogResponse, LogSeverity, LogMessage
from .rest import function_to_json_view, model_function_to_json_view

#: The tuple of actions that need to be acknoledged.  Note this does not
#: include the modal ones as they are not logged till after they are shown.
_NEEDS_ACK = (LogAction.POST_ONCE, LogAction.POST, LogAction.POST_UPDATE)

_NOTIFY_CALLBACK = None
_ON_SAMPLE_LOG = None

def delete_modal():
    """
    This will remove all the modal messages from the database.  Useful
    after the instrument resets.
    """
    LogMessage.objects.modal().delete()
    LogMessage.objects.exclude(input_context=0).delete()


def post(message, action=LogAction.POST, severity=LogSeverity.DEFAULT_INFO):
    """Convenience function meant to be called by a django/rest python application to add a popup
    message """
    time = pytz.utc.localize(datetime.now())
    log = LogMessage(
        time=time, action=action, severity=severity, message=message, acknowledged=False
    )
    log.save()

    if _NOTIFY_CALLBACK:
        _NOTIFY_CALLBACK(LogMessage.objects.unacknowledged_errors().count() > 0)


def add_log_entry(time, action, severity, message, user='', account='', post_id=0):  # pylint: disable=too-many-arguments
    """Add the log message to the permanent log."""
    time = pytz.utc.localize(datetime.fromtimestamp(time))
    log = LogMessage(
        time=time,
        action=action,
        severity=severity,
        message=message,
        user=user,
        account=account,
        post_id=post_id,
        acknowledged=action not in _NEEDS_ACK
    )

    # special handling of some types.
    if action in (LogAction.POST_ONCE, LogAction.POST_UPDATE):
        # if there is a matching entry update it and be done.
        match = _find_matching(log)
        if match:
            match.message = log.message
            match.save()
            return

        # if it was post udpate and not match, ignore it.
        if action == LogAction.POST_UPDATE:
            return

    # Special handling of sample log message.
    if action == LogAction.SAMPLE_LOG:
        if _ON_SAMPLE_LOG:
            _ON_SAMPLE_LOG(log)

    # save the requested log
    log.save()

    if _NOTIFY_CALLBACK:
        _NOTIFY_CALLBACK(LogMessage.objects.unacknowledged_errors().count() > 0)


def show_modal(time, action, severity, message, modal_context, user='', account='', post_id=0):  # pylint: disable=too-many-arguments
    """Setup the data to show the log to the user."""
    time = pytz.utc.localize(datetime.fromtimestamp(time))
    log = LogMessage(
        time=time,
        action=action,
        severity=severity,
        message=message,
        user=user,
        account=account,
        post_id=post_id,
        acknowledged=False,
        modal_context=modal_context
    )
    log.save()


def get_log_to_show():
    """Return the most critical log message to show to the user, or None if there aren't any."""

    # The ui only shows a single message at a time.
    # First show modal, then show unacknowledged.
    log = LogMessage.objects.modal().first()
    if log is not None:
        return log

    return LogMessage.objects.unacknowledged().first()


def acknowledge_message(log_id, response=LogResponse.OK, callback=None):
    """Mark the message as having been acknowledged."""
    try:
        log = LogMessage.objects.get(id=log_id)
    except LogMessage.DoesNotExist:
        # ignore if object does not exist. Most likely non bug is double message.
        return

    if log.modal_context != 0:
        if callback is not None:
            callback(log.modal_context, response)
        log.delete()
    else:
        log.acknowledged = True
        log.save()

    if _NOTIFY_CALLBACK:
        _NOTIFY_CALLBACK(LogMessage.objects.unacknowledged_errors().count() > 0)


def show_input(time, message, modal_context, input_context, input_params):
    """Add the user query form information to the database."""
    time = pytz.utc.localize(datetime.fromtimestamp(time))
    log = LogMessage(
        time=time,
        action=LogAction.INPUT_FORM,
        severity=LogSeverity.DEFAULT_INFO,
        message=message,
        modal_context=modal_context,
        input_context=input_context,
        input_params=json.dumps(input_params),
        acknowledged=False
    )
    log.save()


def acknowledge_input(log_id, response, data, callback):
    """Send the user response and data back to the requester."""
    try:
        log = LogMessage.objects.get(id=log_id)
    except LogMessage.DoesNotExist:
        # ignore if object does not exist. Most likely non bug is double message.
        return
    if log.modal_context != 0:
        if callback is not None:
            callback(log.modal_context, response, json.dumps(data))
            log.delete()
    else:
        log.delete()


def get_input_to_show(input_context):
    """Return the input parameters to show if there are any for this context."""
    return LogMessage.objects.filter(input_context=input_context).first()


def url_list(base, show_modal_callback, show_input_callback):
    """Returns the list of urls that the log_query can respond to.

    :param base: The base name of the list.
    :param callback: The callback for responding to modal queries.
    """

    # disable the slow time middleware for the polling function.
    get_log_to_show_view = lambda request: model_function_to_json_view(request, get_log_to_show)
    get_log_to_show_view.ignore_time = True

    return [
        path(
            base + 'add_log_entry/', lambda request: function_to_json_view(request, add_log_entry)
        ),
        path(base + 'show_modal/', lambda request: function_to_json_view(request, show_modal)),
        path(
            base + 'acknowledge_message/', lambda request:
            function_to_json_view(request, acknowledge_message, callback=show_modal_callback)
        ),
        path(base + 'get_log_to_show/', get_log_to_show_view),
        path(base + 'show_input/', lambda request: function_to_json_view(request, show_input)),
        path(
            base + 'acknowledge_input/', lambda request:
            function_to_json_view(request, acknowledge_input, callback=show_input_callback)
        ),
        path(
            base + 'get_input_to_show/',
            lambda request: model_function_to_json_view(request, get_input_to_show)
        ),
    ]


def cmd_line_check(callback):
    """Checks to see if there are any messages to display or wait on.
    If there is a message to display, this will prompt the user and
    wait for a response.  Even if this is just a post, it will wait
    anyway.  The presumption is that operation is proceeding anyway.

    :param callback: The callback for responding to modal queries.
    """
    resp_lookup = {
        'a': LogResponse.OK,
        'o': LogResponse.OK,
        'c': LogResponse.CANCEL,
        'y': LogResponse.YES,
        'n': LogResponse.NO,
    }
    while True:
        log = get_log_to_show()
        if not log:
            break
        if log.action in (LogAction.POST_ONCE, LogAction.POST, LogAction.POST_UPDATE):
            resp = 'a'
        elif log.action == LogAction.INFORM_WAIT:
            resp = 'o'
        elif log.action == LogAction.YES_NO:
            resp = 'yn'
        elif log.action == LogAction.YES_NO_CANCEL:
            resp = 'ync'
        elif log.action == LogAction.OK_CANCEL:
            resp = 'oc'
        while True:
            print("POST: ", log.time, log.message)
            user_resp = input(f"RESPONSE: ({'/'.join(resp)}): ")
            if user_resp and user_resp in resp:
                break
        acknowledge_message(log.id, resp_lookup[user_resp], callback)


def cmd_line_check_input(input_context, callback):
    """Checks to see if there are any messages to display or wait on.
    If there is a message to display, this will prompt the user and
    wait for a response.  Even if this is just a post, it will wait
    anyway.  The presumption is that operation is proceeding anyway.

    :param input_context: The input context under which to check for a prompt.
    :param callback: The callback for responding to modal queries.
    """
    form = get_input_to_show(input_context)
    if form:
        prms = json.loads(form.input_params)
        layout = prms["format"]["layout"]
        fields = prms["format"]["fields"]
        values = {}
        print(form.message)
        for name in layout:
            field = fields[name]
            if field["field"] == 'number':
                values[name] = float(input(f'{name}: '))
            if field["field"] == 'text':
                values[name] = str(input(f'{name}: '))
        acknowledge_input(form.id, LogResponse.OK, values, callback)


def setup_notify_shown_errors(callback):
    """Setup the mechanism to call callback whenever there may have
    been unacknowledged errors.

    :param callback: The callback for when there may be a change in
        the state of if there are any unacknowledged errors.
    """
    global _NOTIFY_CALLBACK  # pylint: disable=global-statement
    _NOTIFY_CALLBACK = callback
    _NOTIFY_CALLBACK(LogMessage.objects.unacknowledged_errors().count() > 0)


def setup_on_sample_log(callback):
    """Setup the mechanism to call callback whenever there is a sample log message.

    :param callback: The callback for when there is a new sample log message.
    """
    global _ON_SAMPLE_LOG  # pylint: disable=global-statement
    _ON_SAMPLE_LOG = callback


def _find_matching(log):
    """Return the log message (if any) that matches the update request."""

    if log.post_id > 0:
        return LogMessage.objects.unacknowledged().filter(post_id=log.post_id).last()
    return LogMessage.objects.unacknowledged().filter(message=log.message).last()
