'''
The python interface for mounting cifs/samba drives onto the instrument.
'''
import gettext
import os
import shutil
import subprocess
from .instrument import is_instrument

_ = gettext.gettext
# =================================================================================================
# where the cifs user credentials are stored.
_CREDENTIALS_FILE = "/home/root/network_user.cifs"
# where the cifs user credentials are stored for the check on the credentials.
_CREDENTIALS_CHECK_FILE = "/home/root/check_user.cifs"
# where the network will be mounted.
_NETWORK_MOUNT = "/mnt/network"
# where the network will be mounted for the check on the credentials.
_NETWORK_CHECK_MOUNT = "/mnt/check_network"
# the /etc/fstab file that is update with network info
_FSTAB_FILE = "/etc/fstab"


# =================================================================================================
def get_mount_point():
    "Return the full path to the network mount point."
    return _NETWORK_MOUNT


def check_mount(server, path, user, password):
    """
    Check to see if the credentials are OK for mounting a cifs path.

    @returns an empty string if it went OK, or an error message if the mount failed.
    """
    if not is_instrument():
        return "Can only mount on a real instrument."

    subprocess.run(["umount", _NETWORK_CHECK_MOUNT], check=False)
    _save_user_credentials(user, password, path=_CREDENTIALS_CHECK_FILE)
    os.makedirs(_NETWORK_CHECK_MOUNT, exist_ok=True)
    proc = subprocess.run([
        "mount", "-t", "cifs", "-o", f"credentials={_CREDENTIALS_CHECK_FILE}", f"//{server}/{path}",
        _NETWORK_CHECK_MOUNT
    ],
                          check=False)
    subprocess.run(["umount", _NETWORK_CHECK_MOUNT], check=False)
    if proc.returncode != 0:
        return _("Unable to mount network drive.")
    return ""


def check_printer(server, workgroup, name, user, password):
    """
    Check to see if we can communicate with the printer given the credentials.

    @returns an empty string if it went OK, or an error message if the mount failed.
    """
    if not is_instrument():
        return "Can only print on a real instrument."

    proc = subprocess.run([
        "smbclient", f"//{server}/{name}", password, f"-W{workgroup}", "<password>", "-c", "q",
        f"-U{user}"
    ],
                          check=False)
    if proc.returncode != 0:
        return _("Unable to connect to network printer.")
    return ""


def print_cmd(server, workgroup, name, user, password):
    """
    @returns the string for the command to print the results.
    """
    return f'smbclient //{server}/{name} {password} -W{workgroup} {password} -c "print -" -U{user}'


def set_and_mount(server, path, user, password):
    """
    Remove the old network mount point, and try to create the new one at //<server>/path with the
    user name and password.

    @returns an empty string if it went OK, or an error message if the mount failed.
    """
    if not is_instrument():
        return "Can only mount on a real instrument."

    subprocess.run(["umount", "-f", _NETWORK_MOUNT], check=False)
    _save_user_credentials(user, password)
    _set_mount_point(server, path)
    return _mount_network()


def start_automount():
    "Access the network mount point in case it is slow in mounting."
    if not is_instrument():
        return

    # just try to read the mount point in the background.  Don't wait.
    subprocess.Popen(
        ["ls", _NETWORK_MOUNT], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

# =================================================================================================
def _mount_network():
    "Mount the network once everything is setup."
    proc = subprocess.run(["mount", _NETWORK_MOUNT], check=False)
    if proc.returncode != 0:
        return _("Unable to mount network drive.")
    return ""


def _save_user_credentials(user, password, path=_CREDENTIALS_FILE):
    "Save the user credentials in the credentials file."
    with open(path, 'w') as cred_file:
        print(f"username={user}", file=cred_file)
        print(f"password={password}", file=cred_file)


def _set_mount_point(server, path):
    "Update the /etc/fstab file.  Remove the old mount point (if any), and add the new one."
    server = server.replace(' ', '\040')
    path = path.replace(' ', '\040')
    new_file = _FSTAB_FILE + ".new"
    try:
        os.makedirs(_NETWORK_MOUNT, exist_ok=True)
    except FileExistsError:
        pass
    with open(new_file, 'w') as mount_file:
        for line in open(_FSTAB_FILE).read().splitlines():
            if not _NETWORK_MOUNT in line:
                print(line, file=mount_file)
        print(
            f'//{server}/{path} {_NETWORK_MOUNT} cifs credentials={_CREDENTIALS_FILE},noauto,x-systemd.automount 0 0',
            file=mount_file
        )
    shutil.move(new_file, _FSTAB_FILE)
