"""
The mic-net event client that handles udp broadcast events on the mic-network.


The MicNetEventHandler is expected to be created and then run in a service separately from  the
django server.  At startup it will broadcast the 'instrument-boot' message, then it will listen
future messages, including replies. This class handles the following messages:

Send messages:
instrument-boot: broadcast ourselves, and all our known instruments.
instrument-knows: Reply to a broadcast with all the ones we know.

Receive messages:
instrument-boot: Stores the new instruments in the table, and replies with a instrument-known
message.
instrument-knows: Writes down all the known instruments, and if there is a conflict with our
ip address, it will update ours, and resend an instrument-boot message.

The ip address are coordinated in the 16 bit subnet that is the Mic Ethernet connector.  The ip
addresses are mutually coordinated.  Each instrument keeps a table of known instruments.   We allow
up to 100(MAX_RELEVANT) entries in that table.  If there is a history of more than these on the
network, the ones that have not been heard from the longest are dropped.

When an instrument is first started, or when the Mic Ethernet subnet is changed, the instrument
gets a random ip address.   When it is first turned on it will broadcast its ip address along with
its known instruments (instrument-boot).  All other instruments will then reply with
instrument-knows broadcast. This will update the new instrument, (and all others).

When a known instrument has the same address as we do, (but not on boot) we will reset our
ip address to a new random value that is not known.  This should end quickly since there are
significantly more ip addresses available than are used.

The way to add an instrument to an existing network is to first, set the Mic Ethernet subnet,
then add it to the existing network, and turn it on to handle the coordination.

"""
import datetime
import json
import os
import socket
import sys
from django.utils import timezone, dateparse
from .mic_network import get_network_info, mic_subnet_broadcast, random_ip, setup_mic_ethernet
from .models import KnownInstrument

# =================================================================================================
PORT = 20045  # random port to use for the boradcast messages.
MAX_RELEVANT = 100  # limit the shared instruments to the last known 100 of them for communications.
INST_BOOT_MSG = 'instrument-boot'
INST_KNOWN_MSG = 'instrument-known'
NEW_DATA_MSG = 'new-data'


# =================================================================================================
class MicNetEventHandler:  # pylint: disable=too-few-public-methods
    """
    The event client for the Mic Ethernet network.

    This is created with model and serial number.  It will broadcast the instrument-boot and react
    to instrument-boot and instrument-known messages to coordinate the instrument table.

    It also have timer tick functionality.  The timer will tick every tick_time seconds I does not
    get another message, and will call the tick_fcn when it ticks.
    """

    def __init__( # pylint: disable=too-many-arguments
            self, model, serial, new_data_fcn=None, tick_time=60, tick_fcn=None):
        self.model = model
        self.serial = serial
        self.running = False
        self.tick_fcn = tick_fcn
        self.tick_time = tick_time
        self.new_data_fcn = new_data_fcn

        self._events = {
            INST_BOOT_MSG: self._recv_boot_notification,
            INST_KNOWN_MSG: self._recv_known_notification,
            NEW_DATA_MSG: self._recv_new_data,
        }

        self._server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self._server.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self._server.settimeout(tick_time)
        self._server.bind(("", PORT))

    def run(self):
        """Run the server, send the initial I'm up notification, and then wait for UPD events."""
        self.running = True
        self._send_instrument_notification(is_boot=True)
        last_tick = datetime.datetime.now()
        while self.running:
            try:
                data, addr = self._server.recvfrom(65536)
                data = json.loads(data.decode('utf-8'))
                if 'event' not in data:
                    continue
                if data['event'] == 'quit':
                    break
                if data['event'] not in self._events:
                    continue
                event = data['event']
                del data['event']
                self._events[event](addr[0], **data)

            except json.JSONDecodeError:
                pass
            except socket.timeout:
                pass

            # if it has been a long enough run the tick
            if last_tick + datetime.timedelta(seconds=self.tick_time) <= datetime.datetime.now():
                if self.tick_fcn is not None:
                    self.tick_fcn()
                    last_tick = datetime.datetime.now()

    def _add_or_update_instrument_ip(self, model, serial, ip_addr, last_communicated):
        "Add the instrument to the table of known instruments unless it is us."
        # don't add ourselves to the table.
        if self.model != model or self.serial != serial:
            inst, created = KnownInstrument.objects.get_or_create(model=model, serial=serial)
            inst.ip_addr = ip_addr
            last_communicated = dateparse.parse_datetime(last_communicated)
            if created or last_communicated > inst.last_communicated:
                inst.last_communicated = last_communicated
            inst.save()

    def _ensure_unique_ip(self):
        """See if our IP address is already used in the known table.  If so change it and
        rebroadcast our new IP address."""
        info = get_network_info()
        same_ip = KnownInstrument.objects.filter(ip_addr=info.mic_ethernet_ip)
        if len(same_ip) > 0:
            known_ip = [inst.ip_addr for inst in KnownInstrument.objects.all()]
            setup_mic_ethernet(random_ip(info.mic_ethernet_ip[:7], known_ip))
            self._send_instrument_notification(is_boot=True)
            return False
        return True

    def _recv_boot_notification( # pylint: disable=too-many-arguments
            self, addr, model, serial, ip_addr, last_communicated, known_instruments):
        "Functionality on receiving a instrument-boot message.  Update our table."
        # update all the tables just like boot.
        self._recv_known_notification(
            addr, model, serial, ip_addr, last_communicated, known_instruments
        )
        # now send the instruments I know of in an 'instrument-known' message.
        self._send_instrument_notification(is_boot=False)
        # when a new instrument booted, check to see if it has new data.
        self._recv_new_data(addr, model, serial)

    def _recv_known_notification( # pylint: disable=too-many-arguments, unused-argument
            self, addr, model, serial, ip_addr, last_communicated, known_instruments):
        """Functionality for the instrument-known message.  Update the table, and reset our ip
        address if there is a conflict."""
        # add the one that booted, and then all the ones it knows about.
        self._add_or_update_instrument_ip(model, serial, ip_addr, last_communicated)
        for inst in known_instruments:
            self._add_or_update_instrument_ip(
                inst['model'], inst['serial'], inst['ip_addr'], inst['last_communicated']
            )

        # if there are more than MAX_RELEVANT in the table, remove the oldest ones.
        insts = KnownInstrument.objects.all().order_by("-last_communicated")
        if len(insts) > MAX_RELEVANT:
            for inst in insts[MAX_RELEVANT:]:
                inst.delete()

        # make sure our IP address is unique, and if we did not change it, update results from the
        # instrument that just talked to us.
        if self._ensure_unique_ip():
            self._recv_new_data(addr, model, serial)

    def _recv_new_data(self, addr, model, serial):
        "Functionality for when we got new data notification from an instrument."
        if self.new_data_fcn is not None:
            self.new_data_fcn(addr, model, serial)

    def _send_instrument_notification(self, is_boot):
        "Sent the instrument-known or instrument-boot over the Mic Ethernet network."
        info = get_network_info()
        data = json.dumps({
            'event': INST_BOOT_MSG if is_boot else INST_KNOWN_MSG,
            'model': self.model,
            'serial': self.serial,
            'ip_addr': info.mic_ethernet_ip,
            'last_communicated': str(timezone.now()),
            'known_instruments': _known_instruments(),
        }).encode('utf-8')
        # note this message is always boadcast.  That way even if there is a port conflict everyone
        # gets it.
        self._server.sendto(data, (mic_subnet_broadcast(), PORT))


def notify_event(name, **kwargs):
    "This will broadcast an event to all the instruments on the mic-net."

    # create a separate process to send the information.  That way any failure or timeout will not
    # cause an issue with the calling process.  Nor will there be any delay. Exit the process if
    # there was an exception or not.
    if os.fork() == 0:
        try:
            server = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            server.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            server.bind(("", PORT))
            kwargs['event'] = name
            server.sendto(json.dumps(kwargs).encode('utf-8'), (mic_subnet_broadcast(), PORT))
        except: # pylint: disable=bare-except
            pass
        os._exit(os.EX_OK)

def notify_new_data(model, serial):
    "Broadcasts a new data available message from instrument model and serial."
    notify_event(NEW_DATA_MSG, model=model, serial=serial)


# =================================================================================================
def _known_instruments():
    """Get the array of dictionaries for the known instruments (not including us)."""
    inst_qry = KnownInstrument.objects.all().order_by('-last_communicated')[:MAX_RELEVANT]
    return [{
        'model': inst.model,
        'serial': inst.serial,
        'ip_addr': inst.ip_addr,
        'last_communicated': str(inst.last_communicated),
    } for inst in inst_qry]
