"""
Functionality to support the embedded mic network setup configuration and setup.

The system uses nmcli to create named networks:

    - 'Lab Ethernet' on eth0

    - 'Mic Ethernet' on eth1

    - Whatever the wifi name is on wifi.

Lab Ethernet: This is always dhcp and cannot be set.

Mic Ethernet: This is always a 16 bit subnet.  It is either 10.151.x.x, 192.168.x.x, 172.22.x.x.  It
has no need for nameserver nor gateway.  The 192.168.0.x and 192.168.1.x (or other prefix) range is
resreved for older instruments with static IP.

Wifi: This looks for available wifi networks, and prompts the user for which one to connect to, and
what password to use.  It then connects to that one if possible.

On startup (when get_network_info is called) it will create the Lab Ethernet (as dhcp) and Mic
Ethernet connection (as a random ip in 192.168.x.x)
"""
import datetime
import gettext
import random
import re
import glob
import os
import subprocess
import time
from zipfile import ZipFile
from dataclasses import dataclass
from urllib.parse import unquote
from .instrument import is_instrument
from .log_query import post

_ = gettext.gettext

# =================================================================================================
LAB_NAME = 'Lab Ethernet'
MIC_NAME = 'Mic Ethernet'
KNOWN_ETHERNET_NETWORKS = {
    LAB_NAME: 'eth0',
    MIC_NAME: 'eth1',
}


@dataclass
class MicNetworkInfo:
    "The information that is needed to display the network status to the user."
    lab_ethernet_ip: str = ''
    mic_ethernet_ip: str = ''
    wifi_name: str = ''
    wifi_ip: str = ''
    wifi_available: list = None


@dataclass
class WIFINetwork:  # pylint: disable=too-many-instance-attributes
    "The data structure for information on each wifi network."
    in_use: bool
    bsid: str
    ssid: str
    mode: str
    chan: int
    rate: str
    signal: int
    bars: int
    security: list


class NoSuchNetwork(Exception):
    "The exception for if an expected network is not found."


# =================================================================================================
def cache_timeout(timeout):
    "Decorator to cache the function results for timeout seconds."

    def decorator(function):

        def wrapper(*args, **kwargs):
            if function.cache_time is not None:
                elapsed = datetime.datetime.now() - function.cache_time
                if elapsed < datetime.timedelta(seconds=timeout):
                    return function.cache_info
            function.cache_info = function(*args, **kwargs)
            function.cache_time = datetime.datetime.now()
            return function.cache_info

        def reset_cache():
            function.cache_time = None

        wrapper.reset_cache = reset_cache
        reset_cache()
        return wrapper

    return decorator


def only_on_instrument(non_inst_ret):
    "Decorator to call on instrument, but return non_inst_ret if not."

    def decorator(function):

        def wrapper(*args, **kwargs):
            if is_instrument():
                return function(*args, **kwargs)
            return non_inst_ret

        return wrapper

    return decorator


@only_on_instrument(None)
@cache_timeout(timeout=5)
def _bluetooth_scan_devices(timeout=5):
    "Scan for bluetooth devices for timeout seconds."

    process = subprocess.Popen(['bluetoothctl', 'scan', 'on'],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    time.sleep(timeout)
    process.kill()


@only_on_instrument([('30:C9:AB:A7:9C:F0', 'LAPTOP-KN4VQVPL'), ('08:EB:ED:E0:E9:F7', 'NS-CSPBTF1')])
def bluetooth_known_devices():
    "Return a list of mac, name pairs for all known devices found."
    _bluetooth_scan_devices()
    process = subprocess.Popen(['bluetoothctl', 'devices'],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    stdout, _ = process.communicate()
    match_re = re.compile(r"Device (?P<mac>[0-9A-F:]{17}) (?P<name>.*)")
    devices = []

    for line in stdout.decode('utf-8').split('\n'):
        match = match_re.match(line)
        if match:
            devices.append((match.group("mac"), match.group("name")))
    return devices


@only_on_instrument(True)
def bluetooth_connect(mac):
    "Connect a device at the MAC address."
    _bluetooth_scan_devices()
    process = subprocess.Popen(['bluetoothctl', 'pair', mac])
    process.communicate()
    process = subprocess.Popen(['bluetoothctl', 'trust', mac])
    process.communicate()
    process = subprocess.Popen(['bluetoothctl', 'connect', mac])
    process.communicate()
    return process.returncode == 0


@only_on_instrument(None)
def bluetooth_disconnect(mac):
    "Disconnect a device at the MAC address."
    process = subprocess.Popen(['bluetoothctl', 'remove', mac])
    process.communicate()


@cache_timeout(5)
@only_on_instrument(
    MicNetworkInfo(
        lab_ethernet_ip='10.0.0.199',
        mic_ethernet_ip='192.168.25.141',
        wifi_name='dorsey-5',
        wifi_ip='10.0.0.26',
        wifi_available=['WIFI-1', 'WIFI-2'],
    )
)
def get_network_info():
    """Return the MicNetworkInfo structure with the information from this system."""
    net_info = MicNetworkInfo()

    try:
        net_info.lab_ethernet_ip = _get_ip_address(LAB_NAME)
        net_info.mic_ethernet_ip = _get_ip_address(MIC_NAME)
    except NoSuchNetwork:
        # if the network names were not available, set them up. try it once more, then if they are
        # not available, the exception is real.
        _init_ethernet_networks()
        net_info.lab_ethernet_ip = _get_ip_address(LAB_NAME)
        net_info.mic_ethernet_ip = _get_ip_address(MIC_NAME)

    # sometimes the wifi's are not received, so try once more again.
    wifis = _get_wifi_networks()
    if not wifis:
        wifis = _get_wifi_networks()
    for wifi in wifis:
        if wifi.in_use:
            net_info.wifi_name = wifi.ssid
            net_info.wifi_ip = _get_ip_address(wifi.ssid)

    net_info.wifi_available = [wifi.ssid for wifi in wifis]
    return net_info


@only_on_instrument(["/tmp/"])
def get_usb_drive_mounts():
    "Return a list of paths for each usb drive mounted"
    process = subprocess.Popen(['mount'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, _ = process.communicate()
    re_usb = re.compile(r'^/dev/sd\w+\s+on\s+(?P<mount>[/\w]+).*')
    ret = []
    for line in stdout.decode('utf-8').splitlines():
        match = re_usb.match(line)
        if match:
            ret.append(match.group('mount'))
    return ret


@cache_timeout(5)
@only_on_instrument([])
def get_wifi_printers():
    """
    Return a list of wifi printer pairs.  The first element is the internal name, the second is
    the display name.
    """
    process = subprocess.Popen(['lpinfo', '-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, _ = process.communicate()
    re_ipp = re.compile(r'^network (?P<internal>ipp://(?P<user>.*)._ipp._tcp.local/)$')
    ret = []
    for line in stdout.decode('utf-8').splitlines():
        match = re_ipp.match(line)
        if match:
            ret.append((match.group('internal'), unquote(match.group('user'))))
    return ret


def mic_subnet_broadcast():
    "Returns the broadcast address for the mic ethernet network."
    info = get_network_info()
    parts = info.mic_ethernet_ip.split('.')
    if len(parts) == 4:
        parts[2] = '255'
        parts[3] = '255'
    return '.'.join(parts)


def random_ip(prefix, exclude=None):
    "Generate a random ip that is not already in the exclude list."
    # the first number is between 1 and 255, the second is between 0 and 255. The prefix.0 is
    # reserved for older instruments with static IP.
    if exclude is None:
        exclude = []
    while True:
        ip3 = random.randint(2, 255)
        ip4 = random.randint(0, 255)
        ip_addr = prefix + '.' + str(ip3) + '.' + str(ip4)
        if ip_addr not in exclude:
            return ip_addr


@only_on_instrument(None)
def reset_to_pristine_defaults():
    _init_ethernet_networks()
    setup_wifi(None)
    subprocess.Popen(['lpadmin', '-x', 'mic-print'], stdout=subprocess.PIPE,
                     stderr=subprocess.PIPE).communicate()

@only_on_instrument(None)
def setup_mic_ethernet(ip_addr):
    "Set the mic ethernet address to the ip address passed in."
    get_network_info.reset_cache()
    _setup_ethernet(MIC_NAME, dhcp=False, ip_addr=ip_addr, subnet_bits=16)


@only_on_instrument(False)
def setup_wifi(ssid, password=''):
    "Connect to the wifi network with ssid and password."

    # disconnect from previous networks.
    for network in [n for n in _get_wifi_networks() if n.in_use]:
        subprocess.Popen(["nmcli", "con", "delete", network.ssid]).communicate()

    get_network_info.reset_cache()

    # if we are not connecting, we are done.
    if not ssid:
        return True

    # otherwise add it.
    cmd = ['nmcli', 'd', 'wifi', 'connect', ssid]
    if password:
        cmd = cmd + ['password', password]
    nmcli = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = nmcli.communicate()
    get_network_info.reset_cache()
    return (
        nmcli.returncode == 0
        and stdout.decode('utf-8').find("Error:") < 0
        and stderr.decode('utf-8').find("Error:") < 0
    )


def setup_wifi_printer(name):
    "Set the (internally) named printer to be the default wifi printer."
    subprocess.Popen(['lpadmin', '-x', 'mic-print'], stdout=subprocess.PIPE,
                     stderr=subprocess.PIPE).communicate()
    subprocess.Popen([
        'lpadmin', '-p', 'mic-print', '-m', 'everywhere', '-v', name, '-L', 'Mic Printer', '-E'
    ],
                     stdout=subprocess.PIPE,
                     stderr=subprocess.PIPE).communicate()

    # clear the cache on update.
    get_wifi_printers.reset_cache()


def zip_dir_to_usb(src_dir, zipname):
    "zip up the diagnostics files and save them to the USB drive."
    usb_dirs = get_usb_drive_mounts()
    if len(usb_dirs) == 1:
        with ZipFile(usb_dirs[0] + '/' + zipname, 'w') as zip_object:
            for fname in glob.glob(src_dir+"/*"):
                zip_object.write(fname, os.path.basename(fname))
        os.sync()
        post(_("Export to USB is complete. It is safe to remove USB drive."))
    elif len(usb_dirs) > 1:
        post(_("Cannot export to USB. Multiple USB drives found."))
    else:
        post(_("Cannot export to USB. No USB drive found."))
    return {}


def wifi_print_command():
    "The command to print to the wifi printer."
    return 'lp -d mic-print'


# =================================================================================================
def _get_ip_address(name):
    cmd = ['nmcli', '-t', 'connection', 'show', name]
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    if "Error:" in stderr.decode("utf-8"):
        raise NoSuchNetwork

    for option in stdout.decode('utf-8').split('\n'):
        if not option:
            continue
        opt, value = option.split(':', 1)
        if opt.startswith('IP4.ADDRESS') or opt.startswith('ipv4.address'):
            ip_addr = value.split('/')[0].strip()
            if ip_addr:
                return ip_addr

    # The network existed, but it did not have an ip address.  Make sure to also check the
    # ifconfig command.  We saw once that the ifconfig showed an address but nmcli did not.
    if name in KNOWN_ETHERNET_NETWORKS:
        cmd = ['ifconfig', KNOWN_ETHERNET_NETWORKS[name]]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, _ = process.communicate()
        re_ip = re.compile(r'inet\s+(?P<ip>\d+\.\d+\.\d+\.\d+)')
        match = re_ip.search(stdout.decode('utf-8'))
        if match:
            return match.group('ip')

    return ''


def _get_wifi_networks():
    "returns a list of WIFINetwork types for the discoverable wifi networks."
    cmd = [
        'nmcli', '-t', '-f', 'IN-USE,BSSID,SSID,MODE,CHAN,RATE,SIGNAL,BARS,SECURITY', 'device',
        'wifi', 'list'
    ]
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, _ = process.communicate()
    networks = []
    # make sure that the in use networks are shown first, so that they are not eliminated in the
    # duplicate rejection.
    lines = sorted(stdout.decode('utf-8').split('\n'), reverse=True)
    for network in lines:
        if not network:
            continue
        in_use, bssid, ssid, mode, chan, rate, signal, bars, security \
            = re.split(r'(?<!\\):', network)
        ssid = ssid.strip()
        if not ssid:
            continue
        in_use = bool(in_use.strip())
        bars = len(bars.strip())
        bssid = bssid.replace('\\', '')
        chan = int(chan)
        signal = int(signal)
        # we do not support WPA2 with 802.1X which I think requires user/passwd authentication. So
        # remove them from our list.
        if '802' in security:
            continue
        security = security.split()
        # don't allow duplicates.
        if len([n for n in networks if n.ssid == ssid]) >= 1:
            continue
        networks.append(WIFINetwork(in_use, bssid, ssid, mode, chan, rate, signal, bars, security))
    return networks


def _init_ethernet_networks():
    "Initialize the Mic specific networks with default setup."
    subprocess.Popen(['nmcli', 'con', 'delete', 'Wired connection 1']).communicate()
    subprocess.Popen(['nmcli', 'con', 'delete', 'Wired connection 2']).communicate()
    _setup_ethernet(LAB_NAME, dhcp=True)
    _setup_ethernet(
        MIC_NAME,
        dhcp=False,
        ip_addr=random_ip('10.151'),
        subnet_bits=16,
    )


def _setup_ethernet(name, dhcp, ip_addr=None, subnet_bits=None):
    "Setup the ethernet network of name with the configuration."

    # first delete the network if it is already available.
    subprocess.Popen(['nmcli', 'con', 'delete', name]).communicate()

    # now set the configuration of the network.
    cmd = [
        'nmcli', 'con', 'add', 'type', 'ethernet', 'con-name', name, 'ifname',
        KNOWN_ETHERNET_NETWORKS[name]
    ]
    if dhcp:
        cmd = cmd + ['ipv4.method', 'auto']
    else:
        cmd = cmd + ['ip4', ip_addr + '/' + str(subnet_bits)]
    proc = subprocess.Popen(cmd)
    proc.communicate()

    # make sure the data is written to disk.
    proc = subprocess.Popen(['sync'])
    proc.communicate()

    # now make sure it came up.
    if proc.returncode == 0:
        proc = subprocess.Popen(['nmcli', 'con', 'up', name])
        proc.communicate()

    return proc.returncode == 0
