'''The django models for the mic_python_lib.'''
from enum import IntEnum
import gettext
from django.db import models
from django.utils import timezone
from django.utils.functional import lazy
from .mic_json_encoder import MicJSONEncoder

_ = lazy(gettext.gettext, str)

BASE_LANGUAGE_CHOICES = (
    ('en', 'English'),
    ('zh', '中文'),
    ('fr', 'Français'),
    ('de', 'Deutsch'),
    ('it', 'Italiano'),
    ('ja', '日本語'),
    ('ko', '한국어'),
    ('es', 'Español'),
)

LANGUAGE_CHOICES = {
    'en': 'en-US',
    'zh': 'zh-CN',
    'fr': 'fr-FR',
    'de': 'de-DE',
    'it': 'it-IT',
    'ja': 'ja-JP',
    'ko': 'ko-KR',
    'es': 'es-ES',
}

RESULT_RANK_CHOICES = ( # Not user visible
    ("P", "primary"),
    ("S", "secondary"),
    ("C", "calculated"),
    ("R", "runtime")
)

UPDATE_CHOICES = ( # Not user visible
    ("A", "Always"),
    ("L", "Remind Later"),
    ("V", "Remind on New Version"),
    ("N", "Remind Never"),
)

SORT_DIRECTION_CHOICES = (
    ('A', _('Ascend')),
    ('D', _('Descend'))
)

PRINTER_TYPE_CHOICES = (
    ('N', _('None')),
    ('W', _('Wi-Fi')),
    ('S', _('Network')),
)


class Settings(models.Model):
    '''The setting for this instrument.

    NOTE: There should only ever be one of these.

    See also the mic-react-lib Utils.Settings for the context.

    The update_reminder, update_next_date, and update_version work together to decide when the user
    is to be updated of a new version.  If the reminder is ("N", "Remind Never") the user is never
    reminded.  If the reminder is ("L", "Remind Later"), the user is reminded if a new version that
    is not the update_version is available, or if the new version that is the update version is
    still available, and the date is after update_next_date.  If the setting is ("V", "Remind on New
    Version"), the user is reminded whenever there is a new version available. ("A", "Always") will
    always remind if there is a later version available than installed.  This is the default.
    '''
    base_language = models.CharField(max_length=2, choices=BASE_LANGUAGE_CHOICES, default='en')
    region = models.CharField(max_length=2, default='US')
    time_zone = models.CharField(max_length=64, verbose_name=_("Time Zone"), default="America/New_York")
    set_time_automatically = models.BooleanField(
        verbose_name=_("Set Time From Network"),
        default=False
    )
    display_usb = models.BooleanField(default=True, verbose_name=_("Export to USB"))

    export_use = models.BooleanField(
        max_length=255, verbose_name=_("Use Network Drive"), default=False)
    export_user = models.CharField(max_length=255, verbose_name=_("User"), blank=True, default="")
    export_password = models.CharField(
        max_length=127, verbose_name=_("Password"), blank=True, default="")
    export_domain = models.CharField(
        max_length=255, verbose_name=_("Domain"), blank=True, default="")
    export_path = models.CharField(max_length=255, verbose_name=_("Path"), blank=True, default="")

    printer_type = models.CharField(max_length=1, verbose_name=_("Printer Type"),
                                    choices=PRINTER_TYPE_CHOICES, default='N')
    smbprint_user = models.CharField(max_length=255, verbose_name=_("User"), blank=True, default="")
    smbprint_password = models.CharField(
        max_length=127, verbose_name=_("Password"), blank=True, default="")
    smbprint_domain = models.CharField(
        max_length=255, verbose_name=_("Domain"), blank=True, default="")
    smbprint_workgroup = models.CharField(
        max_length=255, verbose_name=_("Workgroup"), blank=True, default="")
    smbprint_name = models.CharField(
        max_length=255, verbose_name=_("Printer Name"), blank=True, default="")
    wifi_printer = models.CharField(
        max_length=255, verbose_name=_("Printer Name"), blank=True, default="")

    # bluetooth speaker.  This has the mac address followed by ';' followed by the name.
    bluetooth_speaker = models.CharField(
        max_length=255, verbose_name=_("Speaker"), blank=True, default="")

    update_reminder = models.CharField(max_length=1, choices=UPDATE_CHOICES, default='A')
    update_next_date = models.DateTimeField(default=timezone.now)
    update_version = models.CharField(max_length=16, default="")

    # 0-n: The current page of the guided setup wizard.
    # -1:  The guided setup wizard is not active.
    guided_setup_page = models.IntegerField(default=-1)

    @property
    def language(self):
        'Return the language for gettext translations'
        return LANGUAGE_CHOICES[self.base_language]

    @property
    def locale(self):
        'Return the locale for ICU formatting'
        return f"{self.base_language}-{self.region}"

    @property
    def bluetooth_speaker_mac(self):
        "The mac (address) part of the bluetooth speaker"
        if self.bluetooth_speaker:
            return self.bluetooth_speaker.split(';')[0]
        return ""

    @property
    def bluetooth_speaker_name(self):
        "The name part of the bluetooth speaker"
        if self.bluetooth_speaker:
            return self.bluetooth_speaker.split(';', 1)[1]
        return ""

    relevant_if = {
        'export_user': ['==', 'export_use', True],
        'export_password': ['==', 'export_use', True],
        'export_domain': ['==', 'export_use', True],
        'export_path': ['==', 'export_use', True],
        'smbprint_user': ['==', 'printer_type', 'S'],
        'smbprint_password': ['==', 'printer_type', 'S'],
        'smbprint_domain': ['==', 'printer_type', 'S'],
        'smbprint_workgroup': ['==', 'printer_type', 'S'],
        'smbprint_name': ['==', 'printer_type', 'S'],
        'wifi_printer': ['==', 'printer_type', 'W'],
    }


class LogAction(IntEnum):
    """
    The choices for actions on log messages.  This is the python
    equivalent of mic-cpp-lib NLogQuery::EUserAction.
    """
    NONE = 0
    POST_ONCE = 1
    POST = 2
    INFORM_WAIT = 3
    YES_NO = 4
    YES_NO_CANCEL = 5
    OK_CANCEL = 6
    POST_UPDATE = 7
    INPUT_FORM = 8
    SAMPLE_LOG = 9


class LogSeverity(IntEnum):
    """
    The choices for severity level of messages.  This is the python
    equivalent of mic-cpp-lib NLogQuery::ESeverity.
    """
    IMPORTED = 0
    VERBOSE = 1
    DEFAULT_ERR = 2
    DEFAULT_INFO = 3


class LogResponse(IntEnum):
    """
    The possible responses from the modal query message.  This is the
    python equivalent of mic-cpp-lib NLogQuery::EUserResponse.
    """
    OK = 0
    YES = OK
    NO = 1
    CANCEL = 2
    NONE = 3


class UnitSelection(models.Model):
    '''The persistent storage for the user selections of models.

    :param name: The unit selection choice from cpp/baseunit.hpp.
    :param choice: The choice that the user made for that unit.
    '''
    name = models.CharField(max_length=64, unique=True)
    choice = models.IntegerField()


class LogMessageManager(models.Manager):  # pylint: disable=too-few-public-methods
    """Accessors for specific types of log messages."""

    def modal(self):
        """:return: the currently active modal messages."""
        # sort by reverse time, so the newest is the first.
        return self.filter(input_context=0).exclude(modal_context=0).order_by("-time")

    def unacknowledged(self):
        """:return: all the unacknowledged messages."""
        # sort by time so the oldest is the first.
        return self.filter(modal_context=0, acknowledged=False).order_by("time")

    def unacknowledged_errors(self):
        """:return: all the unacknowledged error messages."""
        return self.filter(
            modal_context=0,
            acknowledged=False,
            severity=LogSeverity.DEFAULT_ERR,
            action__in=[LogAction.POST_ONCE, LogAction.POST, LogAction.POST_UPDATE]
        )


class LogMessage(models.Model):
    '''The persistent storage for log messages and user queries/modal
    notifications.

    :param time: When the message was created.
    :param action: What should be done with the message.
    :param severity: How severe is the message.
    :param message: The actual text of the message.
    :param user: (Optional) the name of the user.
    :param account: (Optional) the type of account for the user.
    :param post_id: used to replace the (unacknowleged) post once
        string even if the message has changed.
    :param acknowledged: True if the user has acknowledged the message
        (or did not need to).
    :param modal_context: 0 if there is not show modal, but the context if there is.
    :param input_context: if action == INPUT_FORM then this is the context ID.
    :param input_params: if action == INPUT_FORM then this is the input form parameters as a string.
    '''

    time = models.DateTimeField()
    action = models.IntegerField(choices=tuple((i.value, i.name) for i in LogAction))
    severity = models.IntegerField(choices=tuple((i.value, i.name) for i in LogSeverity))
    message = models.TextField()
    user = models.CharField(max_length=64, null=True, default='')
    account = models.CharField(max_length=64, null=True, default='')
    post_id = models.IntegerField(default=0)
    acknowledged = models.BooleanField()
    modal_context = models.IntegerField(default=0)
    input_context = models.IntegerField(default=0)
    input_params = models.TextField(null=True)

    objects = LogMessageManager()


class Result(models.Model):
    """Summary results from analyses.

    :param analysis_technique: The translateable text for what type of analysis was done to
        collect the result.  While this is a translated string, it is stored in English.
    :param result_title: The translateable string that gives the user visible title of the
        result.  While this is a translated string, it is stored in English.
    :param result_name: The internal name of the result.
    :param result_rank: The importance/visibility of the result.
    :param value: The value of the statistic of the result.
    :param std_dev: The standard deviation of the measured result (if available).
    :param units_json: The stringified json object defining the units of the result.
    :param format_json: The stringified json object defining the formatting of the result.

    """
    analysis_technique = models.CharField(max_length=32)
    result_title = models.CharField(max_length=32)
    result_name = models.CharField(max_length=32)
    result_rank = models.CharField(max_length=1, choices=RESULT_RANK_CHOICES, default='P')
    value = models.FloatField()
    std_dev = models.FloatField(null=True)
    units_json = models.TextField(null=True)
    format_json = models.TextField(null=True)


class ResultSortFilter(models.Model):
    '''Persistent storage for result sort and filter settings.

    :param session_key: The primary key of the session
    :param context: The application specific context string
    :param sort_by: The list of (field_name, sort_direction) tuples
    :param filters: The list of (field_name, comparison, comparison_value) tuples

    '''

    session_key = models.CharField(max_length=40, null=True)
    context = models.CharField(max_length=32, null=True)
    sort_by = models.JSONField(default=list, encoder=MicJSONEncoder)
    filters = models.JSONField(default=list, encoder=MicJSONEncoder)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['context', 'session_key'], name='unique_session_context'
            )
        ]


class SampleLog(models.Model):
    "The sample log for any particular results."
    datetime = models.DateTimeField(auto_now=True)
    message = models.CharField(max_length=1024)


class KnownInstrument(models.Model):
    """The table of instruments that have been attached to this unit on the Mic Ethernet network."""
    model = models.CharField(max_length=4, verbose_name=_("Model Number"))
    serial = models.CharField(max_length=8, verbose_name=_("Serial Number"))
    ip_addr = models.CharField(max_length=15, verbose_name=_("IP Address"))
    methods_updated = models.DateTimeField(default=None, null=True)
    last_communicated = models.DateTimeField(default=timezone.now)

    class Meta:
        ordering = ['model', 'serial']
        unique_together = ('model', 'serial')


class InstrumentScriptRunManager(models.Manager):  # pylint: disable=too-few-public-methods
    """Accessors for specific Run results."""
    def active_or_create(self, name):
        active = self.filter(state='P')
        if len(active) == 1:
            return active[0]
        self.filter(state='P').delete()
        return self.create(name=name, state='P', operator="unknown")


class InstrumentScriptRun(models.Model):
    STATE_CHOICES = ( # Not user visible.
        ('P', "In Process"),
        ('C', "Completed"),
        ('X', "Canceled"),
        ('E', "Error"),
    )
    name = models.CharField(max_length=64)
    state = models.CharField(max_length=1, choices=STATE_CHOICES)
    operator = models.CharField(max_length=64)
    passed = models.BooleanField(default=True)
    start_time = models.DateTimeField(default=timezone.now)
    end_time = models.DateTimeField(null=True, default=None)
    report_model = models.BinaryField(null=True, default=None)
    error_output = models.TextField(blank=True, default="")
    user_data_context = models.IntegerField(default=0)

    objects = InstrumentScriptRunManager()

class UIAttached(models.Model):
    """
    The list of recent ui's that have attached to an instrument.

    This also supports if a source is specifically blocked by another ui.
    """
    source_ip = models.CharField(max_length=15)
    source_uuid = models.CharField(max_length=64, default="")
    last_date_time = models.DateTimeField()
    blocked = models.BooleanField(default=False)
    blocked_by_ip = models.CharField(max_length=15, default="")
    blocked_by_uuid = models.CharField(max_length=64, default="")
