'''
The python interface between the mic-cpp-lib units and react-ui units
'''
import json
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from .models import UnitSelection
from .settings import unit_sel_and_simple, unit_update_selection, unit_udp_notify_updated
from .mic_json_encoder import MicJSONEncoder


def notify_unit_selections():
    """Calls the notify_updates with the current unit selections formatted in json."""
    unit_udp_notify_updated(json.dumps({us.name: us.choice for us in UnitSelection.objects.all()}))


@csrf_exempt
def init_units_view(request):
    '''Initialize the units table, and return unit selections in the request.
    '''

    # get the c++ data
    data = json.loads(unit_sel_and_simple())

    # update the database, and fill in the default selections.
    for name, sel in data["selections"].items():
        _update_selection(name, sel)

    # return the augmented data
    return JsonResponse(data, encoder=MicJSONEncoder)


def update_selection(sel, choice):
    '''
    Update the unit selection of sel to be choice.  Include any linked
    units and notify everyone.  Return the linked units.
    '''
    linked = json.loads(unit_update_selection(sel, choice))
    linked[sel] = choice

    for name, linked_choice in linked.items():
        _update_choice(name, linked_choice)

    # now that everything is updated, also notify the interested function.
    notify_unit_selections()

    return linked


def _update_selection(name, sel):
    '''Make sure the selection is saved in the database, and augment sel to have the selected
    choice.

    :param name: The name of the unit selection data object.
    :param sel: the data associated with the unit selection.
    '''
    try:
        obj = UnitSelection.objects.get(name=name)
    except UnitSelection.DoesNotExist:
        obj = UnitSelection(name=name, choice=sel['choices'][0]['choice'] if sel['choices'] else 0)
        obj.save()
    sel['choice'] = obj.choice


def _update_choice(name, choice):
    '''Update the saved user choice for this selection.

    :param name: The name of the unit selection data object.
    :param choice: the data integer choice the user made.
    '''

    try:
        obj = UnitSelection.objects.get(name=name)
    except UnitSelection.DoesNotExist:
        obj = UnitSelection(name=name)
    obj.choice = choice
    obj.save()
