"""Common rest functionality"""
import json
import re
from collections import namedtuple
import requests
from django.db import transaction
from django.db.models import Min, Max, Q
from django.core.exceptions import FieldDoesNotExist
from django.forms.models import model_to_dict
from django.http import JsonResponse
from django.urls import path
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from .mic_json_encoder import MicJSONEncoder
from .models import KnownInstrument


class RestError(Exception):
    '''Base class for exceptions in this module'''


class BadInstrument(Exception):
    "The exception thrown when reading of the rest data from an instrument is not consistent. "


UpdateMethod = namedtuple(
    "UpdateMethod",
    ['name', 'model', 'fields', 'key_fields', 'set_field_fcn'])
UpdateMethod.__doc__ = """
    This defines the parameters that method tables need to share their data between
    instruments. Methods can be both the full analysis parameters methods along with ForeignKey sub
    objects that also be a part of the  full method parameters.   (For example inserts in the 1350,
    or fluids in the 7100.)

    name - The name type of result that is used to identify it in the rest api.
    model - the django model class that holds this information.
    fields - An array of strings.  Each entry is the name of a field that will be part of the json
             export for this result.
    key_fields - The array of fields names that uniquely identify the method in the table.
    set_field_fcn - A function (obj, name, value) that will set the field(name) in the object(obj)
                    with the json value of value.   For simple types this is simply the built in
                    python setattr.  For compound types this should be application specific.
"""


UpdateResult = namedtuple(
    "UpdateResults",
    ['name', 'res_model', 'instrument_field', 'filter_args', "fields",
     'update_field', 'set_params', 'set_field_fcn', 'update_results_fcn', 'key_fields'])
UpdateResult.__doc__ = """
    This defines the parameters that results tables need to share data between instruments.  Each of
    these must be directly or indirectly linked to an instrument object.

    name - The name type of result that is used to identify it in the rest api.
    res_model - the django model class that holds this information.
    instrument_field - The field that stores the instrument object.  If it is nested, it is
                       separated with the usual django chaining of __.
    filter_args - the arguments that are used in filtering the table, so that only the objects that
                  we want to share are shared.  This is typically used to not share canceled
                  results.
    fields - An array of strings.  Each entry is the name of a field that will be part of the json
             export for this result.
    update_field - the name of the field that indicates the date/time of the last update of the
                   object.
    set_params - the way to set parameters that are not defaulted, and not part of the transfer.
                 This works in coordination with the filter_args and fields to make sure everything
                 is set.
    set_field_fcn - the function (obj, model, serial, name, and value) the function that sets the
                    field (of name name) in the object (obj) with the value that is a json value.
                    The model and serial are the model number and serial number of the instrument
                    associated with this result.
    update_results_fcn - the callback function (taking an object of type res_model) that is called
                         to indicate that a new result is available.  This is what should trigger
                         the broadcast to the other instruments.   This can be None if there is no
                         trigger for this result.  (Normally, only analysis results trigger, the
                         dependent calibrations are requested when the result is desired.)
    key_fields - the list of fields names the form the key (along with the instrument field) for the
                 result.
"""

class RestAPIData: # pylint: disable=too-many-instance-attributes, too-many-arguments
    """
    The class that defines a particular version of the rest API.

    version - the version string for the api (e.g. api-v1)
    inst_model - the django model class for the instrument that will be converted.
    inst_fields - an array of strings that are the fields in the model that will be transferred.
    result_order_by - The field in the instrument model, by which the instruments should be sorted.
    result_filter - The filter dictionary for which to look at the instruments.
    methods - The array of UpdateMethod parameter sets that will be used to share the methods
              between instruments.  Note that these should be in dependency order so that the
              components of the later method types are already transferred.
    results - The array of UpdateResult parameter sets that will be used to share the results
              between instruments.  Note that these should be in dependency order so that the
              components of the later results types are already transferred.  (I.e. calibrations
              should come before analysis results.)
    format_table - the list of functions that are used in the field lists to format members into
                   json format in an application specific way.
    """
    _FCN_REGEX = re.compile(r'(?P<fcn>\w+)\((?P<prm>\w+)\)')

    def __init__(self, version, inst_model, inst_fields, result_order_by, result_filter,
                 methods, results, format_table):
        self.version = version
        self.inst_model = inst_model
        self.inst_fields = inst_fields
        self.result_order_by = result_order_by
        self.result_filter = result_filter
        self.methods = methods
        self.results = results
        self.format_table = format_table


    def dict_from_fields(self, obj, fields):
        """
        Return a (possible nested) dictionary for the ojbect based on the fields.

        The attributes are . separated to indicate nesting in the object and resultant dictionary.

        :param obj: The object to read the fields from.
        :param fields: an array of attributes that the _value_from_attr can use.
        """

        ret = {}
        for field in fields:
            try:
                # if the fieldname happens to be a 2 character string, it is not a pair.
                if isinstance(field, str) and len(field) == 2:
                    raise ValueError()
                field_name, field_source = field
            except ValueError:
                field_name = None
                field_source = field
            value = obj
            sub_dict = ret
            attrs = field_source.split('.')
            for attr in attrs[:-1]:
                attr, value = self._value_from_attr(value, attr)
                if field_name is None:
                    sub_dict = sub_dict.setdefault(attr, {})
            attr, value = self._value_from_attr(value, attrs[-1])
            if field_name is not None:
                attrs = field_name.split('.')
                for attr in attrs[:-1]:
                    sub_dict = sub_dict.setdefault(attr, {})
                attr = attrs[-1]
            sub_dict[attr] = value
        return ret

    def urls(self):
        "All the urls that this class is associated with."
        urls = [
            path(f'{self.version}/instrument_config/', function_get_to_json_view(
                lambda: _instrument_config(self))),
            path(f'{self.version}/instruments_with_results/', function_get_to_json_view(
                lambda: _instrument_with_results(self))),
            path(f'{self.version}/get_last_updates/', function_get_to_json_view(
                lambda: _get_last_updates(self))),
            path(f'{self.version}/get_instrument_ids/', function_get_to_json_view(
                lambda since=None: _get_model_ids(self.inst_model, 'last_updated', since))),
            path(f'{self.version}/instrument/', function_get_to_json_view(
                lambda oid: self._get_model_obj(
                    self.inst_model, self.inst_fields, oid))),
        ]
        urls += self._url_methods()
        urls += self._url_method_ids()
        urls += self._url_results()
        urls += self._url_result_ids()
        return urls

    def update_from_instrument(self, ip_addr):
        """
        Updates our stored data by adding all the new data available from the other instrument.

        This will first get all is_method=True inserts, AnalysisParameters, and Instruments from
        the instrument at ip_addr.  It gets all of them since it cannot tell which ones were changed
        locally, and which ones are changed remotely.  It will override the local copies whenever
        the last_updated is later than the local copy.

        Then it will get the new Calibrations and AnalysisResults for each instrument (model/serial)
        know to the remote instrument.
        """
        config = _get_rest_json(ip_addr, 'instrument_config')
        # make sure the instruments are compatible
        if self.inst_model.match_local_config(config):
            with transaction.atomic():
                model = config['model']
                serial = config['serial']
                try:
                    since = KnownInstrument.objects.get(model=model, serial=serial).methods_updated
                except KnownInstrument.DoesNotExist:
                    since = None
                for update_method in self.methods:
                    _update_methods(ip_addr, self.inst_model, since, update_method)
                # after everything is updated, save the known time.
                known, _ = KnownInstrument.objects.get_or_create(model=model, serial=serial)
                known.methods_updated = timezone.now()
                known.save()

        # now update the instruments and results.
        _update_instruments(self.inst_model, ip_addr)
        data = _get_rest_json(ip_addr, 'get_last_updates')
        for result_type in self.results:
            for res in data[result_type.name]:
                _update_result(ip_addr, result_type,
                               res['model'], res['serial'], parse_datetime(res['last_updated']))

    def _get_model_obj(self, model, fields, oid, **kwargs):
        "return the dictionary parameters for object with id=oid"
        return self.dict_from_fields(model.objects.filter(**kwargs).get(id=oid), fields)

    def _url_methods(self):
        return [
            path(f'{self.version}/{method.name}/', function_get_to_json_view(
                lambda oid, method=method: self._get_model_obj(
                    method.model, method.fields, oid, is_method=True, is_shared=True)))
            for method in self.methods]


    def _url_method_ids(self):
        return [
            path(f'{self.version}/get_{method.name}_ids/', function_get_to_json_view(
                lambda since=None, method=method: _get_model_ids(
                    method.model, 'last_updated', since, is_method=True, is_shared=True)))
            for method in self.methods]

    def _url_results(self):
        return [
            path(f'{self.version}/{result.name}/', function_get_to_json_view(
                lambda oid, result=result: self._get_model_obj(
                    result.res_model, result.fields, oid)))
            for result in self.results]

    def _url_result_ids(self):
        return [
            path(f'{self.version}/get_{result.name}_ids/', function_get_to_json_view(
                lambda model, serial, since=None, result=result: _get_model_ids(
                    result.res_model, result.update_field, since,
                    **result.filter_args,
                    **{f'{result.instrument_field}__model': model,
                       f'{result.instrument_field}__serial': serial}
                )))
            for result in self.results
        ]

    def _value_from_attr(self, obj, attr):
        """
        @returns the attribute name and value pair for the attrubute in object.

        :param obj: The object that contains the attribute.
        :param attr: Can either be a simple name, or a function call with a simple name parameter.
        """
        match = self._FCN_REGEX.match(attr)
        if match:
            attr = match.group('prm')
            fcn = match.group('fcn')
            _, value = self._value_from_attr(obj, attr)
            value = self.format_table[fcn](value)
        else:
            value = getattr(obj, attr)
        return attr, value


def function_get_to_json_view(fcn):
    "Helper function used to unpack get request data"

    def view(request):
        # turn get to dict.  Here we assume that only one instance per key.
        prms = {k: request.GET[k] for k in request.GET.keys()}
        return JsonResponse(fcn(**prms), safe=False)

    return view


def function_to_json_view(request, service_fcn, **kwargs):
    "Helper function used to unpack post request data"
    if request.body:
        data = json.loads(request.body.decode('utf-8'))
    else:
        data = {}
    try:
        return JsonResponse({
            'error': False,
            'return': service_fcn(**{
                **data,
                **kwargs
            }),
        },
                            encoder=MicJSONEncoder)
    except RestError as err:
        return JsonResponse({'error': True, 'message': str(err)}, encoder=MicJSONEncoder)
    return JsonResponse({'error': False}, encoder=MicJSONEncoder)


def make_json_view(fcn):
    "Make a json view function from fcn"
    return lambda request: function_to_json_view(request, fcn)


def merge_model_dict(#pylint: disable=too-many-arguments
        model, obj, known_fields=None, on_set=None, on_missing=None, on_extra=None):
    """
    This creates and returns an instance of model from the obj dictonary and fills in the known
    fields.

    :param model: the model class that is to be created.
    :param obj: The dictionary of model fields that will be added to the model object.
    :param known_fields: The dictionary of extra parameters that are filled in not from obj.
    :param on_set: The function that sets the parameter from obj, and known_fields.
    :param on_missing: the function (if any) to call when this finds a missing parameter in
         the model.
    :param on_extra: the function (if any) to call when this finds a extra parameter in obj.
    """

    def _set_model_value(name, value):
        if on_set is not None:
            on_set(name, value)

    # Set the know values.
    if known_fields is None:
        known_fields = {}
    for name, value in known_fields.items():
        _set_model_value(name, value)

    # The <name>_set fields are unusual.  They have backlinks of form <name> but are used as
    # <name>_set.  They are also optional (since they may be empty).  Make a list of the ones used
    # in the object, and allow those to be matched.
    set_fields = {name for name in obj if name.endswith('_set') and _has_set_field(model, name)}

    # check for missing fields, and record all the model fields.
    model_fields = set()
    for field in model._meta.get_fields():
        # skip over the id link it is not sent.
        if _is_skip_model_field(field, known_fields):
            continue
        model_fields.add(field.name)
        if field.name not in obj:
            if on_missing is not None:
                on_missing(field.name)

    # set all the data onto the object and note any missing fields.
    for name in obj:
        if name not in model_fields and name not in set_fields:
            if on_extra:
                on_extra(name)
        else:
            if name not in set_fields:
                _set_model_value(name, obj[name])

    # do the set fields last.  This is so that the rest of the object is complete and can be saved.
    for name in obj:
        if name in set_fields:
            _set_model_value(name, obj[name])


def model_function_to_json_view(request, model_fcn):
    "Helper function used to unpack post request data"
    if request.body:
        data = json.loads(request.body.decode('utf-8'))
    else:
        data = {}
    model = model_fcn(**data)
    if model is not None:
        return JsonResponse(model_to_dict(model), encoder=MicJSONEncoder)
    return JsonResponse({})


# =================================================================================================
def _get_last_updates(api: RestAPIData):
    "Get the last update times for each table (and by instrument)."
    def _last_updated_method(model):
        last = model.objects.filter(is_method=True, is_shared=True).order_by('last_updated').last()
        return  (
            last.last_updated
            if last is not None else timezone.datetime.min.replace(tzinfo=timezone.utc)
        )

    def _add_last_updated_result(latest, model, inst, order_by, **kwargs):
        obj = model.objects.filter(**kwargs).order_by(order_by).last()
        if obj is not None:
            latest.append({
                'model': inst.model,
                'serial': inst.serial,
                'last_updated': getattr(obj, order_by),
            })

    def _last_by_instrument(instruments, model, order_by, instrument_field, **kwargs):
        latest = []
        for inst in instruments:
            filter_args = kwargs | {instrument_field: inst}
            _add_last_updated_result(
                latest,
                model,
                inst,
                order_by,
                **filter_args,
            )
        return latest

    updates = {}
    inst_all = api.inst_model.objects.all()
    updates['instrument'] = inst_all.order_by('last_updated').last().last_updated

    for result in api.results:
        updates[result.name] = _last_by_instrument(
            api.inst_model.objects.all(),
            result.res_model,
            result.update_field,
            result.instrument_field,
            **result.filter_args,
        )

    for method in api.methods:
        updates[method.name] = _last_updated_method(method.model)

    return updates


def _get_model_ids(model, order_by, since=None, **kwargs):
    "Get all the known method IDs (in last updated order) since the date passed."
    qry = model.objects.filter(**kwargs)
    if since is not None:
        qry = qry.filter(last_updated__gt=since)
    qry = qry.order_by(order_by, 'id')
    return {'ids': list(qry.values_list('id', flat=True))}


def _get_rest_json(ip_addr, endpoint, **kwargs):
    "Get the json data from the endpoint url at ip_addr.  Raise BadInstrument if nothing there."
    url = 'http://' + ip_addr + '/rest/api-v1/' + endpoint + '/'
    page = requests.get(url, params=kwargs)
    if page.status_code != 200:
        raise BadInstrument
    return page.json()


def _has_set_field(model, set_field):
    "Does the model have the set_field parameter, i.e. is there a back link."
    try:
        model._meta.get_field(set_field[:-4])
        return True
    except FieldDoesNotExist:
        return False


def _instrument_config(api: RestAPIData):
    '''The view that returns all the known instruments with available results.'''
    return api.inst_model.local_config()


def _instrument_with_results(api: RestAPIData):
    '''The view that returns all the known instruments with available results.'''
    instruments = api.inst_model.objects.annotate(
        last=Max(api.result_order_by, filter=Q(**api.result_filter)),
        first=Min(api.result_order_by, filter=Q(**api.result_filter))
    )
    data = {
        'instruments': [{
            'model': inst.model,
            'serial': inst.serial,
            'first': inst.first,
            'last': inst.last,
        } for inst in instruments],
    }
    return data


def _is_skip_model_field(field, known_fields):
    "Should the field be skipped in setting parameters. All known_fields should be skipped."
    # if it is specifically in known fields, skip the check.
    if field.name in known_fields:
        return True
    # skip the database id field and/or the base class pointer.
    if field.name == 'id' or field.name.endswith('_ptr'):
        return True

    # skip the automatically crated back links.  I'm not sure this is correct, but it seems to work.
    if hasattr(field, 'parent_link') and not field.parent_link:
        return True

    return False


def _update_instruments(inst_model, ip_addr):
    "Get latest known instruments from instrument at ip_addr and update any later ones in our db."
    for oid in _get_rest_json(ip_addr, 'get_instrument_ids')['ids']:
        inst_dict = _get_rest_json(ip_addr, 'instrument', oid=oid)
        inst, created = inst_model.objects.get_or_create(
            model=inst_dict['model'], serial=inst_dict['serial']
        )
        updated = parse_datetime(inst_dict['last_updated'])
        if created or updated > inst.last_updated:
            merge_model_dict(
                inst_model,
                inst_dict,
                on_set=lambda name, value, inst=inst: setattr(inst, name, value)
            )
            inst.save(update=False)


def _update_methods(ip_addr, inst_model, since, update_method: UpdateMethod):
    "Update the method model objects from ip_addr defined by update_method."

    ids = _get_rest_json(ip_addr, f'get_{update_method.name}_ids')
    for oid in ids['ids']:
        # skip over ones that nave not changed since the last time.
        obj_dict = _get_rest_json(ip_addr, update_method.name, oid=oid)
        updated = parse_datetime(obj_dict['last_updated'])
        if since and since >= updated:
            continue

        # now update or create.
        key_dict = {field: obj_dict[field] for field in update_method.key_fields}
        key_dict['is_method'] = True
        try:
            obj = update_method.model.objects.get(**key_dict)
            # if we have that named object, and it is not marked for sharing, don't update.
            if not obj.is_shared:
                return
            created = False
        except update_method.model.DoesNotExist:
            obj = update_method.model(**key_dict)
            obj.is_shared = True
            created = True
        if created or updated > obj.last_updated:
            merge_model_dict(
                update_method.model,
                obj_dict,
                on_set=lambda name, value, obj=obj: update_method.set_field_fcn(obj, name, value),
            )
            obj.save(update=False)


def _update_result(ip_addr, result_type: UpdateResult, model, serial, upto):
    "Get the latest results from the instrument at ip_addr from instrument model/serial."
    # pylint: disable=too-many-locals
    select = {
        result_type.instrument_field.replace('.', '__') + "__model": model,
        result_type.instrument_field.replace('.', '__') + "__serial": serial,
    }

    last = result_type.res_model.objects.filter(**select).order_by(
        '-' + result_type.update_field).first()
    if last is None:
        oids = _get_rest_json(ip_addr, f'get_{result_type.name}_ids',
                              model=model, serial=serial)['ids']
    else:
        since = getattr(last, result_type.update_field)
        if since >= upto:
            return
        oids = _get_rest_json(
            ip_addr, f'get_{result_type.name}_ids', model=model, serial=serial, since=str(since)
        )['ids']

    for oid in oids:
        res_dict = _get_rest_json(ip_addr, result_type.name, oid=oid)
        try:
            select_start = select.copy()
            for field in result_type.key_fields:
                select_start[field] = res_dict[field]
            res = result_type.res_model.objects.get(**select_start)
            created = False
        except result_type.res_model.DoesNotExist:
            res = result_type.res_model()
            created = True
        updated = parse_datetime(res_dict[result_type.update_field])
        if created or updated > res.last_updated:
            merge_model_dict(
                result_type.res_model,
                res_dict, result_type.set_params,
                on_set=lambda name, value, res=res:
                    result_type.set_field_fcn(res, model, serial, name, value)
            )
            res.save(update=False)
            if result_type.update_results_fcn:
                result_type.update_results_fcn(res)
