'''
rest_input.py contains the generic reusable python code for creating a django rest server for
a react InputForm from mic-react-lib.
'''
import gettext
import json
from abc import ABC, abstractmethod
from django.core.exceptions import ObjectDoesNotExist, FieldDoesNotExist
from django import forms
from django.db import IntegrityError
from django.http import JsonResponse, Http404
from django.shortcuts import redirect
from django.db import models
from .icu_datetime import ICUDateTimeFormField
from .mic_json_encoder import MicJSONEncoder

_ = gettext.gettext
# =================================================================================================
class ReactFormMixin:  # pylint: disable=too-few-public-methods
    '''
    Use the ReactFormMixin when you want to create a json rest api that will
    respond with field definitions and data for an InputForm component from mic-react-lib
    from a django model.

    To use the ReactFormMixin, a view must inherit from the imported mixin and define the
    ReactMeta class with these attributes:

        - model: imported model class used to create form if not provided

        - form: a django form to be used for field/form validation

        - get_default_obj: This is a function that should return an object of type model. If
          there is no id get parameter in the url, it is this object that is edited.  This
          function takes a request object as a parameter.

        - on_save: This is a function that is called after the edited object was saved. It
          can be used to trigger application level post save functionality.  This function
          takes the saved object as a parameter.

        - queryset: This is a django queryset that represents the objects to be returned by
          the ?action=list get request.

        - get_selected: This is a function that is called to get the id of the object that
          is marked as selected in the ?action=list response. If passed, this will display
          a switch on the list of objects in the list view.

        - select_object: This is a function that is called to mark the object as selected.
          This function takes the object_id as a parameter.

        - delete_object: This is a function that is used to override the default functionality
          of the ?action=delete get request. This function takes the id of the object to be
          deleted as a parameter.

        - add_object: this is a function that is used to override the default functionality
          of the ?action=add get request.

        - get_derived: this is a function that is used to get a dictionary of derived quantities in
          the form.  It takes an instance of the model object with edited values and returns the
          dictionary of fields names to values.

    These attributes of ReactMeta are optional:

         - relevant_if: dictionary of rules that represent whether a field should be displayed. This
           allows for showing/hiding fields based on complex rules that refer to other field
           values. The key is the field name to control.  The value is the rule, or a callable that
           returns the rule.

         - disabled: dictionary of booleans or functions that return boolean that represent whether
           a field should be disabled. Disabled means that the field is appears greyed-out and when
           it is clicked, the keyboard does not open.

         - hidden: dictionary of booleans of functions that return boolean that reepresent whether
           a field should be hidden. Hidden means that the field is not shown at all.

         - copy_fail_messages: a dictionary that is used when a save a copy of the object causes an
           integrity error.  The keys are the string representations of the error and the values are
           the string to be used to notify the user why the copy failed.

    The get function is expecting a url get parameter like `/test/?id=1`, and will return
    a 404 the object doesn't exist.  If the id parameter is not present, it will call the
    get_default_obj method set in ReactMeta if it exists.

    The post function will save the edited values from the object identified by the id get
    parameter, or the ReactMeta.get_default_obj() if it exists.  After the object is saved,
    the ReactMeta.on_save() is called if it exists.

    example =>

        class ExampleRESTView(ReactFormMixin, View):
            class ReactMeta:
                model = ExampleModel

    '''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''defines which model or form the view displays and edits'''

        model = None
        form = None
        get_default_obj = lambda request: None
        on_save = lambda obj: None
        queryset = None
        get_selected = lambda: None
        select_object = lambda obj_id: None
        delete_object = lambda obj_id: None
        add_object = lambda: None
        get_derived = None
        copy_fail_messages = None

    def get_model_type(self):
        '''return model from form if not provided'''
        try:
            return self.ReactMeta.model
        except AttributeError:
            return self.ReactMeta.form._meta.model

    def get_form_type(self):
        '''create form from model if not provided'''
        try:
            model_form = self.ReactMeta.form
            fields = list(model_form().fields)  # pylint: disable=not-callable
        except AttributeError:
            model_form = forms.ModelForm
            fields = '__all__'
        return forms.modelform_factory(
            self.get_model_type(),
            form=model_form,
            formfield_callback=_get_form_field,
            fields=fields
        )

    def get_input_form_props(self, object_id):
        '''return input form props from form'''
        form_type = self.get_form_type()
        try:
            relevant_if = self.ReactMeta.relevant_if
        except AttributeError:
            try:
                relevant_if = self.get_model_type().relevant_if
            except AttributeError:
                relevant_if = {}
        disabled = _get_field_attributes(self.ReactMeta, "disabled")
        hidden = _get_field_attributes(self.ReactMeta, "hidden")
        if object_id:
            try:
                model_object = self.get_model_type().objects.get(id=object_id)
                form = form_type(instance=model_object)
            except ObjectDoesNotExist:
                raise Http404(str(self.get_model_type()) + " does not exist.")
        else:
            form = form_type()
        parameters = _field_parameters(form)
        data = {field.name: field.value() for field in form}
        for name, value in data.items():
            # if the initial value was not set in the field, see if we can get it from the model.
            if value is None:
                try:
                    default = self.get_model_type()._meta.get_field(name).default
                    if default is not models.fields.NOT_PROVIDED:
                        if callable(default):
                            data[name] = default()
                        else:
                            data[name] = default
                except FieldDoesNotExist:
                    pass
        return {
            'format': {
                'fields': parameters,
                'layout': list(form.fields)
            },
            'data': data,
            'disabled': disabled,
            'hidden': hidden,
            'relevantIf': _eval_relevant_if(relevant_if)
        }

    def get_object_list(self):
        '''return object with id/name pairs for the model object'''
        def dict_from_obj(obj):
            ret = {"id": obj.id, "name": str(obj)}
            if hasattr(obj, 'is_default'):
                ret['is_default'] = obj.is_default
            return ret

        if hasattr(self.ReactMeta, 'queryset'):
            objects = self.ReactMeta.queryset.all()  # re-evaluate queryset
        else:
            objects = self.get_model_type().objects.all()
        if hasattr(self.ReactMeta, 'get_selected'):
            selected = self.ReactMeta.get_selected()
        else:
            selected = -1
        return {
            'objects': [dict_from_obj(obj) for obj in objects],
            'selected': selected,
        }

    def get_select_object(self, object_id):
        '''select object with given id'''
        if hasattr(self.ReactMeta, 'select_object'):
            return self.ReactMeta.select_object(object_id)
        return {}

    def get_delete_object(self, object_id):
        '''delete object with given id'''
        if hasattr(self.ReactMeta, 'delete_object'):
            return self.ReactMeta.delete_object(object_id)
        return self.get_model_type().objects.get(id=object_id).delete()

    def get_add_object(self):
        '''create new object and return id'''
        if hasattr(self.ReactMeta, 'add_object'):
            return self.ReactMeta.add_object()
        return {'object_id': self.get_model_type().objects.create().id}

    def id_from_request(self, request):  # pylint: disable=no-self-use
        "Return the object ID from the request parameters."
        return request.GET.get('id', None)

    def get(self, request):
        '''return json response depending on action'''
        action = request.GET.get('action', None)
        object_id = self.id_from_request(request)
        if action == 'list':
            return JsonResponse(self.get_object_list(), encoder=MicJSONEncoder)
        if action == 'select':
            return JsonResponse(self.get_select_object(object_id), encoder=MicJSONEncoder)
        if action == 'delete':
            return JsonResponse(self.get_delete_object(object_id), encoder=MicJSONEncoder)
        if action == 'add':
            return JsonResponse(self.get_add_object(), encoder=MicJSONEncoder)
        if not object_id and hasattr(self.ReactMeta, 'get_default_obj'):
            url = request.META["PATH_INFO"]
            return redirect(f'{url}?id={self.ReactMeta.get_default_obj(request).id}')
        return JsonResponse(self.get_input_form_props(object_id), encoder=MicJSONEncoder)

    def form_from_request(self, request):
        "Create the proper form object for this class and request."
        post = json.loads(request.body.decode('utf-8'))
        if 'id' not in request.GET and hasattr(self.ReactMeta, 'get_default_obj'):
            obj = self.ReactMeta.get_default_obj(request)
        elif 'id' in request.GET:
            obj_id = self.id_from_request(request)
            obj = self.get_model_type().objects.get(id=obj_id)
        else:
            obj = None
        form_type = self.get_form_type()
        form = form_type(post.get('data'), instance=obj)
        return form

    def save_valid_form(self, form):
        "Save the form and call the on_save method if any."
        obj = form.save()
        if hasattr(self.ReactMeta, 'on_save'):
            self.ReactMeta.on_save(obj)

    def save_copy_valid_form(self, form):
        "Save the form as a new object and call the on_save method if any."
        obj = form.instance
        obj.pk = None
        print("save_copy", obj.__dict__)
        try:
            obj.save()
        except IntegrityError as err:
            error = _("Cannot save copy.")
            if (hasattr(self.ReactMeta, 'copy_fail_messages')
                and str(err) in self.ReactMeta.copy_fail_messages):
                error = self.ReactMeta.copy_fail_messages[str(err)]
            return JsonResponse({'error': error})
        if hasattr(self.ReactMeta, 'on_save'):
            self.ReactMeta.on_save(obj)
        return JsonResponse({})

    def post(self, request):
        '''Validate or save form data'''
        post = json.loads(request.body.decode('utf-8'))
        form = self.form_from_request(request)
        if (hasattr(self.ReactMeta, 'get_derived')
            and self.ReactMeta.get_derived
            and form.is_valid()
            and post.get('action') == 'derived'):
            # pylint: disable=not-callable
            return JsonResponse({
                'derived': self.ReactMeta.get_derived(form.instance),
            })
        if form.is_valid() and post.get('action') == 'copy':
            return self.save_copy_valid_form(form)
        if form.is_valid() and post.get('action') == 'save':
            self.save_valid_form(form)
        return post_response(form)


class ReactFormDataMixin(ABC):  # pylint: disable=too-few-public-methods
    '''
    Use the ReactFormDataMixin when you want to create a json rest api that will
    respond with field definitions and data for an InputForm component from mic-react-lib
    but you don't have a django model.

    To use the ReactFormDataMixin, a view must inherit from the imported mixin and define the
    ReactMeta class with these attributes:

        - data: A dictionary of field name to django model field table.

        - form: a django form to be used for field/form validation

        - fields: tuple of strings of model field names from the above model

    These attributes of ReactMeta are optional:

        - relevant_if: dictionary of rules that represent whether a field should be displayed. This
          allows for showing/hiding fields based on complex rules that refer to other field
          values. The key is the field name to control.  The value is the rule, or a callable that
          returns the rule.

        - disabled: dictionary of booleans or functions that return boolean that represent whether
          a field should be disabled. Disabled means that the field is appears greyed-out and when
          it is clicked, the keyboard does not open.

        - hidden: dictionary of booleans of functions that return boolean that reepresent whether
          a field should be hidden. Hidden means that the field is not shown at all.

    '''

    @abstractmethod
    def get_initial_values(self, request):
        '''Pure virtual function.  Called to get the initial data.'''

    @abstractmethod
    def on_save(self, request, data):
        '''Pure virtual function.  Called to handle the response.'''

    def on_delete(self, request):
        '''Optional function called to handle the delete button.'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''defines which fields the view displays and edits'''

        data = None
        form = None

    def get_form_type(self):
        '''Create form from data if not provided'''
        try:
            form_type = self.ReactMeta.form
        except AttributeError:
            fields = {k: _get_form_field(v) for k, v in self.ReactMeta.data.items()}
            form_type = type('ReactDataForm', (forms.BaseForm,), {'base_fields': fields})
        return form_type

    def get_input_form_props(self, request):
        '''return input form props using form definition'''
        try:
            relevant_if = self.get_form_type().relevant_if
        except AttributeError:
            try:
                relevant_if = self.ReactMeta.relevant_if
            except AttributeError:
                relevant_if = {}
        disabled = _get_field_attributes(self.ReactMeta, "disabled")
        hidden = _get_field_attributes(self.ReactMeta, "hidden")
        form_type = self.get_form_type()
        form = form_type()
        data = self.get_initial_values(request)
        parameters = _field_parameters(form)
        return {
            'format': {
                'fields': parameters,
                'layout': list(form.fields),
            },
            'data': data,
            'disabled': disabled,
            'hidden': hidden,
            'relevantIf': _eval_relevant_if(relevant_if)
        }

    def get(self, request):
        '''return json response with input form props'''
        action = request.GET.get('action', None)
        if action == 'delete':
            return JsonResponse(self.on_delete(request), encoder=MicJSONEncoder)
        return JsonResponse(self.get_input_form_props(request), encoder=MicJSONEncoder)

    def post(self, request):
        '''Update the instrument information based on what the user edited.'''
        post = json.loads(request.body.decode('utf-8'))
        form_type = self.get_form_type()
        form = form_type(post.get('data'))
        if form.is_valid() and post.get('action') == 'save':
            self.on_save(request, form.cleaned_data)
        return post_response(form)


class UnitsFloatField(models.FloatField):
    '''
    The UnitsFloatField can be used by a django model in place of a
    FloatField, to store a decimal along with the units, min_value, max_value
    and decimal_places attributes that are needed by the InputNumber in
    mic-react-lib.

     UnitsFloatField accepts the following parameters:

        - fmt: An exposed C++ formatted double type that defines the type.
    '''

    def __init__(self, fmt, *args, **kwargs):
        self.fmt = fmt
        kwargs['default'] = fmt().GetDefault()
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['fmt'] = self.fmt
        if 'default' in kwargs:
            del kwargs['default']
        return name, path, args, kwargs

    def format_value_json(self, value):
        """Return the json data to display this as a formatted numeric."""
        return {
            'value': value,
            'units': json.loads(self.fmt().JsonUnits()),
            'format': json.loads(self.fmt().JsonFormat())
        }


class UnitsIntegerField(models.IntegerField):
    '''
    The UnitsIntegerField can be used by a django model in place of an
    IntegerField, to store an integer along with the units, min_value, and
    max_value attributes that are needed by the InputNumber in mic-react-lib.

    UnitsIntegerField accepts the following parameters:

        - fmt: An exposed C++ formatted integer type that defines the type.
    '''

    def __init__(self, fmt, *args, **kwargs):
        self.fmt = fmt
        kwargs['default'] = fmt().GetDefault()
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['fmt'] = self.fmt
        if 'default' in kwargs:
            del kwargs['default']
        return name, path, args, kwargs

    def format_value_json(self, value):
        """Return the json data to display this as a formatted numeric."""
        return {
            'value': value,
            'units': json.loads(self.fmt().JsonUnits()),
            'format': json.loads(self.fmt().JsonFormat())
        }


class UnitsFloatFormField(forms.FloatField):
    '''Custom form field for model UnitsFloatField'''

    def __init__(self, fmt, **kwargs):
        super().__init__(**kwargs)
        self.fmt = fmt


class UnitsIntegerFormField(forms.IntegerField):
    '''Custom form field for model UnitsIntegerField'''

    def __init__(self, fmt, **kwargs):
        super().__init__(**kwargs)
        self.fmt = fmt


class RadioWidget(forms.Widget):
    '''Custom widget for displaying radio choice'''
    input_type = "radio_choice"


class SliderWidget(forms.Widget):
    '''Custom widget for displaying a slider for integer entry.'''
    input_type = "slider_choice"


class ButtonGroupWidget(forms.Widget):
    '''Custom widget for displaying button group choice'''
    input_type = "button_group_choice"


class AlphanumericPasswordInput(forms.Widget):
    '''Custom widget for displaying alphanumeric password entry'''
    input_type = "alphanumeric_password"


class NumericPasswordInput(forms.Widget):
    '''Custom widget for displaying numeric password entry'''
    input_type = "numeric_password"


def generate_field_display(obj, field_name):
    '''generate dictionary for displaying one field'''

    field = type(obj)._meta.get_field(field_name)
    value = getattr(obj, field_name)
    if hasattr(field, 'format_value_json'):
        return field.format_value_json(value)
    return value


def generate_model_display(obj, fields):
    '''generate dictionary for displaying fields of a model object'''

    def add_to_dict(ret, obj, field):
        if field.find('.') == -1:
            ret[field] = generate_field_display(obj, field)
            return
        name, rest = field.split('.', 2)
        ret = ret.setdefault(name, {})
        add_to_dict(ret, generate_field_display(obj, name), rest)

    ret = {}
    for field in fields:
        add_to_dict(ret, obj, field)
    return ret


def multi_form_view(views, request):
    """
    This is a view function that combines the views from multiple ReactFormMixin views.

    :param views: a sequence of ReactFormDataMixin that will be shown in the dialog.
    :param request: the view function request.
    """
    if request.method == 'GET':
        form = views[0].get_input_form_props(views[0].id_from_request(request))
        for view in views[1:]:
            next_form = view.get_input_form_props(view.id_from_request(request))
            form["data"].update(next_form["data"])
            form["format"]["fields"].update(next_form["format"]["fields"])
            form["format"]["layout"] += next_form["format"]["layout"]
            form["hidden"].update(next_form["hidden"])
            form["relevantIf"].update(next_form["relevantIf"])
        return JsonResponse(form)

    post = json.loads(request.body.decode('utf-8'))
    view_forms = [view.form_from_request(request) for view in views]
    valid = [form.is_valid() for form in view_forms]
    if all(valid) and post.get('action') == 'save':
        # save in reverse order so the main one is last.
        for view, form in zip(reversed(views), reversed(view_forms)):
            view.save_valid_form(form)
    errors = {}
    for form in view_forms:
        errors.update({
            name: [err['message'] for err in errs
                  ] for name, errs in form.errors.get_json_data().items()
        })
    return JsonResponse({"errors": errors})


def post_response(form):
    '''return json response with form errors'''
    errors = {
        name: [err['message'] for err in errs] for name, errs in form.errors.get_json_data().items()
    }
    return JsonResponse({'errors': errors}, encoder=MicJSONEncoder)


def _get_form_field(field, **kwargs):
    if isinstance(field, UnitsFloatField):
        return UnitsFloatFormField(
            fmt=field.fmt, label=field.verbose_name, required=not field.blank, **kwargs
        )
    if isinstance(field, UnitsIntegerField):
        return UnitsIntegerFormField(
            fmt=field.fmt, label=field.verbose_name, required=not field.blank, **kwargs
        )
    if isinstance(field, models.DateTimeField):
        return ICUDateTimeFormField(label=field.verbose_name, **kwargs)
    return field.formfield(**kwargs)


def _field_data(obj, fields):
    '''serialize model object instance data into json payload'''

    def _get_data(field_name):
        field = type(obj)._meta.get_field(field_name)
        if isinstance(field, models.ForeignKey):
            related = getattr(obj, field_name)
            return related.id if related is not None else ''
        return getattr(obj, field_name)

    return {field: _get_data(field) for field in fields}


def _generate_field_dict(bound_field):
    '''return dictionary representing field parameters for single field'''
    field_dict = {'title': bound_field.label, 'showTitle': bound_field.label != ''}
    field = bound_field.field
    if isinstance(field, forms.CharField):
        field_dict.update([('field', 'text'), ('maxLength', field.max_length)])
        if field.widget.input_type == "alphanumeric_password":
            field_dict.update([('password', True), ('layout', 'qwerty')])
        if field.widget.input_type == "numeric_password":
            field_dict.update([('password', True), ('layout', 'numpad')])
    if isinstance(field, forms.ChoiceField):
        field_dict.update([
            ('field', 'choice'),
            ('choices', [choice[0] for choice in field.choices]),
            ('labels', [choice[1] for choice in field.choices]),
            ('widget', field.widget.input_type),
        ])
    if isinstance(field, UnitsIntegerFormField):
        field_dict.update([
            ('field', 'number'),
            ('min', field.fmt().GetMinimum()),
            ('max', field.fmt().GetMaximum()),
            ('format', json.loads(field.fmt().JsonFormat())),
            ('units', json.loads(field.fmt().JsonUnits())),
            ('widget', field.widget.input_type),
        ])
    elif isinstance(field, forms.IntegerField):
        field_dict.update([
            ('field', 'number'),
            ('min', field.min_value),
            ('max', field.max_value),
            ('format', {
                'style': 'fixed',
                'precision': 0
            }),
            ('widget', field.widget.input_type),
        ])
    if isinstance(field, UnitsFloatFormField):
        field_dict.update([
            ('field', 'number'),
            ('min', field.fmt().GetMinimum()),
            ('max', field.fmt().GetMaximum()),
            ('format', json.loads(field.fmt().JsonFormat())),
            ('units', json.loads(field.fmt().JsonUnits())),
        ])
    elif isinstance(field, forms.FloatField):
        field_dict.update([
            ('field', 'number'),
            ('min', field.min_value),
            ('max', field.max_value),
            ('format', {
                'style': 'fixed',
                'precision': field.decimal_places
            }),
        ])
    if isinstance(field, forms.BooleanField):
        field_dict.update([
            ('field', 'check'),
            ('value', 'false'),
        ])
    if isinstance(field, forms.ModelChoiceField):
        choices = [c[0].value if hasattr(c[0], "value") else str(c[0]) for c in field.choices]
        labels = [str(c[1]) for c in field.choices]
        field_dict.update([
            ('field', 'choice'),
            ('choices', choices),
            ('labels', labels),
        ])
    if isinstance(field, forms.DateTimeField):
        field_dict.update([
            ('field', 'datetime'),
        ])
    return field_dict


def _field_parameters(form):
    '''serialize InputForm field parameters from form into json payload'''

    return {field.name: _generate_field_dict(field) for field in form}


def _get_field_attributes(react_meta, attribute_name):
    """Return object of fieldname/attribute pairs from a ReactMeta object and an attribute name"""
    try:
        attributes = {
            field: attribute() if callable(attribute) else attribute
            for field, attribute in getattr(react_meta, attribute_name).items()
        }
    except AttributeError:
        attributes = {}
    return attributes

def _eval_relevant_if(relevant_if):
    '''
    Some of the rules in relevant_if can be functions instead of constant rules.  For those, the
    function is evaluated and the return value is places in the dictionary.
    '''
    def retval_if_callable(value):
        if callable(value):
            return value()
        return value
    return {k: retval_if_callable(v) for k, v in relevant_if.items()}
