'''
Python functionality that supports the results view for the mic-react-lib/Results/SampleList.tsx
component.
'''
import gettext
import json
import copy
from collections import namedtuple
from django.http import JsonResponse
from django.utils import translation
from django.db import models
from django.core.exceptions import FieldDoesNotExist
from django.db.models import Case, When, F
from mic_cpp import cpp
from .mic_cifs import get_mount_point, print_cmd
from .mic_network import get_usb_drive_mounts, wifi_print_command
from .models import Result, ResultSortFilter, Settings, SORT_DIRECTION_CHOICES
from .log_query import post
from .rest_input import ReactFormDataMixin, UnitsFloatField, MicJSONEncoder

_ = gettext.gettext


# =================================================================================================
class ReactSortFilterMixin:  # pylint: disable=too-few-public-methods
    '''
    Use the ReactSortFilter mixin to create a view that returns a form for results
    sort and filter settings, to be displayed by the SampleList mic-react-lib component.

    To the use the ReactSortFilterMixin, a view must inherit from the imported mixin and
    define the ReactMeta class with these attributes:

    - model: imported model class representing the result

    - sorts: a list of strings representing fields that can be sorted

    - filters: a list of strings representing the fields that can have filters applied

    - field_options: a dictionary of options specific to the Result model

        - formats: a dictionary relating field names to field formats

        - titles: a dictionary relating field names to field titles

        - hidden: a dictionary relating field names to functions that evaluate to
          a boolean representing whether to show the field or not

    example =>

        class ExampleModel(Model):
            field_1 = IntegerField()
            field_2 = BooleanField()

        class ExampleResultSortFilterView(ReactSortFilterMixin, View):
            class ReactResultsMeta:
                model = ExampleModel
                sorts = ['field_1', 'field_2']
                filters = ['field_1', 'field_2']
                field_options = {
                'formats': {
                'field_1': field1_format
                },
                'titles': {
                'field_1': "Field One"
                },
                }

        This view creates a mic-react-lib InstDialog that looks like this:

           Sort by: <ChoiceField [Field One, Field 2] />
           Sort direction: <ChoiceField [Ascending, Descending] />
           Minimum Field One: <Switch /> <IntegerField fmt=field1_format />
           Maximum Field One: <Switch /> <IntegerField fmt=field1_format />
           Field 2: <Switch />

    '''

    class ReactResultsMeta:  # pylint: disable=too-few-public-methods
        '''defines which sorts and filters to show in the form'''

        model = None
        sorts = None
        filters = None
        field_options = None

    def get(self, request):
        '''Forward get requests to ReactSortFilterView'''

        return _react_sort_filter_view(self.ReactResultsMeta).get(request)

    def post(self, request):
        '''Forward post requests to ReactSortFilterView'''

        return _react_sort_filter_view(self.ReactResultsMeta).post(request)


# =================================================================================================
def report_show_export_print_view(request, model, basename):
    """
    The view function that supports the mic-react-lib ReportPage component.

    This will handles the following requests:
    - show: return the json that is the model (with no post arguments in the body)
    - print: send model to printer from the settings.
    - export: will export to the basename in the proper location for the settings.
    """
    # the on screen report data.
    if not request.body:
        return JsonResponse(json.loads(model.AsJson()), encoder=MicJSONEncoder)

    data = json.loads(request.body.decode('utf-8'))

    # export to usb or network.
    if data.get('target') == 'export':
        _export_report(
            model, basename, data['export_format'], data['export_usb'], data['export_network']
        )
        return JsonResponse({})

    # print to printer.
    if 'target' in data and data['target'] == 'print':
        _print_report(model)
        return JsonResponse({})

    assert False
    return JsonResponse({})


def sample_results_view(request, get_analyses, format_instrument, restrictions=None):
    '''Return all results queryset with selected sorts and filters

    - request: the django HttpRequest object
    - get_analyses: function returns queryset of all results that can be shown
    - format_instrument: function accepts result object and returns instrument
      display string
    - restrictions: dictionary of additional filter functions that accept queryset
      of results and return queryset of results

    '''

    def analysis_fields(anl):
        ret = {
            'id': anl.id,
            'start_time': anl.start_time,
            'name': anl.name,
            'operator': anl.operator
        }
        if hasattr(anl, 'mass'):
            ret['mass'] = type(anl)._meta.get_field('mass').format_value_json(anl.mass)
        return ret

    def results_from_anl(anl):
        return anl.results.exclude(result_rank='R').exclude(value=float('inf')).exclude(value=float('-inf'))

    def format_result(result):
        return {
            'analysis_technique': _(result.analysis_technique),
            'result_title': _(result.result_title),
            'value': result.value,
            'std_dev': result.std_dev,
            'units_json': json.loads(result.units_json),
            'format_json': json.loads(result.format_json),
            'rank': result.get_result_rank_display(),
        }

    def filter_and_sort(analyses, request):
        '''Return sorted and filtered queryset of results by
        applying stored filter and sort options incrementally

        - analyses: unfiltered, unsorted queryset of results
        - request: django HttpRequest object

        '''
        if not request.session.session_key:
            request.session.create()
        sort_filter, __ = ResultSortFilter.objects.get_or_create(
            session_key=request.session.session_key,
            context=request.GET.get('context'),
        )
        filtered = analyses.all()
        for field, comparison, value in sort_filter.filters:
            if field in restrictions:
                if field == 'local_only' and not value:
                    pass
                else:
                    filtered = restrictions[field](filtered)
            else:
                field_name, is_result_value = _get_field_lookup(field, analyses.model)
                if is_result_value:  # annotate generic result field with specific result type
                    filtered = filtered.annotate(
                        **{
                            f'{field_name}':
                                Case(
                                    When(results__result_name=field_name, then=F('results__value'))
                                )
                        }
                    )
                    # remove duplicates from join clause
                    filtered = filtered.filter(**{f'{field_name}__isnull': False})
                filtered = filtered.filter(**{f'{field_name}__{comparison}': value})
        order_by = list()
        for field, direction in sort_filter.sort_by:
            field_name, is_result_value = _get_field_lookup(field, analyses.model)
            if is_result_value:
                filtered = filtered.annotate(
                    **{
                        f'{field_name}':
                            Case(When(results__result_name=field_name, then=F('results__value')))
                    }
                )
                filtered = filtered.filter(**{f'{field_name}__isnull': False})
            order_by.append(f'{"-" if direction == "D" else ""}{field_name}')
        if order_by:
            return filtered.order_by(*order_by)
        return filtered

    def format_sample(anl, index):
        sample = {}
        sample['analysis'] = analysis_fields(anl)
        sample['results'] = [format_result(r) for r in results_from_anl(anl)]
        sample['instrument'] = format_instrument(anl)
        sample['index'] = index
        if index == analyses.count():
            sample['last'] = True
        return sample

    start = int(request.GET.get('start', None))
    end = int(request.GET.get('end', None))
    analyses = filter_and_sort(get_analyses(), request)
    block = analyses[start:end]
    response = [format_sample(anl, index + 1 + start) for index, anl in enumerate(block)]
    return JsonResponse(response, safe=False, encoder=MicJSONEncoder)


# =================================================================================================
def _export_report(model, basename, export_format, export_usb, export_network):
    'Export the report model to basename in the proper format to usb and/or network.'
    export_fun_ext = {
        "pdf": (cpp.ExportReportModelPDF, ".pdf"),
        "xls": (cpp.ExportReportModelXLSX, ".xlsx"),
        "txt": (cpp.ExportReportModelTXT, ".txt"),
        "raw": (cpp.ExportReportModelXLSX, "-raw.xlsx"),
    }

    fcn, ext = export_fun_ext[export_format]
    filename = f"{basename}{ext}"
    if export_usb:
        usb_dirs = get_usb_drive_mounts()

        if len(usb_dirs) == 1:
            usb_dir = usb_dirs[0]
            error = fcn(model, usb_dir, filename)
            if error:
                post(error)
            else:
                post(_("Export to USB is complete. It is safe to remove USB drive."))
        elif len(usb_dirs) > 1:
            post(_("Cannot export to USB. Multiple USB drives found."))
        else:
            post(_("Cannot export to USB. No USB drive found."))
    if export_network:
        smb_dir = get_mount_point()
        error = fcn(model, smb_dir, filename)
        if error:
            post(error)
        else:
            post(_("Export to network is complete."))
    if not (export_usb or export_network):
        post(_("Select location for exported sample."))


def _print_report(model):
    """Send the report model to the selected device."""
    settings = Settings.objects.get()
    if settings.printer_type == 'S':
        cmd = print_cmd(
            settings.smbprint_domain,
            settings.smbprint_workgroup,
            settings.smbprint_name,
            settings.smbprint_user,
            settings.smbprint_password,
        )
    elif settings.printer_type == 'W':
        cmd = wifi_print_command()
    else:
        post(_("No printer available."))

    error = cpp.PrintReportModel(model, cmd)
    if error:
        post(error)
    else:
        post(_("Report sent to printer."))


def _react_sort_filter_view(meta):  # pylint: disable=too-many-statements
    """Function to create view from meta options"""

    def form_props():
        '''Create the form from the ReactMeta parameters'''

        def get_field(expr, model=meta.model):
            '''Parse the filter expression to get the field'''
            try:
                field_name, rest = expr.split('.', 1)
            except ValueError:
                field_name = expr
            if model == Result:
                if hasattr(meta, "field_options") and field_name in meta.field_options['formats']:
                    field = UnitsFloatField(
                        fmt=meta.field_options['formats'][field_name],
                        verbose_name=meta.field_options['titles'][field_name]
                    )
                else:
                    field = Result._meta.get_field("value")
            else:
                try:
                    field = model._meta.get_field(field_name)
                except FieldDoesNotExist:
                    field = models.BooleanField(default=False)
            if field.is_relation and rest is not None:
                return get_field(rest, field.related_model)
            field = copy.copy(field)
            setattr(field, 'blank', True)
            if hasattr(meta, "field_options") and field_name in meta.field_options['titles']:
                setattr(field, "verbose_name", meta.field_options['titles'][field_name])
            return field

        sort_by_choices = [(name, get_field(name).formfield().label) for name in meta.sorts]
        sort_fields = {
            'sort_by':
                models.CharField(choices=sort_by_choices,
                                 verbose_name=_("Sort By"),
                                 default=sort_by_choices[0][0]),
            'sort_direction':
                models.CharField(
                    choices=SORT_DIRECTION_CHOICES,
                    verbose_name=_("Sort Direction"),
                    default=SORT_DIRECTION_CHOICES[0][0]
                )
        }
        filter_fields, relevant_if, hidden = dict(), dict(), dict()
        try:
            hidden = meta.field_options["hidden"]
        except (AttributeError, KeyError):
            pass
        for expr in meta.filters:
            field = get_field(expr)
            min_max_fields = (models.FloatField, models.IntegerField, models.DateTimeField)
            if isinstance(field, min_max_fields):
                min_max_field = get_field(expr)
                min_max_field.verbose_name = ""
                filter_fields.update({
                    f"min_{expr}_active": models.BooleanField(
                        default=False,
                        verbose_name=str(_("Minimum {0}")).replace("{0}", str(field.verbose_name))
                    ),
                    f"min_{expr}": min_max_field,
                    f"max_{expr}_active": models.BooleanField(
                        default=False,
                        verbose_name=str(_("Maximum {0}")).replace("{0}", str(field.verbose_name))
                    ),
                    f"max_{expr}": min_max_field,
                })
                relevant_if.update({
                    f"min_{expr}": ["==", f"min_{expr}_active", True],
                    f"max_{expr}": ["==", f"max_{expr}_active", True]
                })
                hidden.update({
                    f"min_{expr}": hidden.get(expr, False),
                    f"max_{expr}": hidden.get(expr, False)
                })
            else:
                if not isinstance(field, models.BooleanField):
                    title = field.verbose_name
                    field.verbose_name = ""
                    filter_fields.update({
                        f"{expr}_active": models.BooleanField(default=False, verbose_name=title),
                        expr: field
                    })
                    relevant_if.update({expr: ["==", f"{expr}_active", True]})
                    hidden.update({f"{expr}_active": hidden.get(expr, False)})
                else:
                    filter_fields.update({expr: field})
        return {
            "fields": dict(**sort_fields, **filter_fields),
            "relevant_if": relevant_if,
            "hidden": hidden
        }

    class ReactSortFilterView(ReactFormDataMixin):
        '''View to display and edit form'''

        def get_initial_values(self, request):  # pylint: disable=no-self-use
            '''Get the current sort and filter settings for the form'''
            if not request.session.session_key:
                request.session.create()
            sort_filter, __ = ResultSortFilter.objects.get_or_create(
                session_key=request.session.session_key, context=request.GET.get('context')
            )
            if sort_filter.sort_by:
                sort_by = sort_filter.sort_by[0]
            else:
                sort_by = (meta.sorts[0], 'D')
            sort_values = {'sort_by': sort_by[0], 'sort_direction': sort_by[1]}
            filter_values = dict()
            for field, comparison, value in sort_filter.filters:
                if comparison == "gte":
                    filter_name = "min_" + field
                elif comparison == "lte":
                    filter_name = "max_" + field
                else:
                    filter_name = field
                filter_values.update({filter_name: value})
                filter_values.update({filter_name + "_active": True})
            return dict(**sort_values, **filter_values)

        def on_save(self, request, data):  # pylint: disable=no-self-use
            '''Update the persistent storage'''

            if not request.session.session_key:
                request.session.create()
            sort_filter, __ = ResultSortFilter.objects.get_or_create(
                session_key=request.session.session_key,
                context=request.GET.get('context'),
            )
            sort_by = data.pop('sort_by')
            history = dict(sort_filter.sort_by)
            history.pop(sort_by, None)
            history = list(tuple(history.items()))
            sort_filter.sort_by = [(sort_by, data.pop('sort_direction'))] + history
            switches = {k[:-7]: v for k, v in data.items() if k.endswith("_active")}
            filters = list()
            for filter_name, value in data.items():
                if filter_name.endswith("_active"):
                    continue
                if filter_name.startswith("min_"):
                    comparison = 'gte'
                    field_name = filter_name[4:]
                    value = _round_min(value, meta, field_name)
                elif filter_name.startswith("max_"):
                    comparison = 'lte'
                    field_name = filter_name[4:]
                    value = _round_max(value, meta, field_name)
                else:
                    comparison = 'regex'
                    field_name = filter_name
                if filter_name not in switches or switches[filter_name]:
                    filters.append((field_name, comparison, value))
            sort_filter.filters = filters
            sort_filter.save()

        def on_delete(self, request):  # pylint: disable=no-self-use
            '''Reset the sort and filters to defaults'''

            if not request.session.session_key:
                request.session.create()
            ResultSortFilter.objects.update_or_create(
                session_key=request.session.session_key,
                context=request.GET.get('context'),
                defaults={
                    'sort_by': [(meta.sorts[0], 'D')],
                    'filters': [],
                }
            )
            return {}

        react_meta = namedtuple('react_meta', 'data relevant_if hidden')
        form, relevant_if, hidden = form_props().values()
        ReactMeta = react_meta(data=form, relevant_if=relevant_if, hidden=hidden)

    return ReactSortFilterView()


def _get_field_lookup(expr, model):
    '''returns (lookup, is_result_value) from the sorts / filters list items
    - expr: the string specified in sorts / filters of ReactResultsMeta
    - model: the django model to lookup the field in
    - lookup: the django field lookup for the sort or filter operation
    - is_result_value: whether the field lookup represents a result value
    '''
    try:
        field_name, rest = expr.split('.', 1)
    except ValueError:
        field_name = expr
    if model is Result:
        try:
            field = Result._meta.get_field(field_name)
        except FieldDoesNotExist:
            return (field_name, True)
    else:
        field = model._meta.get_field(field_name)
        if field.is_relation and rest is not None:
            return _get_field_lookup(rest, field.related_model)
    return (field_name, False)


def _round_min(value, meta, field):
    "Round down to the minimum value that will be displayed as value given for the model field."
    if not hasattr(meta, 'field_options') or 'formats' not in meta.field_options:
        return value
    field = field.split('.')[-1]  # the lookup in field_options['formats'] is just the last part.
    if not field in meta.field_options['formats']:
        return value
    return meta.field_options['formats'][field](value if value else 0.0).RoundedInternalMin()


def _round_max(value, meta, field):
    "Round up to the maximum value that will be displayed as the value given for the model field."
    if not hasattr(meta, 'field_options') or 'formats' not in meta.field_options:
        return value
    field = field.split('.')[-1]  # the lookup in field_options['formats'] is just the last part.
    if not field in meta.field_options['formats']:
        return value
    return meta.field_options['formats'][field](value if value else 0.0).RoundedInternalMax()
