'''
The python functionality that supports language translation.
'''
import gettext
import os
import pytz
from django.conf import settings
from django.utils import timezone

_LANGUAGES_AVAILABLE = ['en']

# The domains that are used for translations (main python/c++ and ui strings.)
_GETTEXT_DOMAIN = ''
_GETTEXT_DOMAIN_UI = ''

# The specific functions that sets the locale in the c++ extension, and the external
# UDP application.
_SET_TRANSLATE_LOCALE = lambda domain, directory, locale: None
_UDP_SET_TRANSLATE_LOCALE = lambda locale: None

# unit notification functions
_UNIT_SEL_AND_SIMPLE = lambda: "{}"
_UNIT_UPDATE_SELECTION = lambda sel, choice: "{}"
_UNIT_UDP_NOTIFY_UPDATES = lambda sels: None
_UNIT_EDIT_UNITS = lambda: "{}"


def gettext_domain():
    """return the domain name for gettext translations."""
    return _GETTEXT_DOMAIN


def gettext_domain_ui():
    """return the domain name for gettext translations."""
    return _GETTEXT_DOMAIN_UI


def init_languages(
    available, domain, domain_ui, set_translate_locale_fcn, udp_set_translate_locale_fcn
):
    """Initialize the application level language parameters."""

    # pylint: disable=global-statement
    global _LANGUAGES_AVAILABLE, _GETTEXT_DOMAIN, _GETTEXT_DOMAIN_UI
    global _SET_TRANSLATE_LOCALE, _UDP_SET_TRANSLATE_LOCALE

    _LANGUAGES_AVAILABLE = available
    _GETTEXT_DOMAIN = domain
    _GETTEXT_DOMAIN_UI = domain_ui
    _SET_TRANSLATE_LOCALE = set_translate_locale_fcn
    _UDP_SET_TRANSLATE_LOCALE = udp_set_translate_locale_fcn

    try:
        home = os.environ['HOME']
        if 'root' in home:
            gettext.bindtextdomain(_GETTEXT_DOMAIN, '/usr/share/locale')
        else:
            gettext.bindtextdomain(_GETTEXT_DOMAIN, home + '/work-micro/i1350/build-debug/po')
    except:
        gettext.bindtextdomain(_GETTEXT_DOMAIN, '/usr/share/locale')
    gettext.textdomain(_GETTEXT_DOMAIN)


def init_units(unit_sel_and_simple_fcn, update_selection_fcn, notify_updates_fcn, edit_units_fcn):
    """Initialize the application level unit selection parameters."""
    # pylint: disable=global-statement
    global _UNIT_SEL_AND_SIMPLE, _UNIT_UPDATE_SELECTION, _UNIT_UDP_NOTIFY_UPDATES, _UNIT_EDIT_UNITS
    _UNIT_SEL_AND_SIMPLE = unit_sel_and_simple_fcn
    _UNIT_UPDATE_SELECTION = update_selection_fcn
    _UNIT_UDP_NOTIFY_UPDATES = notify_updates_fcn
    _UNIT_EDIT_UNITS = edit_units_fcn


def languages_available():
    """return the domain name for gettext translations."""
    return _LANGUAGES_AVAILABLE


def set_language(language, tzname):
    """Set the current language to use"""
    settings.LANGUAGE_CODE = "zh-hans" if language.startswith("zh") else language
    settings.TIME_ZONE = tzname
    if tzname:
        timezone.activate(pytz.timezone(tzname))
    else:
        timezone.deactivate()
        return self.get_response(request)
    os.environ['LANG'] = settings.LANGUAGE_CODE.replace('-', '_',) + ".UTF-8"
    locale = '/usr/share/locale'
    try:
        home = os.environ['HOME']
        if 'root' not in home:
            locale = home + '/work-micro/i1350/build-debug/po'
    except:
        pass

    _SET_TRANSLATE_LOCALE(
        _GETTEXT_DOMAIN, locale, _python_locale_to_c(language), tzname
    )


def udp_set_translate_locale():
    """Update the languages in the udp framework."""
    _UDP_SET_TRANSLATE_LOCALE(_python_locale_to_c(settings.LANGUAGE_CODE), settings.TIME_ZONE)


def unit_sel_and_simple():
    """Return the unit selections and simple units as a json object."""
    return _UNIT_SEL_AND_SIMPLE()


def unit_update_selection(sel, choice):
    """
    Update the linked units based on the sel unit changed to choice.
    Return the json object indicating the linked units..
    """
    return _UNIT_UPDATE_SELECTION(sel, choice)


def unit_udp_notify_updated(sels):
    """
    Let everyone know that there are new unit selections that should take effect.
    """
    _UNIT_UDP_NOTIFY_UPDATES(sels)


def unit_edit_units():
    """
    Return the json object that defined the units to edit.
    """
    return _UNIT_EDIT_UNITS()


def _python_locale_to_c(language):

    # The python side and c++ side seem to use different strings to
    # deal with the language choice.  python use the middle -.  That
    # does not work with c++ getttext which needs a _ separator.
    return language.replace('-', '_')
