'''
The python functionality that supports language translation.
'''
import datetime
import gettext
import json
import icu
import locale
import pytz
from django.apps import apps
from django.db import models
from django import forms
from django.contrib.auth import authenticate
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.http import JsonResponse
from django.shortcuts import redirect
from django.views.generic import View
from django.views.i18n import JavaScriptCatalog
from django.utils import timezone
from django.utils.functional import lazy
from mic_python_lib import log_query
from .icu_datetime import ICUDateTimeFormField, SelectTimeZoneMapWidget
from .instrument import set_instrument_time
from .mic_cifs import check_mount, check_printer, set_and_mount
from .mic_network import (
    bluetooth_connect,
    bluetooth_disconnect,
    bluetooth_known_devices,
    get_network_info,
    get_wifi_printers,
    random_ip,
    setup_mic_ethernet,
    setup_wifi,
    setup_wifi_printer,
)
from .mic_json_encoder import MicJSONEncoder
from .models import (
    KnownInstrument,
    Settings,
    UnitSelection,
    LANGUAGE_CHOICES,
    BASE_LANGUAGE_CHOICES,
)
from .quantity_formatter import update_selection
from .rest_input import (
    AlphanumericPasswordInput,
    NumericPasswordInput,
    ReactFormDataMixin,
    ReactFormMixin,
    post_response,
)
from .settings import (
    gettext_domain_ui,
    languages_available,
    set_language,
    udp_set_translate_locale,
    unit_edit_units,
)

_ = lazy(gettext.gettext, str)

# the list of callback functions that are called when settings change in a way that invalidates
# cached results.
_ON_SETTINGS_CHANGED = []

def javascript_catalog_view(request):
    """Return the JavaScriptCatalog.as_view setup for this application."""
    return JavaScriptCatalog.as_view(domain=gettext_domain_ui())(request)


MICNET_CHOICES = [(v, v) for v in ['10.151', '172.22', '192.168']]
REGION_CHOICES = {
    'en': ['AG', 'AU', 'BW', 'CA', 'DK', 'GB', 'HK', 'IN', 'US', 'ZA'],
    'zh': ['CN', 'HK', 'SG'],
    'fr': ['BE', 'CA', 'CH', 'FR'],
    'de': ['AT', 'CH'],
    'it': ['CH', 'IT'],
    'ja': ['JP'],
    'ko': ['KR'],
    'es': ['AR', 'CL', 'CO', 'CU', 'DO', 'GT', 'PA', 'US', 'VE'],
}  # see tools/icu_datetime_number_formats.py for how this was generated

# these are the known timezones generated by google chrome with the following code:
#    for (const lang in locales) {
#      for (const reg of locales[lang]) {
#              const ar = new (Intl as any).Locale(lang + "-" + reg);
#              console.log(lang, reg, ar.timeZones);
#      }
#    }
# Note that this does not include the US/XXX ones that would otherwise be used.
KNOWN_TIMEZONES = {
    'Africa/Gaborone',
    'Africa/Johannesburg',
    'America/Adak',
    'America/Anchorage',
    'America/Antigua',
    'America/Argentina/La_Rioja',
    'America/Argentina/Rio_Gallegos',
    'America/Argentina/Salta',
    'America/Argentina/San_Juan',
    'America/Argentina/San_Luis',
    'America/Argentina/Tucuman',
    'America/Argentina/Ushuaia',
    'America/Blanc-Sablon',
    'America/Bogota',
    'America/Boise',
    'America/Buenos_Aires',
    'America/Cambridge_Bay',
    'America/Caracas',
    'America/Catamarca',
    'America/Chicago',
    'America/Coral_Harbour',
    'America/Cordoba',
    'America/Creston',
    'America/Dawson',
    'America/Dawson_Creek',
    'America/Denver',
    'America/Detroit',
    'America/Edmonton',
    'America/Fort_Nelson',
    'America/Glace_Bay',
    'America/Goose_Bay',
    'America/Guatemala',
    'America/Halifax',
    'America/Havana',
    'America/Indiana/Knox',
    'America/Indiana/Marengo',
    'America/Indiana/Petersburg',
    'America/Indianapolis',
    'America/Indiana/Tell_City',
    'America/Indiana/Vevay',
    'America/Indiana/Vincennes',
    'America/Indiana/Winamac',
    'America/Inuvik',
    'America/Iqaluit',
    'America/Jujuy',
    'America/Juneau',
    'America/Kentucky/Monticello',
    'America/Los_Angeles',
    'America/Louisville',
    'America/Mendoza',
    'America/Menominee',
    'America/Metlakatla',
    'America/Moncton',
    'America/Montreal',
    'America/New_York',
    'America/Nipigon',
    'America/Nome',
    'America/North_Dakota/Beulah',
    'America/North_Dakota/Center',
    'America/North_Dakota/New_Salem',
    'America/Panama',
    'America/Pangnirtung',
    'America/Phoenix',
    'America/Punta_Arenas',
    'America/Rainy_River',
    'America/Rankin_Inlet',
    'America/Regina',
    'America/Resolute',
    'America/Santiago',
    'America/Santo_Domingo',
    'America/Sitka',
    'America/St_Johns',
    'America/Swift_Current',
    'America/Thunder_Bay',
    'America/Toronto',
    'America/Vancouver',
    'America/Whitehorse',
    'America/Winnipeg',
    'America/Yakutat',
    'America/Yellowknife',
    'Antarctica/Macquarie',
    'Asia/Calcutta',
    'Asia/Hong_Kong',
    'Asia/Seoul',
    'Asia/Shanghai',
    'Asia/Singapore',
    'Asia/Tokyo',
    'Asia/Urumqi',
    'Australia/Adelaide',
    'Australia/Brisbane',
    'Australia/Broken_Hill',
    'Australia/Currie',
    'Australia/Darwin',
    'Australia/Eucla',
    'Australia/Hobart',
    'Australia/Lindeman',
    'Australia/Lord_Howe',
    'Australia/Melbourne',
    'Australia/Perth',
    'Australia/Sydney',
    'Europe/Brussels',
    'Europe/Copenhagen',
    'Europe/London',
    'Europe/Paris',
    'Europe/Rome',
    'Europe/Vienna',
    'Europe/Zurich',
    'Pacific/Easter',
    'Pacific/Honolulu',
}

class SettingsExportForm(forms.ModelForm):
    '''Custom form to specify and validate export choices.'''

    class Meta:
        "The formatting for the form"
        model = Settings
        fields = (
            'export_use',
            'export_user',
            'export_password',
            'export_domain',
            'export_path',
        )
        widgets = {'export_password': AlphanumericPasswordInput}

    def clean(self):
        "The special handling for the form."
        cleaned_data = super().clean()
        if not cleaned_data['export_use']:
            return cleaned_data
        if not all([
            cleaned_data['export_user'], cleaned_data['export_password'],
            cleaned_data['export_domain'], cleaned_data['export_path']
        ]):
            raise ValidationError(_("You must enter all parameters."))
        err = check_mount(
            cleaned_data['export_domain'], cleaned_data['export_path'], cleaned_data['export_user'],
            cleaned_data['export_password']
        )
        if err:
            raise ValidationError(err)
        return cleaned_data


class SettingsSMBPrinterForm(forms.ModelForm):
    '''Custom form to specify and validate export choices.'''

    class Meta:
        "The formatting for the form"
        model = Settings
        fields = (
            'printer_type',
            'smbprint_user',
            'smbprint_password',
            'smbprint_domain',
            'smbprint_workgroup',
            'smbprint_name',
            'wifi_printer',
        )
        widgets = {
            'smbprint_password': AlphanumericPasswordInput,
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['wifi_printer'] = forms.ChoiceField(
            choices=[("", "---")] + get_wifi_printers(),
            required=False,
            label=_("Wi-Fi Printer")
        )

    def clean(self):
        "The special handling for the form."
        cleaned_data = super().clean()
        if cleaned_data['printer_type'] == 'S':
            if not all([
                cleaned_data['smbprint_user'], cleaned_data['smbprint_password'],
                cleaned_data['smbprint_domain'], cleaned_data['smbprint_workgroup'],
                cleaned_data['smbprint_name']
            ]):
                raise ValidationError(_("You must enter all parameters."))
            err = check_printer(
                cleaned_data['smbprint_domain'], cleaned_data['smbprint_workgroup'],
                cleaned_data['smbprint_name'], cleaned_data['smbprint_user'],
                cleaned_data['smbprint_password']
            )
            if err:
                raise ValidationError(err)
        elif cleaned_data['printer_type'] == 'W':
            if not cleaned_data['wifi_printer']:
                cleaned_data['printer_type'] = 'N'

        return cleaned_data


class SettingsSpeakerForm(forms.ModelForm):
    '''View for editing the speaker options.'''

    class Meta:
        "The formatting for the form"
        model = Settings
        fields = ('bluetooth_speaker',)

    def __init__(self, *args, **kwargs):
        "Setup the choices for speakers."
        super().__init__(*args, **kwargs)
        devices = bluetooth_known_devices()
        speaker_choices = [('', '----')] + [(f'{mac};{name}', name) for mac, name in devices]
        self.fields['bluetooth_speaker'] = forms.ChoiceField(
            choices=speaker_choices, required=False, label=_("Speaker"))

    def save(self, commit=True):
        "Do the special handling for connecting to the speaker."
        previous = Settings.objects.get()
        params = super().save(commit=commit)
        # update the system bluetooth settings.
        if commit and previous.bluetooth_speaker != params.bluetooth_speaker:
            if previous.bluetooth_speaker is not None:
                bluetooth_disconnect(previous.bluetooth_speaker_mac)
            if params.bluetooth_speaker is not None:
                bluetooth_connect(params.bluetooth_speaker_mac)

        return params


class SettingsSpeakerView(ReactFormMixin, View):
    '''View editing speaker.'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''handle requests to show the insert editor'''
        form = SettingsSpeakerForm
        get_default_obj = lambda request: Settings.objects.get()


def lang_choices():
    "Return the choices for language selection based on available languages."
    return [
        c for c in BASE_LANGUAGE_CHOICES if LANGUAGE_CHOICES[c[0]] in languages_available()
    ]

def region_relevant_if():
    "Return the relevant if dictionary based on supported regions and languages."
    relevant = dict()
    for lang, _choices in REGION_CHOICES.items():
        relevant[f'{lang}_region'] = ['==', 'base_language', lang]
    return relevant

def add_region_props(props, settings: Settings):
    "Add the region choices to the dialog properties."
    # There is only one shown at a time, based on the value of base_language.
    for lang, choices in REGION_CHOICES.items():
        field_name = f'{lang}_region'
        labels = [_get_locale_display_name(lang + "_" + loc) for loc in choices]
        props['format']['fields'][field_name] = {
            'choices': choices,
            'labels': labels,
            'field': 'choice',
            'title': _('Regional Display'),
            'showTitle': True
        }
        props['format']['layout'].append(field_name)
        if lang == settings.base_language:
            default_region = settings.region
        else:
            default_region = choices[0]
        props['data'][field_name] = default_region

def add_unit_selections(props):
    "Add the unit selections to the dialog properties."
    props['format']['layout'] = list(props['format']['layout'])
    for choice in json.loads(unit_edit_units())['choices']:
        sel = choice['selection']
        props['format']['fields'][sel] = {
            'choices': choice['choices'],
            'labels': choice['labels'],
            'field': 'choice',
            'title': choice['title'],
            'showTitle': True,
        }
        props['format']['layout'].append(sel)
        props['data'][sel] = UnitSelection.objects.get(name=sel).choice

def pull_region_setting(data):
    "Return the updated region setting from the post data."
    return data[f'{data["base_language"]}_region']

def update_language(general: Settings):
    "Update the global language settings "
    set_language(general.locale, general.time_zone)
    udp_set_translate_locale()

def update_unit_selections(data):
    "Update the unit selections from the post data."
    for choice in json.loads(unit_edit_units())['choices']:
        sel = choice['selection']
        if sel not in data:
            continue
        update_selection(sel, int(data[sel]))
    # Notify all of this change.
    for on_change in _ON_SETTINGS_CHANGED:
        on_change()

def settings_general_view(request):
    "View for editing general settings"


    class SettingsGeneralForm(forms.ModelForm):
        "Custom form to specify language and formatting choices"
        base_language = forms.ChoiceField(
            choices=lang_choices(),
            label="Language/" + str(_("Language")) if "Language" != str(_("Language")) else _("Language")
        )

        class Meta:
            model = Settings
            fields = ('base_language',)

    class SettingsGeneralView(ReactFormMixin, View):
        '''View for editing the general settings.'''

        class ReactMeta:  # pylint: disable=too-few-public-methods
            '''Handle get requests to show input form'''
            form = SettingsGeneralForm
            relevant_if = region_relevant_if()

        def get(self, request):
            '''Return the information for editing the local settings.  If there is no
            object id, redirect to the local one.'''
            if 'id' not in request.GET:
                obj, __ = Settings.objects.get_or_create()
                return redirect(f'{request.path}?id={obj.id}')

            settings = Settings.objects.get(id=request.GET.get('id', None))
            props = self.get_input_form_props(request.GET.get('id', None))
            add_region_props(props, settings)
            add_unit_selections(props)

            return JsonResponse(props, encoder=MicJSONEncoder)

        def post(self, request):
            '''Update the instrument information based on what the user edited.'''
            general_form = forms.modelform_factory(
                Settings,
                form=SettingsGeneralForm,
            )
            post = json.loads(request.body.decode('utf-8'))
            data = post.get('data')
            form = general_form(data, instance=Settings.objects.get())
            if form.is_valid() and post.get('action') == 'save':
                general = form.save(commit=False)
                general.region = pull_region_setting(data)
                general.save()
                update_language(general)
                update_unit_selections(data)

            return post_response(form)

    return SettingsGeneralView.as_view()(request)


def settings_time_view(request):
    "View for editing time-related settings"

    timezone_choices = [(tz, tz) for tz in pytz.common_timezones if tz in KNOWN_TIMEZONES]

    class SettingsTimeForm(forms.ModelForm):
        "Form to specify timezone and current time"
        time_zone = forms.ChoiceField(
            label=_("Time Zone"),
            choices=timezone_choices,
            widget=SelectTimeZoneMapWidget
        )
        current_time = ICUDateTimeFormField(
            label=_("Current Time"),
            initial="",
            required=False
        )

        class Meta:
            model = Settings
            fields = (
                'time_zone',
                'set_time_automatically',
                'current_time',
            )

        def clean(self):
            "Validate user input"
            cleaned_data = super().clean()
            if not cleaned_data['set_time_automatically'] and not cleaned_data['current_time']:
                raise ValidationError(_("Must specify current time if not set from network."))

    class SettingsTimeView(ReactFormMixin, View):
        "View for editing time-related settings"

        class ReactMeta:  # pylint: disable=too-few-public-methods
            "Specify form options for time settings"
            form = SettingsTimeForm
            relevant_if = {'current_time': ['==', 'set_time_automatically', False]}

        def get(self, request):  # pylint: disable=no-self-use
            "Return the form information, with added region fields"
            if 'id' not in request.GET:
                obj, __ = Settings.objects.get_or_create()
                return redirect(f'{request.path}?id={obj.id}')

            props = self.get_input_form_props(request.GET.get('id', None))
            return JsonResponse(props)

        def post(self, request):
            "Save the settings and update the system time"
            settings_time_form = forms.modelform_factory(
                Settings,
                form=SettingsTimeForm,
            )
            post = json.loads(request.body.decode('utf-8'))
            data = post.get('data')
            form = settings_time_form(data, instance=Settings.objects.get())
            if form.is_valid() and post.get('action') == 'save':
                general = form.save()
                # the time is sent to the rest server from the ui in UTC, but UTC assumes that the
                # local time is entered in the old timezone. So get back the entered time into a
                # string without the timezone suffix.
                set_time = form.cleaned_data["current_time"]
                set_time = set_time.astimezone(timezone.get_current_timezone())
                set_time = set_time.isoformat()[:-6]
                set_language(general.locale, general.time_zone)
                udp_set_translate_locale()
                # now that the timezone has been set, convert the entered string back into utc for
                # that timezone.
                set_time = timezone.make_aware(
                    datetime.datetime.fromisoformat(set_time), timezone.get_current_timezone())
                set_time = set_time.astimezone(pytz.utc)
                set_instrument_time(
                    form.cleaned_data["set_time_automatically"],
                    set_time,
                )
                # notify all of this change.
                for on_change in _ON_SETTINGS_CHANGED:
                    on_change()
            return post_response(form)

    return SettingsTimeView.as_view()(request)


def _on_save_export(data):  # pylint: disable=no-self-use
    "Update the system to have the (validated) settings."
    if data.export_use:
        set_and_mount(data.export_domain, data.export_path, data.export_user, data.export_password)


class SettingsExportView(ReactFormMixin, View):
    '''Handle get requests to show input form'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define model, fields, and options for displaying form'''
        form = SettingsExportForm
        on_save = _on_save_export
        get_default_obj = lambda request: Settings.objects.get()


class SettingsSMBPrinterView(ReactFormMixin, View):
    '''Handle get requests to show input form'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define model, fields, and options for displaying form'''
        form = SettingsSMBPrinterForm
        get_default_obj = lambda request: Settings.objects.get()

        @staticmethod
        def on_save(settings):
            "On saving the data, setup the wifi printer if selected."
            if settings.printer_type == 'W':
                setup_wifi_printer(settings.wifi_printer)


def get_settings_general_view(request):
    """The view that returns the json object for all the mic-react-lib settings."""
    obj = Settings.objects.get()
    try:
        servicePasswordActive = User.objects.get(username="service").is_active
    except User.DoesNotExist:
        servicePasswordActive = False

    ret = {
        'language': obj.language,
        'locale': obj.locale,
        'displayUsb': obj.display_usb,
        'displayNetwork': obj.export_use,
        'displayPrinter': obj.printer_type != 'N',
        'guidedSetupPage': obj.guided_setup_page,
        'servicePasswordActive': servicePasswordActive,
        'timezone': str(timezone.get_current_timezone()),
    }
    return JsonResponse(ret, encoder=MicJSONEncoder)


def ui_settings_network_view(request):
    """The view for editing network options"""
    info = get_network_info()
    wifi_choices = [(n, n) for n in info.wifi_available]
    if info.wifi_name and info.wifi_name not in info.wifi_available:
        wifi_choices = [(info.wifi_name, info.wifi_name)] + wifi_choices
    wifi_choices = [("", "None")] + wifi_choices # allow removing wifi.

    class SettingsNetworkForm(forms.Form):
        lab_ethernet_ip = forms.CharField(required=False, max_length=15, label=_("LAB NET"))
        mic_ethernet_ip = forms.CharField(required=False, max_length=15, label=_("MIC NET"))
        wifi_name_ip = forms.CharField(required=False, max_length=512, label=_("Wi-Fi"))
        mic_ethernet = forms.ChoiceField(choices=MICNET_CHOICES, label=_("MIC NET Subnet"))
        wifi = forms.ChoiceField(required=False, choices=wifi_choices, label=_("Wi-Fi"))
        password = forms.CharField(required=False, max_length=32, label=_("Password"),
                                   widget=AlphanumericPasswordInput)

    # make the class local to this function so that it has access to the info and wifi_choices we
    # only want to get once per view call.
    class SettingsNetworkView(ReactFormDataMixin, View):
        '''View for editing the network options.'''

        class ReactMeta:  # pylint: disable=too-few-public-methods
            '''Define model, fields, and options for displaying form'''
            form = SettingsNetworkForm
            relevant_if = {
                'lab_ethernet_ip': ['!=', 'lab_ethernet_ip', ''],
                'mic_ethernet_ip': ['!=', 'mic_ethernet_ip', ''],
                'wifi_name_ip': ['!=', 'wifi_name_ip', ''],
                'password': ['&&', ['!=', 'wifi', ''], ['!=', 'wifi', info.wifi_name]],
            }
            disabled = {'lab_ethernet_ip': True, 'mic_ethernet_ip': True, 'wifi_name_ip': True}

        def get_initial_values(self, request):
            if info.mic_ethernet_ip:
                ip1, ip2, _, _ = info.mic_ethernet_ip.split('.')
                micnet = ip1 + '.' + ip2
            else:
                micnet = ''
            return {
                'lab_ethernet_ip': info.lab_ethernet_ip,
                'mic_ethernet_ip': info.mic_ethernet_ip,
                'wifi_name_ip': (
                    f'{info.wifi_name}({info.wifi_ip})' if info.wifi_name and info.wifi_ip else ''
                ),
                'mic_ethernet': micnet,
                'wifi': info.wifi_name,
                'password': ''
            }

        def on_save(self, request, data):
            micether = data["mic_ethernet"]
            wifiname = data["wifi"]
            wifipassw = data["password"]
            if not info.mic_ethernet_ip.startswith(micether):
                setup_mic_ethernet(random_ip(micether))
                # if the subnet changes, the known instruments are no longer useful because they
                # were for the old address.
                KnownInstrument.objects.all().delete()
            if wifiname != info.wifi_name:
                if not setup_wifi(wifiname, wifipassw):
                    log_query.post(_("Unable to connect to Wi-Fi."))

    return SettingsNetworkView.as_view()(request)


def ui_settings_speaker_view(request):
    """The view for editing speaker options"""
    return SettingsSpeakerView.as_view()(request)


def password_protected_view(request, username=None):
    '''Return view for mic-react-lib/Protect component'''

    class PasswordProtectedForm(forms.Form):
        '''Form for validating service password'''
        password = forms.CharField(widget=NumericPasswordInput)

        def clean(self):
            "Raise error if password is incorrect"
            super().clean()
            user = authenticate(username=username, password=self.cleaned_data['password'])
            if not user:
                raise ValidationError(_("Incorrect password."))
            return self.cleaned_data

    class PasswordProtectedView(ReactFormDataMixin, View):
        '''View for verifying password protected pages'''

        class ReactMeta:  # pylint: disable=too-few-public-methods
            '''Define form for view'''
            form = PasswordProtectedForm

        def get_initial_values(self, request):
            return {'password': ''}

        def on_save(self, request, data):
            pass

    return PasswordProtectedView.as_view()(request)


def reset_to_factory_defaults():
    """Reset unit to factory defaults after final assembly"""
    # delete all objects of all models defined in the library.  Don't delete the Settings ones
    # because those are handled as higher level settings in the application.
    for model in apps.get_app_config('mic_python_lib').get_models():
        if model == Settings:
            continue
        model.objects.all().delete()

    # set the user as active to indicate that we now have an finished instrument.
    user = User.objects.get(username="service")
    user.is_active = True
    user.save()

    # resetting to factory defaults sets the guided setup.
    settings = Settings.objects.get()
    settings.guided_setup_page = 0
    settings.save()


def set_guided_setup_page(page):
    """Set the guided setup page to page."""
    # resetting to factory defaults sets the guided setup.
    settings = Settings.objects.get()
    settings.guided_setup_page = page
    settings.save()


def add_settings_changed_callback(on_settings_updated):
    'adds the callback to the list of functions calles whenever the setting change.'
    _ON_SETTINGS_CHANGED.append(on_settings_updated)


def about_video_list_view(request, tag):
    """Returns the information needed by the about video list view.

    This returns the list of video name and url pairs for the specific tag passed in.
    The videos are assumed to be in the /srv/www/static/videos/ directory (where they are installed
    on the instrument).  They are assumed to be served at /static/videos/xxx (where they are
    served by the nginx server on the instrument).  The index.json file is in that same directory.
    The format is an object whose attributes are the possible tags, and values are an array of
    name and file objects that define the list for that tag.  (See the i1350/videos directory for
    an example.)

    :param request: The usual django view request parameters
    :param tag: the name of the tag in the index file in the videos directory.
    """
    language = Settings.objects.get().base_language
    try:
        index = json.load(open("/srv/www/static/videos/index.json"))[tag]
        return JsonResponse({
            "videos": [
                {
                    "name": (data["name"][language] if language in data["name"]
                             else data["en"] if "en" in data else data),
                    "description": (data["description"][language]
                                    if 'description' in data and language in data["description"]
                                    else "" ),
                    "url": f"/static/videos/{data['file']}",
                }
                for data in index
            ]
        })
    except FileNotFoundError:
        return JsonResponse({"videos": []})
    except AttributeError:
        return JsonResponse({"videos": []})

def _get_locale_display_name(locale_str):
    settings = Settings.objects.get()
    time_zone = icu.TimeZone.createTimeZone(settings.time_zone)
    date_fmt = icu.DateFormat.createDateTimeInstance(
        icu.DateFormat.SHORT, icu.DateFormat.SHORT, icu.Locale(locale_str)
    )
    date_fmt.setTimeZone(time_zone)
    set_locale = locale.setlocale(locale.LC_ALL)
    try:
        locale.setlocale(locale.LC_ALL, locale_str)
    except locale.Error:
        pass

    ret = (date_fmt.format(datetime.datetime(year=2020, month=1, day=31, hour=8, minute=0))
           + ": " + locale.str(1.234))
    locale.setlocale(locale.LC_ALL, set_locale)
    return ret
