"""The tools needed for yocto linux systems to upgrade themselves."""
import filecmp
import glob
import os
import re
import shutil
import subprocess
import gettext
import requests
from .mic_network import get_usb_drive_mounts

_ = gettext.gettext

ROOT_HTTPS = 'https://micromeritics.com/software/updates'
VER_DIR_MATCH = r"^v(\d+)\.(\d+).(\d+)(-dev)?$"
VER_DIR_HTTP_MATCH = r"^v(\d+)\.(\d+).(\d+)$"
APP_VER_FILE = 'mic-instrument'
APP_VER_DIR = '/etc'
LOCAL_CACHE = '/var/updates'


def download_latest_if_new(model):
    """Download the latest version of software for the instrument model that is available on the
    Micromeritics website to the LOCAL_CACHE directory."""
    version_dir = _find_latest_https_version_dir_or_none(model)
    if version_dir is None:
        return
    ver = version_dir.split('/')[-1]
    local_dir = LOCAL_CACHE + '/' + model + '/' + ver
    if os.path.isdir(local_dir):
        return

    for prev_ver in glob.glob(LOCAL_CACHE + '/' + model + '/*'):
        shutil.rmtree(prev_ver, ignore_errors=False)

    os.makedirs(local_dir, exist_ok=True)
    for file in _get_files(version_dir):
        name = file.split('/')[-1]
        _get_file(version_dir, name, local_dir)


def new_version_dir_if_available(model):
    """
    Find the newest version directory of the software that is available for the model.  This
    looks in USB stick (any mount point) and in the downloaded cache (LOCAL_CACHE).  If there is a
    version on the installed USB stick, the local cache is not consulted.
    """
    version_dir = None
    for path in get_usb_drive_mounts():
        version_dir = _find_latest_version_dir_or_none(model, path)
        if version_dir is not None:
            break
    if version_dir is None:
        version_dir = _find_latest_version_dir_or_none(model, LOCAL_CACHE)

    if version_dir is not None and not _is_installed(version_dir):
        return version_dir

    return ""


def new_version_if_available(model):
    """Returns the version string for the latest software available for the model. """
    version_dir = new_version_dir_if_available(model)
    if version_dir:
        return version_dir.split('/')[-1]

    return ""


def install_and_restart(version_dir):
    """This will install the software in version_dir.  If that is not possible, it will return a
    string giving the error condition.  If it is possible, it will restart the software (including
    this.)"""
    debs = glob.glob(version_dir + "/*.deb")
    install_deb_cmd = "dpkg -i " + " ".join(debs)
    ver_file = version_dir + '/' + APP_VER_FILE
    installed_file = APP_VER_DIR + '/' + APP_VER_FILE

    try:
        # Use the at command so that this is run in a different process group from the service that
        # is this process.  A simple background task does not work (I think) because it is part of
        # the same process group, and installing the package will kill the server which is this
        # including the update itself.
        with open('/tmp/update.sh', 'w') as update:
            print(install_deb_cmd, file=update)
            print(f'cp {ver_file} {installed_file}', file=update)
        subprocess.Popen(['/usr/bin/at', '-f', '/tmp/update.sh', 'now'])
        return ""
    except OSError as exc:
        return _("Cannot install software file: ") + str(exc)


# =================================================================================================
def _find_latest_version_dir_or_none(model, root):
    """Finds the latest version available for the model on the root system"""
    ver_dict = _get_potential_versions(model, root)
    versions = _sort_versions(ver_dict.keys())
    for ver in versions:
        if _is_version_dir(ver_dict[ver]):
            return ver_dict[ver]
    return None


def _is_installed(version_dir):
    """Check to see if the version_dir is installed true if so, false if not."""
    # dev directories are never installed because the dev could be newer.
    ver = os.path.basename(version_dir)
    match = re.match(VER_DIR_MATCH, ver)
    if match.group(4):
        return False

    ver_file = version_dir + '/' + APP_VER_FILE
    installed_file = APP_VER_DIR + '/' + APP_VER_FILE
    if not os.path.isfile(installed_file):
        return False
    return filecmp.cmp(ver_file, installed_file, shallow=False)


def _find_latest_https_version_dir_or_none(model, root=ROOT_HTTPS):
    """Finds the latest version available for the model on the root system"""
    ver_dict = _get_potential_https_versions(model, root)
    versions = _sort_versions(ver_dict.keys())
    for ver in versions:
        if _is_https_version_dir(ver_dict[ver]):
            return ver_dict[ver]
    return None


def _get_potential_versions(model, root):
    versions = {}
    for version_dir in glob.glob(root + '/' + model + '/v*'):
        if not _is_valid_version_name(version_dir):
            continue
        ver = os.path.basename(version_dir)
        versions[ver] = version_dir
    return versions


def _get_potential_https_versions(model, root):
    versions = {}
    for version_dir in _get_dirs(root + '/' + model):
        ver = version_dir.split('/')[-1]
        if not re.match(VER_DIR_HTTP_MATCH, ver):
            continue
        versions[ver] = version_dir
    return versions


def _is_valid_version_name(version_dir):
    if not os.path.isdir(version_dir):
        return False
    ver = os.path.basename(version_dir)
    if not re.match(VER_DIR_MATCH, ver):
        return False
    return True


def _sort_versions(versions):

    def ver_key(ver):
        match = re.match(VER_DIR_MATCH, ver)
        major, minor, patch = int(match.group(1)), int(match.group(2)), int(match.group(3))
        ver_val = major * 1_000_000 + minor * 1_000 + patch
        if match.group(4):
            ver_val = ver_val + 1_000_000_000
        return ver_val

    return sorted(versions, key=ver_key, reverse=True)


def _is_version_dir(version_dir):
    if len(glob.glob(version_dir + '/*.deb')) == 0:
        return False
    if not os.path.isfile(version_dir + '/mic-instrument'):
        return False
    return True


def _is_https_version_dir(version_dir):
    files = [f.split('/')[-1] for f in _get_files(version_dir)]
    if len([f for f in files if f.endswith('.deb')]) == 0:
        return False
    if 'mic-instrument' not in files:
        return False
    return True


def _get_dirs(root):
    page = requests.get(root)
    if page.status_code != 200:
        return []
    data = page.content.decode('utf-8')
    matches = re.findall(r'<a href="([^/:;="]+)/">.*</a>', data)
    return [root + '/' + m for m in matches]


def _get_files(root):
    data = requests.get(root).content.decode('utf-8')
    matches = re.findall(r'<a href="([^/:;="]+)">.*</a>', data)
    return [root + '/' + m for m in matches]


def _get_file(root, name, dest):
    rslt = requests.get(root + '/' + name, stream=True)
    if rslt.status_code != 200:
        # silently fail if the file does not exist.
        return
    with open(dest + '/' + name, 'wb') as out:
        for chunk in rslt.iter_content(chunk_size=8192):
            out.write(chunk)
