"""The support functionality for working with 1345 results."""
import gettext
import re
from datetime import datetime, timedelta
import requests
import pytz
from django.utils import timezone
from mic_python_lib import log_query
from mic_python_lib.models import LogMessage, Settings
from .models import AnalysisResult1345, Expansion1345, Instrument, LegacyInstrument
from .services import update_results

_ = gettext.gettext

# pylint: disable=too-many-locals,invalid-name,too-many-branches,too-many-statements,too-many-lines
# =================================================================================================
# pylint: disable=line-too-long
AREA = "(Active Area|Surface active|Area attiva|Area activa|Aktive Fläeche)"
CELL_DIAMETER = "(Cell Diameter|Diametre cellule|Diametro della cella|Diametro  de celda|Zelldurchmesserr)"
CELL_VOL = "(Cell Volume|Volume de Cellule|Volumen Celda|Volume Della Cella|Zellvolumen)"
CELL_WALL = "(Cell Wall Density|Densite de la Paroi Cellulaire|Densidad pared celda|Densita parete cell|Zellwanddichte)"
CHORD_LENGTH = "(Chord Length|Longueur arrete|Lunghezza della cord|Longitud de cuerda|Sehnenlänge)"
COMPLETED = "(Completed|Fini|Completado|Terminata|Ende)"
CYCLE = "(Cycle#|Proc#|Ciclo#|Nr)"
EQUIL_RATE = "(Equilibration Rate|Taux d'Equilibre|Tasa de Equilibrio|Tempo di Equilibrio|Gleichgewicht)"
EQUIL_TIME = "(Interval time|Intervallo di tempo|Intervalo de tiempo|Intervalle de temps|Intervall Dauer)"
EXP_VOL = "(Expansion Volume|Volume d'Expansion|Volumen Expansion|Vol.Camera in Espan|Expansionsvolumen)"
GEOMETRIC_VOLUME = "(Geometric Volume|Volume geometrique|Volumen geometrico|Volume geometrico|Geometrisches Vol.)"
METHOD_A = "(Correction using Cell Dimensions|Correction par dimension de la cellule|Correzione mediante dimensioni della cellula|Correccion utilizando dimensiones de celda|Korrektur auf Basis Zellendimensionen)"
METHOD_B = "(Correction by Recutting Sample|Correction par decoupage|Correccion al volver a cortar la muestra|Correzione mediante taglio del provino|Korrektur neuer Schnitt)"
METHOD_C = "(No Correction|Sin correccion|Pas de correction|Nessuna correzione|Keine Korrektur)"
METHOD_D = "(Compressibility Test|Prova di comprimibilita|Prueba de compresibilidad|Test de compressibilite|Kompressibilitatstest)"
METHOD_E = "(Cell Fracture Test|Prueba de fractura de celda|Test de rupture cellulaire|Prova di frattura della cellula|Zell-Fraktur-Test)"
NUM_PURGE = "(Number of Purges|Nombre de Purges|Numero de Purgas|Num. Cicli Condiz|Spülzyklen)"
REPEAT = "(Repeat Analysis After Recutting|Repeter analyse apres recoupage|Repetir analisis despues de recortar|Ripetizione dell'analisi dopo il taglio del provino|Wiederholung nach Probenzerteilung)"
SAMPLE_ID = "(Sample ID|ID echantillon|ID Muestra|Ser.Num.del Camp|Probenname)"
SAMPLE_MASS = "(Sample Mass|Poids echantillon|Peso Muestra|Peso del Campione|Probengewicht)"
SERIAL_NO = "(Serial Number|Numero de Serie|Numero di Matricola|Seriennummer)"
STARTED = "(Started|Commence|Comenzado|Iniziata|Beginn)"
TEMPERATURE = "(Temperature|Temperature|Temperatura|Temperatur)"
TIME = "(Time|Ecoule|Analisi|\\(h:m:s\\))"
VOLUME_REPORT = "(and Volume Report|Rapport .* et de Volume|Reporte .* y Volumen|Report .* e Volume|und Volumenbericht)"
# pylint: enable=line-too-long


class BadReport(Exception):
    "The exception thrown when parsing of the report does not work."


def parse_ascii_report(report, iso_file_date):
    """
    Parse the string that is the report for the report and add the data if it is new to the
    database.

    :param report: The string that is the full html report from the 1345.
    :param iso_file_date: The string that is the file date stamp from the directory of the 1345
           html server.  It is in 'iso' "%Y-%m-%d %H:%M" format.

    Throw BadReport if there is a failure in the parsing.
    """
    inst = _parse_instrument(report)
    if inst is not None:
        _parse_results(report, inst, datetime.strptime(iso_file_date, "%Y-%m-%d %H:%M"))


def check_available(ip_addr):
    "Returns true if the ip_address is (potentially) a valid 1345"
    try:
        return len(_html_dir_files(f'http://{ip_addr}')) > 0
    except requests.exceptions.ConnectionError:
        return False


def scrape_all():
    "Scrape the reports from all known 1345 instruments."
    for legacy in LegacyInstrument.objects.filter(model='1345'):
        if _scrape_reports(legacy.ip_addr):
            legacy.last_communicated = timezone.now()
            legacy.save()


# =================================================================================================
def _convert_vol(value, units):
    "Convert a volume value in units to internal cm3 units and return."
    if units == "cm3":
        return value
    if units == "in3":
        return value * 2.54 * 2.54 * 2.54
    raise BadReport(_("Invalid volume units: ") + units)


def _convert_area(value, units):
    "Convert an area value in units to internal cm3 units and return."
    if units == "cm2":
        return value
    if units == "in2":
        return value * 2.54 * 2.54
    raise BadReport(_("Invalid area units: ") + units)


def _convert_length(value, units):
    "Convert a length value in units to internal cm3 units and return."
    if units == "cm":
        return value
    if units == "in":
        return value * 2.54
    raise BadReport(_("Invalid length units: ") + units)


def _convert_temp(value, units):
    "Convert a temperature value in units to internal C units and return."
    if units in ["°C", "C"]:
        return value
    raise BadReport(_("Invalid temperature units: ") + units)


def _convert_press(value, units):
    "Convert a pressure value in units to internal psi units and return."
    if units == "psig":
        return value
    if units == "kPag":
        return value / 6.89476
    raise BadReport(_("Invalid volume units: ") + units)


def _get_date(name, report, file_date, active_timezone, default=None):
    """
    Return the date of the name header field.  This tries multiple different formats.  If several
    formats could be OK, it returns the closet one to the file write date.
    """
    date_time = _get_string(name, report)
    fmts = [
        "%d/%m/%y  %H:%M:%S",
        "%m/%d/%y  %H:%M:%S",
        "%d/%m/%y  %I:%M:%S %p",
        "%m/%d/%y  %I:%M:%S %p",
    ]
    best_date = None
    best_diff = None
    for fmt in fmts:
        try:
            date = datetime.strptime(date_time, fmt)
        except ValueError:
            continue
        diff = file_date - date
        if diff < timedelta(seconds=0):
            diff = -diff
        if best_date is None or diff < best_diff:
            best_date = date
            best_diff = diff
    if best_date is None:
        if default is None:
            raise BadReport(_("Invalid date time: ") + f'"{date_time}"')
        return default

    #TODO: Set as local time.
    return timezone.make_aware(best_date, timezone=active_timezone)


def _get_value_units(name, report):
    "Return the value and units from the header prompt, or None, None if not found."
    match = re.compile(name + r':?\s+(?P<value>[-+]?\d+(.\d+)?)\s+(?P<units>[°\w]+)').search(report)
    if match:
        try:
            return int(match.group('value')), match.group('units')
        except ValueError:
            return float(match.group('value')), match.group('units')
    return None, None


def _get_value(name, report):
    "Return the value without units from the header prompt, or None if not found."
    match = re.compile(name + r':?\s+(?P<value>[-+]?\d+(.\d+)?)').search(report)
    if match:
        try:
            return int(match.group('value'))
        except ValueError:
            return float(match.group('value'))
    return None


def _get_string(name, report, post=''):
    "Return the string from the header prompt, set post for the end of string."
    match = re.compile(name + r':\s+(?P<value>.*)\s+' + post).search(report)
    if match:
        return match.group('value').strip()
    return None


def _html_dir_files(url):
    "Parse the html directory contents for the files it shows."
    try:
        page = requests.get(url, timeout=5)
    except requests.exceptions.ConnectTimeout:
        return []

    # if the page is not there, ignore it.
    if page.status_code != 200:
        return []
    data = page.content.decode('utf-8')

    matches = re.finditer(
        r'<a href="(?P<link>[^"@]+)">.*</a>\D+(?P<date>\d+-\d+-\d+ \d+:\d+)', data
    )
    ret = [(m.group('link'), m.group('date')) for m in matches]

    return ret


def _parse_instrument(report):
    "Create, save, and return an Instrument object from the report."

    re_model = re.compile(r'(AccuPycII|FoamPycII)\s+(?P<model>\d+)')
    re_serial = re.compile(SERIAL_NO + r':\s+(?P<sn>\d+)')

    # match all the data we want.
    match_model = re_model.search(report)
    match_serial = re_serial.search(report)
    if not match_model or not match_serial:
        return None

    cell, cell_u = _get_value_units(CELL_VOL, report)
    exp, exp_u = _get_value_units(EXP_VOL, report)
    cell = _convert_vol(cell, cell_u)
    exp = _convert_vol(exp, exp_u)

    # create the instrument with the data if it does not already exist.
    if cell > 20.0 or exp > 20.0:
        kind = 'H'
    elif cell > 2.0 or exp > 2.0:
        kind = 'T'
    else:
        kind = '1'

    try:
        return Instrument.objects.get(
            model=match_model.group('model'), serial=match_serial.group('sn')
        )
    except Instrument.DoesNotExist:
        return Instrument.objects.create(
            model=match_model.group('model'),
            serial=match_serial.group('sn'),
            kind=kind,
            material="ot",
            other_expansion_um_per_mk=0.0,
        )


def _parse_accupyc(report, anl):
    "Parse the accupyc specific fields and add them to the anl(AnalysisResult1345) object."
    # pylint: disable=line-too-long
    re_header = re.compile(r'\s*' + CYCLE + r'\s+(?P<vol>\w+).*' + TIME + r'\s+(?P<temp>.+)\s*')
    re_data = re.compile(
        r'\s+(?P<cycle>\d+\*?)\s+(?P<vol>[-+]?\d+\.\d+).*(?P<time>[-+]?\d+:\d+:\d+)\s+(?P<temp>[-+]?\d+\.\d+)\s*'
    )
    # pylint: enable=line-too-long
    vol_u = None
    temp_u = None
    vol = []
    time = []
    temp = []
    excludes = []
    for line in report.splitlines():
        match = re_header.match(line)
        if match:
            vol_u = match.group('vol')
            temp_u = match.group('temp')
            continue

        match = re_data.match(line)
        if match:
            vol.append(_convert_vol(float(match.group('vol')), vol_u))
            hrs, mins, secs = [int(v) for v in match.group('time').split(':')]
            time.append(((hrs * 60 + mins) * 60 + secs) * 1000)
            temp.append(_convert_temp(float(match.group('temp')), temp_u))
            excludes.append(match.group('cycle').endswith('*'))

    anl.expansion = Expansion1345()
    anl.expansion.num = len(vol)
    anl.expansion.volumes = vol
    anl.expansion.times = time
    anl.expansion.temperatures = temp
    anl.expansion.excludes = excludes


def _parse_foam_params(report, anl):
    "Parse the common foam pyc parameters from the header."
    vol, vol_u = _get_value_units(GEOMETRIC_VOLUME, report)
    resin = _get_value(CELL_WALL, report)
    if vol is None or resin is None:
        raise BadReport(_("No cell wall density or geometric volume for FoamPyc report."))
    anl.geometric_volume = _convert_vol(vol, vol_u)
    anl.resin_density = resin


def _parse_foam_method_a(report, anl):
    "Parse the accupyc specific fields and add them to the anl(AnalysisResult1345) object."
    area, area_u = _get_value_units(AREA, report)
    length, length_u = _get_value_units(CHORD_LENGTH, report)
    diameter, diameter_u = _get_value_units(CELL_DIAMETER, report)
    if area is None:
        raise BadReport(_("No area for FoamPyc correction using cell dimensions report."))
    anl.active_area = _convert_area(area, area_u)
    if length is not None:
        anl.measure_by = 'L'
        anl.chord_length = _convert_length(length, length_u)
    elif diameter is not None:
        anl.measure_by = 'D'
        anl.cell_diameter = _convert_length(diameter, diameter_u)
    else:
        # pylint: disable=line-too-long
        raise BadReport(_("No chord length nor cell diameter for FoamPyc correction using cell dimensions report."))

    # pylint: disable=line-too-long
    re_header = re.compile(
        r'\s*' + CYCLE + r'\s+(?P<p1>\w+)\s+(?P<p2>\w+)\s+(?P<vol>\w+).*' + TIME + r'\s*'
    )
    re_data = re.compile(
        r'\s*\d+\*?\s+(?P<p1>[-+]?\d+\.\d+)\s+(?P<p2>[-+]?\d+\.\d+)\s+(?P<vol>[-+]?\d+\.\d+).*(?P<time>\d+:\d+:\d+)\s*'
    )
    # pylint: enable=line-too-long
    vol_u = None
    p1_u = None
    p2_u = None
    vol = []
    p1 = []
    p2 = []
    time = []
    excludes = []
    for line in report.splitlines():
        match = re_header.match(line)
        if match:
            p1_u = match.group('p1')
            p2_u = match.group('p2')
            vol_u = match.group('vol')
            continue

        match = re_data.match(line)
        if match:
            vol.append(_convert_vol(float(match.group('vol')), vol_u))
            p1.append(_convert_press(float(match.group('p1')), p1_u))
            p2.append(_convert_press(float(match.group('p2')), p2_u))
            hrs, mins, secs = [int(v) for v in match.group('time').split(':')]
            time.append(((hrs * 60 + mins) * 60 + secs) * 1000)
            excludes.append(False)

    anl.expansion = Expansion1345()
    anl.expansion.num = len(vol)
    anl.expansion.volumes = vol
    anl.expansion.p1s = p1
    anl.expansion.p2s = p2
    anl.expansion.times = time
    anl.expansion.excludes = excludes


def _parse_foam_method_b(report, anl):
    "Parse the foam pyc by recutting and add them to the anl(AnalysisResult1345) object."

    # pylint: disable=line-too-long
    re_header = re.compile(
        r'\s*' + CYCLE + r'\s+(?P<p1>\w+)\s+(?P<p2>\w+)\s+(?P<vol>\w+)\s+\w+\s*' + TIME + r'\s*'
    )
    re_data = re.compile(
        r'\s+\d+\s+(?P<p1>[-+]?\d+\.\d+)\s+(?P<p2>[-+]?\d+\.\d+)\s+(?P<vol>[-+]?\d+\.\d+)\s+[-+]?\d+\.\d+\s+(?P<time>\d+:\d+:\d+)\s*'
    )
    # pylint: enable=line-too-long
    re_repeat = re.compile(r'\s+' + REPEAT + r'\s*')
    p1_u = None
    p2_u = None
    vol_u = None
    vol = []
    time = []
    p1 = []
    p2 = []
    excludes = []
    for line in report.splitlines():
        match = re_header.match(line)
        if match:
            p1_u = match.group('p1')
            p2_u = match.group('p2')
            vol_u = match.group('vol')
            continue

        match = re_data.match(line)
        if match:
            vol.append(_convert_vol(float(match.group('vol')), vol_u))
            hrs, mins, secs = [int(v) for v in match.group('time').split(':')]
            time.append(((hrs * 60 + mins) * 60 + secs) * 1000)
            p1.append(_convert_press(float(match.group('p1')), p1_u))
            p2.append(_convert_press(float(match.group('p2')), p2_u))
            excludes.append(False)
        # if we are staring the repeat, then save the initial information.
        if re_repeat.match(line):
            anl.expansion = Expansion1345()
            anl.expansion.num = len(vol)
            anl.expansion.volumes = vol
            anl.expansion.times = time
            anl.expansion.p1s = p1
            anl.expansion.p2s = p2
            anl.expansion.excludes = excludes
            vol = []
            time = []
            p1 = []
            p2 = []
            excludes = []
        anl.expansion_repeat = Expansion1345()
        anl.expansion_repeat.num = len(vol)
        anl.expansion_repeat.volumes = vol
        anl.expansion_repeat.times = time
        anl.expansion_repeat.p1s = p1
        anl.expansion_repeat.p2s = p2
        anl.expansion_repeat.excludes = excludes


def _parse_foam_method_c(report, anl):
    "Parse the accupyc specific fields and add them to the anl(AnalysisResult1345) object."
    # pylint: disable=line-too-long
    re_header = re.compile(
        r'\s*' + CYCLE + r'\s+(?P<p1>\w+)\s+(?P<p2>\w+)\s+(?P<vol>\w+).*' + TIME + r'\s*'
    )
    re_data = re.compile(
        r'\s*\d+\s+(?P<p1>[-+]?\d+\.\d+)\s+(?P<p2>[-+]?\d+\.\d+)\s+(?P<vol>[-+]?\d+\.\d+).*(?P<time>\d+:\d+:\d+)\s*'
    )
    # pylint: enable=line-too-long
    vol_u = None
    p1_u = None
    p2_u = None
    vol = []
    p1 = []
    p2 = []
    time = []
    excludes = []
    for line in report.splitlines():
        match = re_header.match(line)
        if match:
            p1_u = match.group('p1')
            p2_u = match.group('p2')
            vol_u = match.group('vol')
            continue

        match = re_data.match(line)
        if match:
            vol.append(_convert_vol(float(match.group('vol')), vol_u))
            p1.append(_convert_press(float(match.group('p1')), p1_u))
            p2.append(_convert_press(float(match.group('p2')), p2_u))
            hrs, mins, secs = [int(v) for v in match.group('time').split(':')]
            time.append(((hrs * 60 + mins) * 60 + secs) * 1000)
            excludes.append(False)

    anl.expansion = Expansion1345()
    anl.expansion.num = len(vol)
    anl.expansion.volumes = vol
    anl.expansion.p1s = p1
    anl.expansion.p2s = p2
    anl.expansion.times = time
    anl.expansion.excludes = excludes


def _parse_foam_method_d(report, anl):
    "Parse the accupyc specific fields and add them to the anl(AnalysisResult1345) object."
    # pylint: disable=line-too-long
    re_header = re.compile(r'\s*' + CYCLE + r'\s+(?P<p1>\w+)\s+(?P<p2>\w+)\s+(?P<vol>\w+)\s*')
    re_data = re.compile(
        r'\s*\d+\s+(?P<p1>[-+]?\d+\.\d+)\s+(?P<p2>[-+]?\d+\.\d+)\s+(?P<vol>[-+]?\d+\.\d+)\s*'
    )
    # pylint: enable=line-too-long
    vol_u = None
    p1_u = None
    p2_u = None
    vol = []
    p1 = []
    p2 = []
    excludes = []
    for line in report.splitlines():
        match = re_header.match(line)
        if match:
            p1_u = match.group('p1')
            p2_u = match.group('p2')
            vol_u = match.group('vol')
            continue

        match = re_data.match(line)
        if match:
            vol.append(_convert_vol(float(match.group('vol')), vol_u))
            p1.append(_convert_press(float(match.group('p1')), p1_u))
            p2.append(_convert_press(float(match.group('p2')), p2_u))
            excludes.append(False)

    anl.expansion = Expansion1345()
    anl.expansion.num = len(vol)
    anl.expansion.volumes = vol
    anl.expansion.p1s = p1
    anl.expansion.p2s = p2
    anl.expansion.excludes = excludes


def _parse_foam_method_e(report, anl):
    "Parse the accupyc specific fields and add them to the anl(AnalysisResult1345) object."
    re_header = re.compile(r'\s*' + CYCLE + r'\s+(?P<p1>\w+)\s+(?P<p2>\w+)\s+(?P<vol>\w+)\s*')
    re_data = re.compile(
        r'.*(?P<p1>[-+]?\d+\.\d+)\s+(?P<p2>[-+]?\d+\.\d+)\s+(?P<vol>[-+]?\d+\.\d+)\s*'
    )
    vol_u = None
    p1_u = None
    p2_u = None
    vol = []
    p1 = []
    p2 = []
    excludes = []
    for line in report.splitlines():
        match = re_header.match(line)
        if match:
            p1_u = match.group('p1')
            p2_u = match.group('p2')
            vol_u = match.group('vol')
            continue

        match = re_data.match(line)
        if match:
            vol.append(_convert_vol(float(match.group('vol')), vol_u))
            p1.append(_convert_press(float(match.group('p1')), p1_u))
            p2.append(_convert_press(float(match.group('p2')), p2_u))
            excludes.append(False)

    anl.expansion = Expansion1345()
    anl.expansion.num = len(vol)
    anl.expansion.volumes = vol
    anl.expansion.p1s = p1
    anl.expansion.p2s = p2
    anl.expansion.excludes = excludes


def _parse_results(report, inst, file_date):  # pylint: disable=too-many-locals
    "Create and save an AnalysisResult1345 object with this report data linked to inst."
    # make sure we don't have this report already.
    active_timezone = pytz.timezone(Settings.objects.get().time_zone)
    started = _get_date(STARTED, report, file_date, active_timezone)
    completed = _get_date(COMPLETED, report, file_date, active_timezone, started)
    try:
        # if we already have this, don't add again.
        AnalysisResult1345.objects.get(instrument=inst, start_time=started)
        return
    except AnalysisResult1345.DoesNotExist:
        pass

    sample_id = _get_string(SAMPLE_ID, report, STARTED)
    mass, __ = _get_value_units(SAMPLE_MASS, report)
    equil_rate, equil_rate_u = _get_value_units(EQUIL_RATE, report)
    equil_time = _get_value(EQUIL_TIME, report)
    purges = _get_value(NUM_PURGE, report)
    temperature, temperature_u = _get_value_units(TEMPERATURE, report)
    if sample_id is None or mass is None or purges is None:
        raise BadReport(_("Missing header parameter."))

    anl = AnalysisResult1345(
        state='C',
        start_time=started,
        end_time=completed,
        instrument=inst,
        mass=mass,
        name=sample_id,
    )
    if temperature is not None:
        anl.temperature = _convert_temp(temperature, temperature_u)
    anl.num_purges = purges
    if equil_rate is not None:
        anl.endby = 'E'
        anl.equilibration_rate = _convert_press(equil_rate, equil_rate_u.split('/')[0])
    elif equil_time is not None:
        anl.endby = 'F'
        anl.interval_time = equil_time
    else:
        raise BadReport(_("Missing equilibration option in header."))

    if re.compile(VOLUME_REPORT).search(report):
        anl.technique = 'P'
        _parse_accupyc(report, anl)
    elif re.compile(METHOD_A).search(report):
        anl.technique = 'A'
        _parse_foam_params(report, anl)
        _parse_foam_method_a(report, anl)
    elif re.compile(METHOD_B).search(report):
        anl.technique = 'B'
        _parse_foam_params(report, anl)
        _parse_foam_method_b(report, anl)
    elif re.compile(METHOD_C).search(report):
        anl.technique = 'C'
        _parse_foam_params(report, anl)
        _parse_foam_method_c(report, anl)
    elif re.compile(METHOD_D).search(report):
        anl.technique = 'D'
        _parse_foam_params(report, anl)
        _parse_foam_method_d(report, anl)
    elif re.compile(METHOD_E).search(report):
        anl.technique = 'E'
        _parse_foam_params(report, anl)
        _parse_foam_method_e(report, anl)
    else:
        raise BadReport(_("Unknown report format."))
    anl.expansion.save()
    if anl.expansion_repeat:
        anl.expansion_repeat.save()
    anl.save()

    # Add the results to the table.
    update_results(anl)


def _scrape_reports(ip_addr):
    "Scrape all the reports of this site.  Return true if we got one."
    did_update = False
    for serial_index, __ in _html_dir_files(f'http://{ip_addr}'):
        serial = serial_index.split('/')[0]
        for name, date in _html_dir_files(f'http://{ip_addr}/{serial_index}'):
            # if the page is not there, ignore it.
            url = f'http://{ip_addr}/{serial}/{name}'
            page = requests.get(url)
            if page.status_code == 200:
                try:
                    parse_ascii_report(page.content.decode('utf-8', errors='ignore'), date)
                    did_update = True
                except BadReport as error:
                    # on error parsing, post the error, but only once in the database.
                    err_str = _("Cannot parse report on %s: %s") % (url, str(error))
                    if LogMessage.objects.filter(message=err_str).count() == 0:
                        log_query.post(err_str)

    return did_update
