'''
Django model for data storage for the accupyc 1350
'''
import datetime
import gettext
import platform
import struct
import subprocess

from django.db import IntegrityError, models
from django.db.utils import OperationalError
from django.utils import timezone
from django.utils.functional import lazy

import mic_python_lib
import mic_python_lib.models as mpl_models
from mic_python_lib.mic_net_events import notify_new_data
from mic_python_lib.rest_input import UnitsFloatField, UnitsIntegerField
from mic_cpp import pycnometry

import i1350

_ = lazy(gettext.gettext, str)
# pylint: disable=too-many-statements,too-many-lines
# =================================================================================================

CALIBRATION_CHOICES = (
    ('C', _("Factory")),
    ('R', _("Reset")),
    ('H', _("Check")),
    ('UA', _("Applied")),
    ('UN', _("Not Applied")),
    ('UR', _("Rejected")),
)

OPERATION_CHOICES = ( # Not user visible.
    ('P', "HW Pressure Calibration"),
    ('B', "HW Block Temperature Calibration"),
    ('I', "HW Inlet Temperature Calibration"),
    ('U', "User Operation"),
    ('S', "Service Operation"),
    ('Y', "Python Script"),
)

EQUILIBRATION_CHOICES = ( # Not user visible.
    ('1', "P1"),
    ('2', "P2"),
)

INSTRUMENT_CHOICES = ( # Not user visible.
    ('1', "1 cm\u00B3"),
    ('T', "10 cm\u00B3"),
    ('H', "100 cm\u00B3"),
)

TEMPCHANNEL_CHOICES = ( # Not user visible.
    ('B', "Block"),
    ('I', "Inlet"),
)

MATERIAL_CHOICES = (
    ("Al", _("Aluminum")),
    ("SS", _("Stainless Steel")),
    ("TC", _("Tungsten Carbide")),
    ("ot", _("Other")),
)

MATERIAL_CTE = {
    "Al": 23.4,
    "SS": 10.1,
    "TC": 5.9
}

INSERT_CHOICES = (
    ('I', _("Insert")),
    ('R', _("Reference Ball")),
    ('C', _("Cup")),
)

INSERT_NO_REF_CHOICES = [ins for ins in INSERT_CHOICES if ins[0] != 'R']

TECHNIQUE_CHOICES = (
    ('P', _("Gas Pycnometry")),
    ('A', _("Correction using Cell Dimensions")),
    ('B', _("Correction by Recutting Sample")),
    ('C', _("No Correction")),
    ('D', _("Compressibility Test")),
    ('E', _("Cell Fracture Test")),
)

COMPUTE_VOLUME_CHOICES = (
    ('C', _("Computed")),
    ('E', _("Entered")),
)

SHAPE_CHOICES = (
    ('U', _("Cube")),
    ('Y', _("Cylinder")),
    ('R', _("Rectangular")),
)
MEASURE_BY_CHOICES = (
    ('L', _("Chord Length")),
    ('D', _("Cell Diameter")),
)

STATE_CHOICES = ( # Not user visible.
    ('P', "In Process"),
    ('C', "Completed"),
    ('X', "Canceled"),
    ('E', "Error"),
)

FILL_CHOICES = (
    ('S', _("Sample Chamber")),
    ('R', _("Reference Chamber"))
)

END_CHOICES = (
    ('E', _("Rate")),
    ('F', _("Time"))
)


class Settings(mpl_models.Settings):
    '''Adds Application specific information to the generic settings.

    :param display_id: Should the ID field be displayed on the start analysis screen
    :param display_method: Should the method field be displayed on the start analysis screen.
    :param display_wizard: Should the start analysis display the wizard?
    :param display_operator: Should the start analysis screen display the operator?
    :param do_idle_purges: If the user wants to do idle purges.
    :param days_idle_purges: How many days of non activity are required before doing idle purges
    :param flow_required_temperature: The temp below which we want to check for flow.

    '''
    display_id = models.BooleanField(default=True, verbose_name=_("Description"))
    display_method = models.BooleanField(default=True, verbose_name=_("Method"))
    display_wizard = models.BooleanField(default=False, verbose_name=_("Analysis Wizard"))
    display_operator = models.BooleanField(default=False, verbose_name=_("Operator"))

    do_idle_purges = models.BooleanField(
        default=True, verbose_name=_("Enable Background Purges"))
    days_idle_purges = UnitsIntegerField(
        fmt=i1350.UDaysIdlePurges,  # pylint: disable=no-member
        verbose_name=_("Days Since Last Activity"))

    flow_required_temperature = UnitsFloatField(
        fmt=i1350.DPycPurgeCheckMinTemp,  # pylint: disable=no-member
        verbose_name=_("Check for No Flow Below"),
        default=15.0,
    )


class ExpansionManager(models.Manager):   # pylint: disable=too-few-public-methods
    '''Management object for the Expansion model.'''

    def copy(self, obj):
        "Make a copy of the passed in object and return. If obj is None, then return None."
        if obj is None:
            return None

        return self.create(
            num=obj.num,
            times_binary=obj.times_binary,
            p0s_binary=obj.p0s_binary,
            p1s_binary=obj.p1s_binary,
            p2s_binary=obj.p2s_binary,
            excludes_binary=obj.excludes_binary,
            temp_anl_block_binary=obj.temp_anl_block_binary,
            temp_valve_block_binary=obj.temp_valve_block_binary,
        )


class Expansion(models.Model):  # pylint: disable=too-many-instance-attributes
    '''The expansion data for multiple cycles:


    :param num: The number of points of equilibration.
    :param times: Binary array of uint_32 each giving the time since
        start of analysis in milliseconds.
    :param p0s: the pre-fill pressures, in psia, as float_32, for each cycle.
    :param p1s: the fill pressures, in psia, as float_32, for each cycle.
    :param p2s: the equilibration pressures, in psia, as float_32, for each cycle.
    :param excludes: the exclude flags for each cycle (as bool_8)
    :param temp_anl_block: the temperatures, in C, as float_32, for each cycle.
    :param temp_valve_block: the temperatures, in C, as float_32, for each cycle.
    '''
    num = models.IntegerField()
    times_binary = models.BinaryField()
    p0s_binary = models.BinaryField()
    p1s_binary = models.BinaryField()
    p2s_binary = models.BinaryField()
    excludes_binary = models.BinaryField()
    temp_anl_block_binary = models.BinaryField()
    temp_valve_block_binary = models.BinaryField()

    objects = ExpansionManager()

    @property
    def times(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num * "I", self.times_binary)

    @times.setter
    def times(self, value):  # pylint: disable=missing-function-docstring
        self.times_binary = struct.pack(self.num * "I", *value)

    @property
    def p0s(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num * "f", self.p0s_binary)

    @p0s.setter
    def p0s(self, value):  # pylint: disable=missing-function-docstring
        self.p0s_binary = struct.pack(self.num * "f", *value)

    @property
    def p1s(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num * "f", self.p1s_binary)

    @p1s.setter
    def p1s(self, value):  # pylint: disable=missing-function-docstring
        self.p1s_binary = struct.pack(self.num * "f", *value)

    @property
    def p2s(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num * "f", self.p2s_binary)

    @p2s.setter
    def p2s(self, value):  # pylint: disable=missing-function-docstring
        self.p2s_binary = struct.pack(self.num * "f", *value)

    @property
    def excludes(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num * "b", self.excludes_binary)

    @excludes.setter
    def excludes(self, value):  # pylint: disable=missing-function-docstring
        self.excludes_binary = struct.pack(self.num * "b", *value)

    @property
    def temp_anl_block(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num * "f", self.temp_anl_block_binary)

    @temp_anl_block.setter
    def temp_anl_block(self, value):  # pylint: disable=missing-function-docstring
        self.temp_anl_block_binary = struct.pack(self.num * "f", *value)

    @property
    def temp_valve_block(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num * "f", self.temp_valve_block_binary)

    @temp_valve_block.setter
    def temp_valve_block(self, value):  # pylint: disable=missing-function-docstring
        self.temp_valve_block_binary = struct.pack(self.num * "f", *value)


class Expansion1345(models.Model):  # pylint: disable=too-many-instance-attributes
    '''The expansion data for multiple cycles:


    :param num: The number of points of equilibration.

    :param times: Binary array of uint_32 each giving the time since
        start of analysis in milliseconds.

    :param volumes: the volume in cm3 as float_32 for each cycle.

    :param p1s: the fill pressures, in psia, as float_32, for each cycle.

    :param p2s: the equilibration pressures, in psia, as float_32, for each cycle.

    :param excludes: the exclude flags for each cycle (as bool_8)

    :param temp_anl_block: the temperatures, in C, as float_32, for each cycle.
    '''
    num = models.IntegerField()
    times_binary = models.BinaryField()
    volumes_binary = models.BinaryField()
    p1s_binary = models.BinaryField(null=True)
    p2s_binary = models.BinaryField(null=True)
    excludes_binary = models.BinaryField()
    temp_anl_block_binary = models.BinaryField(null=True)

    @property
    def times(self):  # pylint: disable=missing-function-docstring
        if self.times_binary:
            return struct.unpack(self.num * "I", self.times_binary)
        return []

    @times.setter
    def times(self, value):  # pylint: disable=missing-function-docstring
        self.times_binary = struct.pack(len(value) * "I", *value)

    @property
    def volumes(self):  # pylint: disable=missing-function-docstring
        if self.volumes_binary:
            return struct.unpack(self.num * "f", self.volumes_binary)
        return []

    @volumes.setter
    def volumes(self, value):  # pylint: disable=missing-function-docstring
        self.volumes_binary = struct.pack(len(value) * "f", *value)

    @property
    def p1s(self):  # pylint: disable=missing-function-docstring
        if self.p1s_binary:
            return struct.unpack(self.num * "f", self.p1s_binary)
        return []

    @p1s.setter
    def p1s(self, value):  # pylint: disable=missing-function-docstring
        self.p1s_binary = struct.pack(len(value) * "f", *value) if value else None

    @property
    def p2s(self):  # pylint: disable=missing-function-docstring
        if self.p2s_binary:
            return struct.unpack(self.num * "f", self.p2s_binary)
        return []

    @p2s.setter
    def p2s(self, value):  # pylint: disable=missing-function-docstring
        self.p2s_binary = struct.pack(len(value) * "f", *value) if value else None

    @property
    def excludes(self):  # pylint: disable=missing-function-docstring
        if self.excludes_binary:
            return struct.unpack(self.num * "b", self.excludes_binary)
        return []

    @excludes.setter
    def excludes(self, value):  # pylint: disable=missing-function-docstring
        self.excludes_binary = struct.pack(len(value) * "b", *value)

    @property
    def temp_anl_block(self):  # pylint: disable=missing-function-docstring
        if self.temp_anl_block_binary:
            return struct.unpack(self.num * "f", self.temp_anl_block_binary)
        return []

    @temp_anl_block.setter
    def temp_anl_block(self, value):  # pylint: disable=missing-function-docstring
        self.temp_anl_block_binary = struct.pack(len(value) * "f", *value) if value else None


class InsertManager(models.Manager):
    """Accessors for specific types of inserts."""

    def methods(self, **kwargs):
        """Return all is_method=True objects"""
        return self.filter(is_method=True, **kwargs)

    def cups(self, **kwargs):
        '''Return all cups that are methods parameters'''
        return self.methods(kind='C', **kwargs)

    def reference_balls(self, **kwargs):
        '''Return all reference balls that are methods parameters'''
        return self.methods(kind='R', **kwargs)

    def inserts(self, **kwargs):
        '''Return all inserts that are methods parameters'''
        return self.methods(kind='I', **kwargs)

    def create_insert(self, kind):  # pylint: disable=no-self-use
        "Create a new insert of the given kind"
        name_from_kind = dict(INSERT_CHOICES)
        if kind == 'R':
            try:
                if Instrument.objects.local().kind == 'T':
                    material = 'TC'
                else:
                    material = 'SS'
            except Instrument.DoesNotExist:
                material = 'Al'
        else:
            material = 'Al'

        # Keep trying new names till we have a unique one.
        name = name_from_kind[kind]
        count = 0
        while True:
            try:
                insert = Insert(
                    kind=kind, name=name, material=material, other_expansion_um_per_mk=0.0)
                insert.save()
                return insert
            except IntegrityError:
                count = count + 1
                name = name_from_kind[kind] + ' ' + str(count)


def copy_model_obj(model, obj, save=True):
    """Make a non method copy of the object obj can either be a model object, or an id integer"""
    obj = model.objects.get(id=obj.id if hasattr(obj, "id") else obj)
    obj.id = None
    obj.is_method = False
    obj.last_updated = None
    if save:
        obj.save()
    return obj


class Insert(models.Model):
    '''The parameters of an insert that can be used.

    :param name: the name of the insert.
    :param is_method: True if this is an insert that the user can edit.  False if this is a copy of
                      an insert that is referenced by collected data.
    :param last_updated: The last time the insert was updated.
    :param is_measured: True if the insert was measured, and so cannot be edited.
    :param is_shared: True if the user wants to share this insert.
    :param volume: The volume of the insert.
    :param temperature: The temperature at which the volume was measured.
    :param material: Which of the known materials it is.
    :param other_expansion_um_per_mk: The thermal expansion coefficient of the insert as
        entered by the user in um/m·K, for the other material selection.
    :param kind: Which kind of insert it is.
    '''
    name = models.CharField(max_length=16, verbose_name=_("Name"))
    is_method = models.BooleanField(default=True)
    last_updated = models.DateTimeField(null=True)
    is_measured = models.BooleanField(default=False)
    is_shared = models.BooleanField(default=False, verbose_name=_("Share on MIC NET"))
    volume = UnitsFloatField(
        fmt=pycnometry.DPycInsertVolume,  # pylint: disable=no-member
        verbose_name=_("Volume"),
    )
    temperature = UnitsFloatField(
        fmt=pycnometry.DPycTemperature1350,  # pylint: disable=no-member
        verbose_name=_("Temperature")
    )
    material = models.CharField(
        max_length=2,
        choices=MATERIAL_CHOICES,
        default="Al",
    )
    other_expansion_um_per_mk = UnitsFloatField(
        fmt=pycnometry.DPycThermalExpansionInsert,  # pylint: disable=no-member
        verbose_name=_("Thermal Expansion"),
    )
    kind = models.CharField(max_length=1, choices=INSERT_CHOICES)

    class Meta:
        ordering = ['name']
        constraints = [
            models.CheckConstraint(
                check=models.Q(is_method=False) | models.Q(last_updated__isnull=False),
                name='insert_methods_must_have_updated'
            ),
            models.CheckConstraint(
                check=models.Q(is_method=True) | models.Q(last_updated__isnull=True),
                name='insert_copies_dont_have_updated'
            ),
            models.UniqueConstraint(
                fields=['name', 'kind'],
                condition=models.Q(is_method=True),
                name='unique_insert_name'
            ),
        ]

    objects = InsertManager()

    relevant_if = {
        # pylint: disable=line-too-long
        'is_shared': ['!=', 'kind', 'R'],
        "other_expansion_um_per_mk": ["==", "material", "ot"]
    }

    @property
    def thermal_expansion_um_per_mk(self):
        "Get the thermal expansion in um_per_mk units for the material selected."
        if self.material in MATERIAL_CTE:
            return MATERIAL_CTE[self.material]
        return self.other_expansion_um_per_mk

    @property
    def thermal_expansion(self):
        """The thermal expansion coefficient of the insert (in linear fraction per K)"""
        return self.thermal_expansion_um_per_mk * 1.0e-6

    def __str__(self):
        return f'{pycnometry.DPycVolume(self.volume).ValueAndUnits()} - {self.name}'  # pylint: disable=no-member

    def save(self, *args, update=True, **kwargs):  # pylint: disable=arguments-differ
        if update and self.is_method:
            self.last_updated = timezone.now()
        super().save(*args, **kwargs)
        if update and self.is_method and self.is_shared:
            _notify_new_data()


class Gas(models.Model):
    '''The name of an available gas'''
    name = models.CharField(max_length=8)

    def __str__(self):
        return self.name


def _get_default_gas():
    return Gas.objects.get(name="Helium").id


class AnalysisParametersManager(models.Manager):
    """Accessors for specific types of analysis parameters."""

    def methods(self, **kwargs):
        '''Return all METHOD parameters'''
        return self.filter(is_method=True, **kwargs)

    def copy_method(self, method_id):
        """Make a non method copy of the method"""
        method = self.get(id=method_id)
        method.id = None
        if method.cup is not None:
            method.cup = copy_model_obj(Insert, method.cup)
            method.cup.save()
        if method.insert is not None:
            method.insert = copy_model_obj(Insert, method.insert)
            method.insert.save()
        method.is_method = False
        method.last_updated = None
        method.save()
        return method

    def default(self):
        """Get default method or set it if doesn't exist"""
        try:
            default = self.get(is_default=True, is_method=True)
        except AnalysisParameters.DoesNotExist:
            default = self.methods().first()
            default.is_default = True
            default.save()
        except AnalysisParameters.MultipleObjectsReturned:
            default = self.methods().first()
            default.is_default = True
            default.save()
        return default


def get_default_fill_direction():
    """return default fill direction for method form"""
    try:
        return 'R' if Instrument.objects.local().kind == '1' else 'S'
    except OperationalError:
        # the test for migrations gets here with funky combination of model setup.  So just return
        # a proper default if the models are funky.
        return 'S'


class AnalysisParameters(models.Model):
    '''The analysis parameters under which an analysis can be done.

    :param name: The name of the method.
    :param is_method: Is this a method, or a copy of a method for an analysis.
    :param is_shared: True if the user wants to share this method.
    '''

    name = models.CharField(max_length=64, verbose_name=_("Name"))
    is_method = models.BooleanField()
    is_default = models.BooleanField(default=False)
    is_shared = models.BooleanField(default=False, verbose_name=_("Share on MIC NET"))
    last_updated = models.DateTimeField(null=True)
    temperature = UnitsFloatField(
        fmt=pycnometry.DPycTemperature1350,  # pylint: disable=no-member
        verbose_name=_("Temperature")
    )
    cup = models.ForeignKey(
        Insert,
        related_name="analysis_cup",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name=_("Cup"),
    )
    use_vacuum_prep = models.BooleanField(default=False, verbose_name=_("Use Vacuum Preparation"))
    insert = models.ForeignKey(
        Insert,
        related_name="analysis_insert",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name=_("Insert")
    )
    fill_direction = models.CharField(
        max_length=1,
        choices=FILL_CHOICES,
        default=get_default_fill_direction,
        verbose_name=_("Fill Direction")
    )
    do_pre_purges = models.BooleanField(
        default=True,
        verbose_name=_("Pre-Purge While Changing Temperature")
    )
    purge_fill_pressure = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Purge Fill Pressure")
    )
    num_purges = UnitsIntegerField(
        fmt=pycnometry.UPycNumPurges,  # pylint: disable=no-member
        verbose_name=_("Purges"))
    num_cycles = UnitsIntegerField(
        fmt=pycnometry.UPycNumCycles,  # pylint: disable=no-member
        verbose_name=_("Cycles"))
    cycle_fill_pressure = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Cycle Fill Pressure")
    )
    endby = models.CharField(
        max_length=1,
        choices=END_CHOICES,
        default='E',
        verbose_name=_("Pressure Stabilization")
    )
    interval_time = UnitsIntegerField(
        fmt=pycnometry.UPycEquilTime,  # pylint: disable=no-member
        null=True,
        blank=True,
        verbose_name=_("Interval")
    )
    equilibration_rate = UnitsFloatField(
        fmt=pycnometry.DPycEquilRate,  # pylint: disable=no-member
        null=True,
        blank=True,
        verbose_name=_("Equilibration Rate")
    )
    p1_record_equilibration_data = models.BooleanField(
        default=False,
        verbose_name=_("Record P1 Equilibration Data")
    )
    p1_equilibration_cycles = UnitsIntegerField(
        fmt=pycnometry.UPycEquilDataCycles,  # pylint: disable=no-member
        null=True,
        blank=True,
        verbose_name=_("Record First")
    )
    p1_equilibration_points = UnitsIntegerField(
        fmt=pycnometry.UPycEquilDataPoints,  # pylint: disable=no-member
        null=True,
        blank=True,
        verbose_name=_("Record Last")
    )
    p2_record_equilibration_data = models.BooleanField(
        default=False,
        verbose_name=_('Record P2 Equilibration Data')
    )
    p2_equilibration_cycles = UnitsIntegerField(
        fmt=pycnometry.UPycEquilDataCycles,  # pylint: disable=no-member
        null=True,
        blank=True,
        verbose_name=_("Record First")
    )
    p2_equilibration_points = UnitsIntegerField(
        fmt=pycnometry.UPycEquilDataPoints,  # pylint: disable=no-member
        null=True,
        blank=True,
        verbose_name=_("Record Last")
    )
    technique = models.CharField(
        max_length=1,
        choices=TECHNIQUE_CHOICES,
        default='P',
        verbose_name=_("Technique")
    )
    resin_density = UnitsFloatField(
        fmt=pycnometry.DPycDensity,  # pylint: disable=no-member
        verbose_name=_("Cell Wall Density")
    )
    initial_pressure = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Initial Pressure")
    )
    final_pressure = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Final Pressure")
    )
    pressure_increment = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Pressure Increment")
    )
    low_pressure = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Low Pressure")
    )
    fracture_pressure = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Fracture Pressure")
    )
    use_run_precision = models.BooleanField(default=False, verbose_name=_("Use Run Precision"))
    run_precision_tolerance = UnitsFloatField(
        fmt=pycnometry.DPycRunPrecPercentFullScale,  # pylint: disable=no-member
        null=True,
        blank=True,
        verbose_name=_("Percent Full Scale")
    )

    # The FoamPyc geometry fields.
    compute_volume = models.CharField(
        max_length=1,
        choices=COMPUTE_VOLUME_CHOICES,
        default='C',
        verbose_name=_("Compute Volume")
    )
    geometric_volume = UnitsFloatField(
        fmt=pycnometry.DPycVolume,  # pylint: disable=no-member
        verbose_name=_("Geometric Volume")
    )
    active_area = mic_python_lib.rest_input.UnitsFloatField(
        fmt=pycnometry.DPycActiveArea,  # pylint: disable=no-member
        verbose_name=_("Active Area")
    )
    shape = models.CharField(
        max_length=1,
        choices=SHAPE_CHOICES,
        default='U',
        verbose_name=_("Shape")
    )
    cube_edge_length = UnitsFloatField(
        fmt=pycnometry.DPycLength,  # pylint: disable=no-member
        verbose_name=_("Cube Edge Length")
    )
    cylinder_diameter = UnitsFloatField(
        fmt=pycnometry.DPycLength,  # pylint: disable=no-member
        verbose_name=_("Cylinder Diameter")
    )
    cylinder_length = UnitsFloatField(
        fmt=pycnometry.DPycLength,  # pylint: disable=no-member
        verbose_name=_("Cylinder Length")
    )
    rect_long_edge = UnitsFloatField(
        fmt=pycnometry.DPycLength,  # pylint: disable=no-member
        verbose_name=_("Rect Long Edge")
    )
    rect_short_edge = UnitsFloatField(
        fmt=pycnometry.DPycLength,  # pylint: disable=no-member
        verbose_name=_("Rect Short Edge")
    )
    rect_other_edge = UnitsFloatField(
        fmt=pycnometry.DPycLength,  # pylint: disable=no-member
        verbose_name=_("Rect Other Edge")
    )
    number_of_pieces = UnitsIntegerField(
        fmt=pycnometry.UPycNumberOfPieces,  # pylint: disable=no-member
        verbose_name=_("Number of Pieces")
    )
    measure_by = models.CharField(
        max_length=1,
        choices=MEASURE_BY_CHOICES,
        default='L',
        verbose_name=_("Measure By")
    )
    chord_length = UnitsFloatField(
        fmt=pycnometry.DPycLength,  # pylint: disable=no-member
        verbose_name=_("Chord Length")
    )
    cell_diameter = UnitsFloatField(
        fmt=pycnometry.DPycLength,  # pylint: disable=no-member
        verbose_name=_("Cell Diameter")
    )
    cube_number_of_skins = UnitsIntegerField(
        fmt=pycnometry.UPycNumberOfSkinsCube,  # pylint: disable=no-member
        verbose_name=_("Cube Number of Skins")
    )
    cyl_number_of_skins = UnitsIntegerField(
        fmt=pycnometry.UPycNumberOfSkins,  # pylint: disable=no-member
        verbose_name=_("Cyl Number of Skins")
    )
    rect_number_of_large_skins = UnitsIntegerField(
        fmt=pycnometry.UPycNumberOfSkins,  # pylint: disable=no-member
        verbose_name=_("Rect Number of Large Skins")
    )
    rect_number_of_small_skins = UnitsIntegerField(
        fmt=pycnometry.UPycNumberOfSkins,  # pylint: disable=no-member
        verbose_name=_("Rect Number of Small Skins")
    )
    rect_number_of_other_skins = UnitsIntegerField(
        fmt=pycnometry.UPycNumberOfSkins,  # pylint: disable=no-member
        verbose_name=_("Rect Number of Other Skins")
    )

    relevant_if = {
        # pylint: disable=line-too-long
        'num_cycles': ["&&", ['!=', 'technique', 'D'], ['!=', 'technique', 'E']],
        'cycle_fill_pressure': ["&&", ['!=', 'technique', 'D'], ['!=', 'technique', 'E']],
        'p1_record_equilibration_data': ['==', 'technique', 'P'],
        'p2_record_equilibration_data': ['==', 'technique', 'P'],
        'p1_equilibration_cycles': ['==', 'p1_record_equilibration_data', True],
        'p1_equilibration_points': ['==', 'p1_record_equilibration_data', True],
        'p2_equilibration_cycles': ['==', 'p2_record_equilibration_data', True],
        'p2_equilibration_points': ['==', 'p2_record_equilibration_data', True],
        'initial_pressure': ['==', 'technique', 'D'],
        'final_pressure': ['==', 'technique', 'D'],
        'pressure_increment': ['==', 'technique', 'D'],
        'purge_fill_pressure': ['==', 'use_vacuum_prep', False],
        'low_pressure': ['==', 'technique', 'E'],
        'fracture_pressure': ['==', 'technique', 'E'],
        'use_run_precision': ['==', 'technique', 'P'],
        'run_precision_tolerance': ['==', 'use_run_precision', True],
        'interval_time': ['==', 'endby', 'F'],
        'resin_density': ['!=', 'technique', 'P'],
        'compute_volume': ['!=', 'technique', 'P'],
        'geometric_volume': ['==', 'compute_volume', 'E'],
        'active_area': ["&&", ['==', 'compute_volume', 'E'], ['==', 'technique', 'A']],
        'shape': ['==', 'compute_volume', 'C'],
        'cube_edge_length': ['==', 'shape', 'U'],
        'cube_number_of_skins': ["&&", ['==', 'shape', 'U'], ['==', 'technique', 'A']],
        'cylinder_diameter': ['==', 'shape', 'Y'],
        'cylinder_length': ['==', 'shape', 'Y'],
        'cyl_number_of_skins': ["&&", ['==', 'shape', 'Y'], ['==', 'technique', 'A']],
        'rect_long_edge': ['==', 'shape', 'R'],
        'rect_short_edge': ['==', 'shape', 'R'],
        'rect_other_edge': ['==', 'shape', 'R'],
        'rect_number_of_large_skins': ["&&", ['==', 'shape', 'R'], ['==', 'technique', 'A']],
        'rect_number_of_small_skins': ["&&", ['==', 'shape', 'R'], ['==', 'technique', 'A']],
        'rect_number_of_other_skins': ["&&", ['==', 'shape', 'R'], ['==', 'technique', 'A']],
        'number_of_pieces': ['==', 'compute_volume', 'C'],
        'measure_by': ["&&", ['!=', 'technique', 'P'], ['==', 'technique', 'A']],
        'chord_length': ['==', 'measure_by', 'L'],
        'cell_diameter': ['==', 'measure_by', 'D'],
    }
    objects = AnalysisParametersManager()

    class Meta:
        ordering = ['name']
        constraints = [
            models.CheckConstraint(
                check=models.Q(is_method=False) | models.Q(last_updated__isnull=False),
                name='methods_must_have_updated'
            ),
            models.CheckConstraint(
                check=models.Q(is_method=True) | models.Q(last_updated__isnull=True),
                name='copies_dont_have_updated'
            ),
            models.UniqueConstraint(
                fields=['name'], condition=models.Q(is_method=True), name='unique_method_name'
            ),
        ]
        verbose_name = _("Method")

    def __str__(self):
        return self.name

    def save(self, *args, update=True, **kwargs):  # pylint: disable=arguments-differ
        if update and self.is_method:
            self.last_updated = timezone.now()
        super().save(*args, **kwargs)
        if update and self.is_method  and self.is_shared:
            _notify_new_data()


class InstrumentManager(models.Manager):  # pylint: disable=too-few-public-methods
    """Accessors for specific types of instruments."""

    def local(self):
        """:return: the one and only instrument connected to this system."""
        return self.get(is_local=True)


class Instrument(models.Model):
    '''The information for each known instrument.

    :param model: The model number for the instrument
    :param serial: The serial number of the instrument.
    :param is_local: True if the instrument is attached to this hardware.
    :param kind: What type of hardware configuration this is.
    :param is_temp_control: True if the instrument supports full range temperature control.
    :param material: Which of the known materials it is.
    :param other_expansion_um_per_mk: The thermal expansion coefficient of the instrument as
        entered by the user in um/m·K, for the other material selection.
    :param gas: The currently selected gas of the instrument
    :param device_service: The date at which the next service is due.
    :param device_warranty: The date at which warranty expires.
    '''

    model = models.CharField(max_length=4, default='1350')
    serial = models.CharField(max_length=8, verbose_name=_("Serial Number"))
    is_local = models.BooleanField(default=False)
    kind = models.CharField(
        max_length=1,
        choices=INSTRUMENT_CHOICES,
        verbose_name=_("Configuration")
    )
    is_temp_control = models.BooleanField(
        default=True,
        choices=((True, "ATC"), (False, "TS")),
        verbose_name=_("Temperature Configuration")
    )
    material = models.CharField(
        max_length=2,
        choices=MATERIAL_CHOICES,
        default="Al",
    )
    other_expansion_um_per_mk = UnitsFloatField(
        fmt=pycnometry.DPycThermalExpansionInstrument,  # pylint: disable=no-member
        verbose_name=_("Chamber Thermal Expansion")
    )
    gas = models.ForeignKey(
        Gas,
        default=_get_default_gas,
        on_delete=models.PROTECT,
        verbose_name=_("Gas")
    )
    device_service = models.DateField(default=datetime.date.today)
    device_warranty = models.DateField(default=datetime.date.today)
    last_updated = models.DateTimeField(null=False)

    relevant_if = {
        "other_expansion_um_per_mk": ["==", "material", "ot"]
    }

    @property
    def thermal_expansion_um_per_mk(self):
        "Get the thermal expansion in um_per_mk units for the material selected."
        if self.material in MATERIAL_CTE:
            return MATERIAL_CTE[self.material]
        return self.other_expansion_um_per_mk

    @property
    def thermal_expansion(self):
        """The thermal expansion coefficient of the instrument (in linear fraction per K)"""
        return self.thermal_expansion_um_per_mk * 1.0e-6

    @property
    def configuration(self):
        """The human readable configuration for the instrument."""
        if self.model == '1350':
            return self.get_kind_display() + ' (' + self.get_is_temp_control_display() + ')'
        return self.get_kind_display()

    @property
    def full_config(self):
        """The human readable full configuration for the instrument."""
        return "AccuPyc " + self.model + " " + self.configuration + " SN: " + self.serial

    @classmethod
    def local_config(cls):
        "The dictionary for the local instrument configuration."
        local = cls.objects.local()
        return {
            'model': local.model,
            'serial': local.serial,
            'kind': local.kind,
            'is_temp_control': local.is_temp_control,
        }

    @classmethod
    def match_local_config(cls, config):
        "returns true if the configuration matches out local instrument."
        try:
            local = cls.objects.local()
            return ((local.model, local.kind, local.is_temp_control)
                    == (config['model'], config['kind'], config['is_temp_control']))
        except KeyError:
            return False

    objects = InstrumentManager()

    def save(self, *args, update=True, **kwargs):  # pylint: disable=arguments-differ
        try:
            if self.id == Instrument.objects.local().id:
                _set_local_hostname(self.serial)
        except Instrument.DoesNotExist:
            pass
        if update:
            self.last_updated = timezone.now()
        super().save(*args, **kwargs)
        if update:
            _notify_new_data()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['model', 'serial'], condition=models.Q(is_local=True), name='single_local'
            ),
        ]


class RTDCalibration(models.Model):
    '''The information for a single rtd resistance to temperature calibration.

    The temperature calibration is a linear point slope form.

    :param instrument: The instrument on which it was calibrated.
    :param channel: Which RTD channel is the calibration for.
    :param time: When the calibration occurred.
    :param slope: The slope of the calibration; delta T over delta R.
    :param raw_resistance: The raw resistance of the calibrated point.
    :param actual_resistance: The actual converted resistance used in the lookup table.
    :param check_reference: The reference gauge reading for the check point.
    :param check_reading: The calibrated reading for the check point.
    :param gauge_sn: The serial number of the gauge that was used on the calibration.
    '''
    instrument = models.ForeignKey(Instrument, on_delete=models.CASCADE)
    channel = models.CharField(max_length=1, choices=TEMPCHANNEL_CHOICES)
    time = models.DateTimeField(default=timezone.now)
    slope = models.FloatField()
    raw_resistance = models.FloatField()
    actual_resistance = models.FloatField()
    check_reference = models.FloatField()
    check_reading = models.FloatField()
    gauge_sn = models.CharField(max_length=16)


class PressureCalibration(models.Model):
    '''The information for a single voltage to pressure calibration.

    The pressure calibration is a linear point slope form.

    :param instrument: The instrument on which it was calibrated.
    :param time: When the calibration occurred.
    :param slope: The slope of the calibration delta P over delta V.
    :param voltage: The voltage of the calibrated point.
    :param pressure: The pressure of the calibrated point.
    :param check_reference: The reference gauge reading for the check point.
    :param check_reading: The calibrated reading for the check point.
    :param gauge_sn: The serial number of the gauge that was used on the calibration.
    '''
    instrument = models.ForeignKey(Instrument, on_delete=models.CASCADE)
    time = models.DateTimeField(default=timezone.now)
    slope = models.FloatField()
    voltage = models.FloatField()
    pressure = models.FloatField()
    check_reference = models.FloatField()
    check_reading = models.FloatField()
    gauge_sn = models.CharField(max_length=16)


class CalibrationManager(models.Manager):  # pylint: disable=too-few-public-methods
    """The model manager for Calibrations."""

    def get_current(self):
        """Get the currently active calibration for the local instrument."""
        return self.filter(
            instrument=Instrument.objects.local(), kind__in=['C', 'R', 'UA'], state='C'
        ).order_by('-start_time').first()

    def get_last_factory(self):
        """Get the most recent service calibration for the local instrument."""
        return self.filter(
            instrument=Instrument.objects.local(), kind='C', state='C'
        ).order_by('-start_time').first()

    def complete_cal(self, **kwargs):
        """:return: query for all completed calibrations."""
        return self.filter(state='C', kind='C', **kwargs)


class Calibration(models.Model):
    '''The results of a calibration operation

    :param state: What is the state of the calibration.
    :param start_time: When did the calibration start.
    :param kind: Is it a calibration or check.
    :param fill_pressure: What was the requested fill pressure.
    :param instrument: Which instrument was this done on.
    :param ref_ball: Which reference ball was used.
    :param cup: Which cup was used.
    :param check_of: The calibration for which this is a check. only non null if kind is check.
    :param expansion_fs_empty: The expansion data for filling sample without reference volume.
    :param expansion_fs_ref: The expansion data for filling sample with reference volume.
    :param expansion_fr_empty: The expansion data for filling reference without reference volume.
    :param expansion_fr_ref: The expansion data for filling reference with reference volume.
    :param volume_fs_sample: The sample volume (cm3) when filling sample.
    :param volume_fs_reference: The reference volume (cm3) when filling sample.
    :param volume_fr_sample: The sample volume (cm3) when filling reference.
    :param volume_fr_reference: The reference volume (cm3) when filling reference.
    :param temperature: The temperature (C) corresponding to the volumes.
    '''

    state = models.CharField(max_length=1, choices=STATE_CHOICES, default='P')
    review_active = models.BooleanField(default=False)
    start_time = models.DateTimeField(default=timezone.now)
    kind = models.CharField(max_length=2, choices=CALIBRATION_CHOICES)
    last_updated = models.DateTimeField(null=False)
    fill_pressure = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Fill Pressure"),
    )
    collection_temperature = models.FloatField(default=0.0)
    gas = models.CharField(max_length=16)
    instrument = models.ForeignKey(Instrument, on_delete=models.CASCADE)
    instrument_thermal_expansion_um_per_mk = UnitsFloatField(
        fmt=pycnometry.DPycThermalExpansionInstrument,  # pylint: disable=no-member
    )
    ref_ball = models.ForeignKey(
        Insert, related_name="calibration_ref_ball", on_delete=models.PROTECT
    )
    cup = models.ForeignKey(
        Insert, related_name="calibration_cup", null=True, on_delete=models.PROTECT
    )
    check_of = models.ForeignKey("self", null=True, default=None, on_delete=models.CASCADE)
    expansion_fs_empty = models.OneToOneField(
        Expansion, on_delete=models.CASCADE, null=True, related_name="analysis_fs_empty"
    )
    expansion_fs_ref = models.OneToOneField(
        Expansion, on_delete=models.CASCADE, null=True, related_name="analysis_fs_ref"
    )
    expansion_fr_empty = models.OneToOneField(
        Expansion, on_delete=models.CASCADE, null=True, related_name="analysis_fr_empty"
    )
    expansion_fr_ref = models.OneToOneField(
        Expansion, on_delete=models.CASCADE, null=True, related_name="analysis_fr_ref"
    )
    volume_fs_sample = models.FloatField(null=True)
    volume_fs_reference = models.FloatField(null=True)
    volume_fr_sample = models.FloatField(null=True)
    volume_fr_reference = models.FloatField(null=True)
    temperature = models.FloatField(null=True)

    @property
    def instrument_thermal_expansion(self):
        """The thermal expansion coefficient of the instrument (in linear fraction per K)"""
        return self.instrument_thermal_expansion_um_per_mk * 1.0e-6

    objects = CalibrationManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['start_time', 'instrument'], name='cal_id'),
        ]
        verbose_name = _("Volume Calibration")

    def save(self, *args, update=True, **kwargs):  # pylint: disable=arguments-differ
        if update:
            self.last_updated = timezone.now()
        super().save(*args, **kwargs)
        if update and self.state == 'C':
            _notify_new_data()


class OtherOperation(models.Model):
    '''The status of operations that are not analysis nor volume calibration.

    :param state: What is the state of the calibration.
    :param start_time: When did the calibration start.
    :param kind: Which operation it is.
    :param number: Some operations are indexed, this provides which one.
    :param instrument: Which instrument was this done on.
    '''

    state = models.CharField(max_length=1, choices=STATE_CHOICES, default='P')
    review_active = models.BooleanField(default=False)
    start_time = models.DateTimeField(default=timezone.now)
    kind = models.CharField(max_length=1, choices=OPERATION_CHOICES)
    number = models.IntegerField(default=0)
    instrument = models.ForeignKey(Instrument, on_delete=models.CASCADE)
    script_run = models.ForeignKey(
        mpl_models.InstrumentScriptRun, on_delete=models.CASCADE, null=True, default=None)


class InsertMeasure(models.Model):
    "The results (or in process) for an insert (cup/multivolume) calibration."
    state = models.CharField(max_length=1, choices=STATE_CHOICES, default='P')
    review_active = models.BooleanField(default=False)
    start_time = models.DateTimeField(default=timezone.now, verbose_name=_("Start Time"))
    end_time = models.DateTimeField(null=True)
    instrument = models.ForeignKey(Instrument, on_delete=models.CASCADE)
    kind = models.CharField(max_length=1, choices=INSERT_NO_REF_CHOICES, default='C',
                            verbose_name=_("Type"))
    fill_pressure = UnitsFloatField(
        fmt=pycnometry.DPycPressureGauge,  # pylint: disable=no-member
        verbose_name=_("Fill Pressure"),
    )
    ref_ball = models.ForeignKey(
        Insert, related_name="insertcal_ref_ball", null=True, on_delete=models.PROTECT
    )
    multivolume_insert = models.ForeignKey(
        Insert, related_name="insertcal_multivolume", null=True, on_delete=models.PROTECT
    )
    gas = models.CharField(max_length=16)
    collection_temperature = models.FloatField(default=0.0)
    calibration = models.ForeignKey(Calibration, on_delete=models.CASCADE)
    expansion_fs = models.OneToOneField(
        Expansion, on_delete=models.CASCADE, null=True, related_name="insertcal_fill_sample"
    )
    expansion_fr = models.OneToOneField(
        Expansion, on_delete=models.CASCADE, null=True, related_name="insertcal_fill_reference"
    )
    results = models.ManyToManyField(mpl_models.Result)
    created_insert = models.ForeignKey(
        Insert, related_name="insertcal_created", null=True, on_delete=models.SET_NULL
    )

    relevant_if = {
        'ref_ball': ['==', 'kind', 'C'],
        'multivolume_insert': ['==', 'kind', 'C'],
    }

    class Meta:
        ordering = ['-start_time']
        constraints = [
            models.UniqueConstraint(fields=['start_time', 'instrument'], name='insert_cal_id'),
        ]


class AnalysisResult(models.Model):
    '''The results (or in process) for an analysis

    '''
    state = models.CharField(max_length=1, choices=STATE_CHOICES, default='P')
    review_active = models.BooleanField(default=False)
    start_time = models.DateTimeField(default=timezone.now, verbose_name=_("Start Time"))
    end_time = models.DateTimeField(null=True)
    last_updated = models.DateTimeField(null=False)
    instrument = models.ForeignKey(Instrument, on_delete=models.CASCADE)
    mass = UnitsFloatField(fmt=i1350.DPycSampleMass1350, verbose_name=_("Mass"))  # pylint: disable=no-member
    name = models.CharField(max_length=64, default='', blank=True, verbose_name=_("Description"))
    operator = models.CharField(max_length=64, default='', blank=True, verbose_name=_("Operator"))
    results = models.ManyToManyField(mpl_models.Result)
    sample_log = models.ManyToManyField(mpl_models.SampleLog)

    class Meta:
        ordering = ['-start_time']
        constraints = [
            models.UniqueConstraint(fields=['start_time', 'instrument'], name='result_id'),
        ]

    def save(self, *args, update=True, **kwargs):  # pylint: disable=arguments-differ
        ResultCache.objects.filter(results_id=self.id).delete()
        if update:
            self.last_updated = timezone.now()
        super().save(*args, **kwargs)
        if update and self.state == 'C':
            _notify_new_data()


class AnalysisResult1350Manager(models.Manager):  # pylint: disable=too-few-public-methods
    """Accessors for specific types of results."""

    def complete(self, **kwargs):
        """:return: completed results."""
        return self.filter(state='C', **kwargs)


class AnalysisResult1350(AnalysisResult):
    '''The results (or in process) for an analysis'''

    parameters = models.OneToOneField(AnalysisParameters, on_delete=models.CASCADE)
    gas = models.ForeignKey(Gas, on_delete=models.PROTECT, verbose_name=_("Gas"))
    calibration = models.ForeignKey(Calibration, on_delete=models.CASCADE)
    expansion = models.OneToOneField(
        Expansion, on_delete=models.CASCADE, null=True, related_name="analysis"
    )
    expansion_repeat = models.OneToOneField(
        Expansion, on_delete=models.CASCADE, null=True, related_name="analysis_repeat"
    )

    objects = AnalysisResult1350Manager()

    @property
    def technique(self):  # pylint: disable=missing-function-docstring
        return self.parameters.technique

    def get_technique_display(self):  # pylint: disable=missing-function-docstring
        return self.parameters.get_technique_display()

    class Meta:
        ordering = ['-start_time']


class AnalysisResult1345(AnalysisResult):  # pylint: disable=too-many-instance-attributes
    "The results for an analysis that was scraped from a 1345."

    technique = models.CharField(
        max_length=1,
        choices=TECHNIQUE_CHOICES,
        default='P',
        verbose_name=_("Technique")
    )
    temperature = UnitsFloatField(fmt=pycnometry.DPycTemperature1350)  # pylint: disable=no-member
    num_purges = UnitsIntegerField(fmt=pycnometry.UPycNumPurges)  # pylint: disable=no-member
    endby = models.CharField(
        max_length=1,
        choices=END_CHOICES,
        default='E',
    )
    interval_time = UnitsIntegerField(
        fmt=pycnometry.UPycEquilTime,  # pylint: disable=no-member
        null=True,
        blank=True,
    )
    equilibration_rate = UnitsFloatField(
        fmt=pycnometry.DPycEquilRate,  # pylint: disable=no-member
        null=True,
        blank=True,
    )
    resin_density = UnitsFloatField(fmt=pycnometry.DPycDensity)  # pylint: disable=no-member
    geometric_volume = UnitsFloatField(fmt=pycnometry.DPycVolume)  # pylint: disable=no-member
    active_area = mic_python_lib.rest_input.UnitsFloatField(fmt=pycnometry.DPycActiveArea)  # pylint: disable=no-member
    measure_by = models.CharField(max_length=1, choices=MEASURE_BY_CHOICES, default='L')
    chord_length = UnitsFloatField(fmt=pycnometry.DPycLength)  # pylint: disable=no-member
    cell_diameter = UnitsFloatField(fmt=pycnometry.DPycLength)  # pylint: disable=no-member

    expansion = models.OneToOneField(
        Expansion1345, on_delete=models.CASCADE, related_name="analysis"
    )
    expansion_repeat = models.OneToOneField(
        Expansion1345, on_delete=models.CASCADE, null=True, related_name="analysis_repeat"
    )

    class Meta:
        ordering = ['-start_time']


class Equilibration(models.Model):
    '''The pressure equilibration data for each point

    :param analysis: The analysis for which this was taken.
    :param cycle: the number of the cycle for which this data is.
    :param which_equil: For which pressure this equilibration is.
    :param num_points: The number of points of equilibration.

    :param times: Binary array of uint_32 each giving the time since
        start of analysis in milliseconds.

    :param pressures: Binary array of float_32 each giving the pressure
        in psia.
    '''

    analysis = models.ForeignKey(AnalysisResult1350, on_delete=models.CASCADE)
    cycle = models.IntegerField()
    which_equil = models.CharField(max_length=1, choices=EQUILIBRATION_CHOICES)
    num_points = models.IntegerField()
    times_binary = models.BinaryField()
    pressures_binary = models.BinaryField()

    @property
    def times(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num_points * "I", self.times_binary)

    @times.setter
    def times(self, value):  # pylint: disable=missing-function-docstring
        self.times_binary = struct.pack(self.num_points * "I", *value)

    @property
    def pressures(self):  # pylint: disable=missing-function-docstring
        return struct.unpack(self.num_points * "f", self.pressures_binary)

    @pressures.setter
    def pressures(self, value):  # pylint: disable=missing-function-docstring
        self.pressures_binary = struct.pack(self.num_points * "f", *value)

class ResultCache(models.Model):
    'The cache of results/reports for the analysis result.'
    results = models.ForeignKey(AnalysisResult, on_delete=models.CASCADE)
    report_json = models.TextField()

def _is_on_instrument():
    """:return: True if the system is running on the instrument and not a test system."""
    if not hasattr(platform, 'machine'):
        return False

    return platform.machine().startswith('aarch64')


def _set_local_hostname(serial):
    """Set the hostname of the system to "accupyc3-<serial>". """
    hostname = f'accupyc3-{serial}'
    if _is_on_instrument():
        subprocess.run(['hostnamectl', 'set-hostname', hostname], check=False)


def _notify_new_data():
    "Tells all subnets that we have new data available."
    try:
        local = Instrument.objects.local()
        if _is_on_instrument():
            notify_new_data(model=local.model, serial=local.serial)
    except Instrument.DoesNotExist:
        pass


class LegacyInstrument(models.Model):
    '''The known legacy instruments and their IP address'''

    model = models.CharField(max_length=4, verbose_name=_("Model Number"))
    ip_addr = models.CharField(max_length=15, verbose_name=_("IP Address"))
    last_communicated = models.DateTimeField(default=timezone.now)

    def __str__(self):
        return f'Model: {self.model}, IP: {self.ip_addr}'  # pylint: disable=no-member
