'''The views and support functionality of the rest api for data sharing'''
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import path

from mic_python_lib.rest import (
    RestAPIData,
    UpdateMethod,
    UpdateResult,
    function_get_to_json_view,
    merge_model_dict,
)
from .models import (
    AnalysisParameters,
    AnalysisResult1345,
    AnalysisResult1350,
    Calibration,
    Equilibration,
    Expansion,
    Expansion1345,
    Gas,
    Insert,
    Instrument,
)
from .services import update_results

# -------------------------------------------------------------------------------------------------
# module constants

#: The format for the AnalysisResult1350 object as needed by microactive.
_MICROACTIVE_RESULT_FIELDS = [
    'state', 'format_date(start_time)', 'format_date(end_time)', 'mass', 'name', 'operator',
    'format_expansion(expansion)', 'format_expansion(expansion_repeat)', 'instrument.model',
    'instrument.serial', 'instrument.is_local', 'instrument.kind', 'instrument.is_temp_control',
    ('instrument.thermal_expansion_um_per_mk',
     'calibration.instrument_thermal_expansion_um_per_mk'),
    'parameters.name', 'format_gas(gas)',
    'parameters.temperature', 'parameters.use_vacuum_prep', 'parameters.format_insert(cup)',
    'parameters.format_insert(insert)', 'parameters.fill_direction', 'parameters.do_pre_purges',
    'parameters.purge_fill_pressure', 'parameters.num_purges', 'parameters.num_cycles',
    'parameters.cycle_fill_pressure', 'parameters.endby', 'parameters.interval_time',
    'parameters.equilibration_rate', 'parameters.p1_record_equilibration_data',
    'parameters.p1_equilibration_cycles', 'parameters.p1_equilibration_points',
    'parameters.p2_record_equilibration_data', 'parameters.p2_equilibration_cycles',
    'parameters.p2_equilibration_points', 'parameters.technique', 'parameters.resin_density',
    'parameters.initial_pressure', 'parameters.final_pressure', 'parameters.pressure_increment',
    'parameters.low_pressure', 'parameters.fracture_pressure', 'parameters.use_run_precision',
    'parameters.run_precision_tolerance', 'parameters.compute_volume',
    'parameters.geometric_volume', 'parameters.active_area', 'parameters.shape',
    'parameters.cube_edge_length', 'parameters.cylinder_diameter', 'parameters.cylinder_length',
    'parameters.rect_long_edge', 'parameters.rect_short_edge', 'parameters.rect_other_edge',
    'parameters.number_of_pieces', 'parameters.measure_by', 'parameters.chord_length',
    'parameters.cell_diameter', 'parameters.cube_number_of_skins', 'parameters.cyl_number_of_skins',
    'parameters.rect_number_of_large_skins', 'parameters.rect_number_of_small_skins',
    'parameters.rect_number_of_other_skins', 'calibration.volume_fs_sample',
    'calibration.volume_fs_reference', 'calibration.volume_fr_sample',
    'calibration.volume_fr_reference', 'calibration.temperature',
    'format_equilibration_set(equilibration_set)'
]

_INSTRUMENT_FIELDS = [
    'model',
    'serial',
    'kind',
    'format_gas(gas)',
    'is_temp_control',
    'material',
    'other_expansion_um_per_mk',
    'device_service',
    'device_warranty',
    'last_updated',
]

_CALIBRATION_FIELDS = [
    'start_time',
    'last_updated',
    'fill_pressure',
    'collection_temperature',
    'gas',
    'instrument.model',
    'instrument.serial',
    'instrument_thermal_expansion_um_per_mk',
    'format_insert(ref_ball)',
    'format_insert(cup)',
    'format_expansion(expansion_fs_empty)',
    'format_expansion(expansion_fs_ref)',
    'format_expansion(expansion_fr_empty)',
    'format_expansion(expansion_fr_ref)',
    'volume_fs_sample',
    'volume_fs_reference',
    'volume_fr_sample',
    'volume_fr_reference',
    'temperature',
]

_RESULT_FIELDS = [
    'start_time',
    'end_time',
    'last_updated',
    'mass',
    'name',
    'operator',
    'instrument.model',
    'instrument.serial',
    'calibration.start_time',
    'parameters.name',
    'format_gas(gas)',
    'parameters.temperature',
    'parameters.use_vacuum_prep',
    'parameters.format_insert(cup)',
    'parameters.format_insert(insert)',
    'parameters.fill_direction',
    'parameters.do_pre_purges',
    'parameters.purge_fill_pressure',
    'parameters.num_purges',
    'parameters.num_cycles',
    'parameters.cycle_fill_pressure',
    'parameters.endby',
    'parameters.interval_time',
    'parameters.equilibration_rate',
    'parameters.p1_record_equilibration_data',
    'parameters.p1_equilibration_cycles',
    'parameters.p1_equilibration_points',
    'parameters.p2_record_equilibration_data',
    'parameters.p2_equilibration_cycles',
    'parameters.p2_equilibration_points',
    'parameters.technique',
    'parameters.resin_density',
    'parameters.initial_pressure',
    'parameters.final_pressure',
    'parameters.pressure_increment',
    'parameters.low_pressure',
    'parameters.fracture_pressure',
    'parameters.use_run_precision',
    'parameters.run_precision_tolerance',
    'parameters.compute_volume',
    'parameters.geometric_volume',
    'parameters.active_area',
    'parameters.shape',
    'parameters.cube_edge_length',
    'parameters.cylinder_diameter',
    'parameters.cylinder_length',
    'parameters.rect_long_edge',
    'parameters.rect_short_edge',
    'parameters.rect_other_edge',
    'parameters.number_of_pieces',
    'parameters.measure_by',
    'parameters.chord_length',
    'parameters.cell_diameter',
    'parameters.cube_number_of_skins',
    'parameters.cyl_number_of_skins',
    'parameters.rect_number_of_large_skins',
    'parameters.rect_number_of_small_skins',
    'parameters.rect_number_of_other_skins',
    'format_expansion(expansion)',
    'format_expansion(expansion_repeat)',
    'format_equilibration_set(equilibration_set)',
    'format_sample_log(sample_log)',
]

_RESULT1345_FIELDS = [
    'start_time',
    'end_time',
    'last_updated',
    'mass',
    'name',
    'operator',
    'instrument.model',
    'instrument.serial',
    'technique',
    'temperature',
    'num_purges',
    'endby',
    'interval_time',
    'equilibration_rate',
    'resin_density',
    'geometric_volume',
    'active_area',
    'measure_by',
    'chord_length',
    'cell_diameter',
    'format_expansion1345(expansion)',
    'format_expansion1345(expansion_repeat)',
    'format_sample_log(sample_log)',
]

_INSERT_FIELDS = [
    'name',
    'volume',
    'temperature',
    'material',
    'other_expansion_um_per_mk',
    'is_measured',
    'kind',
    'last_updated',
]

_METHOD_FIELDS = [
    'name',
    'temperature',
    'use_vacuum_prep',
    'format_insert(cup)',
    'format_insert(insert)',
    'fill_direction',
    'do_pre_purges',
    'purge_fill_pressure',
    'num_purges',
    'num_cycles',
    'cycle_fill_pressure',
    'endby',
    'interval_time',
    'equilibration_rate',
    'p1_record_equilibration_data',
    'p1_equilibration_cycles',
    'p1_equilibration_points',
    'p2_record_equilibration_data',
    'p2_equilibration_cycles',
    'p2_equilibration_points',
    'technique',
    'resin_density',
    'initial_pressure',
    'final_pressure',
    'pressure_increment',
    'low_pressure',
    'fracture_pressure',
    'use_run_precision',
    'run_precision_tolerance',
    'compute_volume',
    'geometric_volume',
    'active_area',
    'shape',
    'cube_edge_length',
    'cylinder_diameter',
    'cylinder_length',
    'rect_long_edge',
    'rect_short_edge',
    'rect_other_edge',
    'number_of_pieces',
    'measure_by',
    'chord_length',
    'cell_diameter',
    'cube_number_of_skins',
    'cyl_number_of_skins',
    'rect_number_of_large_skins',
    'rect_number_of_small_skins',
    'rect_number_of_other_skins',
    'last_updated',
]

def _format_date(value):
    return int(value.timestamp() + 0.5) if value else 0


def _format_equilibration(equilibration):
    return REST_API_DATA_1350.dict_from_fields(
        equilibration, [
            'cycle',
            'which_equil',
            'num_points',
            'times',
            'pressures',
        ]
    )


def _format_equilibration_set(equilibration_set):
    if not equilibration_set:
        return None
    return [_format_equilibration(e) for e in equilibration_set.all()]


def _format_expansion(expansion):
    if not expansion:
        return None
    return {
        'num': expansion.num,
        'p0s': expansion.p0s,
        'p1s': expansion.p1s,
        'p2s': expansion.p2s,
        'excludes': expansion.excludes,
        'temp_anl_block': expansion.temp_anl_block,
        'temp_valve_block': expansion.temp_valve_block,
        'times': expansion.times,
    }


def _format_expansion1345(expansion):
    if not expansion:
        return None
    return {
        'num': expansion.num,
        'volumes': expansion.volumes,
        'p1s': expansion.p1s,
        'p2s': expansion.p2s,
        'excludes': expansion.excludes,
        'times': expansion.times,
        'temp_anl_block': expansion.volumes,
    }


def _format_gas(gas):
    return gas.name


def _format_insert(insert):
    if not insert:
        return None
    return {
        'name': insert.name,
        'volume': insert.volume,
        'temperature': insert.temperature,
        'material': insert.material,
        'other_expansion_um_per_mk': insert.other_expansion_um_per_mk,
        'thermal_expansion': insert.thermal_expansion,
        'is_measured': insert.is_measured,
    }


def _format_sample_log(sample_log):
    return [{
        'datetime': log.datetime,
        'message': log.message,
    } for log in sample_log.all()]


def _set_cup(method, value):
    # If a local cup of the same name exists, that is used for the method.
    try:
        method.cup = Insert.objects.get(is_method=True, kind='C', name=value['name'])
        return
    except Insert.DoesNotExist:
        pass

    # If there is no cup of the same name, and there is only one local cup, it is used.
    try:
        method.cup = Insert.objects.get(is_method=True, kind='C')
        return
    except Insert.MultipleObjectsReturned:
        pass
    except Insert.DoesNotExist:
        pass

    # Otherwise, don't share the cup.
    method.cup = None


def _set_insert(method, value):
    # If a local insert of the same name exists, that is used for the method.
    try:
        method.insert = Insert.objects.get(is_method=True, kind='I', name=value['name'])
        return
    except Insert.DoesNotExist:
        pass

    # Otherwise, don't share the insert
    method.insert = None


def _set_method_field(method, name, value):
    "Set a method model field in method of name and value."
    if name == 'gas':
        method.gas = Gas.objects.get(name=value)
    elif name == 'cup':
        if value is not None:
            _set_cup(method, value)
        else:
            method.cup = None
    elif name == 'insert':
        if value is not None:
            _set_insert(method, value)
        else:
            setattr(method, name, None)
    else:
        setattr(method, name, value)


def _set_param_field(prms, name, value):
    "Set an analysis parameter (non method) field in method of name and value."
    if name == 'gas':
        setattr(prms, name, Gas.objects.get(name=value))
    elif name in ['cup', 'insert']:
        # reuse model if it exists.
        if value is not None:
            insert = Insert()
            merge_model_dict(
                Insert,
                value, {'is_method': False},
                on_set=lambda name, value, insert=insert: setattr(insert, name, value)
            )
            insert.save()
        else:
            insert = None
        setattr(prms, name, insert)
    else:
        setattr(prms, name, value)


def _set_cal_result_field(cal, model, serial, name, value):
    "Set a Calibration, or AnalysisResult1350 field for the model and serial number instrument."
    # pylint: disable=too-many-branches, too-many-statements
    if name == 'instrument':
        setattr(cal, name, Instrument.objects.get(model=value['model'], serial=value['serial']))
    elif name in ['ref_ball', 'cup']:
        if value is not None:
            insert = Insert()
            merge_model_dict(
                Insert,
                value, {'is_method': False},
                on_set=lambda name, value, insert=insert: setattr(insert, name, value)
            )
            insert.save()
        else:
            insert = None
        setattr(cal, name, insert)
    elif name == 'calibration':
        setattr(
            cal, name,
            Calibration.objects.get(
                instrument__model=model, instrument__serial=serial, start_time=value['start_time']
            )
        )
    elif name == 'parameters':
        if cal.id is None or getattr(cal, name) is None:
            prms = AnalysisParameters()
        else:
            prms = getattr(cal, name)
        merge_model_dict(
            AnalysisParameters,
            value, {
                'is_method': False,
                'is_default': False
            },
            on_set=lambda name, value, prms=prms: _set_param_field(prms, name, value)
        )
        prms.save()
        setattr(cal, name, prms)

    elif name.startswith("expansion"):
        if value is None:
            exp = None
        else:
            if isinstance(cal, AnalysisResult1345):
                if cal.id is None or getattr(cal, name) is None:
                    exp = Expansion1345()
                else:
                    exp = getattr(cal, name)
                exp.num = value['num']
                exp.times = value['times']
                exp.volumes = value['volumes']
                exp.p1s = value['p1s']
                exp.p2s = value['p2s']
                exp.excludes = value['excludes']
                exp.temp_anl_block = value['temp_anl_block']
                exp.save()
            else:
                if cal.id is None or getattr(cal, name) is None:
                    exp = Expansion()
                else:
                    exp = getattr(cal, name)
                exp.num = value['num']
                exp.times = value['times']
                exp.p0s = value['p0s']
                exp.p1s = value['p1s']
                exp.p2s = value['p2s']
                exp.excludes = value['excludes']
                exp.temp_anl_block = value['temp_anl_block']
                exp.temp_valve_block = value['temp_valve_block']
                exp.save()
        setattr(cal, name, exp)
    elif name == 'equilibration_set':
        cal.save(update=False)  # make sure the base object is saves when we add equilibration data.
        Equilibration.objects.filter(analysis=cal).delete()
        for equil_dict in value:
            equil = Equilibration(analysis=cal)
            equil.num_points = equil_dict['num_points']
            equil.cycle = equil_dict['cycle']
            equil.which_equil = equil_dict['which_equil']
            equil.times = equil_dict['times']
            equil.pressures = equil_dict['pressures']
            equil.save()
    elif name == "sample_log":
        # if this does not yet have a parameters object, make a default one so it can save, the
        # default values will be overwritten when the rest of the data is filled in.
        if not hasattr(cal, 'parameters'):
            cal.parameters = AnalysisParameters.objects.create(
                is_method=False,
                gas=Gas.objects.first(),
            )
        cal.save(update=False)  # make sure the base object is saved when we add the log data.
        cal.sample_log.all().delete()
        for log in value:
            cal.sample_log.create(**log)
    else:
        setattr(cal, name, value)


_FORMAT_TABLE = {
    "format_date": _format_date,
    "format_equilibration": _format_equilibration,
    "format_equilibration_set": _format_equilibration_set,
    "format_expansion": _format_expansion,
    "format_expansion1345": _format_expansion1345,
    "format_gas": _format_gas,
    "format_insert": _format_insert,
    "format_sample_log": _format_sample_log,
}

REST_API_DATA_1350 = RestAPIData(# pylint: disable=invalid-name
    'api-v1',
    Instrument,
    _INSTRUMENT_FIELDS,
    'analysisresult__id',
    {'analysisresult__state': 'C'},
    methods=[
        UpdateMethod('insert', Insert, _INSERT_FIELDS,
                     key_fields=['name', 'kind'],
                     set_field_fcn=setattr),
        UpdateMethod('method', AnalysisParameters, _METHOD_FIELDS,
                     key_fields=['name'],
                     set_field_fcn=_set_method_field),
    ],
    results=[
        UpdateResult(
            'calibration', Calibration,
            'instrument', {'state': 'C', 'kind': 'C'},
            _CALIBRATION_FIELDS,
            "last_updated", {'state': 'C', 'kind': 'C'}, _set_cal_result_field, None,
            ["start_time"],
        ),
        UpdateResult(
            'result', AnalysisResult1350,
            'instrument', {'state': 'C'},
            _RESULT_FIELDS,
            "last_updated", {'state': 'C'}, _set_cal_result_field, update_results,
            ["start_time"],
        ),
        UpdateResult(
            'result1345', AnalysisResult1345,
            'instrument', {'state': 'C'},
            _RESULT1345_FIELDS,
            "last_updated", {'state': 'C'}, _set_cal_result_field, update_results,
            ["start_time"],
        ),
    ],
    format_table=_FORMAT_TABLE,
)

def get_results_between(model, serial, first_id, last_id):
    '''The view that returns the result formatted for microactive.

    :param model: The model of the instrument
    :param serial: The serial number of the instrument
    :param first_id: The first id to return
    :param last_id: The last id to return
    '''
    inst = get_object_or_404(Instrument, model=model, serial=serial)
    results = {
        'results': [{
            'id': result.id,
            'start_time': int(result.start_time.timestamp() + 0.5)
        } for result in AnalysisResult1350.objects.filter(
                instrument=inst,
                expansion__isnull=False,
                state='C',
                id__gte=first_id,
                id__lte=last_id)],
    }
    return results


def get_microactive_result(model, serial, result_id):
    '''The view that returns the result formatted for microactive.

    :param model: The model of the instrument
    :param serial: The serial number of the instrument
    :param result_id: The id of the result to return
    '''
    inst = get_object_or_404(Instrument, model=model, serial=serial)
    result = get_object_or_404(
        AnalysisResult1350,
        instrument=inst,
        state='C',
        id=result_id,
    )
    return REST_API_DATA_1350.dict_from_fields(result, _MICROACTIVE_RESULT_FIELDS)

# the patterns that the rest api makes available.
urlpatterns = [
    path('api-supported/',
         lambda request: JsonResponse([REST_API_DATA_1350.version], safe=False)),

    # To support MicroActive
    path('api-v1/get_results_between/', function_get_to_json_view(get_results_between)),
    path('api-v1/get_microactive_result/', function_get_to_json_view(get_microactive_result)),
] + REST_API_DATA_1350.urls()
