"""The business logic related to doing an analysis and storing the results. """
import datetime
import gettext
import math
import json

import django
from django.apps import apps
from django.db import IntegrityError, models
from django.utils import timezone
from django.utils.functional import lazy
from mic_cpp import cpp
from mic_python_lib import inst_scripts
from mic_python_lib.models import LogMessage
from mic_python_lib.log_query import post, setup_notify_shown_errors, setup_on_sample_log
from mic_python_lib import mic_network
from mic_python_lib.models import InstrumentScriptRun
from mic_python_lib.quantity_formatter import notify_unit_selections
from mic_python_lib.settings import set_language, udp_set_translate_locale
import mic_python_lib.settings_views as mpy_settings

import i1350

from .models import (
    AnalysisParameters,
    AnalysisResult1345,
    AnalysisResult1350,
    Calibration,
    Equilibration,
    Expansion,
    Gas,
    Insert,
    InsertMeasure,
    Instrument,
    OtherOperation,
    PressureCalibration,
    RTDCalibration,
    Settings,
    copy_model_obj,
)
from .exceptions import NoCalibrationError

_ = lazy(gettext.gettext, str)

# pylint: disable=invalid-name, too-many-lines


def initialize_analysis():
    """Called to initialize the 1350analyis application with the latest calibration."""
    local = Instrument.objects.local()
    default = AnalysisParameters.objects.default()

    # Cancel any in progress operations
    AnalysisResult1350.objects.filter(state='P').update(state='X')
    Calibration.objects.filter(state='P').update(state='X')
    OtherOperation.objects.filter(state='P').update(state='X')
    InstrumentScriptRun.objects.filter(state='P').update(state='X')

    # setup the locale and units.
    settings = Settings.objects.get()
    set_language(settings.locale, settings.time_zone)
    udp_set_translate_locale()
    notify_unit_selections()

    # Initialize the hardware configuration: the 1cm3 has 3
    # latching valves (fill, expansion and vent) and the rest are
    # continuous
    is1cm3 = local.kind == '1'
    is100cm3 = local.kind == 'H'
    if local.is_temp_control:
        temperature = default.temperature
    else:
        temp_field = AnalysisParameters._meta.get_field('temperature')
        temperature = temp_field.get_default()
    i1350.UDPSetConfig(is1cm3, is100cm3, temperature)

    # Send the latest temperature calibration (if any).
    try:
        cal = RTDCalibration.objects.filter(instrument=local, channel='B').latest('time')
        i1350.UDPSetBlockTempCal(cal.slope, cal.raw_resistance, cal.actual_resistance)
    except django.db.utils.OperationalError:
        # table does not exist yet, needed for make migrations.
        pass
    except RTDCalibration.DoesNotExist:
        pass

    if not is1cm3:
        try:
            cal = RTDCalibration.objects.filter(instrument=local, channel='I').latest('time')
            i1350.UDPSetInletTempCal(cal.slope, cal.raw_resistance, cal.actual_resistance)
        except django.db.utils.OperationalError:
            # table does not exist yet, needed for make migrations.
            pass
        except RTDCalibration.DoesNotExist:
            pass

    # Send the latest pressure calibration (if any).
    try:
        cal = PressureCalibration.objects.filter(instrument=local).latest('time')
        i1350.UDPSetPressureCal(cal.slope, cal.voltage, cal.pressure)
    except django.db.utils.OperationalError:
        # Table does not exist yet, needed for make migrations.
        pass
    except PressureCalibration.DoesNotExist:
        pass

    send_idle_purge_info()
    send_purge_flow_info()

    # setup the notification of if error messages are shown.
    setup_notify_shown_errors(i1350.UDPNotifyMessageShown)
    setup_on_sample_log(add_sample_log)


def send_idle_purge_info():
    "Sends the idle purge info to the analysis application."
    settings = Settings.objects.get()
    info = i1350.SIdlePurgeParameters()
    info.fIdlePurgeEnabled = settings.do_idle_purges
    info.uiIdlePurgeDays = settings.days_idle_purges
    lastActivity = 0
    anl = AnalysisResult1350.objects.order_by('start_time').last()
    if anl and anl.start_time.timestamp() > lastActivity:
        lastActivity = anl.start_time.timestamp()
    cal = Calibration.objects.order_by('start_time').last()
    if cal and cal.start_time.timestamp() > lastActivity:
        lastActivity = cal.start_time.timestamp()
    info.lastActivityTime = int(lastActivity)
    i1350.UDPSetIdlePurge(info)


def send_purge_flow_info():
    "Sends the purge flow temperature checkpoint."
    settings = Settings.objects.get()
    i1350.UDPSetPurgeFlowInfo(settings.flow_required_temperature)


def add_sample_log(log):
    "Adds the log information into the sample log if an analysis is in process."
    local = Instrument.objects.local()
    anl = AnalysisResult1350.objects.filter(
        instrument=local, state='P').order_by("-start_time").first()
    if anl is None:
        return
    anl.sample_log.create(datetime=log.time, message=log.message)


def analysis_done(anl_id, state):
    '''Close the analysis'''

    anl = AnalysisResult1350.objects.get(id=anl_id)
    anl.state = state
    anl.end_time = timezone.now()
    anl.review_active = True
    anl.save()


def operation_cancel():
    '''Cancel the operation'''
    i1350.UDPCancelOperation()


def operation_skip():
    '''Skip the operation wait step'''
    i1350.UDPSkipWait()


def operation_resume():
    "Resume the operation"
    i1350.UDPResumeOperation()


def analysis_run_data(anl_id, run, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve):  # pylint: disable=too-many-arguments
    '''Update the expansion data for the analysis'''
    anl = AnalysisResult1350.objects.get(id=anl_id)
    if run == 0:
        anl.expansion = _update_expansion(
            anl.expansion, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve
        )
    elif run == 1:
        anl.expansion_repeat = _update_expansion(
            anl.expansion_repeat, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve
        )
    else:  # pragma: nocover
        assert False
    update_results(anl)
    anl.save()


def analysis_set_equilibration(anl_id, is_p1, run, cycle, num, time, pressures):  # pylint: disable=too-many-arguments,unused-argument
    '''Stores the equilibration data for an analysis'''
    Equilibration.objects.create(
        analysis=AnalysisResult1350.objects.get(id=anl_id),
        cycle=cycle,
        which_equil='1' if is_p1 else '2',
        num_points=num,
        times=time,
        pressures=pressures
    )


def analysis_start(method_id, mass, name, operator):
    '''Starts an analysis with the given parameters'''
    local = Instrument.objects.local()
    method = AnalysisParameters.objects.get(id=method_id)
    try:
        anl = AnalysisResult1350.objects.create(
            instrument=local,
            parameters=AnalysisParameters.objects.copy_method(method_id),
            gas=local.gas,
            mass=mass,
            name=name,
            operator=operator,
            calibration=Calibration.objects.get_current()
        )
    except IntegrityError:
        raise NoCalibrationError(_("Cannot start an analysis. No calibration available."))
    # if the system is not temperature controlled, set the temperature to default.
    if not local.is_temp_control:
        temp_field = AnalysisParameters._meta.get_field('temperature')
        anl.parameters.temperature = temp_field.get_default()
    # if it is a 1cm3 unit, it is always fill reference.
    if local.kind == "1":
        anl.parameters.fill_direction = 'R'
    anl.parameters.save()
    prms = _analysisparameters_from_anl(anl, local.gas)
    i1350.UDPStartAnalysis(prms)


def analysis_operation_status(inst_oper):
    """Return the status dictionary to pass to the json rest api."""
    oper = {
        "canceled": False,
        "id": None,
        "is_active": False,
        "has_results": False,
    }
    try:
        latest = AnalysisResult1350.objects.filter(instrument=Instrument.objects.local()
                                                  ).latest('start_time')
        oper["canceled"] = latest.state == 'X' or latest.state == 'E'
        oper["id"] = latest.id
        oper["review"] = latest.review_active
        oper["has_results"] = bool(latest.results.all().count())
        if latest.state == 'P' and inst_oper == 'ceAnalyze':
            oper["is_active"] = True
    except AnalysisResult1350.DoesNotExist:
        pass
    return oper


def analysis_get_hw_state():
    '''Get hw state from c++ application'''
    return i1350.UDPGetHWState()


def analysis_get_operation_status():
    '''Get operation status from c++ application'''
    return i1350.UDPGetOperationStatus()


def calibration_start(fill_pressure_gauge, ref_ball_id, kind):
    '''Start a calibration with the given parameters'''

    local = Instrument.objects.local()
    if local.is_temp_control:
        collection_temperature = i1350.UDPGetHWState().flTempControlTarget
    else:
        temp_field = AnalysisParameters._meta.get_field('temperature')
        collection_temperature = temp_field.get_default()
    try:
        inst = Instrument.objects.local()
        cal = Calibration.objects.create(
            kind=kind,
            fill_pressure=fill_pressure_gauge,
            collection_temperature=collection_temperature,
            gas=local.gas.name,
            ref_ball=copy_model_obj(Insert, ref_ball_id),
            cup=None,
            instrument=inst,
            instrument_thermal_expansion_um_per_mk=inst.thermal_expansion_um_per_mk,
            check_of=Calibration.objects.get_current() if kind == "H" else None,
        )
    except IntegrityError:
        raise NoCalibrationError(_("Cannot run check without an existing calibration."))
    prms = i1350.SVolumeCalibrationParameters()
    prms.uiCalID = cal.id
    prms.fIsCheck = kind == "H"
    prms.eInsPrompt = i1350.EInsPrompt.ceInsPromptRef
    prms.flFillPressureGauge = fill_pressure_gauge
    prms.szGasName = cal.gas
    prms.flCalibrationTemperatureC = collection_temperature
    i1350.UDPStartVolumeCalibration(prms)


def calibration_done(cal_id, state):
    '''Close the calibration'''
    cal = Calibration.objects.get(id=cal_id)
    cal.state = state
    cal.review_active = True
    if state == 'C':
        _update_cal(cal)

    # on a completed service calibration set the next service date.
    if cal.state == 'C' and cal.kind == 'C':
        local = Instrument.objects.local()
        local.device_service = timezone.now() + datetime.timedelta(days=365)
        local.save()
    cal.save()


def calibration_run_data(cal_id, run, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve):  # pylint: disable=too-many-arguments
    '''Update the expansion data for the analysis'''
    cal = Calibration.objects.get(id=cal_id)
    if run == 0:
        cal.expansion_fs_empty = _update_expansion(
            cal.expansion_fs_empty, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve
        )
    elif run == 1:
        cal.expansion_fr_empty = _update_expansion(
            cal.expansion_fr_empty, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve
        )
    elif run == 2:
        cal.expansion_fs_ref = _update_expansion(
            cal.expansion_fs_ref, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve
        )
    elif run == 3:
        cal.expansion_fr_ref = _update_expansion(
            cal.expansion_fr_ref, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve
        )
    else:  # pragma: nocover
        assert False

    cal.save()


def idle_cancel():
    '''Cancel the idle purges'''
    i1350.UDPCancelIdlePurges()

def calibration_operation_status(inst_oper, kind):
    """Return the status dictionary to pass to the json rest api."""
    oper = {
        "canceled": False,
        "id": None,
        "is_active": False,
        "has_results": False,
    }
    try:
        # if the request is for a user calibration, it may have been automatically rejected, so look
        # for that one as well.
        kind_list = [kind]
        if kind == 'UN':
            kind_list.append('UR')
        latest = Calibration.objects.filter(
            instrument=Instrument.objects.local(), kind__in=kind_list
        ).latest('start_time')
        oper['canceled'] = latest.state == 'X' or latest.state == 'E'
        oper['id'] = latest.id
        oper["review"] = latest.review_active
        oper['has_results'] = not oper['canceled']
        if latest.state == 'P' and inst_oper == 'ceCalibrateVolume':
            oper["is_active"] = True
    except Calibration.DoesNotExist:
        pass
    return oper

def calibration_reset():
    "Reset the calibration to the last factory run one."
    last = Calibration.objects.get_last_factory()
    last.id = None
    last.kind = 'R'
    last.start_time = timezone.now()
    # make copy of the one to one fields.
    last.expansion_fs_empty = Expansion.objects.copy(last.expansion_fs_empty)
    last.expansion_fs_ref = Expansion.objects.copy(last.expansion_fs_ref)
    last.expansion_fr_empty = Expansion.objects.copy(last.expansion_fr_empty)
    last.expansion_fr_ref = Expansion.objects.copy(last.expansion_fr_ref)
    last.save()
    return {}

def calibration_user_save():
    "Mark the user calibration as accepted."
    cal = Calibration.objects.last()
    if cal.kind == "UN":
        cal.kind = "UA"
        cal.save()
    return {}

def insert_measure_start(kind, ref_ball, multivolume_insert, fill_pressure):
    '''Start an insert or cup measurement with the given parameters'''
    local = Instrument.objects.local()
    if local.is_temp_control:
        collection_temperature = i1350.UDPGetHWState().flTempControlTarget
    else:
        temp_field = AnalysisParameters._meta.get_field('temperature')
        collection_temperature = temp_field.get_default()
    try:
        oper = InsertMeasure.objects.create(
            kind=kind,
            fill_pressure=fill_pressure,
            ref_ball=(
                copy_model_obj(Insert, ref_ball) if kind == 'C' and ref_ball else None
            ),
            multivolume_insert=(
                copy_model_obj(Insert, multivolume_insert)
                if kind == 'C' and multivolume_insert else None
            ),
            instrument=local,
            gas=local.gas.name,
            collection_temperature=collection_temperature,
            calibration=Calibration.objects.get_current()
        )
    except IntegrityError:
        raise NoCalibrationError(_("Cannot run measurement without an existing calibration."))

    prms = i1350.SInsertMeasurementParameters()
    prms.uiOperID = oper.id
    prms.flAnalysisTemperatureC = collection_temperature
    prms.flFillPressureGauge = fill_pressure
    if oper.multivolume_insert and oper.ref_ball:
        prms.eInsPrompt = i1350.EInsPrompt.ceInsPromptCupRefIns
        prms.iNumInserts = 2
        sInsert = _insert_to_cpp(oper.multivolume_insert)
        sRefBall = _insert_to_cpp(oper.ref_ball)
        prms.asInsert = [sInsert, sRefBall, sInsert]
    elif oper.multivolume_insert:
        prms.eInsPrompt = i1350.EInsPrompt.ceInsPromptCupIns
        prms.iNumInserts = 1
        sInsert = _insert_to_cpp(oper.multivolume_insert)
        prms.asInsert = [sInsert, sInsert, sInsert]
    elif oper.ref_ball:
        prms.eInsPrompt = i1350.EInsPrompt.ceInsPromptCupRef
        prms.iNumInserts = 1
        sRefBall = _insert_to_cpp(oper.ref_ball)
        prms.asInsert = [sRefBall, sRefBall, sRefBall]
    else:
        if kind == 'C':
            prms.eInsPrompt = i1350.EInsPrompt.ceInsPromptCup
        else:
            prms.eInsPrompt = i1350.EInsPrompt.ceInsPromptIns
        prms.iNumInserts = 0

    prms.szGasName = oper.gas
    prms.sChamberCal = _chamber_cal_to_cpp(oper.calibration)
    i1350.UDPStartInsertMeasurement(prms)


def insert_measure_operation_status(inst_oper):
    """Return the status dictionary to pass to the json rest api."""
    oper = {
        "canceled": False,
        "id": None,
        "is_active": False,
        "has_results": False,
    }
    try:
        latest = InsertMeasure.objects.filter(instrument=Instrument.objects.local()
                                             ).latest('start_time')
        oper['canceled'] = latest.state == 'X' or latest.state == 'E'
        oper['id'] = latest.id
        oper["review"] = latest.review_active
        oper['has_results'] = not oper['canceled']
        if latest.state == 'P' and inst_oper == 'ceInsertMeasurement':
            oper["is_active"] = True
    except InsertMeasure.DoesNotExist:
        pass
    return oper


def insmeas_run_data(oper_id, run, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve):  # pylint: disable=too-many-arguments
    '''Update the expansion data for the measurement'''
    anl = InsertMeasure.objects.get(id=oper_id)
    if run == 0:
        anl.expansion_fs = _update_expansion(
            anl.expansion_fs, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve
        )
    elif run == 1:
        anl.expansion_fr = _update_expansion(
            anl.expansion_fr, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve
        )
    else:  # pragma: nocover
        assert False
    _update_insmeas_results(anl)
    anl.save()


def insmeas_done(oper_id, state):
    '''Close the measurement'''

    anl = InsertMeasure.objects.get(id=oper_id)
    anl.state = state
    anl.review_active = True
    anl.end_time = timezone.now()

    if state == 'C':
        volume = []
        if anl.expansion_fs:
            volume_fs = json.loads(i1350.CalcSampleVolumeJson(get_insmeas_samplevoldata(anl, True)))
            volume.append(volume_fs['ave'])
        if anl.expansion_fr:
            volume_fr = json.loads(
                i1350.CalcSampleVolumeJson(get_insmeas_samplevoldata(anl, False))
            )
            volume.append(volume_fr['ave'])
        if len(volume) > 0:
            ins = Insert.objects.create_insert(anl.kind)
            anl.created_insert = ins
            anl.created_insert.volume = sum(volume) / len(volume)
            anl.created_insert.temperature = anl.collection_temperature
            anl.created_insert.is_measured = True
        anl.created_insert.save()
    anl.save()


def insmeas_redirect(oper_id):
    "Return the url for the insert that was measured."
    anl = InsertMeasure.objects.get(id=oper_id)
    if anl.kind == 'C':
        url = f"/#/settings/objects/cup/{anl.created_insert.id}"
    else:
        url = f"/#/settings/objects/multi-volume/{anl.created_insert.id}"
    return {"url": url}

def fill_instrument_log(model):
    "Add the instrument log as a new page to the report model."
    # nothing in this function needs to be lazy, so this simplifies the c++ extension conversion.
    _ = gettext.gettext

    page = model.AddPage(_("Instrument Log"))
    table = cpp.CReportStdTable(page, "", 2)
    table.SetTitle(0, _("Date"))
    table.SetTitle(1, _("Message"))
    start = timezone.now() - datetime.timedelta(days=90)
    for log in LogMessage.objects.filter(time__gte=start).order_by("-time")[:660]:
        table.AddItem(0, cpp.CTimeDate(log.time).GetBoth())
        table.AddItem(1, log.message)


def get_chambercal(cal):
    '''Return data for a chamber calibration'''
    return _chamber_cal_to_cpp(cal)


def get_hwcal_pressure(cal):
    '''Return the SHWCalPressureResults structure from this cal object'''
    sRet = i1350.SHWCalPressureResults()
    sRet.szGaugeSerial = cal.gauge_sn
    sRet.timeDate = cal.time
    sRet.flPtVoltage = cal.voltage
    sRet.flPtPressure = cal.pressure
    sRet.flSlope = cal.slope
    sRet.flCheckReference = cal.check_reference
    sRet.flCheckReading = cal.check_reading
    return sRet


def get_hwcal_rtd(cal):
    '''Return the SHWCalPressureResults structure from this cal object'''
    sRet = i1350.SHWCalRTDResults()
    if cal.channel == 'B':
        sRet.eChannel = i1350.ETempChannel.ceTempChannelBlock
    else:
        sRet.eChannel = i1350.ETempChannel.ceTempChannelInlet
    sRet.szGaugeSerial = cal.gauge_sn
    sRet.timeDate = cal.time
    sRet.flPtRawResistance = cal.raw_resistance
    sRet.flPtActualResistance = cal.actual_resistance
    sRet.flSlope = cal.slope
    sRet.flCheckReference = cal.check_reference
    sRet.flCheckReading = cal.check_reading
    return sRet


def get_samplevoldata(anl, repeat=False):
    '''Return data for calculating sample volume'''
    if isinstance(anl, AnalysisResult1350):
        sData = i1350.SSampleVolData()
        sData.szGas = anl.gas.name
        if anl.parameters.fill_direction == 'R' or anl.instrument.kind == '1':
            sData.eFillDirection = i1350.EFillDirection.ceFillReference
        else:
            sData.eFillDirection = i1350.EFillDirection.ceFillSample
        sData.sChamberCal = _chamber_cal_to_cpp(anl.calibration)
        if anl.parameters.insert and anl.parameters.cup:
            sData.vsInserts = [
                _insert_to_cpp(anl.parameters.insert),
                _insert_to_cpp(anl.parameters.cup)
            ]
            sData.vszInsertNames = [
                anl.parameters.insert.name,
                anl.parameters.cup.name
            ]
        elif anl.parameters.insert:
            sData.vsInserts = [_insert_to_cpp(anl.parameters.insert)]
            sData.vszInsertNames = [anl.parameters.insert.name]
        elif anl.parameters.cup:
            sData.vsInserts = [_insert_to_cpp(anl.parameters.cup)]
            sData.vszInsertNames = [anl.parameters.cup.name]
        else:
            sData.vsInserts = []
            sData.vszInsertNames = []
        sData.sExpansion = _expansion_to_cpp(anl.expansion_repeat if repeat else anl.expansion)
        return sData

    if isinstance(anl, AnalysisResult1345):
        sData = i1350.SSampleVolDataVolume()
        if repeat:
            sData.vsExpansion = [
                i1350.SExpansionPointVolume(
                    anl.expansion_repeat.volumes[i],
                    (
                        anl.expansion_repeat.temp_anl_block[i]
                        if anl.expansion_repeat.temp_anl_block_binary else anl.temperature
                    ),
                    anl.expansion_repeat.p1s[i] if anl.expansion_repeat.p1s_binary else 0.0,
                    anl.expansion_repeat.p2s[i] if anl.expansion_repeat.p2s_binary else 0.0,
                    anl.expansion_repeat.times[i] if anl.expansion_repeat.times_binary else 0,
                    anl.expansion_repeat.excludes[i],
                ) for i in range(anl.expansion.num)
            ]
        else:
            sData.vsExpansion = [
                i1350.SExpansionPointVolume(
                    anl.expansion.volumes[i],
                    (
                        anl.expansion.temp_anl_block[i]
                        if anl.expansion.temp_anl_block_binary else anl.temperature
                    ),
                    anl.expansion.p1s[i] if anl.expansion.p1s_binary else 0.0,
                    anl.expansion.p2s[i] if anl.expansion.p2s_binary else 0.0,
                    anl.expansion.times[i] if anl.expansion.times_binary else 0,
                    anl.expansion.excludes[i],
                ) for i in range(anl.expansion.num)
            ]
        return sData

    raise ValueError('anl must be a concrete AnalysisResult object.')


def get_insmeas_samplevoldata(anl, fs):
    '''Return data for calculating sample volume'''
    sData = i1350.SSampleVolData()
    sData.szGas = anl.gas
    sData.sChamberCal = _chamber_cal_to_cpp(anl.calibration)
    if anl.multivolume_insert and anl.ref_ball:
        sData.vsInserts = [_insert_to_cpp(anl.multivolume_insert), _insert_to_cpp(anl.ref_ball)]
        sData.vszInsertNames = [anl.multivolume_insert.name, anl.ref_ball.name]
    elif anl.multivolume_insert:
        sData.vsInserts = [_insert_to_cpp(anl.multivolume_insert)]
        sData.vszInsertNames = [anl.multivolume_insert.name]
    elif anl.ref_ball:
        sData.vsInserts = [_insert_to_cpp(anl.ref_ball)]
        sData.vszInsertNames = [anl.ref_ball.name]
    else:
        sData.vsInserts = []
        sData.vszInsertNames = []
    if fs:
        sData.eFillDirection = i1350.EFillDirection.ceFillSample
        sData.sExpansion = _expansion_to_cpp(anl.expansion_fs)
    else:
        sData.eFillDirection = i1350.EFillDirection.ceFillReference
        sData.sExpansion = _expansion_to_cpp(anl.expansion_fr)
    return sData


def get_samplecaldata(cal):
    '''Return data for calculating calibration volume'''
    sData = i1350.SVolumeCalData()
    sData.sReference = _insert_to_cpp(cal.ref_ball)
    sData.dChamberThermalExpansion = cal.instrument_thermal_expansion
    sData.sExpansionFSEmpty = _expansion_to_cpp(cal.expansion_fs_empty)
    sData.sExpansionFSRef = _expansion_to_cpp(cal.expansion_fs_ref)
    sData.sExpansionFREmpty = _expansion_to_cpp(cal.expansion_fr_empty)
    sData.sExpansionFRRef = _expansion_to_cpp(cal.expansion_fr_ref)
    sData.szGas = cal.gas
    sData.flTemperature = cal.collection_temperature
    sData.fHasCup = bool(cal.cup)
    if sData.fHasCup:
        sData.sCupData = _insert_to_cpp(cal.cup)
    return sData


def notify_analysis_seen():
    '''Tell the instrument that the analysis review period is over.'''
    i1350.UDPNotifyAnalysisSeen()


def hwcal_start(which):
    """Start a hardware calibration.


    """
    local = Instrument.objects.local()
    if which == 'pressure':
        oper = OtherOperation(instrument=local, kind='P')
        oper.save()
        i1350.UDPHWPressCalibration(oper.id)
    elif which == 'block':
        oper = OtherOperation(instrument=local, kind='B')
        oper.save()
        i1350.UDPHWTempCalibration(oper.id, i1350.ETempChannel.ceTempChannelBlock)
    elif which == 'inlet':
        oper = OtherOperation(instrument=local, kind='I')
        oper.save()
        i1350.UDPHWTempCalibration(oper.id, i1350.ETempChannel.ceTempChannelInlet)
    else:
        assert False


def hwcal_results(which):
    """Start a hardware calibration.

    :param which: the name of the calibration to start: 'pressure', 'block', or 'inlet'
    """
    result = {}
    if which == 'pressure':
        sResult = get_hwcal_pressure(PressureCalibration.objects.last())
        result = json.loads(i1350.CalPressureResultsPage(sResult))
    elif which == 'block':
        sResult = get_hwcal_rtd(RTDCalibration.objects.filter(channel='B').last())
        result = json.loads(i1350.CalRTDResultsPage(sResult))
    elif which == 'inlet':
        sResult = get_hwcal_rtd(RTDCalibration.objects.filter(channel='I').last())
        result = json.loads(i1350.CalRTDResultsPage(sResult))
    else:
        assert False
    return result

def hwcal_save_pressure(# pylint: disable=too-many-arguments
        slope, voltage, pressure, check_reference, check_reading, gauge_sn):
    """Save the pressure calibration into the database."""
    cal = PressureCalibration(
        instrument=Instrument.objects.local(),
        slope=slope,
        voltage=voltage,
        pressure=pressure,
        check_reference=check_reference,
        check_reading=check_reading,
        gauge_sn=gauge_sn,
    )
    cal.save()
    i1350.UDPSetPressureCal(cal.slope, cal.voltage, cal.pressure)


def hwcal_save_rtd(  # pylint: disable=too-many-arguments
    channel,
    slope,
    raw_resistance,
    actual_resistance,
    check_reference,
    check_reading,
    gauge_sn
):
    """Save the temperature calibration into the database."""
    cal = RTDCalibration(
        instrument=Instrument.objects.local(),
        channel=channel,
        slope=slope,
        raw_resistance=raw_resistance,
        actual_resistance=actual_resistance,
        check_reference=check_reference,
        check_reading=check_reading,
        gauge_sn=gauge_sn,
    )
    cal.save()
    if channel == 'B':
        i1350.UDPSetBlockTempCal(cal.slope, cal.raw_resistance, cal.actual_resistance)
    elif channel == 'I':
        i1350.UDPSetInletTempCal(cal.slope, cal.raw_resistance, cal.actual_resistance)


def other_operation_status(inst_oper, kind):
    """Return the status dictionary to pass to the json rest api."""
    kind_oper = ''
    if kind == 'P':
        kind_oper = 'ceHWCalibratePressure'
    if kind == 'B':
        kind_oper = 'ceHWCalibrateBlockTemp'
    if kind == 'I':
        kind_oper = 'ceHWCalibrateInletTemp'
    if kind == 'U':
        kind_oper = 'ceOtherOperation'
    if kind == 'S':
        kind_oper = 'ceOtherOperation'
    if kind == 'Y':
        kind_oper = 'ceScriptOperation'

    oper = {
        "canceled": False,
        "id": None,
        "is_active": False,
        "has_results": True,  # go to review page unconditionally
        "is_terminating": False,
        "is_suspended": False,
    }
    try:
        latest = OtherOperation.objects.filter(
            instrument=Instrument.objects.local(), kind=kind
        ).latest('start_time')
        oper["canceled"] = latest.state == 'X' or latest.state == 'E'
        oper["id"] = latest.id
        oper["number"] = latest.number
        oper["review"] = latest.review_active
        if latest.state == 'P' and inst_oper == kind_oper:
            oper["is_active"] = True
    except OtherOperation.DoesNotExist:
        pass
    return oper


def other_operation_start(kind, number):
    "Start the other operation."
    local = Instrument.objects.local()
    oper = OtherOperation(instrument=local, kind=kind, number=number)
    oper.save()
    i1350.UDPOtherOperation(oper.id, number)


def script_operation_start(name: str, operator: str):
    "Start the script operation."
    local = Instrument.objects.local()
    active = InstrumentScriptRun.objects.create(name=name, operator=operator, state='P')
    oper = OtherOperation.objects.create(instrument=local, kind='Y', script_run=active)
    active.user_data_context = oper.id
    active.save()
    i1350.UDPScriptOperation(oper.id, inst_scripts.script_from_name(name))


def other_operation_state():
    "Get the other operation state."
    state = i1350.UDPGetHWState()
    state_dict = {name: getattr(state, name) for name in dir(state) if not name.startswith('__')}
    # Augment the hardware state with the operation status string and termination state.
    state_dict['szOperStatus'] = i1350.UDPGetOperationStatus().szStatus
    oper = i1350.UDPGetOperationStatus()
    state_dict['fIsTerminating'] = oper.fIsTerminating
    state_dict['fIsWaitingSkip'] = oper.fIsWaitingSkip
    state_dict['fIsSuspended'] = oper.fIsSuspended
    return state_dict


def other_operation(action, **kwargs):
    "Handle all the actions for the other operations."
    actions = {
        "start": other_operation_start,
        "state": other_operation_state,
        "cancel": operation_cancel,
        "resume": operation_resume,
        "skip": operation_skip,
    }
    if action in actions:
        return actions[action](**kwargs)
    return None


def other_operation_done(operID, state):
    '''Close the other operation'''
    oper = OtherOperation.objects.get(id=operID)
    oper.state = state
    oper.review_active = True
    # If it was an instrument script, set that to the same state.
    if oper.script_run:
        oper.script_run.state = state
        # If the user canceled during a prompt it goes through the error path.  But, the stderr
        # output is an indication that it should have been a cancel.
        if oper.script_run.state == 'E' and oper.script_run.error_output == "User cancel\n":
            oper.script_run.state = 'X'
            oper.script_run.error_output = ''
            oper.state = 'X'
        oper.script_run.save()
    # on a completed service calibration set the next service date.
    if oper.state == 'C' and oper.kind in "PBIS":
        local = Instrument.objects.local()
        local.device_service = timezone.now() + datetime.timedelta(days=365)
        local.save()
    oper.save()


def prompt_for_cal_check():
    "Returns true if the user should be prompted for a calibration check"
    inactive_days = datetime.timedelta(days=15) # if this is updated, also update the ui prompt.
    now = timezone.now()
    last_date = now - inactive_days
    local = Instrument.objects.local()
    cal = Calibration.objects.filter(instrument=local).order_by('start_time').last()
    if cal and cal.start_time > last_date:
        last_date = cal.start_time
    anl = AnalysisResult1350.objects.filter(instrument=local).order_by('start_time').last()
    if anl and anl.start_time > last_date:
        last_date = anl.start_time
    return now - inactive_days >= last_date


def calc_foam_pyc_a(anl):
    """@returns the foampyc method a calculation object filled in for this anl"""
    calc = _calc_foam_pyc_common(anl, i1350.CCalcFoamPycMethodA())
    if isinstance(anl, AnalysisResult1350):
        prms = anl.parameters
        if prms.compute_volume == 'E':
            calc.SetEnteredGeometryArea(prms.active_area)
        else:
            assert prms.compute_volume == 'C'
            if prms.shape == 'U':
                calc.SetCubeSkins(prms.cube_number_of_skins)
            elif prms.shape == 'Y':
                calc.SetCylinderSkins(prms.cyl_number_of_skins)
            else:
                assert prms.shape == 'R'
                calc.SetRectangleSkins(
                    prms.rect_number_of_large_skins, prms.rect_number_of_small_skins,
                    prms.rect_number_of_other_skins
                )
        if prms.measure_by == 'L':
            calc.SetChordLength(prms.chord_length)
        else:
            assert prms.measure_by == 'D'
            calc.SetCellDiameter(prms.cell_diameter)
    elif isinstance(anl, AnalysisResult1345):
        calc.SetEnteredGeometryArea(anl.active_area)
        if anl.measure_by == 'L':
            calc.SetChordLength(anl.chord_length)
        else:
            assert anl.measure_by == 'D'
            calc.SetCellDiameter(anl.cell_diameter)
    else:
        raise ValueError('anl must be a concrete AnalysisResult object.')

    return calc


def calc_foam_pyc_b(anl):
    """@returns the foampyc method b calculation object filled in for this anl"""
    return _calc_foam_pyc_common(anl, i1350.CCalcFoamPycMethodB())


def calc_foam_pyc_c(anl):
    """@returns the foampyc method c calculation object filled in for this anl"""
    return _calc_foam_pyc_common(anl, i1350.CCalcFoamPycMethodC())


def calc_foam_pyc_d(anl):
    """@returns the foampyc method d calculation object filled in for this anl"""
    return _calc_foam_pyc_common(anl, i1350.CCalcFoamPycMethodD())


def calc_foam_pyc_e(anl):
    """@returns the foampyc method e calculation object filled in for this anl"""
    return _calc_foam_pyc_common(anl, i1350.CCalcFoamPycMethodE())


def method_options_from_anl(anl):
    """Returns the SMethodTable object with the information from the anl."""
    method = i1350.SMethodTable()
    method.szOperator = anl.operator
    method.szMethodName = anl.parameters.name
    method.szGasName = anl.gas.name
    method.szInsert = anl.parameters.insert.name if anl.parameters.insert else ""
    method.szCup = anl.parameters.cup.name if anl.parameters.cup else ""
    method.sAnalysisParameters = _analysisparameters_from_anl(anl, anl.gas)
    method.dResinDensity = anl.parameters.resin_density
    method.dActiveArea = anl.parameters.active_area
    if anl.parameters.measure_by == 'L':
        method.eMeasureBy = i1350.EMeasureBy.ceChordLength
    else:
        method.measure_by = i1350.EMeasureBy.ceCellDiameter
    method.dChordLength = anl.parameters.chord_length
    method.dCellDiameter = anl.parameters.cell_diameter
    if anl.parameters.compute_volume == 'C':
        method.eComputeVolume = i1350.EComputeVolume.ceComputed
    else:
        method.eComputeVolume = i1350.EComputeVolume.ceUserEntered
    if anl.parameters.shape == 'U':
        method.eShape = i1350.EShape.ceCube
    elif anl.parameters.shape == 'Y':
        method.eShape = i1350.EShape.ceCylinder
    else:
        method.eShape = i1350.EShape.ceRectangular
    method.uiCubeNumberSkins = anl.parameters.cube_number_of_skins
    method.uiCylNumberSkins = anl.parameters.cyl_number_of_skins
    method.uiRectNumberLargeSkins = anl.parameters.rect_number_of_large_skins
    method.uiRectNumberSmallSkins = anl.parameters.rect_number_of_small_skins
    method.uiRectNumberOtherSkins = anl.parameters.rect_number_of_other_skins

    method.vsLogItem = [
        i1350.SSampleLogItem(log.datetime, log.message) for log in anl.sample_log.all()
    ]
    return method


def update_method1345_results(accupyc, anl):
    "Update the method parameters in the accupyc results with the data from anl"
    assert isinstance(accupyc, i1350.SAccuPycResultVolume)
    assert isinstance(anl, AnalysisResult1345)
    if anl.technique == 'P':
        accupyc.eAnalysisMethod = i1350.EAnalysisType.ceStandardAnalysis
    if anl.technique == 'A':
        accupyc.eAnalysisMethod = i1350.EAnalysisType.ceMethodA
    if anl.technique == 'B':
        accupyc.eAnalysisMethod = i1350.EAnalysisType.ceMethodB
    if anl.technique == 'C':
        accupyc.eAnalysisMethod = i1350.EAnalysisType.ceMethodC
    if anl.technique == 'D':
        accupyc.eAnalysisMethod = i1350.EAnalysisType.ceMethodD
    if anl.technique == 'E':
        accupyc.eAnalysisMethod = i1350.EAnalysisType.ceMethodE
    accupyc.iNumberOfPurges = anl.num_purges
    if anl.endby == 'E':
        accupyc.eP1_P2_EndBy = i1350.EEndEquil.ceEndByEquilibrate
    else:
        accupyc.eP1_P2_EndBy = i1350.EEndEquil.ceEndByFixedInterval
    accupyc.iIntervalTime = anl.interval_time
    accupyc.flEquilibrationRate = anl.equilibration_rate
    accupyc.dResinDensity = anl.resin_density
    accupyc.dGeometricVolume = anl.geometric_volume
    accupyc.dActiveArea = anl.active_area
    if anl.measure_by == 'L':
        accupyc.eMeasureBy = i1350.EMeasureBy.ceChordLength
    else:
        accupyc.measure_by = i1350.EMeasureBy.ceCellDiameter
    accupyc.dChordLength = anl.chord_length
    accupyc.dCellDiameter = anl.cell_diameter


def accu_or_foam_results_from_anl(anl):
    """Returns the SAccuPycResult object with the information from the anl."""
    # pylint: disable=too-many-branches
    if isinstance(anl, AnalysisResult1350):
        accupyc = i1350.SAccuPycResult()
        accupyc.sMethod = method_options_from_anl(anl)
    elif isinstance(anl, AnalysisResult1345):
        accupyc = i1350.SAccuPycResultVolume()
        update_method1345_results(accupyc, anl)
    else:
        raise ValueError('anl must be a concrete AnalysisResult object.')

    if anl.technique == 'P':
        accupyc.eType = i1350.EAnalysisType.ceStandardAnalysis
    elif anl.technique == 'A':
        accupyc.eType = i1350.EAnalysisType.ceMethodA
        accupyc.sCalcFoamA = calc_foam_pyc_a(anl)
        if isinstance(anl, AnalysisResult1350):
            accupyc.sMethod.dGeometricVolume = accupyc.sCalcFoamA.GeometricVolume()
    elif anl.technique == 'B':
        accupyc.eType = i1350.EAnalysisType.ceMethodB
        accupyc.sCalcFoamB = calc_foam_pyc_b(anl)
        if isinstance(anl, AnalysisResult1350):
            accupyc.sMethod.dGeometricVolume = accupyc.sCalcFoamB.GeometricVolume()
    elif anl.technique == 'C':
        accupyc.eType = i1350.EAnalysisType.ceMethodC
        accupyc.sCalcFoamC = calc_foam_pyc_c(anl)
        if isinstance(anl, AnalysisResult1350):
            accupyc.sMethod.dGeometricVolume = accupyc.sCalcFoamC.GeometricVolume()
    elif anl.technique == 'D':
        accupyc.eType = i1350.EAnalysisType.ceMethodD
        accupyc.sCalcFoamD = calc_foam_pyc_d(anl)
        if isinstance(anl, AnalysisResult1350):
            accupyc.sMethod.dGeometricVolume = accupyc.sCalcFoamD.GeometricVolume()
    elif anl.technique == 'E':
        accupyc.eType = i1350.EAnalysisType.ceMethodE
        accupyc.sCalcFoamE = calc_foam_pyc_e(anl)
        if isinstance(anl, AnalysisResult1350):
            accupyc.sMethod.dGeometricVolume = accupyc.sCalcFoamE.GeometricVolume()
    else:
        return None

    accupyc.sVolData = get_samplevoldata(anl)
    if anl.technique == 'B':
        accupyc.sRecutVolData = get_samplevoldata(anl, repeat=True)
    accupyc.dMass = anl.mass
    accupyc.sHeader = _resultheader_from_anl(anl)
    return accupyc


def accu_or_foam_results_from_anl_with_equil(anl):
    """
    Returns the SAccuPycResult object with the information from the anl and with equilibration data
    filled in if available.
    """

    def get_equil_data(equil):
        data = i1350.SEquilibrationData()
        data.iCycle = equil.cycle
        data.fIsP1 = equil.which_equil == '1'
        pressures = equil.pressures
        times = equil.times
        data.vsEquilPoints = [
            i1350.SEquilibrationPoint(pressures[i], times[i]) for i in range(equil.num_points)
        ]
        return data

    results = accu_or_foam_results_from_anl(anl)
    if isinstance(anl, AnalysisResult1350):
        results.sVolData.vsEquilData = [
            get_equil_data(equil) for equil in anl.equilibration_set.all()
        ]
    return results


def update_results(anl):
    "Update the results summary info for the analysis result."
    if anl.technique == 'P':
        _update_accupyc_results(anl)
    elif anl.technique == 'A':
        _update_foampyc_a_results(anl)
    elif anl.technique == 'B':
        _update_foampyc_b_results(anl)
    elif anl.technique == 'C':
        _update_foampyc_c_results(anl)
    elif anl.technique == 'D':
        _update_foampyc_d_results(anl)
    elif anl.technique == 'E':
        _update_foampyc_e_results(anl)


def set_guided_setup_page(page):
    "Set the guided setup page, and if exiting, reset the warranty and service date."
    if page < 0:
        local = Instrument.objects.local()
        local.device_service = timezone.now() + datetime.timedelta(days=365)
        local.device_warranty = timezone.now() + datetime.timedelta(days=365)
        local.save()

    mpy_settings.set_guided_setup_page(page)


def reset_to_pristine_defaults(app_name):
    """Reset unit to pristine defaults when coming back to final."""
    from django.conf import settings
    import os
    i1350.UDPSetTempControl(0.0, 0)
    mic_network.reset_to_pristine_defaults()
    os.unlink(settings.DATABASES['default']['NAME'])
    os.system("/usr/bin/manage.py migrate")
    os.system(f"reboot")


def reset_to_factory_defaults():  # pylint: disable=too-many-statements, too-many-locals, too-many-branches
    """Reset unit to factory defaults after final assembly"""

    # validate that there are at least one of each calibration, and exactly one cup.
    allow_reset = True
    error = _("Cannot reset to factory defaults:")
    if len(Calibration.objects.filter(kind='C', state='C')) == 0:
        post(error + ' ' + str(_("There is no volume calibration.")))
        allow_reset = False
    if len(RTDCalibration.objects.filter(channel='B')) == 0:
        post(error + ' ' + str(_("There is no block temperature calibration.")))
        allow_reset = False
    if len(RTDCalibration.objects.filter(channel='I')) == 0:
        post(error + ' ' + str(_("There is no valve manifold temperature calibration.")))
        allow_reset = False
    if len(PressureCalibration.objects.filter()) == 0:
        post(error + ' ' + str(_("There is no pressure calibration.")))
        allow_reset = False
    if len(Insert.objects.filter(is_method=True, kind="C")) != 1:
        post(error + ' ' + str(_("There is not exactly one cup.")))
        allow_reset = False
    if len(Insert.objects.filter(is_method=True, kind="R")) != 1:
        post(error + ' ' + str(_("There is not exactly one reference volume.")))
        allow_reset = False
    if not allow_reset:
        return {'reset': False}

    # delete all objects of all models defined in the library that have not been handled yet.
    handled = [
        Gas, Settings, Instrument, RTDCalibration, PressureCalibration, Insert, Expansion,
        Calibration
    ]
    for model in apps.get_app_config('pyc_sample').get_models():
        if model in handled:
            continue
        model.objects.all().delete()

    # Update the default settings object.  Do this in a generic way so that if we add fields, this
    # will not have to change.  Note that this means that all fields must have defaults, or they
    # won't be reset by this operation.  This is less likely to be a problem with future
    # maintenance than the simple implementation that would not reset any new members.
    settings = Settings.objects.get()
    for field in Settings._meta.get_fields():
        if field.default != models.fields.NOT_PROVIDED:
            default = field.default() if callable(field.default) else field.default
            setattr(settings, field.name, default)
        elif field.name != 'id' and not field.name.endswith('_ptr'):
            raise RuntimeError(f'Settings model field: {field.name} does not have default.')
    # override the default for this.  The default is setup for PCP, but post PCP we would like to
    # have this displayed.
    settings.display_wizard = True
    settings.save()

    # delete all data except the last calibrations and support.
    local = Instrument.objects.local()
    local.device_service = timezone.now() + datetime.timedelta(days=365)
    local.device_warranty = timezone.now() + datetime.timedelta(days=365)
    local.save()

    Instrument.objects.exclude(pk__in=[local.pk]).delete()

    last = [
        RTDCalibration.objects.filter(channel='B').last(),
        RTDCalibration.objects.filter(channel='I').last()
    ]
    last = [l.pk for l in last if l is not None]
    RTDCalibration.objects.exclude(pk__in=last).delete()

    last = [PressureCalibration.objects.last()]
    last = [l.pk for l in last if l is not None]
    PressureCalibration.objects.exclude(pk__in=last).delete()

    keep_ins = []
    keep_expansion = []
    cal = Calibration.objects.get_last_factory()
    if cal is not None:
        Calibration.objects.exclude(pk__in=[cal.pk]).delete()

        keep_ins = [cal.cup, cal.ref_ball]
        keep_ins = [k.pk for k in keep_ins if k is not None]
        keep_expansion = [
            cal.expansion_fs_empty,
            cal.expansion_fs_ref,
            cal.expansion_fr_empty,
            cal.expansion_fr_ref,
        ]
        keep_expansion = [k.pk for k in keep_expansion if k is not None]
    else:
        Calibration.objects.all().delete()

    Insert.objects.exclude(pk__in=keep_ins).exclude(
        is_method=True, kind='C'
    ).exclude(
        is_method=True, kind='R'
    ).delete()

    Expansion.objects.exclude(pk__in=keep_expansion).delete()

    # Add the analysis parameters.
    from .default_methods import DEFAULT_1CM3_METHODS, DEFAULT_10CM3_METHODS, DEFAULT_100CM3_METHODS  # pylint: disable=import-outside-toplevel
    methods = DEFAULT_10CM3_METHODS
    if local.kind == '1':
        methods = DEFAULT_1CM3_METHODS
    elif local.kind == 'H':
        methods = DEFAULT_100CM3_METHODS
    for meth in methods:
        AnalysisParameters.objects.create(**meth)

    # Make sure the one and only cup is not shared, and is part of every method.
    cup = Insert.objects.get(is_method=True, kind="C")
    cup.is_shared = False
    cup.save()
    for meth in AnalysisParameters.objects.filter(is_method=True):
        meth.cup = cup
        meth.save()

    # cleanup the mic_python_lib models as well.
    mpy_settings.reset_to_factory_defaults()
    return {'reset': True}


def log_field_changes(anl, org, new, changed_fields):
    "Add sample log messages for all the fields that have changed."
    _ = gettext.gettext
    for field_name in changed_fields:
        field = type(new)._meta.get_field(field_name)
        display_name = field.verbose_name
        org_val = getattr(org, field_name)
        new_val = getattr(new, field_name)
        if hasattr(field, 'fmt'):
            org_val = field.fmt(org_val).ValueAndUnits()
            new_val = field.fmt(new_val).ValueAndUnits()
        else:
            org_val = f'"{org_val}"'
            new_val = f'"{new_val}"'

        message = _("{0} changed from {1} to {2}").format(display_name, org_val, new_val)
        anl.sample_log.create(message=message)



def save_diagnostics():
    "zip up the diagnostics files and save them to the USB drive."
    local = Instrument.objects.local()
    zipname = "diagnostics-" + local.model + "-" + local.serial + ".zip"
    mic_network.zip_dir_to_usb("/var/log/mic-instrument/", zipname)
    return {}

# ==================================================================================================
def _update_expansion(exp, cycles, time, p0, p1, p2, exclude, temp_anl, temp_valve):  # pylint: disable=too-many-arguments
    if exp is None:
        exp = Expansion()
    exp.num = cycles
    exp.times = time
    exp.p0s = p0
    exp.p1s = p1
    exp.p2s = p2
    exp.excludes = exclude
    exp.temp_anl_block = temp_anl
    exp.temp_valve_block = temp_valve
    exp.save()
    return exp


def _update_accupyc_results(anl):
    sample_vol_data = get_samplevoldata(anl)
    volume = json.loads(i1350.CalcSampleVolumeJson(sample_vol_data))
    density = json.loads(i1350.CalcSampleDensityJson(sample_vol_data, anl.mass))
    last_density = json.loads(i1350.CalcLastSampleDensityJson(sample_vol_data, anl.mass))

    # Make sure the strings are stored in English and that the strings do show up in the translation
    # tables.
    _("Volume")
    _("Density")
    _("Last Density")

    anl.results.update_or_create(
        result_name="volume",
        result_rank='S',
        defaults={
            'analysis_technique': "Gas Pycnometry",
            'result_title': "Volume",
            'value': volume['ave'],
            'std_dev': volume['std_dev'],
            'units_json': json.dumps(volume['units']),
            'format_json': json.dumps(volume['format']),
        }
    )
    anl.results.update_or_create(
        result_name="absolute_density",
        result_rank='P',
        defaults={
            'analysis_technique': "Gas Pycnometry",
            'result_title': "Density",
            'value': density['ave'],
            'std_dev': density['std_dev'],
            'units_json': json.dumps(density['units']),
            'format_json': json.dumps(density['format']),
        }
    )
    anl.results.update_or_create(
        result_name="last_density",
        result_rank='R',
        defaults={
            'analysis_technique': "Gas Pycnometry",
            'result_title': "Last Density",
            'value': last_density['value'],
            'units_json': json.dumps(last_density['units']),
            'format_json': json.dumps(last_density['format']),
        }
    )


def _update_foampyc_a_results(anl):
    data = get_samplevoldata(anl)
    calc_accu = calc_foam_pyc_a(anl)
    value = json.loads(i1350.CalcFoamPycAOpenCellPercentJson(data, calc_accu, anl.mass))

    # Make sure the strings are stored in English and that the strings do show up in the translation
    # tables.
    _("Open Cell Percent")
    anl.results.update_or_create(
        result_name="open_cell_percent",
        result_rank='P',
        defaults={
            'analysis_technique': anl.get_technique_display(),
            'result_title': "Open Cell Percent",
            'value': value['value'],
            'units_json': json.dumps(value['units']),
            'format_json': json.dumps(value['format']),
        }
    )


def _update_foampyc_b_results(anl):
    data = get_samplevoldata(anl)
    data_recut = get_samplevoldata(anl, repeat=True)
    calc_accu = calc_foam_pyc_b(anl)
    value = json.loads(i1350.CalcFoamPycBOpenCellPercentJson(data, data_recut, calc_accu, anl.mass))

    # Make sure the strings are stored in English and that the strings do show up in the translation
    # tables.
    _("Open Cell Percent")
    anl.results.update_or_create(
        result_name="open_cell_percent",
        result_rank='P',
        defaults={
            'analysis_technique': anl.get_technique_display(),
            'result_title': "Open Cell Percent",
            'value': value['value'],
            'units_json': json.dumps(value['units']),
            'format_json': json.dumps(value['format']),
        }
    )


def _update_foampyc_c_results(anl):
    data = get_samplevoldata(anl)
    calc_accu = calc_foam_pyc_c(anl)
    value = json.loads(i1350.CalcFoamPycCOpenCellPercentJson(data, calc_accu, anl.mass))

    # Make sure the strings are stored in English and that the strings do show up in the translation
    # tables.
    _("Open Cell Percent")
    anl.results.update_or_create(
        result_name="open_cell_percent",
        result_rank='P',
        defaults={
            'analysis_technique': anl.get_technique_display(),
            'result_title': "Open Cell Percent",
            'value': value['value'],
            'units_json': json.dumps(value['units']),
            'format_json': json.dumps(value['format']),
        }
    )


def _update_foampyc_d_results(anl):
    data = get_samplevoldata(anl)
    calc_accu = calc_foam_pyc_d(anl)
    value = json.loads(i1350.CalcFoamPycDP1PercentChangeJson(data, calc_accu, anl.mass))

    # Make sure the strings are stored in English and that the strings do show up in the translation
    # tables.
    _("P1 Percent Change")
    anl.results.update_or_create(
        result_name="p1_percent_change",
        result_rank='P',
        defaults={
            'analysis_technique': anl.get_technique_display(),
            'result_title': "P1 Percent Change",
            'value': value['value'],
            'units_json': json.dumps(value['units']),
            'format_json': json.dumps(value['format']),
        }
    )


def _update_foampyc_e_results(anl):
    data = get_samplevoldata(anl)
    calc_accu = calc_foam_pyc_e(anl)
    value = json.loads(i1350.CalcFoamPycEPercentChangeJson(data, calc_accu, anl.mass))

    # Make sure the strings are stored in English and that the strings do show up in the translation
    # tables.
    _("Percent Change")
    anl.results.update_or_create(
        result_name="percent_change",
        result_rank='P',
        defaults={
            'analysis_technique': anl.get_technique_display(),
            'result_title': "Percent Change",
            'value': value['value'],
            'units_json': json.dumps(value['units']),
            'format_json': json.dumps(value['format']),
        }
    )


def _update_insmeas_results(anl):

    # Make sure the strings are stored in English.
    if anl.expansion_fs:
        cup_vol_data_fs = get_insmeas_samplevoldata(anl, True)
        volume_fs = json.loads(i1350.CalcSampleVolumeJson(cup_vol_data_fs))
        anl.results.update_or_create(
            result_name="volume_fs",
            result_rank='P',
            defaults={
                'analysis_technique': "Gas Pycnometry",
                'result_title': _("Volume Fill Sample"),
                'value': volume_fs['ave'],
                'std_dev': volume_fs['std_dev'],
                'units_json': json.dumps(volume_fs['units']),
                'format_json': json.dumps(volume_fs['format']),
            }
        )
    if anl.expansion_fr:
        cup_vol_data_fr = get_insmeas_samplevoldata(anl, False)
        volume_fr = json.loads(i1350.CalcSampleVolumeJson(cup_vol_data_fr))
        anl.results.update_or_create(
            result_name="volume_fr",
            result_rank='S',
            defaults={
                'analysis_technique': "Gas Pycnometry",
                'result_title': _("Volume Fill Reference"),
                'value': volume_fr['ave'],
                'std_dev': volume_fr['std_dev'],
                'units_json': json.dumps(volume_fr['units']),
                'format_json': json.dumps(volume_fr['format']),
            }
        )


def _expansion_to_cpp(expansion):
    if expansion is None:
        return []
    if expansion.num == 0:
        return []

    return [
        i1350.SExpansionPoint(p0, p1, p2, temp_anl, temp_valve, time, ex)
        for time, p0, p1, p2, ex, temp_anl, temp_valve in zip(
            expansion.times, expansion.p0s, expansion.p1s, expansion.p2s, expansion.excludes,
            expansion.temp_anl_block, expansion.temp_valve_block
        )
    ]


def _update_cal(cal):
    # get the non adjusted chamber cal.
    sCalData = get_samplecaldata(cal)
    sCal = i1350.CalcChamberVolume(sCalData)

    # adjust the volume.
    cal.volume_fs_sample = sCal.dFSVolumeSample
    cal.volume_fs_reference = sCal.dFSVolumeReference
    cal.volume_fr_sample = sCal.dFRVolumeSample
    cal.volume_fr_reference = sCal.dFRVolumeReference
    cal.temperature = sCal.dTemperature


def _chamber_cal_to_cpp(cal):
    sChamberCal = i1350.SChamberCalibration()
    sChamberCal.dFSVolumeSample = _zero_if_none(cal.volume_fs_sample)
    sChamberCal.dFSVolumeReference = _zero_if_none(cal.volume_fs_reference)
    sChamberCal.dFRVolumeSample = _zero_if_none(cal.volume_fr_sample)
    sChamberCal.dFRVolumeReference = _zero_if_none(cal.volume_fr_reference)
    sChamberCal.dTemperature = cal.temperature
    sChamberCal.dThermalExpansion = cal.instrument_thermal_expansion
    return sChamberCal


def _insert_to_cpp(insert):
    sInsert = i1350.SInsertCalibration()
    sInsert.dVolume = insert.volume
    sInsert.dTemperature = insert.temperature
    sInsert.dThermalExpansion = insert.thermal_expansion
    return sInsert


def _calc_foam_pyc_common(anl, calc):

    calc.SetSampleMass(anl.mass)

    if isinstance(anl, AnalysisResult1350):
        prms = anl.parameters
        calc.SetResinDensity(prms.resin_density)

        if prms.compute_volume == 'E':
            calc.SetEnteredGeometryVolume(prms.geometric_volume)
        else:
            assert prms.compute_volume == 'C'
            if prms.shape == 'U':
                calc.SetCubeGeometry(prms.cube_edge_length, prms.number_of_pieces)
            elif prms.shape == 'Y':
                calc.SetCylinderGeometry(
                    prms.cylinder_diameter, prms.cylinder_length, prms.number_of_pieces
                )
            else:
                assert prms.shape == 'R'
                calc.SetRectangleGeometry(
                    prms.rect_long_edge, prms.rect_short_edge, prms.rect_other_edge,
                    prms.number_of_pieces
                )
    if isinstance(anl, AnalysisResult1345):
        calc.SetResinDensity(anl.resin_density)
        calc.SetEnteredGeometryVolume(anl.geometric_volume)

    return calc


def _none_if_nan(val):
    return val if not math.isnan(val) else None


def _zero_if_none(val):
    return val if val is not None else 0.0


def _analysisparameters_from_anl(anl, gas):  # pylint: disable=too-many-statements, too-many-branches
    """Fills in and returns an SAnalysisParameters appropriate for anl"""
    prms = i1350.SAnalysisParameters()
    prms.uiAnlID = anl.id
    method = anl.parameters
    if method.technique == 'P':
        prms.eAnalysisMethod = i1350.EAnalysisType.ceStandardAnalysis
    if method.technique == 'A':
        prms.eAnalysisMethod = i1350.EAnalysisType.ceMethodA
    if method.technique == 'B':
        prms.eAnalysisMethod = i1350.EAnalysisType.ceMethodB
    if method.technique == 'C':
        prms.eAnalysisMethod = i1350.EAnalysisType.ceMethodC
    if method.technique == 'D':
        prms.eAnalysisMethod = i1350.EAnalysisType.ceMethodD
        prms.flInitialPressureGauge = method.initial_pressure
        prms.flFinalPressureGauge = method.final_pressure
        prms.flPressureIncrement = method.pressure_increment
    if method.technique == 'E':
        prms.eAnalysisMethod = i1350.EAnalysisType.ceMethodE
        prms.flLowPressureGauge = method.low_pressure
        prms.flFracturePressureGuage = method.fracture_pressure
    prms.eFillDirection = (
        i1350.EFillDirection.ceFillSample
        if method.fill_direction == 'S' else i1350.EFillDirection.ceFillReference
    )
    prms.flAnalysisTemperatureC = method.temperature
    prms.fDoPrePurges = method.do_pre_purges
    prms.fUseVacuumPrep = method.use_vacuum_prep
    prms.iNumberOfPurges = method.num_purges
    prms.flPurgeFillPressureGauge = method.purge_fill_pressure
    prms.iNumberOfCycles = method.num_cycles
    prms.flCycleFillPressureGauge = method.cycle_fill_pressure
    if method.endby == 'E':
        prms.eP1_P2_EndBy = i1350.EEndEquil.ceEndByEquilibrate
    else:
        prms.eP1_P2_EndBy = i1350.EEndEquil.ceEndByFixedInterval
    prms.iIntervalTime = method.interval_time
    prms.flEquilibrationRate = method.equilibration_rate
    prms.fP1RecordEquilibrationData = method.p1_record_equilibration_data
    prms.iP1EquilibrationDataCycles = method.p1_equilibration_cycles
    prms.iP1EquilibrationDataPoints = method.p1_equilibration_points
    prms.fP2RecordEquilibrationData = method.p2_record_equilibration_data
    prms.iP2EquilibrationDataCycles = method.p2_equilibration_cycles
    prms.iP2EquilibrationDataPoints = method.p2_equilibration_points
    if method.technique == 'P':
        prms.fUseRunPrecision = method.use_run_precision
    else:
        prms.fUseRunPrecision = False
    prms.flRunPrecisionTolerance = method.run_precision_tolerance
    if method.insert and method.cup:
        prms.iNumInserts = 2
        sInsert = _insert_to_cpp(method.insert)
        sCup = _insert_to_cpp(method.cup)
        prms.asInsert = [sInsert, sCup, sInsert]
    elif method.insert:
        prms.iNumInserts = 1
        sInsert = _insert_to_cpp(method.insert)
        prms.asInsert = [sInsert, sInsert, sInsert]
    elif method.cup:
        prms.iNumInserts = 1
        sCup = _insert_to_cpp(method.cup)
        prms.asInsert = [sCup, sCup, sCup]
    else:
        prms.iNumInserts = 0
    prms.szGasName = gas.name
    prms.sChamberCal = _chamber_cal_to_cpp(anl.calibration)
    return prms


def _resultheader_from_anl(anl):
    """Fills in and returns the SResultHeader structure."""
    header = i1350.SResultHeader()
    header.szModel = anl.instrument.model
    header.szSerial = anl.instrument.serial
    header.szAppConfig = anl.instrument.full_config
    header.szDescription = anl.name
    header.analysisTime = anl.start_time
    return header
