'''views and support logic that handles the settings/service functionality.'''
import gettext
import glob
import re
from django import forms
from django.http import JsonResponse
from django.utils.functional import lazy
from django.views.generic import View
from mic_python_lib.mic_network import get_usb_drive_mounts
from mic_python_lib.rest import function_to_json_view
from mic_python_lib.rest_input import (
    ReactFormDataMixin,
    UnitsFloatFormField,
    UnitsIntegerFormField,
)
import i1350
from mic_cpp import pycnometry
from .models import Gas, Instrument

# pylint: disable=invalid-name

_ = lazy(gettext.gettext, str)


def hw_set_valve(index, state):
    '''Set the valve index to state.'''
    if state:
        i1350.UDPOpenValve(index)
    else:
        i1350.UDPCloseValve(index)


hw_set_valve_view = lambda request: function_to_json_view(request, hw_set_valve)


def hw_state_view(request):
    """The view that returns the json object for the hardware state"""
    state = i1350.UDPGetHWState()
    state_dict = {name: getattr(state, name) for name in dir(state) if not name.startswith('__')}
    # augment the hardware state with the operation status string.
    state_dict['szOperStatus'] = i1350.UDPGetOperationStatus().szStatus
    return JsonResponse(state_dict)


class SetTemperatureForm(forms.Form):
    """Form to set temperature and related field"""
    temperature = UnitsFloatFormField(fmt=pycnometry.DPycTemperature1350, label=_("Temperature"))  # pylint: disable=no-member

class SetGasForm(forms.Form):
    """Form to set gas and related field"""
    gas = forms.CharField(max_length=64, label=_("Gas"))  # pylint: disable=no-member


class hw_set_temperature(ReactFormDataMixin, View):  # pylint: disable=too-few-public-methods
    '''View to return data for start analysis screen'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define model, fields, and options for displaying form'''
        form = SetTemperatureForm

    def get_initial_values(self, request):
        return {
            'temperature': i1350.UDPGetHWState().flTempControlTarget,
        }

    def on_save(self, request, data):
        i1350.UDPSetTempTarget(data['temperature'])

def hw_get_gas():
    "Get the current and list of gases"
    return {
        'current': Instrument.objects.local().gas.name,
        'available': [gas.name for gas in Gas.objects.all()]
    }


class hw_set_gas(ReactFormDataMixin, View):  # pylint: disable=too-few-public-methods
    '''View to return data for start analysis screen'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define model, fields, and options for displaying form'''
        form = SetGasForm

    def get_initial_values(self, request):
        return {
            'gas': Instrument.objects.local().gas
        }

    def on_save(self, request, data):
        local = Instrument.objects.local()
        local.gas = Gas.objects.get(name=data['gas'])
        local.save()


def hw_set_temp_control(state):
    '''Set the temperature control to state.'''
    if state:
        i1350.UDPSetTempTarget(i1350.UDPGetHWState().flTempControlTarget)
    else:
        i1350.UDPSetTempControl(0.0, 0)


hw_set_temp_control_view = lambda request: function_to_json_view(request, hw_set_temp_control)


class SetLEDForm(forms.Form):
    """Form to set LEDs"""
    red = UnitsIntegerFormField(fmt=pycnometry.UPycLEDColor, label=_("Red"))  # pylint: disable=no-member
    green = UnitsIntegerFormField(fmt=pycnometry.UPycLEDColor, label=_("Green"))  # pylint: disable=no-member
    blue = UnitsIntegerFormField(fmt=pycnometry.UPycLEDColor, label=_("Blue"))  # pylint: disable=no-member


class hw_set_led(ReactFormDataMixin, View):  # pylint: disable=too-few-public-methods
    '''View to return data for start analysis screen'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define model, fields, and options for displaying form'''
        form = SetLEDForm

    def get_initial_values(self, request):
        state = i1350.UDPGetHWState()
        return {
            'red': state.ubLEDRed,
            'green': state.ubLEDGreen,
            'blue': state.ubLEDBlue,
        }

    def on_save(self, request, data):
        i1350.UDPSetLEDColor(data['red'], data['green'], data['blue'])

def pcb_sw_update_avail():
    "Get the versions of the pcb update software if available."

    bin_files = []
    for drive in get_usb_drive_mounts():
        bin_files = bin_files + glob.glob(f"{drive}/135-20021-00_*.bin")

    if len(bin_files) != 1:
        return {
            "isAvail": False,
            "currentVersion": i1350.UDPGetHWState().szFirmwareVersion,
            "newVersion": ""
        }

    version = re.match(".*/135-20021-00_(?P<ver>.*)\\.bin", bin_files[0]).group("ver")

    return {
        "isAvail": True,
        "currentVersion": i1350.UDPGetHWState().szFirmwareVersion,
        "newVersion": version
    }

def pcb_sw_update(version):
    "Updates the pcb software with the file of the version requested."
    bin_files = []
    for drive in get_usb_drive_mounts():
        bin_files = bin_files + glob.glob(f"{drive}/135-20021-00_{version}.bin")
    if len(bin_files) != 1:
        return {"error": _("Cannot find PCB software file.")}
    i1350.UDPDownloadPCBSoftware(bin_files[0])
    return {"error": ""}
