'''
tests for pyc_sample.services
'''

# pylint: disable=invalid-name,wrong-import-position,missing-function-docstring

import json
from collections import namedtuple
from unittest.mock import MagicMock
from django.test import TestCase, Client
from mic_python_lib.models import Result

import pyc_sample.services as services
import pyc_sample.views as views
from ..services import (
    analysis_start,
    calibration_start,
    get_chambercal,
    get_samplecaldata,
    operation_cancel,
)
from ..models import (
    AnalysisParameters,
    AnalysisResult1350,
    Calibration,
    Equilibration,
    Expansion,
    Gas,
    Insert,
    Instrument,
)

mock_i1350 = MagicMock()

SCalcChamberCalibration = namedtuple(
    "SCalcChamberCalibration", [
        "dFSVolumeSample",
        "dFSVolumeReference",
        "dFRVolumeSample",
        "dFRVolumeReference",
        "dTemperature",
        "dThermalExpansion",
        "dFSVolumeSampleStdDev",
        "dFSVolumeReferenceStdDev",
        "dFRVolumeSampleStdDev",
        "dFRVolumeReferenceStdDev",
    ]
)

SHWState = namedtuple("SHWState", ["flBlockTemp", "flGaugePressure"])
SStatus = namedtuple(
    "SStatus",
    ["iCycle", "iCycleNum", "iPurge", "iPurgeNum", "iRefPurge", "iRefPurgeNum",
     "eOperation", "szStatus", "flFractionComplete", "fIsTerminating", "eIdleOperation"]
)


class Test10cc(TestCase):
    '''test case for start calibration.'''

    def setUp(self):
        # mock c++ extensions so they aren't needed in testing
        self.i1350 = services.i1350
        services.i1350 = mock_i1350
        views.i1350 = mock_i1350
        mock_i1350.SVolumeCalibrationParameters = MagicMock()
        mock_i1350.UDPStartVolumeCalibration = fake_calibration
        mock_i1350.SAnalysisParameters = MagicMock()
        mock_i1350.UDPStartAnalysis = fake_analysis
        mock_i1350.UDPGetHWState = fake_get_hw_state
        mock_i1350.UDPGetOperationStatus = fake_get_op_status
        fake_calibration.testobj = self
        fake_analysis.testobj = self
        mock_i1350.CalcChamberVolume = fake_calc_chamber_volume
        mock_i1350.CalcSampleVolumeJson = fake_calc_sample_volume_json
        mock_i1350.CalcSampleDensityJson = fake_calc_sample_density_json
        mock_i1350.CalcLastSampleDensityJson = fake_calc_last_sample_density_json
        self.client = Client()

        # make the local instrument parameters as expected for this test.
        inst = Instrument.objects.local()
        inst.serial = "test"
        inst.kind = "T"
        inst.material = "ot"
        inst.other_expansion_um_per_mk = 0.010 / 1.0e-6
        inst.save()

        insert = Insert.objects.create(
            name="ref-vol",
            volume=5,
            temperature=20,
            material="ot",
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="R"
        )
        cup = Insert.objects.create(
            name="cup",
            volume=0.5,
            temperature=20,
            material="ot",
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="C"
        )
        AnalysisParameters.objects.create(
            name="test", is_method=True, gas=Gas.objects.get(name="Helium"), insert=insert
        )
        AnalysisParameters.objects.create(
            name="test2", is_method=True, gas=Gas.objects.get(name="Helium"), cup=cup
        )

    def tearDown(self):
        services.i1350 = self.i1350
        views.i1350 = self.i1350
        del fake_calibration.testobj

    def test_calibration_and_analysis(self):
        self.assertEqual(Expansion.objects.all().count(), 0)
        self.assertEqual(Calibration.objects.all().count(), 0)

        ref_ball = Insert.objects.get(name="ref-vol")
        calibration_start(
            10.0,
            kind='C',
            ref_ball_id=ref_ball,
        )
        self.assertEqual(Calibration.objects.all().count(), 1)
        inst = Instrument.objects.local()
        cal = Calibration.objects.get()
        # make sure that the calibration values are stored.
        self.assertEqual(cal.state, "C")
        self.assertEqual(Expansion.objects.all().count(), 4)
        self.assertEqual(cal.ref_ball.name, ref_ball.name)
        self.assertEqual(cal.ref_ball.volume, ref_ball.volume)
        self.assertNotEqual(cal.ref_ball.id, ref_ball.id)
        self.assertAlmostEqual(cal.instrument_thermal_expansion, inst.thermal_expansion)

        anl_prm = AnalysisParameters.objects.get(is_method=True, name="test")
        analysis_start(anl_prm.id, 2.0, "analysis", "django")
        self.assertEqual(Expansion.objects.all().count(), 6)
        self.assertEqual(Equilibration.objects.all().count(), 1)
        anl = AnalysisResult1350.objects.get()
        self.assertNotEqual(anl_prm.insert.id, anl.parameters.insert.id)
        self.assertEqual(anl_prm.insert.name, anl.parameters.insert.name)
        self.assertEqual(anl_prm.insert.volume, anl.parameters.insert.volume)
        self.assertEqual(anl.state, "C")

        self.assertEqual(Result.objects.all().count(), 3)
        vol = Result.objects.get(result_name="volume")
        self.assertEqual(vol.value, 6.0)
        self.assertEqual(vol.std_dev, 0.01)
        den = Result.objects.get(result_name="absolute_density")
        self.assertEqual(den.value, 2 / 6.0)
        self.assertEqual(den.std_dev, 0.005)

        anl_prm = AnalysisParameters.objects.get(is_method=True, name="test2")
        analysis_start(anl_prm.id, 2.0, "analysis", "django")
        anl = AnalysisResult1350.objects.get(parameters__name="test2")
        self.assertNotEqual(anl_prm.cup.id, anl.parameters.cup.id)
        self.assertEqual(anl_prm.cup.name, anl.parameters.cup.name)
        self.assertEqual(anl_prm.cup.volume, anl.parameters.cup.volume)

    def test_analysis_no_calibration(self):
        start_params = {
            'method_id': AnalysisParameters.objects.get(name='test').id,
            'name': 'test',
            'mass': 2.0,
            'operator': 'django'
        }
        response = self.client.post(
            '/rest/anlstart/', json.dumps(start_params), content_type="application/json"
        )
        self.assertTrue(json.loads(response.content)['error'])

    def test_analysis_parameters_string(self):
        method = AnalysisParameters.objects.get(is_method=True, name="test")
        self.assertEqual("test", str(method))

    def test_operation_cancel(self):  # pylint: disable=no-self-use
        operation_cancel()
        mock_i1350.UDPCancelOperation.assert_called_once()

    def test_ui_start(self):
        response = self.client.get('/rest/ui_anl_start/')
        self.assertEqual(response.status_code, 200)

    def test_ui_inprocess(self):
        calibration_start(
            10.0,
            kind='C',
            ref_ball_id=Insert.objects.get(name="ref-vol").id,
        )
        analysis_start(
            AnalysisParameters.objects.get(is_method=True, name="test").id, 2.0, "analysis",
            "django"
        )
        anl = AnalysisResult1350.objects.get()
        response = self.client.post(
            f'/rest/ui_anl_inprocess/{anl.id}/', content_type="application/json"
        )
        self.assertEqual(response.status_code, 200)

    def test_ui_operation_status(self):
        response = self.client.post(f'/rest/ui_operation_status/', content_type="application/json")
        self.assertEqual(response.status_code, 200)
        calibration_start(
            10.0,
            kind='C',
            ref_ball_id=Insert.objects.get(name="ref-vol").id,
        )
        analysis_start(
            AnalysisParameters.objects.get(is_method=True, name="test").id, 2.0, "analysis",
            "django"
        )
        response = self.client.post(f'/rest/ui_operation_status/', content_type="application/json")
        self.assertEqual(response.status_code, 200)

    def test_chamber_expansion_copy(self):
        self.assertEqual(Expansion.objects.all().count(), 0)
        self.assertEqual(Calibration.objects.all().count(), 0)

        ref_ball = Insert.objects.get(name="ref-vol")
        calibration_start(
            10.0,
            kind='C',
            ref_ball_id=ref_ball,
        )
        self.assertEqual(Calibration.objects.all().count(), 1)

        inst = Instrument.objects.local()
        inst.material = "ot"
        inst.other_expansion_um_per_mk = 0.020 / 1.0e-6
        inst.save()
        cal = Calibration.objects.get()
        data = get_samplecaldata(cal)
        self.assertAlmostEqual(data.dChamberThermalExpansion, 0.01)

        data = get_chambercal(cal)
        self.assertAlmostEqual(data.dThermalExpansion, 0.01)


def fake_calibration(params):
    self = fake_calibration.testobj
    calID = params.uiCalID

    self.assertEqual(Calibration.objects.get(id=calID).state, "P")
    rundata = {
        'cal_id': calID,
        'run': 0,
        'cycles': 2,
        'time': [1000, 2000],
        'p0': [15.0, 15.0],
        'p1': [35.0, 35.0],
        'p2': [25.0, 25.0],
        'exclude': [False, False],
        'temp_anl': [25.0, 25.0],
        'temp_valve': [25.3, 25.3],
    }

    resp = self.client.post('/rest/caldata/', json.dumps(rundata), content_type="application/json")
    self.assertEqual(resp.status_code, 200)
    rundata['run'] = 1
    resp = self.client.post('/rest/caldata/', json.dumps(rundata), content_type="application/json")
    self.assertEqual(resp.status_code, 200)
    rundata['run'] = 2
    rundata['p2'] = [20.0, 20.0]
    resp = self.client.post('/rest/caldata/', json.dumps(rundata), content_type="application/json")
    self.assertEqual(resp.status_code, 200)
    rundata['run'] = 3
    resp = self.client.post('/rest/caldata/', json.dumps(rundata), content_type="application/json")
    self.assertEqual(resp.status_code, 200)

    resp = self.client.post(
        '/rest/caldone/',
        json.dumps({
            'cal_id': calID,
            'state': "C"
        }),
        content_type="application/json"
    )
    self.assertEqual(resp.status_code, 200)

    # make sure that the calibration values are stored.
    cal = Calibration.objects.get_current()
    self.assertIsNone(cal.volume_fs_sample)
    self.assertIsNone(cal.volume_fs_reference)
    self.assertEqual(cal.volume_fr_sample, 10)
    self.assertEqual(cal.volume_fr_reference, 11)
    self.assertEqual(cal.temperature, 22)


def fake_analysis(params):
    self = fake_analysis.testobj
    anlID = params.uiAnlID

    resp = self.client.post(
        '/rest/equildata/',
        json.dumps({
            'anl_id': anlID,
            'is_p1': False,
            'run': 0,
            'cycle': 0,
            'num': 5,
            'time': [100, 200, 300, 400, 500],
            'pressures': [35.0, 30.0, 28.0, 26.0, 25.0]
        }),
        content_type="application/json"
    )
    self.assertEqual(resp.status_code, 200)

    resp = self.client.post(
        '/rest/anldata/',
        json.dumps({
            'anl_id': anlID,
            'run': 0,
            'cycles': 1,
            'time': [1000],
            'p0': [15.0],
            'p1': [35.0],
            'p2': [25.0],
            'exclude': [False],
            'temp_anl': [25.0],
            'temp_valve': [25.3],
        }),
        content_type="application/json"
    )
    self.assertEqual(resp.status_code, 200)

    resp = self.client.post(
        '/rest/anldata/',
        json.dumps({
            'anl_id': anlID,
            'run': 0,
            'cycles': 2,
            'time': [1000, 2000],
            'p0': [15.0, 15.0],
            'p1': [35.0, 35.0],
            'p2': [25.0, 25.0],
            'exclude': [False, False],
            'temp_anl': [25.0, 25.0],
            'temp_valve': [25.3, 25.3],
        }),
        content_type="application/json"
    )
    self.assertEqual(resp.status_code, 200)
    resp = self.client.post(
        '/rest/anldata/',
        json.dumps({
            'anl_id': anlID,
            'run': 1,
            'cycles': 2,
            'time': [1000, 2000],
            'p0': [15.0, 15.0],
            'p1': [35.0, 35.0],
            'p2': [25.0, 25.0],
            'exclude': [False, False],
            'temp_anl': [25.0, 25.0],
            'temp_valve': [25.3, 25.3],
        }),
        content_type="application/json"
    )
    self.assertEqual(resp.status_code, 200)

    resp = self.client.post(
        '/rest/anldone/',
        json.dumps({
            'anl_id': anlID,
            'state': "C"
        }),
        content_type="application/json"
    )
    self.assertEqual(resp.status_code, 200)


def fake_calc_chamber_volume(prms):
    return SCalcChamberCalibration(
        dFSVolumeSample=float("nan"),
        dFSVolumeReference=float("nan"),
        dFRVolumeSample=10.0,
        dFRVolumeReference=11.0,
        dTemperature=22.0,
        dThermalExpansion=prms.dChamberThermalExpansion,
        dFSVolumeSampleStdDev=0.0,
        dFSVolumeReferenceStdDev=0.0,
        dFRVolumeSampleStdDev=0.0,
        dFRVolumeReferenceStdDev=0.0,
    )


def fake_calc_sample_volume_json(prms):  # pylint: disable=unused-argument
    return '{"ave":6.0,"std_dev":0.01,"format":{"style":"fixed","precision":4},"units":["cm3"]}'


def fake_calc_sample_density_json(prms, mass):  # pylint: disable=unused-argument
    data = {
        "ave": mass / 6.0,
        "std_dev": 0.005,
        "format": {
            "style": "fixed",
            "precision": 4
        },
        "units": ["g/cm3"]
    }
    return json.dumps(data)


def fake_calc_last_sample_density_json(prms, mass):  # pylint: disable=unused-argument
    data = {"value": mass / 6.0, "format": {"style": "fixed", "precision": 4}, "units": ["g/cm3"]}
    return json.dumps(data)


def fake_get_hw_state():
    return SHWState(15.0, 25.0)


def fake_get_op_status():
    return SStatus(1, 10, 1, 1, 0, 0, "ceAnalyze", "No Status", 0.5, False, 0)
