'''
tests for reset to factory defaults.
'''
import datetime
from django.contrib.auth.models import User
from django.test import TestCase
from django.utils import timezone
from mic_python_lib.models import LogMessage, Settings

from ..models import (
    AnalysisParameters,
    AnalysisResult1350,
    Calibration,
    Equilibration,
    Expansion,
    Gas,
    Insert,
    Instrument,
    RTDCalibration,
    PressureCalibration,
    copy_model_obj,
)
from ..services import reset_to_factory_defaults


# pylint: disable=invalid-name,wrong-import-position,missing-function-docstring
class TestReset(TestCase):
    '''test case for microactive communicaitons.'''

    def setUp(self):
        User.objects.create(username="service", is_active=False)
        _add_data()

    def test_service_warranty(self):
        reset_to_factory_defaults()
        local = Instrument.objects.local()
        self.assertGreater(
            local.device_service,
            timezone.now().date() + datetime.timedelta(days=364)
        )
        self.assertGreater(
            local.device_warranty,
            timezone.now().date() + datetime.timedelta(days=364)
        )
        self.assertLess(local.device_service, timezone.now().date() + datetime.timedelta(days=367))
        self.assertLess(local.device_warranty, timezone.now().date() + datetime.timedelta(days=367))

    def test_data(self):
        self.assertGreater(Instrument.objects.all().count(), 1)
        self.assertGreater(Calibration.objects.all().count(), 1)
        self.assertGreater(AnalysisResult1350.objects.all().count(), 0)
        reset_to_factory_defaults()
        self.assertEqual(Instrument.objects.all().count(), 1)
        self.assertEqual(Calibration.objects.all().count(), 1)
        self.assertEqual(Insert.objects.filter(kind='C', is_method=True).count(), 1)
        self.assertEqual(Insert.objects.filter(kind='R', is_method=True).count(), 1)
        self.assertEqual(AnalysisResult1350.objects.all().count(), 0)

    def test_methods(self):
        local = Instrument.objects.local()
        local.kind = '1'
        local.save()
        self.assertEqual(AnalysisParameters.objects.all().count(), 3)
        reset_to_factory_defaults()
        self.assertEqual(AnalysisParameters.objects.all().count(), 2)

        local.kind = 'T'
        local.save()
        reset_to_factory_defaults()
        self.assertEqual(AnalysisParameters.objects.all().count(), 10)

        local.kind = 'H'
        local.save()
        reset_to_factory_defaults()
        self.assertEqual(AnalysisParameters.objects.all().count(), 9)


class TestErrors(TestCase):
    "test case for microactive communicaitons."

    def setUp(self):
        User.objects.create(username="service", is_active=False)

    def test_no_error(self):
        _add_data()
        self.assertEqual(Settings.objects.get().guided_setup_page, -1)
        reset_to_factory_defaults()
        self.assertEqual(len(LogMessage.objects.all()), 0)
        self.assertEqual(Settings.objects.get().guided_setup_page, 0)

    def test_no_date(self):
        reset_to_factory_defaults()
        self.assertEqual(len(LogMessage.objects.all()), 6)
        self.assertEqual(Settings.objects.get().guided_setup_page, -1)

    def test_extra_cup(self):
        _add_data()
        Insert.objects.create(
            name="cup 2",
            volume=0.5,
            temperature=20,
            material="ot",
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="C"
        )
        reset_to_factory_defaults()
        self.assertEqual(len(LogMessage.objects.all()), 1)
        self.assertIn("exactly one cup.", LogMessage.objects.get().message)

    def test_cup(self):
        _add_data()
        for meth in AnalysisParameters.objects.filter(is_method=True):
            meth.cup = None
            meth.save()
        Insert.objects.filter(is_method=True, kind='C').delete()
        cup = Insert.objects.create(
            name="cup 2",
            volume=3.141,
            temperature=20,
            material="ot",
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="C"
        )
        reset_to_factory_defaults()
        self.assertEqual(len(Insert.objects.filter(is_method=True, kind='C')), 1)
        self.assertEqual(Insert.objects.get(is_method=True, kind='C').volume, 3.141)
        for meth in AnalysisParameters.objects.filter(is_method=True):
            self.assertEqual(meth.cup.id, cup.id)

    def test_extra_ref(self):
        _add_data()
        Insert.objects.create(
            name="ref 2",
            volume=0.5,
            temperature=20,
            material='ot',
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="R"
        )
        reset_to_factory_defaults()
        self.assertEqual(len(LogMessage.objects.all()), 1)
        self.assertIn("exactly one reference volume.", LogMessage.objects.get().message)

    def test_ref(self):
        _add_data()
        Insert.objects.filter(is_method=True, kind='R').delete()
        Insert.objects.create(
            name="ref 2",
            volume=3.141,
            temperature=20,
            material='ot',
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="R"
        )
        reset_to_factory_defaults()
        self.assertEqual(len(Insert.objects.filter(is_method=True, kind='R')), 1)
        self.assertEqual(Insert.objects.get(is_method=True, kind='R').volume, 3.141)

def _add_data():
    inst = Instrument.objects.local()
    Instrument.objects.create(serial='1234',)
    inst.serial = "test"
    inst.kind = "T"
    inst.material = 'ot'
    inst.other_expansion_um_per_mk = 0.010 / 1.0e-6
    inst.save()

    RTDCalibration.objects.create(
        instrument=inst,
        channel='B',
        slope=1.0,
        raw_resistance=0.0,
        actual_resistance=0.0,
        check_reference=5.0,
        check_reading=5.0,
        gauge_sn='fake',
    )
    RTDCalibration.objects.create(
        instrument=inst,
        channel='I',
        slope=1.0,
        raw_resistance=0.0,
        actual_resistance=0.0,
        check_reference=5.0,
        check_reading=5.0,
        gauge_sn='fake',
    )
    PressureCalibration.objects.create(
        instrument=inst,
        slope=1.0,
        voltage=5.0,
        pressure=50.0,
        check_reference=25.0,
        check_reading=25.0,
        gauge_sn='fake',
    )

    ref_ball = Insert.objects.create(
        name="ref-vol",
        volume=5,
        temperature=20,
        material='ot',
        other_expansion_um_per_mk=0.011 / 1.0e-6,
        kind="R"
    )
    cup = Insert.objects.create(
        name="cup",
        volume=0.5,
        temperature=20,
        material='ot',
        other_expansion_um_per_mk=0.011 / 1.0e-6,
        kind="C"
    )
    Insert.objects.create(
        name="insert2",
        volume=1.5,
        temperature=20,
        material='ot',
        other_expansion_um_per_mk=0.011 / 1.0e-6,
        kind="I"
    )
    insert = Insert.objects.create(
        name="insert",
        volume=1.5,
        temperature=20,
        material='ot',
        other_expansion_um_per_mk=0.011 / 1.0e-6,
        kind="I"
    )
    prms = AnalysisParameters.objects.create(
        name="Parameters",
        is_default=True,
        is_method=True,
        gas=Gas.objects.get(name="Nitrogen"),
        insert=insert,
        cup=cup,
    )
    prms = AnalysisParameters.objects.create(
        name="test",
        is_method=False,
        gas=Gas.objects.get(name="Helium"),
        insert=insert,
        cup=copy_model_obj(Insert, cup),
    )
    expansion = Expansion.objects.create(
        num=3,
        times=[1, 120, 180],
        p0s=[14.6, 14.6, 14.6],
        p1s=[30.0, 30.0, 30.0],
        p2s=[20.0, 20.0, 20.0],
        excludes=[0, 0, 0],
        temp_anl_block=[22.6, 22.6, 22.6],
        temp_valve_block=[22.9, 22.9, 22.9],
    )
    cal = Calibration.objects.create(
        state='C',
        kind='C',
        gas='Helium',
        instrument=inst,
        instrument_thermal_expansion_um_per_mk=0.123 / 1e-6,
        ref_ball=copy_model_obj(Insert, ref_ball),
        cup=copy_model_obj(Insert, cup),
        volume_fs_sample=10.0,
        volume_fs_reference=10.0,
        volume_fr_sample=10.0,
        volume_fr_reference=10.0,
        temperature=22.0,
    )
    cal = Calibration.objects.create(
        state='X',
        kind='H',
        gas='Helium',
        instrument=inst,
        instrument_thermal_expansion_um_per_mk=0.123 / 1e-6,
        ref_ball=copy_model_obj(Insert, ref_ball),
        volume_fs_sample=10.0,
        volume_fs_reference=10.0,
        volume_fr_sample=10.0,
        volume_fr_reference=10.0,
        temperature=22.0,
    )
    result = AnalysisResult1350.objects.create(
        state='C',
        instrument=inst,
        parameters=prms,
        mass=1.0,
        name="Result 1",
        calibration=cal,
        expansion=expansion,
    )

    Equilibration.objects.create(
        analysis=result,
        cycle=0,
        which_equil='1',
        num_points=5,
        times=[1, 2, 3, 4, 5],
        pressures=[28, 26, 24, 22, 20],
    )
