"""Tests for the Insert model and associated functionality."""
# pylint: disable=invalid-name,missing-function-docstring,missing-class-docstring,protected-access,too-many-locals
import json
from django.test import RequestFactory, TestCase
from django_test_migrations.contrib.unittest_case import MigratorTestCase
from pyc_sample.models import Insert, Instrument, copy_model_obj
from pyc_sample.views import (
    CalStartForm,
    MethodForm,
    _get_default_ref,
    make_insert_view,
)


class TestInsertMethodMigration(MigratorTestCase):
    """This class is used to test direct migrations."""

    migrate_from = ('pyc_sample', '0046_add_air_to_gastable')
    migrate_to = ('pyc_sample', '0047_insert_methods_and_copy')

    def prepare(self):
        """Prepare some data before the migration."""

        def _add_insert(name, kind, active=True):
            return OldInsert.objects.create(
                name=name,
                volume=1.0,
                temperature=22.0,
                thermal_expansion_um_per_mk=0.0,
                kind=kind,
                active=active,
            )

        def _add_prm(name, cup, insert, is_method=False):
            return OldAnalysisParameters.objects.create(
                name=name,
                is_method=is_method,
                gas=OldGas.objects.get(name="Helium"),
                cup=cup,
                insert=insert,
            )

        def _add_cal(cup, ref):
            return OldCalibration.objects.create(
                state='C',
                kind='C',
                gas=OldGas.objects.get(name="Helium"),
                instrument=OldInstrument.objects.get(),
                ref_ball=ref,
                cup=cup,
            )

        # pylint: disable=invalid-name
        OldInstrument = self.old_state.apps.get_model('pyc_sample', 'Instrument')
        OldGas = self.old_state.apps.get_model('pyc_sample', 'Gas')
        OldInstrument.objects.create(serial='0000', is_local=True, kind='T')
        OldAnalysisParameters = self.old_state.apps.get_model('pyc_sample', 'AnalysisParameters')
        OldInsert = self.old_state.apps.get_model('pyc_sample', 'Insert')
        OldCalibration = self.old_state.apps.get_model('pyc_sample', 'Calibration')

        c1 = _add_insert("Cup 1", 'C')
        r1 = _add_insert("Ref 1", 'R')
        _add_insert("Ins 1", 'I')
        c2 = _add_insert("Cup 2", 'C', False)
        _add_insert("Ref 2", 'R', False)
        i2 = _add_insert("Ins 2", 'I', False)

        _add_prm("Prm 1", c1, i2)
        _add_prm("Prm 2", None, i2)
        _add_prm("Prm 3", c1, None)
        _add_prm("Prm 4", None, None)
        _add_cal(c2, r1)
        _add_cal(None, r1)

        # add some duplicates to test their removal.
        c_dup = _add_insert("Cup 1", 'C')
        _add_insert("Ref 1", 'R')
        i_dup = _add_insert("Ins 1", 'I')
        _add_prm("Prm 5", c_dup, i_dup, True)

    def test_duplicate_removal(self):
        # pylint: disable=invalid-name
        NewAnalysisParameters = self.new_state.apps.get_model('pyc_sample', 'AnalysisParameters')
        NewCalibration = self.new_state.apps.get_model('pyc_sample', 'Calibration')
        NewInsert = self.new_state.apps.get_model('pyc_sample', 'Insert')

        # make sure the three inactive inserts were deleted.
        self.assertEqual(NewInsert.objects.filter(is_method=True).count(), 3)
        self.assertGreater(NewInsert.objects.count(), 3)

        # make sure the duplicate cups were switched over to the remaining ones.
        self.assertEqual(NewAnalysisParameters.objects.filter(is_method=True).count(), 2)
        self.assertEqual(NewAnalysisParameters.objects.get(name="Prm 5").cup.name, "Cup 1")
        self.assertEqual(NewCalibration.objects.filter().count(), 2)

    def test_migration_main0003(self):
        """Run the test itself."""
        # pylint: disable=invalid-name
        NewAnalysisParameters = self.new_state.apps.get_model('pyc_sample', 'AnalysisParameters')
        NewCalibration = self.new_state.apps.get_model('pyc_sample', 'Calibration')

        # make sure that the analysis parameters were copied.
        p1 = NewAnalysisParameters.objects.get(name="Prm 1")
        self.assertEqual(p1.cup.name, "Cup 1")
        self.assertEqual(p1.insert.name, "Ins 2")
        self.assertFalse(p1.cup.is_method)
        self.assertFalse(p1.insert.is_method)

        # make sure the calibration parameters were copied.
        cal = NewCalibration.objects.first()
        self.assertEqual(cal.cup.name, "Cup 2")
        self.assertEqual(cal.ref_ball.name, "Ref 1")
        self.assertFalse(cal.cup.is_method)
        self.assertFalse(cal.ref_ball.is_method)


class TestInsertManager(TestCase):
    """This class is used to test the InsertManager class."""

    def setUp(self):
        _add_insert("Cup 1", 'C')
        _add_insert("Ref 1", 'R')
        _add_insert("Ins 1", 'I')
        _add_insert("Cup 2", 'C')
        _add_insert("Ref 2", 'R')
        _add_insert("Ins 2", 'I')
        _add_insert("Cup 3", 'C', False)
        _add_insert("Ref 3", 'R', False)
        _add_insert("Ins 3", 'I', False)

    def test_manager(self):
        self.assertEqual(len(Insert.objects.cups()), 2)
        self.assertTrue(all(c.is_method for c in Insert.objects.cups()))
        self.assertTrue(all(c.kind == 'C' for c in Insert.objects.cups()))
        self.assertEqual(len(Insert.objects.reference_balls()), 2)
        self.assertTrue(all(c.is_method for c in Insert.objects.reference_balls()))
        self.assertTrue(all(c.kind == 'R' for c in Insert.objects.reference_balls()))
        self.assertEqual(len(Insert.objects.inserts()), 2)
        self.assertTrue(all(c.is_method for c in Insert.objects.inserts()))
        self.assertTrue(all(c.kind == 'I' for c in Insert.objects.inserts()))

    def test_get_default(self):
        ref_id = _get_default_ref()
        ref = Insert.objects.get(id=ref_id)
        self.assertEqual(ref.name, "Ref 2")
        self.assertTrue(ref.is_method)

        Insert.objects.methods().delete()
        self.assertEqual(_get_default_ref(), -1)

    def test_copy_model(self):
        num = len(Insert.objects.all())
        c1 = Insert.objects.get(name="Cup 1")
        c2 = copy_model_obj(Insert, c1)
        self.assertTrue(c1.is_method)
        self.assertFalse(c2.is_method)
        self.assertEqual(num + 1, len(Insert.objects.all()))  # make sure c2 is saved
        c3 = copy_model_obj(Insert, c1, False)
        self.assertFalse(c3.is_method)
        self.assertEqual(num + 1, len(Insert.objects.all()))  # make sure c2 is not saved
        c3.save()
        self.assertEqual(num + 2, len(Insert.objects.all()))  # make sure c2 is now saved


class TestCalStartForm(TestCase):

    def setUp(self):
        _add_insert("Cup 1", 'C')
        _add_insert("Ref 1", 'R')
        _add_insert("Cup 2", 'C')
        _add_insert("Ref 2", 'R')
        _add_insert("Cup 3", 'C', False)
        _add_insert("Ref 3", 'R', False)
        _add_insert("Ref 4", 'R')

    def test_cup_choices(self):
        form = CalStartForm()
        cup = form.fields['cup']
        cup_choices = list(cup._queryset)
        self.assertEqual(len(cup_choices), 2)

    def test_ref_ball_choices(self):
        form = CalStartForm()
        ref_ball = form.fields['ref_ball']
        ref_ball_choices = list(ref_ball._queryset)
        self.assertEqual(len(ref_ball_choices), 3)


class TestMethodForm(TestCase):

    def setUp(self):
        _add_insert("Cup 1", 'C')
        _add_insert("Ins 1", 'I')
        _add_insert("Cup 2", 'C')
        _add_insert("Ins 2", 'I')
        _add_insert("Cup 3", 'C', False)
        _add_insert("Ins 3", 'I', False)
        _add_insert("Ins 4", 'I')

    def test_cup_choices(self):
        form = MethodForm()
        cup = form.fields['cup']
        cup_choices = list(cup._queryset)
        self.assertEqual(len(cup_choices), 2)

    def test_ref_ball_choices(self):
        form = MethodForm()
        ins = form.fields['insert']
        ins_choices = list(ins._queryset)
        self.assertEqual(len(ins_choices), 3)


class TestInsertView(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        _add_insert("Cup 1", 'C')
        _add_insert("Ins 1", 'I')
        _add_insert("Ref 1", 'R')
        _add_insert("Cup 2", 'C')
        _add_insert("Ins 2", 'I')
        _add_insert("Ref 2", 'R')
        _add_insert("Cup 3", 'C', False)
        _add_insert("Ins 3", 'I', False)
        _add_insert("Ref 3", 'R', False)

    def test_form(self):
        view = make_insert_view('C')
        request = self.factory.get('/rest/ui_edit_cup/', data={'action': 'list'})
        data = json.loads(view(request).content.decode('utf-8'))
        self.assertEqual(len(data['objects']), 2)

    def test_delete(self):
        view = make_insert_view('C')
        # make sure the delete deletes and not just marks inactive
        c1 = Insert.objects.get(name="Cup 1")
        self.assertEqual(len(Insert.objects.filter(name="Cup 1")), 1)
        request = self.factory.get('/rest/ui_edit_cup/', data={'action': 'delete', 'id': c1.id})
        view(request)
        self.assertEqual(len(Insert.objects.filter(name="Cup 1")), 0)

    def test_duplicate(self):
        view = make_insert_view('C')
        # make sure I cannot save a duplicate name
        c1 = Insert.objects.get(name="Cup 1")
        request = self.factory.post(
            f'/rest/ui_edit_cup/?id={c1.id}', {
                'action': 'validate',
                'data': {
                    'name': "Cup 2",
                    'volume': 1.234,
                    'temperature': 20.0,
                    'material': 'ot',
                    'other_expansion_um_per_mk': 0.0,
                },
            },
            content_type="application/json; charset=utf-8"
        )
        request.GET = {'id': c1.id}
        data = json.loads(view(request).content.decode('utf-8'))
        self.assertIn('name', data['errors'])

        # make sure I can save a duplicate name of non method.
        request = self.factory.post(
            f'/rest/ui_edit_cup/?id={c1.id}', {
                'action': 'validate',
                'data': {
                    'name': "Cup 3",
                    'volume': 1.234,
                    'temperature': 20.0,
                    'material': 'ot',
                    'other_expansion_um_per_mk': 0.0,
                },
            },
            content_type="application/json; charset=utf-8"
        )
        request.GET = {'id': c1.id}
        data = json.loads(view(request).content.decode('utf-8'))
        self.assertNotIn('name', data['errors'])


class TestInsertMaterialInstrument(TestCase):
    "Test the thermal expansion and material"

    def test_material(self):
        local = Instrument.objects.local()
        local.material = 'Al'
        self.assertEqual(local.thermal_expansion_um_per_mk, 23.4)
        local.material = 'SS'
        self.assertEqual(local.thermal_expansion_um_per_mk, 10.1)
        local.material = 'TC'
        self.assertEqual(local.thermal_expansion_um_per_mk, 5.9)
        local.material = 'ot'
        local.other_expansion_um_per_mk = 1.24
        self.assertEqual(local.thermal_expansion_um_per_mk, 1.24)


class TestInsertMaterial10cm3(TestCase):
    "Test the thermal expansion and material"

    def test_reference(self):
        local = Instrument.objects.local()
        self.assertEqual(local.kind, 'T')
        ref = Insert.objects.create_insert('R')
        self.assertEqual(ref.material, 'TC')
        self.assertEqual(ref.other_expansion_um_per_mk, 0.0)

    def test_cup(self):
        cup = Insert.objects.create_insert('C')
        self.assertEqual(cup.material, 'Al')
        self.assertEqual(cup.other_expansion_um_per_mk, 0.0)

    def test_insert(self):
        ins = Insert.objects.create_insert('I')
        self.assertEqual(ins.material, 'Al')
        self.assertEqual(ins.other_expansion_um_per_mk, 0.0)

    def test_material(self):
        cup = Insert.objects.create_insert('C')
        cup.material = 'Al'
        self.assertEqual(cup.thermal_expansion_um_per_mk, 23.4)
        cup.material = 'SS'
        self.assertEqual(cup.thermal_expansion_um_per_mk, 10.1)
        cup.material = 'TC'
        self.assertEqual(cup.thermal_expansion_um_per_mk, 5.9)
        cup.material = 'ot'
        cup.other_expansion_um_per_mk = 1.24
        self.assertEqual(cup.thermal_expansion_um_per_mk, 1.24)


class TestInsertMaterial100cm3(TestCase):
    "Test the thermal expansion and material"

    def setUp(self):
        local = Instrument.objects.local()
        local.kind = 'H'
        local.save()

    def test_reference(self):
        local = Instrument.objects.local()
        self.assertEqual(local.kind, 'H')
        ref = Insert.objects.create_insert('R')
        self.assertEqual(ref.material, 'SS')
        self.assertEqual(ref.other_expansion_um_per_mk, 0.0)

def _add_insert(name, kind, is_method=True):
    return Insert.objects.create(
        name=name,
        is_method=is_method,
        volume=1.0,
        temperature=22.0,
        material='ot',
        other_expansion_um_per_mk=0.0,
        kind=kind,
    )
