'''
tests for pyc_sample.rest_api
'''
import datetime
import json
from django.test import TestCase, Client
from django.utils import timezone
from django.utils.dateparse import parse_datetime

from mic_python_lib.rest import merge_model_dict

from ..models import (
    AnalysisParameters,
    AnalysisResult1350,
    Calibration,
    Equilibration,
    Expansion,
    Gas,
    Insert,
    Instrument,
    copy_model_obj,
)
from ..services import log_field_changes


# pylint: disable=invalid-name,wrong-import-position,missing-function-docstring
class TestMicroActive(TestCase):
    '''test case for microactive communicaitons.'''

    def setUp(self):
        self.client = Client()

        inst = Instrument.objects.local()
        inst.serial = "test"
        inst.kind = "T"
        inst.material = 'ot'
        inst.other_expansion_um_per_mk = 0.010 / 1.0e-6
        inst.save()

        ref_ball = Insert.objects.create(
            name="ref-vol",
            volume=5,
            temperature=20,
            material='ot',
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="R"
        )
        cup = Insert.objects.create(
            name="cup",
            volume=0.5,
            temperature=20,
            material='ot',
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="C"
        )
        insert = Insert.objects.create(
            name="insert",
            is_method=False,
            volume=1.5,
            temperature=20,
            material='ot',
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="I"
        )
        prms = AnalysisParameters.objects.create(
            is_method=False,
            name="test",
            gas=Gas.objects.get(name="Helium"),
            insert=insert,
            cup=cup,
        )
        expansion = Expansion.objects.create(
            num=3,
            times=[1, 120, 180],
            p0s=[14.6, 14.6, 14.6],
            p1s=[30.0, 30.0, 30.0],
            p2s=[20.0, 20.0, 20.0],
            excludes=[0, 0, 0],
            temp_anl_block=[22.6, 22.6, 22.6],
            temp_valve_block=[22.9, 22.9, 22.9],
        )
        cal = Calibration.objects.create(
            state='C',
            kind='C',
            gas='Helium',
            instrument=inst,
            instrument_thermal_expansion_um_per_mk=0.123 / 1e-6,
            ref_ball=copy_model_obj(Insert, ref_ball),
            volume_fs_sample=10.0,
            volume_fs_reference=10.0,
            volume_fr_sample=10.0,
            volume_fr_reference=10.0,
            temperature=22.0,
        )
        result = AnalysisResult1350.objects.create(
            state='C',
            instrument=inst,
            parameters=prms,
            mass=1.0,
            name="Result 1",
            calibration=cal,
            expansion=expansion,
        )
        self.result_id = result.id

        Equilibration.objects.create(
            analysis=result,
            cycle=0,
            which_equil='1',
            num_points=5,
            times=[1, 2, 3, 4, 5],
            pressures=[28, 26, 24, 22, 20],
        )

    def test_api(self):
        response = self.client.post('/rest/api-supported/', content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content), ['api-v1'])

    def test_microactive(self):
        response = self.client.post(
            '/rest/api-v1/instruments_with_results/', content_type="application/json"
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)['instruments']
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['model'], '1350')
        self.assertEqual(data[0]['serial'], 'test')
        self.assertEqual(data[0]['first'], data[0]['last'])
        response = self.client.post(
            f'/rest/api-v1/get_results_between/?'
            f'model=1350&serial=test'
            f'&first_id={data[0]["first"]}'
            f'&last_id={data[0]["last"]}',
            content_type="application/json"
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)['ids']
        self.assertEqual(len(data), 1)

        # change the instrument thermal expansion and make sure the value is the one from the time
        # of the calibration.
        inst = Instrument.objects.local()
        inst.material = 'ot'
        inst.other_expansion_um_per_mk = 0.020 / 1.0e-6
        inst.save()
        response = self.client.post(
            f'/rest/api-v1/get_microactive_result/?model=1350&serial=test&result_id={data[0]}',
            content_type="application/json"
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertTrue('parameters' in data)
        self.assertEqual(data['parameters']['gas'], 'Helium')
        self.assertEqual(len(data['equilibration_set']), 1)
        self.assertAlmostEqual(data['expansion']['p0s'][0], 14.6, 5)
        self.assertAlmostEqual(data['equilibration_set'][0]['pressures'][-1], 20, 5)
        self.assertAlmostEqual(data['instrument']['thermal_expansion_um_per_mk'], 0.123 / 1.e-6)


class TestShareActive(TestCase):
    '''test case for sharing communicaitons.'''

    def setUp(self):
        self.client = Client()

        inst = Instrument.objects.local()
        inst.serial = "0001"
        inst.kind = "T"
        inst.material = 'ot'
        inst.other_expansion_um_per_mk = 0.010 / 1.0e-6
        inst.save()

        inst2 = Instrument.objects.create(
            serial="0002",
            is_local=False,
            kind="T",
            material='ot',
            other_expansion_um_per_mk=0.010 / 1.0e-6,
        )
        inst2.last_updated = timezone.now() - datetime.timedelta(days=4)
        inst2.save(update=False)
        ref_ball = Insert.objects.create(
            name="ref-vol",
            is_shared=True,
            volume=5,
            temperature=20,
            material='ot',
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="R"
        )
        cup = Insert.objects.create(
            name="cup",
            is_shared=True,
            volume=0.5,
            temperature=20,
            material='ot',
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="C"
        )
        insert = Insert.objects.create(
            name="insert",
            is_shared=True,
            volume=1.5,
            temperature=20,
            material='ot',
            other_expansion_um_per_mk=0.011 / 1.0e-6,
            kind="I"
        )
        AnalysisParameters.objects.create(
            name="MTH 1",
            is_method=True,
            is_shared=True,
            is_default=False,
            gas=Gas.objects.get(name="Helium"),
            insert=insert,
            cup=cup,
            do_pre_purges=True,
        )
        method2 = AnalysisParameters.objects.create(
            name="MTH 2",
            is_method=True,
            is_shared=True,
            gas=Gas.objects.get(name="Helium"),
            insert=insert,
            cup=cup,
            do_pre_purges=True,
        )
        method2.last_updated = timezone.now() - datetime.timedelta(days=4)
        method2.save(update=False)
        prms = AnalysisParameters.objects.create(
            name="test",
            is_method=True,
            is_shared=True,
            gas=Gas.objects.get(name="Helium"),
            insert=insert,
            cup=cup,
        )
        for inst in Instrument.objects.all():
            cal = Calibration.objects.create(
                state='C',
                kind='C',
                gas='Helium',
                instrument=inst,
                instrument_thermal_expansion_um_per_mk=0.123 / 1e-6,
                ref_ball=copy_model_obj(Insert, ref_ball),
                volume_fs_sample=10.0,
                volume_fs_reference=10.0,
                volume_fr_sample=10.0,
                volume_fr_reference=10.0,
                temperature=22.0,
            )

            expansion = Expansion.objects.create(
                num=3,
                times=[1, 120, 180],
                p0s=[14.6, 14.6, 14.6],
                p1s=[30.0, 30.0, 30.0],
                p2s=[20.0, 20.0, 20.0],
                excludes=[0, 0, 0],
                temp_anl_block=[22.6, 22.6, 22.6],
                temp_valve_block=[22.9, 22.9, 22.9],
            )
            result = AnalysisResult1350.objects.create(
                state='C',
                instrument=inst,
                parameters=AnalysisParameters.objects.copy_method(prms.id),
                mass=1.0,
                name="Result 1",
                operator="Django",
                calibration=cal,
                expansion=expansion,
            )

            Equilibration.objects.create(
                analysis=result,
                cycle=0,
                which_equil='1',
                num_points=5,
                times=[1, 2, 3, 4, 5],
                pressures=[28, 26, 24, 22, 20],
            )
            org = AnalysisResult1350.objects.get(id=result.id)
            result.mass = 1.234
            log_field_changes(result, org, result, ['mass'])
            result.save(update=False)

            cal2 = Calibration.objects.create(
                state='C',
                kind='C',
                gas='Helium',
                start_time=timezone.now() - datetime.timedelta(days=4),
                instrument=inst,
                instrument_thermal_expansion_um_per_mk=0.123 / 1e-6,
                ref_ball=copy_model_obj(Insert, ref_ball),
                volume_fs_sample=10.0,
                volume_fs_reference=10.0,
                volume_fr_sample=10.0,
                volume_fr_reference=10.0,
                temperature=22.0,
            )
            cal2.last_updated = timezone.now() - datetime.timedelta(days=4)
            cal2.save(update=False)

            expansion = Expansion.objects.create(
                num=3,
                times=[1, 120, 180],
                p0s=[14.6, 14.6, 14.6],
                p1s=[30.0, 30.0, 30.0],
                p2s=[20.0, 20.0, 20.0],
                excludes=[0, 0, 0],
                temp_anl_block=[22.6, 22.6, 22.6],
                temp_valve_block=[22.9, 22.9, 22.9],
            )
            result2 = AnalysisResult1350.objects.create(
                state='C',
                start_time=timezone.now() - datetime.timedelta(days=4),
                instrument=inst,
                parameters=AnalysisParameters.objects.copy_method(prms.id),
                mass=1.0,
                name="Result 1",
                operator="Django",
                calibration=cal,
                expansion=expansion,
            )
            result2.last_updated = timezone.now() - datetime.timedelta(days=4)
            result2.save(update=False)

    def test_last_updates(self):
        response = self.client.get(
            '/rest/api-v1/get_last_updates/',
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertIn('instrument', data)
        parse_datetime(data['instrument'])  # make sue it parses
        self.assertIn('method', data)
        parse_datetime(data['method'])  # make sue it parses
        self.assertIn('insert', data)
        parse_datetime(data['insert'])  # make sue it parses
        self.assertIn('result', data)
        for res in data['result']:
            self.assertIn('last_updated', res)
            parse_datetime(res['last_updated'])  # make sue it parses
            self.assertIn('model', res)
            self.assertIn('serial', res)
        self.assertIn('calibration', data)
        for cal in data['calibration']:
            self.assertIn('last_updated', cal)
            parse_datetime(cal['last_updated'])  # make sue it parses
            self.assertIn('model', cal)
            self.assertIn('serial', cal)

    def test_get_insert(self):
        response = self.client.get(
            '/rest/api-v1/get_insert_ids/',
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 3)
        response = self.client.get(
            '/rest/api-v1/get_insert_ids/',
            {'since': timezone.now() - datetime.timedelta(days=2)},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        oids = json.loads(response.content)['ids']
        self.assertEqual(len(oids), 3)
        for oid in oids:
            response = self.client.get(
                '/rest/api-v1/insert/',
                {'oid': oid},
                content_type="application/json",
            )
            self.assertEqual(response.status_code, 200)
            obj = json.loads(response.content)
            _check_model_fields(Insert, obj, is_method=True, is_shared=True)

    def test_get_method(self):
        response = self.client.get(
            '/rest/api-v1/get_method_ids/',
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 3)
        response = self.client.get(
            '/rest/api-v1/get_method_ids/',
            {'since': timezone.now() - datetime.timedelta(days=2)},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        oids = json.loads(response.content)['ids']
        self.assertEqual(len(oids), 2)
        for oid in oids:
            response = self.client.get(
                '/rest/api-v1/method/',
                {'oid': oid},
                content_type="application/json",
            )
            self.assertEqual(response.status_code, 200)
            obj = json.loads(response.content)
            _check_model_fields(AnalysisParameters, obj,
                                is_method=True, is_shared=True, is_default=False)

    def test_get_instrument(self):
        response = self.client.get(
            '/rest/api-v1/get_instrument_ids/',
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 2)
        response = self.client.get(
            '/rest/api-v1/get_instrument_ids/',
            {'since': timezone.now() - datetime.timedelta(days=2)},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        oids = json.loads(response.content)['ids']
        self.assertEqual(len(oids), 1)
        for oid in oids:
            response = self.client.get(
                '/rest/api-v1/instrument/',
                {'oid': oid},
                content_type="application/json",
            )
            self.assertEqual(response.status_code, 200)
            obj = json.loads(response.content)
            _check_model_fields(Instrument, obj, is_local=False)

    def test_get_calibration(self):
        response = self.client.get(
            '/rest/api-v1/get_calibration_ids/',
            {
                'model': '1350',
                'serial': '0002'
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 2)

        response = self.client.get(
            '/rest/api-v1/get_calibration_ids/',
            {
                'model': '1350',
                'serial': '0001',
                'since': timezone.now() - datetime.timedelta(days=2),
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        oids = json.loads(response.content)['ids']
        self.assertEqual(len(oids), 1)
        for oid in oids:
            response = self.client.get(
                '/rest/api-v1/calibration/',
                {'oid': oid},
                content_type="application/json",
            )
            self.assertEqual(response.status_code, 200)
            obj = json.loads(response.content)
            _check_model_fields(Calibration, obj, state='C', kind='C', check_of=None)

    def test_get_result(self):
        response = self.client.get(
            '/rest/api-v1/get_result_ids/',
            {
                'model': '1350',
                'serial': '0002'
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 2)

        response = self.client.get(
            '/rest/api-v1/get_result_ids/',
            {
                'model': '1350',
                'serial': '0001',
                'since': timezone.now() - datetime.timedelta(days=2),
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        oids = json.loads(response.content)['ids']
        self.assertEqual(len(oids), 1)
        for oid in oids:
            response = self.client.get(
                '/rest/api-v1/result/',
                {'oid': oid},
                content_type="application/json",
            )
            self.assertEqual(response.status_code, 200)
            obj = json.loads(response.content)

            _check_model_fields(AnalysisResult1350, obj, state='C', results=[])

            # make sure we assert on missing fields and on added fields.
            self.assertRaises(
                AssertionError,
                lambda obj=obj: _check_model_fields(AnalysisResult1350, obj, state='C')
            )
            obj['extra_set'] = True
            self.assertRaises(
                AssertionError,
                lambda obj=obj: _check_model_fields(AnalysisResult1350, obj, state='C', results=[])
            )


class TestShareActiveEmpty(TestCase):
    '''test case for sharing communicaitons with no data.'''

    def setUp(self):
        self.client = Client()

        inst = Instrument.objects.local()
        inst.serial = "0001"
        inst.kind = "T"
        inst.material = 'ot'
        inst.other_expansion_um_per_mk = 0.010 / 1.0e-6
        inst.save()

    def test_last_updates(self):
        response = self.client.get(
            '/rest/api-v1/get_last_updates/',
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.assertIn('instrument', data)
        parse_datetime(data['instrument'])  # make sue it parses
        self.assertIn('method', data)
        parse_datetime(data['method'])  # make sue it parses
        self.assertIn('insert', data)
        parse_datetime(data['insert'])  # make sue it parses
        self.assertIn('result', data)
        self.assertEqual(len(data['result']), 0)
        self.assertIn('calibration', data)
        self.assertEqual(len(data['calibration']), 0)

    def test_get_inserts(self):
        response = self.client.get(
            '/rest/api-v1/get_insert_ids/',
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 0)

    def test_get_method(self):
        response = self.client.get(
            '/rest/api-v1/get_method_ids/',
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 0)

    def test_get_instrument(self):
        response = self.client.get(
            '/rest/api-v1/get_instrument_ids/',
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 1)

    def test_get_calibration(self):
        response = self.client.get(
            '/rest/api-v1/get_calibration_ids/',
            {
                'model': '1350',
                'serial': '0002'
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 0)

    def test_get_result(self):
        response = self.client.get(
            '/rest/api-v1/get_result_ids/',
            {
                'model': '1350',
                'serial': '0002'
            },
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(json.loads(response.content)['ids']), 0)


# =================================================================================================
def _missing_field(name):
    assert False, f'"{name}" is missing from the object.'


def _extra_field(name):
    assert False, f'"{name}" is extra in the object.'


def _check_model_fields(model, obj, **kwargs):
    merge_model_dict(model, obj, kwargs, on_missing=_missing_field, on_extra=_extra_field)
