"""Test the sharing of results."""
import contextlib
import json
from django.test import TestCase
from django.utils.dateparse import parse_datetime
from mic_python_lib import rest
from mic_python_lib.models import KnownInstrument
from pyc_sample.rest_api import REST_API_DATA_1350
from pyc_sample.models import (
    AnalysisParameters,
    AnalysisResult1345,
    AnalysisResult1350,
    Calibration,
    Insert,
    Instrument,
)

# pylint: disable=invalid-name,wrong-import-position,missing-function-docstring, too-many-lines
# ==================================================================================================
LAST_UPDATE_0 = '2022-04-20T18:45:07.109Z'
LAST_UPDATE_1 = '2022-04-21T18:45:07.109Z'
LAST_UPDATE_2 = '2022-04-22T18:45:07.109Z'
LAST_UPDATE_FUTURE = '2122-04-22T18:45:07.109Z'

class MockShareServer:
    "A mock server that can be setup with shared data and return json objects."

    def __init__(self):
        self.inserts = {}
        self.method = {}
        self.instruments = {}
        self.calibrations = {}
        self.results = {}
        self.results1345 = {}
        self.inst_config = Instrument.local_config()

    def set_inst_config(self, **kwargs):
        self.inst_config = kwargs

    def add_insert(self, oid, insert_dict):
        self.inserts[oid] = insert_dict

    def update_insert(self, oid, new_update, **kwargs):
        self.inserts[oid]['last_updated'] = new_update
        self.inserts[oid].update(kwargs)

    def add_method(self, oid, method_dict):
        self.method[oid] = method_dict

    def update_method(self, oid, new_update, **kwargs):
        self.method[oid]['last_updated'] = new_update
        self.method[oid].update(kwargs)

    def add_instrument(self, oid, inst_dict):
        self.instruments[oid] = inst_dict

    def update_instrument(self, oid, new_update, **kwargs):
        self.instruments[oid]['last_updated'] = new_update
        self.instruments[oid].update(kwargs)

    def add_calibration(self, oid, cal_dict):
        self.calibrations[oid] = cal_dict

    def update_calibration(self, oid, new_update, **kwargs):
        self.calibrations[oid]['last_updated'] = new_update
        self.calibrations[oid].update(kwargs)

    def add_result(self, oid, inst_dict):
        self.results[oid] = inst_dict

    def add_result1345(self, oid, inst_dict):
        self.results1345[oid] = inst_dict

    def update_result_param(self, oid, new_update, **kwargs):
        self.results[oid]['last_updated'] = new_update
        self.results[oid]['parameters'].update(kwargs)

    def page(self, url, params):
        params = json.loads(json.dumps(params))  # Fake the url conversion.
        if url in MockShareServer._URLS:
            return MockShareServer._URLS[url](self, **params)
        return None  # pragma: no cover

    def _insert_ids(self):
        return {
            'ids': sorted(self.inserts.keys()),
            'inst_config': self.inst_config,
        }

    def _insert(self, oid):
        return self.inserts[oid]

    def _method_ids(self):
        return {
            'ids': sorted(self.method.keys()),
            'inst_config': self.inst_config,
        }

    def _method(self, oid):
        return self.method[oid]

    def _instrument_ids(self):
        return {'ids': sorted(self.instruments.keys())}

    def _instrument(self, oid):
        return self.instruments[oid]

    def _calibration_ids(self, model, serial, since=None):
        if since is not None:
            since = parse_datetime(since)
        return {
            'ids':
                sorted([
                    k for k, v in self.calibrations.items() if (
                        v['instrument']['model'] == model and v['instrument']['serial'] == serial
                        and (since is None or parse_datetime(v['last_updated']) > since)
                    )
                ])
        }

    def _calibration(self, oid):
        return self.calibrations[oid]

    def _result_ids(self, model, serial, since=None):
        if since is not None:
            since = parse_datetime(since)
        return {
            'ids':
                sorted([
                    k for k, v in self.results.items() if (
                        v['instrument']['model'] == model and v['instrument']['serial'] == serial
                        and (since is None or parse_datetime(v['last_updated']) > since)
                    )
                ])
        }

    def _result(self, oid):
        return self.results[oid]

    def _result1345_ids(self, model, serial, since=None):
        if since is not None:
            since = parse_datetime(since)
        return {
            'ids':
            sorted([
                k for k, v in self.results1345.items() if (
                    v['instrument']['model'] == model and v['instrument']['serial'] == serial
                    and (since is None or parse_datetime(v['last_updated']) > since)
                )
            ])
        }

    def _result1345(self, oid):
        return self.results1345[oid]

    def _inst_config(self):
        return self.inst_config

    def _last_updates(self):
        cal_inst = {}
        for _, result in self.calibrations.items():
            mod_ser = (result['instrument']['model'], result['instrument']['serial'])
            if mod_ser not in cal_inst:
                cal_inst[mod_ser] = result['last_updated']
            elif parse_datetime(result['last_updated']) > parse_datetime(cal_inst[mod_ser]):
                cal_inst[mod_ser] = result['last_updated']

        res_inst = {}
        for _, result in self.results.items():
            mod_ser = (result['instrument']['model'], result['instrument']['serial'])
            if mod_ser not in res_inst:
                res_inst[mod_ser] = result['last_updated']
            elif parse_datetime(result['last_updated']) > parse_datetime(res_inst[mod_ser]):
                res_inst[mod_ser] = result['last_updated']

        res1345_inst = {}
        for _, result in self.results1345.items():
            mod_ser = (result['instrument']['model'], result['instrument']['serial'])
            if mod_ser not in res1345_inst:
                res1345_inst[mod_ser] = result['last_updated']
            elif parse_datetime(result['last_updated']) > parse_datetime(res1345_inst[mod_ser]):
                res1345_inst[mod_ser] = result['last_updated']

        ret = {
            'calibration': [{
                'model': k[0],
                'serial': k[1],
                'last_updated': v
            } for k, v in cal_inst.items()],
            'result': [{
                'model': k[0],
                'serial': k[1],
                'last_updated': v
            } for k, v in res_inst.items()],
            'result1345': [{
                'model': k[0],
                'serial': k[1],
                'last_updated': v
            } for k, v in res1345_inst.items()],
        }
        return ret

    _URLS = {
        'http://local/rest/api-v1/instrument_config/': _inst_config,
        'http://local/rest/api-v1/get_last_updates/': _last_updates,
        'http://local/rest/api-v1/get_insert_ids/': _insert_ids,
        'http://local/rest/api-v1/insert/': _insert,
        'http://local/rest/api-v1/get_method_ids/': _method_ids,
        'http://local/rest/api-v1/method/': _method,
        'http://local/rest/api-v1/get_instrument_ids/': _instrument_ids,
        'http://local/rest/api-v1/instrument/': _instrument,
        'http://local/rest/api-v1/get_calibration_ids/': _calibration_ids,
        'http://local/rest/api-v1/calibration/': _calibration,
        'http://local/rest/api-v1/get_result_ids/': _result_ids,
        'http://local/rest/api-v1/get_result1345_ids/': _result1345_ids,
        'http://local/rest/api-v1/result/': _result,
        'http://local/rest/api-v1/result1345/': _result1345,
    }


class MockRequests:  #pylint: disable=too-few-public-methods
    "The class that will mock the requests module in these tests."
    _CONTEXT = None

    @staticmethod
    @contextlib.contextmanager
    def http_context():
        "Create a MockRequests context that has a MockShareServer"
        site = MockShareServer()
        MockRequests._CONTEXT = site
        yield site
        MockRequests._CONTEXT = None

    @staticmethod
    def get(url, params):
        "Fake the json get call in the requests module. "

        class Object:
            "The pretend return object."

            def __init__(self):
                self.jsondata = None
                self.status_code = None

            def json(self):
                "Fake the get return object json call be giving the mocked json."
                return self.jsondata

        ret = Object()
        ret.jsondata = MockRequests._CONTEXT.page(url, params) if MockRequests._CONTEXT else None
        if ret.jsondata is None:
            ret.status_code = 404
        else:
            ret.status_code = 200
        return ret


# =================================================================================================
class TestUpdates(TestCase):
    "Test getting when we don't have any (other instrument data) already."

    def run(self, result=None):
        requests = rest.requests
        rest.requests = MockRequests
        with MockRequests.http_context() as site:
            self.site = site  # pylint: disable=attribute-defined-outside-init
            _add_test_data(self.site)
            super(TestUpdates, self).run(result)
        rest.requests = requests

    def test_bad_inst(self):
        self.assertRaises(
            rest.BadInstrument, lambda: REST_API_DATA_1350.update_from_instrument('bad'))

    def test_update_methods(self):
        REST_API_DATA_1350.update_from_instrument('local')
        # get them an second time to test no update branch.
        REST_API_DATA_1350.update_from_instrument('local')

        self.assertEqual(len(Insert.objects.methods()), 5)
        self.site.update_insert(1, LAST_UPDATE_2, volume=1.234)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.methods()), 5)
        self.assertEqual(Insert.objects.methods().get(name='Reference').volume, 1.234)
        self.site.update_insert(1, LAST_UPDATE_2, volume=2.34)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.methods()), 5)
        self.assertEqual(Insert.objects.methods().get(name='Reference').volume, 1.234)

        self.assertEqual(len(AnalysisParameters.objects.methods()), 3)
        self.site.update_method(22, LAST_UPDATE_2, temperature=55.0)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(AnalysisParameters.objects.methods()), 3)
        self.assertEqual(
            AnalysisParameters.objects.methods().get(name='Method 2').temperature, 55.0
        )
        self.site.update_method(22, LAST_UPDATE_2, temperature=45.0)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(AnalysisParameters.objects.methods()), 3)
        self.assertEqual(
            AnalysisParameters.objects.methods().get(name='Method 2').temperature, 55.0
        )

        self.assertEqual(len(Instrument.objects.all()), 3)
        self.site.update_instrument(1, LAST_UPDATE_2, other_expansion_um_per_mk=10.0)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Instrument.objects.all()), 3)
        self.assertEqual(
            Instrument.objects.get(model='1350', serial='0001').other_expansion_um_per_mk, 10.0
        )
        self.site.update_instrument(1, LAST_UPDATE_2, other_expansion_um_per_mk=30.0)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Instrument.objects.all()), 3)
        self.assertEqual(
            Instrument.objects.get(model='1350', serial='0001').other_expansion_um_per_mk, 10.0
        )

    def test_update_calibrations(self):
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Calibration.objects.all()), 2)
        cal = Calibration.objects.get(
            instrument__model='1350',
            instrument__serial='0001',
            start_time='2022-02-02T15:07:16.885Z'
        )
        self.site.update_calibration(1, LAST_UPDATE_2, fill_pressure=30.0)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Calibration.objects.all()), 2)
        cal = Calibration.objects.get(
            instrument__model='1350',
            instrument__serial='0001',
            start_time='2022-02-02T15:07:16.885Z'
        )
        self.assertEqual(cal.fill_pressure, 30.0)
        self.site.update_calibration(1, LAST_UPDATE_2, fill_pressure=10.0)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Calibration.objects.all()), 2)
        cal = Calibration.objects.get(
            instrument__model='1350',
            instrument__serial='0001',
            start_time='2022-02-02T15:07:16.885Z'
        )
        self.assertEqual(cal.fill_pressure, 30.0)

    def test_update_restults(self):
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(AnalysisResult1350.objects.all()), 12)
        self.site.update_result_param(3, LAST_UPDATE_2, temperature=30)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(AnalysisResult1350.objects.all()), 12)
        res = AnalysisResult1350.objects.get(
            instrument__model='1350',
            instrument__serial='0001',
            start_time='2022-02-03T19:08:29.100Z'
        )
        self.assertEqual(res.parameters.temperature, 30)
        self.site.update_result_param(3, LAST_UPDATE_2, temperature=10)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(AnalysisResult1350.objects.all()), 12)
        res = AnalysisResult1350.objects.get(
            instrument__model='1350',
            instrument__serial='0001',
            start_time='2022-02-03T19:08:29.100Z'
        )
        self.assertEqual(res.parameters.temperature, 30)

        res = AnalysisResult1345.objects.get()
        self.assertEqual(res.temperature, 20.009)


# =================================================================================================
class TestInsertMethods(TestCase):
    "Test getting when we don't have any (other instrument data) already."

    def run(self, result=None):
        requests = rest.requests
        rest.requests = MockRequests
        with MockRequests.http_context() as site:
            self.site = site  # pylint: disable=attribute-defined-outside-init
            super().run(result)
        rest.requests = requests

    def test_share_cup(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)

        # make sure we get one when it is available.
        self.site.add_insert(1, {
            'name': 'Cup',
            'volume': 1.0,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_1
        })
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertEqual(Insert.objects.get(kind='C').volume, 1.0)
        self.assertTrue(Insert.objects.get(kind='C').is_shared)

        # makes sure we don't override with older cups.
        self.site.update_insert(1, LAST_UPDATE_0, volume=10.0)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertEqual(Insert.objects.get(kind='C').volume, 1.0)

        # makes sure we do override with older cups.
        self.site.update_insert(1, LAST_UPDATE_2, volume=10.0)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertEqual(Insert.objects.get(kind='C').volume, 10.0)

    def test_nonshared(self):
        # make a local one that is not shared.
        cup = Insert.objects.create(
            kind='C',
            name='Cup',
            is_method=True,
            volume=1.0,
            material='ot',
            other_expansion_um_per_mk=23.6,
        )
        cup.last_updated = LAST_UPDATE_0
        cup.save(update=False)

        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertFalse(Insert.objects.get(kind='C').is_shared)
        self.assertEqual(Insert.objects.get(kind='C').volume, 1.0)

        # make sure we do not override the local one since it is not shared.
        self.site.add_insert(1, {
            'name': 'Cup',
            'volume': 10.0,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_2
        })
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertFalse(Insert.objects.get(kind='C').is_shared)
        self.assertEqual(Insert.objects.get(kind='C').volume, 1.0)

        # make sure we do override the local one if it is shared.
        cup = Insert.objects.get(kind='C')
        cup.is_shared = True
        cup.save(update=False)
        KnownInstrument.objects.all().delete()  # reset previous updates.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertTrue(Insert.objects.get(kind='C').is_shared)
        self.assertEqual(Insert.objects.get(kind='C').volume, 10.0)


class TestDeleteMethods(TestCase):
    "Test locally deleting inserts and not updating."

    def run(self, result=None):
        requests = rest.requests
        rest.requests = MockRequests
        with MockRequests.http_context() as site:
            self.site = site  # pylint: disable=attribute-defined-outside-init
            super().run(result)
        rest.requests = requests

    def test_delete_cup(self):
        # make sure we don't have any now.
        cup = Insert.objects.create(
            kind='C',
            name='Cup',
            is_method=True,
            is_shared=True,
            volume=1.0,
            material='ot',
            other_expansion_um_per_mk=23.6,
        )
        cup.last_updated = LAST_UPDATE_0
        cup.save(update=False)
        self.site.add_insert(1, {
            'name': 'Cup',
            'volume': 10.0,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_1
        })
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)

        # make sure we get one when it is available.
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertEqual(Insert.objects.get(kind='C').volume, 10.0)
        self.assertTrue(Insert.objects.get(kind='C').is_shared)

        # Delete locally and make sure it does not come back.
        cup.delete()
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)

        # Update the remote one, and make sure it comes back.
        self.site.update_insert(1, LAST_UPDATE_FUTURE)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)


# =================================================================================================
class TestMethodsInstrument(TestCase):
    "Test getting when we don't have any (other instrument data) already."

    def run(self, result=None):
        requests = rest.requests
        rest.requests = MockRequests
        with MockRequests.http_context() as site:
            self.site = site  # pylint: disable=attribute-defined-outside-init
            super().run(result)
        rest.requests = requests

    def test_compatible_inst(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.site.add_insert(1, {
            'name': 'Cup',
            'volume': 1.0,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_1
        })

        # make sure we get one when the instrument is compatible.
        self.site.set_inst_config(model='1350', serial='1234', kind='T', is_temp_control=False)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertEqual(Insert.objects.get(kind='C').volume, 1.0)
        self.assertTrue(Insert.objects.get(kind='C').is_shared)

    def test_bad_volume(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.site.add_insert(1, {
            'name': 'Cup',
            'volume': 1.0,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_1
        })

        # make sure we get one when the instrument is compatible.
        self.site.set_inst_config(model='1350', serial='1234', kind='H', is_temp_control=False)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)

    def test_bad_temp(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.site.add_insert(1, {
            'name': 'Cup',
            'volume': 1.0,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_1
        })

        # make sure we get one when the instrument is compatible.
        self.site.set_inst_config(model='1350', serial='1234', kind='T', is_temp_control=True)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)

    def test_bad_model(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.site.add_insert(1, {
            'name': 'Cup',
            'volume': 1.0,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_1
        })

        # make sure we get one when the instrument is compatible.
        self.site.set_inst_config(model='1351', serial='1234', kind='T', is_temp_control=False)
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)


# =================================================================================================
class TestMethods(TestCase):
    "Test getting when we don't have any (other instrument data) already."

    def run(self, result=None):
        requests = rest.requests
        rest.requests = MockRequests
        with MockRequests.http_context() as site:
            self.site = site  # pylint: disable=attribute-defined-outside-init
            super().run(result)
        rest.requests = requests

    def test_no_cup(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # Share a method without a cup.
        self.site.add_insert(1, _insert(volume=10))
        self.site.add_method(1, _method(name='Test', cup=None))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertIsNone(AnalysisParameters.objects.get(name='Test').cup)

    def test_sunshared_cup(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # Share a cup and method using an unshared cup.
        self.site.add_method(1, _method(name='Test', cup=_insert(volume=1)))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertIsNone(AnalysisParameters.objects.get(name='Test').cup)

    def test_shared_cup(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # Share a cup and method using that cup.
        self.site.add_insert(1, _insert(volume=10))
        self.site.add_method(1, _method(name='Test', cup=_insert(volume=1)))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertEqual(AnalysisParameters.objects.get(name='Test').cup.volume, 10)

    def test_single_local_cup(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # make a local cup, and remote method with Cup
        Insert.objects.create(
            kind='C',
            name='Local Cup',
            is_method=True,
            volume=1.0,
            material='ot',
            other_expansion_um_per_mk=23.6,
        )
        self.site.add_method(1, _method(name='Test', cup=_insert(volume=1, name="Remote Cup")))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 1)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertEqual(AnalysisParameters.objects.get(name='Test').cup.name, "Local Cup")

    def test_multiple_local_cup(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # make a local cup, and remote method with Cup
        Insert.objects.create(
            kind='C',
            name='Cup 1',
            is_method=True,
            volume=1.0,
            material='ot',
            other_expansion_um_per_mk=23.6,
        )
        Insert.objects.create(
            kind='C',
            name='Cup 2',
            is_method=True,
            volume=1.0,
            material='ot',
            other_expansion_um_per_mk=23.6,
        )
        self.site.add_method(1, _method(name='Test', cup=_insert(volume=1, name="Remote Cup")))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='C')), 2)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertIsNone(AnalysisParameters.objects.get(name='Test').cup)

    def test_no_insert(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='I')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # Share a method without a cup.
        self.site.add_insert(1, _insert(volume=10, kind='I'))
        self.site.add_method(1, _method(name='Test', insert=None))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='I')), 1)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertIsNone(AnalysisParameters.objects.get(name='Test').insert)

    def test_shared_insert(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='I')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # Share a cup and method using that cup.
        self.site.add_insert(1, _insert(volume=10, kind='I'))
        self.site.add_method(1, _method(name='Test', insert=_insert(volume=1, kind='I')))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='I')), 1)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertEqual(AnalysisParameters.objects.get(name='Test').insert.volume, 10)

    def test_single_local_insert(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # make a local cup, and remote method with Cup
        Insert.objects.create(
            kind='I',
            name='Local Cup',
            is_method=True,
            volume=1.0,
            material='ot',
            other_expansion_um_per_mk=23.6,
        )
        self.site.add_method(1, _method(
            name='Test', insert=_insert(volume=1, name="Remote Cup", kind='I')))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='I')), 1)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertIsNone(AnalysisParameters.objects.get(name='Test').insert)

    def test_multiple_local_insert(self):
        # make sure we don't have any now.
        self.assertEqual(len(Insert.objects.filter(kind='C')), 0)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 1)

        # make a local cup, and remote method with Cup
        Insert.objects.create(
            kind='I',
            name='Insert 1',
            is_method=True,
            volume=1.0,
            material='ot',
            other_expansion_um_per_mk=23.6,
        )
        Insert.objects.create(
            kind='I',
            name='Insert 2',
            is_method=True,
            volume=1.0,
            material='ot',
            other_expansion_um_per_mk=23.6,
        )
        self.site.add_method(1, _method(
            name='Test', insert=_insert(volume=1, name="Remote Cup", kind='I')))
        REST_API_DATA_1350.update_from_instrument('local')
        self.assertEqual(len(Insert.objects.filter(kind='I')), 2)
        self.assertEqual(len(AnalysisParameters.objects.filter(is_method=True)), 2)
        self.assertIsNone(AnalysisParameters.objects.get(name='Test').insert)


# =================================================================================================
def _add_test_data(site):
    # pylint: disable=line-too-long
    site.add_insert(
        1, {
            'name': 'Reference',
            'volume': 0.718522,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.2,
            'kind': 'R',
            'last_updated': LAST_UPDATE_1,
        }
    )
    site.add_insert(
        2, {
            'name': 'Cup',
            'volume': 0.2174,
            'temperature': 22.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_1
        }
    )
    site.add_insert(
        4, {
            'name': 'ref 10',
            'volume': 6.371259,
            'temperature': 20.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 0.0,
            'kind': 'R',
            'last_updated': LAST_UPDATE_1
        }
    )
    site.add_insert(
        5, {
            'name': 'cup 10',
            'volume': 0.795778,
            'temperature': 20.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.6,
            'kind': 'C',
            'last_updated': LAST_UPDATE_1
        }
    )
    site.add_insert(
        6, {
            'name': 'Insert',
            'volume': 2.0,
            'temperature': 20.0,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.2,
            'kind': 'I',
            'last_updated': LAST_UPDATE_1
        }
    )
    site.add_method(
        21, {
            'name': '5 sf',
            'gas': 'Helium',
            'temperature': 20.0,
            'use_vacuum_prep': False,
            'cup': {
                'name': 'Cup',
                'volume': 0.2174,
                'temperature': 22.0,
                'thermal_expansion': 2.36e-05
            },
            'insert': None,
            'fill_direction': 'S',
            'do_pre_purges': True,
            'purge_fill_pressure': 19.5,
            'num_purges': 10,
            'num_cycles': 5,
            'cycle_fill_pressure': 19.5,
            'endby': 'E',
            'interval_time': 10,
            'equilibration_rate': 0.005,
            'p1_record_equilibration_data': False,
            'p1_equilibration_cycles': 5,
            'p1_equilibration_points': 500,
            'p2_record_equilibration_data': False,
            'p2_equilibration_cycles': 5,
            'p2_equilibration_points': 500,
            'technique': 'P',
            'resin_density': 1.0,
            'initial_pressure': 19.5,
            'final_pressure': 19.5,
            'pressure_increment': 19.5,
            'low_pressure': 19.5,
            'fracture_pressure': 19.5,
            'use_run_precision': False,
            'run_precision_tolerance': 0.05,
            'compute_volume': 'C',
            'geometric_volume': 10.0,
            'active_area': 1.0,
            'shape': 'U',
            'cube_edge_length': 1.0,
            'cylinder_diameter': 1.0,
            'cylinder_length': 1.0,
            'rect_long_edge': 1.0,
            'rect_short_edge': 1.0,
            'rect_other_edge': 1.0,
            'number_of_pieces': 1,
            'measure_by': 'L',
            'chord_length': 1.0,
            'cell_diameter': 1.0,
            'cube_number_of_skins': 0,
            'cyl_number_of_skins': 0,
            'rect_number_of_large_skins': 0,
            'rect_number_of_small_skins': 0,
            'rect_number_of_other_skins': 0,
            'last_updated': LAST_UPDATE_1,
        }
    )
    site.add_method(
        22, {
            'name': 'Method 2',
            'gas': 'Helium',
            'temperature': 20.0,
            'use_vacuum_prep': False,
            'cup': None,
            'insert': {
                'name': 'Insert',
                'volume': 0.2174,
                'temperature': 22.0,
                'thermal_expansion': 2.36e-05
            },
            'fill_direction': 'S',
            'do_pre_purges': True,
            'purge_fill_pressure': 19.5,
            'num_purges': 10,
            'num_cycles': 5,
            'cycle_fill_pressure': 19.5,
            'endby': 'E',
            'interval_time': 10,
            'equilibration_rate': 0.005,
            'p1_record_equilibration_data': False,
            'p1_equilibration_cycles': 5,
            'p1_equilibration_points': 500,
            'p2_record_equilibration_data': False,
            'p2_equilibration_cycles': 5,
            'p2_equilibration_points': 500,
            'technique': 'P',
            'resin_density': 1.0,
            'initial_pressure': 19.5,
            'final_pressure': 19.5,
            'pressure_increment': 19.5,
            'low_pressure': 19.5,
            'fracture_pressure': 19.5,
            'use_run_precision': False,
            'run_precision_tolerance': 0.05,
            'compute_volume': 'C',
            'geometric_volume': 10.0,
            'active_area': 1.0,
            'shape': 'U',
            'cube_edge_length': 1.0,
            'cylinder_diameter': 1.0,
            'cylinder_length': 1.0,
            'rect_long_edge': 1.0,
            'rect_short_edge': 1.0,
            'rect_other_edge': 1.0,
            'number_of_pieces': 1,
            'measure_by': 'L',
            'chord_length': 1.0,
            'cell_diameter': 1.0,
            'cube_number_of_skins': 0,
            'cyl_number_of_skins': 0,
            'rect_number_of_large_skins': 0,
            'rect_number_of_small_skins': 0,
            'rect_number_of_other_skins': 0,
            'last_updated': LAST_UPDATE_1,
        }
    )
    site.add_instrument(
        1, {
            'model': '1350',
            'serial': '0001',
            'kind': 'T',
            'is_temp_control': True,
            'material': 'ot',
            'other_expansion_um_per_mk': 23.2,
            'device_service': '2022-02-01',
            'device_warranty': '2022-02-01',
            'last_updated': LAST_UPDATE_1
        }
    )
    site.add_instrument(
        2, {
            'model': '1345',
            'serial': '4467',
            'kind': 'T',
            'is_temp_control': False,
            'material': 'ot',
            'other_expansion_um_per_mk': 0.0,
            'device_service': '2023-04-19',
            'device_warranty': '2023-04-19',
            'last_updated': LAST_UPDATE_1
        }
    )

    site.add_calibration(
        1, {
            'start_time': '2022-02-02T15:07:16.885Z',
            'last_updated': LAST_UPDATE_0,
            'fill_pressure': 19.5,
            'collection_temperature': 0.0,
            'gas': 'Helium',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'instrument_thermal_expansion_um_per_mk': 23.2,
            'ref_ball': {
                'name': 'Reference',
                'volume': 0.718522,
                'temperature': 22.0,
                'thermal_expansion': 2.3199999999999998e-05
            },
            'cup': {
                'name': 'Cup',
                'volume': 0.2174,
                'temperature': 22.0,
                'thermal_expansion': 2.36e-05
            },
            'expansion_fs_empty': None,
            'expansion_fs_ref': None,
            'expansion_fr_empty': {
                'num':
                    10,
                'p0s': [
                    12.955585479736328, 12.953627586364746, 12.951489448547363, 12.956624984741211,
                    12.952152252197266, 12.957865715026855, 12.954122543334961, 12.958620071411133,
                    12.953705787658691, 12.95920467376709
                ],
                'p1s': [
                    32.66371154785156, 32.70341491699219, 32.691680908203125, 32.636959075927734,
                    32.63353729248047, 32.71750259399414, 32.59309005737305, 32.673763275146484,
                    32.69328308105469, 32.610198974609375
                ],
                'p2s': [
                    18.558603286743164, 18.56930923461914, 18.562423706054688, 18.553447723388672,
                    18.545684814453125, 18.57682228088379, 18.535144805908203, 18.56588363647461,
                    18.563669204711914, 18.547842025756836
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.003501892089844, 22.085081100463867, 21.96920394897461, 22.086305618286133,
                    21.977062225341797, 22.082502365112305, 21.97055435180664, 22.087154388427734,
                    21.9652099609375, 22.08875846862793
                ],
                'temp_valve_block': [
                    22.003501892089844, 22.085081100463867, 21.96920394897461, 22.086305618286133,
                    21.977062225341797, 22.082502365112305, 21.97055435180664, 22.087154388427734,
                    21.9652099609375, 22.08875846862793
                ],
                'times': [
                    1260668, 1396536, 1535370, 1685211, 1839054, 1980885, 2144741, 2295628, 2441505,
                    2593345
                ]
            },
            'expansion_fr_ref': {
                'num':
                    10,
                'p0s': [
                    12.955777168273926, 12.956938743591309, 12.954573631286621, 12.944293975830078,
                    12.953569412231445, 12.949806213378906, 12.939936637878418, 12.948169708251953,
                    12.937586784362793, 12.940505981445312
                ],
                'p1s': [
                    32.709896087646484, 32.635223388671875, 32.635711669921875, 32.62635803222656,
                    32.648887634277344, 32.66506576538086, 32.66200256347656, 32.708953857421875,
                    32.669044494628906, 32.666107177734375
                ],
                'p2s': [
                    22.83177375793457, 22.79646110534668, 22.79655647277832, 22.785879135131836,
                    22.800922393798828, 22.809734344482422, 22.801422119140625, 22.830076217651367,
                    22.808992385864258, 22.801748275756836
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    21.95546531677246, 21.961626052856445, 22.097623825073242, 21.968889236450195,
                    21.956125259399414, 22.10438346862793, 21.96385955810547, 22.10438346862793,
                    22.0332088470459, 21.94864273071289
                ],
                'temp_valve_block': [
                    21.95546531677246, 21.961626052856445, 22.097623825073242, 21.968889236450195,
                    21.956125259399414, 22.10438346862793, 21.96385955810547, 22.10438346862793,
                    22.0332088470459, 21.94864273071289
                ],
                'times': [
                    6360314, 6660313, 6826160, 6972992, 7259981, 7413824, 7569676, 7716552, 7909436,
                    8145384
                ]
            },
            'volume_fs_sample': None,
            'volume_fs_reference': None,
            'volume_fr_sample': 1.552313005267143,
            'volume_fr_reference': 0.6170427709492514,
            'temperature': 20.0,
        }
    )
    site.add_calibration(
        2, {
            'start_time': '2022-02-03T15:07:16.885Z',
            'last_updated': LAST_UPDATE_1,
            'fill_pressure': 19.5,
            'collection_temperature': 0.0,
            'gas': 'Helium',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'instrument_thermal_expansion_um_per_mk': 23.2,
            'ref_ball': {
                'name': 'Reference',
                'volume': 0.718522,
                'temperature': 22.0,
                'thermal_expansion': 2.3199999999999998e-05
            },
            'cup': None,
            'expansion_fs_empty': None,
            'expansion_fs_ref': None,
            'expansion_fr_empty': {
                'num':
                    10,
                'p0s': [
                    12.955585479736328, 12.953627586364746, 12.951489448547363, 12.956624984741211,
                    12.952152252197266, 12.957865715026855, 12.954122543334961, 12.958620071411133,
                    12.953705787658691, 12.95920467376709
                ],
                'p1s': [
                    32.66371154785156, 32.70341491699219, 32.691680908203125, 32.636959075927734,
                    32.63353729248047, 32.71750259399414, 32.59309005737305, 32.673763275146484,
                    32.69328308105469, 32.610198974609375
                ],
                'p2s': [
                    18.558603286743164, 18.56930923461914, 18.562423706054688, 18.553447723388672,
                    18.545684814453125, 18.57682228088379, 18.535144805908203, 18.56588363647461,
                    18.563669204711914, 18.547842025756836
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.003501892089844, 22.085081100463867, 21.96920394897461, 22.086305618286133,
                    21.977062225341797, 22.082502365112305, 21.97055435180664, 22.087154388427734,
                    21.9652099609375, 22.08875846862793
                ],
                'temp_valve_block': [
                    22.003501892089844, 22.085081100463867, 21.96920394897461, 22.086305618286133,
                    21.977062225341797, 22.082502365112305, 21.97055435180664, 22.087154388427734,
                    21.9652099609375, 22.08875846862793
                ],
                'times': [
                    1260668, 1396536, 1535370, 1685211, 1839054, 1980885, 2144741, 2295628, 2441505,
                    2593345
                ]
            },
            'expansion_fr_ref': {
                'num':
                    10,
                'p0s': [
                    12.955777168273926, 12.956938743591309, 12.954573631286621, 12.944293975830078,
                    12.953569412231445, 12.949806213378906, 12.939936637878418, 12.948169708251953,
                    12.937586784362793, 12.940505981445312
                ],
                'p1s': [
                    32.709896087646484, 32.635223388671875, 32.635711669921875, 32.62635803222656,
                    32.648887634277344, 32.66506576538086, 32.66200256347656, 32.708953857421875,
                    32.669044494628906, 32.666107177734375
                ],
                'p2s': [
                    22.83177375793457, 22.79646110534668, 22.79655647277832, 22.785879135131836,
                    22.800922393798828, 22.809734344482422, 22.801422119140625, 22.830076217651367,
                    22.808992385864258, 22.801748275756836
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    21.95546531677246, 21.961626052856445, 22.097623825073242, 21.968889236450195,
                    21.956125259399414, 22.10438346862793, 21.96385955810547, 22.10438346862793,
                    22.0332088470459, 21.94864273071289
                ],
                'temp_valve_block': [
                    21.95546531677246, 21.961626052856445, 22.097623825073242, 21.968889236450195,
                    21.956125259399414, 22.10438346862793, 21.96385955810547, 22.10438346862793,
                    22.0332088470459, 21.94864273071289
                ],
                'times': [
                    6360314, 6660313, 6826160, 6972992, 7259981, 7413824, 7569676, 7716552, 7909436,
                    8145384
                ]
            },
            'volume_fs_sample': None,
            'volume_fs_reference': None,
            'volume_fr_sample': 1.552313005267143,
            'volume_fr_reference': 0.6170427709492514,
            'temperature': 20.0,
        }
    )
    site.add_result(
        1, {
            'start_time': '2022-02-02T17:28:43.127Z',
            'end_time': '2022-02-02T18:12:39.142Z',
            'last_updated': LAST_UPDATE_0,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 1',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': 'Standard',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 19.5,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 19.5,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    12.927614212036133, 12.929944038391113, 12.923385620117188, 12.930500984191895,
                    12.929095268249512, 12.919280052185059, 12.928112030029297, 12.921252250671387,
                    12.931562423706055, 12.929157257080078
                ],
                'p1s': [
                    32.62295913696289, 32.68987274169922, 32.60611343383789, 32.63804626464844,
                    32.66550827026367, 32.63631820678711, 32.67919158935547, 32.68134689331055,
                    32.66421890258789, 32.57997512817383
                ],
                'p2s': [
                    22.773513793945312, 22.812437057495117, 22.768993377685547, 22.783002853393555,
                    22.79920196533203, 22.77890968322754, 22.80499267578125, 22.802438735961914,
                    22.797531127929688, 22.7535343170166
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    21.948108673095703, 22.1102294921875, 22.030128479003906, 21.957351684570312,
                    22.10869026184082, 21.977062225341797, 22.103723526000977, 21.973007202148438,
                    21.953516006469727, 21.946128845214844
                ],
                'temp_valve_block': [
                    21.948108673095703, 22.1102294921875, 22.030128479003906, 21.957351684570312,
                    22.10869026184082, 21.977062225341797, 22.103723526000977, 21.973007202148438,
                    21.953516006469727, 21.946128845214844
                ],
                'times': [
                    838667, 994671, 1190648, 1434711, 1588652, 1744606, 1890567, 2036511, 2320607,
                    2609719
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        2, {
            'start_time': '2022-02-03T16:28:10.186Z',
            'end_time': '2022-02-03T16:55:26.629Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 2',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': 'Standard',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'R',
                'do_pre_purges': True,
                'purge_fill_pressure': 19.5,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 19.5,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.237804412841797, 14.236385345458984, 14.238737106323242, 14.237832069396973,
                    14.238334655761719, 14.236432075500488, 14.234519958496094, 14.2366943359375,
                    14.234245300292969, 14.237784385681152
                ],
                'p1s': [
                    33.866729736328125, 33.84098815917969, 33.916473388671875, 33.87053298950195,
                    33.891868591308594, 33.91254425048828, 33.874061584472656, 33.85850143432617,
                    33.85559844970703, 33.9117317199707
                ],
                'p2s': [
                    19.81703758239746, 19.803836822509766, 19.831832885742188, 19.813138961791992,
                    19.82470703125, 19.828786849975586, 19.814414978027344, 19.814136505126953,
                    19.804927825927734, 19.829832077026367
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.052825927734375, 21.934906005859375, 22.041257858276367, 21.928176879882812,
                    22.038301467895508, 22.10309410095215, 21.96341896057129, 22.086181640625,
                    21.940061569213867, 22.04854965209961
                ],
                'temp_valve_block': [
                    22.052825927734375, 21.934906005859375, 22.041257858276367, 21.928176879882812,
                    22.038301467895508, 22.10309410095215, 21.96341896057129, 22.086181640625,
                    21.940061569213867, 22.04854965209961
                ],
                'times': [
                    437210, 614380, 716254, 908262, 1005163, 1092134, 1244121, 1361149, 1504221,
                    1608141
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        3, {
            'start_time': '2022-02-03T19:08:29.100Z',
            'end_time': '2022-02-03T19:48:20.902Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 3',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': 'Standard',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 19.5,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 19.5,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.207351684570312, 14.208345413208008, 14.200316429138184, 14.20730972290039,
                    14.200138092041016, 14.207369804382324, 14.200509071350098, 14.208799362182617,
                    14.19925594329834, 14.205327987670898
                ],
                'p1s': [
                    33.81410217285156, 33.87273025512695, 33.86135482788086, 33.83768844604492,
                    33.834842681884766, 33.87639617919922, 33.852943420410156, 33.83415603637695,
                    33.798404693603516, 33.827213287353516
                ],
                'p2s': [
                    24.01152229309082, 24.041719436645508, 24.031347274780273, 24.023780822753906,
                    24.017566680908203, 24.042287826538086, 24.024715423583984, 24.021883010864258,
                    23.99738121032715, 24.01688003540039
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    21.97366714477539, 22.107589721679688, 21.980520248413086, 22.110889434814453,
                    21.979230880737305, 22.111392974853516, 21.94606590270996, 22.09844207763672,
                    21.962852478027344, 22.107967376708984
                ],
                'temp_valve_block': [
                    21.97366714477539, 22.107589721679688, 21.980520248413086, 22.110889434814453,
                    21.979230880737305, 22.111392974853516, 21.94606590270996, 22.09844207763672,
                    21.962852478027344, 22.107967376708984
                ],
                'times': [
                    1042867, 1193121, 1345175, 1488354, 1644573, 1787814, 1927945, 2084182, 2227324,
                    2371356
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        4, {
            'start_time': '2022-02-03T19:54:37.572Z',
            'end_time': '2022-02-03T20:24:44.679Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 4',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': 'Standard',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 19.5,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 19.5,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.20467472076416, 14.199069023132324, 14.204703330993652, 14.19838809967041,
                    14.203909873962402, 14.198530197143555, 14.20513916015625, 14.198847770690918,
                    14.20517349243164, 14.196331024169922
                ],
                'p1s': [
                    33.81455612182617, 33.8548698425293, 33.86112976074219, 33.82698059082031,
                    33.84956359863281, 33.82572555541992, 33.85053634643555, 33.83274459838867,
                    33.864532470703125, 33.81560134887695
                ],
                'p2s': [
                    24.00835418701172, 24.026790618896484, 24.033021926879883, 24.012310028076172,
                    24.025903701782227, 24.010658264160156, 24.028356552124023, 24.014110565185547,
                    24.03483009338379, 24.00428581237793
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.109350204467773, 21.996049880981445, 22.11884307861328, 22.003501892089844,
                    22.11635971069336, 21.984638214111328, 22.121013641357422, 21.98573875427246,
                    22.11981773376465, 21.983600616455078
                ],
                'temp_valve_block': [
                    22.109350204467773, 21.996049880981445, 22.11884307861328, 22.003501892089844,
                    22.11635971069336, 21.984638214111328, 22.121013641357422, 21.98573875427246,
                    22.11981773376465, 21.983600616455078
                ],
                'times': [
                    509279, 653636, 788616, 942596, 1066631, 1213713, 1351812, 1502236, 1643256,
                    1786464
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        5, {
            'start_time': '2022-02-03T20:33:33.622Z',
            'end_time': '2022-02-03T20:55:04.750Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 9p 5',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': '9 psi',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': None,
                'insert': {
                    'name': 'Insert',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 9.0,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 9.0,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.195656776428223, 14.195520401000977, 14.194944381713867, 14.196634292602539,
                    14.193774223327637, 14.191059112548828, 14.19188404083252, 14.190858840942383,
                    14.190886497497559, 14.193095207214355
                ],
                'p1s': [
                    23.92870330810547, 23.93729019165039, 23.97865104675293, 23.895620346069336,
                    23.836589813232422, 23.854341506958008, 23.894079208374023, 23.82672119140625,
                    23.9461727142334, 23.7914981842041
                ],
                'p2s': [
                    19.063026428222656, 19.065322875976562, 19.081287384033203, 19.045913696289062,
                    19.014318466186523, 19.016759872436523, 19.044086456298828, 19.007863998413086,
                    19.063034057617188, 18.99300765991211
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.01516342163086, 22.085206985473633, 21.918180465698242, 22.035409927368164,
                    22.10186767578125, 21.910541534423828, 22.027738571166992, 22.078950881958008,
                    21.909284591674805, 22.020444869995117
                ],
                'temp_valve_block': [
                    22.01516342163086, 22.085206985473633, 21.918180465698242, 22.035409927368164,
                    22.10186767578125, 21.910541534423828, 22.027738571166992, 22.078950881958008,
                    21.909284591674805, 22.020444869995117
                ],
                'times': [
                    374137, 453278, 590573, 684529, 756524, 885663, 978732, 1043618, 1181656,
                    1271781
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        6, {
            'start_time': '2022-02-03T21:01:20.063Z',
            'end_time': '2022-02-03T21:24:56.982Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 9p 6',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': '9 psi',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 9.0,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 9.0,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.192269325256348, 14.191065788269043, 14.190340995788574, 14.192840576171875,
                    14.19243335723877, 14.191588401794434, 14.19286823272705, 14.191265106201172,
                    14.191607475280762, 14.192167282104492
                ],
                'p1s': [
                    23.956323623657227, 23.858488082885742, 23.788843154907227, 23.745683670043945,
                    23.784812927246094, 23.994407653808594, 23.746234893798828, 23.870065689086914,
                    23.907390594482422, 23.916444778442383
                ],
                'p2s': [
                    19.074209213256836, 19.024072647094727, 18.983932495117188, 18.970531463623047,
                    18.987680435180664, 19.08735466003418, 18.969741821289062, 19.02984046936035,
                    19.043367385864258, 19.056228637695312
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.017993927001953, 22.084232330322266, 21.90708351135254, 22.03484344482422,
                    22.088443756103516, 21.90862464904785, 22.024690628051758, 22.090362548828125,
                    21.901613235473633, 22.028242111206055
                ],
                'temp_valve_block': [
                    22.017993927001953, 22.084232330322266, 21.90708351135254, 22.03484344482422,
                    22.088443756103516, 21.90862464904785, 22.024690628051758, 22.090362548828125,
                    21.901613235473633, 22.028242111206055
                ],
                'times': [
                    498252, 577476, 713938, 809995, 878279, 1006249, 1092151, 1160251, 1302445,
                    1397504
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        7, {
            'start_time': '2022-02-03T21:45:16.548Z',
            'end_time': '2022-02-03T22:09:00.937Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 9p 7',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': '9 psi',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 9.0,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 9.0,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.199007034301758, 14.198505401611328, 14.196718215942383, 14.196630477905273,
                    14.196859359741211, 14.196242332458496, 14.197488784790039, 14.19698429107666,
                    14.196012496948242, 14.195878028869629
                ],
                'p1s': [
                    23.75822639465332, 23.80229377746582, 23.816070556640625, 23.834640502929688,
                    23.921663284301758, 23.829010009765625, 24.017972946166992, 24.064041137695312,
                    23.973142623901367, 23.882631301879883
                ],
                'p2s': [
                    18.978601455688477, 18.999248504638672, 19.001157760620117, 19.01536750793457,
                    19.058744430541992, 19.006784439086914, 19.108875274658203, 19.129867553710938,
                    19.079132080078125, 19.040098190307617
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.013153076171875, 22.079988479614258, 21.92192268371582, 22.01560401916504,
                    22.08753204345703, 21.910982131958008, 22.021263122558594, 22.075868606567383,
                    21.91465950012207, 22.025318145751953
                ],
                'temp_valve_block': [
                    22.013153076171875, 22.079988479614258, 21.92192268371582, 22.01560401916504,
                    22.08753204345703, 21.910982131958008, 22.021263122558594, 22.075868606567383,
                    21.91465950012207, 22.025318145751953
                ],
                'times': [
                    504282, 583397, 723616, 801520, 878724, 1017856, 1110845, 1183754, 1319893,
                    1404913
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        8, {
            'start_time': '2022-02-04T12:23:16.228Z',
            'end_time': '2022-02-04T12:37:16.627Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 5p 8',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': '5 psi',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 5.0,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 5.0,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.212986946105957, 14.214278221130371, 14.216361045837402, 14.218143463134766,
                    14.217048645019531, 14.218076705932617, 14.219786643981934, 14.220304489135742,
                    14.220611572265625, 14.220672607421875
                ],
                'p1s': [
                    20.001771926879883, 19.98788070678711, 19.88715171813965, 20.04866600036621,
                    20.10904312133789, 19.969032287597656, 20.076133728027344, 19.982120513916016,
                    20.316043853759766, 20.14958953857422
                ],
                'p2s': [
                    17.10694122314453, 17.101055145263672, 17.05112648010254, 17.132736206054688,
                    17.16257667541504, 17.092859268188477, 17.147666931152344, 17.100492477416992,
                    17.267444610595703, 17.184974670410156
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.00126838684082, 22.004066467285156, 21.992372512817383, 22.00343894958496,
                    22.00189781188965, 21.997339248657227, 22.010133743286133, 21.99463653564453,
                    21.990894317626953, 22.006330490112305
                ],
                'temp_valve_block': [
                    22.00126838684082, 22.004066467285156, 21.992372512817383, 22.00343894958496,
                    22.00189781188965, 21.997339248657227, 22.010133743286133, 21.99463653564453,
                    21.990894317626953, 22.006330490112305
                ],
                'times': [
                    249994, 309941, 372815, 443780, 500944, 562903, 627855, 688787, 761829, 821864
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        9, {
            'start_time': '2022-02-04T12:46:25.325Z',
            'end_time': '2022-02-04T12:59:22.646Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 5p 9',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': '5 psi',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 5.0,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 5.0,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.224132537841797, 14.22487735748291, 14.225892066955566, 14.226895332336426,
                    14.228107452392578, 14.229516983032227, 14.229623794555664, 14.230971336364746,
                    14.231216430664062, 14.231071472167969
                ],
                'p1s': [
                    20.12830352783203, 19.928739547729492, 20.249738693237305, 20.05295753479004,
                    20.21966552734375, 20.0286922454834, 20.130767822265625, 20.12566566467285,
                    20.120460510253906, 20.1538028717041
                ],
                'p2s': [
                    17.17595100402832, 17.076112747192383, 17.237180709838867, 17.139497756958008,
                    17.223112106323242, 17.128437042236328, 17.179588317871094, 17.177465438842773,
                    17.175508499145508, 17.191556930541992
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.002653121948242, 21.995986938476562, 22.002243041992188, 22.000797271728516,
                    21.995201110839844, 22.008625030517578, 22.000734329223633, 21.995389938354492,
                    22.00824737548828, 21.992876052856445
                ],
                'temp_valve_block': [
                    22.002653121948242, 21.995986938476562, 22.002243041992188, 22.000797271728516,
                    21.995201110839844, 22.008625030517578, 22.000734329223633, 21.995389938354492,
                    22.00824737548828, 21.992876052856445
                ],
                'times': [
                    230966, 289895, 347737, 404757, 467822, 527837, 584749, 642685, 695733, 758821
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        10, {
            'start_time': '2022-02-04T13:00:39.266Z',
            'end_time': '2022-02-04T13:25:22.550Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 10',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': 'Standard',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 19.5,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 19.5,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.241714477539062, 14.242865562438965, 14.244438171386719, 14.245716094970703,
                    14.246204376220703, 14.246147155761719, 14.245890617370605, 14.245721817016602,
                    14.24518871307373, 14.24477481842041
                ],
                'p1s': [
                    33.87849426269531, 33.86403274536133, 33.88093948364258, 33.87791061401367,
                    33.86410903930664, 33.86308670043945, 33.88859939575195, 33.88832473754883,
                    33.88869857788086, 33.886539459228516
                ],
                'p2s': [
                    24.058122634887695, 24.05117416381836, 24.06136131286621, 24.0601749420166,
                    24.05363655090332, 24.052955627441406, 24.065263748168945, 24.066011428833008,
                    24.065814971923828, 24.064348220825195
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.005859375, 21.995861053466797, 22.007997512817383, 22.00972557067871,
                    22.01224136352539, 22.007431030273438, 22.00541877746582, 21.992687225341797,
                    22.007305145263672, 22.008846282958984
                ],
                'temp_valve_block': [
                    22.005859375, 21.995861053466797, 22.007997512817383, 22.00972557067871,
                    22.01224136352539, 22.007431030273438, 22.00541877746582, 21.992687225341797,
                    22.007305145263672, 22.008846282958984
                ],
                'times': [
                    414173, 533129, 667069, 767978, 875024, 972930, 1079856, 1228869, 1356847,
                    1461808
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        11, {
            'start_time': '2022-02-04T13:36:50.646Z',
            'end_time': '2022-02-04T13:51:43.580Z',
            'last_updated': LAST_UPDATE_1,
            'mass': 1.0,
            'operator': 'django',
            'name': 'ball 9p 11',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'parameters': {
                'name': '9 psi',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 9.0,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 9.0,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.241793632507324, 14.242178916931152, 14.24313735961914, 14.242799758911133,
                    14.242786407470703, 14.242868423461914, 14.241971015930176, 14.240965843200684,
                    14.241920471191406, 14.243014335632324
                ],
                'p1s': [
                    23.890052795410156, 24.0811824798584, 23.97492218017578, 23.932153701782227,
                    23.783065795898438, 23.922916412353516, 23.79615592956543, 23.840972900390625,
                    24.10832977294922, 23.970537185668945
                ],
                'p2s': [
                    19.06414794921875, 19.15955924987793, 19.10686683654785, 19.08584976196289,
                    19.010791778564453, 19.08112907409668, 19.01720428466797, 19.038938522338867,
                    19.173080444335938, 19.105281829833984
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    22.002086639404297, 21.990516662597656, 21.99532699584961, 22.007431030273438,
                    21.98904037475586, 22.01007080078125, 21.99463653564453, 21.993410110473633,
                    21.993032455444336, 22.003877639770508
                ],
                'temp_valve_block': [
                    22.002086639404297, 21.990516662597656, 21.99532699584961, 22.007431030273438,
                    21.98904037475586, 22.01007080078125, 21.99463653564453, 21.993410110473633,
                    21.993032455444336, 22.003877639770508
                ],
                'times': [
                    251988, 322950, 395855, 457709, 528580, 592534, 660528, 732521, 807516, 873460
                ]
            },
            'expansion_repeat': None,
            'equilibration_set': []
        }
    )
    site.add_result(
        12, {
            'start_time':
                '2022-02-04T14:01:03.996Z',
            'end_time':
                '2022-02-04T14:21:41.550Z',
            'last_updated':
                LAST_UPDATE_1,
            'mass':
                1.0,
            'operator':
                'django',
            'name':
                'ball 15p 12',
            'instrument': {
                'model': '1350',
                'serial': '0001'
            },
            'calibration': {
                'start_time': '2022-02-02T15:07:16.885Z'
            },
            'sample_log': [{
                'datetime': '2022-02-04T14:21:41.550Z',
                'message': 'this is a test'
            }],
            'parameters': {
                'name': '15 psi',
                'gas': 'Helium',
                'temperature': 22.0,
                'use_vacuum_prep': False,
                'cup': {
                    'name': 'Cup',
                    'volume': 0.2174,
                    'temperature': 22.0,
                    'thermal_expansion': 2.36e-05
                },
                'insert': None,
                'fill_direction': 'S',
                'do_pre_purges': True,
                'purge_fill_pressure': 15.0,
                'num_purges': 10,
                'num_cycles': 10,
                'cycle_fill_pressure': 15.0,
                'endby': 'E',
                'interval_time': 10,
                'equilibration_rate': 0.005,
                'p1_record_equilibration_data': False,
                'p1_equilibration_cycles': 5,
                'p1_equilibration_points': 500,
                'p2_record_equilibration_data': False,
                'p2_equilibration_cycles': 5,
                'p2_equilibration_points': 500,
                'technique': 'P',
                'resin_density': 1.0,
                'initial_pressure': 19.5,
                'final_pressure': 19.5,
                'pressure_increment': 19.5,
                'low_pressure': 19.5,
                'fracture_pressure': 19.5,
                'use_run_precision': False,
                'run_precision_tolerance': 0.05,
                'compute_volume': 'C',
                'geometric_volume': 10.0,
                'active_area': 1.0,
                'shape': 'U',
                'cube_edge_length': 1.0,
                'cylinder_diameter': 1.0,
                'cylinder_length': 1.0,
                'rect_long_edge': 1.0,
                'rect_short_edge': 1.0,
                'rect_other_edge': 1.0,
                'number_of_pieces': 1,
                'measure_by': 'L',
                'chord_length': 1.0,
                'cell_diameter': 1.0,
                'cube_number_of_skins': 0,
                'cyl_number_of_skins': 0,
                'rect_number_of_large_skins': 0,
                'rect_number_of_small_skins': 0,
                'rect_number_of_other_skins': 0
            },
            'expansion': {
                'num':
                    10,
                'p0s': [
                    14.242864608764648, 14.2457857131958, 14.247845649719238, 14.245817184448242,
                    14.246564865112305, 14.248208045959473, 14.250221252441406, 14.251330375671387,
                    14.251490592956543, 14.251739501953125
                ],
                'p1s': [
                    29.62009048461914, 29.582569122314453, 29.609495162963867, 29.715131759643555,
                    29.66482162475586, 29.63442611694336, 29.641559600830078, 29.5817813873291,
                    29.660228729248047, 29.67257308959961
                ],
                'p2s': [
                    21.928590774536133, 21.91130256652832, 21.925884246826172, 21.977567672729492,
                    21.952816009521484, 21.938453674316406, 21.943065643310547, 21.913726806640625,
                    21.953020095825195, 21.95937728881836
                ],
                'excludes': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'temp_anl_block': [
                    21.994447708129883, 21.99532699584961, 21.99825096130371, 21.996238708496094,
                    21.992969512939453, 21.99334716796875, 21.99567413330078, 21.991806030273438,
                    21.990455627441406, 21.992938995361328
                ],
                'temp_valve_block': [
                    21.994447708129883, 21.99532699584961, 21.99825096130371, 21.996238708496094,
                    21.992969512939453, 21.99334716796875, 21.99567413330078, 21.991806030273438,
                    21.990455627441406, 21.992938995361328
                ],
                'times': [
                    302054, 407049, 506021, 596095, 696045, 802142, 903076, 1001106, 1110075,
                    1218050
                ]
            },
            'expansion_repeat':
                None,
            'equilibration_set': [{
                'cycle': 0,
                'which_equil': '1',
                'num_points': 2,
                'times': [1, 3],
                'pressures': [1.1, 1.0]
            }]
        }
    )
    site.add_result1345(
        4293, {
            "start_time": "2023-04-14T17:17:56Z",
            "end_time": "2023-04-14T17:37:09Z",
            "last_updated": "2023-04-19T18:42:15.574Z",
            "mass": 0.1374,
            "name": "",
            "operator": "",
            "instrument": {"model": "1345", "serial": "4467"},
            "technique": "P",
            "temperature": 20.009,
            "num_purges": 10,
            "endby": "E",
            "interval_time": 10,
            "equilibration_rate": 0.0050037999872366845,
            "resin_density": 1.0,
            "geometric_volume": 1.0,
            "active_area": 1.0,
            "measure_by": "L",
            "chord_length": 1.0,
            "cell_diameter": 1.0,
            "expansion": {
                "num": 10,
                "volumes": [0.051500000059604645, 0.057999998331069946, 0.05889999866485596,
                            0.05530000105500221, 0.054099999368190765, 0.055799998342990875,
                            0.05209999904036522, 0.053700000047683716, 0.05550000071525574,
                            0.0575999990105629],
                "p1s": [],
                "p2s": [],
                "excludes": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                "times": [326000, 416000, 503000, 592000, 685000, 777000, 868000, 961000, 1056000,
                          1145000],
                "temp_anl_block": [0.051500000059604645, 0.057999998331069946, 0.05889999866485596,
                                   0.05530000105500221, 0.054099999368190765, 0.055799998342990875,
                                   0.05209999904036522, 0.053700000047683716, 0.05550000071525574,
                                   0.0575999990105629]},
            "expansion_repeat": None,
            "sample_log": []
        }
    )

def _insert(**kwargs):
    insert = {
        'name': 'Cup',
        'volume': 1.0,
        'temperature': 22.0,
        'material': 'ot',
        'other_expansion_um_per_mk': 23.6,
        'kind': 'C',
        'last_updated': LAST_UPDATE_1
    }
    insert.update(**kwargs)
    return insert

def _method(**kwargs):
    method = {
            'name': 'Method 1',
            'gas': 'Helium',
            'temperature': 20.0,
            'use_vacuum_prep': False,
            'cup': None,
            'insert': None,
            'fill_direction': 'S',
            'do_pre_purges': True,
            'purge_fill_pressure': 19.5,
            'num_purges': 10,
            'num_cycles': 5,
            'cycle_fill_pressure': 19.5,
            'endby': 'E',
            'interval_time': 10,
            'equilibration_rate': 0.005,
            'p1_record_equilibration_data': False,
            'p1_equilibration_cycles': 5,
            'p1_equilibration_points': 500,
            'p2_record_equilibration_data': False,
            'p2_equilibration_cycles': 5,
            'p2_equilibration_points': 500,
            'technique': 'P',
            'resin_density': 1.0,
            'initial_pressure': 19.5,
            'final_pressure': 19.5,
            'pressure_increment': 19.5,
            'low_pressure': 19.5,
            'fracture_pressure': 19.5,
            'use_run_precision': False,
            'run_precision_tolerance': 0.05,
            'compute_volume': 'C',
            'geometric_volume': 10.0,
            'active_area': 1.0,
            'shape': 'U',
            'cube_edge_length': 1.0,
            'cylinder_diameter': 1.0,
            'cylinder_length': 1.0,
            'rect_long_edge': 1.0,
            'rect_short_edge': 1.0,
            'rect_other_edge': 1.0,
            'number_of_pieces': 1,
            'measure_by': 'L',
            'chord_length': 1.0,
            'cell_diameter': 1.0,
            'cube_number_of_skins': 0,
            'cyl_number_of_skins': 0,
            'rect_number_of_large_skins': 0,
            'rect_number_of_small_skins': 0,
            'rect_number_of_other_skins': 0,
            'last_updated': LAST_UPDATE_1,
        }
    method.update(kwargs)
    return method
