'''
View that handles /rest/ requests to the rest api
'''
import gettext
import json
import re
from django.http import JsonResponse
from django.shortcuts import redirect
from django.views.generic import View
from django import forms
from django.core.exceptions import ValidationError
from django.db.models import ProtectedError, Q
from django.utils.functional import lazy
from mic_python_lib import inst_scripts
from mic_python_lib.mic_json_encoder import MicJSONEncoder
from mic_python_lib.rest_input import (
    ReactFormMixin,
    UnitsFloatFormField,
    ButtonGroupWidget,
    AlphanumericPasswordInput,
    multi_form_view,
    post_response,
)
from mic_python_lib.results import (
    ReactSortFilterMixin,
    report_show_export_print_view,
    sample_results_view,
)
import mic_python_lib.blocked_ui as bui
from mic_python_lib.instrument import gen_config_passwd
from mic_python_lib.settings_views import (
    lang_choices,
    region_relevant_if,
    add_region_props,
    add_unit_selections,
    pull_region_setting,
    update_language,
    update_unit_selections,
)

import i1350
from mic_cpp import cpp, pycnometry

from .accupyc1345 import check_available
from .models import (
    AnalysisParameters,
    AnalysisResult,
    AnalysisResult1345,
    AnalysisResult1350,
    Calibration,
    InsertMeasure,
    ResultCache,
    RTDCalibration,
    PressureCalibration,
    Insert,
    Instrument,
    LegacyInstrument,
    OtherOperation,
    Settings,
    FILL_CHOICES,
    END_CHOICES,
    copy_model_obj,
)

from .services import (
    initialize_analysis,
    analysis_get_hw_state,
    analysis_get_operation_status,
    analysis_operation_status,
    calibration_operation_status,
    fill_instrument_log,
    get_chambercal,
    get_samplecaldata,
    accu_or_foam_results_from_anl,
    notify_analysis_seen,
    hwcal_results,
    insert_measure_operation_status,
    log_field_changes,
    other_operation_status,
    operation_cancel,
    update_results,
    prompt_for_cal_check,
    send_idle_purge_info,
    send_purge_flow_info,
    script_operation_start,
)
# pylint: disable=invalid-name,too-many-lines

_ = lazy(gettext.gettext, str)


class ScriptInterface1350(inst_scripts.ScriptInterface):

    def model_serial(self):
        from .models import Instrument
        local = Instrument.objects.local()
        return local.model, local.serial

    def run_from_oper(self, oper_id: int):
        from mic_python_lib.models import InstrumentScriptRun
        from .models import OtherOperation
        return InstrumentScriptRun.objects.get(
            id=OtherOperation.objects.get(id=oper_id).script_run.id
        ).id

    def script_start(self, name: str, operator: str):
        script_operation_start(name, operator)

    def script_cancel(self):
        operation_cancel()

    def app_export(self, location: str, is_service: bool):
        app_export_results(location, is_service)


class LegacyForm(forms.ModelForm):
    '''Custom form for Legacy Instruments'''

    class Meta:
        model = LegacyInstrument
        fields = ('ip_addr',)

    def clean(self):
        '''Check for an instrument that is up at the requested address.'''
        cleaned_data = super().clean()
        if not check_available(cleaned_data['ip_addr']):
            raise ValidationError(_("No legacy instrument at the address."))
        return cleaned_data


def _create_legacy_instrument():
    inst = LegacyInstrument.objects.create(model='1345', ip_addr='192.168.0.0')
    return {'object_id': inst.id}


def _delete_legacy_instrument(object_id):
    inst = LegacyInstrument.objects.get(id=object_id)
    inst.delete()
    return {'object_id': object_id}


class LegacyView(ReactFormMixin, View):
    '''View for creating and editing inserts'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''handle requests to show the insert editor'''
        form = LegacyForm
        delete_object = _delete_legacy_instrument
        add_object = _create_legacy_instrument


def ui_about(request):
    '''View to fill in the data for what is shown on about screen.'''
    _ = gettext.gettext

    local = Instrument.objects.local()

    model = cpp.CReportModel(_("Calibration"))
    page = model.AddPage("About") # single page report.  Page name not shown to user.
    table = cpp.CReportOptionsTable(page, "")
    table.AddItem(_("Serial Number:"), local.model + " - " + local.serial)
    table.AddItem(_("Configuration:"), local.configuration)
    table.AddItem(_("Last Analysis:"), local.gas.name)
    table.AddItem(_("Warranty Expires:"), cpp.CTimeDate(local.device_warranty).GetDate())
    table.AddItem(_("Next Service Due:"), cpp.CTimeDate(local.device_service).GetDate())
    ver = i1350.GetVersionInfo1350()
    table.AddItem(
        _("Software Version:"),
        ver['version'] + " " + ver['date'] + " [" + ver['git_rev'] + ']')
    return JsonResponse(json.loads(model.AsJson()))


def ui_anl_review_report(request, anl_id):
    '''
    The endpoint for a report page for the analysis view.

    Without post/body parameters this returns the results report for the analysis with id anl_id.
    With parameters it does the print of export functionality.
    '''

    # If it is the on screen report, and we have a cache, use that without regenerating.
    if not request.body:
        try:
            cache, _ = ResultCache.objects.get_or_create(results_id=anl_id)
            # Even if it was created, but the cache value is not there yet, redo the cache
            # content creation.
            if cache.report_json:
                return JsonResponse(json.loads(cache.report_json))
        except ResultCache.MultipleObjectsReturned:
            # if there had been a race condition that created multiple caches, just delete them
            # all.  This could happen if there are multiple screens looking at the analysis page
            # when the analysis finishes.
            ResultCache.objects.filter(results_id=anl_id).delete()
            cache = ResultCache.objects.create(results_id=anl_id)

    # otherwise do the normal handling.
    try:
        anl = AnalysisResult1350.objects.get(id=anl_id)
    except AnalysisResult1350.DoesNotExist:
        anl = AnalysisResult1345.objects.get(id=anl_id)
    accu_result = accu_or_foam_results_from_anl(anl)

    model = cpp.CReportModel("")
    if not request.body:
        i1350.AddAccuPycResultsScreen(model, accu_result)
    else:
        data = json.loads(request.body.decode('utf-8'))
        if 'target' in data and data['target'] == 'export' and data['export_format'] == 'raw':
            i1350.AddAccuPycRaw(model, accu_result)
        else:
            i1350.AddAccuPycResultsReport(model, accu_result)

    basename = f"{_base_export_name(anl)}"

    # If it is the on screen report, and we did not have a cache, save it in the cache.
    if not request.body:
        cache.report_json = model.AsJson()
        cache.save()
        return JsonResponse(json.loads(cache.report_json))
    return report_show_export_print_view(request, model, basename)


def ui_cal_review_history(request, which):
    '''
    The endpoint for a report page for the calibration check.

    Without post/body parameters this returns the results report for the analysis with id anl_id.
    With parameters it does the print of export functionality.
    '''
    _ = gettext.gettext
    local = Instrument.objects.local()
    basename = str(_("calibrations")) + "-" + local.serial
    model = cpp.CReportModel("")
    last_cal = Calibration.objects.get_current()
    cal_checks = Calibration.objects.filter(
        check_of=last_cal, kind='H', state='C'
    ).order_by('-start_time')
    is_1cm3 = Instrument.objects.local().kind == '1'

    data = json.loads(request.body.decode('utf-8')) if request.body else {}
    if 'target' in data and data['target'] == 'export' and data['export_format'] == 'raw':
        _fill_vol_cal_raw(
            model,
            is_1cm3,
            Calibration.objects.filter(state='C').exclude(kind='H').order_by("-start_time")[:100])
    else:
        # add all the pages if this is an export or print. Otherwise just show the current one.
        doall = bool(request.body)
        if (doall or which == 'last_cal') and last_cal:
            i1350.AddVolCalResults(model, _vol_cal_or_check_results_from_cal(last_cal))
        if doall or which == 'history':
            _fill_vol_check_table(model, is_1cm3, cal_checks, _("Calibration Check History"))
        if doall or which == 'cal_history':
            _fill_vol_cal_table(
                model,
                is_1cm3,
                Calibration.objects.filter(state='C').exclude(kind='H').order_by("-start_time"),
                _("Calibration History"))
        if doall or which == 'last_check':
            last_check = cal_checks.first()
            if last_check:
                header = _vol_cal_or_check_header(last_check)
                results = _vol_check_results(last_check)
                i1350.AddVolCheckCalResults(model, header, results)
        if doall or which == 'inst_log':
            fill_instrument_log(model)

    return report_show_export_print_view(request, model, basename)


def ui_cal_review_report(request, cal_id):
    '''View to return data for the body of the review calibration screen'''
    cal = Calibration.objects.get(id=cal_id)
    if cal.state != 'C':
        return JsonResponse({})

    model = cpp.CReportModel("")
    if cal.kind == 'H':
        basename = _("calibration-check")
        header = _vol_cal_or_check_header(cal)
        results = _vol_check_results(cal)
        i1350.AddVolCheckCalResults(model, header, results)
    else:
        basename = _("calibration")
        data = _vol_cal_or_check_results_from_cal(cal)
        i1350.AddVolCalResults(model, data)
    return report_show_export_print_view(request, model, basename)


def ui_anl_inprocess(request, anl_id):
    '''View to return data for the in process analysis screen'''
    anl = AnalysisResult1350.objects.get(id=anl_id)
    if anl.technique == 'P':
        response = {
            'primary': {'title': _("Density")},
            'secondary': {'title': _("Volume")},
            'runtime': {'title': _("Last Density")}
        }
    if anl.technique in ('A', 'B', 'C'):
        response = {
            'primary': {'title': _("Open Cell Percent")}
        }
    if anl.technique == 'D':
        response = {
            'primary': {'title': _("P1 Percent Change")}
        }
    if anl.technique == 'E':
        response = {
            'primary': {'title': _("Percent Change")}
        }
    response.update({
        result.get_result_rank_display(): {
            'title': result.result_title,
            'value': result.value,
            'std_dev': result.std_dev,
            'format': json.loads(result.format_json),
            'units': json.loads(result.units_json)
        } for result in anl.results.all()
    })
    state = analysis_get_hw_state()
    status = analysis_get_operation_status()
    response.update({
        'terminating': status.fIsTerminating,
        'waitingskip': status.fIsWaitingSkip,
        'pressure': state.flGaugePressure,
        'temperature': state.flBlockTemp,
        'status': status.szStatus,
        'progress': status.flFractionComplete,
        'ref_purge_cur': status.iRefPurge,
        'ref_purge_num': status.iRefPurgeNum,
        'purge_cur': status.iPurge,
        'purge_num': status.iPurgeNum,
        'cycle_cur': status.iCycle,
        'cycle_num': status.iCycleNum,
    })
    return JsonResponse(response)


def ui_cal_inprocess(request):
    '''View to return data for the in process calibration screen'''
    state = analysis_get_hw_state()
    status = analysis_get_operation_status()
    return JsonResponse({
        'terminating': status.fIsTerminating,
        'waitingskip': status.fIsWaitingSkip,
        'current_pressure': state.flGaugePressure,
        'status': status.szStatus,
        'progress': status.flFractionComplete,
        'ref_purge_cur': status.iRefPurge,
        'ref_purge_num': status.iRefPurgeNum,
        'purge_cur': status.iPurge,
        'purge_num': status.iPurgeNum,
        'cycle_cur': status.iCycle,
        'cycle_num': status.iCycleNum,
    })


def ui_insmeas_inprocess(request, anl_id):
    '''View to return data for the in process insert measurement screen.'''
    anl = InsertMeasure.objects.get(id=anl_id)
    state = analysis_get_hw_state()
    status = analysis_get_operation_status()
    response = {
        'primary': {'title': _("Volume Fill Sample")},
        'secondary': {'title': _("Volume Fill Reference")},
    }
    response.update({
        result.get_result_rank_display(): {
            'title': result.result_title,
            'value': result.value,
            'std_dev': result.std_dev,
            'format': json.loads(result.format_json),
            'units': json.loads(result.units_json)
        } for result in anl.results.all()
    })
    response.update({
        'terminating': status.fIsTerminating,
        'waitingskip': status.fIsWaitingSkip,
        'pressure': state.flGaugePressure,
        'temperature': state.flBlockTemp,
        'status': status.szStatus,
        'progress': status.flFractionComplete,
        'ref_purge_cur': status.iRefPurge,
        'ref_purge_num': status.iRefPurgeNum,
        'purge_cur': status.iPurge,
        'purge_num': status.iPurgeNum,
        'cycle_cur': status.iCycle,
        'cycle_num': status.iCycleNum,
    })
    return JsonResponse(response)



class AnlStartForm(forms.ModelForm):
    '''Form to start analysis'''
    parameters = forms.ModelChoiceField(
        AnalysisParameters.objects.methods().order_by('-is_default', 'name'),
        initial=AnalysisParameters.objects.default,
        label=_("Method")
    )
    mass = UnitsFloatFormField(
        fmt=i1350.DPycSampleMass1350,  # pylint: disable=no-member
        initial="",
        label=_("Mass")
    )

    class Meta:
        model = AnalysisResult1350
        fields = ('parameters', 'name', 'operator', 'mass')


class ui_anl_start(ReactFormMixin, View):  # pylint: disable=too-few-public-methods
    '''View to return data for start analysis screen'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define model, fields, and options for displaying form'''
        form = AnlStartForm
        hidden = {
            'name': lambda: not Settings.objects.get().display_id,
            'parameters': lambda: not Settings.objects.get().display_method,
            'operator': lambda: not Settings.objects.get().display_operator,
        }

    def get(self, request):
        'Do the normal get, but add the enter mass prompt on startup.'
        object_id = self.id_from_request(request)
        props = self.get_input_form_props(object_id)
        props['format']['fields']['mass']['help'] = _("Enter mass")
        return JsonResponse(props, encoder=MicJSONEncoder)

def _get_default_ref():
    refs = Insert.objects.reference_balls()
    return refs.last().id if refs.last() else -1


class CalStartForm(forms.ModelForm):
    '''Form for starting calibration'''

    ref_ball = forms.ModelChoiceField(
        Insert.objects.reference_balls(),
        initial=_get_default_ref,
        label=_("Volume Standard"),
    )

    class Meta:
        model = Calibration
        fields = ('fill_pressure',)

    field_order = ['ref_ball', 'fill_pressure']

class ui_cal_start(ReactFormMixin, View):  # pylint: disable=too-few-public-methods
    '''View to return data for start calibration screen'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define models, fields, and options for displaying form'''
        form = CalStartForm


def ui_cal_user(request):
    "View that gives the json for the user calibration review screen."
    _ = gettext.gettext

    cal = Calibration.objects.last()
    try:
        factory = Calibration.objects.get_last_factory()
    except Calibration.DoesNotExist:
        factory = None
    try:
        previous = Calibration.objects.get_current()
        if previous.id == factory.id or previous.kind == "R":
            previous = None
    except Calibration.DoesNotExist:
        previous = None

    factory_error_allowed = 0.002 # 0.2% error allowed from factory.
    previous_error_allowed = 0.002 # 0.2% error allowed from previous.

    rej_prior = not _is_cal_in_range(cal, previous, previous_error_allowed) if previous else False
    rej_factory = not _is_cal_in_range(cal, factory, factory_error_allowed) if factory else False

    # if this was rejected, due to differences in results, mark it as such.
    if rej_prior or rej_factory:
        cal.kind = 'UR'
        cal.save()

    return JsonResponse({
        "rejectPrior": rej_prior,
        "rejectFactory": rej_factory,
        "report": json.loads(_user_cal_results(cal, previous, factory).AsJson()),
    })

def ui_operation_status(request):
    '''
    View to return status for polling loop of analysis screen.


    This is the first message from the UI.  If it does not get one it will continue retrying.  Also,
    it assumes that the instrument c++ code is running and initialized. (And, it will then make
    other requests that need it are working.)  So, don't return anything until that has been
    initialized.
    '''
    # don't return a proper response if the application is not ready to give status.
    if not i1350.UDPIsReady():
        return JsonResponse({"error": "not ready."})

    if not bui.update_ui_data(request):
        return JsonResponse({"blocked": "too many UI."})

    # Get the c++ operation state enumerations as a string.
    inst_oper = str(analysis_get_operation_status().eOperation).split('.')[-1]

    response = {'operationActive': None}

    operations = {
        "analysis": analysis_operation_status,
        "calibration": lambda inst_oper: calibration_operation_status(inst_oper, kind='C'),
        "calibrationCheck": lambda inst_oper: calibration_operation_status(inst_oper, kind='H'),
        "calibrationUser": lambda inst_oper: calibration_operation_status(inst_oper, kind='UN'),
        "measureInsert": insert_measure_operation_status,
        "hwcalPressure": lambda inst_oper: other_operation_status(inst_oper, kind='P'),
        "hwcalBlockTemperature": lambda inst_oper: other_operation_status(inst_oper, kind='B'),
        "hwcalInletTemperature": lambda inst_oper: other_operation_status(inst_oper, kind='I'),
        "otherOperationS": lambda inst_oper: other_operation_status(inst_oper, kind='S'),
        "otherOperationU": lambda inst_oper: other_operation_status(inst_oper, kind='U'),
        "scriptOperation": lambda inst_oper: other_operation_status(inst_oper, kind='Y'),
    }
    for op, status_fcn in operations.items():
        status = status_fcn(inst_oper)
        response[op] = status
        if status["is_active"]:
            response["operationActive"] = op
    response["is1cm3"] = Instrument.objects.local().kind == '1'
    response["isATC"] = Instrument.objects.local().is_temp_control
    response["promptCalCheck"] = prompt_for_cal_check()

    oper_status = i1350.UDPGetOperationStatus()
    if oper_status.eIdleOperation == i1350.EIdleOperation.ceIdlePurgeRun:
        response["idleOperation"] = "ceIdlePurgeRun"
    elif oper_status.eIdleOperation == i1350.EIdleOperation.ceIdlePurgeTerminate:
        response["idleOperation"] = "ceIdlePurgeTerminate"
    else:
        response["idleOperation"] = ""

    return JsonResponse(response)


class ResultEditParamForm(forms.ModelForm):
    "The form that allows editing of the post analysis parameters (that are part of the results)"

    insert = forms.ModelChoiceField(
        Insert.objects.inserts(),
        required=False,
        label=_('Insert')
    )
    cup = forms.ModelChoiceField(
        Insert.objects.cups(),
        required=False,
        label=_('Cup')
    )

    class Meta:
        model = AnalysisParameters
        fields = (
            'cup',
            'insert',
            'technique',  # needed for relevant if, but it is hidden.
            'compute_volume',
            'geometric_volume',
            'active_area',
            'shape',
            'cube_edge_length',
            'cylinder_diameter',
            'cylinder_length',
            'rect_long_edge',
            'rect_short_edge',
            'rect_other_edge',
            'number_of_pieces',
            'measure_by',
            'chord_length',
            'cell_diameter',
            'cube_number_of_skins',
            'cyl_number_of_skins',
            'rect_number_of_large_skins',
            'rect_number_of_small_skins',
            'rect_number_of_other_skins',
        )

    def __init__(self, *args, **kwargs):
        "Update the cup/insert selections based on the instance."

        super().__init__(*args, **kwargs)
        if self.instance:
            if self.instance.cup:
                self.fields['cup'].queryset = _insert_choices_for(self.instance.cup)
            if self.instance.insert:
                self.fields['insert'].queryset = _insert_choices_for(self.instance.insert)

    def save(self, commit=True):
        "Do the special handling for contained objects, and for logging changes."
        params = super().save(commit=False)
        if commit:
            org = AnalysisParameters.objects.get(id=params.id)
            anl = AnalysisResult1350.objects.get(parameters=org)
            log_field_changes(anl, org, params, self.changed_data)
            params.cup = _update_insert(params, params.cup, org.cup)
            params.insert = _update_insert(params, params.insert, org.insert)
            params.save()
        return params


class ResultEditParamView(ReactFormMixin, View):  # pylint: disable=too-few-public-methods
    '''View to return data for start analysis screen'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define model, fields, and options for displaying form'''
        form = ResultEditParamForm
        hidden = {'technique': True}

    def id_from_request(self, request):
        return AnalysisResult1350.objects.get(id=request.GET.get('id', None)).parameters_id


def ui_results_edit_view(request):
    "The (multi-form) view that allows editing of post analysis parameters."
    try:
        AnalysisResult1350.objects.get(id=request.GET.get('id', None))
        anl_type = AnalysisResult1350
    except AnalysisResult1350.DoesNotExist:
        anl_type = AnalysisResult1345

    class ResultEditForm(forms.ModelForm):
        "The from to edit the post analysis results parameters."

        class Meta:
            model = anl_type
            fields = ('mass', 'name', 'operator')

        def save(self, commit=True):
            result = super().save(commit=False)
            if commit:
                org = anl_type.objects.get(id=result.id)
                log_field_changes(result, org, result, self.changed_data)
                result.save()
                update_results(result)
            return result

    class ResultEditView(ReactFormMixin, View):  # pylint: disable=too-few-public-methods
        '''View to return data for start analysis screen'''

        class ReactMeta:  # pylint: disable=too-few-public-methods
            '''Define model, fields, and options for displaying form'''
            form = ResultEditForm
            hidden = {
                'name': lambda: not Settings.objects.get().display_id,
            }

    result_view = ResultEditView()
    views = [result_view]
    if anl_type == AnalysisResult1350:
        views.append(ResultEditParamView())

    return multi_form_view(views, request)


def ui_review_done(request):
    '''Called when the user is done with the post analysis/calibration review.'''
    notify_analysis_seen()
    AnalysisResult.objects.filter(review_active=True).update(review_active=False)
    Calibration.objects.filter(review_active=True).update(review_active=False)
    OtherOperation.objects.filter(review_active=True).update(review_active=False)
    InsertMeasure.objects.filter(review_active=True).update(review_active=False)
    return JsonResponse({})


def add_method():
    '''Create new method'''
    method = AnalysisParameters.objects.default()
    method.pk = None
    method.is_default = False
    match = re.search(r"(.*)(\[\d+\])$", method.name)
    basename = match.group(1) if match else method.name
    seq = 1
    while True:
        name = f'{basename} [{seq}]'
        if AnalysisParameters.objects.methods().filter(name=name).count() == 0:
            break
        seq = seq + 1
    method.name = f'{basename} [{seq}]'
    method.is_shared = False
    method.save()
    return {'object_id': method.id}


def delete_method(object_id):
    '''Delete method with given id'''
    return AnalysisParameters.objects.get(id=object_id).delete()[1]


def default_method(object_id):
    '''Mark method as default'''
    try:
        AnalysisParameters.objects.filter(is_default=True).update(is_default=False)
        AnalysisParameters.objects.filter(id=object_id).update(is_default=True)
    except AnalysisParameters.DoesNotExist:
        return {'error': True}
    return {}


method_fields = (
    'name',
    'is_shared',
    'temperature',
    'insert',
    'cup',
    'fill_direction',
    'do_pre_purges',
    'num_purges',
    'use_vacuum_prep',
    'purge_fill_pressure',
    'num_cycles',
    'cycle_fill_pressure',
    'endby',
    'interval_time',
    'equilibration_rate',
    'use_run_precision',
    'run_precision_tolerance',
    'technique',
    'resin_density',
    'initial_pressure',
    'final_pressure',
    'pressure_increment',
    'low_pressure',
    'fracture_pressure',
    'compute_volume',
    'geometric_volume',
    'active_area',
    'shape',
    'cube_edge_length',
    'cylinder_diameter',
    'cylinder_length',
    'rect_long_edge',
    'rect_short_edge',
    'rect_other_edge',
    'number_of_pieces',
    'measure_by',
    'chord_length',
    'cell_diameter',
    'cube_number_of_skins',
    'cyl_number_of_skins',
    'rect_number_of_large_skins',
    'rect_number_of_small_skins',
    'rect_number_of_other_skins',
)


class MethodForm(forms.ModelForm):
    '''Custom form for validating analysis parameters'''
    insert = forms.ModelChoiceField(
        Insert.objects.inserts(),
        required=False,
        label=_('Insert')
    )
    cup = forms.ModelChoiceField(
        Insert.objects.cups(),
        required=False,
        label=_('Cup')
    )
    fill_direction = forms.ChoiceField(
        choices=FILL_CHOICES,
        widget=ButtonGroupWidget,
        label=_('Fill Direction')
    )
    endby = forms.ChoiceField(choices=END_CHOICES,
                              widget=ButtonGroupWidget,
                              label=_('Pressure Stabilization'))

    class Meta:
        model = AnalysisParameters
        fields = method_fields

    def clean_name(self):
        '''Don't allow duplicate names'''
        name = self.cleaned_data.get("name")
        if AnalysisParameters.objects.methods(name=name).exclude(pk=self.instance.id).count():
            raise ValidationError(
                _("A method named %(name)s already exists."),
                code="unique_method_name",
                params={'name': name},
            )
        return name


class ui_edit_method(ReactFormMixin, View):
    '''View for creating and editing methods'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Handle get requests to show input form'''
        form = MethodForm
        queryset = AnalysisParameters.objects.methods()
        get_selected = lambda: AnalysisParameters.objects.default().id
        select_object = default_method
        delete_object = delete_method
        add_object = add_method
        hidden = {
            'fill_direction': lambda: Instrument.objects.local().kind == '1',
            'temperature': lambda: not Instrument.objects.local().is_temp_control,
        }


def make_insert_view(kind):
    '''Return view for editing inserts, references, and cups'''
    def deactivate_insert(object_id):
        '''Mark insert as not active'''
        insert = Insert.objects.get(id=object_id)
        try:
            insert.delete()
            return {'object_id': object_id}
        except ProtectedError:
            return {
                "error": _("Cannot delete item. It is still referenced.")
            }

    def create_insert():
        '''Create new insert'''
        insert = Insert.objects.create_insert(kind)
        return {'object_id': insert.id}

    class InsertForm(forms.ModelForm):
        '''Custom form for editing inserts'''

        class Meta:
            model = Insert
            fields = (
                'kind',
                'name',
                'is_shared',
                'volume',
                'temperature',
                'material',
                'other_expansion_um_per_mk',
            )

        def clean_name(self):
            '''Don't allow duplicate names'''
            name = self.cleaned_data.get("name")
            if Insert.objects.methods(kind=kind, name=name).exclude(pk=self.instance.id).count():
                raise ValidationError(
                    _("An insert named %(name)s already exists."),
                    code="unique_insert_name",
                    params={'name': name},
                )
            return name

        def clean(self):
            '''Don't allow editing of inserts associated with an analysis'''
            super().clean()
            if AnalysisParameters.objects.filter(insert=self.instance).count():
                raise ValidationError(
                    _("Cannot modify multivolume insert with associated analysis."))
            if AnalysisParameters.objects.filter(cup=self.instance).count():
                raise ValidationError(_("Cannot modify cup with associated analysis."))
            if Calibration.objects.filter(cup=self.instance).count():
                raise ValidationError(_("Cannot modify cup with associated calibration."))
            if Calibration.objects.filter(ref_ball=self.instance).count():
                raise ValidationError(
                    _("Cannot modify volume standard with associated calibration."))

    class ui_edit_insert(ReactFormMixin, View):
        '''View for creating and editing inserts'''

        class ReactMeta:  # pylint: disable=too-few-public-methods
            '''handle requests to show the insert editor'''
            form = InsertForm
            queryset = Insert.objects.methods(kind=kind)
            delete_object = deactivate_insert
            add_object = create_insert
            hidden = {'kind': True}

        def get_input_form_props(self, object_id):
            props = super().get_input_form_props(object_id)
            obj = self.get_model_type().objects.get(id=object_id)
            props['disabled'] = {
                'temperature': obj.is_measured,
                'volume': obj.is_measured,
            }
            return props

    return ui_edit_insert.as_view()


class InsertMeasureForm(forms.ModelForm):
    '''Form for insert measurement'''

    ref_ball = forms.ModelChoiceField(
        Insert.objects.reference_balls(),
        initial=_get_default_ref,
        required=False,
        label=_("Volume Standard"),
    )
    multivolume_insert = forms.ModelChoiceField(
        Insert.objects.inserts(),
        required=False,
        label=_("Multivolume Insert"),
    )

    class Meta:
        model = InsertMeasure
        fields = ('kind', 'fill_pressure')

    field_order = ['kind', 'ref_ball', 'multivolume_insert', 'fill_pressure']


class ui_insert_measure(ReactFormMixin, View):  # pylint: disable=too-few-public-methods
    '''View to return data for start insert measurement screen'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Define model, fields, and options for displaying form'''
        form = InsertMeasureForm


class SettingsDisplayForm(forms.ModelForm):
    '''Form for display settings'''

    class Meta:
        model = Settings
        fields = (
            'display_id',
            'display_method',
            'display_operator',
            'display_wizard',
            'display_usb',
        )


class ui_settings_display(ReactFormMixin, View):
    '''View for editing the instrument serial number and kind.'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Handle get requests to show input form'''
        form = SettingsDisplayForm
        get_default_obj = lambda request: Settings.objects.get()


class ServiceInstrumentForm(forms.ModelForm):
    '''Form for editing instrument serial number and kind.'''

    passwd = forms.CharField(
        max_length=4,
        strip=True,
        required=False,
        initial="",
        label=_("Password"),
        widget=AlphanumericPasswordInput,
    )

    class Meta:
        model = Instrument
        fields = (
            'serial',
            'kind',
            'is_temp_control',
            'passwd',
            'material',
            'other_expansion_um_per_mk',
        )

    def clean(self):
        cleaned_data = super().clean()

        inst = Instrument.objects.local()
        needs_passwd = False
        # If the is_temp_control is changed we need the password.  This is true even if we turn it
        # off. That the user cannot accidentally turn it off and then need the password to turn it
        # back on.
        if bool(cleaned_data['is_temp_control']) != bool(inst.is_temp_control):
            needs_passwd = True
        # If the is_temp_control is turned on, and the serial number changed it needs a new
        # password.  That way you cannot memorize a serial number password to turn on temp control.
        if cleaned_data['is_temp_control'] and inst.serial != cleaned_data["serial"]:
            needs_passwd = True
        # If we need a password, and the user did not provide the right one, show an error.
        if needs_passwd:
            inst_passwd = gen_config_passwd(inst.model, cleaned_data['serial'])
            if cleaned_data['passwd'] != inst_passwd:
                self.add_error('passwd', _("You must enter a valid password."))
        return cleaned_data


class ui_service_instrument(ReactFormMixin, View):
    '''View for editing the instrument serial number and kind.'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Handle get requests to show input form'''
        form = ServiceInstrumentForm
        get_default_obj = lambda request: Instrument.objects.local()
        on_save = lambda data: initialize_analysis()


class ServiceFactoryResetForm(forms.ModelForm):
    '''Form for entering the password for reset to factory defaults. '''

    passwd = forms.CharField(
        max_length=4,
        strip=True,
        required=False,
        initial="",
        label=_("Password"),
        widget=AlphanumericPasswordInput,
    )

    class Meta:
        model = Instrument
        fields = ()

    def clean(self):
        cleaned_data = super().clean()

        inst = Instrument.objects.local()
        inst_passwd = gen_config_passwd(inst.model, inst.serial)
        if cleaned_data['passwd'] != inst_passwd:
            self.add_error('passwd', _("You must enter a valid password."))
        return cleaned_data


class ui_service_factory_reset(ReactFormMixin, View):
    '''View for entering the passowrd for reset to factory defaults.'''

    class ReactMeta:  # pylint: disable=too-few-public-methods
        '''Handle get requests to show input form'''
        form = ServiceFactoryResetForm
        get_default_obj = lambda request: Instrument.objects.local()


def ui_service_calibrations(request):  # pylint: disable=too-many-statements
    '''Return all calibrations'''
    # nothing in this function needs to be lazy, so this simplifies the c++ extension conversion.
    _ = gettext.gettext

    def fill_temp_cal_table(model, channel):
        page = model.AddPage(_("Block Temp") if channel == 'B' else _("Inlet Temp"))
        table = cpp.CReportStdTable(page, "", 4)
        table.SetTitle(0, _("Date"))
        table.SetTitle(1, _("Raw"))
        table.SetTitle(2, _("Cal"))
        table.SetTitle(3, _("Slope"))
        table.SetFormat(1, i1350.DPycTempResistance())
        table.SetFormat(2, i1350.DPycTempResistance())
        table.SetFormat(3, i1350.DPycTempCalSlope())
        for i in range(4):
            table.AlignCenter(i)
        for cal in RTDCalibration.objects.filter(channel=channel).order_by("-time"):
            table.AddItem(0, cpp.CTimeDate(cal.time).GetDate())
            table.AddItem(1, cal.raw_resistance)
            table.AddItem(2, cal.actual_resistance)
            table.AddItem(3, cal.slope)

    def fill_press_cal_table(model):
        page = model.AddPage(_("Pressure"))
        table = cpp.CReportStdTable(page, "", 4)
        table.SetTitle(0, _("Date"))
        table.SetTitle(1, _("Raw"))
        table.SetTitle(2, _("Cal"))
        table.SetTitle(3, _("Slope"))
        table.SetFormat(1, i1350.DPycPressureVoltage())
        table.SetFormat(2, pycnometry.DPycPressureAbs())
        table.SetFormat(3, i1350.DPycPressureCalSlope())
        for i in range(4):
            table.AlignCenter(i)
        for cal in PressureCalibration.objects.all().order_by("-time"):
            table.AddItem(0, cpp.CTimeDate(cal.time).GetDate())
            table.AddItem(1, cal.voltage)
            table.AddItem(2, cal.pressure)
            table.AddItem(3, cal.slope)

    def fill_last_vol_cal(model):
        cal = Calibration.objects.get_current()
        if cal is None:
            return

        data = i1350.SVolCalResult()
        data.sVolCalData = get_samplecaldata(cal)
        data.dTemperature = cal.collection_temperature
        data.szReference = str(cal.ref_ball)
        data.sHeader.szAppConfig = cal.instrument.full_config
        data.sHeader.szDescription = ""
        data.sHeader.analysisTime = cal.start_time
        i1350.AddLastVolCalResults(model, data)

    is_1cm3 = Instrument.objects.local().kind == '1'

    model = cpp.CReportModel(_("Calibration"))
    fill_temp_cal_table(model, 'B')
    if not is_1cm3:
        fill_temp_cal_table(model, 'I')
    fill_press_cal_table(model)
    _fill_vol_cal_table(
        model, is_1cm3, Calibration.objects.filter(state='C', kind='C').order_by("-start_time"),
        _("Volume")
    )
    fill_last_vol_cal(model)

    return report_show_export_print_view(request, model, _("all-calibrations"))


def ui_sample_results(request):
    '''Return all results'''

    def get_analyses():
        return AnalysisResult.objects.exclude(results=None)

    def format_instrument(anl):
        inst = anl.instrument
        return f"AccuPyc {inst.model} {inst.get_kind_display()} SN: {inst.serial}"

    restrictions = {'local_only': lambda qs: qs.filter(instrument=Instrument.objects.local())}

    return sample_results_view(request, get_analyses, format_instrument, restrictions)


class ui_sort_filter_view(ReactSortFilterMixin, View):
    '''The view that displays the result sorts and filters form'''

    class ReactResultsMeta:  # pylint: disable=too-few-public-methods
        '''The class that defines the available result sorts and filters'''

        model = AnalysisResult1350
        sorts = ['start_time', 'results.absolute_density']
        filters = ['name', 'operator', 'results.absolute_density', 'local_only', 'start_time']
        field_options = {
            'formats': {
                'absolute_density': pycnometry.DPycDensity
            },
            'titles': {
                'absolute_density': _("Density"),
                'local_only': _("Local Only"),
                'start_time': _("Date"),
            },
            'hidden': {
                'name': lambda: not Settings.objects.get().display_id,
                'operator': lambda: not Settings.objects.get().display_operator
            }
        }


def hwcal_results_view(request, which):
    """The view that returns the json report for the calibration results."""
    response = hwcal_results(which=which)
    return JsonResponse(response)


def get_settings_1350_view(request):
    """The view that returns the the json ojbject for all the 1350 specific settings."""
    obj = Settings.objects.get()
    ret = {
        "displayId": obj.display_id,
        "displayMethod": obj.display_method,
        "displayOperator": obj.display_operator,
        "displayWizard": obj.display_wizard,
    }
    return JsonResponse(ret)


def settings_general_view(request):
    "View for editing general settings"
    local = Instrument.objects.local()
    needs_purge_check = local.is_temp_control and i1350.UDPGetHWState().fPressureSwitchPresent

    class SettingsGeneralForm(forms.ModelForm):
        "Custom form to specify language and formatting choices"
        base_language = forms.ChoiceField(
            choices=lang_choices(),
            label="Language/" + str(_("Language")) if "Language" != str(_("Language")) else _("Language")
        )

        class Meta:
            model = Settings
            fields = (
                (
                    'do_idle_purges', 'days_idle_purges', 'flow_required_temperature',
                    'base_language',
                )
                if needs_purge_check else
                ('do_idle_purges', 'days_idle_purges', 'base_language', )
            )


    class SettingsGeneralView(ReactFormMixin, View):
        '''View for editing the general settings.'''

        class ReactMeta:  # pylint: disable=too-few-public-methods
            '''Handle get requests to show input form'''
            form = SettingsGeneralForm
            relevant_if = dict({
                'days_idle_purges': ['==', 'do_idle_purges', True],
            }, **region_relevant_if())

        def get(self, request):
            '''Return the information for editing the local settings.  If there is no
            object id, redirect to the local one.'''
            if 'id' not in request.GET:
                obj, __ = Settings.objects.get_or_create()
                return redirect(f'{request.path}?id={obj.id}')

            settings = Settings.objects.get(id=request.GET.get('id', None))
            props = self.get_input_form_props(request.GET.get('id', None))
            add_region_props(props, settings)
            add_unit_selections(props)

            return JsonResponse(props, encoder=MicJSONEncoder)

        def post(self, request):
            '''Update the instrument information based on what the user edited.'''
            general_form = forms.modelform_factory(
                Settings,
                form=SettingsGeneralForm,
            )
            post = json.loads(request.body.decode('utf-8'))
            data = post.get('data')
            form = general_form(data, instance=Settings.objects.get())
            if form.is_valid() and post.get('action') == 'save':
                general = form.save(commit=False)
                general.region = pull_region_setting(data)
                general.save()
                update_language(general)
                update_unit_selections(data)
                send_idle_purge_info()
                send_purge_flow_info()

            return post_response(form)

    return SettingsGeneralView.as_view()(request)


def app_export_results(location, is_service: bool):

    # If this is a service (post reset to defaults) export, don't add anything.
    if is_service:
        return

    # Otherwise export the calibration and sample results.
    local = Instrument.objects.local()

    # Export the last 25 results.
    for res in AnalysisResult1350.objects.filter(
            instrument=local, state="C").order_by("-start_time")[:25]:
        model = cpp.CReportModel("")
        accu_result = accu_or_foam_results_from_anl(res)
        i1350.AddAccuPycResultsReport(model, accu_result)
        name = f'record-{local.serial}-{res.start_time.strftime("%Y%m%d%H%M")}'
        cpp.ExportReportModelPDF(model, location, f'{name}.pdf')
        cpp.ExportReportModelXLSX(model, location, f'{name}.xlsx')
        model = cpp.CReportModel("")
        i1350.AddAccuPycRaw(model, accu_result)
        cpp.ExportReportModelXLSX(model, location, f'{name}-raw.xlsx')

    # Export the last 25 calibrations.
    for cal in Calibration.objects.filter(
            instrument=local, state="C", kind__in=["C", "UA"]).order_by("-start_time")[:25]:
        model = cpp.CReportModel("")
        i1350.AddVolCalResults(model, _vol_cal_or_check_results_from_cal(cal))
        name = f'cal-{local.serial}-{cal.start_time.strftime("%Y%m%d%H%M")}'
        cpp.ExportReportModelPDF(model, location, f'{name}.pdf')
        cpp.ExportReportModelXLSX(model, location, f'{name}.xlsx')


# =================================================================================================
def _base_export_name(anl):

    def remove_illegal(name):
        illegal = ["/", "<", ">", ":", "\"", "\\", "|", "?", "*"]
        for rem in illegal:
            name = name.replace(rem, "")
        return name

    date = anl.start_time
    model = anl.instrument.model
    serial = anl.instrument.serial
    name = anl.name
    filename = f'{date.strftime("%Y%m%d%H%M")}-{model}-{serial}'
    if name:
        filename = filename + remove_illegal(name)
    return filename


def _fill_vol_cal_table(model, is_1cm3, cal_list, page_name):
    _ = gettext.gettext

    def val_or_zero(value):
        return value if value is not None else 0.0

    page = model.AddPage(page_name)
    num_cols = 10 if not is_1cm3 else 8
    table = cpp.CReportStdTable(page, "", num_cols)
    table.SetTitle(0, _("Date\b"))
    if not is_1cm3:
        table.SetTitle(1, _("Sample Chamber-\bSample Filled"))
        table.SetTitle(2, _("Reference Chamber-\bSample Filled"))
        table.SetFormat(1, pycnometry.DPycVolume())
        table.SetFormat(2, pycnometry.DPycVolume())
    table.SetTitle(num_cols - 7, _("Sample Chamber-\bReference Filled"))
    table.SetTitle(num_cols - 6, _("Reference Chamber-\bReference Filled"))
    table.SetTitle(num_cols - 5, _("Result\b"))
    table.SetTitle(num_cols - 4, _("Pressure\b"))
    table.SetTitle(num_cols - 3, _("Temp\b"))
    table.SetTitle(num_cols - 2, _("Reference Object\b"))
    table.SetTitle(num_cols - 1, _("Gas\b"))
    table.SetFormat(num_cols - 4, pycnometry.DPycPressureGauge())
    table.SetFormat(num_cols - 3, pycnometry.DPycTemperature())
    table.SetFormat(num_cols - 7, pycnometry.DPycVolume())
    table.SetFormat(num_cols - 6, pycnometry.DPycVolume())
    for i in range(num_cols):
        table.AlignCenter(i)
    for cal in cal_list:
        table.AddItem(0, cpp.CTimeDate(cal.start_time).GetDate())
        if not is_1cm3:
            table.AddItem(1, val_or_zero(cal.volume_fs_sample))
            table.AddItem(2, val_or_zero(cal.volume_fs_reference))
            table.AddItem(3, val_or_zero(cal.volume_fr_sample))
            table.AddItem(4, val_or_zero(cal.volume_fr_reference))
        else:
            table.AddItem(1, val_or_zero(cal.volume_fr_sample))
            table.AddItem(2, val_or_zero(cal.volume_fr_reference))
        table.AddItem(num_cols - 5, cal.get_kind_display())
        table.AddItem(num_cols - 4, cal.fill_pressure)
        table.AddItem(num_cols - 3, cal.collection_temperature)
        table.AddItem(num_cols - 2, cal.ref_ball.volume)
        table.AddItem(num_cols - 1, cal.gas)


def _fill_vol_cal_raw(model, is_1cm3, cal_list):
    _ = gettext.gettext

    for cal in cal_list:
        _fill_single_vol_cal_raw(model.AddPage(cal.start_time.strftime("%Y-%m-%d %H %M")), is_1cm3, cal)

def _fill_single_vol_cal_raw(page, is_1cm3, cal):
    _ = gettext.gettext
    def val_or_zero(value):
        return value if value is not None else 0.0

    def write_cycles(table, exp, fill):
        for i in range(exp.num):
            table.AddItem(0, exp.p0s[i])
            table.AddItem(1, exp.p1s[i])
            table.AddItem(2, exp.p2s[i])
            table.AddItem(3, exp.temp_anl_block[i])
            table.AddItem(4, exp.temp_valve_block[i])
            table.AddItem(5, fill)

    table = cpp.CReportOptionsTable(page, _("Summary"))
    table.AddItem(_("Date"), cpp.CTimeDate(cal.start_time).GetDate())
    if not is_1cm3:
        table.AddItem(
            _("Sample Chamber Sample Filled"),
            pycnometry.DPycVolume(val_or_zero(cal.volume_fs_sample)).ValueAndUnits())
        table.AddItem(
            _("Reference Chamber Sample Filled"),
            pycnometry.DPycVolume(val_or_zero(cal.volume_fs_reference)).ValueAndUnits())
        table.AddItem(
            _("Sample Chamber Reference Filled"),
            pycnometry.DPycVolume(val_or_zero(cal.volume_fr_sample)).ValueAndUnits())
        table.AddItem(
            _("Reference Chamber Reference Filled"),
            pycnometry.DPycVolume(val_or_zero(cal.volume_fr_reference)).ValueAndUnits())
        table.AddItem(
            _("Result"),
            cal.get_kind_display())
        table.AddItem(
            _("Pressure"),
            pycnometry.DPycPressureGauge(cal.fill_pressure).ValueAndUnits())
        table.AddItem(
            _("Temp"),
            pycnometry.DPycTemperature(cal.collection_temperature).ValueAndUnits())
        table.AddItem(
            _("Reference Object"),
            pycnometry.DPycVolume(cal.ref_ball.volume).ValueAndUnits())
        table.AddItem(
            _("Gas"),
            cal.gas)

    table = cpp.CReportStdTable(page, _("Expansions"), 6)
    table.SetTitle(0, _("P0"))
    table.SetTitle(1, _("P1"))
    table.SetTitle(2, _("P2"))
    table.SetTitle(3, _("T"))
    table.SetTitle(4, _("Tv"))
    table.SetTitle(5, _("Fill"))
    table.SetFormat(0, pycnometry.DPycPressureGauge())
    table.SetFormat(1, pycnometry.DPycPressureGauge())
    table.SetFormat(2, pycnometry.DPycPressureGauge())
    table.SetFormat(3, pycnometry.DPycTemperature())
    table.SetFormat(4, pycnometry.DPycTemperature())
    for i in range(6):
        table.AlignCenter(i)
    if cal.expansion_fs_empty:
        write_cycles(table, cal.expansion_fs_empty, 'SE')
    if cal.expansion_fr_empty:
        write_cycles(table, cal.expansion_fr_empty, 'RE')
    if cal.expansion_fs_ref:
        write_cycles(table, cal.expansion_fs_ref, 'SR')
    if cal.expansion_fr_ref:
        write_cycles(table, cal.expansion_fr_ref, 'RR')


def _fill_vol_check_table(model, is_1cm3, cal_list, page_name):
    _ = gettext.gettext

    def val_or_zero(value):
        return value if value is not None else 0.0

    page = model.AddPage(page_name)
    num_cols = 8 if not is_1cm3 else 7
    table = cpp.CReportStdTable(page, "", num_cols)
    table.SetTitle(0, _("Date"))
    table.SetTitle(1, _("Pressure"))
    table.SetFormat(1, pycnometry.DPycPressureGauge())
    table.SetTitle(2, _("Temp"))
    table.SetFormat(2, pycnometry.DPycTemperature())
    table.SetTitle(3, _("Gas"))
    table.SetTitle(4, _("Reference"))
    table.SetFormat(4, pycnometry.DPycVolume())
    if not is_1cm3:
        table.SetTitle(5, _("Sample Filled"))
        table.SetFormat(5, pycnometry.DPycVolume())
    table.SetTitle(num_cols - 2, _("Reference Filled"))
    table.SetFormat(num_cols - 2, pycnometry.DPycVolume())
    table.SetTitle(num_cols - 1, _("Pass/Fail"))
    for i in range(num_cols):
        table.AlignCenter(i)
    for cal in cal_list:
        results = _vol_check_results(cal)
        table.AddItem(0, cpp.CTimeDate(cal.start_time).GetDate())
        table.AddItem(1, cal.fill_pressure)
        table.AddItem(2, cal.collection_temperature)
        table.AddItem(3, cal.gas)
        table.AddItem(4, results.dRefVol)
        if not is_1cm3:
            table.AddItem(5, val_or_zero(results.dFSVolume))
        table.AddItem(num_cols - 2, results.dFRVolume)
        table.AddItem(num_cols - 1, _("Passed") if results.fPassed else _("Failed"))


def _insert_choices_for(ins):
    """
    The right choices for an insert are all the methods that do not match the insert along with
    the insert itself.

    :param ins: the non method insert that is being edited.
    :return: the query object that selects the method choices that are available for ``ins``.
    """
    return Insert.objects.filter(Q(is_method=True, kind=ins.kind) | Q(id=ins.id)).exclude(
        is_method=True,
        kind=ins.kind,
        volume=ins.volume,
        temperature=ins.temperature,
        material=ins.material,
        other_expansion_um_per_mk=ins.other_expansion_um_per_mk,
    )


def _update_insert(obj: AnalysisResult1350, new: Insert, org: Insert) -> Insert:
    """
    This function replaces the ``org`` insert of cup with ``new`` inside the ``obj``.

    This handles the cases:
    - both are None:  Just return none there is nothing to be done.
    - new is none and org is not:  Then the origional has to be deleted so it does not clutter
      the table.
    - new is not and org is none:  Then the new should be a non method copy of org.
    - neither is none:  Then new is updated to have the same content, but be a non method.

    :param obj: The object in which ``new`` is a member.
    :param new: The newly selected insert/cup.
    :param org: The origional insert/cup in the ``obj``
    :return: The update member of ``obj``
    """
    if not new and not org:
        return new
    if new and org:
        new.id = org.id
        new.is_method = False
        new.last_updated = None
        new.save()
    if new and not org:
        new = copy_model_obj(Insert, new)
        new.save()
    if not new and org:
        obj.save()
        org.delete()
    return new

def _vol_cal_or_check_header(cal):
    header = i1350.SResultHeader()
    header.szModel = cal.instrument.model
    header.szSerial = cal.instrument.serial
    header.szAppConfig = cal.instrument.full_config
    header.szDescription = ""
    header.analysisTime = cal.start_time
    return header

def _vol_cal_or_check_results_from_cal(cal):
    '''Create and fill in the appropriate results structure from the calibration.'''
    if cal.kind == 'H':
        data = i1350.SVolCheckCalResult()
        data.sChamberCal = get_chambercal(cal.check_of)
    else:
        data = i1350.SVolCalResult()
    data.sVolCalData = get_samplecaldata(cal)
    data.dTemperature = cal.collection_temperature
    data.szReference = str(cal.ref_ball)
    data.szGas = cal.gas
    data.sHeader = _vol_cal_or_check_header(cal)

    return data

def _vol_check_results(cal):
    "Return the volume cal check results for the cal object."
    data = _vol_cal_or_check_results_from_cal(cal)
    return i1350.CalcCalCheckSummaryData(data)

def _is_cal_in_range(cal: Calibration, previous: Calibration, error: float):
    "Return true iff the cal is within error of previous"
    def is_error_small(v1, v2, error):
        print("is_error_small:", v1, v2, abs(v1 - v2), error * abs(v1))
        return abs(v1 - v2) < error * abs(v1)

    is_ok = True
    if cal.expansion_fs_empty and cal.expansion_fs_ref:
        smp = is_error_small(previous.volume_fs_sample, cal.volume_fs_sample, error)
        ref = is_error_small(previous.volume_fs_reference, cal.volume_fs_reference, error)
        is_ok = is_ok and smp and ref
    if cal.expansion_fr_empty and cal.expansion_fr_ref:
        smp = is_error_small(previous.volume_fr_sample, cal.volume_fr_sample, error)
        ref = is_error_small(previous.volume_fr_reference, cal.volume_fr_reference, error)
        is_ok = is_ok and smp and ref
    return is_ok

def _user_cal_results(cal: Calibration, previous: Calibration, factory: Calibration):
    "Return the report model that is the result of this user calibration."
    # nothing in this function needs to be lazy, so this simplifies the c++ extension conversion.
    _ = gettext.gettext

    def add_row(table, name, cal):
        table.AddItem(0, name)
        if cal.expansion_fs_empty and cal.expansion_fs_ref:
            table.AddItem(1, cal.volume_fs_sample)
            table.AddItem(2, cal.volume_fs_reference)
        if cal.expansion_fr_empty and cal.expansion_fr_ref:
            table.AddItem(3, cal.volume_fr_sample)
            table.AddItem(4, cal.volume_fr_reference)

    model = cpp.CReportModel("results")
    page = model.AddPage("results")
    table = cpp.CReportStdTable(page, "", 5)
    table.SetTitle(0, "")
    table.SetTitle(1, _("Sample Chamber-\bSample Filled"))
    table.SetTitle(2, _("Reference Chamber-\bSample Filled"))
    table.SetTitle(3, _("Sample Chamber-\bReference Filled"))
    table.SetTitle(4, _("Reference Chamber-\bReference Filled"))
    table.SetFormat(1, pycnometry.DPycVolume())
    table.SetFormat(2, pycnometry.DPycVolume())
    table.SetFormat(3, pycnometry.DPycVolume())
    table.SetFormat(4, pycnometry.DPycVolume())
    for i in range(5):
        table.AlignCenter(i)
    add_row(table, _("Current"), cal)
    if previous:
        add_row(table, _("Previous"), previous)
    if factory:
        add_row(table, _("Factory"), factory)
    return model
